(* :Authors: J.William Helton, Orlando Merino, and Trent Walker
   :Programming:  T.Walker and  Julia Myers *)

(* :Context: Anopt` *)

(* :Package Version: 4.0 *)

(* :Copyright (c) 1991-2001 J.W. Helton and O. Merino
    All rights reserved.  Usual disclaimer applies.
*)

(* :History: 
   Version 4.0 June 1998
   Version 3.1 May 1994
   Version 3.0 January 1994
   Version 2.2 July 1993
   Version 2.1 March 1993
   Version 2.0 by Merino, Walker and Myers August 1991
   Version 1.2 package details by Myers May 1991
   Version 1.1 by T. Walker, Feb. 20, 1991
*)

(* :Keywords: *)

(* :Source: *)

(* :Warning: *)
    If[$VersionNumber<2,
    Print["Warning: Your copy of Mathematica has $VersionNumber < 2.0 \n
    		THE PACKAGE Anopt REQUIRES $ReleaseNumber >= 2.0"]];

   Unprotect[Plus];
   x_ + 0. I := x;
   Protect[Plus];


(* :Requirement:  ComplexRules.m *)

(* :Limitation: *)
   
(* :Discussion: 
   -  Gamma such as Abs[(e+2)/(e+4) - z[1]]^2 will not work
      This could be tested for with some additional programming
*)

(* :Table of Contents: location of functions in this package 
   Anopt
   PreSmoothing
   Smoothing
   SmoothRoutine
   SmoothVector
   SmoothTime
   Project
   LineSearch 
   GridDoubling
   DoubleTime
   Hankel
   MinCheck2d
   CheckTheData
   WriteIterationID
   DiagnosticSave
   DualWrite
   DualWrite2
   WSpace
   AppendVectSave
   aform
   NumFormat
   DefineMessages
   BeginOutputFiles
   FinalSummary
   EndOutputFiles
   WriteScreenHeading
   WriteOptions
   WriteHeadings
*)
(***********************************************************************)







(****************** Package Preliminaries ************************)  

BeginPackage[ "Anopt`", 
	     {"AnoptUtilities`", 
	      "DualExtremal`",
	      "DiskIteration`",
              "Graphics`ParametricPlot3D`",
	      "ComplexRules`", 
	      "CF`", 
	      "Utilities`FilterOptions`" }]





(*************************************************************************
        Usage Notes - given to global variables and functions 
*************************************************************************)

(*. . . Anopt input variables . . *)
$AnoptVersion::usage="$AnoptVersion"

Anopt::usage = " \n
   Anopt[gamma, AnoptTol], Anopt[gamma, AnoptTol, InitialGuess, Options] \n
   minimizes over functions f analytic on the unit disk in the complex plane \n
   a performance function gamma, which is an expression in e and z[1] \n
   (or z[1], z[2],...).  Example: gamma = Abs[1/(e+0.5) + z[1]]^2 \n
   gamma should be positive valued and differentiable in z. \n
   AnoptTol is a positive real number, a tolerance for stopping calculations.\n
   If not specified, InitialGuess is defaulted as the function zero, \n
   discretized as Table[0.,{ndim},{32}], where ndim given by gamma \n
   as a function of z[1],...,z[ndim].";
e::usage = " \n
   e is used to represent Exp[I theta] in the formula of performance functions gamma.  "
z::usage = "\n
   variable in expression of performance functions gamma, which depend on both e and on z[1] (scalar case) and z[1],z[2],... (vector case).  \n
   Ex:  gamma = Abs[1/(e - 0.5) - z[1]]^2 "
a::usage = "\n
   variable in expression of performance functions gamma.  It is optional, and it represents a parameter which is input to Anopt[] using the option AMatrix.\n   Ex:  gamma = a[1] Abs[1/(e - 0.5) - z[1]]^2 + a[2] Abs[z[1]]^2 "

(*. . . Anopt output variables . . .*)
Solution::usage = " \n
   Solution is the (discrete) analytic function produced by \n
   Anopt[] which solves approx. the optimization problem. \n
   It is a list of lists with the the format \n
   Solution={{x(1,1),x(1,2),...x(1,m)},..., {x(n,1),x(n,2),...x(n,m)}} \n
   where x(i,j) is a complex number.  The i-th entry of Solution is \n
   called by Solution[[j]]."
gammastar::usage = " \n
   Output to Anopt.  A positive number, gammastar is optimal value calculated by Anopt[].  Symbolically, gammastar = Sup_{e} gamma[e,f(e)], where f(e) is the optimizer.";
flat::usage = " \n
   Output to Anopt.  A positive number, flat is the (normalized) optimality test that measures flatness of performance at a function f.  It is calculated as (Sup[gamma[e,f(e)] - In[gamma[e,f(e)])/Sup[gamma[e,f(e)]. At optimum, flat=0."
grAlign::usage = " \n
   Output to Anopt.  A positive number, grAlign is the test for the optimality condition on the complex gradient of performance at a function f.  Typically relevant only in the vector valued case.  At optimum, grAlign=0."
ned::usage = " \n
   Output to Anopt.  A positive number, ned is called the  Numerical Error Diagnostic, and it measures numerical noise arising in the calculations in a run of Anopt[].  The accuracy of results of the run is typically of the same order of magnitude than ned."

(*. . . Anopt Options *)
GridPoints::usage="Option of Anopt.  An integer, usually a power of 2, it
specifies the number of points on the unit circle for sampling
functions.  This option is disregarded if an initial guess to the
solution is provided by the user when calling Anopt.";  
QuickResults::usage = " \n
   Option of Anopt.  True/False, it specifies a quick calculation \n 
   for saving run time at the expense of precision.";
DebugMe::usage=" \n
   Option of Anopt for debugging. Puts out a location.   Controlled by Message Debugger. True/False. "
OptimalityTol::usage = " \n
   Option of Anopt.  A positive number, it specifies the tolerance for stopping calculations.  Anopt[] stops if Max[flat,grAlign]<optimalitytol.";
AMatrix::usage = " \n
   AMatrix is an Option of Anopt[]. It is a matrix, each row is a list of values representing the sampled values of a fixed function that is present in the formula of the performance function.";
Iterations::usage = " \n
   Iterations is an Option of Anopt[]. A positive integer, it specifies the maximum number of iterations.";
GridDouble::usage = " \n
   GridDouble is an Option of Anopt[]. True/False, it specifies if grid doubling is allowed or not when Anopt[] runs.";
MaxGridSize::usage = " \n
   MaxGridSize is an Option of Anopt[]. A power or 2 such as 256 or 1024, it specifies a cap on the maximum size of grid allowed in Anopt[] runs.";
Diagnostics::usage = " \n
   Diagnostics is an Option of Anopt[]. Possible values are -1,0,1,2.\n
   = -1   No output is printed on the screen or in a file Screen.out. \n
   =0     Some output is printed on the screen. \n
   =1     Some output is printed on the screen and in a file Screen.out. \n
   =2     Some output is printed in a file Screen.out. \n
See also the option Diagnostics2 ";
Diagnostics2::usage = " \n
   Diagnostics 2 is an Option of Anopt[]. Possible values are 0,1,2,3,4,5.\n
   =0 no files produced \n
   =1 Summary1 only \n
   =2 Summary1 and Summary2 \n
   =3 Summary1 and Summary2 and Summary3 \n
   =4 Summary1 and Summary2 and Summary3 and Summary4 \n
   =5 Summary1 and Summary2 and Summary3 and Summary4 and *.Save files.";
RunID::usage = " \n
   RunID is an Option of Anopt[].  String used to identify a run of Anopt. \n
   It can be specified as None or as any string.";
Weight::usage = " \n
   Weight is an Option of Anopt[]. Its default is {}.";
OT3Check::usage = " \n
   OT3Check is an Option of Anopt[].  True/False, it applies only in the 2-dim case.  When OT3Check->True is specified and if flat=grAlign=approx=0 at a function f, the program checks if a condition (OT3) sufficient for optimality is satisfied.  ";
DoLineSearch::usage = " \n
   DoLineSearch is an Option of Anopt[]. Specifies whether a linesearch in t>0 has to be performed at each iteration to find an update f' = f + t *h, where h is produced by the subproblem solver.  The linesearch is on the expression Sup_{e} gamma(e,f(e) + t h(e)).";
Subproblem::usage = "\n
   Options for Anopt.  A string, it specifies the method used for solving the subproblem. Example: Subproblem->DiskIteration";

(*. . . Anopt Smoothing Options . . .*)
Smoothing::usage="\n
Option of Anopt[].  It specifies whether smoothing of functions is performed, and the amount of smoothing.  It can be set to Automatic, None, or a list {n1,n2} where  n1 < n2 are integers between 0 and 90.";
InitialSmooth::usage = " \n
   Option of Anopt[].  This causes the initial guess to be smoothed.  It can be set to Automatic or to None.";
UpdateVectorSmooth::usage = " \n
   Option of Anopt[].  This causes the update vector to be smoothed according to how fast the Fourier coeffs. go to 0.  It can be set to Automatic, None, or a list {n1,n2} where  n1 < n2 are integers between 0 and 90.";
SmoothType::usage = " \n
   SmoothType is an Option of Anopt[]. It specifies the type of smoothing \n
   to be done.  Possible values are A,B,C,..etc.";
LineSearchFunction::usage="Option of Anopt[].";

(*. . . Messages . . .*)
$MessagePrePrint = Identity;
Debugger::location::usage = " \n
    Message for DebugMe option.  Function COMMENT[] controls it ";
CheckCode::zerodivide::usage = " \n
    Message to check for division by zero";
CheckCode::withintolerance::usage = " \n
    Message to check if within a tolerance";
CheckData::length::usage= " ";
CheckData::contradiction::usage=" ";
CheckData::gridsize::usage= " "; 
CheckData::dimension::usage=" "; 
CheckData::invalid::usage=" "; 
CheckData::warning::usage=" "; 

(* Options that can be filtered to other functions
WilsonSFSave::
    WilsonSFSave is an option of AnoptUtilities`WilsonSF[]. 
    Its default in Anopt[] is False.  
WilsonSFOutputChannel::
    Is the output channel for AnoptUtilities`WilsonSF[] output to be directed to
    Its default in Anopt is WilsonSFSave.out.
HarrisSFSave:: 
    is an option of AnoptUtilities`HarrisSF[]. 
    Its default in Anopt[] is False.  
HarrisSFOutputChannel::
    is an option of AnoptUtilities`HarrisSF[]. 
    Its default in Anopt is HarrisSFSave.out.
HarrisSFInitial::
    is an option of AnoptUtilities`HarrisSF[]. 
    It has no default in Anopt.
HarrisMaxIts::
    is an option of AnoptUtilities`HarrisSF[]. 
    It has no default in Anopt.
HarrisSFToler::
    is an option of AnoptUtilities`HarrisSF[]. 
    It has no default in Anopt.
DualExtremalSave::
    is an option of DualExtremal`DualExtremal[].
    Its default in Anopt is False
DualTol::
    is an option of DualExtremal`DualExtremal[].
    Its default in Anopt depends on.... (see before DE call) 
DualMaxIts::
    is an option of DualExtremal`DualExtremal[].
    Its default in Anopt is dualmaxits.
DualOutputChannel::
    is an option of DualExtremal`DualExtremal[].
    Its default in Anopt is DualExtremalSave.out.
DualCircleGrid::
    is an option of DualExtremal`DualExtremal[].
    Its default in Anopt is eLocal.
*)

(*. . . Anopt Utilities . . .*)
ShiftBlockHankel::usage = " \n "
BlockHankel::usage = " \n "
FastHankel::usage = " \n "
FastToeplitz::usage = " \n "
Toeplitz::usage = " \n "
ShiftBlockToeplitz::usage = " \n
   Creates a Toeplitz with lower left corner being zero, upper right being \n
   -n coef.  Code is built for SPEED on large grids.  Be sure that gridsize \n
   is always even.";
BlockToeplitz::usage = " \n
   Creates a Toeplitz with zeroth coef down the diagonal, negative coef's \n
   in the lower triangle, and positive coef's in upper triangle.  System \n
   is built for speed and calls ShiftBlockToeplitz.  "
MatrixMerge::usage = " \n "



(************************************************************************
                  Option Defaults
*************************************************************************)
Options[Anopt] = 
   {
	ConstrainedSearch -> None,
   	QuickResults -> False,
   	OptimalityTol -> Automatic,
   	Subproblem -> "DiskIteration",
   	DebugMe ->False,
        AMatrix->None,
        Iterations->10,
        GridDouble->True,
        MaxGridSize->1024,
        Diagnostics->0,
        Diagnostics2->0,
        RunID->"none",
        Weight -> 1,
	DualEigenfunction->Automatic,
	OT3Check->True,
        DoLineSearch ->True,
(*. . . Smoothing . . .*)	
	Smoothing->Automatic,
	InitialSmooth->None,	
	UpdateVectorSmooth->None,
	SmoothType->"A",
	LineSearchFunction -> GoldenLineSearch,
	GridPoints->32
   }

$AnoptVersion = "Anopt 4.0 (June 1998)";

Begin["Anopt`Private`"]


Print[" \n
Anopt 4.0 (c) Copyright 1991-2001 \n
J.W. Helton and O. Merino.  All rights reserved."];

(****************** end Package Preliminaries ***********************)




(*********************************************************************
                         Anopt
*********************************************************************)

(****** This Anopt creates a default value for the Initial Guess ******)
Anopt[ gam_, AnoptTol_?(#>0&), opts___Rule]:=
   Block[{ndim,gridpoints},
	gridpoints = GridPoints /. {opts} /. Options[Anopt];
        ndim = 0;
        While[ !TrueQ[ComplexD[gam, z[ndim+1]] == 0 ], ++ndim];
        Anopt[gam, AnoptTol,Table[0.,{ndim},{gridpoints}], opts]
   ]; (* end Block *)


(* add linesearch *)
(****** This Anopt allows the user to input the InitialGuess ******)
Anopt[ gam_, AnoptTol_?(#>0&),InitialGuess_, opts___Rule ]:=
   Block[{ ExactSolve,z,a, 
        zvalues, eLocal, avalues,   (* implicit in gamma*)
        i, j, counter, ndim, nsp, 
        theta,iternum,g,Z,gammaz,gammazz,gammazzbar,zero,SmthFac,
        pot1, pot2, poptpartial,  (* prior values *)
        temp,foucoef, adim,asp,aflag,phnorm,hnorm,ForceIter,
	x,tempVector,sBlckHankel,sBlckToep,s,sBlckToepTran, BlckToepTrans,
        sbhMatrix,sbtMatrix, windreal, UpdateV, posdef, pg,
(*. . . Variables used to write files . . .*)
        Summary1, Summary2, Summary3, Summary4, SaveDerivative, SaveGamma,
        SaveFunctVal, SaveHessian, SaveSpectralFactor, SaveSubproblem,
        SaveSubproblemSolution, SaveUpdateVector, SaveSolution,
(*. . . Option local variables . . .*)
        A, niter, dGrid, mxGrid, diag, diag2, runID, weight,  
        dualEigenfunction, doLineSearch,
	ot3Check, initialSmooth,
	progressTolerance, 
	updateVectorSmooth, smoothParameters, smoothType, 
(*. . . Other stuff . . .*)
        dualExtremalSave, updateVectorSave, solutionSave, projDiffMatrixSave, 
	derivativeSave,  functValSave, gammaSave, hessianSave,
	spectralFactorSave, subproblemSave, subproblemSolutionSave,
        time, infgamma , supgamma, datetime, zeromatrix, zeroGrad },
	
(*. . . Local equivalents to external options . . . *)
        amatrix          = AMatrix     /. {opts} /. Options[Anopt];
        niter      = Iterations  /. {opts} /. Options[Anopt];
        dGrid      = GridDouble  /. {opts} /. Options[Anopt];
        mxGrid     = MaxGridSize /. {opts} /. Options[Anopt];
        diag       = Diagnostics /. {opts} /. Options[Anopt];
        diag2      = Diagnostics2/. {opts} /. Options[Anopt];
        runID      = RunID       /. {opts} /. Options[Anopt];
	subproblem = Subproblem  /. {opts} /. Options[Anopt];
	debug      = DebugMe     /. {opts} /. Options[Anopt];
	quickResults       = QuickResults       /. {opts} /. Options[Anopt];
        optimalitytol      = OptimalityTol      /. {opts} /. Options[Anopt];
        weight             = Weight             /. {opts} /. Options[Anopt];
	ot3Check           = OT3Check           /. {opts} /. Options[Anopt];
	doLineSearch       = DoLineSearch       /. {opts} /. Options[Anopt];
	dualEigenfunction  = DualEigenfunction  /. {opts} /. Options[Anopt];
	smoothing	   = Smoothing          /. {opts} /. Options[Anopt];
	initialSmooth      = InitialSmooth      /. {opts} /. Options[Anopt];
	updateVectorSmooth = UpdateVectorSmooth /. {opts} /. Options[Anopt];
	smoothType         = SmoothType         /. {opts} /. Options[Anopt];
	searchfunction	   = LineSearchFunction /. {opts} /. Options[Anopt];
	constrainedsearch  = ConstrainedSearch  /. {opts} /. Options[Anopt];
	gridpoints	   = GridPoints		/. {opts} /. Options[Anopt];

(* local version of input (non-options) *)
	Z = InitialGuess;
	anopttol = AnoptTol;
        gamma = gam;
        
(* ------------------------------------------------------------------*)
               COMMENT[" Check input data "];
(* ------------------------------------------------------------------*)
        CheckTheData;



(* ------------------------------------------------------------------
	Initialize Messages and output files 
   ------------------------------------------------------------------*)
        DefineMessages;
        BeginOutputFiles[gam,anopttol]; 



(* ------------------------------------------------------------------
			Initialize variables  
   ------------------------------------------------------------------*)

	errors = {};
	ForceIter = False;
        windreal = " ";
	steptol = anopttol;	
	doubleprogresstolerance = .10;

(* . . .Set Dimensions *)
        ndim = 0;  
        While[Not[TrueQ[ComplexD[gamma,z[ndim+1]] == 0]] &&
		ndim < 25,++ndim];

(*. . . Smoothing variables . . .*)

	smoothingflag = Not[TrueQ[smoothing==None]];
	
	If[smoothingflag,
		progressTolerance   = .20;
	        (*smoothing if improvements<progressTolerance*)
		pSmthFac = " ";
		ppSmthFac = " ";
		smoothParameters =
			Which[TrueQ[smoothing==Automatic],{0,90},
			      TrueQ[Length[smoothing]==2],smoothing];
		{smoothstart,smoothend} = smoothParameters;
	  ];

	updateVectorSmoothflag = 
			If[TrueQ[updateVectorSmooth==None],False,True];

	initialSmoothflag = 
			If[TrueQ[initialSmooth==None],False,True];

(* . .  Reset Double Grid if necessary. . .*)
	If[Not[dGrid], mxGrid=nsp];


(* . . .Select norm according to problem solved . . . *)
	Which[subproblem=="DiskIteration",
	        normused = "SupNorm",
	      subproblem=="NewtonMethod",
		normused = "SupNorm"];
	      

(* . . .Tolerances, constants and functions specific to the norm used *)
	Which[normused == "SupNorm",
		If[optimalitytol == Automatic,optimalitytol=anopttol];
	   	norm[x_] := Max[Abs[x]],
	      normused == "L2Norm",
	        norm[x_] := Sqrt[Re[Flatten[x].Conjugate[Flatten[x]]/
		  			N[Length[Flatten[x]]]]]
	     ];

(* . . .Tolerances, constants and functions specific to the subproblem . . . *)
	Which[
	   subproblem=="DiskIteration",
		dualmaxits= If[quickResults,5,30];
		hessiantolerance = (anopttol)^(1/4);
		specfactol = If[quickResults,0.1,0.00001]; 
		disktol = If[quickResults,
		                ot1/10.,
				Min[10^-5,Max[anopttol/10,Abs[ot1]^(1.5) ] ]],
	   supbroblem=="NewtonMethod",
        	weight  = weight /. e->eLocal;
		NewtonMethodweight=Table[1.,{nsp}]		
		
	     ];

(* . .  Set dimensions ******)
	Zdim = Dimensions[Z];
	Which[TrueQ[Zdim=={ndim}],   
				nsp = gridpoints,
	      (Zdim[[1]]==ndim)&&(Length[Zdim]==2) && FreeQ[Z,e],
	        	        nsp = Zdim[[2]],
	       True, 
	         Message[CheckData::length,"Error in input initial guess"];
	         Abort[];
	       ];

(*. . . Construct important vectors. . .*)
        CheckZeroDivide[nsp,"nsp"];
	eLocal = Table[N[Exp[2 Pi I j/nsp]], {j,0,nsp-1}];
	zero  := Table[0.,{nsp/2 +1}];
	zeroV := Table[Table[0.,{nsp}],{ndim}];


(* . . .Discretize Z if it is given in terms of "e"  *)
	Z = (Z /. e->eLocal) + zeroV;
	
	aflag=Not[TrueQ[amatrix==None]];
	If[aflag && Length[Dimensions[amatrix]]==2,
		{adim,asp}=Dimensions[amatrix],
		{adim,asp} = {0,0}];

      If[Not[TrueQ[asp==nsp]] && aflag,
         Message[CheckData::gridsize,"Initial Guess","AMatrix"]; Abort[]];
      
      counter = 0;  
      While[ !TrueQ[ComplexD[gamma,a[counter+1]] == 0], ++counter];
      If[(counter != adim) && aflag,
            If[(adim<counter),
               Message[CheckData::dimension,"Number of parameters a[k] in performance is greater than the number of rows in specified AMatrix.",""]; Abort[]; 
            ,  Message[CheckData::warning,"The specified AMatrix has more rows than the number of parameters a[k] in performance.",""];
            ];
      ];




(* ------------------------------------------------------------------
			Function evaluation is set up
   ------------------------------------------------------------------*)

(*. . *)COMMENT[" Function evaluation rule "];
	If[aflag,
             evaluaterule0:= 
	     		Join[{e->eLocal},Table[a[j]->amatrix[[j]],{j,1,adim}]],
	     evaluaterule0:= {e->eLocal}			    
          ];

	evaluaterule:= 
		(Join[evaluaterule0,Table[z[j]->#[[j]],{j,1,ndim}]]&);


(*. . *)COMMENT[" Define function evaluation for line search "];
	BetaValue[x_]:=norm[gamma /. evaluaterule[Z + x updatev]];

(*. . *)COMMENT[" define bound function for line search "];
	If[TrueQ[constrainedsearch == None],
		lnsrchbndfn := ( 100. &),
	   (* else *)
		tol2 = 0.01;
		If[Length[constrainedsearch]==3,   (* do check of input *)
			{Gam2,bound2,tol2}=constrainedsearch,
			{Gam2,bound2}=constrainedsearch];
		foot0[t_]:= Max[Gam2 /. evaluaterule[Z+ t*updatev]] - bound2; 
		foot[ t_]:= Abs[foot0[t]]; 
		lnsrchbndfn[ ] := 
  		    Module[{ },
Print[foot[0.1]];
Print[foot[0.3]];
Print[foot[0.6]];
Print[foot[1.]];
Print[foot[2.]];
			{tmax,fmax} = GoldenLineSearch[foot,10.];
			If[Chop[tmax]==0,
	Print["Warning:constraint not binding in direction of update"];
			    tmax = 1;
			    tmaxcount = 0;
			    While[tmaxcount < 100 && foot0[tmax] > 0,
				tmaxcount++;
				tmax = tmax*0.5];				
	Print["        Bound for linesearch parameter is set to ",tmax];
			   ];
			Return[tmax];
			];
		];
 		
(* NEW 
	nvar = ndim+adim+1;
	dvar = Table[StringJoin[",x",ToString[j],"_"],{j,1,nvar}];
	dvarlist = ToExpression[StringJoin["{x1",
			Table[StringJoin[",x",ToString[j]],{j,1,nvar}],
			"}"]];
	dummyvar = "gamperf[x1_";
	Do[dummyvar = StringJoin[dummyvar,dvar[[i]]],{i,1,Length[dvar]}]; 
	dummyvar = StringJoin[dummyvar,"]"];

	ToExpression[dummyvar] = gamma /. 
		Join[{e->dvarlist[[i]]},
		     Table[z[j_]->dvarlist[[1+j]],{j,1,ndim}],
		     Table[a[j_]->dvarlist[[1+ndim+j]],{j,1,adim}]];
*)


(* ------------------------------------------------------------------
		  Symbolic algebra and derivatives 
   ------------------------------------------------------------------*)
(*. . *)COMMENT[" Construct partial derivatives gammaz, gammazzbar, gammazz"];
	Do[gammaz[i]=ComplexD[gamma,z[i]],{i,1,ndim}]; 
	Do[Do[gammazz[i,j]=ComplexD[gammaz[i],z[j]],{i,1,ndim}],{j,1,ndim}];
	Do[Do[gammazzbar[i,j]=ComplexD[gammaz[j],Conjugate[z[i]]],
	                         {i,1,ndim}],{j,1,ndim}];

(* . . .Algebra specific to the subproblem method *)
	Which[subproblem=="DiskIteration",
(* . . . . . . . . .Determine if function gamma is second order in z, zbar
		    with zz-hessian equal to zero and zzbar hessian 
		    constant and diagonal *)
		    diskflag = True;
		    Do[diskflag= diskflag && TrueQ[gammazzbar[i,j]==0],
		           {i,1,ndim-1},{j,i+1,ndim}];
		    Do[diskflag= diskflag && TrueQ[gammazz[i,j]==0],
		           {i,1,ndim},{j,i,ndim}];
		    Do[diskflag= diskflag && 
		                 TrueQ[ComplexD[gammazzbar[i,i],z[j]]==0],
		           {i,1,ndim},{j,i,ndim}];
(* . . . . . . . . .if disk case, turn off smoothing *)
        	    If[diskflag,
             		smoothing=None;  (* smoothing not needed *)
			updateVectorSmoothflag=False;
			If[diag2 >= 1,
			    WriteString[Summary1,
			    "Performance function gamma is of quasicircular type; \n",
			    "All smoothing has been turned off. \n \n"]];
			SmthFac="NON"
           		];
			   
(* . . . . . . . . .Avoid linesearch and iterations if subproblem solves all *)
		    niter = If[diskflag && Not[dGrid],1,niter];
(*		    doLineSearch = Not[diskflag]; *)
		    If[diag2 >= 1 && diskflag,WriteString[Summary1,
			    "Performance function gamma is of quasicircular type; \n",
			    "Linesearch has been turned off."]];

		    ,
	      subproblem=="NewtonMethod",
	            Null
	      ];


(* ------------------------------------------------------------------*)
                  COMMENT[" Zeroth Iteration: initial data "];	
(* ------------------------------------------------------------------*)

	iternum = 0;
(*. . . Clean initial guess function . . .*) 
	Z = Clean[Z];

(*. . . Smooth out initial guess if required . . .*)
	If[initialSmoothflag,
		{SmthFac,Z}=SmoothRoutine[Z,50,90];
		,
		SmthFac="NON"
          ];

(*. . . Function evaluation . . .*)
	functionevaluation;

(* . . .Compute Diagnostics . . .*)
	computediagnostics;

(*. . . Print output  . . . *)
	printoutput;

(* ----------------------------------------------------------------------------- *)




(* ----------------------------------------------------------------------------- *)
                    COMMENT[" Begin iteration "];
(* ----------------------------------------------------------------------------- *)

    While[
       COMMENT[" New iteration has begun "];
       Which[normused=="SupNorm",
               (iternum ==0 &&(ot1 > optimalitytol || ot2 > optimalitytol))
	        || (0<iternum<niter) && ForceIter || 
	       ((0<iternum<niter)&&(ot1 >optimalitytol || ot2 >optimalitytol)
	         &&(betabest*hnorm > steptol )),
 	     normused == "L2Norm",
	       Null
        ],



(*. . *)COMMENT[" Increment iteration number "]; 
	iternum = iternum+1;
        WriteIterationID;   

(*. . . Store prior values,reset other . . .*)
	ForceIter=False;
	pdoublegridflag = If[iternum==1,True,doublegridflag];
	psmoothingflag = If[iternum==1,True,smoothingflag];
	doublegridflag = False;
	
(* ----------------------------------------------------------------------------- *)

        Which[subproblem=="DiskIteration",

(*............................................................................*)
(*. . . . . . . *)COMMENT[" Call DiskIteration "];	
                  diskiterationtime = Timing[
                  {updatev,ExactSolve,errors,dualEigenfunction,ninnerloop,tau}=
	 	     DiskIteration[gkvalues,akvalues,Hkvalues,
		        DualExtremal`DualMaxIts ->dualmaxits,
		        DualExtremal`DualTol -> anopttol,
		        DualExtremal`DualCircleGrid ->eLocal,
		        DualExtremal`DualSingularFunction->dualEigenfunction,
		        AnoptUtilities`HarrisSFMaxIts->10,
		        AnoptUtilities`HarrisSFToler->specfactol,
		        DiskIterationMaxIts->15,
		        DiskIterationDiagnostics-> 
			    If[diag2==0 || diag2 ==1,0,4],
		        DiskIterationOutputChannel->{Summary2},
		        DiskIterationTol -> disktol,
		        DiskIterationWrite->TrueQ[diag2>0],
                        DiskIterationDebugMe->debug];
		      ];
                 If[debug,Print["DiskIteration time:",diskiterationtime]];
		 If[diag2 >= 1,
		     WriteString[Summary1,
		"\n DiskIteration diagnostics output: \n \n",
		"                    ExactSolve = ",
		ToString[ExactSolve],"\n",
		"        Flatness of subproblem = ",
		ToString[InputForm[errors[[2]]]],"\n",
		"NonAnalyticNoise of subp.sol h = ",
		ToString[InputForm[errors[[1]]]],"\n",
		" Tail F.C. of symbol of Hankel = ",
		ToString[InputForm[errors[[4]]]],"\n",
		"              Eigenvector step = ",
		ToString[InputForm[errors[[3]]]],"\n",
		"Abs.val.of Error in Spect.Fact.= ",
		ToString[InputForm[errors[[5]]]],"\n",
		"NonAnalyticNoise in Spect.Fact.= ",
		ToString[InputForm[errors[[6]]]],"\n",
		"             Subprob:  tau - 1 = ",
		ToString[InputForm[tau-1.]],"\n",
		"   Subproblem inner iterations = ",
		ToString[InputForm[ninnerloop]],"\n"];
		   ];
	         updatev= Clean[updatev],
(*............................................................................*)



(*............................................................................*)
              subproblem == "NewtonMethod",
                 Null
(*............................................................................*)
        ]; (* end  Which[subproblem ... ] *)



(* ----------------------------------------------------------------------------- *)
 
(*. . *)COMMENT[" Smooth update vector if requested "];
	Presmoothing;
	
(* ----------------------------------------------------------------------------- *)

(*. . *)COMMENT[" Begin Line Search "];

	LineSearch; 
    

(* ----------------------------------------------------------------------------- *)
        COMMENT[" Update "];                                   

	Z = Z + UpdateV;	
	Solution = Z;

(* ----------------------------------------------------------------------------- *)
(*. . *)COMMENT[" Function evaluation "];
	functionevaluation;

(* ----------------------------------------------------------------------------- *)
	
(*. . *)COMMENT[" Compute diagnostics "];
	computediagnostics;

(* ----------------------------------------------------------------------------- *)

(*. . . Print output  . . . *)
	printoutput;
(* ----------------------------------------------------------------------------- *)

(*. . *)COMMENT[" Grid Doubling "];
        GridDoubling[anopttol];
	
(* ----------------------------------------------------------------------------- *)

(*. . *)COMMENT[" Smoothing "];
	smoothingZ;
	
(* ----------------------------------------------------------------------------- *)

(*. . *)COMMENT[" Function evaluation if smoothing or grid doubling"];

	If[doublegridflag || smoothingflag, functionevaluation];


(* ----------------------------------------------------------------------------- *)

    ]; (*. . .End While. . .*)

(* ----------------------------------------------------------------------------- 
                        End of Iteration 
----------------------------------------------------------------------------- *)

(*. . *)COMMENT[" Check for local min, if applicable "];
	Which[normused == "SupNorm",
           	If[ndim == 1 && diag>=1 && 
	   	     ( ot1<optimalitytol)&&(ot2<optimalitytol),
                      DualWrite2[Summary1,"**A global min. has been obtained"];
	              WSpace[Summary1,1];
                  ];
		If[( ot1<anopttol)&&(ot2<anopttol)&& ndim==2 && ot3Check,
	    		MinCheck2d
          	  ],
	      normused == "L2Norm",
	        Null
	     ];


(*. .*)COMMENT[" Assign output variables "];
      If[ iternum>0, LastUpdate = UpdateV ];
      Solution = Z; 
      InfGamma = infgamma ;
      gammastar = supgamma;
      flat = ot1;
      grAlign = ot2;
      ned = ned0;

   

(*. .*)COMMENT[" Write final summary and close files "];
      FinalSummary;
      EndOutputFiles;

   Currentf = Z;

   ];  (* end Anopt Block *)

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)






(******************************************************************
                    INTERNAL ROUTINES OF ANOPT
******************************************************************)

computediagnostics :=
	Which[normused == "SupNorm",
        	If[iternum==0,
			pot1 = 0;
			pot2 = 0;
			popt = 0;
			phnorm = 0;
			hnorm = 0;
			,
		   (* else *)
			pot1 = ot1;
			pot2 = ot2;
			popt = opt;
			phnorm = hnorm;
			hnorm  = norm[updatev];
		  ];
		supgamma   = Max[N[Abs[gkvalues]]];
        	infgamma   = Min[N[Abs[gkvalues]]];

	If[supgamma < 0.001,  (* FIX *)
		spgam = ToString[InputForm[supgamma]];
		messagespgam = StringJoin["sup Gamma[e,f] = ",spgam,
				"\n Diagnostics flat and grAlign are undefined"];
		Message[CheckCode::withintolerance,messagespgam];
		Solution = Z;
		Abort[];
			];

        	CheckWithinTol[.001,supgamma,
			"sup Gamma[e,f(e)] is near zero at current f."];	

	minak = Min[ Re[Sum[Abs[akvalues[[j]]],{j,1,ndim}]]];
	maxak = Max[ Re[Sum[Abs[akvalues[[j]]],{j,1,ndim}]]];
	If[subproblem == "DiskIteration",		
		CheckWithinTol[0.0000000001,minak,
			"ComplexGradient of performance gamma at current f is near 0 at at least one gridpoint.  Anopt does not like this.  Anopt suggests trying a different initial guess, and checking the performance to see if it is defined correctly."];
	];

	minrad = Min[ Re[supgamma - gkvalues + Sum[Abs[akvalues[[j]]],{j,1,ndim}]]];
	maxrad = Max[ Re[supgamma - gkvalues + Sum[Abs[akvalues[[j]]],{j,1,ndim}]]];
		CheckWithinTol[0.000001, minrad,
			"ComplexGradient of performance gamma is near 0 at gridpoint where gamma[e,f(e)] attains max.  No local descent direction exists  at current f.  Anopt suggests trying a different initial guess."];
		opt = supgamma;
		ot1 = (supgamma-infgamma )/supgamma;
		{ot2,windreal}={WnoTest,WindReal}/.OptimalityTest2[akvalues];
		ned0 = Apply[Plus,Abs[errors]];
		akpointwisenorm = Table[Sqrt[Re[akvalues[[j]]*
					Conjugate[akvalues[[j]]]]],{j,1,ndim}];
		akmax = Max[akpointwisenorm];
		akmin = Min[akpointwisenorm];
		If[iternum > 0 && diag2 >= 1,
		     WriteString[Summary1,
		     "\n Diagnostics on update f' = f + t*h: \n \n"];
		     WriteString[Summary1,		     
"                             t = ",ToString[betabest],"\n",
"                       Norm[h] = ",ToString[hnorm],"\n",
"           Sup[gamma[e,f'(e)]] = ",ToString[supgamma],"\n",
"           Inf[gamma[e,f'(e)]] = ",ToString[infgamma],"\n",
"  Min || Dgamma/dz[e,f'(e)] || = ",ToString[minak],"\n",
"  Sup || Dgamma/dz[e,f'(e)] || = ",ToString[maxak],"\n",
"  		minimum radius  = ",ToString[minrad],"\n",
"  		maximum radius  = ",ToString[maxrad],"\n",
"                           ot1 = ",ToString[ot1],"\n",
"                           ot2 = ",ToString[ot2],"\n",
"                           ned = ",ToString[ned0],"\n \n"
		     ];
		  ];
		,
	      normused == "L2Norm",
	        Null
	     ]; 

(* ----------------------------------------------------------------------------- *)

   printoutput :=
	Which[normused == "SupNorm",
		Which[iternum==0,
        		WriteScreenHeading;
        		PartialWrite=True;
        		temp=StringJoin[" ",ToString[iternum],
		        	 " : ",NumFormat[supgamma,19],
				" :   N/A   : ",NumFormat[ot1,7]," : ",
				NumFormat[ot2,7],windreal,":   N/A   : ",
				SmthFac," : ",ToString[nsp] ],
		      True,
	        	If[iternum<10,itSpace=" ",itSpace=""];
        		temp=StringJoin[itSpace,ToString[iternum]," : ",
			NumFormat[supgamma,19]," : ",
        		NumFormat[betabest*hnorm,7]," : ",NumFormat[ot1,7],
			" : ",NumFormat[ot2,7],windreal,": ",NumFormat[ned0,7],
			" : ",SmthFac," : ",ToString[nsp] ]];
                DualWrite[Screen,temp];
	      ];


(* ----------------------------------------------------------------------------- *)

    functionevaluation :=
	       Which[subproblem == "DiskIteration",
	   	   gkvalues = gamma /. evaluaterule[Z];
		   akvalues = Table[gammaz[j],{j,1,ndim}] /. evaluaterule[Z];
		   Hkvalues = If[quickResults && Not[diskflag],
			Table[ Re[Sum[akvalues[[j]]* 
			Conjugate[akvalues[[j]]]/gkvalues,{j,1,ndim}]],{ndim}],
		   (* else *) 
		       Re[zeroV +
		      (Table[gammazzbar[j,j],{j,1,ndim}] /. evaluaterule[Z])]
		                 ];
			,		      
	         subproblem == "NewtonMethod",
	           Null
	       ];

(* --------------------------------------------------------------- *)


(* --------------------------------------------------------------- 
                    Smoothing Routines
------------------------------------------------------------------ *)
Presmoothing:=
	If[updateVectorSmoothflag,
	   Which[TrueQ[updateVectorSmooth==Automatic],
	         updatevfc = Fourier[updatev]/N[Sqrt[nsp]];
	         updatevspectrum = Re[updatevfc Conjugate[updatevfc]];
	         ratio = Sum[
	Apply[Plus,Take[updatevspectrum[[j]],{nsp/2+1,5 nsp/8 + 1}]]/
	     Apply[Plus,Take[updatevspectrum[[j]],{nsp/2+1,3 nsp/4 + 1}]],
	                     {j,1,ndim}];
		 updateVectorSmoothParam =
		 	Which[ratio > 0.9,{0,50},
		       		ratio > 0.6,{0,100},
				ratio > 0.3,{50,100},
		       		True,       {100,100}],
	         True, updateVectorSmoothParam = updateVectorSmooth];
            If[diag2 >= 1,
(*. . . .  *)COMMENT[" Write smoothparameters to file "];
	     WriteString[Summary1,
	       StringJoin["Smoothing Coeffs. for update: {",
	       ToString[updateVectorSmoothParam[[1]]],",",
	       ToString[updateVectorSmoothParam[[2]]],"}"]];
             WSpace[Summary1,2];
            ];
	   If[updateVectorSmoothParam[[1]] <100,
	     {SmthFac,updatev}= 
                  SmoothRoutine[updatev,
		  updateVectorSmoothParam[[1]],updateVectorSmoothParam[[2]]];
	     ];
	   ];

(* ----------------------------------------------------------------------------- *)
  smoothingZ := 
	(
        smoothingflag = ((iternum == 1 ) || ( iternum > 1 ) && 
		Not[doublegridflag] && Not[pdoublegridflag]) &&
		Not[TrueQ[smoothing==None]] && (iternum<niter) && 
       		((Not[posdef && ot1<anopttol && ot2<anopttol]));
	If[Not[smoothingflag],SmthFac="NON"];
        If[smoothingflag,

(*. . . .  *)COMMENT[" Begin smoothing. Reset Smoothing parameters"];
	   {smoothstart,smoothend}=
	    	If[smoothing==Automatic, 
	      	        SmoothTime,
			smoothParameters];
            If[diag2 >= 1,
(*. . . .  *)COMMENT[" Write smoothparameters to file "];
	     WriteString[Summary1,
	       StringJoin["Smoothing Coefficients Recommended: smoothstart=",
	       ToString[smoothstart],"     smoothend=",ToString[smoothend]]];
             WSpace[Summary1,2];
            ];
(*. . . .  *)COMMENT[" Carry out the smoothing "];
(* . . . . Smoothing occurs provided it hasnt happened in last 2 iters. *)
 	   {SmthFac,Z}=If[TrueQ[(smoothstart <100) &&((iternum==1) ||
                         ((iternum == 2)&& (pSmthFac=="NON") ) ||
	   		 (iternum>1&&(pSmthFac=="NON")&&(ppSmthFac=="NON")))],
	   		    SmoothRoutine[Z,smoothstart,smoothend],
			  (* else *)
			    {"NON",Z}];
	   ppSmthFac = pSmthFac;
	   pSmthFac = SmthFac;
	   If[Not[TrueQ[SmthFac=="NON"]],
	   		ForceIter = True,
			smoothingflag = False];
         ];
	);
(* ----------------------------------------------------------------------------- *)

SmoothRoutine[Vector_,n1_,n2_]:=
   Block[{x,NewVector,tString,smthFactor,smthfunction},

         NewVector=Vector;

(*. . . Smoothing routines for all objects. . .*)
        Which[smoothType == "A",
   	           smthfunction[x_]:=N[Cos[(Pi/2)  x]];
		   tString="A",
              smoothType == "B",
	           smthfunction[x_]:=1-x;
		   tString="B",
              smoothType == "C",
	           smthfunction[x_]:=2/(x+1) -1;
		   tString="C",
              smoothType == "D",
                   smthfunction[x_]:=1;
		   tString="D",
              smoothType == "E",
                   smthfunction[x_]:=N[0.5*(Cos[ Pi  x]+1.)];
		   tString="E",
              smoothType == "F",
                   smthfunction[x_]:=If[x==0,1.,N[Sin[Pi x]/(Pi x)]];
		   tString="F";
              ];

	NewVector=SmoothVector[NewVector,n1,n2,smthfunction];
	
        smthFactor=StringJoin[ToString[Floor[n1/10]],tString,
					ToString[Floor[n2/10]]];


        Return[{smthFactor,NewVector}];

   ]; (* end Block SmoothRoutine *)

(* ----------------------------------------------------------------------------- *)

SmoothVector[Vector_,n1_,n2_,SmthFunction___]:=
   Block[{smVector,Result,N1,N2,i,nsp,ndim},

(*. . . Smooth Vector according to function . . .*)
	{ndim,nsp}=Dimensions[Vector];

(*. . . Parameters (n1,n2) given . . .*)
	N1 = 1+Floor[n1/100 nsp/2];
        N2 = 1+Ceiling[n2/100 nsp/2];
	mask = Join[Table[1.,{N1-1}],
		    Table[SmthFunction[(i-N1)/(N2-N1)],{i,N1,N2}],
		    Table[0.,{nsp/2+1-N2}]];
	smVector = Join[ mask, Reverse[Drop[Rest[mask],-1]]];		
(*. . . Create new Vector . . .*)
	Result = Table[
	 	     InverseFourier[ Fourier[Vector[[i]]] smVector ],
                 {i,ndim} ];

        Return[Result];

   ]; (* end Block SmoothVector *) 


(* ----------------------------------------------------------------------------- *)

SmoothTime :=
       Module[{r1,r2,r3,cond1,cond2,cond3},
(* . . .ratios for measuring improvement *)			
        CheckZeroDivide[ot1,"ot1"];			
	r1 = (pot1 - ot1)/ot1;  (* % change in ot1 *)
	r2 = If[ndim > 1,
        	CheckZeroDivide[ot2,"ot2"];			
		(pot2 - ot2)/ot2,  (* % change in ot2 *)
		r1];               (* ot2 not important when ndim=1 *)
        CheckZeroDivide[opt,"opt"];			
	r3 = (popt - opt)/opt;  (* % change in opt *)

(* . . .Worst case when at least one ratio is negative *)
	cond1 = TrueQ[ Min[r1,r2,r3] <= 0];
(* . . .Next case when all ratios are less than tol *)	
	cond2 = TrueQ[ Max[r1,r2,r3] <= progressTolerance];
(* . . .Next case when at least one ratio is less than tol *)
	cond3 = TrueQ[ Min[r1,r2,r3] <= progressTolerance];
(* . . .The remaining cases are so good that smoothing is not nec. *)
	
	Which[  cond1 , { 0,50},    (* temper first 50% f.c., drop last 50% *)
	        cond2 , { 0,100},   (* temper all f.c. *)
		cond3 , {50,100},   (* keep first 50% fc., temper rest *)
		True  , {100,100}   (* no smoothing *)
		   ]
       ] (* end Block *)

(* ----------------------------------------------------------------------------- *)






(* -----------------------------------------------------------------------------                       			
 			Line Search  
----------------------------------------------------------------------------- *)

LineSearch:= 
   (
(* n32 is no.sampling points for a fast linesearch *)
	If[quickResults,
		n32 = Max[32,2^(Round[Log[2,nsp]/2])];
		Zsrch = Table[Table[Z[[j,i*nsp/n32 + 1]],
					{i,0,n32-1}],{j,1,ndim}];
		updatevsrch = 
		   Table[Table[updatev[[j,i*nsp/n32 + 1]],
		   			{i,0,n32-1}],{j,1,ndim}];

(* . . . . . . .Function evaluation rule  . . . *)
		If[iternum==1,
	   	   evaluaterule0srch = 
		      {e->Table[N[Exp[2 Pi I (i-1)/n32]],{i,1,n32}]};
		   If[aflag,
		      evaluaterule0srch = 
		 		Join[evaluaterule0srch,
		 		  Table[a[j]->amatrix[[j]],{j,1,adim}]]];
		   evaluaterulesrch = 
		    (Join[evaluaterule0srch,Table[z[j]->#[[j]],{j,1,ndim}]]&);
		   BetaValue[x_]:=norm[gamma /. 
		      evaluaterulesrch[Zsrch + x updatevsrch]];
		];
	   ];


(*. . *)COMMENT[" Reset lineserch if necessary "];     
	Which[subproblem=="DiskIteration",
		If[ndim>1 && iternum > 1 && TrueQ[searchfunction==Automatic] &&
		    ot1<optimalitytol &&  ot2 > 10.*ot1,
		    If[quickResults,
           	       BetaValue[x_] := (
		       gvals = Re[gamma /.evaluaterulesrch[Zsrch+x updatevsrch]];
		       (Max[gvals]-Min[gvals])/Max[gvals] + 
		          (WnoTest /. OptimalityTest2[
		           Table[gammaz[i],{i,1,ndim}]/.
			        evaluaterulesrch[Zsrch + x updatevsrch]]));
		       ,
           	       BetaValue[x_] := (
		       gvals = Re[gamma /.evaluaterule[Z+x updatev]];
		       (Max[gvals]-Min[gvals])/Max[gvals] + 
		          (WnoTest /. OptimalityTest2[
		           Table[gammaz[i],{i,1,ndim}]/.
			        evaluaterule[Z + x updatev]]));
		       ];
	            If[diag2 >= 1,
	     		WriteString[Summary1,"Line Search on flat+grAlign"];
             		WSpace[Summary1,2];
            		];
	   	   ];
	     ];

(*. . *)COMMENT[" Calculate bound for linesearch "];

	lnsrchbnd  = lnsrchbndfn[  ];
  	BetaValue[x_]:=norm[gamma /. evaluaterule[Z + x updatev]];


(*. . *)COMMENT[" Line search happens here "];
         linesearchtime = Timing[
            If[doLineSearch,
 	      {betabest,valuebest}= searchfunction[BetaValue,lnsrchbnd];
 		,
	      betabest = 1.
            ];
	]; (* end timing *)

        If[debug,Print["        linesearch time:",linesearchtime]];	       
         UpdateV = betabest*updatev;

   )


(* ----------------------------------------------------------------------------- *)



(* -----------------------------------------------------------------------------
                      Grid Doubling
----------------------------------------------------------------------------- *)
GridDoubling[AnoptTol_]:=
	If[(nsp<=mxGrid/2) && dGrid && 
	   (iternum<niter),
 (*. . .   *)COMMENT[" Determine if it is time to double grid "];
	     doublegridflag =DoubleTime && 
	     		Not[pdoublegridflag] && Not[psmoothingflag];
	     doublegridflag = doublegridflag  ||  
				((betabest * hnorm < steptol) &&
			((ot1 > optimalitytol ) || (ot2 > optimalitytol)));

(*. . .    *)COMMENT[" If diskflag and not optimal, double.. "];
 	     If[diskflag,
		If[TrueQ[ ot1>optimalitytol || ot2>optimalitytol ],
		    doublegridflag=True,
		    ForceIter=False; 
		    doublegridflag=False;
		]
              ]; 

	    If[doublegridflag,
(*. . . . . . *)COMMENT[" Double the grid size now "];
	        nsp = 2 nsp;
		If[ diag2 >= 1,
		WriteString[Summary1,"Grid doubling set. \n",
		     "New gridsize: ",ToString[nsp],"\n"]];
	        Z=DoubleGrid[Z];
		If[aflag, amatrix=DoubleGrid[amatrix] ];
		eLocal = DoubleGrid[eLocal];
	
                Which[ subproblem == "DiskIteration",
                          dualEigenfunction = DoubleGrid[dualEigenfunction],
                       subproblem == "NewtonMethod",
                          NewtonMethodweight = DoubleGrid[NewtonMethodweight];
		          TemplateBuild[nsp,ndim]
                ];
		zero=Table[0.,{nsp/2 + 1}];
		ForceIter=True;
	      ];
	];
(* end GridDoubling*)
(* ----------------------------------------------------------------------------- *)


(*. . . DoubleTime gives a True/False flag in return if the program
   evaluates the situation and determines that doubling is necessary 
*)

DoubleTime:=
   Block[{r1,r2,r3,d1,d2,d3,double},
(* . . .ratios for measuring improvement *)			
	r1 = (pot1 - ot1)/ot1;                (* % change in ot1 *)
	r2 = If[ndim>1,(pot2 - ot2)/ot2,r1];  (* % change in ot2 *)
	r3 = (popt - opt)/opt;                (* % change in opt *)

	d1 = Abs[ r1 ]<doubleprogresstolerance; 
	d2 = Abs[ r2 ]<doubleprogresstolerance;
	d3 = (ot1 >optimalitytol || ot2 >optimalitytol)&&
	         (0<iternum<niter)&&(betabest*hnorm < steptol );
	 
	double=TrueQ[(d1&&d2) || d3 ||
	             ((ned0>ot1) && (niter>1) && ((ot1>pot1)||(opt>popt)))];

        If[diag2>=1,
	   Write[Summary1,OutputForm["Anopt Suggests Double = "], double];
	   WSpace[Summary1,1];
        ];

	Return[double];
   ]; (* end DoubleTime Block *)
(* ----------------------------------------------------------------------------- *)




(**********************************************************************
                           Optimality Tests
**********************************************************************)

(*. . .MinCheck2d checks if Anopt arrived at a local min, or just a type
of saddle point *)

MinCheck2d:=
   Block[{A,B,b,bt,V1,V2,ka,kh,ckh,v1,v2,sign,windingNumber,A1,
	  x,F1,F2,F,k,Q,junk,phi,Argument}, 

(*. . . Construct Matricies. . .*)
	A=Table[ckh[j,i],{i,ndim},{j,ndim}];
	B=Table[kh[i,j],{i,ndim},{j,ndim}];
	b={-ka[2],ka[1]};
	bt={{-ka[2]},{ka[1]}};

(*. . . Construct vectors and evaluate (evaluation must be after composition)*)
	V1 = Conjugate[b].A.bt;
	V2 = Abs[b.B.bt]; 
	V3 = Re[b.B.bt]; 
	Do[ka[i]=gammaz[i];
	   Do[kh[i,j]=gammazz[i,j];ckh[i,j]=gammazzbar[i,j],{j,ndim}],
        {i,ndim}];
	v1 = Flatten[V1];
        v2 = Flatten[V2];
        v3 = Flatten[V3];

(*. . . Check first criterion. . .*) 
	sign = Sign[Chop[v1-v2,hessiantolerance] ]+1;
	If[sign.sign==0, posdef=False, posdef=True]; 

(*. . . Check second criterion if necessary. . .*)
        If[posdef==False,
	   windingNumber=WindingNumber[v2];
	   If[windingNumber/2!=Round[windingNumber/2] || windingNumber>2,
              posdef=True
           ];
        ];



(*. . . Display success if correct choice. . .*)
	If[posdef,
           If[diag>=1,
               DualWrite2[Summary1,"**A local min has been obtained"];
	       WSpace[Summary1,1];
           ];
	];
  ];   (* end Block MinCheck2d *)

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(************************************************************************)






(**********************************************************************
                       Data Checking 
************************************************************************)

CheckTheData :=
   (
(* ------------------------------------------------------- *)
(* Initially, messages are printed complete, as opposed to the 
   Mathematica default, where Short[] is used to Pre-evaluate.
   This will be reset again after checking the input is complete *)
   

      CheckData::length= "`` ``";
      CheckData::contradiction="`` doesn't match ``";
      CheckData::gridsize= "`` doesn't match ``"; 
      CheckData::dimension="`` ``"; 
      CheckData::invalid="`` ``";
      CheckData::warning=" `` ``";

(* . .performance . . *)
      If[FreeQ[gamma,e],Message[CheckData::warning,
      "Symbol e is not present in performance function",gamma]];

      If[FreeQ[gamma,z[1]],Message[CheckData::invalid,
      "Symbol z[1] is not present in performance function",gamma];
          Abort[]];

(* . .AnoptTol . .*)
	If[Not[ TrueQ[anopttol > 0]],
	    Message[CheckData::invalid,
	    "Input must be positive.  Offending input: ",anopttol];
	    Abort[]];

(* . .QuickResults . .*)
	If[Not[ MemberQ[{True,False},quickResults]],
	    Message[CheckData::invalid,
	    "Option QuickResults must be specified as True or False.  Offending input: ",quickResults];
	    Abort[]];

(* . .OptimalityTol . .*)
	If[Not[ TrueQ[(optimalitytol > 0) || (optimalitytol==Automatic)]],
	    Message[CheckData::invalid,
	    "Option OptimalityTol must be specified as a postive number or as Automatic.  Offending input: ",optimalitytol];
	    Abort[]];

(* . .Subproblem . .*)
	If[Not[ MemberQ[{"DiskIteration","Newton"},subproblem]],
	    Message[CheckData::invalid,
	    "Option Subproblem must be specified as a string, e.g. \"DiskIteration\" or \"Newton\".  Offending input: ",subproblem];
	    Abort[]];

(* . .DebugMe . .*)
	If[Not[ MemberQ[{True,False},debug]],
	    Message[CheckData::invalid,
	    "Option DebugMe must be specified as True or False.  Offending input: ",debug];
	    Abort[]];

(* . .Iterations . .*)
	If[Not[ TrueQ[IntegerQ[niter]&& niter>0]],
	    Message[CheckData::invalid,
	    "Option Iterations must be specified as a postive integer.  Offending input: ",niter];
	    Abort[]];

(* . .GridDouble . .*)
	If[Not[ MemberQ[{True,False},dGrid]],
	    Message[CheckData::invalid,
	    "Option GridDouble must be specified as True or False.  Offending input: ",dGrid];
	    Abort[]];

(* . .MaxGridSize . .*)
	If[Not[ TrueQ[IntegerQ[niter]&& niter>0]],
	    Message[CheckData::invalid,
	    "Option MaxGridSize must be specified as a postive integer.  Offending input: ",mxGrid];
	    Abort[]];

(* . .RunID . .*)
	If[Not[ TrueQ[Head[runID]==String]],
	    Message[CheckData::invalid,
	    "Option RunID must be specified as a string.  Offending input: ",runID];
	    Abort[]];

(* . .GridPoints . .*)
	If[Not[ TrueQ[gridpoints > 0 && IntegerQ[gridpoints] ]],
	    Message[CheckData::invalid,
	    "Options GridPoints must specify a positive integer.  Offending input: ",gridpoints];
	    Abort[]];

(* . .Weight . .*)
(* QQ ?? to be filled in *)

(* . .DualEigenfunction . .*)
(* QQ ?? to be filled in *)

(* . .OT3Check . .*)
	If[Not[ MemberQ[{True,False},ot3Check]],
	    Message[CheckData::invalid,
	    "Option OT3Check must be specified as True or False.  Offending input: ",ot3Check];
	    Abort[]];

(* . .DoLineSearch . .*)
	If[Not[ MemberQ[{True,False},doLineSearch]],
	    Message[CheckData::invalid,
	    "Option DoLineSearch must be specified as True or False.  Offending input: ",doLineSearch];
	    Abort[]];

(* . .Smoothing . .*)
	If[Not[  TrueQ[ MemberQ[{None,Automatic},smoothing] || 
		  (VectorQ[smoothing] && (Length[smoothing]==2))]],
	    Message[CheckData::invalid,
	    "Option Smoothing must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",smoothing];
	    Abort[]];

	If[TrueQ[VectorQ[smoothing] && (Length[smoothing]==2)],
	   If[Not[TrueQ[0 <= smoothing[[1]] <= smoothing[[2]] <= 100]],
	      Message[CheckData::invalid,
	      "Option Smoothing must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",smoothing];
	    Abort[]]];

(* . .InitialSmooth . .*)
	If[Not[  TrueQ[ MemberQ[{None,Automatic},initialSmooth] || 
		  (VectorQ[initialSmooth] && (Length[initialSmooth]==2))]],
	    Message[CheckData::invalid,
	    "Option InitialSmooth must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",initialSmooth];
	    Abort[]];

	If[TrueQ[VectorQ[initialSmooth] && (Length[initialSmooth]==2)],
	   If[Not[TrueQ[0 <= initialSmooth[[1]] <= initialSmooth[[2]] <= 100]],
	      Message[CheckData::invalid,
	      "Option InitialSmooth must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",initialSmooth];
	    Abort[]]];

(* . .UpdateVectorSmooth . .*)
	If[Not[  TrueQ[ MemberQ[{None,Automatic},updateVectorSmooth] || 
		  (VectorQ[updateVectorSmooth] && (Length[updateVectorSmooth]==2))]],
	    Message[CheckData::invalid,
	    "Option UpdateVectorSmooth must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",updateVectorSmooth];
	    Abort[]];

	If[TrueQ[VectorQ[updateVectorSmooth] && (Length[updateVectorSmooth]==2)],
	   If[Not[TrueQ[0 <= updateVectorSmooth[[1]] <= updateVectorSmooth[[2]] <= 100]],
	      Message[CheckData::invalid,
	      "Option UpdateVectorSmooth must be specified as None, Automatic, or as a list {n1,n2} of integers such that 0<=n1<=n2<=100.  Offending input: ",updateVectorSmooth];
	    Abort[]]];

(* . .SmoothType . .*)
	If[Not[ MemberQ[{"A","B","C","D","E","F"},smoothType]],
	    Message[CheckData::invalid,
	    "Option SmoothType must be specified as a string, e.g. \"A\" or \"B\".  Offending input: ",smoothType];
	    Abort[]];

(* . .Diagnostics . .*)
      If[Not[diag==-1]&&Not[diag==0]&&Not[diag==1]&&Not[diag==2],
          Message[CheckData::invalid,
	  "Option Diagnostics must be specified as 0, 1 or 2.  Offending input: ",diag];
      ];
      If[Not[diag2==0]&Not[diag2==1]&&Not[diag2==2]&&Not[diag2==3]&Not[diag2==4]&&Not[diag2==5],
          Message[CheckData::invalid,
	  "Option Diagnostics2 must be specified as 0, 1, 2, 3 or 4.  Offending input: ",diag2];
      ];


(* . .AMatrix . . *)            
      If[Not[MatrixQ[amatrix] || TrueQ[amatrix==None]],
         If[VectorQ[amatrix],
              Message[CheckData::invalid,"Options AMatrix was specified incorrectly as a vector instead of a matrix or None.",""];
              Abort[],
	      Message[CheckData::invalid,"Option AMatrix must be specified as a matrix or as None"," "];
              Abort[]]];

    );  (* end CheckTheData *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
		





(**********************************************************************
                       Printing and File Management
************************************************************************)

(* . . . Write Iterations identification to file . . .*)
WriteIterationID:=
   Block[{temp1,temp2},
      temp1 = OutputForm["******** ITERATION: "];
      temp2 = OutputForm[" ********"];
      If[ derivativeSave,     Write[SaveDerivative,temp1,iternum,temp2]];
      If[ functValSave,       Write[SaveFunctVal,temp1,iternum,temp2]];
      If[ gammaSave,          Write[SaveGamma,temp1,iternum,temp2]];
      If[ hessianSave,        Write[SaveHessian,temp1,iternum,temp2]];
      If[ spectralFactorSave, Write[SaveSpectralFactor,temp1,iternum,temp2]];
      If[ subproblemSave,     Write[SaveSubproblem,temp1,iternum,temp2]];
      If[ updateVectorSave,   Write[SaveUpdateVector,temp1,iternum,temp2]];
      If[ subproblemSolutionSave, 
                   Write[SaveSubproblemSolution,temp1,iternum,temp2]];
      If[ diag2>=1, Write[Summary1,temp1,iternum,temp2]];
      If[ diag2>=2, Write[Summary2,temp1,iternum,temp2]];
      If[ diag2>=3, Write[Summary3,temp1,iternum,temp2]];
      If[ diag2>=4, Write[Summary4,temp1,iternum,temp2]];
   
   ]; (*end WriteIterationID Block *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)



DiagnosticSave:=
(	

(*QQ Note:  these Save variables are disabled.  They variables probably . *)

   If[Not[prob1],

	If[functValSave,
         WriteString[SaveFunctVal,"--- Function Values ---"];
	   Do[Write[SaveFunctVal,OutputForm["Z["],i,OutputForm["]"]];
	   AppendVctSave[SaveFunctVal,Z[[i]] ],{i,ndim}]  
	];

	If[derivativeSave,
	 Write[SaveDerivative,OutputForm["--- First partial Derivatives ---"]];
	   Do[Write[SaveDerivative,OutputForm["ak["],i,OutputForm["]"]];
	   AppendVctSave[SaveDerivative,gammaz[i]],{i,ndim}];  
	 Write[SaveDerivative,OutputForm["---2nd partial Derivatives---"]];
	   Do[Write[SaveDerivative,OutputForm["Hk[["],i,OutputForm["]]"]];
	   AppendVctSave[SaveDerivative,Hkvalues[[i]] ],{i,ndim}]  
	];
	
	If[hessianSave,
	 Write[SaveHessian,OutputForm["--- Full Hessian terms ---"]];
	   Do[
		Write[SaveHessian,OutputForm["Real Hessian: Row "],
			i,OutputForm["     Column: "],j];	
		AppendVctSave[SaveHessian,gammazzbar[i,j]],
	   {i,ndim},{j,ndim} 
	   ]; 
	   Do[ 
		Write[SaveHessian,OutputForm["Complex Hessian: Row "],
			i,OutputForm["     Column: "],j];	
		AppendVctSave[SaveHessian,gammazz[i,j]],
	   {i,ndim},{j,ndim} 
	   ]
	];
	
	If[gammaSave,
	    Write[SaveGamma,OutputForm["--- gamma ---"]];
	    AppendVctSave[SaveGamma,gkvalues]
	];




       If[subproblem=="DiskIteration",
 
          If[subproblemSolutionSave,
            WSpace[SaveSubproblemSolution,2]; 
            Write[SaveSubproblemSolution,
             OutputForm["DualEigenfunction of DualExtremal --"]];
            Write[SaveSubproblemSolution,
             N[Chop[Fourier[dualEigenfunction]/Sqrt[nsp]],4] ];
            WSpace[SaveSubproblemSolution,2] 
          ]; 
 

   	   If[subproblemSave ,	   
              If[ExactSolve,
	   Write[SaveSubproblem,OutputForm["---W:  in problem W |K + h|^2
            (exact) --- "]];
	   Do[
             Write[SaveSubproblem,OutputForm["w[["],i,OutputForm["]]"]];
	     AppendVctSave[SaveSubproblem,Hkspecial[[i]]],
           {i,ndim}],
	   Write[SaveSubproblem,OutputForm["---W:  in problem W |K + h|^2---"]];
	   Do[
             Write[SaveSubproblem,OutputForm["w[["],i,OutputForm["]]"]];
	     AppendVctSave[SaveSubproblem,W[newdelta][[i]]],{i,ndim}]
	   ];
	   Write[SaveSubproblem,OutputForm["--- K: in problem W |K + h|^2---"]];
	   Do[
              Write[SaveSubproblem,OutputForm["K[["],i,OutputForm["]]"]];
	      Write[SaveSubproblem,N[Chop[Fourier[K[[i]]]/Sqrt[nsp]],4]],
           {i,ndim}];
           WSpace[SaveSubproblem,2] 
	   ]; 

	   If[updateVectorSave,
	   WriteString[SaveUpdateVector,"--- Update Vector for Subprob ---"];
	   Do[
            Write[SaveUpdateVector,OutputForm["update["],i,OutputForm["]"]];
	    AppendVctSave[SaveUpdateVector,update[[i]] ],
           {i,ndim}]  
	   ];

	   If[spectralFactorSave ,
	   Write[SaveSpectralFactor,OutputForm["--- Spectral Factorization ---"]];
            Do[
            Write[SaveSpectralFactor,OutputForm["alpha[["],i,OutputForm["]]"]];
	    AppendVctSave[SaveSpectralFactor,alpha[[i]]], {i,ndim}] 
            ];

       ]; 

   ];

   If[prob1,
     If[updateVectorSave,
       Write[SaveUpdateVector,OutputForm["--- actual update ---"]];
       Do[
          Write[SaveUpdateVector,OutputForm["UpdateV[["],i,OutputForm["]]"]];
	  AppendVctSave[SaveUpdateVector,UpdateV[[i]] ],
       {i,ndim}]  
     ]; 
   ]; 

); (* end DiagnosticSave *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)



(*. . .DualWrite writes string to Screen, and to file[name] . . . *)
DualWrite[name_,string_]:=
   (     
	If[diag==0||diag==1, Print[string]];
        If[diag==1||diag==2, Write[name,OutputForm[string]]];
   );


(*. . .DualWrite2 writes string to Screen, and to file[name] . . . *)
DualWrite2[name_,string_]:=
   (     
	If[diag==0||diag==1, Print[string]];
        Write[name,OutputForm[string]];
   );

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


WSpace[name_,spacenum_]:=
   (
	Do[Write[name,OutputForm[" "]],{junkCounter,1,spacenum}];  
   );
   

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)

(* . . .AppendVctSave appends a mathematica file.  If q=True, 
then add spaces *)
AppendVctSave[name_,vector_]:= 
   ( 
	AppendVctSave[name,vector,True]; 
   );

AppendVctSave[name_,vector_,spaces_]:= 
   Block[{num,l,i},
	If[spaces,WSpace[name,2]];
	If[Length[vector]==0,l={vector},
		l=Flatten[vector] 
	];
	Do[
		num=l[[i]];
		Write[name,aform[Re[num]],
		OutputForm["E"],pwr,	
		OutputForm["  "],aform[Im[num]],
		OutputForm["E"],pwr], 
	{i,1,Length[l]}];
	If[spaces,WSpace[name,2]] 
   ]; (* end AppendVctSave Block *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)

aform[tempnumber_]:=
  (
   If[Chop[tempnumber,10^-15]==0 || Chop[tempnumber-1,10^-15]==0,pwr=0
	;Return[Chop[tempnumber,10^-15]],
		pwr=Floor[N[(Log[10,tempnumber]),16]];
		Return[N[(tempnumber 10^(-pwr)),16]] 	
   ] 
  );

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


(*. . . NumFormat format a number to occupy a certain size to aid
	in screen output (i.e. NumFormat[31,9] = '3.1__E+01') *) 

NumFormat[number_,nsize_]:=
   Block[{Stringtemp,pwrstring,pwr(*,adjust*)},

        If[Sign[number]==-1, adjust=1, adjust=0];
	Stringtemp=ToString[N[aform[number],nsize-5-adjust]];
	pwrstring=ToString[Abs[pwr]];
	
	If[Abs[pwr]<10, 
                pwrstring=StringJoin["0",pwrstring]
        ]; 
	
	If[pwr<0,
		pwrstring=StringJoin["E-",pwrstring],
		pwrstring=StringJoin["E+",pwrstring]
	];

	While[
  	  StringLength[Stringtemp]+StringLength[pwrstring]<nsize-adjust, 
	  Stringtemp=StringJoin[Stringtemp," "] 
	];

        Return[StringJoin[Stringtemp,pwrstring]] 

   ]; (* end NumFormat Block *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
DefineMessages:=
   (
        Debugger::location= " `` ";
        COMMENT[location_]:=
           If[debug, Off[General::stop];Message[Debugger::location,location];
                     On [General::stop]];

        CheckCode::zerodivide= " `` = 0";
        CheckZeroDivide[value_,variable_]:=
           If[value==0., Message[CheckCode::zerodivide,variable];Abort[]];

        CheckCode::withintolerance= " `` ";
        CheckWithinTol[tol_,x_,comment_]:=
           If[x<tol, Message[CheckCode::withintolerance,comment];Abort[]];
   )


BeginOutputFiles[gamma_, AnoptTol_]:=
(

(*. . . Set individual diagnostics files to open if wanted . . .*)
        If[diag2>=5,
           derivativeSave=True; functValSave=True; gammaSave = True;
           hessianSave= True;  spectralFactorSave = True; subproblemSave = True;
           updateVectorSave = True; subproblemSolutionSave= True;
           solutionSave = True; projDiffMatrixSave= True;
        ];

(*. . . Set up the date . . .*)
        datetime=
	   Block[{year,month,day,hour,minute,second},
                 {year,month,day,hour,minute,second}=Map[ToString,Date[]];
                 StringJoin[month,"-",day,"-",year,"  ",hour,":",
	             minute,":",second]
           ];


(*. . . Initialize filenames . . .*)
        Screen   = "Screen.out";
        Summary1 = "Summary1.out";
        Summary2 = "Summary2.out";
        Summary3 = "Summary3.out";
        Summary4 = "Summary4.out";
        SaveDerivative         = "DerivativeSave.out";
        SaveFunctVal           = "FunctValSave.out";
        SaveGamma              = "GammaSave.out";
        SaveHessian            = "HessianSave.out";
        SaveSpectralFactor     = "SpectralFactorSave.out";
        SaveSubproblem         = "SubproblemSave.out";
        SaveSubproblemSolution = "SubproblemSolutionSave.out";
        SaveUpdateVector       = "UpdateVectorSave.out";
        SaveSolution           = "SolutionSave.out";
        SaveDEandLone          = "DEandLoneSave.out";





(*. . . Open all files and print out heading of file,runId, date/time . . .*)
        If[ diag==1||diag==2, 
           OpenStream=Streams[Screen];
           If[OpenStream != {},Close[Screen]];
           OpenWrite[Screen]; 
           WriteHeading[Screen,runID,datetime] ];
        If[ diag2>=1,
           OpenStream=Streams[Summary1];
           If[OpenStream != {},Close[Summary1]];
           OpenWrite[Summary1];
           WriteHeading[Summary1,runID,datetime] ];
        If[ diag2>=2,
           OpenStream=Streams[Summary2];
           If[OpenStream != {},Close[Summary2]];
           OpenWrite[Summary2]; 
           WriteHeading[Summary2,runID,datetime] ];
        If[ diag2>=3,
           OpenStream=Streams[Summary3];
           If[OpenStream != {},Close[Summary3]];
           OpenWrite[Summary3]; 
           WriteHeading[Summary3,runID,datetime] ];
        If[ diag2>=4,
           OpenStream=Streams[Summary4];
           If[OpenStream != {},Close[Summary4]];
           OpenWrite[Summary4]; 
           WriteHeading[Summary4,runID,datetime] ];
        If[ derivativeSave,
           OpenStream=Streams[SaveDerivative];
           If[OpenStream != {},Close[SaveDerivative]];
           OpenWrite[SaveDerivative];
           WriteHeading[SaveDerivative,runID,datetime] ];
        If[ functValSave,
           OpenStream=Streams[SaveFunctVal];
           If[OpenStream != {},Close[SaveFunctVal]];
           OpenWrite[SaveFunctVal];
           WriteHeading[SaveFunctVal,runID,datetime] ];
        If[ gammaSave,
           OpenStream=Streams[SaveGamma];
           If[OpenStream != {},Close[SaveGamma]];
           OpenWrite[SaveGamma];
           WriteHeading[SaveGamma,runID,datetime] ];
        If[ hessianSave,
           OpenStream=Streams[SaveHessian];
           If[OpenStream != {},Close[SaveHessian]];
           OpenWrite[SaveHessian];
           WriteHeading[SaveHessian,runID,datetime] ];
        If[ spectralFactorSave,
           OpenStream=Streams[SaveSpectralFactor];
           If[OpenStream != {},Close[SaveSpectralFactor]];
           OpenWrite[SaveSpectralFactor];
           WriteHeading[SaveSpectralFactor,runID,datetime] ];
        If[ subproblemSave,
           OpenStream=Streams[SaveSubproblem];
           If[OpenStream != {},Close[SaveSubproblem]];
           OpenWrite[SaveSubproblem];
           WriteHeading[SaveSubproblem,runID,datetime] ];
        If[ subproblemSolutionSave, 
           OpenStream=Streams[SaveSubproblemSolution];
           If[OpenStream != {},Close[SaveSubproblemSolution]];
           OpenWrite[SaveSubproblemSolution];
           WriteHeading[SaveSubproblemSolution,runID,datetime] ];
        If[ updateVectorSave,
           OpenStream=Streams[SaveUpdateVector];
           If[OpenStream != {},Close[SaveUpdateVector]];
           OpenWrite[SaveUpdateVector];
           WriteHeading[SaveUpdateVector,runID,datetime] ];

(*. . . Write to file the run details . . . *) 
        WriteOut = WriteOptions;
        If[diag2>=1,
        Write[ Summary1,OutputForm[" (* Anopt was run with the following values: *) "]];
        Write[ Summary1,OutputForm[" "]];
        Write[ Summary1,OutputForm["gamma ="]]; 
        Write[ Summary1,OutputForm["  "], gamma ,OutputForm[";"]];	
        Write[ Summary1,OutputForm[" "]];
        Write[ Summary1,OutputForm["AnoptTol = "], AnoptTol,OutputForm[";"]]; 
        Write[ Summary1,OutputForm[" "]];
        Write[ Summary1,OutputForm["opts ="]];
        Write[ Summary1,WriteOut,OutputForm[";"]];
        Write[ Summary1,OutputForm[" "] ];
	Write[ Summary1,OutputForm["initfun = "],
		OutputForm[ToString[InputForm[Z]]],OutputForm[";"]];	
        Write[ Summary1,OutputForm[" "] ];
        Write[ Summary1,OutputForm["Anopt[gamma,AnoptTol,initfun,opts] "]];
        Write[ Summary1,OutputForm[" "] ];
        Write[ Summary1,OutputForm[" "] ];
        Write[ Summary1,OutputForm["(* ------------- cut here for an executable -------------- *)"] ];		
        Write[ Summary1,OutputForm[" "] ];
        Write[ Summary1,OutputForm[" "] ];
        Write[ Summary1,OutputForm[" "] ];		
        ];
        If[diag==1||diag==2,
           Write[ Screen, OutputForm["gamma ="]];
           Write[ Screen, OutputForm["  "], gamma ];
           Write[ Screen, OutputForm["AnoptTol = "], AnoptTol];
           Write[ Screen, OutputForm[" "]];
        ];

); (* end BeginOutputFiles *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


FinalSummary:=
(

   DualWrite[Screen," "];
   DualWrite[Screen," "];
   DualWrite[Screen,"Summary"];
   DualWrite[Screen,"-------"];

If[normused == "SupNorm",
   DualWrite[Screen,
      StringJoin["gammastar = ",NumFormat[supgamma,21]]]
 ,  
   DualWrite[Screen,
      StringJoin["L1 of gamma: L1 Gamma = ",NumFormat[supgamma,21]]];
];

   DualWrite[Screen,
      StringJoin["flat      = ",NumFormat[ot1,16]]];
   DualWrite[Screen,
      StringJoin["grAlign   = ",NumFormat[ot2,16]]]; 
   If[iternum>0, DualWrite[Screen,
      StringJoin["ned      = ",NumFormat[ned0,8]]]
   ];



); (* end FinalSummary *)

(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


EndOutputFiles:=
(
	If[ diag==1||diag==2      ,Close[Screen]              ];
	If[ diag2>=1              ,Close[Summary1]                ];
	If[ diag2>=2              ,Close[Summary2]                ];
	If[ diag2>=3              ,Close[Summary3]                ];
	If[ diag2>=4              ,Close[Summary4]                ];
        If[ derivativeSave        , Close[SaveDerivative]         ];
        If[ functValSave          , Close[SaveFunctVal]           ];
        If[ gammaSave             , Close[SaveGamma]              ];
        If[ hessianSave           , Close[SaveHessian]            ];
        If[ spectralFactorSave    , Close[SaveSpectralFactor]     ];
        If[ subproblemSave        , Close[SaveSubproblem]         ];
        If[ subproblemSolutionSave, Close[SaveSubproblemSolution] ];
        If[ updateVectorSave      , Close[SaveUpdateVector]       ];

); (* end EndOutputFiles *)



(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)

WriteScreenHeading:=
(
 DualWrite[Screen," "];
 temp="It :    Current Value    :  Step   : Optimality Tests  :"; 
  	temp=StringJoin[temp,"  Error  : Sm. : Grid"];
 DualWrite[Screen,temp];

If[normused == "SupNorm",
 temp="   :     gammastar       :         :   flat   grAlign  :";
	temp=StringJoin[temp,"   ned   :     :"]; 
 DualWrite[Screen,temp],
 temp="   :      L1 Gamma       :         :   flat   grAlign  :";
	temp=StringJoin[temp,"   ned   :     :"]; 
 DualWrite[Screen,temp]];


 temp="-------------------------------------------------------------";
	temp=StringJoin[temp,"----------------"];
 DualWrite[Screen,temp];
); (* end WriteScreenHeading *)


(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


(* Write information about the run to file *)


WriteOptions :=
OutputForm[
StringJoin["Sequence[AMatrix->",ToString[InputForm[amatrix]],",QuickResults->",ToString[quickResults],",OptimalityTol->",ToString[optimalitytol],",Subproblem->\"",subproblem,"\",DebugMe->",ToString[debug],",Iterations->",ToString[niter],",GridDouble->",ToString[dGrid],",MaxGridSize->",ToString[mxGrid],",Diagnostics->",ToString[diag],",Diagnostics2->",ToString[diag2],",RunID->\"",ToString[runID],"\",Weight->",ToString[weight],",DualEigenfunction->",ToString[InputForm[dualEigenfunction]],",OT3Check->",ToString[ot3Check],",DoLineSearch->",ToString[doLineSearch],",Smoothing->",ToString[smoothing],",InitialSmooth->",ToString[initialSmooth],",UpdateVectorSmooth->",ToString[updateVectorSmooth],",SmoothType->\"",smoothType,"\"];"]];






(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)


WriteHeading[ NameOfFile_,runID_,datetime_ ]:=
  ( Write[NameOfFile, OutputForm["(* Name of File:  "], OutputForm[NameOfFile]];
    Write[ NameOfFile,OutputForm["   RunID       :  "], OutputForm[runID] ];
    Write[ NameOfFile,OutputForm["   Date/Time   :  "], OutputForm[datetime] ];
    Write[ NameOfFile,OutputForm["*) "] ];
    Write[ NameOfFile,OutputForm[" "] ];
   );
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)
(* --------------------------------------------------------------- *)

(****************** end Printing and File Management ************************)



End[]; (* "`Private`" *)

EndPackage[];
