(* :Title:  AnoptUtilities.m *)

(* :Authors: J.William Helton, Orlando Merino, and Trent Walker
   :Programming:  O. Merino, Julia Myers and T.Walker *)

(* :Context: AnoptUtilities` *)

(* :Package Version: 4.0 *)

(* :Copyright (c) 1991 J.W. Helton and O. Merino
    All rights reserved.  Usual disclaimer applies.
*)

(* :History: *)

(* :Keywords: *)

(* :Source: *)

(* :Warning: *)

(* :Mathematica Version 2.0  or later *)

(* :Requirement: *)

(* :Limitation: *)
   
(* :Discussion: *)


BeginPackage[ "AnoptUtilities`" ]


(***********************************************************
                         Usage Notes 
***********************************************************)

$AnoptUtilitiesVersion::usage="$AnoptUtilitiesVersion"

(* . . . Smooth Variables . . .*)
Smooth::usage = " \n
   Smooth[f,t], Smooth[f,t,opts] smooths out a discrete function f by \n
   tempering the Fourier coefficients with a window, either specified by the  \n
   user as an option or defaulted to the Blackman-Harris-Nutall window.  The \n
   input parameter t is an integer that satisfies the relation Smooth[f,t] = \n
   Smooth[ Smooth[f,t-1],1]. Smooth[f,0] returns f, and Smooth[f,1] is the \n
   smallest amount of smoothing possible with the formula above."
SmoothWindow::usage = " \n
   List of attenuation coefficients used in Smooth[].  If this option is \n
   used, the length of the list given in the option should match the length  \n
   of the function to be smoothed."

(*. . . Convolve variables . . .*)
Convolve::usage = " \n
   Convolve[a,b] produces the convolution of two lists a and b"

(*. . . HarmonicConjugate . . *)
HarmonicConjugate::usage = "\n
   HarmonicConjugate[u] produces v such that u+ I v is analytic ( i.e., \n
   the Fourier coeffs. with neg. index are 0)."

(*. . . WilsonSF variables . . *)
WilsonSF::usage = "\n
   WilsonSF[r], WilsonSF[r,opts] attempts to produce the spectral factor f \n
   of the function r, i.e., r = Abs[f]^2 and f is analytic ( i.e., the \n
   Fourier coeffs. with neg. index are 0)"
SpectralFactor::usage = " \n
   List of numbers, the spectral factor of a list of positive numbers \n
   used as input in a routine such as WilsonSF or HarrisSF."
SpectralFactorNonAnalyticNoise::usage = " \n
   A nonnegative number, a measure of the non-analytic noise generated\n
   in the answer after a call to a spectral factorization routine \n
   (WilsonSF or HarrisSF).  It is calculated as  \n
   Sum of Abs. vals. of Fourier coeffs. of alpha with negative index."
SpectralFactorError::usage = " \n
   A nonnegative number, one measure of error in a call to a Spectral \n
   Factorization routine (WilsonSF or HarrisSF).  It is calculated as: \n
   Max[ r - Abs[alpha]^2 ], where r=positive data and alpha=spectral factor."
WilsonSFSave::usage = "\n
    WilsonSFWriteDiagnostic is an option for WilsonSF[].  When set to True \n
    a run of WilsonSF writes error diagnostics.  The channel for printing is \n
    specified with WilsonSFOutputChannel.  Its default is True."
WilsonSFOutputChannel::usage = " \n
   Option for WilsonSF[].  Specifies the output stream for writing error \n
   diagnostics."

(*. . . HarrisSF . . .*)
HarrisSF::usage = " \n
   HarrisSF[r], HarrisSF[r,opts] attempts to produce the spectral factor f \n
   of the function r, i.e., r = Abs[f]^2 and f is analytic ( i.e., the \n
   Fourier coeffs. with neg. index are 0).  The method is iterative, \n
   (Newton's method, algorithm by ?Harris."
HarrisSFStep::usage = "\n
   A nonnegative number, output to HarrisSF.  It is the size of the step \n
   between the last two iterates of the spectral factor."
HarrisSFIterations::usage = "\n
   An integer, output to HarrisSF.  The total number of iterations used \n
   in a run to obtain the answer."
HarrisSFWriteDiagnostics::usage = " \n
   Option for HarrisSF[].  When set to True a run of HarrisSF writes \n
   error diagnostics.  The channel for printing is specified with \n
   HarrisSFOutputChannel."
HarrisSFSave::usage = " \n
   HarrisSFWriteDiagnostics is an Option for HarrisSF[].  When set to True \n
   a run of HarrisSF writes error diagnostics.  The channel for printing \n
   is specified with HarrisSFOutputChannel."
HarrisSFOutputChannel::usage = " \n
   HarrisSFOutputChannel is an Option for HarrisSF[].  Specifies the output \n
   stream for writing error diagnostics.";
HarrisSFInitial::usage = " \n
   HarrisSFInitial is an Option of HarrisSF.  A list (analytic), the initial \n
   guess at the spectral factor of the data."
HarrisSFMaxIts::usage = " \n
   HarrisSFMaxIts is an Option of HarrisSF.  Specifies maximum number of \n
   loop iterations."
HarrisSFToler::usage = " \n
   HarrisSFToler is an Option of HarrisSF.  Specifies tolerance for stopping \n
   loop iterations."

(* Also defined in WilsonSF 
SpectralFactor
SpectralFactorError
SpectralFactorNonAnalyticNoise
*)
DiskListPlot::usage="DiskListPlot[data] plots a list data of complex values in the Re-Im plane.  The list data is assumed to be the discrete values of a complex function defined on the interval [0, 2 Pi], sampled on equally spaced points.  DiskListPlot[{data1,data2,...}] produces an array of plots corresponding to data1, data2, ...";

DiskListPlot::usage="DiskListPlot[foo, {t, tmin, tmax}] produces a parametric plot of the complex valued expression foo generated as a function of t.";

DiskListPlot3D::usage="DiskListPlot3D[data] plots a list data of complex numbers in 3D space.  The list data is assumed to be the discrete values of a complex function defined on the interval [0, 2 Pi], sampled on equally spaced points. DiskListPlot3D[{data1,data2,...}] produces an array of plots corresponding to data1, data2,etc."

NormSup::usage="  ";
Norm2::usage="  ";
Norm1::usage="  ";
MatrixNorm::usage="  ";
OptimalityTest1::usage="  ";
OptimalityTest2::usage="  ";
WindingNumber::usage="  ";
ArgVector::usage="  ";
DoubleGrid::usage="  ";
HalfGrid::usage="  ";
Clean::usage="  ";
GoldenLineSearch::usage="  ";
WnoTest::usage=" ";
WindReal::usage=" ";



$AnoptUtilitiesVersion = "Anopt 4.0 (June 1998)";


(**********************************************************************
                         Option Defaults
**********************************************************************)
Options[Smooth]={SmoothWindow->Automatic};

Options[WilsonSF]={WilsonSFSave->True,
		   WilsonSFOutputChannel -> OutputStream["stdout",1]};
		   
Options[HarrisSF]={HarrisSFInitial->Automatic,
		   HarrisSFMaxIts->10,
		   HarrisSFToler ->10^(-10),
		   HarrisSFSave->True,
		   HarrisSFOutputChannel -> OutputStream["stdout",1]};


Begin["`Private`"]

(* Plotting routines ------------------------------------- *)

DiskListPlot[data:{_?NumberQ ..}, opts___] :=
	ListPlot[Transpose[{Re[data],Im[data]}],
			opts,Prolog -> PointSize[0.01],
 			AxesLabel->{"Re","Im"}];

DiskListPlot[data:{{_?NumberQ .. } ..}, opts___] :=
	Show[GraphicsArray[Table[
	{DiskListPlot[data[[i]],opts,DisplayFunction->Identity]},
	{i,1,Length[data]}]],DisplayFunction->$DisplayFunction];


DiskListPlot[ff_,{t_,tmin_,tmax_},opts___]:=
	ParametricPlot[{Re[ff],Im[ff]},{t,tmin,tmax},opts];

DiskListPlot3D[data:{_?NumberQ ..}, opts___]:=
   Block[{i},
      Show[Graphics3D[Table[
                 Point[{ N[2 Pi (i-1)/Length[data]],
                      Re[data[[i]]],Im[data[[i]]]}],
             {i,1,Length[data]}],
         Axes->Automatic,AxesLabel-> {"theta","Re","Im"},
         BoxRatios -> {1,0.4,0.4},opts]]
  ]; (* end DiskListPlot3D Block *)
  
DiskListPlot3D[data:{{_?NumberQ .. } ..}, opts___] :=
	Show[GraphicsArray[Table[
	{DiskListPlot3D[data[[j]],opts,DisplayFunction->Identity]},
	{j,1,Length[data]}]],DisplayFunction->$DisplayFunction];

(* Fourier analysis etc. --------------------------------- *)

Smooth[ff:{_?NumberQ ..},tt_Integer, opts___Rule]:=
     Module[{n,w1,window,twopin},
	window = SmoothWindow /. {opts} /. Options[Smooth];

	If[tt<1 ,Return[ff]];

	(* if window defaulted, use the Blackman-Harris-Nutall window *)
		If[TrueQ[window == Automatic],
			n = Length[ff];
			twopin = N[2 Pi/n];
			w1 = Table[ 0.35875 + 0.48829*Cos[twopin*ii] + 
			     	0.14128*Cos[twopin*2*ii]+
				0.01168*Cos[twopin*3*ii],
				{ii,0,n/2}];
			window = Join[w1,Reverse[Take[w1,{2,n/2}]]];
			];
		If[TrueQ[Length[ff]==Length[window]],
			Return[InverseFourier[Fourier[ff]* window^tt]],
			Print["Error in input to Smooth[]"];
			Return[ ];
		    ];
     ]; (* end Module Smooth *)

		


Convolve[a:{_?NumberQ ..},b:{_?NumberQ ..}]:=
	InverseFourier[ Fourier[a] Fourier[b]/N[Sqrt[Length[a]]]];
 


HarmonicConjugate[u:{_?NumberQ ..}]:=
       Module[{n,ufc,n2,h,fcv,j}, 

	  ufc = Fourier[u];
	  n = Length[u]; 
          n2 = n/2;
          h = Table[N[If[(j==1)||(j==n2+1),0.,If[j < n2+1,I, (-1)*I]]],
			{j,1,n}];
	  fcv = ufc h ;
	  Return[InverseFourier[fcv]];

       ]; (* end Module HarmonicConjugate *)


(* Spectral factorization of lists of positive numbers --------- *)


WilsonSF[r_,opts___Rule]:=
   Module[{n,u,fc,channel,diag},

	diag    = WilsonSFSave          /. {opts} /. Options[WilsonSF];	
	channel = WilsonSFOutputChannel /. {opts} /. Options[WilsonSF];

	n = Length[r];
	u = N[0.5 Log[r],16];
	fc = Exp[ u + I*HarmonicConjugate[u] ];
	sferror = Max[Abs[r-Abs[fc]^2]];
	nan = Apply[Plus,Take[Abs[Fourier[fc]],{2,n/2}]]/N[Sqrt[n]];

       If[diag,
	Write[channel,OutputForm[" "] ];
	Write[channel,OutputForm["**WilsonSF DIAGNOSTICS OUTPUT "]];
	Write[channel,OutputForm["Max[Abs[r-Abs[specfac(r)]^2]]   ="],sferror];
	Write[channel,OutputForm["Nonanalytic noise in specfac(r) ="],nan];
       ];

        Return[{SpectralFactor->fc,
                SpectralFactorError->sferror,
                SpectralFactorNonAnalyticNoise->nan}]; 
              
   ]; (* end WilsonSF Module *)






(* . . .  
HarrisSF[r] does spectral factorization  of the discrete function r
on a grid of equally spaced points on the unit circle.
The method by ?? Harris is iterative (Newton's method). 
. . .*)

HarrisSF[r_,opts___Rule]:=
   Module[{n,j,rootnorm1r,step,f1,f,u,v,maxits,toler,norm2v,diag,channel},

	diag = HarrisSFSave             /. {opts} /. Options[HarrisSF];	
	channel = HarrisSFOutputChannel /. {opts} /. Options[HarrisSF];
	f1 = HarrisSFInitial            /. {opts} /. Options[HarrisSF];
	maxits = HarrisSFMaxIts         /. {opts} /. Options[HarrisSF];
	toler = HarrisSFToler           /. {opts} /. Options[HarrisSF];

	n = Length[r];
	rootnorm1r = Sqrt[ Apply[Plus,Abs[r]] /N[n]];
	If[TrueQ[f1==Automatic],f1 = Table[rootnorm1r,{j,1,n}]]; 

	j = 0;
	step = 1000.0;
	While[((step > toler )&&(maxits-j>0)),
	   f = f1;
	   u = 0.5*r/(Abs[f])^2 + 0.5;
	   v = (u + HarmonicConjugate[u] * I) f;
	   norm2v = Abs[Sqrt[ Re[v.Conjugate[v]]/N[n] ] ];
	   f1 = v * (rootnorm1r/norm2v);
	   step = Max[Abs[f - f1]];
	   j = j+1 ;
	];

	sferror = Max[Abs[r-Abs[f1]^2]];
	nan = Apply[Plus, Take[Abs[Fourier[f1]],{2,n/2}]]/N[Sqrt[n]];

	If[diag,
	   Write[channel,OutputForm[" "]];
	   Write[channel,OutputForm["**HarrisSF DIAGNOSTICS OUTPUT"]];
	   Write[channel,OutputForm["Number of iterations  = "],j];
	   Write[channel,OutputForm["Last step in specfact = "],step];
	   Write[channel,OutputForm["Max[Abs[r-Abs[specfac(r)]^2]] = "],sferror];
	   Write[channel,OutputForm["non-analytic noise in specfac(r) = "],nan];
	];

        Return[{ SpectralFactor->f1,
                 SpectralFactorError->sferror,
                 SpectralFactorNonAnalyticNoise->nan,
                 HarrisSFStep->step,
                 HarrisSFIterations->j-1}]

   ]; (* end HarrisSF Module *)



(* Norms of lists of numbers --------------------- *)

NormSup[f_]:= Max[ Abs[ f ] ]; 

Norm2[f_]:= Abs[Sqrt[ Re[f.Conjugate[f]]/Length[f]] ];

Norm1[f_]:= Apply[Plus,Abs[f]]/Length[f];

MatrixNorm[f_]:=
   Module[{i},
      Return[Sqrt[ Sum[f[[i]].Conjugate[ f[[i]] ],{i,1,ndim}]
		/Length[f[[1]]]]];
   ];


(* Anopt optimality tests ------------------------- *)

OptimalityTest1[gk_]:=Module[{u},u=Max[gk]; (u-Min[gk])/u];
        
OptimalityTest2[aa_]:=
   Module[{tab,Wvector,i,Fcoef,WVcounter,windreal,ndim,
          Wnov,optmin,optpartial,temp,newmin,nsp},
	  ndim = Dimensions[aa][[1]];
	windreal = " ";
	If[ndim==1,
	    Return[
	      {WnoTest-> Abs[WindingNumber[aa[[1]]]-1],
	       WindReal->windreal}]];
	nsp = Length[aa[[1]]];
(*. . . Determine which partial is best . . .*)
	optmin = 0; optpartial=1; windreal="*";
	opttemp=WindingNumber[aa[[1]]];
	tab=0;

	Do[temp=WindingNumber[aa[[i]]];
	   newmin=Min[Abs[aa[[i]]]];
	   If[temp==1,
		If[newmin>optmin,optmin=newmin;
		optpartial=i;windreal=" ";
		opttemp=1;tab=tab+1]
		   ],{i,ndim}];
	
	If[optpartial != poptpartial,windreal="C"];
	poptpartial=optpartial;

	If[tab != ndim, windreal="D"];

        (*. . . Construct Winding number vector . . .*)

	Wvector=1.;
	Do[Wvector=Wvector+
		N[aa[[WVcounter]]/(aa[[optpartial]]+10^-15),16]
			,{WVcounter,1,ndim}];

        (*. . . Find the fourier coef of Wvector . . .*)
	Fcoef=N[Fourier[Wvector]/N[Sqrt[nsp],16]];

        (*. . . Add them up to find Winding number . . .*)
	Return[
	  {WnoTest->Sum[Abs[Fcoef[[WVcounter]]],{WVcounter,2,nsp/2 - 1}],
	  WindReal->windreal}];
   ];  (* end OptimalityTest2 *)


(* Arguments and winding number of (discrete) functions *)

WindingNumber[q_]:=
    If[TrueQ[Min[Abs[q]]<10^-(15)],99,
     Chop[N[Apply[Plus,Im[Log[RotateLeft[q]/q]]]/(2*Pi)]]];
   

ArgVector[tempvect_]:=
   Module[{v,i,angle,ArgV},
	v=tempvect;
	ArgV=Table[  
	 If[i==1,angle=Arg[v[[1]]+10^-15], 
	  If[Chop[Abs[v[[i]]]]==0 || Chop[Abs[v[[i-1]]]]==0,
		angle=angle+(50 N[2 Pi]),  
		angle=angle+
			Arg[v[[i]]/v[[i-1]] + N[Pi^-15]] ] ],
	{i,1,Length[v]}];
	Return[ArgV];
   ];



(* Changing the size of grid ---------------------------- *)

DoubleGrid[matrix_]:=
   Module[{ndim,nsp,i,fou,newfou,newmatrix,zz},

	If[Length[Dimensions[matrix]]==1,
		nsp=Length[matrix];
        	zz = Table[0.,{i,1,nsp-1}];
		fou = Fourier[matrix];
		newfou = Join[Take[fou,{1,nsp/2}],zz,
			Take[fou,{nsp/2,nsp}]] Sqrt[2.];
		newmatrix=InverseFourier[newfou],	

	{ndim,nsp} = Dimensions[matrix];
        zz = Table[0.,{i,1,nsp-1}];
        newmatrix=Table[ 
	  fou = Fourier[matrix[[i]]];
	  newfou = Join[Take[fou,{1,nsp/2}],zz,
			Take[fou,{nsp/2,nsp}]] Sqrt[2.];
	  InverseFourier[newfou],	
        {i,1,ndim}]
       	];
 
	Return[newmatrix];

   ]; (* end DoubleGrid Module *)




HalfGrid[matrix_]:=
   Module[{ndim,nsp,newmatrix,foucoef,i,newfou},
	If[Length[Dimensions[matrix]]==1,
		nsp=Length[matrix];
		foucoef=Fourier[matrix]/N[Sqrt[nsp]];
	  	newfou=Join[Take[foucoef,{1,Floor[nsp/4]}],
			Take[foucoef,{Floor[3 nsp/4 + 1],nsp}]];
		newmatrix=InverseFourier[newfou]* N[Sqrt[Length[newfou]]],

	{ndim,nsp} = Dimensions[matrix];
	newmatrix=Table[
		foucoef=Fourier[matrix[[i]]]/N[Sqrt[nsp]];
	  	newfou=Join[Take[foucoef,{1,Floor[nsp/4]}],
			Take[foucoef,{Floor[3 nsp/4 + 1],nsp}]];
	        InverseFourier[newfou] N[Sqrt[Length[newfou]]], 
		{i,ndim}]
	];

	Return[newmatrix];
   ]; (* end Halfgrid Module *)



Clean[function_]:=
   Module[{nd,n,j,i,fc,fc1,function2,zz},
	{nd,n} = Dimensions[function];
        zz = Table[0.,{i,1,n/2}];
        Do[
  	   (fc = Fourier[function[[j]]];
	    fc1 = Join[{fc[[1]]},zz,Table[fc[[i]],{i,n/2+2,n}] ];
	    function2[j]=InverseFourier[fc1];),
        	{j,1,nd}
        ];

        Return[Table[function2[j],{j,1,nd}]];
   ]; (* end Clean Module *)


(* Line search --------------------------------------  *)

(* This is a linesearch on the semixes 0<x, a real valued function F[t] 
of the real variable t is given by the user. Some of this code was
 taken from Numerical Recipes *)

(*---------old stuff 
GoldenLineSearch[F_]:=
    Module[{xa,xb,fa,fb,xmin,fmin}, 
	{{xa,fa},{xb,fb}} = BracketMin[F];
	{xmin,fmin} = Which[Abs[xa-xb]<0.002,{xa,fa},
			    Abs[xa-xb]>1000.,{xb,fb},
				        True, GoldenSection[F,xa,xb]
			];
	Return[{xmin,fmin}];
     ];
     


BracketMin[F_]:=
    Module[{val0,val,larget,factor,t,cond,difft,oldval,oldt},
(* . . .It is assumed that F[t] is decreasing on [0,smallt] 
        for some smallt . . . *)
	val0 = F[0.];
	val = F[1.];
	larget = TrueQ[val<val0];
	factor = If[larget,2.,0.5];
	t = 1.;
	cond = True;
	difft = 1.;
	While[cond&&(t>0.001)&& (t<1000),
		oldval = val;
		oldt = t;
		t = t*factor;
		val = F[t];
		cond = TrueQ[val<oldval];
		difft= Abs[t-oldt];
	     ];
	If[larget, ,
		(* else *)
		t = oldt;
		val = oldval;];
	Return[{{0.,val0},{t,val}}];
     ];

END OLD STUFF ---------- *)

GoldenLineSearch[F_]:=GoldenLineSearch[F,100.];

GoldenLineSearch[F_,maxt_]:=
    Module[{xa,xb,fa,fb,xmin,fmin},
	If[maxt<10^(-10) ,xmin=0.;fmin=F[0.] 
 	,

	{{xa,fa},{xb,fb}} = BracketMin[F,maxt];
	If[fa-fb==0,xmin=xa;fmin=fa
	,
	{xmin,fmin} = 
		If[TrueQ[xa==xb],{xb,fb},GoldenSection[F,xa,xb]];
	]];

	Return[{xmin,fmin}];
     ];

BracketMin[F_,maxt_]:=
    Module[{val0,val,larget,factor,t,bracketflag,oldval,oldt,
		tol=Min[0.01 maxt,.01]},
	val0 = F[0.];
	t1 = tol;
	val1 = F[t1];
(*      Make sure that F[t] decreases near 0:  *)
	If[val1 >= val0, Return[{{0,val0},{0,val0}}]];
	t = N[Min[1.,maxt]];
	val  = F[t];
	If[t==1. && t + tol < maxt,
		If[(val0 > F[t -  tol]>val)&&(val<F[t+tol]),
			Return[{{t,val},{t,val}}];
		  ];
	  ]; 
	larget = TrueQ[val<val0 && t < maxt];
	factor = If[larget,2.,0.5];
	bracketflag = False;
	While[ Not[bracketflag] && (t>tol) && (t<=maxt),
		oldval = val;
		oldt = t;
		t = Min[t*factor,maxt];
		val = F[t];
		bracketflag = TrueQ[val >= oldval];
	     ];
(*      If r.h.endpoint is maxt, then check for bracketed min 
	before returning output  *)
	If[maxt==oldt,
		t1 = maxt - 2.*tol;
		val1 = F[t1];
		If[val1 >= oldval,
			Return[{{oldt,oldval},{oldt,oldval}}],
			Return[{{0.,val0},{oldt,oldval}}];
		  ];
	  ];
(* All other cases are hopefully accounted for below *)
	If[larget   ,Return[{{0,val0},{t,val}} ],
		     Return[{{0,val0},{oldt,oldval}}]];
	];



GoldenSection[F_,xLeft_,xRight_]:=
   Module[{tol=0.0001,gold,gold2,x0,x1,x2,x3,f1,f2},
	gold = 0.5*(-1.+Sqrt[5.]);
	gold2 = 1. - gold;
	x0 = xLeft;
	x3 = xRight;
	x1 = x0+gold2*(x3-x0);
	x2 = x0+ gold*(x3-x0);
	f1 = F[x1];
	f2 = F[x2];
	While[Abs[x3-x0]> tol*(Abs[x1]+Abs[x2]),
		If[f2<f1,
			x0=x1;x1=x2;x2=gold*x1+gold2*x3;
			f1=f2;f2=F[x2],
		(* else *)
			x3=x2;x2=x1;x1=gold*x2+gold2*x0;
			f2=f1;f1=F[x1];
		   ];
	      ];
	If[f1<f2,Return[{x1,f1}],Return[{x2,f2}]];
   ];



End[] (* "`Private`" *)
EndPackage[]
