(* :Copyright (c) 1991-2001  J.W. Helton and O. Merino
    All rights reserved.
    There is to be no reproduction or commercial use of this program
    or any part thereof, without the written consent of the authors.
    The authors bear no liability for damages incurred by the use
    of this software.
*)

(* :Title: OPTDesign *)

(* :Author: Orlando Merino *)


(* :Context: OPTDesign` *)

(* :Package Version: 2.1 *)

(* :History:
    OPTDesign v 2.1   Jskud, Dec 2001

	repair interval check on asymptote of radius function
	add    symbolic check on asymptote of radius function
	replace PolynomialDivision (not in Standard Mma 4.0)
		with PolynomialQuotient and PolynomialRemainder
	replace NSolve[...] (which only lists multiple roots once in Mma 4.0)
		with N[Solve[...]]
	avoid implicit Null to avoid Mma 4.0 warnings about same
	use Scaled[] to define true Circles in PlotZP
	avoid assigning to both Trat and Trat[s_] so notebooks re-evaluate OK

    OPTDesign v.2.0   O. Merino  June 1998.
    OPTDesign v.1.0.  O. Merino, September 1992.
    Experimental version OPTDesign 0.3 O. Merino March 1992.
    Experimental Version DiskRHP 0.2
    by Orlando Merino and Julia Myers, February 1990.
    Experimental Version DiskRHP 0.1 by Orlando Merino, September 1990.

*)

(*??? :Keywords: OPTDesign *)

(*??? :Source: *)

(* :Warning:  GridW[[1]] = Infinity in OPTDesign[] *)

(* :Mathematica Version: 2.0 or higher *)

BeginPackage[ "OPTDesign`",{"NewtonInterpolant`",
		"Anopt`",
		"AnoptUtilities`",
		"ComplexRules`",
		"Utilities`FilterOptions`",
		"Graphics`Graphics`",
		"Graphics`Graphics3D`",
		"Graphics`ParametricPlot3D`",
		"Graphics`Animation`"}]

OPTDParametrize::usage="OPTDParametrize[T] produces a list of pairs
of the form {wi,Ti} where wi is the i-th gridpoint on the w axis, and
Ti is the i-th value of the list T.  OPTDParametrize[T,Ngrid -> n, GridSpread ->b] specifies the options for the grid."

EnvelopePlot3D::usage="EnvelopePlot3D[Radius->r0,Center->k0,CloseLoop->rat] \n
produces a 3D plot of the envelope of constraints.  The input ClosedLoop->rat is optional.  Here r0, k0 are functions of w and rat is a rational in s.";

ClosedLoop::usage="Option for EnvelopePlot3D.  It specifies a rational in s, the closed loop transfer function of the system being analysed.";

xLarge::usage="Symbol used for calculation of certain limits."

Grid::usage="Grid[ ], Grid[opts] Generates a list with n samples w on the real axis, begining with the point (+) Infinity, then descending order.
The list is generated by mapping n equally sapeced points on the unit circle to the imaginary axis with the transformation: z (on the unit circle) -> I w = b(1+z)/(1-z). Half +1 of the gridpoints are between -b and b. The parameters n and b are set by the options Ngrid and GridSpread."

Discretize::usage="Discretize[H], Discretize[H,opts] samples the function H[w] at the (real) gridpoints w produced by Grid[].  H[] may be defined by a single formula, or by several by using If[] or Which[]. Examples of input: \n
H[w_]:= ( - w^2 + 1)/( I w + 2) \n
H[w_]:=If[Abs[w]<1,w,1/w] \n
H[w_]:=Which[Abs[w]<1,w, Abs[w]<2,1/w, True, 1/w^3]"

OPTDesign::usage = "OPTDesign[p,opts] minimizes a performance index, function of a designable closed loop transfer function T which internally stabilizes the closed loop feedback system (S) with plant p.  Here p is a rational function.  The output parameters gammastar, flat, grAlign and ned are displayed on the screen when the run is over. Other output includes the optimal T, and related functions such as L = T/(1-T), C = T/(P(1-T)).  The functions T, L, and C are given as lists of complex numbers, samples on a grid on the jw axis. See Grid[], and Options[OPTDesign]"

BodeMagnitude::usage="BodeMagnitude[list], BodeMagnitude[list,opts] Produces Magnitude Bode plot of list (e.g., a list generated by OPTDesign). Options are the same as thos of ListPlot.  BodeMagnitude utilizes functions from Mma. package Graphics.  Code: is a modification of Graphics`LogLogListPlot."

BodePhase::usage="BodePhase[list], BodeMagnitude[list,opts]. Produces Phase Bode plot of list (e.g., a list generated by OPTDesign). Options are the same as ListPlot. Utilizes function Grid[], and functions from Mma. package Graphics.  Code: is a modification of Graphics`LogLogListPlot."

ListPolarPlot::usage="ListPolarPlot[fun],ListPolarPlot[fun,opts] Produces polar plot from a list with complex entries (e.g., a list generated by OPTDesign). Options are the same as ListPlot."

Nichols::usage="Nichols[list,opts] Produces Nichols plot of list with complex entries (e.g., a list generated by OPTDesign)."

LimitInfinite::usage = "LimitInfinite[f,x] computes the limit of a rational function, when x->Infinite.  f may be given as a combination of simpler rational functions."

DiscreteInvLaplace::usage = "DiscreteInvLaplace[T, maxt, gridt] produces the plot of the discrete inverse Laplace transform of T.  T is a list resulting from sampling a function on the jw axis on gridpoints generated with the function Grid[]. The time interval is [0,maxt]. The number of samples in [0,maxt] is gridt.  Results not guaranteed with lists T that result from sampling a rational function of degree 2 or less."

StepResponse::usage="StepResponse[T,maxt,gridt],StepResponse[T,maxt,gridt,opts] produces a plot of the inverse Laplace transform of T(s)/s.  The input T is given as a discrete function on gridpoints on the jw axis.  "

CancelZP::usage=" CancelZP[rational,term,x] simplifies common factors of type term in the variable x if the remainders of division by term of the numerator and denominator of rational are smaller than Tolerance"

GeneratePlot::usage="Option for some plotting functions in Pkg. OPTDesign. GeneratePlot->True produces a plot as output, and GeneratePlot->False produces a list of coordinate pairs"

NoPoleZeroCancelation::usage="Option for OPTDesign[]. Specifies a pure function fun of complex argument and True/False value. Example: NoPoleZeroCancelation->  (Re[#]>-1&) prevents pole zero cancelation in the product P*C on the region defined by the function.  In this example, the region is {s : Re[s]>-1}."


OV::usage = "OV is the Optimal Value,  or best performance possible for the given specs., produced as output of OPTDesign[]."

ned::usage = "ned is the Numerical Error Diagnostic, an measure of the computational error produced by OPTDesign[].  ned tends to be large when either the grid size is small or the functions involved are not smooth enough."

OT::usage = "OT is the Optimality Test produced by OPTDesign[], an indirect measure of how close to optimal the solution T is."

T::usage = "a list, output to OPTDesign. T is the closed loop transfer function, a list of values on a grid produced by OPTDesign[]."

L::usage = " a list, output to OPTDesign.\L is a list of gridvalues of the open loop transfer function.  It satisfies L = P C = T/(1-T)."

P::usage = " a list, output to OPTDesign. P is the plant values sampled a grid produced by OPTDesign[]."

Co::usage = " a list, output to OPTDesign. Co is the compensator's list of values on a grid produced by OPTDesign[]."

Ngrid::usage = "Ngrid is an option of OPTDesign[].  Specifies the number of gridpoints."

Nsmth::usage = "Nsmth is an option of OPTDesign[].  It is the parameter t in Smooth[f,t] as used by OPTDesign internally."

Cdecay::usage = "Cdecay is an option of OPTDesign[].  It specifies the relative degree of the desired compensator if it is rational, or decay rate. \n"

StopTol::usage = "StopTol is an option of OPTDesign[].  It is the tolerance for stopping computation."

Screen::usage = "Screen is an option of OPTDesign[].  It is the variable for printing on the screen the internal diagnostics of OPTDesign during a run."

GridSpread::usage = "GridSpread is an input of OPTDesign[] and is communicated as OPTDesign[......, GridSpread->value].  It specifies the parameter b in the selection of the grid.  See Grid[]."

PoleLocation::usage = "PoleLocation is an input of OPTDesign[] and is communicated as OPTDesign[......, PoleLocation->value].  It is a negative integer gives pole location of stable rational interpolants. "

DeltaP::usage = "DeltaP is an input of OPTDesign[] and is communicated as \n
OPTDesign[......, DeltaP->value]. Is a radius about poles of the plant on the jw axis.  OPTDesign sets radius[w] = alphap*/Abs[P[j w]] whenever Abs[w-wp]<deltap overriding the radius function given by the user in the given interval. "

DeltaZ::usage = "DeltaZ is an input of OPTDesign[] and is communicated as \n
   OPTDesign[......, DeltaZ->value]. Is a radius about zeros of the plant on the jw axis.  OPTDesign sets radius[w]:=alphaz*Abs[P[j w]] whenever Abs[w-wz] < deltaz overriding the radius function given by the user in the given interval. "

Radius::usage="Option for OPTDesign.  A function of w, positive valued."

Center::usage="Option for OPTDesign.  A function of w."

Performance::usage="option for OPTDesign.  Performance -> DiskPerf indicates that the performance to be used is Abs[(center - z)/radius]^2, where z represents the designable closed loop transfer function.  Another example is Performance -> W[1] Abs[z]^2 + W[2] Abs[1-z]^2, where W[1],W[2] are weight functions specified with the option Weights.  The weights can also be given explicitly as a formula of the type Abs[rational in w]^2"

Weights::usage="option for OPTDesign.  Weights->{wght1,wght2,...} specifies that in the formula of the performance, W[1] represents the function wght1[ ], where wght1[ ] is a function of frequency w."

W::usage="Symbol used in option Performance to denote weights. "

ValueAtInfinity::usage="Option for Discretize, value of the function being sampled at infinite frequency "

s::usage="Laplace domain variable."

t::usage="Time domain variable."

T1::usage="T1 is a list, output to OPTDesign.  It is the gridvalues of the optimal solution to the uncontrained optimization calculated by OPTDesign.  It is related to the optimal designable transfer function T through a formula of the type T = A + B T1, where A and B are functions that codify the internal stability requirements as interpolation conditions."

InitialT1::usage="Option of OPTDesign.  The user may specify an initial value for T1 as a way to speed up calculations or avoid numerical trouble in the optimization routine Anopt called by OPTDesign, which is based on an iterative algorithm.  See option T1"



(* eliminate usage note
*)
Apart2::usage="Apart2[rat,var] does a partial fraction decomposition of \n
the rational function rat in the variable var"

(* eliminate usage note
*)
Apart2Tol::usage="Option for Apart2[].  It specifies the tolerance \n
to distinguish between zeros (or poles) when using Apart2[]."

CFApproximant::usage="CFApproximant[fdisc,dn,dd,var,opts] produces a
rational expression in the variable var with no poles in the closed unid disk in the complex plane, which is a model that approximates the data fdisc on the circle in the norm of the supremum.  fdisc is a list of samples on a grid of equally spaced points on the unit circle.  dn is the degree of the numerator, dd is the degree of the denominator of the output. The algorithm is by L. N. Threfethen.";

HankelSize::usage="Option of CFApproximant[].  Specifies the size of the Hankel matrix"

RationalModel::usage="RationalModel[], produces a rational expression, a model for a discrete function generated by OPTDesign[]."

StableFit::usage="StableFit[f,wpts], produces a rational expression, a model for a discrete function generated by OPTDesign[]."

SampledFunction::usage="Option for RationalModel.";

Var::usage="Option for RationalModel, StableFit  and Sample.";

FrequencyBand::usage="Option for BodeMagnitude and BodePhase functions.";

GridPoints::usage="Option for some function of the package OPTDesign.
It is the list of w points for sampling functions."

Method::usage="Option for RationalModel, CFApproximant.";

InterpolantFunctions::usage="Option for RationalModel.";

DegreeOfDenominator::usage="Option for RationalModel.  Specifies the degree
of the denominator of the output.";

PlotZP::usage="PlotZP[ratexpr] produces a plot with the zeros and poles of the rational expression ratexpr.  Options should be used to specified the variable if it is not the default value.";

EnvelopePlot::usage="EnvelopePlot[] plots the requirements envelope
(as defined by a radius and center functions) and the modified
envelope obtained with automatic smoothing.";

EnvelopeLogPlot::usage="EnvelopeLogPlot[] plots the profile of the smoothed envelope (as defined by a radius and center functions)."

ModifiedCenter::usage="Option for EnvelopePlot.";

ModifiedRadius::usage="Option for EnvelopePlot.";

DiskPerf::usage="See usage note for Performance"

RHPListPlot3D::usage="RHPListPlot[T] Produces a 3D plot of the list T,
where one of the axes is the w axis"

Nyquist::usage="Nyquist[x]";

NyquistListPlot::usage="NyquistListPlot[L]";

NyquistPlot::usage="NyquistPlot[ L[s] ]";

Options[NyquistPlot]=Join[{FrequencyBand -> {0.,10.}},Options[ParametricPlot]];
SetOptions[NyquistPlot,PlotRange -> All];

Options[Nichols] = {SamplingPoints -> 100,FrequencyBand -> {0.01,100.}};

Options[RHPListPlot3D] = {GridPoints :> Grid[]};

Options[EnvelopePlot3D] =
	{Radius->Automatic,
	Center->Automatic,
	ClosedLoop->None,
	PlotJoined -> False,
	FrequencyBand->{0.,5.}};

Options[EnvelopePlot]=
		{Ngrid->128,GridSpread->1.,Nsmth->1,
                 GridPoints :> Grid[Ngrid->128,GridSpread->1.],
		 Radius->Automatic,Center ->Automatic,
		ModifiedRadius->Automatic,ModifiedCenter ->Automatic,
		 FrequencyBand->{0.,10.}};

Options[EnvelopeLogPlot]=
		{Ngrid->128,GridSpread->1.,Nsmth->1,
                 GridPoints :> Grid[Ngrid->128,GridSpread->1.],
		 Radius->Automatic,Center ->Automatic,
		ModifiedRadius->Automatic,ModifiedCenter ->Automatic,
		 FrequencyBand->{0.,10.}};

Options[PlotZP]={Var->s};

Options[Sample]={Ngrid->128,GridSpread->1.0,Var->s};

Options[RationalModel]=
	{SampledFunction->T1,
         GridPoints :> Grid[Ngrid->128,GridSpread->1.],
	 GridSpread->1.,
	 (* Var-> s, *)
	 Method->CFApproximation,
	 InterpolantFunctions->{0&,1&},
	 DegreeOfDenominator->Automatic};

Options[StableFit] = { Var -> s };

Options[CFApproximant]={HankelSize->Automatic,Var->s};

(* move to private?
*)
Options[Apart2]={Apart2Tol->10^-6};

Options[CancelZP]={Tolerance->10^(-4)}

Attributes[CancelZP]={Listable};

Options[Grid]={Ngrid->128,GridSpread->1.0}

Options[Discretize]={Ngrid->128,GridSpread->1.0,GridPoints :> Grid[Ngrid->128,GridSpread->1.],
			ValueAtInfinity->Automatic}


Options[OPTDesign]         = {
			    GridPoints :> Grid[Ngrid->128,GridSpread->1.],
			    Ngrid->128,
			    DeltaZ->0.001,
			    DeltaP->0.001,
			    GridSpread->1.0,
			    PoleLocation->-2.0,
                            Nsmth->1,
			    Cdecay->0,
			    StopTol->0.0001,
                            Screen->False,
			    NoPoleZeroCancelation->( Re[#] > -10^(-6) & ),
			    Radius->None,
			    Center->None,
			    Performance-> DiskPerf,
			    Weights->None,
			    InitialT1->Automatic};
Options[DiscreteInvLaplace]    = {GridSpread -> 1.,GeneratePlot->True};
Options[StepResponse]={GridSpread->1.,GeneratePlot->True};
Options[BodePhase]= {FrequencyBand -> {0.01,100},
			PlotRange -> Automatic,
			SamplingPoints -> 100,
       			GridPoints :> Grid[] };
Options[BodeMagnitude]={
			GridPoints :> Grid[],
			Var->s,
			FrequencyBand->{0.01,100.},
			PlotRange -> Automatic};

SetOptions[Anopt,
   	QuickResults -> False,
   	OptimalityTol -> Automatic,
   	Subproblem -> "DiskIteration",
   	DebugMe ->False,
        Iterations->10,
        GridDouble->False,
        MaxGridSize->1024,
        Diagnostics->0,
        Diagnostics2->0,
        RunID->"none",
	DualEigenfunction->Automatic,
	OT3Check->True,
        DoLineSearch ->True,
(*. . . Smoothing . . .*)
	InitialSmooth->None,
	UpdateVectorSmooth->None,
	SmoothType->"A"
	];



Begin[ "OPTDesign`Private`" ]

(* Greetings and salutations *)
         Print["OPTDesign 2.1 (c) 1991-2001 \n
	 	J.W. Helton and O. Merino.  All rights reserved."];

(* Auxiliary functions and algebra and static variables *)


	Unprotect["Less","LessEqual","Abs",
		  "Greater","GreaterEqual","Conjugate"];
	Less[Abs[xLarge],anything___]:=False;
	LessEqual[Abs[xLarge],anything___]:=False;
	Less[xLarge,anything___]:=False;
	LessEqual[xLarge,anything___]:=False;
	Greater[anything___,Abs[xLarge]]:=False;
	GreaterEqual[anything___,Abs[xLarge]]:=False;
	Greater[anything___,xLarge]:=False;
	GreaterEqual[anything___,xLarge]:=False;
	Conjugate[xLarge] = xLarge;
	Protect["Less","LessEqual","Abs",
		"Greater","GreaterEqual","xLarge","Conjugate"];

Unprotect[Plus]
a_ + 0. := a
Protect[Plus]

(* %%%%%%%%%%%%%%%%%%%%%  OPTDesign BEGINS %%%%%%%%%%%%%%%%%%% *)


OPTDesign[plant_,opts___Rule]:=
   Module[ {K1a, K1b, K1, nexp,coefnum, x, coefden, K1infinity,
   	    R1a, R1b, R1c, R1d, R1, term,
            R1infinity, g, count,plantcoeff,fcancel,A,B},
      tol          = StopTol      /. {opts} /. Options[OPTDesign];
      ngrid        = Ngrid        /. {opts} /. Options[OPTDesign];
      nsmth        = Nsmth        /. {opts} /. Options[OPTDesign];
      cdecay       = Cdecay       /. {opts} /. Options[OPTDesign];
      iv           = Screen       /. {opts} /. Options[OPTDesign];
      gridspread   = GridSpread   /. {opts} /. Options[OPTDesign];
      poleLocation = PoleLocation /. {opts} /. Options[OPTDesign];
      radius  	   = Radius 	  /. {opts} /. Options[OPTDesign];
      center	   = Center	  /. {opts} /. Options[OPTDesign];
      deltap   	   = DeltaP 	  /. {opts} /. Options[OPTDesign];
      deltaz       = DeltaZ 	  /. {opts} /. Options[OPTDesign];
      weights 	   = Weights 	  /. {opts} /. Options[OPTDesign];
      perf         = Performance  /. {opts} /. Options[OPTDesign];
      f0           = InitialT1    /. {opts} /. Options[OPTDesign];
      fcancel	   = NoPoleZeroCancelation /. {opts} /. Options[OPTDesign];


      If[iv,
         Print[" ------------------------------------------------" ];
         Print["                  OPTDesign                      " ];
         Print[" ------------------------------------------------\n"];
         Print[" Input to OPTDesign: \n " ];
	 Print[" plant[s]   = ",plant[s]  ];
         Print["               "];
         Print["    Gridspread            ", gridspread   ];
         Print["    PoleLocation          ", poleLocation ];
         Print["    Cdecay                ", cdecay       ];
         Print["    StopTol               ", tol          ];
         Print["    Ngrid                 ", ngrid        ];
         Print["    Nsmth                 ", nsmth        ];
         Print["    NoPoleZeroCancelation ", fcancel      ];
         Print["    Radius                ", radius       ];
         Print["    Center                ", center       ];
         Print["    Performance           ", perf         ];
         Print["    Weights               ", weights      ];
         Print["    InitialT1             ", f0           ];
         Print["    DeltaP                ", deltap       ];
         Print["    DeltaZ                ", deltaz       ];
         Print["    Screen                ", iv           ];
         Print[" \n                       "               ];
      ]; (* end If *)

(* %%%%%%%%%% check and reset input, initialize %%%%%%%%%%%%%%%%%%%%%%%% *)

	If[Not[TrueQ[poleLocation < 0]],
	      Print["Value set for option PoleLocation is not valid."];
	      Abort[]
	      ];
	If[Not[TrueQ[gridspread >0]],
	      Print["Value set for option GridSpread is not valid."];
	      Abort[]
	      ];
	If[Not[IntegerQ[cdecay]&&(cdecay >=0 )],
	      Print["Value set for option Cdecay is not valid."];
	      Abort[]
	      ];
	If[Not[TrueQ[tol>0]],
	      Print["Value set for option StopTol is not valid."];
	      Abort[]
	      ];
	If[Not[IntegerQ[ngrid]&&(ngrid >=0 )],
	      Print["Value set for option Ngrid is not valid."];
	      Abort[]
	      ];
	If[Not[IntegerQ[nsmth]&&(nsmth >=0 )],
	      Print["Value set for option Nsmth is not valid."];
	      Abort[]
	      ];
	If[Not[TrueQ[deltap>0]],
	      Print["Value set for option DeltaP is not valid."];
	      Abort[]
	      ];
	If[Not[TrueQ[deltaz>0]],
	      Print["Value set for option DeltaZ is not valid."];
	      Abort[]
	      ];
	If[Not[TrueQ[(iv==True)||(iv==False)]],
	      Print["Value set for option Screen is not valid."];
	      Abort[]
	      ];

(* check that plant is ok... *)
	plants = Together[plant[s]];
	{numplants,denplants}={Numerator[plants],Denominator[plants]};
	If[Not[TrueQ[PolynomialQ[numplants]&&
		     PolynomialQ[denplants]]],
		Print["The plant is not a rational function"];
		Abort[],
	   (* else *)
	        If[TrueQ[Exponent[numplants,s]>Exponent[denplants,s]],
		     Print["The plant is not a proper rational function"];
		     Abort[],
		   (* else *)
		     maxim = Max[Abs[Im[CoefficientList[numplants,s]]],
		                 Abs[Im[CoefficientList[denplants,s]]]];
		     If[Not[TrueQ[Chop[maxim]==0]],
		           Print["Plant coefficients are not real numbers"];
			   Abort[];
			   ];
		  ];
	   ];

(* check performance vs. presence of radius and center functions as input *)
	nondiskperf = Not[TrueQ[perf == DiskPerf]];
	diskterm = Not[FreeQ[perf,DiskPerf]];
	If[TrueQ[diskterm && (center==None || radius == None)],
		Print["Center and Radius options must specify functions"];
		Abort[]];

(* check weights given as input vs. performance *)
	nwghts = 0;
	weightsmth = {};
(* REPLACE BELOW THIS BY MEMBERQ[ETC]}...
	While[Not[TrueQ[D[perf,W[nwghts]]==0]],nwghts++];
	nwghts = nwghts-1;
	If[iv,Print["Number of weight functions: ",nwghts]];
		If[nwghts>0 && ( nwghts > Length[weights] ),
		Print[nwghts," weight functions must be specified with Weights->{wght1,wght2,...}"];
		Abort[]];
------- REPLACE ABOVE *)


(* initialize *)
	If[TrueQ[f0==Automatic],
		f0={Table[0.,{ii,1,ngrid}]}];


(* %%%%%%%%%%%%%%%%%%%%%  POLES AND ZEROS OF PLANT %%%%%%%%%%%%%%%%%%% *)

(* Find out degree of the plant  *)

      {pnum,pden}={Numerator[Together[plant[s]]],
      			Denominator[Together[plant[s]]]};
      ListZeros = If[TrueQ[Exponent[pnum,s]==0],
	{}, (s /. N[Solve[pnum==0.,s]])];
      ListPoles = s /. N[Solve[pden==0.,s]];
      degpnum   = Length[ListZeros];
      degpden	= Length[ListPoles];
      reldegp= degpden-degpnum;
      plantcoeff= Coefficient[pnum,s,degpnum]/Coefficient[pden,s,degpden];


(* Check for poles of the plant on the jw axis. *)
      AxisPoles = {};
      Module[{i},
         Do[
            If[ Abs[Re[ListPoles[[i]]]]<10^(-10),
	       AxisPoles = Join[AxisPoles,{ListPoles[[i]]}]
            ],
   	 {i,1,Length[ListPoles]}]
      ];
      nAxisPoles = Length[AxisPoles];

(* Check for zeros of the plant on the jw axis. *)
      AxisZeros = {};
      Module[{i},
         Do[
  	    If[ Abs[Re[ListZeros[[i]]]]<10^(-10),
	       AxisZeros = Join[AxisZeros,{ListZeros[[i]]}]
            ],
	 {i,1,Length[ListZeros]}]
      ];
      nAxisZeros = Length[AxisZeros];

      If[iv && (naxiszeros>0),
 	 Print[" DeltaZ     = ",deltaz ];
      ];
      If[iv && (naxispoles>0),
 	 Print[" DeltaP     = ",deltap];
      ];
      If[iv,
         Print[" ---------------------------------" ];
         Print["Poles of the plant  = ",ListPoles ];
	 Print["Zeros of the plant  = ",ListZeros ];
	 Print["jw-Axis Poles       = ",AxisPoles ];
	 Print["Number jw-axis poles= ",nAxisPoles];
	 Print["jw-Axis Zeros       = ",AxisZeros ];
	 Print["Number jw-axis poles= ",nAxisZeros];
      ];

(* %%%%%%%%%%%%%%%%%%%%%  DECAY RATE    %%%%%%%%%%%%%%%%%%% *)

   (* Damping *)
   (* Cdecay = compensator's decay rate *)
      Pdecay = reldegp;	(* plant's decay rate *)
      Tdecay = Pdecay+cdecay;  	(* decay for T=closed loop t.f. *)

(* %%%%%%%%%%%%%%%%%%%%%  INTERPOLATION  %%%%%%%%%%%%%%%%%%% *)

   (* Interpolation conditions; *)
(* ====================================================== *)
   (*   determine order of zeros and poles. *)
(* ====================================================== *)
      X = ListZeros+0.;
      dataz = {};
      While[ Not[X == {}],
      	xfirst = X[[1]];
	posxfirst = Position[X,xfirst];
	dataz = Join[dataz,
		{Join[{xfirst},Table[0.,{Length[posxfirst]}]] }  ];
	X = Delete[X,posxfirst];
	];

      X = ListPoles+0.;
      datap = {};
      datap0= {};
      While[ Not[X == {}],
      	xfirst = X[[1]];
	posxfirst = Position[X,xfirst];
	datap = Join[datap,
		     {Join[{xfirst,1},Table[0.,{Length[posxfirst]-1}]] }  ];
	datap0 = Join[datap0,
		     {Join[{xfirst},Table[0.,{Length[posxfirst]}]] }  ];
	X = Delete[X,posxfirst];
	];

(* ====================================================== *)
(*   filter out stable poles and zeros   *)
(* ====================================================== *)
	X = {};
	Do[If[fcancel[datap[[i,1]]],X=Join[X,{datap[[i]]}]],
		{i,1,Length[datap]}];
	datap = X;
	X = {};
	Do[If[fcancel[datap0[[i,1]]],X=Join[X,{datap0[[i]]}]],
		{i,1,Length[datap0]}];
	datap0 = X;
	X = {};
	Do[If[fcancel[dataz[[i,1]]],X=Join[X,{dataz[[i]]}]],
		{i,1,Length[dataz]}];
	dataz = X;
(* ====================================================== *)
(*  Interpolation for internal stability   *)
(* ====================================================== *)


      If[iv,
         Print[" ---------------------------------" ];
 	 Print["INPUT to NewtonInterpolant[] :"];
	 Print["{unstable pole,order} = ",
	 	Table[{datap[[i,1]],Length[datap[[i]]]-1},
			{i,1,Length[datap]}]];
	 Print["{unstable zero,order} = ",
	 	Table[{dataz[[i,1]],Length[dataz[[i]]]-1},
			{i,1,Length[dataz]}]];
Print["A[s]=NewtonInterpolant[",Join[dataz,datap],",s,RelativeDegree->",Tdecay,"PoleLocation->",poleLocation,"]"];
Print["B[s]=NewtonInterpolant[",Join[dataz,datap0],",s,RelativeDegree->",Tdecay,"PoleLocation->",poleLocation,"]"];
      ];
   (* assign the interpolating functions A, B, in equation T = A + B T1  *)
      AA[s_] = NewtonInterpolant[Join[dataz,datap],s,
   			RelativeDegree->Tdecay,
			PoleLocation->poleLocation];
      BB[s_] = NewtonInterpolant[Join[dataz,datap0],s,
      			RelativeDegree->Tdecay,
			PoleLocation->poleLocation];

      A[x_]= AA[I x];
      B[x_]= BB[I x];

 	Print["\nParameterization for internal stability:"];
	Print["  T = A + B * T1, where T1 is RHP stable and"];
	Print["A[s]=",AA[s]];
	Print["B[s]=",BB[s]];
	Print["  "];

(* %%%%%%%%%%%%%%%%%  PERFORMANCE FUNCTION %%%%%%%%%%%%%%%%%%%%%%% *)

Print["Processing performance function..."];

     If[iv,Print["Performance as given in input:"]];
     If[iv,Print["Performance->",perf]];
If[Head[perf]=!=List,
     If[nondiskperf,
	perf = perf /. z -> AA[I Global`w] + BB[I Global`w] z[1];
	perf = ComplexCoordinates[perf]/.Conjugate[Global`w]->Global`w;
	perf = perf /. {Global`w-> -I gridspread (1.+e)/(1.-e)};
	perf = Together[perf];
	coefl = CoefficientList[perf,{z[1],Conjugate[z[1]]}];
	coefl = CancelZP[coefl,e-1,e];
	coefl = CancelZP[coefl,e+1,e];
	{nc,mc} = Dimensions[coefl];
	pwsz = Table[z[1]^ii,{ii,0,nc-1}];
	pwscz= Table[Conjugate[z[1]]^ii,{ii,0,mc-1}];
	perf = Table[coefl[[ii]].pwscz,{ii,1,nc}].pwsz;
       ];
     If[diskterm,
     	perf = perf /. DiskPerf->Abs[a[1] a[2] - a[1] z[1]]^2];
     If[diskterm,
        perf = perf /. {W[j_]:>a[j+2]},
	perf = perf /. {W[j_]:>a[j]}];
	,
(* else if head of perf is not a list: *)
	perfoo     = perf[[1]];
	perfterms  = perf[[2]];
Print["HEY1",perfterms];
	ntermsperf = Length[perfterms];
	perfterms  = perfterms /. z -> AA[I Global`w] + BB[I Global`w] z[1];
Print["HEY2",perfterms];
	perfterms  = Map[ComplexCoordinates,perfterms] /.
				Conjugate[Global`w]->Global`w;
	perfterms  = perfterms  /. {Global`w-> -I gridspread (1.+e)/(1.-e)};
Print["HEY3",perfterms];
	perfterms1 = perfterms /. z[1] -> 0;
	perfterms2 = D[perfterms,z[1]] /. z[1] -> 0;
Print["HEY4",perfterms];

	perfterms1  = Simplify[Chop[Together[perfterms1]]];
	perfterms2  = Simplify[Chop[Together[perfterms2]]];

Print["HEY5",{perfterms1,perfterms2}];

	val0 = Join[AxisPoles,AxisZeros];
	val = Join[{-1,1},(val0-1)/(val0+1)];
	Do[perfterms1 = CancelZP[perfterms1,e - val[[i]],e],{i,1,Length[val]}];
	Do[perfterms2 = CancelZP[perfterms2,e - val[[i]],e],{i,1,Length[val]}];

Print["HEY6",{perfterms1,perfterms2}];

	uu = Sequence[];
	Do[uu = Sequence[perfterms1[[i]]+ perfterms2[[i]] z[1],uu],
						{i,1,Length[perfterms1]}];

	perf = perfoo[uu];
	]; (* endif *)

     If[PolynomialQ[perf,{Conjugate[z[1]],z[1]}],
     		perf = Collect[perf,{Conjugate[z[1]],z[1]}]
		];

     If[iv,Print["Performance after change of coordinates and notation \n
     (Here z[1] represents the free variable T1):"]];
     If[iv,Print["Performance->",perf]];

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*)



(* %%%%%%%%%% If there is no DiskPerf term in performance, skip
		all computations related to it %%%%%%%%%%% *)
	If[diskterm,
	If[iv,
         Print[" ---------------------------------" ]];

(* %%%%%% CHECK CENTER AND RADIUS NEAR AXIS POLES AND ZEROS %%%%%%% *)
	smalld = 0.001;
	pertw = {-3.,-2.,-1.,1.,2.,3.}*smalld;
	alphazero = {};
	alphapole = {};
	Do[wzero = AxisZeros[[i]];
		wpertw = wzero + pertw;
		vals = Table[radius[wpertw[[j]]]/Abs[plant[I wpertw[[j]]]]
		,{j,1,Length[wpertw]}];
		If[(Max[vals]-Min[vals] < 0.000001)&&(Max[vals]>0.001),
			alphazero = Join[alphazero,{vals[[1]]}],
		(* else *)
			Print[" \n
			Possible error in input to OPTDesign: \n
			The radius function does not appear to have \n
			the form const*Abs[plant[I w]] for w near ",
			wzero,", which is a zero of the plant."];
			Abort[]
			(* STOP *)
		  ];
	   ,{i,1,nAxisZeros}];
	 Do[wpole = AxisPoles[[i]];
		wpertw = wpole + pertw;
		vals = Table[radius[wpertw[[j]]]*Abs[plant[I wpertw[[j]]]]
		,{j,1,Length[wpertw]}];
		If[(Max[vals]-Min[vals] < 0.000001)&&(Max[vals]>0.001),
			alphapole = Join[alphapole,{vals[[1]]}],
		(* else *)
			Print[" \n
			Possible error in input to OPTDesign: \n
			The radius function does not appear to have \n
			the form const/Abs[plant[I w]] for w near ",
			wpole,", which is a pole of the plant."];
			(* Abort[] *)
			(* STOP *)
		  ];
	   ,{i,1,nAxisPoles}];

Off[Infinity::indet];
	   asymptotic=Limit[ComplexCoordinates[radius[xLarge] xLarge^Tdecay],
	   		xLarge->Infinity];
On[Infinity::indet];

	   If[TrueQ[10^-10 < asymptotic && asymptotic < 10^10],
	     (* then *)
	    	If[iv,Print["Limit[radius[w]/w^Tdecay,
	     			w->Infinity]=",asymptotic]],
	     (* else *)
		Block[{issue},

		(* NB: Mma converts linebreaks and subsequent tabs in
		   string constants to a single blank; use to get long
		   messages with short source lines *)

			issue =

			"Possible error in input to OPTDesign:\n The
			radius function does not appear to have the
			asymptotic form const/w^n for w near Infinity.
			Here, n = dp + dc, the sum of the relative
			degrees of the plant and the compensator (or
			roll-off).  The evaluated asymptotic value is "
			;

			If [ NumberQ[asymptotic] || asymptotic == Infinity
			, (* then *)
				Print[ issue, asymptotic ] ;
			, (* else *)
				Print[ issue, "symbolic" ] ;
			] ;
		];
		(* Abort[] *)
	    ];


	If[iv && (nAxisZeros > 0),Print["alphazero = ",alphazero]];
	If[iv && (nAxisPoles > 0),Print["alphapole = ",alphapole]];
	If[iv,Print["Asymptotically, Closed loop T~",asymptotic,"/w^",Tdecay]];

(* %%%%%%%%%%%%%%%%%%%%  center function  %%%%%%%%%%%%%%%%%%%% *)

(* Assign the center function K1 *)

		pp1 = (1.-A[x1])/B[x1];
		Do[pp1 =
		CancelZP[pp1,x1+I*ListPoles[[kk]],x1],{kk,1,Length[ListPoles]}];
		pp2 = -A[x1]/B[x1];
		Do[pp2 =
		CancelZP[pp2,x1+I*ListZeros[[kk]],x1],{kk,1,Length[ListZeros]}];

(* Compute the value of K1 at infinity *)
		K1infinity = LimitInfinite[-A[x]/B[x],x];
	If[iv,
	Print["Value at infinity of reparameterized center:",
		K1infinity]];


		K1[x_]:=Which[ x == Infinity, 	K1infinity,
		       Min[Abs[AxisZeros-I x]]<deltaz,pp2/.x1->x,
		       Min[Abs[AxisPoles-I x]]<deltap,pp1/.x1->x,
		       True,	(center[x]-A[x])/B[x] ];

(* %%%%%%%%%%%%%%%%%%%%  RADIUS function  %%%%%%%%%%%%%%%%%%%% *)

(* Assign the radius function R1 *)
(* Compute the value of R1 at infinity *)

		   R1infinity = asymptotic *
		     Abs[Limit[ComplexCoordinates[xLarge^Tdecay B[xLarge]],
	   		xLarge->Infinity]];
	If[Not[NumberQ[R1infinity]],
		R1infinity = asymptotic *Abs[xLarge^Tdecay B[xLarge] /.
				xLarge -> 100000000.]];

	If[iv,
	Print["Value at infinity of reparameterized radius:",R1infinity]];

(* some algebra to deal with possible poles and zeros on the axis *)

		ppb = plant[I x1]/B[x1];
		Do[ppb =
		CancelZP[ppb,x1+I*ListZeros[[kk]],x1],{kk,1,Length[ListZeros]}];
		ppa = plant[I x1] B[x1];
		Do[ppa =
		CancelZP[ppa,x1+I*ListPoles[[kk]],x1],{kk,1,Length[ListPoles]}];
(* define the new radius *)
		R1[x_]:=Which[ x == Infinity, 	R1Infinity,
		    Min[Abs[AxisZeros-I x]]<deltaz,
		    			alphazero[[1]]*Abs[ppb /.x1->x],
		  Min[Abs[AxisPoles-I x]]<deltap,
		  			alphapole[[1]]/Abs[ppa /.x1->x],
			True,	Abs[radius[x]/B[x]] ];

(* %%%%%%%%%%%%%%%%%%%%% end of If %%% *)

	];


(* %%%%%%%%%%%%%%%%%  SAMPLING %%%%%%%%%%%%%%%%%%%%%%% *)

	If[diskterm,
		Print["Sampling Radius and Center ..."];
		kdisc = Discretize[K1,
			ValueAtInfinity->K1infinity,
			Ngrid->ngrid,GridSpread->gridspread];
		rdisc = Discretize[R1,
			ValueAtInfinity->R1infinity,
			Ngrid->ngrid,GridSpread->gridspread]
			];
	If[TrueQ[nwghts>0],
		Print["Sampling weight functions..."];
		discweights = Map[Discretize[#,
			Ngrid->ngrid,
			GridSpread->gridspread]&,weights];
		If[Max[Abs[discweights]]>10^10,
			Print["Weight functions specified with option Weights->{...} cannot be unbounded"];
			Return[$Failed]];
		];


(* %%%%%%%%%%%%%%%%%%  SMOOTHING  %%%%%%%%%%%%%%%%%%%%%% *)


     (* Smoothing of functions *)
	If[diskterm,
		Print["Smoothing Radius and Center ..."];
		ksmth = Smooth[kdisc,nsmth];
		rsmth = Smooth[rdisc,nsmth]
		];

	If[TrueQ[nwghts>0],
		Print["Smoothing weight functions..."];
		weightsmth = Map[Smooth[#,nsmth]&,discweights];
		];



(* %%%%%%%%%%%%%%%%%%  OPTIMIZATION  %%%%%%%%%%%%%%%%%%%%%% *)
Print["Optimization routine Anopt now number-crunching ..." ];
(* prepare some input for Anopt *)

	anoptopts = If[nondiskperf,
			Sequence[Smoothing -> Automatic],
			Sequence[Smoothing -> None]];


	If[diskterm,
	     data = Join[{1/rsmth,ksmth},weightsmth],
	     data = weightsmth];

	If[Length[data]==0, Null (* do nothing *),
	 	(* else clean nonreal noise and reset options for anopt *)
		fcdata = Map[Fourier[#]&,data];
		If[Max[Abs[Im[fcdata]]]>10^(-10),
		    Print["Complex Fourier coefficients found. Please check your input (radius, center, weights ...)"];
		    Abort[]
		    ];
		anoptopts = Sequence[AMatrix->data,anoptopts];
		];

(* call main optimization routine *)

	Off[CheckData::warning];
	Anopt[perf,tol,f0,anoptopts];
	On[CheckData::warning];


(* this is needed only if weights are not defined interms of ngrid *)
	OV = N[gammastar];
	OT = flat + Abs[grAlign];
	T1 =  InverseFourier[Re[Fourier[Solution[[1]] ]]];



(* %%%%%%%%%%%%%%%%%%  PREPARE THE OUTPUT  %%%%%%%%%%%%%%%%%%%%%% *)
	Print["Calculating output functions ..."];

(* Put together optimal function T  *)
	T = Discretize[A,Ngrid->ngrid,GridSpread->gridspread]+
		 Discretize[B,Ngrid->ngrid,GridSpread->gridspread] T1;
(* Put together optimal compensator Co *)



(* Compute open loop transfer function
   WARNING:  Error message regarding division by zero is switched off
   since GridW[[1]] is  = Infinity *)
        If[MemberQ[ListPoles,0.],Off[Power::infy] ];
	L = T/(1-T) ;
	If[MemberQ[ListPoles,0.],
	     L[[ngrid/2+1]]=CubicInt[0.,{-2,-1,1,2},{L[[ngrid/2+3]],
		      L[[ngrid/2+2]],L[[ngrid/2]],L[[ngrid/2-1]]}]];
(* Now compute the compensator Co and plant P as data on a grid. *)
	GridW = Grid[Ngrid->ngrid,GridSpread->gridspread];
	P = Join[{0.},plant[s] /. s -> I*Take[GridW,{2,ngrid}]];
	If[MemberQ[ListPoles,0.],
	     P[[ngrid/2+1]]=CubicInt[0.,{-2,-1,1,2},{P[[ngrid/2+3]],
		      P[[ngrid/2+2]],P[[ngrid/2]],P[[ngrid/2-1]]}]];

(* here we see if the value of Co at infinity is 0 *)
        Co = Join[{0.},Take[L,{2,ngrid}]/Take[P,{2,ngrid}]];
	If[Chop[cdecay] == 0,
		Co[[1]] = CubicInt[0.,{-2.,-1.,1.,2.},
			{Co[[ngrid-1]],Co[[ngrid]],Co[[2]],Co[[3]]}] ];

(* finally change the default options for Bode and DiscreteInvLaplace *)
(* turn on error messages again *)
        If[MemberQ[ListPoles,0.],On[Power::infy] ];
(* Reset options *)
Print["Resetting options ..."];
	SetOptions[Grid,Ngrid->ngrid,GridSpread -> gridspread];
	SetOptions[Discretize,Ngrid->ngrid,GridSpread -> gridspread];
	SetOptions[Sample,Ngrid->ngrid,GridSpread -> gridspread];
	SetOptions[DiscreteInvLaplace,GridSpread -> gridspread];
	SetOptions[StepResponse,GridSpread -> gridspread];
	SetOptions[RationalModel,
			SampledFunction->T1,
	 		GridSpread->gridspread,
	 		Method->CFApproximation,
	 		InterpolantFunctions->{AA,BB},
	 		DegreeOfDenominator->2+reldegp+cdecay];
	SetOptions[EnvelopePlot,
		 	Ngrid->ngrid,
			GridSpread->gridspread,
			Nsmth->nsmth,
		 	Radius->radius,Center ->center];

Print["Done! "];

] (* end OPTDesign Module *)

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%% END OPTDesign[] %%%%%%%%%%%%%%%%%%%%%%%%%% *)


(* ----------------------------------------------------------------- *)

(* this is an auxiliary function  for interpolating a value.
It is assumed that x[[1]]<x[[2]]<x0<x[[3]]<x[[4]].
y[[i]] corresponds to x[[i]].  Returned is the value y0
that corresponds to x0, according to the degree 3 interpolating
polynomial. *)
CubicInt[x0_,x_,y_]:=
( y[[1]] +
  (y[[2]]-y[[1]])/(x[[2]]-x[[1]]) (x0 - x[[1]]) +
  (y[[1]]/((x[[1]]-x[[2]])(x[[1]]-x[[3]])) +
   y[[2]]/((x[[2]]-x[[1]])(x[[2]]-x[[3]])) +
   y[[3]]/((x[[3]]-x[[1]])(x[[3]]-x[[2]]))  ) *
 (x0-x[[1]])(x0-x[[2]])	 +
  (y[[1]]/((x[[1]]-x[[2]])(x[[1]]-x[[3]])(x[[1]]-x[[4]])) +
   y[[2]]/((x[[2]]-x[[1]])(x[[2]]-x[[3]])(x[[2]]-x[[4]])) +
   y[[3]]/((x[[3]]-x[[1]])(x[[3]]-x[[2]])(x[[3]]-x[[4]])) +
   y[[4]]/((x[[4]]-x[[1]])(x[[4]]-x[[2]])(x[[4]]-x[[3]]))  ) *
 (x0-x[[1]])(x0-x[[2]])(x0-x[[3]])
)
(* ----------------------------------------------------------------- *)

CancelZP[rational_,term_,x_,opts___Rule]:=
	Module[{ tol,rat,num,den,numrem,denrem,numquo,denquo },
		tol = Tolerance /. {opts} /. Options[CancelZP];
		rat = Together[rational];
		num = Numerator[rat];
		den = Denominator[rat];
		numrem = 0.;
		denrem = 0.;
		While[Max[Abs[numrem],Abs[denrem]]<tol,
			numquo = Chop[PolynomialQuotient [num,term,x],tol];
			numrem = Chop[PolynomialRemainder[num,term,x],tol];
			denquo = Chop[PolynomialQuotient [den,term,x],tol];
			denrem = Chop[PolynomialRemainder[den,term,x],tol];

			If[Max[Abs[numrem],Abs[denrem]]<tol,
				num = numquo;
				den = denquo;
			  ];
		      ];
		Return[num/den];
	      ];

(* ----------------------------------------------------------------- *)

(* Compute the Limit of a rational function at infinity *)
LimitInfinite[f_,x_]:=
   Module[{term,num,den,nexp,dexp,coefnum,coefden,limit},
     If[TrueQ[D[f,x]==0],f,
	(term=Together[f];
	num = Expand[Numerator[term]];
	den = Expand[Denominator[term]];
	nexp = Exponent[num,x];
	dexp = Exponent[den,x];
	coefnum = Coefficient[num,x^nexp];
	coefden = Coefficient[den,x^dexp];
	limit = If[dexp>nexp,0.,coefnum/coefden];
	If[dexp<nexp,Infinity*limit,limit] )
] ]


(* ----------------------------------------------------------------- *)

DiscreteInvLaplace[H_,maxt_,gridt_,opts___]:=
   Module[{ngrid,b,ww,i,k,h,graph,pleaseplot},
	b = GridSpread /. {opts} /. Options[DiscreteInvLaplace];
	pleaseplot = GeneratePlot /. {opts} /. Options[DiscreteInvLaplace];
	ngrid=Length[H];
(* ww stores the grid values of w = omega correponding to H *)
	ww=Grid[Ngrid->ngrid,GridSpread->b];
(* assign to h the values of the inverse laplace on a grid on [0,maxt]
   This will be computed on the circle after a change of variables.
   Since ww = b(1+e)/(1-e), we use this in the integral expression. *)
   	h[0]=0.;
	Do[h[k]= Re[ N[1/ngrid] Sum[
		N[H[[i]] Exp[I ww[[i]] k maxt/gridt](ww[[i]]^2+b^2)/(2 b)],
			{i,2,ngrid}]],{k,1,gridt}];
	graph=Chop[Table[{N[i* maxt/gridt],h[i]},{i,0,gridt}]];
	Return[  If[pleaseplot,ListPlot[graph],graph]]
] (* end of Module *)
(* ----------------------------------------------------------------- *)

StepResponse[H_,maxt_,gridt_,opts___]:=
   Module[{ngrid,b,pplot,ww,opts1,graph,graph1,i},
	b = GridSpread /. {opts} /. Options[StepResponse];
	pplot = GeneratePlot /. {opts} /. Options[StepResponse];
	ngrid=Length[H];
	opts1=Sequence[GeneratePlot->False,GridSpread->b];
(* ww stores the grid values of w = omega correponding to H *)
	ww=Grid[Ngrid->ngrid,GridSpread->b];
(* assign to h the values of the inverse laplace on a grid on [0,maxt] *)
	graph1 = DiscreteInvLaplace[H,maxt,gridt,opts1];
	{time,impulse} = Transpose[graph1];
	step[0]=0;
	Do[step[i]=step[i-1] + (maxt/gridt) impulse[[i+1]],{i,1,gridt}];
	graph = Transpose[{time,Table[step[i],{i,0,gridt}]}];
	Return[  If[pplot,ListPlot[graph,PlotRange->All,Frame->True],graph]]
] (* end of Module *)


(* ----------------------------------------------------------------- *)

Grid[opts___Rule]:=
	Module[{n,b,x,g},
		n = Ngrid 	/. {opts} /. Options[Grid];
		b = GridSpread  /. {opts} /. Options[Grid];
		g = Table[ b Cot[N[Pi x /n]],{x,1,n-1}];
		g = 0.5*(g-Reverse[g]); (* to make it symmetric *)
		Join[{Infinity},g]
		]

(* ----------------------------------------------------------------- *)

Discretize[f_,opts___Rule]:=
	Module[{n,b,y,tempgrid,foo},
		If[FreeQ[f,s],foo=f,foo[w_]=(f /. s -> I w ) ];
		n = Ngrid 	/. {opts} /. Options[Discretize];
		b = GridSpread  /. {opts} /. Options[Discretize];
		y = ValueAtInfinity /. {opts} /. Options[Discretize];
		If[TrueQ[y==Automatic],
			y = N[Limit[ComplexCoordinates[f[xLarge]],
				xLarge->Infinity]];
			];
		If[Not[TrueQ[(Head[y]==Real) || (Head[y]==Integer)]],
	 (*  Print["Discretize[] fails calculation of infinite limit"]; *)
		    y = Re[foo[ 1000000000.]] ];
		(*    Abort[]]; *)
		tempgrid = Rest[Grid[GridSpread->b,Ngrid->n]];
		Join[{y},Map[foo[#]&,tempgrid]]
		]
(* ----------------------------------------------------------------- *)

Sample[f_,opts___Rule]:=
	Module[{n,b,y,tempgrid,x},
		n = Ngrid 	/. {opts} /. Options[Sample];
		b = GridSpread  /. {opts} /. Options[Sample];
		ss= Var         /. {opts} /. Options[Sample];
		If[FreeQ[f,ss],
			Print["Input expression is not a function of ",ss];
			Abort[]
			];
		tempgrid = Grid[GridSpread->b,Ngrid->n];
		Return[Join[{Limit[f,ss->I Infinity]},
			     Table[ f /.ss-> I tempgrid[[x]],{x,2,n}]]]
		];
(* ----------------------------------------------------------------- *)
RHPListPlot3D[T_ , opts___ ]:=
Module[{opts0,wa,wb,n},
	gp = GridPoints /. {opts} /. Options[RHPListPlot3D];
	n  = Length[gp];
	wa = gp[[n/2+1]];
	wb = gp[[2]];
	opts0 = Sequence[
		AxesLabel->{"w","Im[ ]","Re[ ]"},
		FilterOptions[ScatterPlot3D,opts]];
ScatterPlot3D[ Drop[Transpose[ {Grid[],Im[T],Re[T]} ],1] , opts0 ]
];



(* ----------------------------------------------------------------- *)
NyquistListPlot[x_List,opts___ ] :=
	Show[   ListPlot[ Transpose[ {Re[x],Im[x]} ] ,
		Sequence[opts,DisplayFunction -> Identity ]],
			Graphics[{PointSize[0.03],Point[{-1,0}]}],
			DisplayFunction -> $DisplayFunction];


NyquistPlot[x_,opts___] :=
	Module[{G,fb,opts0},
	fb = FrequencyBand /. {opts} /. Options[NyquistPlot];
	opts0= Sequence[DisplayFunction -> Identity,
                  FilterOptions[ParametricPlot,opts],
			FilterOptions[ParametricPlot,Options[NyquistPlot]]];
	opts1= Sequence[DisplayFunction -> Identity,
			PlotStyle -> {{Dashing[{0.03,0.03}]}},
                  FilterOptions[ParametricPlot,opts],
			FilterOptions[ParametricPlot,Options[NyquistPlot]]];
	If[Not[FreeQ[x,s]],
		G = x /. s -> I w;
	     p1=ParametricPlot[{Re[G],Im[G]},
			{w,fb[[1]],fb[[2]]},Evaluate[opts1]];
	     p2=ParametricPlot[{Re[G],Im[G]},
			{w,-fb[[2]],-fb[[1]]},Evaluate[opts0]],
	     p1=ParametricPlot[{Re[x[I w]],Im[x[I w]]} ,
			{w,fb[[1]],fb[[2]]},Evaluate[opts1]];
	     p2=ParametricPlot[{Re[x[I w]],Im[x[I w]]},
			{w,-fb[[2]],-fb[[1]]},Evaluate[opts0]]];
	Show[p1,p2,Graphics[{PointSize[0.03],Point[{-1,0}]}],
			DisplayFunction -> $DisplayFunction];
		];

Nyquist[x_,opts___]:=
	If[ListQ[x],NyquistListPlot[x,opts],
		    NyquistPlot[x,opts]];


(* ----------------------------------------------------------------- *)



BodeMagnitude[f_List, opts___] :=
   Module[{pr,pr2, r,g,xmin,xmax,fx,s,x,dispfn},
	n = Length[f];
        {wa,wb} = FrequencyBand /. {opts} /. Options[BodeMagnitude];
pr = PlotRange /. {opts} /. Options[BodeMagnitude];
pr2 = Which[Dimensions[pr]=={2,2},pr[[2]],
	    Dimensions[pr]=={2},pr,
	    True,pr];
        lwa = Log[10,wa];
        lwb = Log[10,wb];
	tt = Reverse[Take[f,{2,n/2}]];
	ww = Reverse[Take[Grid[],{2,n/2}]];
	val = Map[20. Log[10,Abs[#]]&,tt];
	data = Transpose[{ww,val}];
	g  = ListPlot[ N[{Log[10, #[[1]]],#[[2]]}]& /@ data,
		Ticks->{LogScale,Automatic},
			PlotRange -> {{lwa,lwb},pr2},
                        Prolog -> PointSize[0.01],
			RotateLabel -> False,
                        FrameTicks -> {LogScale, Automatic,
                                       LogScale, Automatic},
                		DisplayFunction -> Identity,
				Evaluate[FilterOptions[ListPlot,opts]] ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
         		GridLines->{LogGridMajor,Automatic},
				Frame->True,
				FrameLabel->
				{"w (rad/s)","dB",None,None},
         		AxesOrigin -> Map[#[[1]]&,r]]
       ];



BodeMagnitude[f_?(Not[ListQ[#]]&), opts___] :=
	Module[{pr,pr2,r,g,xmin,xmax,fx,s,x},
	     s = Var /. {opts} /. Options[BodeMagnitude];
	     fband = FrequencyBand /. {opts} /. Options[BodeMagnitude];
             {wa,wb} = FrequencyBand /. {opts} /.Options[BodeMagnitude];
             lwa = Log[10,wa];
             lwb = Log[10,wb];
	     {xmin,xmax}=fband;
	     fx= f /. s-> I x;
pr = PlotRange /. {opts} /. Options[BodeMagnitude];
pr2 = Which[Dimensions[pr]=={2,2},pr[[2]],
	    Dimensions[pr]=={2},pr,
	    True,pr];

	g  = ParametricPlot[ {Log[10,x],20. Log[10,Abs[fx]] },
				{x,wa,wb},
		Ticks->{LogScale,Automatic},
			PlotRange -> {{lwa,lwb},pr2},
			RotateLabel -> False,
                        FrameTicks -> {LogScale, Automatic,
                                      LogScale, Automatic},
                	DisplayFunction -> Identity,
			Evaluate[FilterOptions[ListPlot,opts]] ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
         		GridLines->{LogGridMajor,Automatic},
				Frame->True,
				FrameLabel->
				{"w (rad/s)","dB",None,None},
         		AxesOrigin -> Map[#[[1]]&,r]]
       ];


(* ----------------------------------------------------------------- *)

BodePhase[discf_List, opts___] :=
	Module[ {(* tt,ww,deltaangle,data,g,r,val,i,n,b *)},
		n = Length[discf];
		b = GridSpread /. {opts} /. Options[BodePhase];
pr = PlotRange /. {opts} /. Options[BodeMagnitude];
pr2 = Which[Dimensions[pr]=={2,2},pr[[2]],
	    Dimensions[pr]=={2},pr,
	    True,pr];

                {wa,wb} = FrequencyBand /. {opts} /. Options[BodePhase];
                lwa = Log[10,wa];
                lwb = Log[10,wb];
		tt = Reverse[Take[discf,{2,n/2+1}]];
		ww = Reverse[Take[Grid[Ngrid->n,GridSpread->b],{2,n/2+1}]];
		deltaangle = N[ 180 Arg[RotateLeft[tt]/tt]/Pi];
		angle[1]=If[Re[tt[[1]]]<0.,-180.,0.];
	       val=Table[angle[i+1]=angle[i]+deltaangle[[i]],
	                 {i,1,n/2 -1}];
		val = Join[{angle[1]},val];
		data = Drop[Transpose[{ww,val}],1];
		g  = ListPlot[ N[{Log[10, #[[1]]],#[[2]]}]& /@ data,
			Ticks->{LogScale,Automatic},
			PlotRange -> {{lwa,lwb},pr2},
                        Prolog -> PointSize[0.01],
			RotateLabel -> False,
                        FrameTicks -> {LogScale, Automatic,
                                       LogScale, Automatic},
                		DisplayFunction -> Identity,
				Evaluate[FilterOptions[ListPlot,opts]] ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
         		GridLines->{LogGridMajor,Automatic},
				Frame->True,
				FrameLabel->
				{"w (rad/s)","deg",None,None},
         		AxesOrigin -> Map[#[[1]]&,r]]]

(* ----------------------------------------------------------------- *)
BodePhase[f_?(Not[ListQ[#]]&), opts___] :=
        Module[ {tt,ww,deltaangle,data,g,r,val,i,n,b,pr,pr2},
                n = SamplingPoints    /. {opts} /. Options[BodePhase];
pr = PlotRange /. {opts} /. Options[BodeMagnitude];
pr2 = Which[Dimensions[pr]=={2,2},pr[[2]],
	    Dimensions[pr]=={2},pr,
	    True,pr];

                {wa,wb} = FrequencyBand /. {opts} /. Options[BodePhase];
                del = (lwb-lwa)/n;
                lwa = Log[10,wa];
                lwb = Log[10,wb];
                ww = Table[ 10.^(lwa + i del),{i,1,n+1}];
                tt = f /. s -> I ww;
                deltaangle = N[ 180 Arg[RotateLeft[tt]/tt]/Pi];
                angle[1]=If[Re[tt[[1]]]<0.,-180.,0.];
               val=Table[angle[i+1]=angle[i]+deltaangle[[i]],
                         {i,1,n}];
                val = Join[{angle[1]},val];
                data = Drop[Transpose[{ww,val}],-1];
                g  = ListPlot[ N[{Log[10, #[[1]]],#[[2]]}]& /@ data,
                        Ticks->{LogScale,Automatic},
			PlotRange -> {{lwa,lwb},pr2},
                        Prolog -> PointSize[0.01],
                        RotateLabel -> False,
                        PlotJoined -> True,
                        FrameTicks -> {LogScale, Automatic,
                                       LogScale, Automatic},
                                DisplayFunction -> Identity,
                                Evaluate[FilterOptions[ListPlot,opts]] ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
                        GridLines->{LogGridMajor,Automatic},
                                Frame->True,
                                FrameLabel->
                                {"w (rad/s)","deg",None,None},
                        AxesOrigin -> Map[#[[1]]&,r]]];

(* ----------------------------------------------------------------- *)

ListPolarPlot[fun_,opts___]:= ListPlot[Transpose[{Re[fun],Im[fun]}],
				AxesLabel->{"Re","Im"},opts]

(* ----------------------------------------------------------------- *)
Nichols[f_, opts___] :=
        Module[ {tt,ww,deltaangle,data,g,r,val,i,n,b,opts0},
                n = SamplingPoints    /. {opts} /. Options[Nichols];
                {wa,wb} = FrequencyBand /. {opts} /. Options[Nichols];
                del = (lwb-lwa)/n;
                lwa = Log[10,wa];
                lwb = Log[10,wb];
                ww = Table[ 10.^(lwa + i del),{i,1,n+1}];
                tt = f /. s -> I ww;
                deltaangle = N[ 180 Arg[RotateLeft[tt]/tt]/Pi];
                angle[1]=If[Re[tt[[1]]]<0.,-180.,0.];
                val=Table[angle[i+1]=angle[i]+deltaangle[[i]],
                         {i,1,n}];
                val = Join[{angle[1]},val];
		     data = Transpose[{val,N[20 Log[10,Abs[tt]]]}];
		opts0 = Sequence[PlotJoined -> True,
				 FilterOptions[ListPlot,opts]];
		g  = ListPlot[data,DisplayFunction -> Identity, opts0 ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
         	PlotRange -> r,
         	GridLines->Automatic,
		RotateLabel -> False,
		Frame->True,
		FrameLabel->{"deg","dB",None,None},
         	AxesOrigin -> Map[#[[1]]&,r]]]
(* ----------------------------------------------------------------- *)

(* ----------------------------------------------------------------- *)

Nichols[discf_List, opts___] :=
	Module[ {tt,deltaangle,angle,data,g,r,val,n,i,opts0},
		n = Length[discf];
		tt = Reverse[Take[discf,{2,n/2+1}]];
		deltaangle = N[ 180 Arg[RotateLeft[tt]/tt]/Pi];
		angle[1]=If[Re[tt[[1]]]<0.,-180.,0.];
		val=Table[angle[i+1]=angle[i]+deltaangle[[i]],
					{i,1,n/2 -1}];
		val = Join[{angle[1]},val];
		data = Take[Transpose[{val,N[20 Log[10,Abs[tt]]]}],
					{1,n/2 -3}];
		opts0 = Sequence[ Prolog -> PointSize[0.01],
				  FilterOptions[ListPlot,opts]];
		g  = ListPlot[data,DisplayFunction -> Identity, opts0 ];
         r = PlotRange[g];
         Show[g,DisplayFunction -> $DisplayFunction,
         	PlotRange -> r,
         	GridLines->Automatic,
		RotateLabel -> False,
		Frame->True,
		FrameLabel->{"deg","dB",None,None},
         	AxesOrigin -> Map[#[[1]]&,r]]]
(* ----------------------------------------------------------------- *)


Apart2[rational_,s_,opts___Rule]:=
Module[{num,den,quot,nrem,coefs,rat2,pole,X,Y,k,rep,q1,a,i,j,tol},
	tol = Apart2Tol /. {opts} /. Options[Apart2];
	{num,den}={Numerator[rational],Denominator[rational]};
	pole = s /. Solve[den==0,s];
	quot = PolynomialQuotient[num,den,s];
	nrem = PolynomialRemainder[num,den,s];
	X = pole;
	Y = {};
	While[ Length[X]>0,
		rep = Flatten[Position[Chop[X-X[[1]],tol],0]];
		Y = Join[Y,{{X[[1]],Length[rep]}}];
		Do[X=Delete[X,rep[[k]]-k+1],{k,1,Length[rep]}];
		];
	Do[If[Y[[i,2]]>1,
		(* then *)
		    deriv = D[den,{s,Y[[i,2]]-1}];
		    s1rule = FindRoot[deriv==0,{s,Y[[i,1]]}];
		    s1 = s /. s1rule;
		    If[Abs[Y[[i,1]] - s1]<tol ,Y[[i,1]]=s /. s1rule];
		],{i,1,Length[Y]}];
	Do[
		q1 = PolynomialQuotient[den,(s-Y[[i,1]])^(Y[[i,2]]),s];
		Do[ a[i,Y[[i,2]]-j]= D[nrem/q1,{s,j}]/(j!) /. s->Y[[i,1]],
						{j,0,Y[[i,2]]-1}];
			,{i,1,Length[Y]}];
		quot + Sum[ Sum[ a[i,j]/(s - Y[[i,1]])^(j),
		{j,1,Y[[i,2]]}],
		{i,1,Length[Y]}]
	]
(* ----------------------------------------------------------------- *)
StableFit[f_,wpts_,degrule__,opts___Rule]:=
   Module[{n,t1,b,order,var,method,z,rat,deg },
        t1 = f;
	n = Length[wpts];
	b = wpts[[n/4+1]];
	order = DegreeOfDenominator /. {degrule} ;
	var  = Var /. {opts} /. Options[StableFit];
	rat =Check[CFApproximant[t1,order,order,
	            Var->z,GridSpread->b],True];
	If[rat==True,Return[$MessageList]];
	rat = Together[ rat /. {z -> (var - b)/(var + b)}] ;
	Return[rat];
	];
(* ----------------------------------------------------------------- *)
RationalModel[var_,opts___Rule]:=
   Module[{t1,(* var, *)b,a0,b0,order,method,z,rat,deg},
        t1 = SampledFunction /. {opts} /. Options[RationalModel];
	(* var = Var	     /. {opts} /. Options[RationalModel]; *)
	b  = GridSpread      /. {opts} /. Options[RationalModel];
	{a0,b0} = InterpolantFunctions /. {opts} /. Options[RationalModel];
	order = DegreeOfDenominator /. {opts} /. Options[RationalModel];
	method = Method      /. {opts} /. Options[RationalModel];
	deg = Exponent[Denominator[b0[s]],s];
	If[TrueQ[order==Automatic],order=2+deg];
	If[TrueQ[order - deg<0],
		Print["Error in input: \n
		Internal stability requires that the option \n
		OrderOfDenominator \n
		must specify an integer greater than or equal to ",deg];
		Return[$Failed]];
	If[method==CFApproximation,
	   rat =Check[CFApproximant[t1,order-deg,order-deg,
	              Var->z,GridSpread->b],True];
	   If[rat==True,Return[$MessageList]];
	   rat = Together[ rat /. {z -> (var - b)/(var + b)}] ;
	   ];
	Return[a0[var] + b0[var] rat];
	];
(* ----------------------------------------------------------------- *)


CFApproximant[fdisc_,dn_,dd_,opts___Rule]:=
Module[{n,k,dn1,dd1,dim,edisc,fc,hankel,u,v,sdn1,var,rat,gsp,sing,
		 u1,v1,zr,qout,qc,b,rt,rtc,pc1,pc,error,x,y,i,j},
	n = Length[fdisc];
	k = HankelSize /. {opts} /. Options[CFApproximant];
	var = Var /. {opts} /. Options[CFApproximant];
	If[k==Automatic,k = n/8];
	dn1 = dn + 1;
	dd1 = dd + 1;
        If[ dd1 > k+dd-dn ,
		k = dd1 + dn -dd;  (* reset k *)
        	dim = dd1,
           (* else *)
	        dim = k + dd - dn];
	dim = k + dd - dn;
	edisc = Table[N[Exp[2 Pi I (i-1)/n]],{i,1,n}];
	fc=Re[Join[Take[Reverse[RotateLeft[Fourier[fdisc]]]/N[Sqrt[n]],{1,n/2}]
                  ,Table[0.,{n}]]];
	hankel = Table[fc[[i+j+dn-dd]],{i,1,dim},{j,1,dim}];
	svd = Check[SingularValues[hankel],True];
	If[svd==True,Return[$MessageList]];
	{u,sing,v}=svd;
	sdn1 = sing[[dd1]];
	u1 = Reverse[u[[dd1]]];
	v1 = v[[dd1]];
	zr = z /. N[Solve[ v1.Reverse[Table[z^i,{i,0,Length[v1]-1}]]==0,z]];
	qout = {};
	Do[If[Abs[zr[[i]]]>1,qout = Join[qout,{zr[[i]]}]],
                   {i,1,dim-1}];
	qc = Expand[Product[(x - qout[[i]]),{i,1,Length[qout]}]/
        Product[qout[[i]],{i,1,Length[qout]}]];
	b = Reverse[RotateLeft[Fourier[Join[u1,Table[0.,{n-dim}]]]/
              Fourier[Join[v1,Table[0.,{n-dim}]]]]];
	rt = fdisc - sdn1*edisc^k *b;
	rtc = Reverse[RotateLeft[Re[Fourier[rt]/N[Sqrt[n]]]]];
	pc1 = Expand[x^(Length[qout]-1+dn1)*
        (  (y^Length[qout]) (qc /. x->y^(-1) )*
            Take[rtc,{1,dn1}].Reverse[Table[y^i,{i,0,dn1-1}]]
         ) /.y->x^(-1)];
	pc = (pc1 /. x->0. ) + Sum[Coefficient[pc1,x^i]*x^i,{i,1,dn1-1}];
	error = Max[Abs[fdisc - (pc/qc /. x->edisc)]];
	Print["Error = ",error];
	rat = Together[ pc/qc /. x->var ] ;
	Return[Chop[rat]];
   ]

(* ----------------------------------------------------------------- *)

PlotZP[expr_,var_,opts___Rule]:=
	Module[{rat,s,zero,pole,maxr,minr,maxi,mini,tx,ty,t,
			cz,xp,ran,i,opts1},
		rat = Together[expr];
		s = var ;
		zero = s /. N[Solve[Numerator[rat]==0,s]];
		pole = s /. N[Solve[Denominator[rat]==0,s]];
		maxr = Max[{Re[{pole,zero}],1}];
		minr = Min[{Re[{pole,zero}],-1}];
		maxi = Max[{Im[{pole,zero}],1}];
		mini = Min[{Im[{pole,zero}],-1}];


		tx = (maxr-minr)/50.;
		ty = (maxi-mini)/50.;
		t = 0.5*(tx + ty);
		cz = Table[Circle[{Re[zero[[i]]],Im[zero[[i]]]},
			Scaled[{0.02, 0.02}]],
             {i,1,Length[zero]}];

		xp =    Flatten[Table[
          { Line[{{Re[pole[[i]]]-tx,Im[pole[[i]]]-ty},
                 {Re[pole[[i]]]+tx,Im[pole[[i]]]+ty}}],
            Line[{{Re[pole[[i]]]-tx,Im[pole[[i]]]+ty},
                 {Re[pole[[i]]]+tx,Im[pole[[i]]]-ty}}] }
       		,{i,1,Length[pole]}]];

		ran = {{Min[0.,minr],Max[0.,maxr]},{mini,maxi}};
		opts1 = Sequence[
              			Axes->True,
              			AxesOrigin->{0.,0.},
              			PlotRange-> 1.1 ran,
              			AspectRatio->1,
				opts];

		Return[Show[Graphics[Join[cz,xp],opts1
              ]]];
     ];

(* ----------------------------------------------------------------- *)
EnvelopePlot[opts___Rule]:=
	Module[{ngr,nsm,ngrsp,rad,cen,freqb,w,p01,p02,grid0,
			k0disc,r0disc,k0smth,r0smth,p01sm,p02sm,
			modcen,modrad},
		ngr = Ngrid 	  /. {opts} /.   Options[EnvelopePlot];
		nsm = Nsmth 	  /. {opts} /.   Options[EnvelopePlot];
		ngrsp=GridSpread /. {opts} /.    Options[EnvelopePlot];
		rad = Radius     /. {opts} /.    Options[EnvelopePlot];
		cen = Center     /. {opts} /.    Options[EnvelopePlot];
		modrad = ModifiedRadius  /. {opts} /.    Options[EnvelopePlot];
		modcen = ModifiedCenter  /. {opts} /.    Options[EnvelopePlot];
		freqb= FrequencyBand /. {opts} /. Options[EnvelopePlot];
		If[modcen==Automatic,modcen = cen];
		If[modrad ==Automatic,modrad = rad];
		If[cen==Automatic || rad ==Automatic,
				Print["You must specify Radius and Center as input"];
				Return[];
				];
		p01 =Plot[Re[cen[w]+rad[w]],{w,freqb[[1]],freqb[[2]]},
			PlotStyle->{{Thickness[0.01],GrayLevel[0.5]}},
            		DisplayFunction->Identity];
		p02 =Plot[Re[cen[w]-rad[w]],{w,freqb[[1]],freqb[[2]]},
			PlotStyle->{{Thickness[0.01],GrayLevel[0.5]}},
            		DisplayFunction->Identity];
		grid0 = Grid[Ngrid->ngr,GridSpread->ngrsp];
		k0disc = Discretize[modcen,Ngrid->ngr,GridSpread->ngrsp];
		r0disc = Discretize[modrad,Ngrid->ngr,GridSpread->ngrsp];
		k0smth = Re[Smooth[k0disc,nsm]];
		r0smth = Re[Smooth[r0disc,nsm]];
		p01sm =
		  ListPlot[Re[Take[Transpose[{grid0,k0smth-r0smth}],{2,ngr}]],
					PlotJoined->True,
        			PlotRange->{freqb,Automatic},
        			DisplayFunction->Identity];
		p02sm = ListPlot[Re[Take[Transpose[{grid0,k0smth+r0smth}],
					{2,ngr}]],
				PlotJoined->True,
        			PlotRange->{freqb,Automatic},
        			DisplayFunction->Identity];
		Return[Show[p01sm,p02sm,p01,p02,
            		DisplayFunction->$DisplayFunction
		    (*	,PlotLabel->"Original and modified Envelopes" *) ]];
		 ];

(* ----------------------------------------------------------------- *)

EnvelopeLogPlot[opts___Rule]:=
	Module[{ngr,nsm,ngrsp,rad,cen,freqb,w,p01,p02,grid0,
			k0disc,r0disc,k0smth,r0smth,p01sm,p02sm,
			modcen,modrad,llp},
		ngr = Ngrid 	  /. {opts} /.   Options[EnvelopeLogPlot];
		nsm = Nsmth 	  /. {opts} /.   Options[EnvelopeLogPlot];
		ngrsp=GridSpread /. {opts} /.    Options[EnvelopeLogPlot];
		rad = Radius     /. {opts} /.    Options[EnvelopeLogPlot];
		cen = Center     /. {opts} /.    Options[EnvelopeLogPlot];
		modrad = ModifiedRadius  /. {opts} /. Options[EnvelopeLogPlot];
		modcen = ModifiedCenter  /. {opts} /. Options[EnvelopeLogPlot];
		freqb= FrequencyBand /. {opts} /. Options[EnvelopeLogPlot];
		If[modcen==Automatic,modcen = cen];
		If[modrad ==Automatic,modrad = rad];
		If[cen==Automatic || rad ==Automatic,
				Print["You must specify Radius and Center as input"];
				Return[];
				];
		grid0 = Grid[Ngrid->ngr,GridSpread->ngrsp];
		k0disc = Discretize[modcen,Ngrid->ngr,GridSpread->ngrsp];
		r0disc = Discretize[modrad,Ngrid->ngr,GridSpread->ngrsp];
		k0smth = Re[Smooth[k0disc,nsm]];
		r0smth = Re[Smooth[r0disc,nsm]];
(* Print[Take[Transpose[{grid0,k0smth+r0smth}], {2,ngr/2-1}]]; *)
		Return[LogLogListPlot[ Take[Transpose[{grid0,k0smth+r0smth}],
					{2,ngr/2-1}],
				(* Frame->True, *) PlotRange->All]];
		 ];

(* ----------------------------------------------------------------- *)
OPTDParametrize[f_,opts___Rule]:= Transpose[{Grid[opts],f}];
(* ----------------------------------------------------------------- *)
EnvelopePlot3D[opts___Rule]:=
Module[{t,i,r0,k0,trat,plot,fb,plot2,pts,Tratw,delta},
	r0 = Radius /. {opts}/. Options[EnvelopePlot3D];
	k0 = Center /. {opts}/. Options[EnvelopePlot3D];
	pj = PlotJoined /. {opts}/. Options[EnvelopePlot3D];
	If[TrueQ[r0==Automatic || k0 == Automatic],
		Print["Please specify Radius->r0 and Center->k0"];
		Return[];
	   ];
	trat = ClosedLoop /. {opts} /. Options[EnvelopePlot3D];
	fb = FrequencyBand /. {opts} /. Options[EnvelopePlot3D];
	plot = ParametricPlot3D[{w,r0[w] Sin[t],
                    k0[w]+r0[w] Cos[t]},
                 {w,fb[[1]],fb[[2]]},{t,0,Pi},
                 AmbientLight->GrayLevel[0.1],
                 DisplayFunction->Identity,
                 ViewPoint->{1.3,-2.4,2},
		 AxesLabel->{"w","Im[z]","Re[z]"}
                 ];
	Which[TrueQ[trat == None],
		Return[Show[plot,DisplayFunction->$DisplayFunction]],
	      ListQ[trat],
		wgrid = Grid[];
		If[pj,
		     pts = Drop[Transpose[{wgrid,Im[trat],Re[trat]}],1];
		     lines = Table[Line[{pts[[i]],pts[[i+1]]}],{i,1,Length[pts]-1}];
		     plot2 = Show[Graphics3D[{Thickness[0.01],lines}],
				  DisplayFunction->Identity],
		  (* else *)
		     pts = Drop[Map[Point[#]&,Transpose[{wgrid,Im[trat],Re[trat]}]],1];
		     plot2 = Graphics3D[{PointSize[0.02],pts}];
		  ];
		Return[Show[plot,plot2,DisplayFunction->$DisplayFunction,
				PlotRange -> {fb,Automatic,Automatic}]],
	      Not[FreeQ[trat,s]],
		delta = (fb[[2]]-fb[[1]])/100.;
		Tratw = trat /. s-> I w ;
		pts = Table[Line[{{w,Im[Tratw],Re[Tratw]}/.w->fb[[1]]+delta*i,
              		{w,Im[Tratw],Re[Tratw]}/.w->fb[[1]]+delta*(i+1)}],
                    	{i,99}];
		plot2 = Show[Graphics3D[{Thickness[0.01],pts}],
				DisplayFunction->Identity];
		Return[Show[plot,plot2,DisplayFunction->$DisplayFunction]],
	      True , Print["Error in input"]
	   ];
	];
(* AN EXAMPLE :
 	p[s_]= (5 - s)/(1+s)^2;
 	k0[w_]= Which[ Abs[w] < 1, 1.0,
                      True, 0.0];
 	r0[w_] = Which[Abs[w] < 1,  2.0,
 			True,	0.707*Abs[p[I w]]];
 	Trat = (-5 + s)/(1+s)^2 +
 	       (-0.911 -1.261 s)(-5+s)/((s+1)^2 (0.88 + s)) ;
 	fb = {0.,4.};
 	test := EnvelopePlot3D[Radius->r0,
 		Center->k0,
 		ClosedLoop->Trat,
 		FrequencyBand->{0.,4.}]
*)


End[ (* "`Private`" *)]

Protect[ ]

EndPackage[ (* OPTDesign *) ]
