(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     99127,       3403]*)
(*NotebookOutlinePosition[    100038,       3434]*)
(*  CellTagsIndexPosition[     99994,       3430]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Anopt", "Title",
  Evaluatable->False,
  TextAlignment->Center,
  AspectRatioFixed->False],

Cell[TextData[
"A program for sup-norm optimization over spaces of analytic functions"], 
  "Text",
  CellMargins->{{17, Inherited}, {Inherited, Inherited}},
  Evaluatable->False,
  CellLabelMargins->{{17, Inherited}, {Inherited, Inherited}},
  TextAlignment->Center,
  AspectRatioFixed->False],

Cell[TextData[
"By  J. W. Helton,  O. Merino, J. Myers and T. Walker\nLab. for Mathematics \
and Statistics, University of California, San Diego"], "Text",
  Evaluatable->False,
  TextAlignment->Center,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Foreword"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "\tAnopt is a  Mathematica package for solving diverse optimization \
problems over spaces of functions analytic on the unit disk in the complex \
plane.\n\tThe software Anopt is useful for engineers doing worst case design \
in the frequency domain.  This includes problems in control as well as \
broadband gain equalization and matching.  Another application is in the \
several complex variable field where the program can be used to find analytic \
disks which are optimal with respect to various criteria, e.g.-Kobayashi \
metric calculations. \n\tThe program is easy to use, even for those with \
little or no computing experience.  The program can be run at a very simple \
level, or a very complex level.  We attempted to make the program as open and \
flexible as possible.  \n\tAlthough Anopt is very easy to use, and for \
operation the user need not look beyond the examples in section 1, one should \
familiarize with some of the options.  Also, many new ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " subroutines are automatically loaded in.   Familiarization with these \
routines will not only help the user in the operation of this program, but \
may as well help in the operation of other programs. \n\tTo the best of our \
knowledge, the program retains complete compatibility to all systems running \
Mathematica.  \n\tDisk space can also be used quickly on problems that \
require a relatively large grid size.  Our advise to you  is to do runs with \
very small gridsizes at first, and slowly work your way up in gridsize and \
difficulty of the problem until you get a feel for the limits of your system.\
\n\tThe software Anopt was developed at the Laboratory for  Mathematics and \
Statistics at the University of California at San Diego, during the years \
1989 to 1994, by J. W. Helton, Orlando Merino, Julia Myers and Trent Walker.  \
Financial support came from the Air Force Office of Scientific Research, the \
National Science Foundation, and the NSF-REU program at the San Diego Super \
Computer center.\n\nSend comments, questions, and information on bugs to \
anopt@math.ucsd.edu"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Optimizing in the sup-norm: the problem OPT "], "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The Mathematica program Anopt can be used to solve min-max problems over \
functions f analytic on the unit disk in the complex plane.  Anopt can do it \
using several algorithms, but the default is a steepest descent -type of \
algorithm called Disk Iteration.  We begin by stating the main optimization \
problem to be solved.\n\nWe shall use the following notation.\n\tA_",
  StyleBox["N",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "  =       the space of C^N-valued analytic functions on the unit disk D in \
the complex plane C, \n\t\t       f  = (  f [1] , f [2] , . . . , f [N]  )  \
which extend continuously to the closed disk. \n\t\t\t  \n\t",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e , z ) =  a smooth, positive-valued function of e in the unit circle \
and \n\t\t        z  = (  z [1] , z [2] , . . . , z [N] )  in  C^N.\n\nThe \
main optimization problem is\n "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  " OPT\t\t           Given  ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e , z ) ,  find    f*  in  A_N   such that \n\n\t\t            s* = inf",
  
  StyleBox["_f",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "    sup",
  StyleBox["_e",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "   ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e   , f ( e ) ) =   sup",
  StyleBox["_e",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "   ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e   , f* ( e ) )\n\t\nwhere the infimum is over f in A^N and the \
supremum is over e in the unit circle."
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[
"Example 1.  First run with a scalar valued case."], "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["The problem"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Our introductory example consists of a performance function where there is \
a single complex variable.\nConsider the function ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e, z ) = Abs[ 0.8  +  ( 1 / e  + z  )^2 ]^2",
  StyleBox["   ",
    FontVariations->{"CompatibilityType"->"Superscript"}],
  "as the objective or performance function.  We want to solve OPT for the \
function G( e , z ) and calculate the optimal value and optimal function f*."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Invoking Anopt[ ]  the first time"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"Begin by entering Mathematica and then loading the package Anopt."], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< "\<Anopt.m\>"\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(" \n Anopt 4.0 (c) Copyright 1991-98 \n J.W.Helton and O.Merino.  All \
rights reserved."\)], "Print"]
}, Open  ]],

Cell[TextData[{
  "To type in the formula for the performance ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e, z ) requires a translation to the symbols Anopt understands.  To \
enter G in Mathematica, replace  z by z[1] , and assign the result to a name, \
say g:"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(g = Abs[0.8 + \((1\/e + z[1])\)\^2]\^2\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    SuperscriptBox[
      RowBox[{"Abs", "[", 
        RowBox[{
          RowBox[{
            StyleBox["0.8`",
              StyleBoxAutoDelete->True,
              PrintPrecision->1], "\[InvisibleSpace]"}], "+", 
          \(\((1\/e + z[1])\)\^2\)}], "]"}], "2"]], "Output"]
}, Open  ]],

Cell[TextData[
"Besides the performance function, Anopt[ ] needs as input an error \
tolerance.  Below we set the tolerance as 0.02."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Anopt[g, 0.02]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 3.24           E+00 :   N/A   : 9.9E-01 : 0  E+00 :   N/A   : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.3456         E+00 : 4. E-01 : 4.8E-01 : 0  E+00 : 4. E-15 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 2 : 1.0242720431082E+00 : 1.4E-01 : 4.9E-02 : 0  E+00 : 2.4E-05 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 3 : 1.0100848818098E+00 : 1.8E-02 : 1.8E-02 : 0  E+00 : 5.4E-03 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.010084881809756E+00"\)], "Print"],

Cell[BoxData[
    \("flat      = 1.7866711005E-02"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 5.35E-03"\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Output from Anopt"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"When Anopt runs, the screen output gives information on how the run is \
progressing.   \nIn the screen output from the previous run we see several \
columns.  A brief explanation of their meaning is "], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "                             It        Iteration number.\n                 \
     Value       Sup",
  StyleBox[" e",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e , f( e )), the value at current iteration        \n                    \
    Step       SupNorm of difference between last two iterates f .    \n      \
                   Flat       Flatness optimality diagnostic.  It is zero at \
the solution.     \n               GradAlign       Gradient alignment \
optimality diagnostic.  It is zero at the solution.       \n                  \
        ned       A measure of numerical noise in calculations.      \n       \
                   Sm.       Indicates whether smoothing takes place.     \n  \
                      Grid        Number of points on the unit circle used \
for function evaluation."
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"In our run above, Flat went from 0.99 down to 0.018 in 3 iterations, while \
GradAlign was zero throughout the iteration (this is not unusual for scalar \
valued examples).  At the moment of stopping the iteration was making \
acceptable progress, but the (large) error tolerance we gave as input \
prevented the program Anopt[ ] from obtaining more accurate results."], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Solutions"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "When the run is over you will have access to the calculated value of the \
function f*, under the name ",
  StyleBox["Solution",
    FontWeight->"Bold"],
  ".   This is a list of lists with the following format:\n\n       Solution \
= { {x",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",x",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",...x",
  StyleBox["n",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "} , {y",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",y",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",...y",
  StyleBox["n",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "} ,..., {z",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",z",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",...z",
  StyleBox["n",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "} },\n\nwhere x",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",x",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",...,z",
  StyleBox["n",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " are complex numbers.  The k-th entry of Solution (itself a list) is  \
Solution[[k]].  "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dimensions[Solution]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \({1, 32}\)], "Output"]
}, Open  ]],

Cell[TextData[
"Thus in our example, only one scalar valued analytic function was produced.  \
It consists of 32 values which are the result of sampling the function  on a \
grid of 32 equally spaced points on the unit circle."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Plotting solutions"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"We now  plot the solution as a discrete curve on the complex plane, using \
the command "], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DiskListPlot[Solution]; \)\)], "Input",
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.952381 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0147151 0.97619 0.603319 MathSubStart
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.870235 0.309017 0.863916 [
[.15191 .29652 -12 -9 ]
[.15191 .29652 12 0 ]
[.32595 .29652 -12 -9 ]
[.32595 .29652 12 0 ]
[.67405 .29652 -9 -9 ]
[.67405 .29652 9 0 ]
[.84809 .29652 -9 -9 ]
[.84809 .29652 9 0 ]
[1.025 .30902 0 -7.5 ]
[1.025 .30902 16 7.5 ]
[.4875 .04984 -24 -4.5 ]
[.4875 .04984 0 4.5 ]
[.4875 .13623 -24 -4.5 ]
[.4875 .13623 0 4.5 ]
[.4875 .22263 -24 -4.5 ]
[.4875 .22263 0 4.5 ]
[.4875 .39541 -18 -4.5 ]
[.4875 .39541 0 4.5 ]
[.4875 .4818 -18 -4.5 ]
[.4875 .4818 0 4.5 ]
[.4875 .56819 -18 -4.5 ]
[.4875 .56819 0 4.5 ]
[.5 .64303 -8 0 ]
[.5 .64303 8 15 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15191 .30902 m
.15191 .31527 L
s
[(-0.4)] .15191 .29652 0 1 Mshowa
.32595 .30902 m
.32595 .31527 L
s
[(-0.2)] .32595 .29652 0 1 Mshowa
.67405 .30902 m
.67405 .31527 L
s
[(0.2)] .67405 .29652 0 1 Mshowa
.84809 .30902 m
.84809 .31527 L
s
[(0.4)] .84809 .29652 0 1 Mshowa
.125 Mabswid
.19542 .30902 m
.19542 .31277 L
s
.23893 .30902 m
.23893 .31277 L
s
.28244 .30902 m
.28244 .31277 L
s
.36946 .30902 m
.36946 .31277 L
s
.41298 .30902 m
.41298 .31277 L
s
.45649 .30902 m
.45649 .31277 L
s
.54351 .30902 m
.54351 .31277 L
s
.58702 .30902 m
.58702 .31277 L
s
.63054 .30902 m
.63054 .31277 L
s
.71756 .30902 m
.71756 .31277 L
s
.76107 .30902 m
.76107 .31277 L
s
.80458 .30902 m
.80458 .31277 L
s
.10839 .30902 m
.10839 .31277 L
s
.06488 .30902 m
.06488 .31277 L
s
.02137 .30902 m
.02137 .31277 L
s
.89161 .30902 m
.89161 .31277 L
s
.93512 .30902 m
.93512 .31277 L
s
.97863 .30902 m
.97863 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
gsave
1.025 .30902 -61 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .04984 m
.50625 .04984 L
s
[(-0.3)] .4875 .04984 1 0 Mshowa
.5 .13623 m
.50625 .13623 L
s
[(-0.2)] .4875 .13623 1 0 Mshowa
.5 .22263 m
.50625 .22263 L
s
[(-0.1)] .4875 .22263 1 0 Mshowa
.5 .39541 m
.50625 .39541 L
s
[(0.1)] .4875 .39541 1 0 Mshowa
.5 .4818 m
.50625 .4818 L
s
[(0.2)] .4875 .4818 1 0 Mshowa
.5 .56819 m
.50625 .56819 L
s
[(0.3)] .4875 .56819 1 0 Mshowa
.125 Mabswid
.5 .06712 m
.50375 .06712 L
s
.5 .0844 m
.50375 .0844 L
s
.5 .10168 m
.50375 .10168 L
s
.5 .11896 m
.50375 .11896 L
s
.5 .15351 m
.50375 .15351 L
s
.5 .17079 m
.50375 .17079 L
s
.5 .18807 m
.50375 .18807 L
s
.5 .20535 m
.50375 .20535 L
s
.5 .2399 m
.50375 .2399 L
s
.5 .25718 m
.50375 .25718 L
s
.5 .27446 m
.50375 .27446 L
s
.5 .29174 m
.50375 .29174 L
s
.5 .3263 m
.50375 .3263 L
s
.5 .34357 m
.50375 .34357 L
s
.5 .36085 m
.50375 .36085 L
s
.5 .37813 m
.50375 .37813 L
s
.5 .41269 m
.50375 .41269 L
s
.5 .42997 m
.50375 .42997 L
s
.5 .44724 m
.50375 .44724 L
s
.5 .46452 m
.50375 .46452 L
s
.5 .49908 m
.50375 .49908 L
s
.5 .51636 m
.50375 .51636 L
s
.5 .53364 m
.50375 .53364 L
s
.5 .55091 m
.50375 .55091 L
s
.5 .03256 m
.50375 .03256 L
s
.5 .01529 m
.50375 .01529 L
s
.5 .58547 m
.50375 .58547 L
s
.5 .60275 m
.50375 .60275 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
gsave
.5 .64303 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.02381 .30902 Mdot
.09544 .15305 Mdot
.18285 .09702 Mdot
.25064 .05857 Mdot
.31082 .04137 Mdot
.36261 .02582 Mdot
.41145 .02042 Mdot
.45612 .01472 Mdot
.5 .01502 Mdot
.54388 .01472 Mdot
.58855 .02042 Mdot
.63739 .02582 Mdot
.68918 .04137 Mdot
.74936 .05857 Mdot
.81715 .09702 Mdot
.90456 .15305 Mdot
.97619 .30902 Mdot
.90456 .46499 Mdot
.81715 .52101 Mdot
.74936 .55947 Mdot
.68918 .57666 Mdot
.63739 .59221 Mdot
.58855 .59762 Mdot
.54388 .60332 Mdot
.5 .60301 Mdot
.45612 .60332 Mdot
.41145 .59762 Mdot
.36261 .59221 Mdot
.31082 .57666 Mdot
.25064 .55947 Mdot
.18285 .52101 Mdot
.09544 .46499 Mdot
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7EjL30`9T0I_ooc>CTi6YZJX0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g
?0200?mV<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V
001V<`1VIP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3o
V@3oc3<003<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I
<c>IIS>IVC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0
VFH0c6H0ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVIVIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<
06K<<fK<IVK<VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITc
c9TcoiUV09UV<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWo
IYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaV
c<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<coool0
<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mVI_l08@000?l08@000?l08@000?l08@0008@0
00<10000V@0008@000<10000V@0008@000<10000V@0007T00P4800<120020I00001^00812P000`40
000700030@0000P000<100001P020HH0001_00030@00018000<100004P000`4000240000H`020Al0
00<100007@020GX0001T00030@0001d000<100007P000`40001h0000Q0020IX0001F0081;0000`40
000Y0081KP0005L000<100005`020@@000<100000`020@D000<10000:P000`40001/0000L0001040
004800040@000@@000<10000V@00070000@100012`000`4000020081VP0004P00P4P00@10P001040
004900811@000`40000h0081G`0004T000<1000090001040004;00030@00008000<10000>@000`40
001M0000L0001040004800040@000@@000<10000V@0007400P4:00811@020IX0002400030@0009T0
002400030@0009T0002400030@0009T000240081VP0008@000<10000V@0008@000<10000V@0003P0
0P5:00030@0004P00P5?0000>@000`4000180081BP000`40001=0000Q0000`40002I0000Q0000`40
002I0000Q0000`40002I0000Q0020IX0002400030@0009T0002400030@0009T0001a008110000`40
000200@110000`40002I0000L0001040004800030@0000D00P6J0000L0001040004900030@0000@0
00<10000V@0006X0104200040@000@X000<100000`000`40002I0000L0001040004;00030@000080
00<10000V@0002<00P5;00040@000@P000@1000110020Eh00P4j000090000`40001:00812P020@D0
00<10000GP000`40000h0000Q0000`40002I0000Q0000`40002I0000Q0000`40002I0000Q0020IX0
002400030@0009T0002400030@0009T0002400030@0009T000240081VP0008@000<10000V@0008@0
00<10000V@0008@000<10000V@0008@00P6J0000Q0000`40002I0000Q0000`40002I0000L@020@@0
00<100000P040@@000<10000V@00070000@100012P000`4000030081VP00070000@100012P000`40
000300030@0009T0001Z00@10P001040004:00030@0000<000<10000V@00070000@100012P000`40
000300030@0009T0001`00040@000@P00`450081VP0007400P4;00030@0000<000<10000V@0008@0
00<10000V@0008@000<10000V@0008@00P6J0000Q0000`40002I0000Q0000`40002I0000Q0000`40
002I0000;@020@@000<1000010000`40000I008110000`40000200@17`020B400P4400030@000080
104J008110000`40000400030@000400000/00040@000@X000<1000060001040004800030@000200
00<100007`001040004800030@0001X000@100012P000`4000100000;0001040004700D16@001040
004900030@0001l000<100007`001040004900030@0001T000@100011`050D40000V00@10P001040
004700040@000A@0104200040@000@X000<100007P000`40000O00040@000@X000<1000060001040
004700040@000D80000/00040@000@P000<100406P001040004;00030@0001d00P4P00040@000@/0
00<100005`001040004800030@010480000/00040@000@T00P4J00040@000@P000@100017`000`40
000O00040@000@P000@100016@00104000490081@P0002d00P4;00030@0001T00P4:008180000`40
000P00812P020A/00P4;00030@0002<00`4200030@0000<14P0008@000<10000O@00104000420003
0@0001<0000B0081L0000`40001^00813@030@<00`4C00003@3_0@P000@100010P001040004B0000
<@000`40000V00030@0002L000<100009`000`40000V00030@0002X000@100010`020A<000240003
0@0007`0104I0000Q0000`40002I0000Q0020IX0002400030@0009T0002400030@0009T000240003
0@0009T000240081VP0008@000<10000V@0008@000<10000V@0008@000<10000V@0008@00P6J0000
Q0000`40002I0000Q0000`40002I0000Q0000`40002I0000Q0020IX0002400030@0009T000240003
0@0009T0001a008110000`40000200@110000`40002I0000L0001040004:00030@0000<00P6J0000
L0001040004:00030@0000<000<10000V@00070000@100012P000`40000300030@0009T0001`0004
0@000@X000<100000`000`40002I0000L0001040004800<11@020IX0001a00812`000`4000030003
0@0009T0002400030@0009T0002400030@0009T000240081VP0008@000<10000V@0008@000<10000
V@0008@000<10000V@0008@000<10000V@0008@00P6J0000Q0000`40002I0000Q0000`40002I0000
8`020El000<10000G@020CX0000T00030@0005d00P5O00030@0003P0002400030@0009T000240003
0@0009T0001a008110000`40000200@110000`40002I0000L0001040004800030@0000D00P6J0000
L0001040004900030@0000@000<10000V@00070000@100012P000`40000300030@0009T0001`0004
0@000@/000<100000P000`40002I0000L0001040004800040@000@@00P6J0000L@020@X00P450003
0@0009T0002400030@0009T0002400030@0009T000240081VP0003P00P5:00030@0004P00P5?0000
>@000`40001800030@0004T000<10000C@0008@000<10000V@0008@00P6J0000Q0000`40002I0000
Q0000`40002I0000Q0000`40002I0000Q0020IX0002400030@0009T000180081>P000`40000h0081
G`0004T000<100009@020@@000<100000`020@D000<10000>@000`40001M0000L000104000480004
0@000@@00P6J0000L0001040004;00030@00008000<10000V@0005H00P4H00040@000@T00P450003
0@0002T00P5^0000E`000`40000F00040@000@/000<100000P000`40000Z00030@0006`0001`0004
0@000@P000@1000110000`40002I0000L@020@X00P450081VP0006<00P4O00030@0001d00P5j0000
I0000`400007008150000`40000A00812`000`40001h0000K`000`40000A008150000`4000240000
N@020@T00P480081T00007X000<100001`000`40000800030@0008h0002400030@0009T000240003
0@0009T0003o0240003o0240003o0240003o0240003o0240003o0240003o0240003o0240003o0240
003o0240001n00D100D00@0100020IH0002000090@0000400@0109L0002000090@0000400@0109L0
002000090@0000400@0109L0002000040@00008100<00@00U`0007h01@6M0000o`0Q0000o`0Q0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.0263022, -0.0154541, 
  0.00366761, 0.00366761}, {{7.125, 279.812}, {172.688, 4.1875}} ->
   {-0.637759, -0.378305, 0.00481383, 0.00484904}}]
}, Open  ]],

Cell[TextData[
"You can oriduce plots or manipulate the output algebraically.  Also, the \
package Anopt.m comes with a function for displaying the solution in 3D."], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(DiskListPlot3D[Solution]\)], "Input",
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .73536 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0175085 0.952381 [
[ 0 0 0 0 ]
[ 1 .73536 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .73536 L
0 .73536 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0175085 0.97619 0.717849 MathSubStart
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0153846 1.01538 -0.116572 1.01538 [
[.04038 .26939 -6 -8.74209 ]
[.04038 .26939 0 .25791 ]
[.27074 .18629 -5.79093 -9 ]
[.27074 .18629 .20907 0 ]
[.51797 .09716 -5.39947 -9 ]
[.51797 .09716 .60053 0 ]
[.78397 .00131 -5.00801 -9 ]
[.78397 .00131 .99199 0 ]
[.35606 .09252 -31.658 -15 ]
[.35606 .09252 2.34203 0 ]
[.84593 .01091 0 -6.10442 ]
[.84593 .01091 24 2.89558 ]
[.87841 .06064 0 -6.06699 ]
[.87841 .06064 30 2.93301 ]
[.90968 .10851 0 -6.03127 ]
[.90968 .10851 6 2.96873 ]
[.93981 .15463 0 -5.99714 ]
[.93981 .15463 24 3.00286 ]
[.96884 .19909 0 -5.9645 ]
[.96884 .19909 18 3.0355 ]
[.96976 .08807 0 -10.0521 ]
[.96976 .08807 16 4.94788 ]
[.98458 .27329 0 -5.91676 ]
[.98458 .27329 24 3.08324 ]
[.99414 .35373 0 -5.8666 ]
[.99414 .35373 6 3.1334 ]
[1.00409 .43743 0 -5.81454 ]
[1.00409 .43743 18 3.18546 ]
[1.05486 .33529 0 -9.77766 ]
[1.05486 .33529 16 5.22234 ]
[ 0 0 0 0 ]
[ 1 .73536 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03277 .28415 m
.82406 0 L
s
.04961 .2781 m
.05423 .28245 L
s
[(0)] .04038 .26939 1 .94269 Mshowa
.27939 .19559 m
.28371 .20023 L
s
[(2)] .27074 .18629 .93031 1 Mshowa
.5259 .10707 m
.52986 .11202 L
s
[(4)] .51797 .09716 .79982 1 Mshowa
.79103 .01186 m
.79456 .01713 L
s
[(6)] .78397 .00131 .66934 1 Mshowa
.125 Mabswid
.10558 .258 m
.10831 .26066 L
s
.16252 .23756 m
.1652 .24025 L
s
.22044 .21675 m
.22309 .2195 L
s
.33938 .17405 m
.34192 .17688 L
s
.40044 .15212 m
.40293 .155 L
s
.4626 .1298 m
.46504 .13272 L
s
.59035 .08392 m
.59267 .08694 L
s
.656 .06035 m
.65826 .06342 L
s
.72288 .03633 m
.72507 .03945 L
s
gsave
.35606 .09252 -92.658 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(theta) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.82406 0 m
.96523 .21594 L
s
.83398 .01518 m
.828 .01731 L
s
[(-0.5)] .84593 .01091 -1 .35654 Mshowa
.86643 .06481 m
.86043 .0669 L
s
[(-0.25)] .87841 .06064 -1 .34822 Mshowa
.89767 .1126 m
.89166 .11464 L
s
[(0)] .90968 .10851 -1 .34028 Mshowa
.92776 .15863 m
.92174 .16064 L
s
[(0.25)] .93981 .15463 -1 .3327 Mshowa
.95677 .20301 m
.95074 .20498 L
s
[(0.5)] .96884 .19909 -1 .32544 Mshowa
.125 Mabswid
.84057 .02526 m
.83698 .02653 L
s
.84711 .03526 m
.84352 .03653 L
s
.8536 .04519 m
.85001 .04645 L
s
.86004 .05504 m
.85644 .05629 L
s
.87277 .07451 m
.86917 .07576 L
s
.87906 .08414 m
.87547 .08539 L
s
.88531 .0937 m
.88171 .09494 L
s
.89151 .10318 m
.88791 .10442 L
s
.90378 .12194 m
.90017 .12316 L
s
.90984 .13122 m
.90623 .13243 L
s
.91586 .14042 m
.91225 .14163 L
s
.92183 .14956 m
.91822 .15077 L
s
.93365 .16764 m
.93003 .16884 L
s
.93949 .17658 m
.93588 .17777 L
s
.9453 .18545 m
.94168 .18664 L
s
.95106 .19427 m
.94744 .19545 L
s
.82734 .00502 m
.82375 .0063 L
s
.96245 .2117 m
.95883 .21287 L
s
gsave
.96976 .08807 -61 -14.0521 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.96523 .21594 m
1 .5093 L
s
.97248 .2771 m
.96642 .27901 L
s
[(-0.2)] .98458 .27329 -1 .31483 Mshowa
.982 .35742 m
.97592 .35926 L
s
[(0)] .99414 .35373 -1 .30369 Mshowa
.9919 .44099 m
.98581 .44277 L
s
[(0.2)] 1.00409 .43743 -1 .29212 Mshowa
.125 Mabswid
.97482 .29689 m
.97119 .29802 L
s
.97719 .31687 m
.97355 .31799 L
s
.97958 .33704 m
.97594 .33816 L
s
.98444 .378 m
.98079 .3791 L
s
.9869 .39878 m
.98325 .39987 L
s
.98939 .41978 m
.98574 .42086 L
s
.97015 .25751 m
.96652 .25866 L
s
.96785 .2381 m
.96423 .23926 L
s
.96558 .21888 m
.96195 .22005 L
s
.99444 .46241 m
.99078 .46347 L
s
.99701 .48405 m
.99335 .4851 L
s
.9996 .50592 m
.99594 .50695 L
s
gsave
1.05486 .33529 -61 -13.7777 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.03277 .28415 m
0 .57348 L
s
0 .57348 m
.20032 .73536 L
s
.20032 .73536 m
.21832 .45745 L
s
.21832 .45745 m
.03277 .28415 L
s
.82406 0 m
.96523 .21594 L
s
.96523 .21594 m
1 .5093 L
s
1 .5093 m
.85081 .30433 L
s
.85081 .30433 m
.82406 0 L
s
.03277 .28415 m
0 .57348 L
s
0 .57348 m
.85081 .30433 L
s
.85081 .30433 m
.82406 0 L
s
.82406 0 m
.03277 .28415 L
s
.21832 .45745 m
.96523 .21594 L
s
.96523 .21594 m
1 .5093 L
s
1 .5093 m
.20032 .73536 L
s
.20032 .73536 m
.21832 .45745 L
s
0 0 m
1 0 L
1 .73536 L
0 .73536 L
closepath
clip
newpath
.01 w
.34997 .31596 Mdot
.31983 .31568 Mdot
.28969 .31513 Mdot
.38017 .31901 Mdot
.41109 .32269 Mdot
.25924 .31705 Mdot
.44245 .33136 Mdot
.47523 .34228 Mdot
.22799 .31809 Mdot
.50931 .36398 Mdot
.19579 .32304 Mdot
.54674 .39711 Mdot
.16178 .32721 Mdot
.12524 .33944 Mdot
.58285 .47336 Mdot
.08355 .35613 Mdot
.03927 .4223 Mdot
.5972 .5272 Mdot
.60842 .53175 Mdot
.62299 .53124 Mdot
.63886 .52143 Mdot
.65642 .51212 Mdot
.67461 .49803 Mdot
.6938 .48463 Mdot
.71329 .46809 Mdot
.73309 .4516 Mdot
.75288 .43161 Mdot
.7723 .41063 Mdot
.79115 .38348 Mdot
.80885 .35335 Mdot
.8247 .31019 Mdot
.83681 .25325 Mdot
.25 Mabswid
.82406 0 m
.96523 .21594 L
s
.96523 .21594 m
1 .5093 L
s
1 .5093 m
.85081 .30433 L
s
.85081 .30433 m
.82406 0 L
s
.03277 .28415 m
0 .57348 L
s
0 .57348 m
.85081 .30433 L
s
.85081 .30433 m
.82406 0 L
s
.82406 0 m
.03277 .28415 L
s
.03277 .28415 m
.82406 0 L
s
.04961 .2781 m
.05423 .28245 L
s
[(0)] .04038 .26939 1 .94269 Mshowa
.27939 .19559 m
.28371 .20023 L
s
[(2)] .27074 .18629 .93031 1 Mshowa
.5259 .10707 m
.52986 .11202 L
s
[(4)] .51797 .09716 .79982 1 Mshowa
.79103 .01186 m
.79456 .01713 L
s
[(6)] .78397 .00131 .66934 1 Mshowa
.125 Mabswid
.10558 .258 m
.10831 .26066 L
s
.16252 .23756 m
.1652 .24025 L
s
.22044 .21675 m
.22309 .2195 L
s
.33938 .17405 m
.34192 .17688 L
s
.40044 .15212 m
.40293 .155 L
s
.4626 .1298 m
.46504 .13272 L
s
.59035 .08392 m
.59267 .08694 L
s
.656 .06035 m
.65826 .06342 L
s
.72288 .03633 m
.72507 .03945 L
s
gsave
.35606 .09252 -92.658 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(theta) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.96523 .21594 m
1 .5093 L
s
.97248 .2771 m
.96642 .27901 L
s
[(-0.2)] .98458 .27329 -1 .31483 Mshowa
.982 .35742 m
.97592 .35926 L
s
[(0)] .99414 .35373 -1 .30369 Mshowa
.9919 .44099 m
.98581 .44277 L
s
[(0.2)] 1.00409 .43743 -1 .29212 Mshowa
.125 Mabswid
.97482 .29689 m
.97119 .29802 L
s
.97719 .31687 m
.97355 .31799 L
s
.97958 .33704 m
.97594 .33816 L
s
.98444 .378 m
.98079 .3791 L
s
.9869 .39878 m
.98325 .39987 L
s
.98939 .41978 m
.98574 .42086 L
s
.97015 .25751 m
.96652 .25866 L
s
.96785 .2381 m
.96423 .23926 L
s
.96558 .21888 m
.96195 .22005 L
s
.99444 .46241 m
.99078 .46347 L
s
.99701 .48405 m
.99335 .4851 L
s
.9996 .50592 m
.99594 .50695 L
s
gsave
1.05486 .33529 -61 -13.7777 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 211.75},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000d`P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7EjL30`9T0I_ooc>CTi6YZJX0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g
?0200?mV<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V
001V<`1VIP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3o
V@3oc3<003<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I
<c>IIS>IVC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0
VFH0c6H0ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVIVIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<
06K<<fK<IVK<VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITc
c9TcoiUV09UV<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWo
IYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaV
c<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<coool0
<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mVI_l08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@000<800P5L0000`@001040005K0000`@001040005K0000`@030E`0003100030@0005`000310003
0@0000X00P4=008110000`40000200<1=@000<800`4600<100<00@402`001040004;00030@000380
0039008110020@/000@100012`000`40000b0000aP030@H000<1004010040@8000@1000120030CD0
003300<10P000`40000400@12@001040004800030@0003D0003000<130001040004900040@000@P0
00<10000=@000;d00`4@00040@010@T00P4900@1=0000;/00P4C00050@000040B`000;P00`4E0005
0@000040B`000;D00`4H00040@000081BP000;800`4200030@0001H000<100000`000`4000170000
[`030Ah000<100000`000`4000170000@`020@0500410@000P4200<110020@<00P400`01001>0081
8@000`40000400030@0000X00P4400030@000080104200<1:000048000H10001004200040@000@H0
00H10001004200030@0004/00`4S00030@0000D000<1000020001040004800030@0000H000<10000
9@00048000<100000P001040004200<110000`40000300<1BP030BH000<1000010020@X000@10001
2@000`40000500030@0002D0001200030@00008000@100010P001040004300030@0000D000<10000
A@040BP000<100001P00104000040@8000@100012P00104000030BP0001100@10P030@@00P4300@1
0`020DD00`4]00030@0000@00`4800040@000@/000<100000`4200030@0000<18000048000<10000
0P000`40000:00030@0004P00P4`00030@0000H000<100001P001040004800040@000@800P420004
0@000B<0001200040@00008130000`40000[00030@0001L00`4b00030@0000H00P4800812P020@<0
104300<18@00088000<1000050030CD000<1000020000`40000G00040@000@8000@10001800007l0
1@4B00<1>0000`400008008160001040004300818@0007l000@1000140030@03004003P000<10000
20020AL0104W0000P0000`400@0>0081?P000`40000:00030@0003l0002100812`030D0000<10000
2@020D40002200030@0000H00`5300030@0000/000<1000010020CP0002800<1AP000`40000;0081
10001040004g0000Q@030DT000<1000030000`40000200040@000CL0002300810`000`4000170003
0@0000`000D100000@0200030@0003D0002000<1C`000`40000;00810`001040004g0000O@030E80
00<100003@001040004200030@0003D0001j00<1E@000`40000=00810P020CP0001g00<1F0000`40
000>00030@0003X0001e00@1F@000`40000>0081>`0007800`5M00030@00010000<10000>00006l0
0`5P00030@0001000P4i0000K0030F<000<100004@000`400004008110000`40000200@10P030Al0
001Y00<1IP000`40000B00030@00008000@1000120000`40000600030@0001`0001W0081JP000`40
000@008110001040004900030@0000D000<10000700006@00`4200030@0006L000<100004P001@40
0001008000<100002000104000030Al0001Q00<1K`000`40000A00<10P001040004;00040@000B40
001400@15P030G8000<100004`001040004200030@0000H000@100010P000`40000O0000A0000`40
000D00<1M@000`40000C00810P020@X00P4300@17P0004D000<100004@020GP000<100005@000`40
000b0000AP000`40000=00<100<00@00M`000`40000D0081=00004L000<100002@030Gd000<10000
5P000`40000a0000A0001040004800<1P0000`40000F0081<P0004D00P4600<1P`000`40000G0003
0@0000@00P4400030@0000800`4N0000B`020HH000<1000060000`40000200040@000@/000<10000
6`0004P0106800030@0001H000<100400`001040004;00030@0001/0001500<110000`4000250003
0@0001P00P4200040@000@P00`4N0000@P030Hl000<100005`000`400@0200040@000@P000<10000
7P0003l00`6B00030@0001X000<100400P000`40000600030@0001h0000m0081U@000`40000I0081
0P020@T0104M0000>P040IH000<100006P020Bd0000g00<1VP000`40000G00D1;@0003@00`6M0003
0@0001@00`4400030@0002/0000a00<1X0000`40000A00<11`000`40000[0000;`020J<000<10000
3@040@X000<10000:`0002`0106T00030@0000X00`4=00<1;00002T00`6Y00030@0000H00`4B0003
0@0002X0000V00<1[0000`40000300<15@000`40000Z000030020AD00`6/008100@00@000`4H0003
0@0002X0000;00040@000A800P6`00030@0100816`000`40000Z00002`001040004?00@1/0040A`0
0P4/00002`001040004<00<1/0040@<000<100006`000`400008008110000`40000200@14`0000/0
00@100012@030K000`4700030@0001/000<100001`001040004800030@0001@0000;00040@000@H0
0`6`00<12P000`40000K00030@0000L000@100012@000`40000C000030020@D00P6`00<13@000`40
000K00810P040@8000@100012P000`40000B000040040Jh00`4@00030@0001X000<1004020001040
004;00030@000140000>008110000`40002X00<150000`40000K00030@0000H000@1000120001040
004C00003P020J`00`4G00030@0001/000<100001`020@X00P4D00003P000`400@2W00@16P000`40
000K00030@0002T0000>00040@000J<00`4N00030@0001/000<10000:@0000h000D100000@2O00<1
8@000`40000J0081:`0000h000<100000P020IX00`4S00817@000`40000Y00003@000`4000050003
0@0009@00`4P008100@00@410P000`40000J00030@0002T0000=00030@0000H000<10000<`020E/0
0`4S00<11@000`40000K00030@0002P0000=00030@0000L000<100008`020@H00P4600030@0000@0
0P4500811@020D800`4S00<12@000`40000I00814`050@0500400@000P4;00003@000`4000080003
0@0001/00P4600030@0000D000<1000030000`40000400030@0000@000<1000010020CH0104R00@1
3@000`40000I00030@0001<000T100000@010040300000d000<100002@000`40000A008120000`40
000a00030@0003400`4S00<14P000`40000H00030@0001<000T100000@010040300000d000<10000
2P000`40000A00030@0004400P4X00<18`030AD000<1000060000`40000C00090@0000400@0100`0
000=00030@0000/000<10000E@000`40000S00<18`030AT000<100005`000`40000C00040@000081
00<00@00300000d000<1000030000`4000050081L0030B<00`4M00030@0001D00P4C00D14P0000d0
00<100003@000`40000500030@0005800P4G00<18`030B0000<100005P000`40000X00003@000`40
000>00030@0005X000<100004P030B<00`4T00030@0001D000<100000`020B<0000<00030@000100
0P5/00<18P040BP000<100005@001@400001008000<10000800000`000<100002`020@D000<10000
I@040B800`4I00814P000`40000D00050@0000400P000`40000P000030000`40000<00030@0000@0
00<10000H@030B<00`4M00030@00010000<100004P030@<000@100018P0000`000<1000050000`40
001L00@18`030C@000<100004`001@400001008000<10000800000`000<100005@000`40001I00@1
8P030CP000<100004P001@400001008000<10000800000`000<100005P000`40001E00<18`030C`0
00<100004@000`40000200818`0000`000<100005`000`40001A00<18`030Cl000<100004@000`40
000W000030000`40000H00030@0004d00`4R00@1@`000`40000?0081:@0000`000<100006@000`40
001800@18P030B/00P4K00030@00010000<100009P0000`000<100006P000`40001400<18`030Bl0
00<100006@000`40000@00030@0002H0000;00030@0001`00P5100<18`030Dl000<100003`000`40
000V00002`000`40000N00030@0003/00`4N00810`030E<000<100003P000`40000V00002`000`40
000O00030@0003L00`4R00@1E`000`40000<0081:00000/000<1000080000`40000c00<18`030EX0
00<100003@000`40000V00002`000`40000Q00030@0002l00`4S00<1GP000`40000<00030@0002H0
000;00030@00028000<10000:P040B80104k00819@000`40000;00030@0002H0000;00030@0002<0
00<100009P030B<00`5000030@0002<000<100002`000`40000V00002`00104000020B4000<10000
8P030B<00`5Z00030@0000X00P4W00002`000`40000200030@00020000<100007P030B<00`5^0003
0@0000X000<100009@0000/000<100009P000`40000J00<18`030G8000<100002@000`4000020081
10000`40000200@1500000X000<10000:0000`40000F00<18`030DH00P4]00030@0000T000D10000
0@0200030@0000H000<100005@0000X000<10000:@020A<00`4S00<1BP000`40000/00030@0000P0
00D100000@0200030@0000L000<10000500000X000<10000:`000`40000<00@18P040Gd000<10000
1`001@400001008000<1000020000`40000C00002P000`40000/00030@0000P00`4S00<1P@000`40
000600810`001040004;00030@000180000:00030@0002d000<1000010030B<00`6500030@0000H0
00D100000@0200030@0000H000@10001500000X000<10000;P000`4000030B<00`5A0081=P000`40
000500030@0000800P4:00815@0000X000<10000;`020B<00`5E00030@0003D000<100001@000`40
000T00002P000`40000_00030@0001l00`6@00030@0000D000<10000900000X000<10000;`000`40
000L00<1U0000`40000300819P0000X000<10000;`000`40000H00@1F@020Cd000<100000`000`40
000T00002@000`40000`00030@0001D00`4m00817`000`40000l00030@00008000<10000900000T0
00<10000<0000`40000B00<1@@000`40001L00030@00008000<10000900000T000<10000<0000`40
000?00<1Y0001@40000102H0000900030@0002l000<100003@030F400P5500040@010BH000090003
0@0002l000<100002P030FD000<10000@`001040004V00002@000`40000_00030@0000L00`6_0004
0@000BD0000900030@0002l000<1000010030FD00P5<00030@0102D0000900030@0002l000<10000
105Y00030@0004/00P4U00002@000`40000_00<1^`020BD0000900030@0002`0105Z0081D@020BD0
000800030@0002X00`4300030@0006T000<10000B`040BL0000800030@0002L00`4600030@000;@0
0`4[000020000`40000T00<12@000`40001S0081B`040Bh0000800030@0002400`4<00030@0005T0
0P4900030@0004H00`4b000020000`40000M00@13`000`40001J00040@010@800P5700@1=@0000P0
00<100006P030A<000<10000G@001@40000104<0104i000020000`40000G00<15P000`40002R00<1
?@0000P000<1000050030AP000<10000W`040D00000800030@0001400`4K00030@0009`00`540000
20000`40000>00<17P000`40002H00@1A`0000L000<1000030030B4000<10000U@030D/000070003
0@0000P0104T00030@000940105>00001`000`40000500<1:0000`40002>00<1DP0000L000<10000
0P030B/000<10000RP040ED0000700030@010081;P000`40002700<1F@0000L00P4a00030@0008<0
105L000020000`40000_00030@0007l0105P00002@020Bl000<10000O0030F@0000;00030@0002`0
00<10000N0040FL0000<00030@0002/000<10000M@030F/0000=00030@0002X000<10000L@040Fh0
000>0081:@000`40001_00<1LP00010000<100009P000`40001[00@1M@00014000<100009@000`40
001X00<1N@00018000<1000090000`40001T00@1O00001<000<100008`000`40001Q00<1P00001@0
0P4S00030@0005d0106300005P000`40000P00030@0005T0106700005`000`40000O00030@0005H0
0`6;000060000`40000N00030@000580106>00006@020Ah000<10000C`030I80000K00030@0001/0
00<10000B`040ID0000L00030@0001X000<10000B0030IT0000M00030@0001T000<10000A0040I`0
000N008160000`40001200<1X000020000<100005@000`40000n00@1X`00024000<1000050000`40
000k00<1Y`00028000<100004`000`40000g00@1ZP0002<000<100004P000`40000c00@1[P0002@0
0P4B00030@0003000`6b00009P000`40000?00030@0002`0106e00009`000`40000>00030@0002T0
0`6i0000:0000`40000=00030@0002D0106l0000:@020@d000<100008P030L00000[00030@0000X0
00<100007P040L<0000/00030@0000T000<100006`030LL0000]00030@0000P000<100005`040LX0
000^008120000`40000D00<1cP00030000<1000010000`40000A00@1d@00034000<100000`000`40
000=00@1e@00038000<100000P000`40000:00<1f@0003<000D100000@0800@1g00003@00P400`01
000400<1h00003H00P400`010@020N<0000g0081i`000?l08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {210.75, 0}} -> {-0.0257484, -0.0183878, 
  0.00366375, 0.00366375}, {{7, 279.938}, {205.688, 5}} -> {-0.0131572, 
  0.0331709, 0.00404337, 0.00404337}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  GraphicsArray  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[
"Example 2:  the case of vector valued analytic functions"], "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["The problem"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "We  consider now a problem of optimization over A",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " , which is the space of  pairs ( f",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " , f",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " ) of analytic functions.  \n\nWe want to find f* that solves OPT for the \
performance function:\n\n",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  "( e ,  z",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " , z",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " )  =  Re[ 1/e + z1",
  StyleBox[" ]^2 + 4 Im[ 1/e + z1 ",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "]^2   +  Re[ 1/e + z2",
  StyleBox[" ]^2 + 0.3 Im[ 1/e + z2 ",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "]^2"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Input"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  " To translate this to the Mathematica language, proceed as before and \
replace z",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " by  z[1] , z",
  StyleBox["2 by  z[2], etc.  The full set of inputs is",
    FontVariations->{"CompatibilityType"->"Subscript"}]
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(g = 
      Re[1\/e + z[1]]\^2 + 4\ Im[1\/e + z[1]]\^2 + Re[1\/e + z[2]]\^2 + 
        0.3\ Im[1\/e + z[2]]\^2; \)\), 
    \(\(Anopt[g, 0.001]; \)\)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 4.3            E+00 :   N/A   : 5.3E-01 : 7.2E-01D:   N/A   : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 1 : 2.1545988017648E+00 : 7.8E-01 : 3.2E-06 : 5.5E-02D: 3.2E-05 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 2 : 2.1485374717714E+00 : 7.6E-02 : 1.5E-03 : 1.2E-02D: 5. E-05 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 3 : 2.1469189941809E+00 : 7.5E-03 : 3.7E-06 : 5.9E-04D: 5.3E-06 : 0A5 \
: 32"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 2.146918994180885E+00"\)], "Print"],

Cell[BoxData[
    \("flat      = 3.6801434565E-06"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 5.9442841631E-04"\)], "Print"],

Cell[BoxData[
    \("ned      = 5.3 E-06"\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Stopping and how good is the solution"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"The run stops when the inequalities\n\n\t\tFlat < tol ,    \t GrAlign < tol \
,\n\nare satisfied,  where tol is the error tolerance set by the user.  At \
the solution one must have that Flat = GradAlign = 0, so small Flat and small \
GradAlign is a necessary condition for a calcultated guess at the answer to \
be close to the actual solution.  \n\tSometimes there is difficulty in the \
calculations.  Anopt[] has basically two ways to deal with it: smoothing, and \
, grid size doubling.  Both are performed automatically if the internal \
algorithms of Anopt indicate that such action is necessary.  The user may \
supress smoothing or grid doubling by specifying options when Anopt is run.  \
In the run above smoothing occurred at the third iteration, while the grid \
size remained constant at 32."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Fourier coefficients of the solution."], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Computing the (discrete) Fourier coefficients of functions is useful for \
many purposes.  The Mathematica command Fourier[list], where list = {x",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",x",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ",...x",
  StyleBox["n",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  "}, is an implementation of the FFT algorithm and returns a list of the \
Fourier coefficients (times the numerical factor Sqrt[n], where n=number of \
sample points).\t\nTo illustrate the use of Fourier[ ], we compute the \
Fourier coefficients of Solution[[2]], and store them in the name ``fc2'': "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(fc2 = 
      Fourier[Solution\[LeftDoubleBracket]2\[RightDoubleBracket]]\/\@32.; 
    \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"The resulting coefficients are ordered according to the indices in { 0 , -1 \
, -2 , -3 , . . . , . . . , 3 , 2 , 1 }.  We see below that  there is only \
one nontrivial Fourier coefficient in the first entry of Solution:"], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Chop[fc2]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \({0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
      \(-5.93687133864991167`*^-9\), 0, 2.4815591546416269`*^-8, 0, 
      5.03699178346398213`*^-9, 0, 1.02097676463780229`*^-9, 0, 
      2.06355528378299668`*^-10, 0, 0, 0, 0, 0, \(-0.681920250138719019`\)}
      \)], "Output"]
}, Open  ]],

Cell[TextData[{
  "Note that it is clear from the expression above that the function f2 of \
the calculated solution f = ( f1 , f2 ) is a multiple of  e",
  StyleBox["i",
    FontVariations->{"CompatibilityType"->"Superscript"}],
  StyleBox["\[Theta]",
    FontFamily->"Symbol",
    FontVariations->{"CompatibilityType"->"Superscript"}]
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Example 3.  Specification of more input."], "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Anopt is based upon an iterative scheme that generates a sequence of \
(discrete) analytic functions f",
  StyleBox["1",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ", f",
  StyleBox["2",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ", f",
  StyleBox["3",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ", etc.  This process begins with an initial guess f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ", which in examples 1 and 2 has been defaulted to the function zero, on a \
discrete grid of 16 points.  In this section you will learn how to start \
Anopt on an Initial guess f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " that you specify as a rational function sampled on the circle.  "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData[
"Statement of the problem: example 1 revisited."], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"Suppose we want to solve example 1, only that we want to start the iteration \
with the initial function  \nf(e) = ( e^2 - 3 e + 4) / ( e + 3 ).   We define \
the performance first."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(g = Abs[0.8 + \((1\/e + z[1])\)\^2]\^2; \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"We can specify the initial guess at solution as an expression in \"e\" ( \
represents Exp[I theta] ).   For example, we set "], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(f0 = {\(e\^2 - 3\ e + 4\)\/\(e + 3\)}; \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[{
  "The performance function we deal with here has only one complex variable \
(N=1), hence only one  entry is required in f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ".  In the case N>1 the different  entries of f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " have to be separated by commas.\n\nThe calling sequence to be used now \
takes as input the function f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " as initial guess, besides the performance g and the tolerance tol.  The \
sequence is Anopt[ g , tol , f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " ].  Many other inputs can be specified as options.  Below we set the \
maximum number of iterations to be 3:"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Anopt[g, .0001, f0, Iterations \[Rule] 3]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 9.6039969948935E+01 :   N/A   : 9.3E-01 : 1. E+00 :   N/A   : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.9365155086078E+00 : 3.8E+00 : 9.1E-01 : 0  E+00 : 2.5E-06 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 2 : 1.1423579024226E+00 : 2.8E-01 : 2.4E-01 : 0  E+00 : 3.9E-04 : \
5A10 : 32"\)], "Print"],

Cell[BoxData[
    \(" 3 : 1.0093525941247E+00 : 6.7E-02 : 2.1E-02 : 0  E+00 : 2.1E-03 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.009352594124695E+00"\)], "Print"],

Cell[BoxData[
    \("flat      = 2.0899516674E-02"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 2.12E-03"\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[TextData[{
  "Another valid way to specify the initial guess is as a list of values \
obtained by sampling the function on a grid of equally spaced points.To \
produce a discrete version of f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " on a 32 point grid you can use a replacement rule.  Note the braces \
around the rational function:"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(f0
       = {\(e\^2 - 3\ e + 4\)\/\(e + 3\)} /. {
          e \[Rule] Table[N[Exp[\(2\ \[Pi]\ I\ i\)\/32]], {i, 0, 31}]}; \)\)],
   "Input",
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["Restarting the run"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"In the run above, the process stopped because the limit on the number of \
iterations was attained.  If you want Anopt to proceed from the last \
iteration above,  you can restart the iteration.  For this take as an initial \
guess for the new run the last iterate of the previous run, which is stored \
in the variable Solution.   \n\nBelow we double the gridsize of the answer we \
obtained in the run above and assign it to f1.  Then f1 is used to restart \
the iteration."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(f1 = DoubleGrid[Solution]; \)\), 
    \(Anopt[g, 0.00002, f1]\)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 1.0093525941247E+00 :   N/A   : 2.1E-02 : 0  E+00 :   N/A   : NON \
: 64"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.0005560979092E+00 : 5.7E-03 : 1.1E-03 : 0  E+00 : 2.1E-03 : NON \
: 64"\)], "Print"],

Cell[BoxData[
    \(" 2 : 1.0002780415569E+00 : 9. E-04 : 5.6E-04 : 0  E+00 : 1.2E-03 : \
5A10 : 64"\)], "Print"],

Cell[BoxData[
    \(" 3 : 1.000033365362 E+00 : 1.6E-04 : 6.7E-05 : 0  E+00 : 2. E-03 : NON \
: 64"\)], "Print"],

Cell[BoxData[
    \(" 4 : 1.0000050181354E+00 : 1.6E-05 : 1.3E-05 : 0  E+00 : 3.1E-03 : NON \
: 64"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.000005018135406E+00"\)], "Print"],

Cell[BoxData[
    \("flat      = 1.291679907 E-05"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 3.1 E-03"\)], "Print"]
}, Open  ]],

Cell[TextData["The solution has the following structure."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dimensions[Solution]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \({1, 64}\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Manipulating solutions"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "As an example of how to manipulate output, we plot  below the discrete \
values of the function \n ",
  StyleBox["\[CapitalGamma]",
    FontFamily->"Symbol"],
  " ( e  , f* ( e ) ).   We then compare this graph with the claim (from the \
theory) that gdisc( e )  is constant.  For this one must substitute in the \
formula g = g( e , z [1] ) , the variable z[1] by Solution[[1]], and e by its \
discrete values on a grid of equally spaced points.  This is accomplished \
using replacement rules as shown below.  First we produce a discrete version \
of Exp[I theta] for theta in [0, 2 Pi]."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(\(ngrid = 64; \)\), 
    \(\(edisc = 
      Table[N[Exp[\(2\ \[Pi]\ I\ \((i - 1)\)\)\/ngrid]], {i, 1, ngrid}]; 
    \)\)}], "Input",
  AspectRatioFixed->True],

Cell[TextData["Now we use replacement rules."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(gdisc = 
      g /. {e \[Rule] edisc, 
          z[1] \[Rule] Solution\[LeftDoubleBracket]1\[RightDoubleBracket]}; 
    \)\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(gdisc\)\" is similar to \
existing symbol \"\!\(edisc\)\"."\)], "Message"]
}, Open  ]],

Cell[TextData[
"The plot is produced below.  Here we specify a range:"], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(ListPlot[gdisc, PlotRange \[Rule] {0, 2}]\)], "Input",
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.014881 0 0.309017 [
[.17262 -0.0125 -6 -9 ]
[.17262 -0.0125 6 0 ]
[.32143 -0.0125 -6 -9 ]
[.32143 -0.0125 6 0 ]
[.47024 -0.0125 -6 -9 ]
[.47024 -0.0125 6 0 ]
[.61905 -0.0125 -6 -9 ]
[.61905 -0.0125 6 0 ]
[.76786 -0.0125 -6 -9 ]
[.76786 -0.0125 6 0 ]
[.91667 -0.0125 -6 -9 ]
[.91667 -0.0125 6 0 ]
[.01131 .07725 -24 -4.5 ]
[.01131 .07725 0 4.5 ]
[.01131 .15451 -18 -4.5 ]
[.01131 .15451 0 4.5 ]
[.01131 .23176 -24 -4.5 ]
[.01131 .23176 0 4.5 ]
[.01131 .30902 -6 -4.5 ]
[.01131 .30902 0 4.5 ]
[.01131 .38627 -24 -4.5 ]
[.01131 .38627 0 4.5 ]
[.01131 .46353 -18 -4.5 ]
[.01131 .46353 0 4.5 ]
[.01131 .54078 -24 -4.5 ]
[.01131 .54078 0 4.5 ]
[.01131 .61803 -6 -4.5 ]
[.01131 .61803 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.17262 0 m
.17262 .00625 L
s
[(10)] .17262 -0.0125 0 1 Mshowa
.32143 0 m
.32143 .00625 L
s
[(20)] .32143 -0.0125 0 1 Mshowa
.47024 0 m
.47024 .00625 L
s
[(30)] .47024 -0.0125 0 1 Mshowa
.61905 0 m
.61905 .00625 L
s
[(40)] .61905 -0.0125 0 1 Mshowa
.76786 0 m
.76786 .00625 L
s
[(50)] .76786 -0.0125 0 1 Mshowa
.91667 0 m
.91667 .00625 L
s
[(60)] .91667 -0.0125 0 1 Mshowa
.125 Mabswid
.05357 0 m
.05357 .00375 L
s
.08333 0 m
.08333 .00375 L
s
.1131 0 m
.1131 .00375 L
s
.14286 0 m
.14286 .00375 L
s
.20238 0 m
.20238 .00375 L
s
.23214 0 m
.23214 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.29167 0 m
.29167 .00375 L
s
.35119 0 m
.35119 .00375 L
s
.38095 0 m
.38095 .00375 L
s
.41071 0 m
.41071 .00375 L
s
.44048 0 m
.44048 .00375 L
s
.5 0 m
.5 .00375 L
s
.52976 0 m
.52976 .00375 L
s
.55952 0 m
.55952 .00375 L
s
.58929 0 m
.58929 .00375 L
s
.64881 0 m
.64881 .00375 L
s
.67857 0 m
.67857 .00375 L
s
.70833 0 m
.70833 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.79762 0 m
.79762 .00375 L
s
.82738 0 m
.82738 .00375 L
s
.85714 0 m
.85714 .00375 L
s
.8869 0 m
.8869 .00375 L
s
.94643 0 m
.94643 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
.02381 .07725 m
.03006 .07725 L
s
[(0.25)] .01131 .07725 1 0 Mshowa
.02381 .15451 m
.03006 .15451 L
s
[(0.5)] .01131 .15451 1 0 Mshowa
.02381 .23176 m
.03006 .23176 L
s
[(0.75)] .01131 .23176 1 0 Mshowa
.02381 .30902 m
.03006 .30902 L
s
[(1)] .01131 .30902 1 0 Mshowa
.02381 .38627 m
.03006 .38627 L
s
[(1.25)] .01131 .38627 1 0 Mshowa
.02381 .46353 m
.03006 .46353 L
s
[(1.5)] .01131 .46353 1 0 Mshowa
.02381 .54078 m
.03006 .54078 L
s
[(1.75)] .01131 .54078 1 0 Mshowa
.02381 .61803 m
.03006 .61803 L
s
[(2)] .01131 .61803 1 0 Mshowa
.125 Mabswid
.02381 .01545 m
.02756 .01545 L
s
.02381 .0309 m
.02756 .0309 L
s
.02381 .04635 m
.02756 .04635 L
s
.02381 .0618 m
.02756 .0618 L
s
.02381 .09271 m
.02756 .09271 L
s
.02381 .10816 m
.02756 .10816 L
s
.02381 .12361 m
.02756 .12361 L
s
.02381 .13906 m
.02756 .13906 L
s
.02381 .16996 m
.02756 .16996 L
s
.02381 .18541 m
.02756 .18541 L
s
.02381 .20086 m
.02756 .20086 L
s
.02381 .21631 m
.02756 .21631 L
s
.02381 .24721 m
.02756 .24721 L
s
.02381 .26266 m
.02756 .26266 L
s
.02381 .27812 m
.02756 .27812 L
s
.02381 .29357 m
.02756 .29357 L
s
.02381 .32447 m
.02756 .32447 L
s
.02381 .33992 m
.02756 .33992 L
s
.02381 .35537 m
.02756 .35537 L
s
.02381 .37082 m
.02756 .37082 L
s
.02381 .40172 m
.02756 .40172 L
s
.02381 .41717 m
.02756 .41717 L
s
.02381 .43262 m
.02756 .43262 L
s
.02381 .44807 m
.02756 .44807 L
s
.02381 .47898 m
.02756 .47898 L
s
.02381 .49443 m
.02756 .49443 L
s
.02381 .50988 m
.02756 .50988 L
s
.02381 .52533 m
.02756 .52533 L
s
.02381 .55623 m
.02756 .55623 L
s
.02381 .57168 m
.02756 .57168 L
s
.02381 .58713 m
.02756 .58713 L
s
.02381 .60258 m
.02756 .60258 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.03869 .30902 Mdot
.05357 .30901 Mdot
.06845 .30902 Mdot
.08333 .30902 Mdot
.09821 .30902 Mdot
.1131 .30902 Mdot
.12798 .30902 Mdot
.14286 .30902 Mdot
.15774 .30902 Mdot
.17262 .30902 Mdot
.1875 .30902 Mdot
.20238 .30902 Mdot
.21726 .30902 Mdot
.23214 .30902 Mdot
.24702 .30902 Mdot
.2619 .30902 Mdot
.27679 .30902 Mdot
.29167 .30902 Mdot
.30655 .30902 Mdot
.32143 .30902 Mdot
.33631 .30902 Mdot
.35119 .30902 Mdot
.36607 .30902 Mdot
.38095 .30902 Mdot
.39583 .30902 Mdot
.41071 .30902 Mdot
.4256 .30902 Mdot
.44048 .30902 Mdot
.45536 .30902 Mdot
.47024 .30902 Mdot
.48512 .30902 Mdot
.5 .30901 Mdot
.51488 .30902 Mdot
.52976 .30901 Mdot
.54464 .30902 Mdot
.55952 .30902 Mdot
.5744 .30902 Mdot
.58929 .30902 Mdot
.60417 .30902 Mdot
.61905 .30902 Mdot
.63393 .30902 Mdot
.64881 .30902 Mdot
.66369 .30902 Mdot
.67857 .30902 Mdot
.69345 .30902 Mdot
.70833 .30902 Mdot
.72321 .30902 Mdot
.7381 .30902 Mdot
.75298 .30902 Mdot
.76786 .30902 Mdot
.78274 .30902 Mdot
.79762 .30902 Mdot
.8125 .30902 Mdot
.82738 .30902 Mdot
.84226 .30902 Mdot
.85714 .30902 Mdot
.87202 .30902 Mdot
.8869 .30902 Mdot
.90179 .30902 Mdot
.91667 .30902 Mdot
.93155 .30902 Mdot
.94643 .30902 Mdot
.96131 .30902 Mdot
.97619 .30901 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  Evaluatable->False,
  AspectRatioFixed->True,
  ImageSize->{282, 174},
  ImageMargins->{{34, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004J0000[PP10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7EjL30`9T0I_ooc>CTi6YZJX0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g
?0200?mV<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V
001V<`1VIP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3o
V@3oc3<003<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I
<c>IIS>IVC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0
VFH0c6H0ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVIVIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<
06K<<fK<IVK<VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITc
c9TcoiUV09UV<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWo
IYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaV
c<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<coool0
<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mVI_l06`000?l06`000?l06`000?l06`0003l0
1043008170040@<00P4N008110020Al000<100000P020A`00`4400817P020@@00P4F0000@@001@40
0001008000<100006@000`40000300040@000A`000@100010P001040004N00050@0000400P000`40
000L00040@000@8000<100006P001040004200040@000AD0001100050@0000400P000`40000J0003
0@00008000@100017`001040004200030@0001T01@4200040@000Ah000@100010P000`40000J0004
0@000@8000@100015@00044000D100000@0200030@0001/000D100000@0200030@0001/00P430004
0@000A/000@100010`001040004K00<10`001040004L00<10`001040004E0000@@001@4000010080
00<1000070001040004200030@0001d000@100010P000`40000J00030@0100<000@100016`000`40
000300040@000A`000<100000`001040004E0000?`030@<000@100016`001040004200040@000A`0
00@100010P001040004M00810`001040004K00030@0000<000@1000170000`40000300040@000AD0
001100030@0000800P4M008110020Ah00P4400817`000`400002008170040@<00P4N00<10`020AH0
003o01/0003o01/0000H0?h1100001h000<100001@000`40000400030@0000D000<1000010000`40
000500030@0000@000<100001@000`40000400030@0000D000<1000010000`40000500030@0000@0
00<100001@000`40000500030@0000@000<100001@000`40000400030@0000D000<1000010000`40
000500030@0000@000<100001@000`40000400030@0000D000<1000010000`40000500030@0000@0
00<100001@000`40000400030@0000D000<100001@000`40000400030@0000P0000N00030@000?T0
000N00030@000?T0000N0081nP0001h000<10000n@0001h000<10000n@0001h00P7j00007P000`40
003i00007P000`40003i00007P000`40003i00007P020OX0000N00030@000?T0000N00030@000?T0
000N00030@000?T0000N0081nP0001h000<10000n@0001h000<10000n@0000D00P4400030@000080
104200<11@000`40003i000010001040004800030@0000H000<100000P030OT0000400040@000@T0
00<100001@000`40000200030@000?T0000400040@000@X000@100000`4500030@000?T000040004
0@000@/000@100011`000`40003i000010001040004800040@000@8000<100001@020OX000050081
2P020@<0104400030@000?T0000N00030@000?T0000N00030@000?T0000N0081nP0001h000<10000
n@0001h000<10000n@0001h000<10000n@0001h00P7j00007P000`40003i00007P000`40003i0000
7P000`40003i00007P020OX0000N00030@000?T0000;008110000`40000200<11@000`40003i0000
2P001040004;00030@00008000<10000n@0000X000@100012`000`40000200<1n@0000X000@10001
20030@D000<10000n@0000X000@1000120000`40000500030@000?T0000:00040@000@P000<10000
1@000`40003i00002`020@T010440081nP0001h000<10000n@0001h000<10000n@0001h000<10000
n@0001h00P7j00007P000`40003i00007P000`40003i00007P020OX0000N00030@000?T0000N0003
0@000?T0000N00030@000?T0000N0081nP0001h000<10000n@0001h000<10000n@0000D00P440003
0@0000<000<100000P030@D000<10000n@0000@000@100012@000`40000500030@0000800`7i0000
10001040004:00030@0000@000<100000P000`40003i000010001040004:00040@0000<11@000`40
003i000010001040004;00040@000@L000<10000n@0000@000@1000120001040004200030@0000D0
0P7j00001@020@T0104200@110000`40003i00007P000`40003i00007P000`40003i00007P020OX0
000N00030@000?T0000N00030@000?T0000N00030@000?T0000N0081nP0001h000<10000n@0001h0
00<10000n@0001h000<10000n@0001h00P7j00007P000`40003i00007P000`40003i00005P040@@0
00<100000P4200810P020@800P400`010@0200810P020@800P400`010@0200810P020@800P420081
00<00@400P020@800P42008100<00@400P020@800P42008100<00@400P020@800P42008100<00@40
0P020@800P4200810P020@03004100800P4200810P020@03004100800P4200810P020@0300410080
0P4200810P020@03004100800P4200810P020@800P400`010@0200810P020@800P400`010@020081
0P020@800P400`010@0200810P020@800P400`010@0200810P020@800P4200812@0001P000<10000
0`030@030040008000D100000@0300040@000@<000D100000@0300040@000@<000D100000@030005
0@0000400P001@40000100<000D100000@0200050@0000400`001@400001008000D100000@030005
0@0000400P001@40000100<000D100000@0300040@000@<000D100000@0300040@000@<000D10000
0@0300040@000@<000D100000@0300040@000@<000D100000@0300050@0000400P001@40000100<0
00D100000@0200050@0000400`001@400001008000D100000@0300050@0000400P001@40000100<0
00D100000@0300030@0000L0000H00030@0000<000<10000n@0001P000<100000`000`40003i0000
60000`40000300030@000?T0000F00<11@020OX0000H00030@0000<000<10000n@0001h000<10000
n@0001h000<10000n@0001h00P7j00007P000`40003i00007P000`40003i00007P000`40003i0000
7P020OX0000N00030@000?T0000N00030@000?T0000N0081nP0001h000<10000n@0001h000<10000
n@0000@0104300030@000080104200<11@000`40003i00001P000`40000700030@0000H000<10000
0P030OT0000600030@0000P000<100001@000`40000200030@000?T0000600030@0000T000@10000
0`4500030@000?T0000600030@0000X000@100011`000`40003i000010030@T000@100010P000`40
00050081nP0000H000<1000020020@<0104400030@000?T0000N00030@000?T0000N00030@000?T0
000N0081nP0001h000<10000n@0001h000<10000n@0001h000<10000n@0001h00P7j00007P000`40
003i00007P000`40003i00007P000`40003i00007P020OX0000N00030@000?T0000N00030@000?T0
000:00@10`000`40000200<11@000`40003i000030000`40000:00030@0000800`7i000030000`40
000:00030@00008000<10000n@0000`000<100001`030@D000<10000n@0000`000<100001`000`40
000500030@000?T0000:00<12@000`4000050081nP0000`000<100001`040@@000<10000n@0001h0
00<10000n@0001h000<10000n@0001h00P7j00007P000`40003i00007P000`40003i00007P000`40
003i00007P020OX0000N00030@000?T0000N00030@000?T0000N00030@000?T0000N0081nP0001h0
00<10000n@0000@0104300030@0000<000<100000P030@D000<10000n@0000H000<1000020000`40
000500030@0000800`7i00001P000`40000900030@0000@000<100000P000`40003i00001P000`40
000900040@0000<11@000`40003i00001P000`40000:00040@000@L000<10000n@0000@00`490004
0@000@8000<100001@020OX0000600030@0000L0104200@110000`40003i00007P000`40003i0000
7P000`40003i00007P020OX0000N00030@000?T0000N00030@000?T0000N00030@000?T0000N0081
nP0001h000<10000n@0001h000<10000n@0001h000<10000n@0001h00P7j00007P000`40003i0000
7P000`40003i00005P040@@000<10000n@0001H000<100001@030OT0000G00030@000?l00@0001P0
00<10000o`0001T000<10000oP0001H000@10001o`010000\
\>"],
  ImageRangeCache->{{{0, 281}, {173, 0}} -> {-8.13241, -0.155597, 0.265655, 
  0.0127928}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell[TextData[
"We conclude that g1 is approximately constant.  Let us find out how close to \
constant:"], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(Max[gdisc] - Min[gdisc]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(0.0000129168638880461372`\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Combining algebra and plots"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"Now we try something a bit different. Suppose you wish to compute and plot \
in 3D the partial derivative of the performance with respect to z, when z = \
our calculated solution.   This requires only two lines of input for the \
user.  We also plot the system on the complex plane for comparison.  The \
function grad defined below is discrete, that is, it is a list of values."], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(grad = 
      ComplexD[g, z[1]] /. {e \[Rule] edisc, 
          z[1] \[Rule] Solution\[LeftDoubleBracket]1\[RightDoubleBracket]}; 
    \)\)], "Input",
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(DiskListPlot[grad, PlotJoined \[Rule] True]\)], "Input",
  PageBreakBelow->False,
  GroupPageBreakWithin->Automatic,
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.233098 0.309017 0.10968 [
[.0338 .29652 -6 -9 ]
[.0338 .29652 6 0 ]
[.2669 .29652 -6 -9 ]
[.2669 .29652 6 0 ]
[.7331 .29652 -3 -9 ]
[.7331 .29652 3 0 ]
[.9662 .29652 -3 -9 ]
[.9662 .29652 3 0 ]
[1.025 .30902 0 -7.5 ]
[1.025 .30902 16 7.5 ]
[.4875 .08966 -12 -4.5 ]
[.4875 .08966 0 4.5 ]
[.4875 .19934 -12 -4.5 ]
[.4875 .19934 0 4.5 ]
[.4875 .4187 -6 -4.5 ]
[.4875 .4187 0 4.5 ]
[.4875 .52838 -6 -4.5 ]
[.4875 .52838 0 4.5 ]
[.5 .64303 -8 0 ]
[.5 .64303 8 15 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.0338 .30902 m
.0338 .31527 L
s
[(-2)] .0338 .29652 0 1 Mshowa
.2669 .30902 m
.2669 .31527 L
s
[(-1)] .2669 .29652 0 1 Mshowa
.7331 .30902 m
.7331 .31527 L
s
[(1)] .7331 .29652 0 1 Mshowa
.9662 .30902 m
.9662 .31527 L
s
[(2)] .9662 .29652 0 1 Mshowa
.125 Mabswid
.08042 .30902 m
.08042 .31277 L
s
.12704 .30902 m
.12704 .31277 L
s
.17366 .30902 m
.17366 .31277 L
s
.22028 .30902 m
.22028 .31277 L
s
.31352 .30902 m
.31352 .31277 L
s
.36014 .30902 m
.36014 .31277 L
s
.40676 .30902 m
.40676 .31277 L
s
.45338 .30902 m
.45338 .31277 L
s
.54662 .30902 m
.54662 .31277 L
s
.59324 .30902 m
.59324 .31277 L
s
.63986 .30902 m
.63986 .31277 L
s
.68648 .30902 m
.68648 .31277 L
s
.77972 .30902 m
.77972 .31277 L
s
.82634 .30902 m
.82634 .31277 L
s
.87296 .30902 m
.87296 .31277 L
s
.91958 .30902 m
.91958 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
gsave
1.025 .30902 -61 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .08966 m
.50625 .08966 L
s
[(-2)] .4875 .08966 1 0 Mshowa
.5 .19934 m
.50625 .19934 L
s
[(-1)] .4875 .19934 1 0 Mshowa
.5 .4187 m
.50625 .4187 L
s
[(1)] .4875 .4187 1 0 Mshowa
.5 .52838 m
.50625 .52838 L
s
[(2)] .4875 .52838 1 0 Mshowa
.125 Mabswid
.5 .11159 m
.50375 .11159 L
s
.5 .13353 m
.50375 .13353 L
s
.5 .15547 m
.50375 .15547 L
s
.5 .1774 m
.50375 .1774 L
s
.5 .22127 m
.50375 .22127 L
s
.5 .24321 m
.50375 .24321 L
s
.5 .26515 m
.50375 .26515 L
s
.5 .28708 m
.50375 .28708 L
s
.5 .33095 m
.50375 .33095 L
s
.5 .35289 m
.50375 .35289 L
s
.5 .37482 m
.50375 .37482 L
s
.5 .39676 m
.50375 .39676 L
s
.5 .44063 m
.50375 .44063 L
s
.5 .46257 m
.50375 .46257 L
s
.5 .4845 m
.50375 .4845 L
s
.5 .50644 m
.50375 .50644 L
s
.5 .06772 m
.50375 .06772 L
s
.5 .04579 m
.50375 .04579 L
s
.5 .02385 m
.50375 .02385 L
s
.5 .00191 m
.50375 .00191 L
s
.5 .55031 m
.50375 .55031 L
s
.5 .57225 m
.50375 .57225 L
s
.5 .59418 m
.50375 .59418 L
s
.5 .61612 m
.50375 .61612 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
gsave
.5 .64303 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.70849 .30902 m
.73495 .28481 L
.78733 .27647 L
.84302 .28168 L
.89273 .29687 L
.93348 .31999 L
.9617 .34859 L
.97619 .38155 L
.97552 .41646 L
.95924 .45237 L
.92783 .48701 L
.88163 .51945 L
.82268 .54786 L
.75224 .5714 L
.67333 .58887 L
.58815 .59967 L
.5 .60332 L
.41185 .59967 L
.32667 .58887 L
.24776 .5714 L
.17732 .54787 L
.11837 .51945 L
.07217 .48701 L
.04076 .45237 L
.02448 .41646 L
.02381 .38155 L
.0383 .34859 L
.06652 .31999 L
.10727 .29687 L
.15698 .28168 L
.21267 .27647 L
.26505 .28481 L
.29151 .30902 L
.26505 .33322 L
.21267 .34157 L
.15698 .33636 L
.10727 .32116 L
.06652 .29804 L
.0383 .26944 L
.02381 .23648 L
.02448 .20157 L
.04076 .16567 L
.07217 .13102 L
.11837 .09858 L
.17732 .07017 L
.24776 .04663 L
.32667 .02916 L
.41185 .01836 L
.5 .01472 L
.58815 .01836 L
Mistroke
.67333 .02916 L
.75224 .04663 L
.82268 .07017 L
.88163 .09858 L
.92783 .13102 L
.95924 .16567 L
.97552 .20157 L
.97619 .23648 L
.9617 .26944 L
.93348 .29804 L
.89273 .32116 L
.84302 .33636 L
.78733 .34157 L
.73495 .33322 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7EjL30`9T0I_ooc>CTi6YZJX0003<cIVIVc820P9VIo`00P28R8X200020P4:JY`0g
?0200?mV<omVc1YOI`00kUDJRj>SXiVIVIT00000<`00IP00V@00c00c000c<`0cIP0cV@0cc00co`1V
001V<`1VIP1VV@1Vc01Vo`2I002I<`2IIP2IV@2Ic02Io`3<003<<`3<IP3<V@3<c03<o`3o<`3oIP3o
V@3oc3<003<0<c<0IS<0VC<0c3<0oc<c03<c<c<cVC<cc3<coc=V03=V<c=VIS=VVC=Vc3=Voc>I03>I
<c>IIS>IVC>Ic3>Ioc?<03?<<c?<IS?<VC?<c3?<oc?o03?o<c?oIS?oVC?oc3?oofH006H0<fH0IVH0
VFH0c6H0ofHc06Hc<fHcIVHcVFHcc6HcofIV06IV<fIVIVIVVFIVofJI06JI<fJIIVJIVFJIc6JIofK<
06K<<fK<IVK<VFK<c6K<ofKo06Ko<fKoIVKoVFKoc6KooiT0<iT0VIT0c9T0oiTc09Tc<iTcIYTcVITc
c9TcoiUV09UV<iUVIYUVVIUVc9UVoiVI09VI<iVIIYVIc9W<09W<<iW<IYW<VIW<c9W<oiWo09Wo<iWo
IYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aVVLaV
c<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<coool0
<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mVIX@00P6J0000Q0000`40002I0000Q0000`40
002I0000N@0G0I00001[00h12`000`40000900h1PP0006<0204I00030@0001L0205j0000G0070B40
0P4P00L1L`0005H01P4X00030@0002H01P5]0000D@050Bh000<10000;0050FP0001<00D1<`000`40
000a00@1I00004L01@4h00030@0003D01@5O0000@`040Cd00P4k00@1F`0004000`5100030@0003h0
0`5H0000?@030D@000<10000@@030ED0000k0081A`000`40001400<1DP0003P00`5900030@0004L0
0`5?0000=@030D`000<10000BP030D`0000b00<1C`020Dh00`590000<0020E8000<10000D0020DL0
000^0081E0000`40001B0081A@0002`00P5F00030@0005@00P530000:P020E00104400030@0005H0
0P510000:0020E8000<100001@020ET00P4o00009P020ED000<1000010000`40001J0081?@0002@0
0P5@00@110000`40000300030@0005`00P4k00008P020E/000<100000P000`40001N0081>@000240
00<10000F0001040004400030@00060000<10000=P00020000<10000FP020@D000<10000H@000`40
000e00007P020F@00P5S0081=@0001d000<10000I0000`40001T00030@000380000L00030@0006D0
00<10000I@000`40000a00006P020FP000<10000IP020C40000I00030@0006P000<10000J0000`40
000^000060000`40001Y0081JP000`40000]00005`000`40001Z00030@0006X000<10000;00001H0
00<10000J`000`40001[00030@0002/0000E00030@0006`000<10000K0000`40000Z000050000`40
001]00030@0006d000<10000:@0001<000<10000KP000`40001^00030@0002P0000B00030@0006l0
0P5`00030@0002L0000A00030@00070000<10000L0000`40000V000040000`40001a00030@000740
00<100009@00010000<10000L@000`40001a00030@0002D0000?00030@00078000<10000LP000`40
000T00003`000`40001b0081L`000`40000T00003P000`40001c00030@0007<000<100008`0000h0
00<10000L`000`40001c00030@0002<0000=00030@0007@000<10000M0000`40000R00003@000`40
001/00@110000`40001d00030@000280000<00030@0006l000<100000`000`40001e00030@000240
000<00030@0006l000<100000`020GH000<100008@0000`000<10000I`040@@000<100000`000`40
001e00030@000240000<00030@0006l000<100000`000`40001e00030@000240000<00030@0006d0
0`4500030@0007D000<100008@0000`000<10000K`000`40000300030@0007D000<100008@0000`0
00<10000M@020GH000<100008@0000`000<10000M@000`40001e00030@000240000<00030@0007D0
00<10000M@000`40000Q000030000`40001e00030@0007D000<100008@0000`000<10000M@000`40
001e00030@000240000=00030@0007@000<10000M@000`40000Q00003@000`40001d0081M@000`40
000R00003P000`40001c00030@0007@000<100008P0000h000<10000L`000`40001c00030@0002<0
000?00030@00078000<10000L`000`40000S00003`000`40001b00030@00078000<1000090000100
00<10000L@020G<000<100009000010000<10000L@000`40001a00030@0002D0000A00030@000700
00<10000L0000`40000V00003`040B402`4;00@1=P000`40000f00@1200;0B801@4U00003`001@40
000101P0204;00L11P000`40000e00030@0003P000@100001P4;00P16@000`400@0X000040001@40
000101@00`4J00@10P000`40000e00030@0003P0104I00<15P001040004W00002@040@@000@10001
4@030A/0104200030@0103L00P4h00817`040A4000<100000P000`40000T00004P001040004=00<1
9@020CL000<10000=P000`400@0S00<13@000`40000400030@0002<0000?00040@000@<000<10000
20020BL00`4g00030@0003@01@4V00812P000`40000200040@000BD0000@00811@020@H00P4[0081
=P000`40000c00030@00008000<100009P020@H00P46008140030@8000<100000`4;00006@020@80
0P4_00030@0003<000<10000<P000`40000^00810P020A/000@100010P000`40000<00001P3m0@P0
0`4300<1300000h000<100002@020@03004100L000<100002@000`40000800030@0000T000<10000
0P000`40000400030@0000T000<1000020000`40000900030@0000T000<100002@000`4000080003
0@0000T000<100002@000`40000900030@0000P000<100002@000`40000900030@0000D00P400`01
0@0:00030@0000l000@100010P001040004;00003P000`40000700811@020BP000D100000@0f0003
0@0003P000<100009P020@D00P4800030@0000l000@100010`020@`0000G00030@0000L00P4Y0003
0@0003D000<10000G`020@T000<100005P040A80000F00030@0000X00`4U00030@0003H000<10000
G0030@`000<10000:`0001D000<100003P030B4000<10000=`000`40001H00@140000`40000Z0000
50000`40000B00<16`030CX00P4i00<16P030AD000<10000:@0001<000<100005P050A401@4m0003
0@0003/0104A00D16@000`40000X00004P000`40000L00L11P040D8000<10000?`040@H01`4O0003
0@0002L0000A00030@0002@01P5600030@0004<01P4W00030@0002H0000@00030@00074000<10000
L@000`40000U000040000`40001a0081LP000`40000U00003`000`40001b00030@00078000<10000
900000l000<10000LP000`40001b00030@0002@0000>00030@0007<000<10000L`000`40000S0000
3P000`40001c00030@0007<000<100008`0000d000<10000M0000`40001d00030@000280000=0003
0@0007@00P5e00030@000280000<00030@0007D000<10000M@000`40000Q000030000`40001e0003
0@0007D000<100008@0000`000<10000M@000`40001e00030@000240000<00030@0007D000<10000
M@000`40000Q000030000`40001e0081MP000`40000Q000030000`40001e00030@0007D000<10000
8@0000`000<10000M@000`40001e00030@000240000<00030@0007D000<10000M@000`40000Q0000
30000`40001e00030@0007D000<100008@0000`000<10000K@040@@000<10000M@000`40000Q0000
30000`40001_00030@0000<00P5f00030@000240000=00030@0006h000<100000`000`40001d0003
0@000280000=00030@0006h000<100000`000`40001d00030@000280000>00030@0006d000<10000
0`000`40001c00030@0002<0000>00030@0006/00`4500030@0007<000<100008`0000l000<10000
K0000`40000300030@00078000<10000900000l000<10000LP020G<000<100009000010000<10000
L@000`40001a00030@0002D0000@00030@00074000<10000L@000`40000U00004@000`40001`0003
0@00070000<100009P00018000<10000K`000`40001_00030@0002L0000C00030@0006h00P5_0003
0@0002P0000D00030@0006d000<10000K@000`40000Y000050000`40001]00030@0006d000<10000
:@0001D000<10000K0000`40001/00030@0002X0000F00030@0006/000<10000J`000`40000[0000
5`000`40001Z00030@0006X000<10000;00001P000<10000J@020FX000<10000;@0001T00P5Y0003
0@0006L00P4`00006`000`40001V00030@0006H000<10000<00001`00P5V00030@0006@00P4c0000
7P000`40001S00030@0006<000<10000<`0001l00P5S0081HP020CH0000Q00030@00060000<10000
H0000`40000f00008P020F0000<10000GP020CT0000T00030@0005d000<10000G@000`40000i0000
9@020ED0104400030@0005/00P4l00009`020E<000<100001@000`40001I0081?P0002T00P5B0003
0@0000@00P5H0081@00002/00P5A00030@0000<000<10000E@020D80000]0081D0000`4000020003
0@0005<00P540000;`020D/000@1000110000`40001A0081AP0003400P5:00811@000`40001?0081
B00003<00P5?0081CP020DX0000e00@1B`000`40001900@1C00003T00`5800030@0004D0105@0000
?0030DD000<10000@P030E@0000o00@1@@000`40000n00@1E`0004<0104m00030@0003/00`5K0000
A`040CT00P4h00@1GP0004/0104e00030@0003<0105R0000C`040C4000<10000;`040FH0001C00@1
;@000`40000[00@1JP0005L01@4X00030@0002H01@5^0000G0080B000P4O00P1L`0006@01`4I0003
0@0001L01`5k0000J`0c0H80002400030@0009T0002400030@0009T0002400030@0009T000240081
VP000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@0007h01@401@0100400081UP00080000T100000@010040U`00080000T100000@010040U`000800
00T100000@010040U`00080000@100000P400`01002G0000OP050Id0003o0240003o0240003o0240
003o0240003o02400000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-2.25466, -2.81747, 
  0.0170353, 0.0362043}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Quick reference for Anopt"], "Section",
  Evaluatable->False,
  PageBreakAbove->True,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["The output table"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "\tAfter every iteration, eight numbers are placed on the screen. Recall \
that  to each new function f",
  StyleBox["k",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " generated by Anopt corresponds a new line in the output table.  The first \
line (iter=0) corresponds to f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  ".  Since you did not specify f",
  StyleBox["0",
    FontVariations->{"CompatibilityType"->"Subscript"}],
  " in the two examples above, it has been defaulted there to the zero \
function.  The columns of the table are explained below."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"\tHEADING\t        EXPLANATION\n\n\tIter\t\t             Iterate number k.\n\
\n\tSupGamma\t             sup  G ( e  , f ( e  ) ) , the value at current f.\
\n\n\tStep\t\t             SupNorm[difference bet. last two iterates f ],\n\n\
\tFlat, GradAlign      \tOptimality Tests.  If f is near optimal\n\t\t\t      \
       then Flat and GradAlign are close to 0.\n\n\tNED\t\t            \
Numerical Error Diagnostic. If relatively\n\t\t\t            large it \
indicates numerical trouble, usually  due to \n\t\t\t            slowly \
decaying Fourier coefficients.\n\n\tSm\t\t            Automatic smoothing of \
fk.\n\n\tGrid\t                        Number of samples on the unit circle, \
a power of 2.\n\t\t\t            Increased by Anopt as needed."], "Text",
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10],

Cell[TextData[
"\n\tUsually the most important numbers on the screen are the Flat and \
GradAlign  diagnostics.  As they approach zero, the current guess is expected \
to approach the solution.  Under mild hypotheses, the solution is unique in \
OPT problems with only one (scalar valued) analytic unknown function.  When \
the number of analytic unknown functions is N > 1, solutions are not unique.  \
There may be many local solutions, which Anopt may find when run with \
different initial guesses.\n\t\n\nExplanation of the output table will be \
expanded in Section 1.  A more advanced user may set the options Diagnostics \
-> 1 and Diagnostics2 ->1.  This gives additional information on the run, \
which is stored automatically in certain files.  "], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Some ",
  StyleBox["Mathematica",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  "  functions and Notation"
}], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"\t\t          Abs[z]\t\tAbsolute Value of z.\t\n\t\tConjugate[z]  \t         \
  Conjugate of z.\n\t\t                z^n\t \tn-th power of z.\n\t\t         \
   Re[z]\t \tReal Part of z.\t\n\t\t             Im[z]\t\tImaginary Part of \
z.\n\t\t     {x1,x2,x3}\t\tA list with entries x1, x2, x3.\n\t\t           \
D[g,z]\t    \tDerivative of g with respect to z.\n                            \
  Fourier[f]             Fast Fourier Transform of f."], "Text",
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Some Anopt  utilities"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"     ComplexCoordinates[g]       Applies transformation rules to g to write \
it in terms of \n\t\t\t\t   the complex variables and their conjugates.\n\n   \
            ComplexD[  g , z ]      Complex derivative of g with respect to \
z.\n\nComplexD[g,Conjugate[z]]      Complex derivative of g with respect to \
Conjugate [z].\n\n\t           DoubleGrid[f]    Doubles the number of sample \
points of a discrete function \n\n\t                HalfGrid[f]    Halves the \
number of sample points of a discrete function \n\n                     \
DiskListPlot[f]     Plots a list of complex numbers in the Re-Im plane.\n\n   \
             DiskListPlot3D[f]     Plots a list of complex numbers in 3D \
space.  One of the \n                                                    axes \
corresponds to the parameter, assumed to be theta \n                          \
                          in [0, 2 Pi].\n"], "Text",
  CellMargins->{{1, Inherited}, {Inherited, Inherited}},
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Examples of valid performance functions."], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"          g = Abs[ 0.8 + ( 1/e + z[1] )^2 ]^2\t\n\n          g = Re[ 1/(0.5 \
+e) -z[2] ]^2 + Im[ 1/e + z[2] ]^2 + Abs[ z[1] ]^2"], "Text",
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[
"Examples of invalid performance functions.        Reason:"], "Subsubsection",\

  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"\t\t  Abs[ 1/(e-1) + z[1] ]^2\tA pole on the unit circle at e=1.\n\t\t  \n\
Re[e + z[1]]^2 + Abs[e^2 + 4 - z[2] ]^2\t                Unbounded level sets \
in z[1].\n\n\t         Abs[ 1/(e^2 - 0.25) + z[1] ]\tNot differentiable in \
z[1] everywhere."], "Text",
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["Anopt calling sequences"], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"\tAnopt[performance , tolerance]\n\n\tAnopt[performance , tolerance, \
Initialguess]\n\t\n\tAnopt[performance , tolerance, Initialguess , options]"],
   "Text",
  Evaluatable->False,
  AspectRatioFixed->True,
  FontSize->10]
}, Open  ]],

Cell[TextData["Anopt Options and defaults."], "Subsubsection",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[Anopt]\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \({ConstrainedSearch \[Rule] None, QuickResults \[Rule] False, 
      OptimalityTol \[Rule] Automatic, Subproblem \[Rule] "DiskIteration", 
      DebugMe \[Rule] False, AMatrix \[Rule] None, Iterations \[Rule] 10, 
      GridDouble \[Rule] True, MaxGridSize \[Rule] 1024, 
      Diagnostics \[Rule] 0, Diagnostics2 \[Rule] 0, RunID \[Rule] "none", 
      Weight \[Rule] 1, DualEigenfunction \[Rule] Automatic, 
      OT3Check \[Rule] True, DoLineSearch \[Rule] True, 
      Smoothing \[Rule] Automatic, InitialSmooth \[Rule] None, 
      UpdateVectorSmooth \[Rule] None, SmoothType \[Rule] "A", 
      LineSearchFunction \[Rule] GoldenLineSearch, GridPoints \[Rule] 32}\)], 
  "Output"]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
WindowToolbars->{"RulerBar", "EditBar"},
CellGrouping->Manual,
WindowSize->{624, 593},
WindowMargins->{{177, Automatic}, {5, Automatic}},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->True,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
CharacterEncoding->"XAutomaticEncoding"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 96, 3, 107, "Title"],
Cell[1830, 56, 294, 7, 32, "Text"],
Cell[2127, 65, 228, 5, 50, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2392, 75, 85, 2, 54, "Section"],
Cell[2480, 79, 2211, 33, 300, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[4728, 117, 124, 2, 45, "Subsection"],
Cell[4855, 121, 998, 18, 214, "SmallText"],
Cell[5856, 141, 756, 24, 112, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[6649, 170, 129, 3, 45, "Subsection"],

Cell[CellGroupData[{
Cell[6803, 177, 94, 2, 42, "Subsubsection"],
Cell[6900, 181, 553, 12, 63, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7490, 198, 116, 2, 42, "Subsubsection"],
Cell[7609, 202, 148, 4, 27, "SmallText"],

Cell[CellGroupData[{
Cell[7782, 210, 75, 2, 27, "Input"],
Cell[7860, 214, 122, 2, 55, "Print"]
}, Open  ]],
Cell[7997, 219, 349, 9, 43, "SmallText"],

Cell[CellGroupData[{
Cell[8371, 232, 97, 2, 47, "Input"],
Cell[8471, 236, 290, 8, 50, "Output"]
}, Open  ]],
Cell[8776, 247, 196, 4, 27, "SmallText"],

Cell[CellGroupData[{
Cell[8997, 255, 73, 2, 27, "Input"],
Cell[9073, 259, 36, 1, 23, "Print"],
Cell[9112, 262, 114, 2, 23, "Print"],
Cell[9229, 266, 109, 2, 23, "Print"],
Cell[9341, 270, 114, 2, 39, "Print"],
Cell[9458, 274, 112, 2, 23, "Print"],
Cell[9573, 278, 112, 2, 23, "Print"],
Cell[9688, 282, 112, 2, 23, "Print"],
Cell[9803, 286, 112, 2, 23, "Print"],
Cell[9918, 290, 36, 1, 23, "Print"],
Cell[9957, 293, 36, 1, 23, "Print"],
Cell[9996, 296, 42, 1, 23, "Print"],
Cell[10041, 299, 42, 1, 23, "Print"],
Cell[10086, 302, 68, 1, 23, "Print"],
Cell[10157, 305, 63, 1, 23, "Print"],
Cell[10223, 308, 63, 1, 23, "Print"],
Cell[10289, 311, 54, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[10392, 318, 100, 2, 42, "Subsubsection"],
Cell[10495, 322, 280, 5, 42, "SmallText"],
Cell[10778, 329, 970, 19, 166, "Text"],
Cell[11751, 350, 450, 8, 57, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[12238, 363, 92, 2, 42, "Subsubsection"],
Cell[12333, 367, 1420, 46, 115, "SmallText"],

Cell[CellGroupData[{
Cell[13778, 417, 79, 2, 27, "Input"],
Cell[13860, 421, 41, 1, 27, "Output"]
}, Open  ]],
Cell[13916, 425, 290, 5, 42, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[14243, 435, 101, 2, 42, "Subsubsection"],
Cell[14347, 439, 168, 4, 27, "SmallText"],

Cell[CellGroupData[{
Cell[14540, 447, 87, 2, 27, "Input"],
Cell[14630, 451, 12109, 450, 186, 6123, 372, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[26754, 904, 230, 5, 42, "SmallText"],

Cell[CellGroupData[{
Cell[27009, 913, 83, 2, 27, "Input"],
Cell[27095, 917, 22136, 779, 220, 11381, 642, "GraphicsData", 
"PostScript", "Graphics"],
Cell[49234, 1698, 135, 3, 27, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[49430, 1708, 137, 3, 45, "Subsection"],

Cell[CellGroupData[{
Cell[49592, 1715, 94, 2, 42, "Subsubsection"],
Cell[49689, 1719, 989, 29, 99, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[50715, 1753, 88, 2, 42, "Subsubsection"],
Cell[50806, 1757, 373, 10, 33, "SmallText"],

Cell[CellGroupData[{
Cell[51204, 1771, 208, 5, 63, "Input"],
Cell[51415, 1778, 36, 1, 23, "Print"],
Cell[51454, 1781, 114, 2, 23, "Print"],
Cell[51571, 1785, 109, 2, 23, "Print"],
Cell[51683, 1789, 114, 2, 39, "Print"],
Cell[51800, 1793, 112, 2, 23, "Print"],
Cell[51915, 1797, 112, 2, 23, "Print"],
Cell[52030, 1801, 112, 2, 23, "Print"],
Cell[52145, 1805, 112, 2, 23, "Print"],
Cell[52260, 1809, 36, 1, 23, "Print"],
Cell[52299, 1812, 36, 1, 23, "Print"],
Cell[52338, 1815, 42, 1, 23, "Print"],
Cell[52383, 1818, 42, 1, 23, "Print"],
Cell[52428, 1821, 68, 1, 23, "Print"],
Cell[52499, 1824, 63, 1, 23, "Print"],
Cell[52565, 1827, 63, 1, 23, "Print"],
Cell[52631, 1830, 54, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[52734, 1837, 120, 2, 42, "Subsubsection"],
Cell[52857, 1841, 885, 13, 162, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[53779, 1859, 120, 2, 42, "Subsubsection"],
Cell[53902, 1863, 773, 17, 78, "SmallText"],
Cell[54678, 1882, 149, 4, 49, "Input"],
Cell[54830, 1888, 304, 6, 42, "SmallText"],

Cell[CellGroupData[{
Cell[55159, 1898, 68, 2, 27, "Input"],
Cell[55230, 1902, 303, 5, 63, "Output"]
}, Open  ]],
Cell[55548, 1910, 401, 10, 34, "SmallText"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[55998, 1926, 120, 2, 45, "Subsection"],
Cell[56121, 1930, 882, 22, 75, "SmallText"],

Cell[CellGroupData[{
Cell[57028, 1956, 130, 3, 42, "Subsubsection"],
Cell[57161, 1961, 261, 5, 42, "SmallText"],
Cell[57425, 1968, 103, 2, 47, "Input"],
Cell[57531, 1972, 204, 4, 27, "SmallText"],
Cell[57738, 1978, 102, 2, 45, "Input"],
Cell[57843, 1982, 884, 20, 126, "SmallText"],

Cell[CellGroupData[{
Cell[58752, 2006, 100, 2, 27, "Input"],
Cell[58855, 2010, 36, 1, 23, "Print"],
Cell[58894, 2013, 114, 2, 23, "Print"],
Cell[59011, 2017, 109, 2, 23, "Print"],
Cell[59123, 2021, 114, 2, 39, "Print"],
Cell[59240, 2025, 112, 2, 23, "Print"],
Cell[59355, 2029, 112, 2, 23, "Print"],
Cell[59470, 2033, 113, 2, 23, "Print"],
Cell[59586, 2037, 112, 2, 23, "Print"],
Cell[59701, 2041, 36, 1, 23, "Print"],
Cell[59740, 2044, 36, 1, 23, "Print"],
Cell[59779, 2047, 42, 1, 23, "Print"],
Cell[59824, 2050, 42, 1, 23, "Print"],
Cell[59869, 2053, 68, 1, 23, "Print"],
Cell[59940, 2056, 63, 1, 23, "Print"],
Cell[60006, 2059, 63, 1, 23, "Print"],
Cell[60072, 2062, 54, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],
Cell[60153, 2067, 441, 10, 48, "SmallText"],
Cell[60597, 2079, 188, 5, 45, "Input"],

Cell[CellGroupData[{
Cell[60810, 2088, 101, 2, 42, "Subsubsection"],
Cell[60914, 2092, 555, 9, 87, "SmallText"],

Cell[CellGroupData[{
Cell[61494, 2105, 124, 3, 43, "Input"],
Cell[61621, 2110, 36, 1, 23, "Print"],
Cell[61660, 2113, 114, 2, 23, "Print"],
Cell[61777, 2117, 109, 2, 23, "Print"],
Cell[61889, 2121, 114, 2, 39, "Print"],
Cell[62006, 2125, 112, 2, 23, "Print"],
Cell[62121, 2129, 112, 2, 23, "Print"],
Cell[62236, 2133, 113, 2, 23, "Print"],
Cell[62352, 2137, 112, 2, 23, "Print"],
Cell[62467, 2141, 112, 2, 23, "Print"],
Cell[62582, 2145, 36, 1, 23, "Print"],
Cell[62621, 2148, 36, 1, 23, "Print"],
Cell[62660, 2151, 42, 1, 23, "Print"],
Cell[62705, 2154, 42, 1, 23, "Print"],
Cell[62750, 2157, 68, 1, 23, "Print"],
Cell[62821, 2160, 63, 1, 23, "Print"],
Cell[62887, 2163, 63, 1, 23, "Print"],
Cell[62953, 2166, 54, 1, 23, "Print"]
}, Open  ]],
Cell[63022, 2170, 120, 2, 27, "SmallText"],

Cell[CellGroupData[{
Cell[63167, 2176, 79, 2, 27, "Input"],
Cell[63249, 2180, 41, 1, 27, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[63351, 2188, 105, 2, 42, "Subsubsection"],
Cell[63459, 2192, 674, 13, 73, "SmallText"],
Cell[64136, 2207, 179, 5, 65, "Input"],
Cell[64318, 2214, 108, 2, 27, "SmallText"],

Cell[CellGroupData[{
Cell[64451, 2220, 183, 5, 27, "Input"],
Cell[64637, 2227, 177, 3, 39, "Message"]
}, Open  ]],
Cell[64829, 2233, 133, 3, 27, "SmallText"],

Cell[CellGroupData[{
Cell[64987, 2240, 100, 2, 27, "Input"],
Cell[65090, 2244, 12157, 461, 182, 5265, 370, "GraphicsData", 
"PostScript", "Graphics"],
Cell[77250, 2707, 130, 3, 27, "Output"]
}, Open  ]],
Cell[77395, 2713, 168, 4, 27, "SmallText"],

Cell[CellGroupData[{
Cell[77588, 2721, 82, 2, 27, "Input"],
Cell[77673, 2725, 59, 1, 27, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[77781, 2732, 110, 2, 42, "Subsubsection"],
Cell[77894, 2736, 461, 8, 57, "SmallText"],
Cell[78358, 2746, 198, 5, 27, "Input"],

Cell[CellGroupData[{
Cell[78581, 2755, 162, 4, 27, "Input",
  PageBreakBelow->False],
Cell[78746, 2761, 13364, 451, 186, 5730, 352, "GraphicsData", 
"PostScript", "Graphics"],
Cell[92113, 3214, 130, 3, 27, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[92292, 3223, 126, 3, 54, "Section",
  PageBreakAbove->True],

Cell[CellGroupData[{
Cell[92443, 3230, 99, 2, 42, "Subsubsection"],
Cell[92545, 3234, 687, 16, 69, "SmallText"],
Cell[93235, 3252, 846, 13, 318, "Text"],
Cell[94084, 3267, 820, 12, 147, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[94941, 3284, 201, 8, 42, "Subsubsection"],
Cell[95145, 3294, 538, 9, 142, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[95720, 3308, 104, 2, 42, "Subsubsection"],
Cell[95827, 3312, 1051, 16, 286, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[96915, 3333, 123, 2, 42, "Subsubsection"],
Cell[97041, 3337, 219, 5, 62, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[97297, 3347, 143, 4, 42, "Subsubsection"],
Cell[97443, 3353, 340, 7, 94, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[97820, 3365, 106, 2, 42, "Subsubsection"],
Cell[97929, 3369, 240, 6, 94, "Text"]
}, Open  ]],
Cell[98184, 3378, 110, 2, 42, "Subsubsection"]
}, Open  ]],

Cell[CellGroupData[{
Cell[98331, 3385, 73, 2, 27, "Input"],
Cell[98407, 3389, 704, 11, 123, "Output"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

