(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    308515,      11888]*)
(*NotebookOutlinePosition[    309150,      11911]*)
(*  CellTagsIndexPosition[    309106,      11907]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Demo 2 Chapter 6", "Title"],

Cell["\<\
This demo corresponds to the second example of Chapter 6 of the \
book Classical Control Using Hinfinity Methods, by J. W. Helton and O. \
Merino,
NOTE: Depending upon the settings of the programs, some number  may differ \
from those in the book.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
<<OPTDesign`;
<<Calculus`LaplaceTransform`;\
\>", "Input"],

Cell[BoxData[
    \("OPTDesign 2.0 (c) 1991-98 \n   J.W. Helton and O.Merino.  All rights \
reserved."\)], "Print"]
}, Open  ]],

Cell["Define the plant and the inverse plant", "Text"],

Cell["\<\

p[s_] = 0.036 (s + 25)/(s^2 (s + 0.02 +  I)(s + 0.02 -   I));
pinv[s_] = 1/p[s];\
\>", "Input"],

Cell["\<\
The reference transfer function is built by incorporating the \
interpolation constraints\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Tref[s_] = NewtonInterpolant[{{0,1,0}},s,RelativeDegree -> 1]\
\>", 
  "Input"],

Cell[BoxData[
    \(\(1 + 2\ s\)\/\((1 + s)\)\^2\)], "Output"]
}, Open  ]],

Cell["\<\
Now we calculate the step response function corresponding to the \
reference transfer function\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
f2[t_] =  InverseLaplaceTransform[ Tref[s]/s ,s,t] ; 


plota = Plot[f2[t],{t,0,10},PlotRange -> {0.8,1.2}];
\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 -1.23607 1.54508 [
[.21429 .29652 -3 -9 ]
[.21429 .29652 3 0 ]
[.40476 .29652 -3 -9 ]
[.40476 .29652 3 0 ]
[.59524 .29652 -3 -9 ]
[.59524 .29652 3 0 ]
[.78571 .29652 -3 -9 ]
[.78571 .29652 3 0 ]
[.97619 .29652 -6 -9 ]
[.97619 .29652 6 0 ]
[.01131 0 -18 -4.5 ]
[.01131 0 0 4.5 ]
[.01131 .07725 -24 -4.5 ]
[.01131 .07725 0 4.5 ]
[.01131 .15451 -18 -4.5 ]
[.01131 .15451 0 4.5 ]
[.01131 .23176 -24 -4.5 ]
[.01131 .23176 0 4.5 ]
[.01131 .38627 -24 -4.5 ]
[.01131 .38627 0 4.5 ]
[.01131 .46353 -18 -4.5 ]
[.01131 .46353 0 4.5 ]
[.01131 .54078 -24 -4.5 ]
[.01131 .54078 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .30902 m
.21429 .31527 L
s
[(2)] .21429 .29652 0 1 Mshowa
.40476 .30902 m
.40476 .31527 L
s
[(4)] .40476 .29652 0 1 Mshowa
.59524 .30902 m
.59524 .31527 L
s
[(6)] .59524 .29652 0 1 Mshowa
.78571 .30902 m
.78571 .31527 L
s
[(8)] .78571 .29652 0 1 Mshowa
.97619 .30902 m
.97619 .31527 L
s
[(10)] .97619 .29652 0 1 Mshowa
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.2619 .30902 m
.2619 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.5 .30902 m
.5 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.7381 .30902 m
.7381 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 0 m
.03006 0 L
s
[(0.8)] .01131 0 1 0 Mshowa
.02381 .07725 m
.03006 .07725 L
s
[(0.85)] .01131 .07725 1 0 Mshowa
.02381 .15451 m
.03006 .15451 L
s
[(0.9)] .01131 .15451 1 0 Mshowa
.02381 .23176 m
.03006 .23176 L
s
[(0.95)] .01131 .23176 1 0 Mshowa
.02381 .38627 m
.03006 .38627 L
s
[(1.05)] .01131 .38627 1 0 Mshowa
.02381 .46353 m
.03006 .46353 L
s
[(1.1)] .01131 .46353 1 0 Mshowa
.02381 .54078 m
.03006 .54078 L
s
[(1.15)] .01131 .54078 1 0 Mshowa
.125 Mabswid
.02381 .01545 m
.02756 .01545 L
s
.02381 .0309 m
.02756 .0309 L
s
.02381 .04635 m
.02756 .04635 L
s
.02381 .0618 m
.02756 .0618 L
s
.02381 .09271 m
.02756 .09271 L
s
.02381 .10816 m
.02756 .10816 L
s
.02381 .12361 m
.02756 .12361 L
s
.02381 .13906 m
.02756 .13906 L
s
.02381 .16996 m
.02756 .16996 L
s
.02381 .18541 m
.02756 .18541 L
s
.02381 .20086 m
.02756 .20086 L
s
.02381 .21631 m
.02756 .21631 L
s
.02381 .24721 m
.02756 .24721 L
s
.02381 .26266 m
.02756 .26266 L
s
.02381 .27812 m
.02756 .27812 L
s
.02381 .29357 m
.02756 .29357 L
s
.02381 .32447 m
.02756 .32447 L
s
.02381 .33992 m
.02756 .33992 L
s
.02381 .35537 m
.02756 .35537 L
s
.02381 .37082 m
.02756 .37082 L
s
.02381 .40172 m
.02756 .40172 L
s
.02381 .41717 m
.02756 .41717 L
s
.02381 .43262 m
.02756 .43262 L
s
.02381 .44807 m
.02756 .44807 L
s
.02381 .47898 m
.02756 .47898 L
s
.02381 .49443 m
.02756 .49443 L
s
.02381 .50988 m
.02756 .50988 L
s
.02381 .52533 m
.02756 .52533 L
s
.02381 .55623 m
.02756 .55623 L
s
.02381 .57168 m
.02756 .57168 L
s
.02381 .58713 m
.02756 .58713 L
s
.02381 .60258 m
.02756 .60258 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.08351 0 m
.09396 .11413 L
.10458 .2085 L
.11508 .2843 L
.125 .34237 L
.14357 .42216 L
.15327 .45161 L
.16355 .47546 L
.17232 .49075 L
.18199 .503 L
.18718 .50786 L
.19212 .5115 L
.19656 .51402 L
.2014 .51602 L
.20414 .51685 L
.20664 .51741 L
.20802 .51765 L
.20951 .51785 L
.2102 .51792 L
.21092 .51799 L
.21222 .51807 L
.21342 .51811 L
.2147 .51812 L
.21591 .51809 L
.21701 .51804 L
.21831 .51794 L
.21972 .51779 L
.22227 .51743 L
.22461 .51698 L
.22712 .51638 L
.23167 .51503 L
.24205 .51079 L
.25301 .50488 L
.26306 .49849 L
.30268 .46839 L
.34079 .438 L
.38135 .40868 L
.4204 .385 L
.46189 .36493 L
.50188 .35005 L
.54034 .33917 L
.58126 .33054 L
.62066 .32446 L
.65855 .32018 L
.69889 .31687 L
.73771 .31459 L
.75746 .31369 L
.77899 .31287 L
.81875 .31171 L
.83846 .31127 L
Mistroke
.85945 .31087 L
.87946 .31056 L
.90108 .31028 L
.9412 .30989 L
.97619 .30965 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@0000P00P4400030@0000<00P7o00/0000700040@000@P000@1000110020@h0
00<10000lP0000L000@1000120001040004400030@0000d000<10000lP0000L000@100012@020@D0
00<100003@000`40003b00001`001040004800040@000@@000<100003@000`40003b00001`001040
004800040@000@@00P4>00030@000?80000800812P020@D000<100003@000`40003b00006`000`40
000=00030@000?80000K00030@0000d000<10000lP0001/00P4?00030@000?40000K00030@0000h0
00<10000l@0001/000<100003P000`40003a00006`000`40000>00030@000?40000K00813`000`40
003a00006`000`40000>00030@000?40000K00030@0000h000<10000l@0001/000<100003P000`40
003a00006`000`40000>00030@000?40000K00813`000`40003a00006`000`40000>00030@000?40
0002008110000`40000300810`030@D000<100003P000`40003a000000D00@000@0800040@000@D0
00<100000P000`40000>00030@000?4000001@01000100P000@100011@000`40000200813`000`40
003a000000D00@000@0900810`030@D000<100003P000`40003a000000D00@000@0800040@000@80
00<100001@000`40000?00030@000?0000001@01000100P000@100010P000`40000500030@0000l0
00<10000l00000800P4:00810`040@@00P4@00030@000?00000K00030@0000l000<10000l00001/0
00<100003`000`40003`00006`000`40000?00030@000?00000K008140000`40003`00006`000`40
000?00030@000?00000K00030@0000l000<10000l00001/000<100003`000`40003`00006`020A00
00<10000l00001/000<100003`000`40003`00006`000`40000@00030@000>l0000K00030@000100
00<10000k`0001/00P4A00030@000>l0000K00030@00010000<10000k`0001/000<1000040000`40
003_000020020@@000<100000P030@D000<1000040000`40003_00001`001040004;00030@000080
0P4A00030@000>l0000700040@000@/000<100000P000`40000@00030@000>l0000700040@000@T0
0`4400030@00014000<10000kP0000L000@1000120001040004400030@00014000<10000kP0000L0
00@1000120001040004400814P000`40003^000020020@X00P4500030@00014000<10000kP0001/0
00<100004@000`40003^00006`000`40000A00030@000>h0000K00814P000`40003^00006`000`40
000A00030@000>h0000K00030@00014000<10000kP0001/000<100004P000`40003]00006`000`40
000B00030@000>d0000K00814`000`40003]00006`000`40000B00030@000>d0000K00030@000180
00<10000k@0001/000<100004P000`40003]00006`020A<000<10000k@0001/000<100004P000`40
003]00000P020@@000<100000P030@<00`4500030@0001<000<10000k0000005004000402`000`40
000300030@00008000<100004`000`40003/000000D00@000@0;00030@0000<000<100000P020A@0
00<10000k0000005004000402@030@800`4500030@0001<000<10000k00000050040004020001040
004200030@0000D000<100004`000`40003/000000D00@000@0800040@000@8000<100001@000`40
000C00030@000>`0000200812P020@<0104400815@000`40003[00006`000`40000D00030@000>/0
000K00030@0001@000<10000j`0001/000<1000050000`40003[00006`020AD000<10000j`0001/0
00<1000050000`40003[00006`000`40000D00030@000>/0000K00030@0001D000<10000jP0001/0
0P4F00030@0001D0104a00030@0002l00P4`0081;@040@<00P4400006`000`40000E00030@0001D0
00<10000<P000`40000^00040@000Bh000@10001;P001@400001008000<100000@0001/000<10000
5@000`40000F00030@0002h01@4_00040@000Bh000@10001;P001@400001008000<100000@0001/0
00<100005@000`40000G00030@0002d000@10001<0030C000P4_00050@0000400P000`4000010000
6`020AH000<1000060000`40000]00030@01030000<10000;`001040004^00050@0000400P000`40
000100006`000`40000F00030@0001@000@10001<0020C0000<10000;`001040004/00<10`001040
004300006`000`40000F00030@0001D00P4b00030@0002l00`4_0081;`000`4000020081100001/0
00<100005P000`40003Y0000503o0@d1000K00030@0000T000<100002P000`40000:00030@0000T0
00<100002P000`40000:00030@0000T000<100002P000`40000:00030@0000T000<100002P000`40
000:00030@0000T000<100002P000`40000:00030@0000D0<`4400030@0000D0000K00030@0001L0
00<100005@000`40000`00030@00030000<100004`0H0@@000<10000<0000`40000500006`000`40
000G00030@0008L02P5G00006`020AP000<10000P0070F40000K00030@0001L000<10000NP060FP0
000K00030@0001L000<10000M@050Fh0000K00030@0001P000<10000KP060G<0000K00030@0001P0
00<10000JP040GT0000K00816@000`40001V00@1O@0001/000<1000060000`40001S00<1P@0001/0
00<1000060000`40001P00<1Q00001/000<100006@000`40001L00<1Q`0001/00P4J00030@0005X0
0P6:00006`000`40000I00030@0005L00`6<00006`000`40000I00030@0005@00`6?00006`000`40
000J00030@0005400P6B00006`020A/000<10000C`020I@0000K00030@0001X000<10000C@020IH0
000K00030@0001X000<10000B`020IP000000`010@020@<000<100000`020@<00`4500030@0001/0
00<10000B0020IX0000300030@0000L000@100011@000`400002008170000`4000160081W00000<0
00<100001`001040004500030@00008000<100006`000`40001500030@0009`0000300030@0000L0
00@100010P030@D000<100006`000`4000130081W`0000<000<100001`001040004200030@0000D0
00<1000070000`40001100030@0009l0000010010@4900040@000@8000<100001@020Ad000<10000
@0000`40002P00000`000`40000800810`040@@000<1000070000`40000n0081X`0001/000<10000
70000`40000m00030@000:<0000K00030@0001d000<10000>P020JH0000K00817P000`40000i0003
0@000:H0000K00030@0001d000<10000>0000`40002W00006`000`40000M00030@0003H00P6Z0000
6`000`40000N00030@0003@000<10000ZP0001/00P4O00030@0003<000<10000Z`0001/000<10000
7P000`40000a0081[P0001/000<100007P000`40000`00030@000:h0000K00030@0001l000<10000
;P000`40002_00006`020B0000<10000;0020K80000K00030@0001l000<10000:`000`40002b0000
6`000`40000O00030@0002X000<10000/`0000L0104300030@000080104400030@00020000<10000
:0000`40002d00002@000`40000900030@0000<00P4Q00030@0002H00P6g00002@000`4000090003
0@0000<000<100008@000`40000T00030@000;L0000900030@0000T000<100000`000`40000Q0003
0@0002<000<10000^00000T000<100002@000`40000300030@00028000<1000080020K/0000700<1
2@030@D00P4S00030@0001l000<10000^`0000T000<100002@000`40000300030@0002<000<10000
7@000`40002l00006`000`40000S00030@0001/00P6o00006`000`40000T00030@0001T000<10000
_`0001/000<1000090000`40000H00030@000<00000K00819P000`40000E0081``0001/000<10000
9P000`40000B0081a@0001/000<100009`000`40000@00030@000<D0000K00030@0002P00P4>0081
b00001/00P4[00030@0000T00P7:00006`000`40000[00811P030L`0000K00030@0002d01P7?0000
6`000`40003o00<0000K0081o`0400006`000`40003o00<0000K00030@000?l00`00000300410081
0`000`40000200@10P030@D000<10000o`0300000`000`40000900030@0000@000<100000P020Ol0
100000<000<100002@000`40000400030@00008000<10000o`0300000`000`40000900040@0000<1
1@000`40003o00<0000300030@0000T000D100000@0700030@000?l00`00000400410@T00`430003
0@0000D00P7o00@0000300030@0000T000@10000104400030@000?l00`0001/000<10000o`030000
6`000`40003o00<0000K0081o`0400006`000`40003o00<0000K00030@000?l00`0001/000<10000
o`0300006`020Ol0100001/000<10000o`0300006`000`40003o00<0000K00030@000?l00`0001/0
0P7o00@0000K00030@000?l00`0001/000<10000o`0300006`000`40003o00<0000K00030@000?l0
0`000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.07864, 0.779277, 
  0.0394728, 0.00243309}}]
}, Open  ]],

Cell["\<\
In order to build the center and radius functions, we plot several \
functions associated with the reference function
The closed loop plant transfer function is now plotted:\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
plotb = Plot[ Abs[ (1-Tref[I w]) p[I w] \
],{w,0,2},PlotRange->All];\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.476191 0.0115517 0.0525667 [
[.2619 -0.00095 -9 -9 ]
[.2619 -0.00095 9 0 ]
[.5 -0.00095 -3 -9 ]
[.5 -0.00095 3 0 ]
[.7381 -0.00095 -9 -9 ]
[.7381 -0.00095 9 0 ]
[.97619 -0.00095 -3 -9 ]
[.97619 -0.00095 3 0 ]
[.01131 .11669 -6 -4.5 ]
[.01131 .11669 0 4.5 ]
[.01131 .22182 -6 -4.5 ]
[.01131 .22182 0 4.5 ]
[.01131 .32695 -6 -4.5 ]
[.01131 .32695 0 4.5 ]
[.01131 .43209 -6 -4.5 ]
[.01131 .43209 0 4.5 ]
[.01131 .53722 -12 -4.5 ]
[.01131 .53722 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01155 m
.2619 .0178 L
s
[(0.5)] .2619 -0.00095 0 1 Mshowa
.5 .01155 m
.5 .0178 L
s
[(1)] .5 -0.00095 0 1 Mshowa
.7381 .01155 m
.7381 .0178 L
s
[(1.5)] .7381 -0.00095 0 1 Mshowa
.97619 .01155 m
.97619 .0178 L
s
[(2)] .97619 -0.00095 0 1 Mshowa
.125 Mabswid
.07143 .01155 m
.07143 .0153 L
s
.11905 .01155 m
.11905 .0153 L
s
.16667 .01155 m
.16667 .0153 L
s
.21429 .01155 m
.21429 .0153 L
s
.30952 .01155 m
.30952 .0153 L
s
.35714 .01155 m
.35714 .0153 L
s
.40476 .01155 m
.40476 .0153 L
s
.45238 .01155 m
.45238 .0153 L
s
.54762 .01155 m
.54762 .0153 L
s
.59524 .01155 m
.59524 .0153 L
s
.64286 .01155 m
.64286 .0153 L
s
.69048 .01155 m
.69048 .0153 L
s
.78571 .01155 m
.78571 .0153 L
s
.83333 .01155 m
.83333 .0153 L
s
.88095 .01155 m
.88095 .0153 L
s
.92857 .01155 m
.92857 .0153 L
s
.25 Mabswid
0 .01155 m
1 .01155 L
s
.02381 .11669 m
.03006 .11669 L
s
[(2)] .01131 .11669 1 0 Mshowa
.02381 .22182 m
.03006 .22182 L
s
[(4)] .01131 .22182 1 0 Mshowa
.02381 .32695 m
.03006 .32695 L
s
[(6)] .01131 .32695 1 0 Mshowa
.02381 .43209 m
.03006 .43209 L
s
[(8)] .01131 .43209 1 0 Mshowa
.02381 .53722 m
.03006 .53722 L
s
[(10)] .01131 .53722 1 0 Mshowa
.125 Mabswid
.02381 .03784 m
.02756 .03784 L
s
.02381 .06412 m
.02756 .06412 L
s
.02381 .0904 m
.02756 .0904 L
s
.02381 .14297 m
.02756 .14297 L
s
.02381 .16925 m
.02756 .16925 L
s
.02381 .19554 m
.02756 .19554 L
s
.02381 .2481 m
.02756 .2481 L
s
.02381 .27439 m
.02756 .27439 L
s
.02381 .30067 m
.02756 .30067 L
s
.02381 .35324 m
.02756 .35324 L
s
.02381 .37952 m
.02756 .37952 L
s
.02381 .4058 m
.02756 .4058 L
s
.02381 .45837 m
.02756 .45837 L
s
.02381 .48465 m
.02756 .48465 L
s
.02381 .51094 m
.02756 .51094 L
s
.02381 .5635 m
.02756 .5635 L
s
.02381 .58979 m
.02756 .58979 L
s
.02381 .61607 m
.02756 .61607 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
.5 Mabswid
.02381 .05994 m
.02499 .05884 L
.02605 .05884 L
.02729 .05884 L
.02846 .05884 L
.02954 .05884 L
.03053 .05884 L
.03163 .05884 L
.03279 .05884 L
.03395 .05884 L
.0352 .05884 L
.03638 .05884 L
.03746 .05884 L
.03873 .05884 L
.04011 .05884 L
.04141 .05884 L
.04262 .05884 L
.04386 .05884 L
.04521 .05884 L
.04765 .05884 L
.05003 .05884 L
.05221 .05884 L
.05459 .05884 L
.05719 .05884 L
.05991 .05884 L
.06244 .05884 L
.06506 .05885 L
.06791 .05885 L
.07308 .05885 L
.07812 .05885 L
.08278 .05885 L
.0878 .05886 L
.09335 .05886 L
.09914 .05887 L
.10458 .05888 L
.11441 .0589 L
.12476 .05894 L
.1336 .05898 L
.14327 .05903 L
.15354 .0591 L
.16288 .05919 L
.17378 .05931 L
.18383 .05945 L
.20301 .05981 L
.21311 .06005 L
.22381 .06035 L
.24338 .06107 L
.2617 .06197 L
.28198 .06329 L
.30052 .0649 L
Mistroke
.31954 .06707 L
.34028 .07024 L
.35999 .07437 L
.38098 .08058 L
.39129 .08461 L
.40243 .09002 L
.41295 .09648 L
.42261 .10402 L
.43238 .11387 L
.44148 .12602 L
.45146 .14446 L
.45706 .15837 L
.46215 .17436 L
.46698 .19361 L
.47209 .22033 L
.47646 .25066 L
.48128 .29621 L
.48397 .32955 L
.48644 .36657 L
.492 .47626 L
.49454 .53387 L
.49578 .5598 L
.4969 .57982 L
.49809 .59532 L
.49936 .60323 L
.50003 .60332 L
.50075 .60018 L
.50205 .58638 L
.51147 .37624 L
.51418 .32712 L
.51675 .28905 L
.52158 .23492 L
.52433 .21144 L
.52688 .19322 L
.53266 .16113 L
.53768 .14046 L
.54302 .12339 L
.55269 .10089 L
.56177 .08598 L
.57173 .07391 L
.58248 .06415 L
.59203 .0574 L
.60249 .05149 L
.62359 .04269 L
.64317 .03694 L
.66411 .0324 L
.68544 .02891 L
.70558 .02634 L
.74401 .02274 L
.76387 .02135 L
Mistroke
.7849 .02013 L
.82276 .0184 L
.86306 .01703 L
.90185 .01604 L
.9431 .01523 L
.97619 .01472 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@0004`00P440003
0@0000800`4g00@1=`040@<000<100000P030CP010490000B`001040004;00030@0003H000<10000
>0000`40000:00030@0003D000<100002P0004/000@100012`000`40000f00030@0003P000<10000
2P000`40000f00030@0000T0001;00040@000@P00`4i00030@0003P000<100001`030CX000<10000
200004/000@1000120000`40000i00030@0003P000<100001`000`40000k00030@0000L0001;0004
0@000@P000<10000=`030CP00`4900030@0003P000@100012@00018000<10000=`020@T0104h0003
0@0003P000<100001`040CP00P4:00004P000`40003o00`0000B00030@000?l0300000`0o`4A0@@0
000B00030@0000X000<100002P000`40000:00030@0000X000<100002P000`40000:00030@0000X0
00<100002P000`40000:00030@0000T000<100002P000`40000:00030@0000X000<100002P000`40
000:00030@0000X000<100002P000`40000U0@X0000B00030@0003h000<10000?@000`40000n0003
0@0000L04`4T00030@0000P0000B00030@000;`03@5200004P000`40002d00P1C`00018000<10000
[P060EL0000B00030@000:P01P5M00004P020J@01@5S00004P000`40002O00@1J000018000<10000
W@020F`0000B00030@0009/00P5^00004P000`40002I0081L000018000<10000U`020G80000B03l1
FP000`40001b00004P020Cd0405900030@0007<0000B00030@0004`01@5300030@0007@0000B0003
0@0005401@4m00030@0007D0000B00030@0005H0104h00030@0007H0000B00030@0005X00`4d0003
0@0007L0000B00030@0005d00`4`00030@0007P0000B00030@0006000`4/00030@0007T0000B0081
I0020BX000<10000N@00018000<10000I@020BL000<10000NP00018000<10000I`000`40000T0003
0@0007X0000B00030@0006P000<100008P000`40001k00004P000`40001Y00030@00024000<10000
N`00018000<10000JP000`40000O00030@0007`0000:00@110000`40001[00030@0001h000<10000
O00000X000<100001@030F`000<1000070000`40001m00002`000`40000400030@0006d000<10000
6`000`40001m000030000`40000300030@0006h000<100006@000`40001n00003@000`4000020003
0@0006l000<1000060000`40001n00002P001040004400030@0006l000<1000060000`40001n0000
2`020@D000<10000L0000`40000F00030@0007l0000B00030@00070000<100005P000`40001o0000
4P000`40001a00030@0001D000<10000O`0001800P5b00030@0001D000<10000O`00018000<10000
LP000`40000C00030@000800000B00030@00078000<100004`000`40002000004P000`40001c0003
0@00018000<10000P000018000<10000L`000`40000A00030@000840000B00030@0007<000<10000
4@000`40002100004P000`40001d00030@00010000<10000P@0001800P5e00030@00010000<10000
P@00018000<10000M0000`40000@00030@000840000B00030@0007@000<1000040000`4000210000
4P000`40001e00030@0000h000<10000PP00018000<10000M@000`40000>00030@000880000B0003
0@0007H000<100003@000`40002200004P000`40001f00030@0000d000<10000PP0001800P5g0003
0@0000d000<10000PP00018000<10000MP000`40000=00030@000880000B00030@0007H000<10000
30000`40002300004P000`40001g00030@0000/000<10000P`00018000<10000M`000`40000;0003
0@0008<0000B00030@0007L000<100002`000`400023000030000`40000300030@0007L000<10000
2`000`400023000030000`40000300<1M`000`40000;00030@0008<0000900D110000`40001g0003
0@0000X000<10000Q00000T000@100011@000`40001g00030@0000X000<10000Q00000X000<10040
1@000`40001g00030@0000X000<10000Q00000/00P4500030@0007P000<100002@000`4000240000
30000`40000300030@0007P000<100002@000`40002400004P000`40001h00030@0000T000<10000
Q00001800P5i00030@0000T000<10000Q000018000<10000N0000`40000900030@0008@0000B0003
0@0007P000<100002@000`40002400004P000`40001h00030@0000T000<10000Q000018000<10000
N0000`40000900030@0008@0000B00030@0007P000<1000020000`40002500004P000`40001h0003
0@0000P000<10000Q@0001800P5j00030@0000L000<10000Q@00018000<10000N@000`4000070003
0@0008D0000B00030@0007T000<100001`000`40002500004P000`40001i00030@0000L000<10000
Q@00018000<10000N@000`40000700030@0008D0000B00030@0007T000<100001`000`4000250000
4P000`40001i00030@0000L000<10000Q@0001800P5j00030@0000L000<10000Q@00018000<10000
N@000`40000700030@0008D0000B00030@0007T000<100001`000`40002500004P000`40001i0003
0@0000H000<10000QP00018000<10000NP000`40000500030@0008H0000B00030@0007X000<10000
1@000`40002600002`020@D000<10000NP000`40000500030@0008H0000:00040@000@@000<10000
NP000`40000500030@0008H0000:00040@000@@00`5j00030@0000D000<10000QP0000X00`450003
0@0007X000<100001@000`40002600002P000`40000500030@0007X000<100001@000`4000260000
2P000`40000500030@0007X000<100001@000`40002600002`030@@000<10000NP000`4000050003
0@0008H0000B00030@0007/000<1000010000`40002600004P000`40001k00030@0000@000<10000
QP0001800P5l00030@0000@000<10000QP00018000<10000N`000`40000400030@0008H0000B0003
0@0007/000<1000010000`40002600004P000`40001k00030@0000@000<10000QP00018000<10000
N`000`40000400030@0008H0000B00030@0007/000<1000010000`40002600004P000`40001k0003
0@0000@000<10000QP0001800P5l00030@0000@000<10000QP00018000<10000N`000`4000040003
0@0008H0000B00030@0007/000<1000010000`40002600004P000`40001k00030@0000@000<10000
QP00018000<10000N`000`40000400030@0008H0000B00030@0007/000<1000010000`4000260000
4P000`40001k00030@0000@000<10000QP0001800P5l00030@0000@000<10000QP00018000<10000
N`000`40000400030@0008H0000B00030@0007/000<100000`000`40002700004P000`40001k0003
0@0000<000<10000Q`00018000<10000O0000`40000200030@0008L0000B00030@0007`000<10000
0P000`40002700002`020@D000<10000O0000`40000200030@0008L0000:00040@000@@00`5l0003
0@00008000<10000Q`0000X000@1000110000`40001l00030@00008000<10000Q`0000/00P450003
0@0007`000<100000P000`40002700002P001040004400030@0007`000<100000P000`4000270000
2P001040004400030@0007`000<100000P000`40002700002`020@D000<10000O0000`4000020003
0@0008L0000B00030@0007`000<100000P000`40002700004P020Gd000<100000P000`4000270000
4P000`40001l00030@00008000<10000Q`00018000<10000O0000`40000200030@0008L0000B0003
0@0007`000<100000P000`40002700004P000`40001l00030@00008000<10000Q`00018000<10000
O0000`40000200030@0008L0000B00030@0007`000<100000P000`40002700004P020Gd000D10000
0@2:00004P000`40001l00050@000040RP00018000<10000O0001@40000108X0000B00030@0007`0
00D100000@2:00004P000`40001l00050@000040RP00018000<10000O0001@40000108X0000B0003
0@0007d000@10001RP0001800P5n00040@000HX0000B00030@0007d000@10001RP00018000<10000
O@001040006:00004P000`40001m00040@000HX0000B00030@0007d000@10001RP00018000<10000
O@001040006:000010040@<00P4500030@0007d000@10001RP0000H000D100000@0200030@000080
00<10000O@001040006:00001P001@400001008000<100000P030Gd000@10001RP0000H000D10000
0@0200030@00008000<10000O@001040006:00001P001@400001008000<100000P000`40001m0004
0@000HX0000400<10`001040004400030@0007d000@10001RP0000H000<100000P020@D000<10000
O@000`400@2;00004P000`40001m00030@0108/0000B00030@0007d000<10040R`0001800P5n0003
0@0108/0000B00030@0007d000<10040R`00018000<10000OP020H/0000B00030@0007h00P6;0000
4P000`40001n0081R`00018000<10000OP020H/0000B00030@0007h00P6;00004P020Gl00P6;0000
4P000`40001n0081R`00018000<10000OP020H/0000B00030@0007h00P6;00004P000`40001n0003
0@0008X0000B00030@000?l03000018000<10000o`0<00004P020Ol03@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.145285, -0.868936, 
  0.0077402, 0.0701169}}]
}, Open  ]],

Cell["The complementary sensitivity for the reference function ", "Text"],

Cell[CellGroupData[{

Cell["Plot[Abs[Tref[I w]],{w,0,12}]", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0793651 0.0147151 0.509749 [
[.18254 .00222 -3 -9 ]
[.18254 .00222 3 0 ]
[.34127 .00222 -3 -9 ]
[.34127 .00222 3 0 ]
[.5 .00222 -3 -9 ]
[.5 .00222 3 0 ]
[.65873 .00222 -3 -9 ]
[.65873 .00222 3 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .11666 -18 -4.5 ]
[.01131 .11666 0 4.5 ]
[.01131 .21861 -18 -4.5 ]
[.01131 .21861 0 4.5 ]
[.01131 .32056 -18 -4.5 ]
[.01131 .32056 0 4.5 ]
[.01131 .42251 -18 -4.5 ]
[.01131 .42251 0 4.5 ]
[.01131 .52446 -6 -4.5 ]
[.01131 .52446 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(2)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(4)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(6)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(8)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(10)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(12)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.06349 .01472 m
.06349 .01847 L
s
.10317 .01472 m
.10317 .01847 L
s
.14286 .01472 m
.14286 .01847 L
s
.22222 .01472 m
.22222 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30159 .01472 m
.30159 .01847 L
s
.38095 .01472 m
.38095 .01847 L
s
.42063 .01472 m
.42063 .01847 L
s
.46032 .01472 m
.46032 .01847 L
s
.53968 .01472 m
.53968 .01847 L
s
.57937 .01472 m
.57937 .01847 L
s
.61905 .01472 m
.61905 .01847 L
s
.69841 .01472 m
.69841 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.77778 .01472 m
.77778 .01847 L
s
.85714 .01472 m
.85714 .01847 L
s
.89683 .01472 m
.89683 .01847 L
s
.93651 .01472 m
.93651 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11666 m
.03006 .11666 L
s
[(0.2)] .01131 .11666 1 0 Mshowa
.02381 .21861 m
.03006 .21861 L
s
[(0.4)] .01131 .21861 1 0 Mshowa
.02381 .32056 m
.03006 .32056 L
s
[(0.6)] .01131 .32056 1 0 Mshowa
.02381 .42251 m
.03006 .42251 L
s
[(0.8)] .01131 .42251 1 0 Mshowa
.02381 .52446 m
.03006 .52446 L
s
[(1)] .01131 .52446 1 0 Mshowa
.125 Mabswid
.02381 .0402 m
.02756 .0402 L
s
.02381 .06569 m
.02756 .06569 L
s
.02381 .09118 m
.02756 .09118 L
s
.02381 .14215 m
.02756 .14215 L
s
.02381 .16764 m
.02756 .16764 L
s
.02381 .19313 m
.02756 .19313 L
s
.02381 .2441 m
.02756 .2441 L
s
.02381 .26959 m
.02756 .26959 L
s
.02381 .29508 m
.02756 .29508 L
s
.02381 .34605 m
.02756 .34605 L
s
.02381 .37154 m
.02756 .37154 L
s
.02381 .39703 m
.02756 .39703 L
s
.02381 .448 m
.02756 .448 L
s
.02381 .47349 m
.02756 .47349 L
s
.02381 .49898 m
.02756 .49898 L
s
.02381 .54995 m
.02756 .54995 L
s
.02381 .57544 m
.02756 .57544 L
s
.02381 .60093 m
.02756 .60093 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .52446 m
.02499 .52458 L
.02605 .52487 L
.02729 .52544 L
.02846 .52619 L
.03053 .52804 L
.03279 .53074 L
.0379 .53911 L
.04262 .54882 L
.05285 .57181 L
.06244 .58986 L
.06713 .59615 L
.06977 .59883 L
.07218 .60073 L
.07345 .60152 L
.07481 .6022 L
.07559 .60252 L
.07629 .60276 L
.07701 .60296 L
.07767 .60311 L
.07897 .60328 L
.07965 .60332 L
.08036 .60331 L
.08157 .60321 L
.08226 .6031 L
.0829 .60296 L
.08424 .60257 L
.08571 .60199 L
.08837 .60053 L
.09096 .59864 L
.09336 .59652 L
.10458 .58257 L
.12324 .5499 L
.14385 .50871 L
.18405 .43244 L
.22274 .37252 L
.26388 .32267 L
.3035 .28513 L
.34161 .25623 L
.38217 .23126 L
.42122 .21148 L
.46271 .19396 L
.5027 .17972 L
.54116 .16797 L
.58208 .15714 L
.62148 .14804 L
.65937 .14031 L
.69971 .133 L
.73853 .12672 L
.77981 .12072 L
Mistroke
.81957 .11552 L
.85782 .11097 L
.89851 .10656 L
.9377 .10267 L
.97619 .09916 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@0003h0
104Y00030@0002L00P4Z00819@040@<00P4R00@10P040@<0000n00030@0002X000<100009P001040
004X00040@000BH000D100000@0200030@00024000D100000@060000?`000`40000V00D19`001040
004X00040@000BH000D100000@0200030@00024000<100000P000`4000030000@0000`40000U0004
0@000BP00`4Z00819`001@400001008000<100008@000`40000300030@000080000E00030@0002T0
00<100009@000`400@0X00030@0002T000@100019P001@400001008000<100008@000`4000040003
0@000040000E00030@0002H000@10001:0020BP000<10000:@001040004T00<10`001040004Q00<1
0`001040004300005@000`40000W0081:P000`40000W00<1:@020BL000<100000P020B@000<10000
0P020@@0000E00030@000?l02@0000h0o`4C0@005@000`40000800030@0000L000<1000020000`40
000800030@0000P000<1000020000`40000700030@0000P000<1000020000`40000800030@0000P0
00<100001`000`40000800030@0000P000<1000020000`40000800030@0000L000<1000020000`40
000800030@0000P000<1000020000`40000700030@0000P000<1000020000`40000500005@000`40
000X00030@0002P000<10000:0000`40000Y00030@0002P000<10000:0000`40000500005@000`40
003o00T0000E00030@000?l02@0001D000<10000o`0900005@000`40003o00T0000E0081o`0:0000
5@000`40003o00T0000E00030@000?l02@0001D000<10000o`0900005@000`40003o00T0000E0003
0@000?l02@0001D000<10000o`0900005@020Ol02P0001D000<10000o`0900005@000`40003o00T0
000E00030@000?l02@0001D000<10000o`0900005@000`40003o00T0000E00030@000?l02@0001D0
0P7o00X0000E00030@000?l02@0001D000<10000n`060@L0000E00030@000?402P4=00005@000`40
003V00/15`0001D000<10000gP080B800002008110000`40000200@110000`40003H00H1:P000005
0040004020000`40000500<1d0080C0000001@01000100T000<1000010000`40003800P1>0000005
004000402P000`40000300030@000<801P50000000D00@000@0;00030@00008000<10000^P080DH0
00001@01000100P000@1000110000`40002c00L1CP0000800P4:00811@000`40002]00H1E@0001D0
00<10000Z0050E/0000E0081Y0050F00000E00030@0009l0105U00005@000`40002K00@1J@0001D0
00<10000U`040Fd0000E00030@0009@00`5a00005@000`40002@00@1M00001D000<10000S0040GP0
000E0081RP030G`0000E00030@0008D0105o00005@000`40002100@1P`0001D000<10000OP030HL0
000E00030@0007/00`6:00005@000`40001i0081S@0001D000<10000MP030Hl0000E0081M0030I80
000E00030@0007400P6E00005@000`40001_0081U`0001D000<10000K@020IT0000E00030@0006/0
0P6K00000P020@@000<1000010000`40000300030@0006P00`6M000000D00@000@0:00030@0000<0
00<10000IP020J0000001@01000100L01@4400<1I0020J8000001@01000100L000@100011@000`40
001R0081Y00000050040004020000`400@0500030@0006000P6V000000D00@000@0900811@000`40
001N0081Z00000800P4;00030@0000<000<10000G0020JX0000E00030@0005/000<10000ZP0001D0
00<10000F@020Jd0000E0081F0020Jl0000E00030@0005H000<10000[`0001D000<10000E0020K80
000E00030@0005<000<10000/P0001D000<10000DP000`40002c00005@000`40001@0081]P0001D0
00<10000C`000`40002f00005@020Dl000<10000]`0001D000<10000C@000`40002h00005@000`40
001;0081^`0001D000<10000BP000`40002k00005@000`40001900030@000;`0000E00030@0004P0
00<10000_@0001D00P5800030@000;h0000E00030@0004H000<10000_`0001D000<10000A@000`40
003000005@000`4000130081``0001D000<10000@P000`40003300005@000`40001100030@000<@0
0002008110000`40000300811@000`40001000030@000<D000001@01000100P000@1000110030Cl0
00<10000aP0000050040004020001040004400030@0003h000<10000a`0000050040004020030@D0
00<10000?@000`400038000000D00@000@0800030@0000D000<10000?0000`400039000000D00@00
0@0800030@0000D000<10000?0000`40003900000P020@X00`4400030@0003/000<10000bP0001D0
00<10000>P000`40003;00005@020CX000<10000c00001D000<10000>@000`40003<00005@000`40
000h00030@000<d0000E00030@0003L000<10000cP0001D000<10000=P000`40003?00005@000`40
000e00030@000=00000E00030@0003D000<10000d00001D00P4e00030@000=40000E00030@0003<0
00<10000dP0001D000<10000<P000`40003C00005@000`40000b00030@000=<0000E00030@000340
00<10000e00001D000<10000<@000`40003D00005@000`40000`00030@000=D0000E0081<0000`40
003F00005@000`40000_00030@000=H0000E00030@0002h000<10000e`0001D000<10000;@000`40
003H00005@000`40000]00030@000=P0000E00030@0002`000<10000f@0000800P4400030@0000<0
0P4500030@0002`000<10000f@0000050040004020001040004400<1:`000`40003J000000D00@00
0@0800040@000@@000<10000:P000`40003K000000D00@000@0900811@000`40000Z00030@000=/0
00001@01000100P000@1000110000`40000Y00030@000=`000001@01000100P000@1000110000`40
000X00030@000=d0000200812P020@D000<10000:0000`40003M00005@000`40000W00030@000=h0
000E0081:0000`40003N00005@000`40000V00030@000=l0000E00030@0002H000<10000g`0001D0
00<100009@000`40003P00005@000`40000U00030@000>00000E00030@0002@000<10000h@0001D0
00<1000090000`40003Q00005@020B@000<10000hP0001D000<100008`000`40003R00005@000`40
000R00030@000><0000E00030@00028000<10000h`0001D000<100008@000`40003T00005@000`40
000Q00030@000>@0000E00030@00020000<10000i@0001D00P4Q00030@000>D0000E00030@0001l0
00<10000iP0001D000<100007`000`40003V00005@000`40000N00030@000>L0000E00030@0001d0
00<10000j00001D000<100007@000`40003X00003@040@@000<1000070000`40003Y00003`000`40
000300<170000`40003Y00003`000`40000300030@0101/000<10000jP0000l000<100000`000`40
0@0K00030@000>X0000?00030@0000<000@100016@000`40003[00003@030@D000D100000@0H0003
0@000>/0000?00030@0000<000<100000P000`40000D00030@000>`0000E00030@00008000<10000
50000`40003/00005@020@<000<100004`000`40003]00005@000`40000300030@00014000<10000
kP0001D000<100000`000`40000A00030@000>h0000E00030@0000@000<100003`000`40003_0000
5@000`40000400030@0000l000<10000k`0001D000<100001@000`40000=00030@000?00000E0003
0@0000D000<100003@000`40003`00005@020@H000<1000030000`40003a00005@000`4000060003
0@0000/000<10000l@0001D000<100001P000`40000:00030@000?80000E00030@0000L000<10000
20000`40003c00005@000`400007008120000`40003d00005@000`40000900030@0000@000<10000
m@0001D000<100002@000`40000300030@000?H0000E00812`050OT0000E00030@000?l02@0001D0
00<10000o`0900005@000`40003o00T0000E00030@000?l02@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.991151, -0.0908161, 
  0.0463108, 0.00721034}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["and the sensitivty for the reference function ", "Text"],

Cell[CellGroupData[{

Cell["Plot[Abs[1-Tref[I w]],{w,0,12}]", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0793651 0.0147151 0.592691 [
[.18254 .00222 -3 -9 ]
[.18254 .00222 3 0 ]
[.34127 .00222 -3 -9 ]
[.34127 .00222 3 0 ]
[.5 .00222 -3 -9 ]
[.5 .00222 3 0 ]
[.65873 .00222 -3 -9 ]
[.65873 .00222 3 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .13325 -18 -4.5 ]
[.01131 .13325 0 4.5 ]
[.01131 .25179 -18 -4.5 ]
[.01131 .25179 0 4.5 ]
[.01131 .37033 -18 -4.5 ]
[.01131 .37033 0 4.5 ]
[.01131 .48887 -18 -4.5 ]
[.01131 .48887 0 4.5 ]
[.01131 .60741 -6 -4.5 ]
[.01131 .60741 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(2)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(4)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(6)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(8)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(10)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(12)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.06349 .01472 m
.06349 .01847 L
s
.10317 .01472 m
.10317 .01847 L
s
.14286 .01472 m
.14286 .01847 L
s
.22222 .01472 m
.22222 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30159 .01472 m
.30159 .01847 L
s
.38095 .01472 m
.38095 .01847 L
s
.42063 .01472 m
.42063 .01847 L
s
.46032 .01472 m
.46032 .01847 L
s
.53968 .01472 m
.53968 .01847 L
s
.57937 .01472 m
.57937 .01847 L
s
.61905 .01472 m
.61905 .01847 L
s
.69841 .01472 m
.69841 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.77778 .01472 m
.77778 .01847 L
s
.85714 .01472 m
.85714 .01847 L
s
.89683 .01472 m
.89683 .01847 L
s
.93651 .01472 m
.93651 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13325 m
.03006 .13325 L
s
[(0.2)] .01131 .13325 1 0 Mshowa
.02381 .25179 m
.03006 .25179 L
s
[(0.4)] .01131 .25179 1 0 Mshowa
.02381 .37033 m
.03006 .37033 L
s
[(0.6)] .01131 .37033 1 0 Mshowa
.02381 .48887 m
.03006 .48887 L
s
[(0.8)] .01131 .48887 1 0 Mshowa
.02381 .60741 m
.03006 .60741 L
s
[(1)] .01131 .60741 1 0 Mshowa
.125 Mabswid
.02381 .04435 m
.02756 .04435 L
s
.02381 .07398 m
.02756 .07398 L
s
.02381 .10362 m
.02756 .10362 L
s
.02381 .16289 m
.02756 .16289 L
s
.02381 .19252 m
.02756 .19252 L
s
.02381 .22216 m
.02756 .22216 L
s
.02381 .28143 m
.02756 .28143 L
s
.02381 .31106 m
.02756 .31106 L
s
.02381 .3407 m
.02756 .3407 L
s
.02381 .39996 m
.02756 .39996 L
s
.02381 .4296 m
.02756 .4296 L
s
.02381 .45923 m
.02756 .45923 L
s
.02381 .5185 m
.02756 .5185 L
s
.02381 .54814 m
.02756 .54814 L
s
.02381 .57777 m
.02756 .57777 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01485 L
.02605 .01519 L
.02729 .01585 L
.02846 .01674 L
.03053 .01894 L
.03279 .0222 L
.03527 .02682 L
.0379 .03282 L
.04262 .04623 L
.05205 .08134 L
.06244 .12826 L
.08274 .22537 L
.10458 .31626 L
.12562 .38337 L
.14487 .42924 L
.16385 .46333 L
.18458 .49127 L
.20425 .51133 L
.22522 .52775 L
.24663 .54068 L
.2668 .55027 L
.30535 .56378 L
.32529 .56899 L
.34636 .57357 L
.38433 .58001 L
.42475 .58506 L
.46366 .58872 L
.50502 .59171 L
.54487 .59397 L
.5832 .59571 L
.62398 .59722 L
.66325 .59841 L
.701 .59938 L
.7412 .60024 L
.77989 .60095 L
.82103 .60159 L
.86065 .60212 L
.89876 .60257 L
.93932 .60299 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@0003l0104Y0003
0@0002H00P4Y00819@040@<00P4R00@10P040@@0000o00030@0002X000<100009@001040004W0004
0@000BH000D100000@0200030@00024000D100000@070000@0000`40000V00D19P001040004W0004
0@000BH000D100000@0200030@00024000<100000P000`4000040000@@000`40000U00040@000BL0
0`4Y00819`001@400001008000<100008@000`40000300030@0000<0001200030@0002D000<10040
9`000`40000X00040@000BH000D100000@0200030@00024000<1000010000`40000200005P000`40
000V00040@000BP00P4W00030@0002P000@1000190030@<000@100018@030@<000@10001100001H0
00<100009`020BX000<100009P030BP00P4W00030@0000800P4T00030@0000800P4500005P000`40
003o00P0000F00030@000?l0200000l0o`4A0@40000F00030@0100P000<100001`000`4000080003
0@0000P000<100001`000`40000800030@0000P000<1000020000`40000700030@0000P000<10000
20000`40000700030@0000P000<1000020000`40000800030@0000L000<1000020000`4000080003
0@0000L000<1000020000`40000800030@0000P000<100001`000`40000800030@0000H0000F0003
0@0102P000<10000:0000`40000W00030@0002P000<10000:0000`40000X00030@0000H0000F0004
0@000Ol01`0001H000D100000@3o00H0000F00050@000040o`0600005P001@4000010?l01P0001H0
00<100000P000`40003o00<0000F00810`000`40003o00<0000F00030@00008000<10000o`030000
5P000`40000300030@000?l00P0001H000<100000`000`40003o0080000F00030@0000<000<10000
o`0200005P000`40000400030@000?l00@0001H000<1000010000`40003o0040000F00030@0000@0
00<10000o`0100005P020@D000<10000o`0100005P000`40000500030@000?l0000F00030@0000D0
00<10000o`0001H000<100001@000`40003o00005P000`40000500030@000?l0000F00030@0000D0
00<10000o`0001H000<100001P000`40003n00005P000`40000600030@000?h0000F00811`000`40
003n00005P000`40000600030@000?h0000F00030@0000H000<10000oP0001H000<100001P000`40
003n00005P000`40000700030@000?d0000F00030@0000L000<10000o@0001H000<100001`000`40
003m00000`020@@000<100000P040@@000<100001`000`40003m00000P001040004800030@0000D0
0`4700030@000?d0000200040@000@T000<1000010000`40000700030@000?d0000200040@000@X0
00<100000`000`40000800030@000?`0000200040@000@/000<100000P000`40000800030@000?`0
000200040@000@P000@1000110000`40000800030@000?`0000300812P020@D000<1000020000`40
003l00005P000`40000900030@000?/0000F00030@0000T000<10000n`0001H00P4:00030@000?/0
000F00030@0000T000<10000n`0001H000<100002P000`40003j00005P000`40000:00030@000?X0
000F00030@0000X000<10000nP0001H000<100002P000`40003j00005P000`40000:00030@000?X0
000F00030@0000/000<10000n@0001H00P4<00030@000?T0000F00030@0000/000<10000n@0001H0
00<100002`000`40003i00005P000`40000<00030@000?P0000F00030@0000`000<10000n00001H0
00<1000030000`40003h00005P000`40000<00030@000?P0000F00030@0000d000<10000m`0001H0
0P4>00030@000?L0000F00030@0000d000<10000m`0001H000<100003@000`40003g00005P000`40
000>00030@000?H0000F00030@0000h000<10000mP0001H000<100003P000`40003f00005P000`40
000>00030@000?H00003008110000`40000400030@0000<000<100003`000`40003e00000P001040
004:00030@0000<00`4?00030@000?D0000200040@000@L01@4400030@0000l000<10000m@000080
00@100011`001040004500030@0000l000<10000m@00008000@1000120000`400@0500030@000100
00<10000m000008000@100012@020@D000<1000040000`40003d00000`020@/000<100000`000`40
000@00030@000?@0000F00030@00010000<10000m00001H000<100004@000`40003c00005P020A80
00<10000l`0001H000<100004@000`40003c00005P000`40000A00030@000?<0000F00030@000180
00<10000lP0001H000<100004P000`40003b00005P000`40000B00030@000?80000F00030@000180
00<10000lP0001H000<100004`000`40003a00005P020A@000<10000l@0001H000<100004`000`40
003a00005P000`40000C00030@000?40000F00030@0001@000<10000l00001H000<1000050000`40
003`00005P000`40000D00030@000?00000F00030@0001@000<10000l00001H000<100005@000`40
003_00005P020AH000<10000k`0001H000<100005@000`40003_00005P000`40000F00030@000>h0
000F00030@0001H000<10000kP0001H000<100005P000`40003^00005P000`40000F00030@000>h0
000F00030@0001L000<10000k@0000<00P4400030@0000<00P4500030@0001L000<10000k@000080
00@1000120001040004400<15`000`40003]00000P001040004800040@000@@000<100005`000`40
003]00000P001040004800<11@000`40000H00030@000>`0000200040@000@P000<100001@000`40
000H00030@000>`0000200040@000@P000<100001@000`40000H00030@000>`0000300812P030@@0
00<100006@000`40003[00005P000`40000I00030@000>/0000F00030@0001X000<10000jP0001H0
0P4K00030@000>X0000F00030@0001/000<10000j@0001H000<100006`000`40003Y00005P000`40
000L00030@000>P0000F00030@0001`000<10000j00001H000<100007@000`40003W00005P000`40
000M00030@000>L0000F00030@0001h000<10000iP0001H00P4O00030@000>H0000F00030@0001l0
00<10000i@0001H000<100007`000`40003U00005P000`40000P00030@000>@0000F00030@000200
00<10000i00001H000<100008@000`40003S00005P000`40000Q00030@000><0000F00030@000280
00<10000hP0001H00P4S00030@000>80000F00030@0002<000<10000h@0001H000<1000090000`40
003P00005P000`40000T00030@000>00000F00030@0002D000<10000g`0001H000<100009P000`40
003N00005P000`40000V00030@000=h00003008110000`40000300811@000`40000W00030@000=d0
000200040@000@P000@1000110030BL000<10000g@00008000@1000120001040004400030@0002P0
00<10000g000008000@100012@020@D000<10000:@000`40003K00000P001040004800040@000@@0
00<10000:P020M/0000200040@000@P000@1000110000`40000/00030@000=P0000300812P020@D0
00<10000;@000`40003G00005P000`40000^00030@000=H0000F00030@0002l000<10000e@0001H0
0P4a0081e@0001H000<10000<P000`40003B00005P000`40000c00030@000=40000F00030@0003@0
0P7A00005P000`40000f00030@000<h0000F00030@0003L00P7>00005P000`40000i00030@000</0
000F00030@0003X00P7;00005P020Cd00P7900005P000`40000n0081a`0001H000<10000@0040L<0
000F00030@0004@0106o00005P000`40001800<1_00001H000<10000B`020KX0000F00030@0004d0
1@6e00005P000`40001B00D1/00001H00P5H00D1Z`0001H000<10000G0080J<0000F00030@0006@0
2`6H00005P000`40001_00/1S@0001H000<10000NP0F0GL0000F00030@0009005@5R00005P000`40
002U03D1;@0000h0104400030@000=X09@48000040000`40000300<1o`08000040000`4000030003
0@000?l02000010000<100000`000`40003o00P0000@00030@0000<000<10000o`0800003P030Ol0
40000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.04391, -0.0773159, 
  0.0466568, 0.00624764}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["\<\
Using the information provided by the plots we produced, we are \
ready to define a radius and center functions:\
\>", "Text"],

Cell["\<\

wp = 0.7;    alphap = 0.9;
wb = 2.0;    alphab = 0.75;
wr = 10.0;   alphar = 0.25/Abs[p[I wr]];

line1[w_] = InterpolatingPolynomial[
                   {{wp,alphap Abs[pinv[I wp]]},{wb,alphab}},w];

line2[w_] = InterpolatingPolynomial[
                   {{wb,alphab},{wr,alphar Abs[p[I wr]]}},w];

line3[w_] = InterpolatingPolynomial[{{wp,1},{wb,0}},w];

k[w_] = Which[      Abs[w] < wp , 1,
\t\t            Abs[w] < wb ,line3[Abs[w]],
                    True        , 0];

r[w_] = Which[      Abs[w] < wp , alphap Abs[pinv[I w]],
\t\t    Abs[w] < wb , line1[Abs[w]],
\t\twb< Abs[w] < wr , line2[Abs[w]],
\t           True         , alphar Abs[p[I w]]];
\
\>", "Input"],

Cell["Lets take a look at the envelopes", "Text"],

Cell[CellGroupData[{

Cell["\<\

EnvelopePlot[Radius->r,Center->k,FrequencyBand->{0.01,12}];

EnvelopePlot3D[Radius->r,Center->k,FrequencyBand->{0.01,12}];
\
\>", "Input"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.000834028 0.0834028 0.23528 0.294369 [
[.16597 .22278 -3 -9 ]
[.16597 .22278 3 0 ]
[.33278 .22278 -3 -9 ]
[.33278 .22278 3 0 ]
[.49958 .22278 -3 -9 ]
[.49958 .22278 3 0 ]
[.66639 .22278 -3 -9 ]
[.66639 .22278 3 0 ]
[.83319 .22278 -6 -9 ]
[.83319 .22278 6 0 ]
[1 .22278 -6 -9 ]
[1 .22278 6 0 ]
[-0.01333 .0881 -24 -4.5 ]
[-0.01333 .0881 0 4.5 ]
[-0.01333 .38246 -18 -4.5 ]
[-0.01333 .38246 0 4.5 ]
[-0.01333 .52965 -6 -4.5 ]
[-0.01333 .52965 0 4.5 ]
[ -0.00083 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.16597 .23528 m
.16597 .24153 L
s
[(2)] .16597 .22278 0 1 Mshowa
.33278 .23528 m
.33278 .24153 L
s
[(4)] .33278 .22278 0 1 Mshowa
.49958 .23528 m
.49958 .24153 L
s
[(6)] .49958 .22278 0 1 Mshowa
.66639 .23528 m
.66639 .24153 L
s
[(8)] .66639 .22278 0 1 Mshowa
.83319 .23528 m
.83319 .24153 L
s
[(10)] .83319 .22278 0 1 Mshowa
1 .23528 m
1 .24153 L
s
[(12)] 1 .22278 0 1 Mshowa
.125 Mabswid
.04087 .23528 m
.04087 .23903 L
s
.08257 .23528 m
.08257 .23903 L
s
.12427 .23528 m
.12427 .23903 L
s
.20767 .23528 m
.20767 .23903 L
s
.24937 .23528 m
.24937 .23903 L
s
.29108 .23528 m
.29108 .23903 L
s
.37448 .23528 m
.37448 .23903 L
s
.41618 .23528 m
.41618 .23903 L
s
.45788 .23528 m
.45788 .23903 L
s
.54128 .23528 m
.54128 .23903 L
s
.58299 .23528 m
.58299 .23903 L
s
.62469 .23528 m
.62469 .23903 L
s
.70809 .23528 m
.70809 .23903 L
s
.74979 .23528 m
.74979 .23903 L
s
.79149 .23528 m
.79149 .23903 L
s
.8749 .23528 m
.8749 .23903 L
s
.9166 .23528 m
.9166 .23903 L
s
.9583 .23528 m
.9583 .23903 L
s
.25 Mabswid
0 .23528 m
1 .23528 L
s
-0.00083 .0881 m
.00542 .0881 L
s
[(-0.5)] -0.01333 .0881 1 0 Mshowa
-0.00083 .38246 m
.00542 .38246 L
s
[(0.5)] -0.01333 .38246 1 0 Mshowa
-0.00083 .52965 m
.00542 .52965 L
s
[(1)] -0.01333 .52965 1 0 Mshowa
.125 Mabswid
-0.00083 .11753 m
.00292 .11753 L
s
-0.00083 .14697 m
.00292 .14697 L
s
-0.00083 .17641 m
.00292 .17641 L
s
-0.00083 .20584 m
.00292 .20584 L
s
-0.00083 .26472 m
.00292 .26472 L
s
-0.00083 .29415 m
.00292 .29415 L
s
-0.00083 .32359 m
.00292 .32359 L
s
-0.00083 .35303 m
.00292 .35303 L
s
-0.00083 .4119 m
.00292 .4119 L
s
-0.00083 .44134 m
.00292 .44134 L
s
-0.00083 .47077 m
.00292 .47077 L
s
-0.00083 .50021 m
.00292 .50021 L
s
-0.00083 .05866 m
.00292 .05866 L
s
-0.00083 .02922 m
.00292 .02922 L
s
-0.00083 .55909 m
.00292 .55909 L
s
-0.00083 .58852 m
.00292 .58852 L
s
-0.00083 .61796 m
.00292 .61796 L
s
.25 Mabswid
-0.00083 0 m
-0.00083 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
1 .18543 m
.84597 .16765 L
.67538 .13353 L
.56142 .10618 L
.47983 .08646 L
.41846 .07199 L
.37058 .06088 L
.33213 .05205 L
.30055 .04486 L
.27411 .03887 L
.25163 .0338 L
.23226 .02945 L
.21538 .02566 L
.20052 .02237 L
.18732 .02 L
.17551 .02168 L
.16486 .03401 L
.1552 .05977 L
.14639 .09272 L
.13832 .1256 L
.13087 .15609 L
.12399 .18429 L
.11759 .21047 L
.11162 .23488 L
.10604 .25772 L
.10079 .27916 L
.09585 .29934 L
.09119 .3184 L
.08677 .33646 L
.08257 .3536 L
.07857 .36992 L
.07476 .38548 L
.07111 .40037 L
.06761 .41464 L
.06425 .42829 L
.06102 .44071 L
.0579 .45025 L
.05489 .45592 L
.05198 .45919 L
.04916 .4622 L
.04641 .46571 L
.04375 .46971 L
.04115 .47403 L
.03861 .47854 L
.03614 .48314 L
.03371 .48774 L
.03134 .49227 L
.02901 .49667 L
.02672 .50089 L
.02447 .5049 L
Mistroke
.02225 .50867 L
.02006 .51216 L
.01789 .51537 L
.01576 .51827 L
.01364 .52086 L
.01154 .52312 L
.00945 .52504 L
.00738 .52662 L
.00532 .52785 L
.00326 .52874 L
.00121 .52927 L
Mfstroke
.00121 .52927 m
0 .52937 L
s
1 .28513 m
.84597 .30291 L
.67538 .33703 L
.56142 .36437 L
.47983 .3841 L
.41846 .39857 L
.37058 .40968 L
.33213 .41851 L
.30055 .4257 L
.27411 .43169 L
.25163 .43676 L
.23226 .44111 L
.21538 .4449 L
.20052 .44823 L
.18732 .45133 L
.17551 .45511 L
.16486 .46138 L
.1552 .47092 L
.14639 .48214 L
.13832 .49313 L
.13087 .5033 L
.12399 .51271 L
.11759 .52144 L
.11162 .52959 L
.10604 .53721 L
.10079 .54436 L
.09585 .5511 L
.09119 .55746 L
.08677 .56348 L
.08257 .5692 L
.07857 .57465 L
.07476 .57984 L
.07111 .58481 L
.06761 .58957 L
.06425 .59411 L
.06102 .59817 L
.0579 .60086 L
.05489 .60136 L
.05198 .59986 L
.04916 .59709 L
.04641 .59358 L
.04375 .58959 L
.04115 .58527 L
.03861 .58075 L
.03614 .57615 L
.03371 .57156 L
.03134 .56703 L
.02901 .56263 L
.02672 .55841 L
.02447 .5544 L
Mistroke
.02225 .55063 L
.02006 .54714 L
.01789 .54393 L
.01576 .54102 L
.01364 .53844 L
.01154 .53618 L
.00945 .53426 L
.00738 .53268 L
.00532 .53145 L
.00326 .53056 L
.00121 .53003 L
Mfstroke
.00121 .53003 m
0 .52993 L
s
.5 g
.01 w
0 .52968 m
.00123 .52983 L
.00236 .53008 L
.00365 .5305 L
.00488 .53102 L
.00706 .53226 L
.00942 .53404 L
.01479 .53962 L
.01975 .54649 L
.04057 .58435 L
.04549 .5925 L
.04821 .5963 L
.05079 .5993 L
.05224 .60067 L
.05382 .60188 L
.05452 .60231 L
.05526 .60269 L
.0566 .60317 L
.0573 .60332 L
.05807 .60264 L
.05945 .60076 L
.06204 .59724 L
.08481 .56632 L
.12786 .50783 L
.14931 .4787 L
.15978 .46447 L
.16218 .46121 L
.1635 .45942 L
.16474 .45774 L
.16579 .45631 L
.16692 .45585 L
.16817 .45557 L
.16932 .45532 L
.17406 .45427 L
.17921 .45314 L
.18857 .45107 L
.20919 .44652 L
.2508 .43734 L
.29081 .42852 L
.3334 .41912 L
.3744 .41008 L
.41797 .40047 L
.45995 .39121 L
.50034 .3823 L
.54331 .37282 L
.58468 .36369 L
.62446 .35492 L
.66682 .34557 L
.70759 .33658 L
.75093 .32702 L
Mistroke
.79267 .31781 L
.83283 .30895 L
.87556 .29601 L
.91671 .28615 L
.96042 .27781 L
1 .27172 L
Mfstroke
0 .52962 m
.00123 .52947 L
.00236 .52922 L
.00365 .5288 L
.00488 .52827 L
.00706 .52703 L
.00942 .52526 L
.01479 .51967 L
.01975 .51281 L
.04057 .47495 L
.04591 .46617 L
.04885 .46218 L
.05166 .45915 L
.05291 .45808 L
.05425 .45715 L
.05502 .45673 L
.05572 .45641 L
.05708 .45602 L
.05831 .45286 L
.05963 .44745 L
.06204 .43765 L
.07288 .39354 L
.08481 .34495 L
.12786 .16966 L
.14931 .08236 L
.15468 .06047 L
.15978 .03973 L
.16199 .0307 L
.16435 .02111 L
.16558 .01611 L
.16693 .01472 L
.16817 .01499 L
.16932 .01524 L
.18857 .01949 L
.20919 .02404 L
.24982 .033 L
.29301 .04253 L
.33462 .05171 L
.37463 .06053 L
.41722 .06993 L
.45822 .07897 L
.50179 .08858 L
.54377 .09784 L
.58416 .10675 L
.62713 .11623 L
.6685 .12536 L
.70828 .13413 L
.75064 .14348 L
.7914 .15247 L
.83474 .16222 L
Mistroke
.87649 .17479 L
.91665 .18439 L
.95938 .19257 L
1 .19884 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@0001/000<10000o`0300006`000`40003o00<0
000K00030@000?l00`0001/000<100009P03JmT0000K00030@0002H026_D00006`000`40000V00e[
c`0001/000<100009P02J`030@0000802f_:00006`020BH00f/00`40000700][a@0001/000<10000
9@02J`030@0000d02f_000006`000`40000U009[00<100004P0<JkX0000K00030@0002@00f/J0003
0F][00M[]`0001/000<100009002J`030@0001d000<1Jf/01V^c00006`000`40000T009[9@08Jjl0
000K00030@0002@00V/Y00U[ZP0001/000<100009002Jb`02f^U00006`020B@00f/a00][X00001/0
00<100008`02JcL036^J00006`000`40000S009[?@09JiL0000K00030@0002<00V]100030F][00E[
U00001/000<100008`02JdD000<1Jf/01F^@00006`000`40000S009[B@09Jh/00008008110000`40
000200<11@000`40000R00=[C00<JhD0000700040@000@/000<100000P020B<00V]B00030F][00U[
P00000L000@100012`000`40000200030@0002800V]F008136]j000000<00@400P4200040@000@P0
0`4500030@0002800V]J00@12F]g00001`001040004800030@0000D000<100008@02Jel010400`1[
J`05Jg@0000700040@000@P000<100001@000`40000Q009[H`040@0306][00E[L00000P00P4900@1
10000`40000Q009[I`040@U[J`0001/000<100008@02Jf/00`4<JfD0000K00818@02Jg0010400`1[
J`09Jf00000K00030@0002000V]d00@10P0;Je/0000K00030@0002000V]h00@10`09JeP0000K0003
0@0001l00f]l00@11008Je@0000K00030@0001l00V^100@11008Je00000K00030@0001l00V^500D1
0`09Jd/0000K00030@0001l00V^:00H12f]600006`000`40000N009[T@050@][@@0001/00P4O009[
UP050@a[>`0001/000<100007P02Ji/01P49JcP0000K00030@0001h00V^Q00D126/d00006`000`40
000M009[Y`040@Q[<00001/000<100007@02Jj`00`47Jbd0000K00030@0001d00V^b00M[:P0001/0
00<100007003JkH01f/40B80000K00817@02JkX02F/70AX0000K00030@0001`00V^m00][0P080A80
000K00030@0001`00V_200][1@080@X0000K00030@0001/00f_700a[1`040@H0000K00030@0001/0
0V_>00][300001/000<100006`02Jm<036/600006`000`40000K009[f@07J`D0000K00030@0001X0
0f/800@1:0000`40000W0081:0020B@0104300818P040@0306/100<10P0001/00P4K009[2@000`40
000Y00030@0002H000@100019P001040004U00050@0000400P000`40000Q00050@0000401@0001/0
00<100006P02J`X000<100009@050BL000@100019P001040004U00050@0000400P000`40000Q0003
0@00008000<100000P0001/000<100006P02J`/000<1000090001040004X00<1:0020BH000D10000
0@0200030@00024000<100000`000`40000100006`000`40000I00=[30000`40000T00030@0102P0
00<100009`001040004U00050@0000400P000`40000Q00030@0000@00@410040000K00030@0001T0
0V/:00040@000BL00P4X00030@0002L000@100018`030@<000@100018@030@<000@100010P0001/0
00<100006@02J`/00P4Y00030@0002L00`4W00819P000`400002008190000`40000200810`0001/0
00<100006@02JnL0000K01/10f_Q0@H0000K00030@0000L000<1000020000`400003009[0P000`40
000800030@0000P000<100001`000`40000800030@0000L000<1000020000`40000700030@0000P0
00<1000020000`40000700030@0000P000<100001`000`40000800030@0000L000<1000020000`40
000800030@0000L000<1000020000`40000700030@0000P000<1000020000`40000400006`000`40
000H009[3@000`40000W00030@0002P000<100009`000`40000W00030@0002P000<10000100001/0
00<100006002JnP0000K00030@0001L00f_X00006`000`40000G009[j@0001/000<100005`02JnT0
000K00030@0001L00V_Y00006`020AL00f_Y00006`000`40000F009[h`000f/0000400006`000`40
000F009[f00=J`D0000K00030@0001H00V_C015[1P0001/000<100005@03Jld036/A00006`000`40
000E009[b`09J`/01P4600006`000`40000E009[b007J`H02P4<00006`020AH00V_400M[2P4F0000
6`000`40000D00=[`@07J`<1800001/000<100005002Jk/00`48JbH0000K00030@0001@00V^f00@1
26/Z00006`000`40000D009[/P030@U[;P0001/000<100004`03Jjd00`4;Jc40000K00030@0001<0
0V^Y00D12F/f00006`000`40000C009[Y0050@0306][00I[>`0001/00P4D009[W`050@8026/o0000
6`000`40000B00=[V`040@802F]300006`000`40000B009[U`050@a[AP0001/000<100004P02Ji80
1@4;Jd`0000K00030@0001800V^>00@12f]A00006`000`40000A00=[RP040@0306][00M[EP0001/0
00<100004@02JhL010400`1[J`06Je/0000K00814P02Jh<010400`1[J`06Jel0000K00030@000140
0V]o00@12F]S00006`000`40000@00=[N`030@][IP0001/000<100004002JgL00`4;Jf/0000K0003
0@0001000V]c00812f]`00006`000`40000@009[K`030@U[M@0000P00P4400030@0000800`450003
0@0000l00f][00<126]j00001`001040004;00030@00008000<100003`02JfP00`48Jgh000070004
0@000@/000<100000P020A000V]T00812F^200001`001040004800<11@000`40000?009[H0000`5[
J`09JhD0000700040@000@P000<100001@000`40000>00=[G00;JhX0000700040@000@P000<10000
1@000`40000>009[E`0<Jhl0000800812@040@@000<100003P02Je@02F^E00006`000`40000>009[
D008JiX0000K00030@0000h00V]<00Q[WP0001/00P4?009[A`09Jj80000K00030@0000d00f]100a[
Y@0001/000<100003@02Jcd02f^[00006`000`40000=009[>00;Jk00000K00030@0000d00V/e00U[
]@0001/000<100003@02Jc4026^j00006`000`40000<00030F][02d026^n00006`000`40000<00=[
:009Jl80000K00813@02Jb@02f_500006`000`40000;00=[7`0;JlX0000K00030@0000/00V/M00U[
c`0001/000<100002P03JaT026_D00006`000`40000900A[6006JmL0000K00030@0000P016/H00A[
fP0001/000<100002003JaT00f_K00006`020@P00f/I00=[g00001/000<100001P03JaX00V_M0000
6`000`400006009[6P03Jmd0000K00030@0000D00f/I00=[gP0001/000<100001@02JaT00f_O0000
6`000`40000400=[6@02Jn00000K00030@0000@00V/I009[h@0001/000<100000`03JaP00f_Q0000
6`020@@00V/H00=[hP0001/000<100000P03JaL00f_S00006`000`400002009[6002Jn@0000K0004
0@0000=[5`02JnD0000K00030@0000=[5`03JnD0000K00030@1[009[5`03JnH0000C00@11000105[
Jf/H009[i`0001D000<100000P04JaP00V_X00005@000`40000200E[5P03JnP0000E00030@0000<0
00<1Jf/00V/D00=[j@0001D000<100000`000`400003Ja<00V_Z00004`030@D000@100000f/A009[
j`0001D000<100000`000`400002009[4003Jn/0000K00030@0000800f/>00=[k00001/00P44009[
3P02Jnd0000K00030@0000<00f/<009[00<10000j`0001/000<100001002J`/00f_^00006`000`40
000400=[2P02Jnl0000K00030@0000D00V/9009[l00001/000<100001@03J`L00f_`00006`000`40
0006009[1`02Jo40000K00030@0000H00f/500=[l@0001/00P48009[1@02Jo80000K00030@0000L0
0f/3009[l`0001/000<100002003J`0406][Jo<0000K00030@0000T01F_d00006`000`40000900A[
m@0001/000<100002P02JoH0000K00030@000?l00`0001/00P7o00@0003o0240003o0240003o0240
003o0240003o0240003o0240003o0240003o0240003o0240003o02400000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.28451, -0.934519, 
  0.0472761, 0.0133946}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .48222 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0512622 1.05126 -0.272769 1.05126 [
[.05441 .48897 -3.24051 0 ]
[.05441 .48897 2.75949 9 ]
[.41403 .37816 -3.0353 0 ]
[.41403 .37816 2.9647 9 ]
[.8076 .25684 -5.60061 0 ]
[.8076 .25684 6.39939 9 ]
[.49008 .42049 -4.9848 0 ]
[.49008 .42049 5.0152 15 ]
[.97195 -0.00321 0 -5.9897 ]
[.97195 -0.00321 6 3.0103 ]
[.97884 .00704 0 -5.98336 ]
[.97884 .00704 18 3.01664 ]
[.98569 .01722 0 -5.97707 ]
[.98569 .01722 18 3.02293 ]
[.99248 .02733 0 -5.97083 ]
[.99248 .02733 18 3.02917 ]
[1.04664 -0.00547 0 -9.96402 ]
[1.04664 -0.00547 34 5.03598 ]
[.99891 .05235 0 -5.95714 ]
[.99891 .05235 24 3.04286 ]
[1.00264 .08734 0 -5.93887 ]
[1.00264 .08734 6 3.06113 ]
[1.00644 .12285 0 -5.92035 ]
[1.00644 .12285 18 3.07965 ]
[1.01029 .15888 0 -5.90157 ]
[1.01029 .15888 6 3.09843 ]
[1.06725 .08606 0 -9.88194 ]
[1.06725 .08606 34 5.11806 ]
[ 0 0 0 0 ]
[ 1 .48222 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03492 .48222 m
1 .18401 L
s
.05546 .47587 m
.05599 .46932 L
s
[(0)] .05441 .48897 .08017 -1 Mshowa
.41419 .36502 m
.41427 .35845 L
s
[(5)] .41403 .37816 .01177 -1 Mshowa
.80673 .24373 m
.80629 .23717 L
s
[(10)] .8076 .25684 -0.06656 -1 Mshowa
.125 Mabswid
.12471 .45447 m
.12498 .45054 L
s
.19518 .4327 m
.19539 .42876 L
s
.26689 .41054 m
.26705 .4066 L
s
.33988 .38798 m
.33999 .38404 L
s
.48984 .34165 m
.48983 .3377 L
s
.56689 .31784 m
.56682 .3139 L
s
.64536 .29359 m
.64522 .28965 L
s
.72529 .26889 m
.72509 .26496 L
s
.88972 .21808 m
.88939 .21415 L
s
.97431 .19195 m
.97391 .18802 L
s
gsave
.49008 .42049 -65.9848 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.95886 0 m
.98445 .03804 L
s
.95948 .00092 m
.95324 .00298 L
s
[(0)] .97195 -0.00321 -1 .33104 Mshowa
.96636 .01115 m
.96012 .01321 L
s
[(0.2)] .97884 .00704 -1 .32963 Mshowa
.9732 .02132 m
.96696 .02337 L
s
[(0.4)] .98569 .01722 -1 .32824 Mshowa
.97999 .03141 m
.97375 .03345 L
s
[(0.6)] .99248 .02733 -1 .32685 Mshowa
.125 Mabswid
.9612 .00348 m
.95746 .00472 L
s
.96293 .00604 m
.95918 .00728 L
s
.96465 .0086 m
.9609 .00984 L
s
.96808 .0137 m
.96433 .01493 L
s
.96979 .01624 m
.96604 .01748 L
s
.97149 .01878 m
.96775 .02001 L
s
.9749 .02385 m
.97116 .02508 L
s
.9766 .02637 m
.97285 .0276 L
s
.9783 .0289 m
.97455 .03012 L
s
.98168 .03392 m
.97793 .03515 L
s
.98337 .03643 m
.97962 .03766 L
s
gsave
1.04664 -0.00547 -61 -13.964 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.98445 .03804 m
1 .18401 L
s
.9864 .0564 m
.98015 .05842 L
s
[(-0.5)] .99891 .05235 -1 .32381 Mshowa
.99013 .09135 m
.98387 .09335 L
s
[(0)] 1.00264 .08734 -1 .31975 Mshowa
.99391 .12681 m
.98764 .12879 L
s
[(0.5)] 1.00644 .12285 -1 .31563 Mshowa
.99774 .16279 m
.99147 .16475 L
s
[(1)] 1.01029 .15888 -1 .31146 Mshowa
.125 Mabswid
.98714 .06335 m
.98339 .06456 L
s
.98789 .07032 m
.98413 .07152 L
s
.98863 .07731 m
.98488 .07851 L
s
.98938 .08432 m
.98562 .08552 L
s
.99088 .0984 m
.98712 .09959 L
s
.99163 .10547 m
.98788 .10666 L
s
.99239 .11256 m
.98863 .11375 L
s
.99315 .11967 m
.98939 .12086 L
s
.99467 .13396 m
.99091 .13515 L
s
.99543 .14114 m
.99167 .14232 L
s
.9962 .14833 m
.99244 .14951 L
s
.99697 .15555 m
.99321 .15673 L
s
.98567 .04947 m
.98192 .05068 L
s
.98493 .04256 m
.98118 .04378 L
s
.99851 .17005 m
.99475 .17122 L
s
.99929 .17734 m
.99552 .1785 L
s
gsave
1.06725 .08606 -61 -13.8819 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.0127 .31334 m
0 .45265 L
s
0 .45265 m
.03492 .48222 L
s
.03492 .48222 m
.04655 .34374 L
s
.04655 .34374 m
.0127 .31334 L
s
.95886 0 m
.98445 .03804 L
s
.98445 .03804 m
1 .18401 L
s
1 .18401 m
.97383 .14677 L
s
.97383 .14677 m
.95886 0 L
s
.0127 .31334 m
0 .45265 L
s
0 .45265 m
.97383 .14677 L
s
.97383 .14677 m
.95886 0 L
s
.95886 0 m
.0127 .31334 L
s
.04655 .34374 m
.98445 .03804 L
s
.98445 .03804 m
1 .18401 L
s
1 .18401 m
.03492 .48222 L
s
.03492 .48222 m
.04655 .34374 L
s
0 0 m
1 0 L
1 .48222 L
0 .48222 L
closepath
clip
newpath
.5 Mabswid
.22671 .28173 m .18221 .30119 L .18568 .31014 L p .752 .67 .85 r
F P
0 g
s
.18568 .31014 m .2302 .29077 L .22671 .28173 L p .752 .67 .85 r
F P
0 g
s
.13119 .36069 m .18568 .31014 L .18221 .30119 L p .88 .733 .799 r
F P
0 g
s
.2302 .29077 m .18568 .31014 L .18848 .31936 L p .738 .632 .817 r
F P
0 g
s
.18568 .31014 m .13119 .36069 L .13304 .36682 L p .862 .703 .78 r
F P
0 g
s
.13304 .36682 m .18848 .31936 L .18568 .31014 L p .862 .703 .78 r
F P
0 g
s
.18848 .31936 m .23303 .30008 L .2302 .29077 L p .738 .632 .817 r
F P
0 g
s
.23303 .30008 m .18848 .31936 L .19053 .32863 L p .722 .592 .782 r
F P
0 g
s
.18848 .31936 m .13304 .36682 L .13438 .37298 L p .844 .669 .758 r
F P
0 g
s
.13438 .37298 m .19053 .32863 L .18848 .31936 L p .844 .669 .758 r
F P
0 g
s
.22264 .27321 m .17815 .29277 L .18221 .30119 L p .765 .707 .881 r
F P
0 g
s
.18221 .30119 m .22671 .28173 L .22264 .27321 L p .765 .707 .881 r
F P
0 g
s
.12887 .35477 m .18221 .30119 L .17815 .29277 L p .896 .762 .816 r
F P
0 g
s
.18221 .30119 m .12887 .35477 L .13119 .36069 L p .88 .733 .799 r
F P
0 g
s
.19053 .32863 m .23512 .30943 L .23303 .30008 L p .722 .592 .782 r
F P
0 g
s
.23512 .30943 m .19053 .32863 L .19178 .3377 L p .703 .547 .743 r
F P
0 g
s
.19053 .32863 m .13438 .37298 L .13518 .37902 L p .822 .631 .731 r
F P
0 g
s
.13518 .37902 m .19178 .3377 L .19053 .32863 L p .822 .631 .731 r
F P
0 g
s
.21811 .26542 m .17361 .28508 L .17815 .29277 L p .777 .746 .912 r
F P
0 g
s
.17815 .29277 m .22264 .27321 L .21811 .26542 L p .777 .746 .912 r
F P
0 g
s
.12615 .3492 m .17815 .29277 L .17361 .28508 L p .912 .791 .831 r
F P
0 g
s
.17815 .29277 m .12615 .3492 L .12887 .35477 L p .896 .762 .816 r
F P
0 g
s
.19178 .3377 m .23642 .31857 L .23512 .30943 L p .703 .547 .743 r
F P
0 g
s
.23642 .31857 m .19178 .3377 L .19219 .34633 L p .68 .494 .695 r
F P
0 g
s
.19178 .3377 m .13518 .37902 L .13541 .38478 L p .795 .584 .697 r
F P
0 g
s
.13541 .38478 m .19219 .34633 L .19178 .3377 L p .795 .584 .697 r
F P
0 g
s
.21322 .25858 m .16871 .27833 L .17361 .28508 L p .787 .787 .944 r
F P
0 g
s
.17361 .28508 m .21811 .26542 L .21322 .25858 L p .787 .787 .944 r
F P
0 g
s
.1231 .34414 m .17361 .28508 L .16871 .27833 L p .929 .819 .844 r
F P
0 g
s
.17361 .28508 m .1231 .34414 L .12615 .3492 L p .912 .791 .831 r
F P
0 g
s
.19219 .34633 m .23689 .32726 L .23642 .31857 L p .68 .494 .695 r
F P
0 g
s
.23689 .32726 m .19219 .34633 L .19175 .35429 L p .648 .429 .634 r
F P
0 g
s
.19219 .34633 m .13541 .38478 L .13506 .39009 L p .76 .523 .65 r
F P
0 g
s
.13506 .39009 m .19175 .35429 L .19219 .34633 L p .76 .523 .65 r
F P
0 g
s
.20811 .25286 m .16357 .2727 L .16871 .27833 L p .797 .832 .978 r
F P
0 g
s
.16871 .27833 m .21322 .25858 L .20811 .25286 L p .797 .832 .978 r
F P
0 g
s
.11981 .33972 m .16871 .27833 L .16357 .2727 L p .946 .85 .857 r
F P
0 g
s
.16871 .27833 m .11981 .33972 L .1231 .34414 L p .929 .819 .844 r
F P
0 g
s
.19175 .35429 m .23652 .33527 L .23689 .32726 L p .648 .429 .634 r
F P
0 g
s
.23652 .33527 m .19175 .35429 L .19047 .36137 L p .603 .342 .551 r
F P
0 g
s
.19175 .35429 m .13506 .39009 L .13413 .39483 L p .71 .438 .581 r
F P
0 g
s
.13413 .39483 m .19047 .36137 L .19175 .35429 L p .71 .438 .581 r
F P
0 g
s
.27017 .26823 m .22671 .28173 L .2302 .29077 L p .713 .649 .859 r
F P
0 g
s
.2302 .29077 m .27348 .27683 L .27017 .26823 L p .713 .649 .859 r
F P
0 g
s
.27348 .27683 m .2302 .29077 L .23303 .30008 L p .702 .61 .823 r
F P
0 g
s
.23303 .30008 m .27617 .28569 L .27348 .27683 L p .702 .61 .823 r
F P
0 g
s
.27617 .28569 m .23303 .30008 L .23512 .30943 L p .688 .57 .785 r
F P
0 g
s
.26633 .26012 m .22264 .27321 L .22671 .28173 L p .722 .687 .893 r
F P
0 g
s
.22671 .28173 m .27017 .26823 L .26633 .26012 L p .722 .687 .893 r
F P
0 g
s
.23512 .30943 m .27818 .29457 L .27617 .28569 L p .688 .57 .785 r
F P
0 g
s
.27818 .29457 m .23512 .30943 L .23642 .31857 L p .672 .525 .743 r
F P
0 g
s
.16357 .2727 m .11635 .33607 L .11981 .33972 L p .946 .85 .857 r
F P
0 g
s
.26207 .2527 m .21811 .26542 L .22264 .27321 L p .729 .726 .928 r
F P
0 g
s
.22264 .27321 m .26633 .26012 L .26207 .2527 L p .729 .726 .928 r
F P
0 g
s
.23642 .31857 m .27945 .30325 L .27818 .29457 L p .672 .525 .743 r
F P
0 g
s
.27945 .30325 m .23642 .31857 L .23689 .32726 L p .65 .472 .693 r
F P
0 g
s
.07682 .40848 m .13119 .36069 L .12887 .35477 L p .872 .73 .803 r
F P
0 g
s
.13119 .36069 m .07682 .40848 L .07778 .41166 L p .855 .699 .783 r
F P
0 g
s
.07778 .41166 m .13304 .36682 L .13119 .36069 L p .855 .699 .783 r
F P
0 g
s
.13304 .36682 m .07778 .41166 L .07846 .41486 L p .837 .665 .76 r
F P
0 g
s
.07846 .41486 m .13438 .37298 L .13304 .36682 L p .837 .665 .76 r
F P
0 g
s
.19047 .36137 m .23531 .34238 L .23652 .33527 L p .603 .342 .551 r
F P
0 g
s
.23531 .34238 m .19047 .36137 L .18837 .36737 L p .532 .217 .429 r
F P
0 g
s
.19047 .36137 m .13413 .39483 L .13267 .39885 L p .627 .308 .469 r
F P
0 g
s
.13267 .39885 m .18837 .36737 L .19047 .36137 L p .627 .308 .469 r
F P
0 g
s
.07561 .40542 m .12887 .35477 L .12615 .3492 L p .889 .76 .821 r
F P
0 g
s
.12887 .35477 m .07561 .40542 L .07682 .40848 L p .872 .73 .803 r
F P
0 g
s
.13438 .37298 m .07846 .41486 L .07886 .418 L p .815 .626 .732 r
F P
0 g
s
.07886 .418 m .13518 .37902 L .13438 .37298 L p .815 .626 .732 r
F P
0 g
s
.25748 .24617 m .21322 .25858 L .21811 .26542 L p .735 .768 .965 r
F P
0 g
s
.21811 .26542 m .26207 .2527 L .25748 .24617 L p .735 .768 .965 r
F P
0 g
s
.07419 .40255 m .12615 .3492 L .1231 .34414 L p .905 .789 .837 r
F P
0 g
s
.12615 .3492 m .07419 .40255 L .07561 .40542 L p .889 .76 .821 r
F P
0 g
s
.13518 .37902 m .07886 .418 L .07895 .42099 L p .789 .579 .697 r
F P
0 g
s
.07895 .42099 m .13541 .38478 L .13518 .37902 L p .789 .579 .697 r
F P
0 g
s
.23689 .32726 m .27995 .31149 L .27945 .30325 L p .65 .472 .693 r
F P
0 g
s
.27995 .31149 m .23689 .32726 L .23652 .33527 L p .621 .407 .629 r
F P
0 g
s
.07259 .39995 m .1231 .34414 L .11981 .33972 L p .922 .819 .852 r
F P
0 g
s
.1231 .34414 m .07259 .39995 L .07419 .40255 L p .905 .789 .837 r
F P
0 g
s
.13541 .38478 m .07895 .42099 L .07874 .42376 L p .755 .518 .649 r
F P
0 g
s
.07874 .42376 m .13506 .39009 L .13541 .38478 L p .755 .518 .649 r
F P
0 g
s
.2527 .24072 m .20811 .25286 L .21322 .25858 L p .736 .814 1 r
F P
0 g
s
.21322 .25858 m .25748 .24617 L .2527 .24072 L p .736 .814 1 r
F P
0 g
s
.18837 .36737 m .23328 .34841 L .23531 .34238 L p .532 .217 .429 r
F P
0 g
s
.23328 .34841 m .18837 .36737 L .1855 .37214 L p .411 .1 .234 r
F P
0 g
s
.18837 .36737 m .13267 .39885 L .13069 .40205 L p .473 .1 .269 r
F P
0 g
s
.13069 .40205 m .1855 .37214 L .18837 .36737 L p .473 .1 .269 r
F P
0 g
s
.07086 .39769 m .11981 .33972 L .11635 .33607 L p .939 .85 .866 r
F P
0 g
s
.11981 .33972 m .07086 .39769 L .07259 .39995 L p .922 .819 .852 r
F P
0 g
s
.23652 .33527 m .27966 .31908 L .27995 .31149 L p .621 .407 .629 r
F P
0 g
s
.27966 .31908 m .23652 .33527 L .23531 .34238 L p .579 .32 .544 r
F P
0 g
s
.13506 .39009 m .07874 .42376 L .07823 .42623 L p .706 .435 .58 r
F P
0 g
s
.07823 .42623 m .13413 .39483 L .13506 .39009 L p .706 .435 .58 r
F P
0 g
s
.11635 .33607 m .06904 .39583 L .07086 .39769 L p .939 .85 .866 r
F P
0 g
s
.13413 .39483 m .07823 .42623 L .07743 .42833 L p .627 .308 .469 r
F P
0 g
s
.07743 .42833 m .13267 .39885 L .13413 .39483 L p .627 .308 .469 r
F P
0 g
s
.1855 .37214 m .23049 .35318 L .23328 .34841 L p .411 .1 .234 r
F P
0 g
s
.23049 .35318 m .1855 .37214 L .18195 .37553 L p .191 .1 .1 r
F P
0 g
s
.1855 .37214 m .13069 .40205 L .12826 .40434 L p .157 .1 .1 r
F P
0 g
s
.12826 .40434 m .18195 .37553 L p .17231 .37956 L .157 .1 .1 r
F P
0 g
s
.07507 .43121 m .12826 .40434 L .13069 .40205 L p .189 .1 .1 r
F P
0 g
s
.12826 .40434 m .12543 .40564 L p .17325 .38016 L .515 .943 .799 r
F P
0 g
s
.18195 .37553 m .17273 .38047 L p .17325 .38016 L .515 .943 .799 r
F P
0 g
s
.17273 .38047 m .12826 .40434 L p .17325 .38016 L .515 .943 .799 r
F P
0 g
s
.18195 .37553 m .1855 .37214 L p .17231 .37956 L .157 .1 .1 r
F P
0 g
s
.1778 .37744 m .17315 .3778 L p .18464 .37333 L .628 1 1 r
F P
0 g
s
.12543 .40564 m .1778 .37744 L .18195 .37553 L p .515 .943 .799 r
F P
0 g
s
.18195 .37553 m .1778 .37744 L p .20591 .36551 L .1 .1 .1 r
F P
0 g
s
.22701 .35655 m .18195 .37553 L p .20591 .36551 L .1 .1 .1 r
F P
0 g
s
.18195 .37553 m .22701 .35655 L .23049 .35318 L p .191 .1 .1 r
F P
0 g
s
.22293 .35843 m .22701 .35655 L p .20223 .36707 L .1 .1 .1 r
F P
0 g
s
.1778 .37744 m .22293 .35843 L p .20223 .36707 L .1 .1 .1 r
F P
0 g
s
.27419 .336 m .23049 .35318 L .22701 .35655 L p .197 .1 .1 r
F P
0 g
s
.22701 .35655 m .27099 .33917 L .27419 .336 L p .197 .1 .1 r
F P
0 g
s
.27099 .33917 m .22701 .35655 L .22293 .35843 L p .1 .1 .1 r
F P
0 g
s
.22293 .35843 m .21835 .35876 L p .24886 .34762 L .59 1 .998 r
F P
0 g
s
.22293 .35843 m .26721 .34092 L .27099 .33917 L p .1 .1 .1 r
F P
0 g
s
.26721 .34092 m .22293 .35843 L p .24886 .34762 L .59 1 .998 r
F P
0 g
s
.22293 .35843 m .1778 .37744 L p .18464 .37333 L .628 1 1 r
F P
0 g
s
.23531 .34238 m .27858 .32581 L .27966 .31908 L p .579 .32 .544 r
F P
0 g
s
.27858 .32581 m .23531 .34238 L .23328 .34841 L p .512 .198 .421 r
F P
0 g
s
.31418 .25456 m .27017 .26823 L .27348 .27683 L p .713 .649 .859 r
F P
0 g
s
.27348 .27683 m .3173 .26271 L .31418 .25456 L p .713 .649 .859 r
F P
0 g
s
.3173 .26271 m .27348 .27683 L .27617 .28569 L p .702 .61 .823 r
F P
0 g
s
.07561 .40542 m .02399 .42642 L .02399 .42642 L p .809 .7 .832 r
F P
0 g
s
.02399 .42642 m .07682 .40848 L .07561 .40542 L p .809 .7 .832 r
F P
0 g
s
.07682 .40848 m .02399 .42642 L .02399 .42642 L p .8 .668 .803 r
F P
0 g
s
.02399 .42642 m .07778 .41166 L .07682 .40848 L p .8 .668 .803 r
F P
0 g
s
.07778 .41166 m .02399 .42642 L .02399 .42642 L p .79 .635 .771 r
F P
0 g
s
.07419 .40255 m .02399 .42642 L .02399 .42642 L p .82 .733 .86 r
F P
0 g
s
.02399 .42642 m .07561 .40542 L .07419 .40255 L p .82 .733 .86 r
F P
0 g
s
.02399 .42642 m .07846 .41486 L .07778 .41166 L p .79 .635 .771 r
F P
0 g
s
.07846 .41486 m .02399 .42642 L .02399 .42643 L p .781 .601 .737 r
F P
0 g
s
.07259 .39995 m .02399 .42642 L .02399 .42642 L p .831 .768 .888 r
F P
0 g
s
.02399 .42642 m .07419 .40255 L .07259 .39995 L p .831 .768 .888 r
F P
0 g
s
.02399 .42643 m .07886 .418 L .07846 .41486 L p .781 .601 .737 r
F P
0 g
s
.07886 .418 m .02399 .42643 L .02399 .42643 L p .771 .565 .697 r
F P
0 g
s
.07086 .39769 m .02399 .42642 L .02399 .42642 L p .844 .806 .916 r
F P
0 g
s
.02399 .42642 m .07259 .39995 L .07086 .39769 L p .844 .806 .916 r
F P
0 g
s
.02399 .42643 m .07895 .42099 L .07886 .418 L p .771 .565 .697 r
F P
0 g
s
.07895 .42099 m .02399 .42643 L .02399 .42643 L p .76 .523 .65 r
F P
0 g
s
.06904 .39583 m .02399 .42642 L .02399 .42642 L p .858 .849 .944 r
F P
0 g
s
.02399 .42642 m .07086 .39769 L .06904 .39583 L p .858 .849 .944 r
F P
0 g
s
.02399 .42643 m .07874 .42376 L .07895 .42099 L p .76 .523 .65 r
F P
0 g
s
.07874 .42376 m .02399 .42643 L .02399 .42643 L p .747 .473 .59 r
F P
0 g
s
.02399 .42643 m .07823 .42623 L .07874 .42376 L p .747 .473 .59 r
F P
0 g
s
.07823 .42623 m .02399 .42643 L .02399 .42643 L p .728 .41 .513 r
F P
0 g
s
.02399 .42643 m .07743 .42833 L .07823 .42623 L p .728 .41 .513 r
F P
0 g
s
.07743 .42833 m .02399 .42643 L .02399 .42643 L p .7 .326 .404 r
F P
0 g
s
.02399 .42643 m .07637 .43001 L .07743 .42833 L p .7 .326 .404 r
F P
0 g
s
.07637 .43001 m .02399 .42643 L .02399 .42643 L p .65 .204 .244 r
F P
0 g
s
.02399 .42643 m .07507 .43121 L .07637 .43001 L p .65 .204 .244 r
F P
0 g
s
.07507 .43121 m .02399 .42643 L .02399 .42643 L p .554 .1 .1 r
F P
0 g
s
.02399 .42643 m .07356 .43191 L .07507 .43121 L p .554 .1 .1 r
F P
0 g
s
.07356 .43191 m .02399 .42643 L .02399 .42643 L p .368 .1 .1 r
F P
0 g
s
.02399 .42643 m .0719 .43208 L .07356 .43191 L p .368 .1 .1 r
F P
0 g
s
.0719 .43208 m .02399 .42643 L .02399 .42643 L p .101 .667 .89 r
F P
0 g
s
.02399 .42643 m .07011 .43172 L .0719 .43208 L p .101 .667 .89 r
F P
0 g
s
.07011 .43172 m .02399 .42643 L .02399 .42643 L p .334 .776 1 r
F P
0 g
s
.02399 .42643 m .06826 .43083 L .07011 .43172 L p .334 .776 1 r
F P
0 g
s
.06826 .43083 m .02399 .42643 L .02399 .42643 L p .47 .781 1 r
F P
0 g
s
.02399 .42643 m .0664 .42944 L .06826 .43083 L p .47 .781 1 r
F P
0 g
s
.27617 .28569 m .31986 .2711 L .3173 .26271 L p .702 .61 .823 r
F P
0 g
s
.31986 .2711 m .27617 .28569 L .27818 .29457 L p .688 .57 .785 r
F P
0 g
s
.31057 .24686 m .26633 .26012 L .27017 .26823 L p .722 .687 .893 r
F P
0 g
s
.27017 .26823 m .31418 .25456 L .31057 .24686 L p .722 .687 .893 r
F P
0 g
s
.27818 .29457 m .32179 .27952 L .31986 .2711 L p .688 .57 .785 r
F P
0 g
s
.32179 .27952 m .27818 .29457 L .27945 .30325 L p .672 .525 .743 r
F P
0 g
s
.13267 .39885 m .07743 .42833 L .07637 .43001 L p .482 .1 .274 r
F P
0 g
s
.07637 .43001 m .13069 .40205 L .13267 .39885 L p .482 .1 .274 r
F P
0 g
s
.30658 .23982 m .26207 .2527 L .26633 .26012 L p .729 .726 .928 r
F P
0 g
s
.26633 .26012 m .31057 .24686 L .30658 .23982 L p .729 .726 .928 r
F P
0 g
s
.27945 .30325 m .32302 .28773 L .32179 .27952 L p .672 .525 .743 r
F P
0 g
s
.32302 .28773 m .27945 .30325 L .27995 .31149 L p .65 .472 .693 r
F P
0 g
s
.23328 .34841 m .27675 .33151 L .27858 .32581 L p .512 .198 .421 r
F P
0 g
s
.27675 .33151 m .23328 .34841 L .23049 .35318 L p .399 .1 .227 r
F P
0 g
s
.30229 .23362 m .25748 .24617 L .26207 .2527 L p .735 .768 .965 r
F P
0 g
s
.26207 .2527 m .30658 .23982 L .30229 .23362 L p .735 .768 .965 r
F P
0 g
s
.13069 .40205 m .07637 .43001 L .07507 .43121 L p .189 .1 .1 r
F P
0 g
s
.27995 .31149 m .32355 .29552 L .32302 .28773 L p .65 .472 .693 r
F P
0 g
s
.32355 .29552 m .27995 .31149 L .27966 .31908 L p .621 .407 .629 r
F P
0 g
s
.29783 .22843 m .2527 .24072 L .25748 .24617 L p .736 .814 1 r
F P
0 g
s
.25748 .24617 m .30229 .23362 L .29783 .22843 L p .736 .814 1 r
F P
0 g
s
.12826 .40434 m .07507 .43121 L .07356 .43191 L p .466 .913 .776 r
F P
0 g
s
.07356 .43191 m .12543 .40564 L .12826 .40434 L p .466 .913 .776 r
F P
0 g
s
.27966 .31908 m .32333 .30269 L .32355 .29552 L p .621 .407 .629 r
F P
0 g
s
.32333 .30269 m .27966 .31908 L .27858 .32581 L p .579 .32 .544 r
F P
0 g
s
.23049 .35318 m .27419 .336 L .27675 .33151 L p .399 .1 .227 r
F P
0 g
s
.1778 .37744 m .12543 .40564 L .12228 .40593 L p .798 1 1 r
F P
0 g
s
.12228 .40593 m .17315 .3778 L .1778 .37744 L p .798 1 1 r
F P
0 g
s
.12543 .40564 m .07356 .43191 L .0719 .43208 L p .766 1 1 r
F P
0 g
s
.0719 .43208 m .12228 .40593 L .12543 .40564 L p .766 1 1 r
F P
0 g
s
.27858 .32581 m .32238 .30904 L .32333 .30269 L p .579 .32 .544 r
F P
0 g
s
.32238 .30904 m .27858 .32581 L .27675 .33151 L p .512 .198 .421 r
F P
0 g
s
.35875 .24071 m .31418 .25456 L .3173 .26271 L p .713 .649 .859 r
F P
0 g
s
.3173 .26271 m .36169 .24842 L .35875 .24071 L p .713 .649 .859 r
F P
0 g
s
.36169 .24842 m .3173 .26271 L .31986 .2711 L p .702 .61 .823 r
F P
0 g
s
.31986 .2711 m .36411 .25634 L .36169 .24842 L p .702 .61 .823 r
F P
0 g
s
.36411 .25634 m .31986 .2711 L .32179 .27952 L p .688 .57 .785 r
F P
0 g
s
.12228 .40593 m .0719 .43208 L .07011 .43172 L p .856 1 1 r
F P
0 g
s
.35538 .23344 m .31057 .24686 L .31418 .25456 L p .722 .687 .893 r
F P
0 g
s
.31418 .25456 m .35875 .24071 L .35538 .23344 L p .722 .687 .893 r
F P
0 g
s
.32179 .27952 m .36594 .26427 L .36411 .25634 L p .688 .57 .785 r
F P
0 g
s
.36594 .26427 m .32179 .27952 L .32302 .28773 L p .672 .525 .743 r
F P
0 g
s
.17315 .3778 m .21835 .35876 L .22293 .35843 L p .628 1 1 r
F P
0 g
s
.21835 .35876 m .17315 .3778 L .16814 .37661 L p .754 1 1 r
F P
0 g
s
.17315 .3778 m .12228 .40593 L .11889 .40518 L p .874 1 1 r
F P
0 g
s
.11889 .40518 m .16814 .37661 L .17315 .3778 L p .874 1 1 r
F P
0 g
s
.07011 .43172 m .11889 .40518 L .12228 .40593 L p .856 1 1 r
F P
0 g
s
.35164 .22678 m .30658 .23982 L .31057 .24686 L p .729 .726 .928 r
F P
0 g
s
.31057 .24686 m .35538 .23344 L .35164 .22678 L p .729 .726 .928 r
F P
0 g
s
.11889 .40518 m .07011 .43172 L .06826 .43083 L p .87 .961 1 r
F P
0 g
s
.32302 .28773 m .36714 .27201 L .36594 .26427 L p .672 .525 .743 r
F P
0 g
s
.36714 .27201 m .32302 .28773 L .32355 .29552 L p .65 .472 .693 r
F P
0 g
s
.27675 .33151 m .32073 .31441 L .32238 .30904 L p .512 .198 .421 r
F P
0 g
s
.32073 .31441 m .27675 .33151 L .27419 .336 L p .399 .1 .227 r
F P
0 g
s
.06826 .43083 m .11537 .40342 L .11889 .40518 L p .87 .961 1 r
F P
0 g
s
.11537 .40342 m .06826 .43083 L .0664 .42944 L p .862 .9 .984 r
F P
0 g
s
.34765 .22091 m .30229 .23362 L .30658 .23982 L p .735 .768 .965 r
F P
0 g
s
.30658 .23982 m .35164 .22678 L .34765 .22091 L p .735 .768 .965 r
F P
0 g
s
.16814 .37661 m .11889 .40518 L .11537 .40342 L p .883 .96 1 r
F P
0 g
s
.32355 .29552 m .36768 .27936 L .36714 .27201 L p .65 .472 .693 r
F P
0 g
s
.36768 .27936 m .32355 .29552 L .32333 .30269 L p .621 .407 .629 r
F P
0 g
s
.0664 .42944 m .1118 .40069 L .11537 .40342 L p .862 .9 .984 r
F P
0 g
s
.16814 .37661 m .2134 .3575 L .21835 .35876 L p .754 1 1 r
F P
0 g
s
.2134 .3575 m .16814 .37661 L .1629 .37387 L p .791 .955 1 r
F P
0 g
s
.11537 .40342 m .1629 .37387 L .16814 .37661 L p .883 .96 1 r
F P
0 g
s
.34351 .21599 m .29783 .22843 L .30229 .23362 L p .736 .814 1 r
F P
0 g
s
.30229 .23362 m .34765 .22091 L .34351 .21599 L p .736 .814 1 r
F P
0 g
s
.27419 .336 m .3184 .31863 L .32073 .31441 L p .399 .1 .227 r
F P
0 g
s
.3184 .31863 m .27419 .336 L .27099 .33917 L p .197 .1 .1 r
F P
0 g
s
.32333 .30269 m .36754 .2861 L .36768 .27936 L p .621 .407 .629 r
F P
0 g
s
.36754 .2861 m .32333 .30269 L .32238 .30904 L p .579 .32 .544 r
F P
0 g
s
.1629 .37387 m .11537 .40342 L .1118 .40069 L p .873 .901 .975 r
F P
0 g
s
.1118 .40069 m .15757 .36965 L .1629 .37387 L p .873 .901 .975 r
F P
0 g
s
.21835 .35876 m .26297 .34118 L .26721 .34092 L p .59 1 .998 r
F P
0 g
s
.26297 .34118 m .21835 .35876 L .2134 .3575 L p .723 1 1 r
F P
0 g
s
.1629 .37387 m .20821 .35469 L .2134 .3575 L p .791 .955 1 r
F P
0 g
s
.20821 .35469 m .1629 .37387 L .15757 .36965 L p .796 .892 1 r
F P
0 g
s
.32238 .30904 m .36672 .29207 L .36754 .2861 L p .579 .32 .544 r
F P
0 g
s
.36672 .29207 m .32238 .30904 L .32073 .31441 L p .512 .198 .421 r
F P
0 g
s
.27099 .33917 m .31546 .32159 L .3184 .31863 L p .197 .1 .1 r
F P
0 g
s
.31546 .32159 m .27099 .33917 L .26721 .34092 L p .1 .1 .1 r
F P
0 g
s
.4039 .22669 m .35875 .24071 L .36169 .24842 L p .713 .649 .859 r
F P
0 g
s
.36169 .24842 m .40665 .23393 L .4039 .22669 L p .713 .649 .859 r
F P
0 g
s
.40665 .23393 m .36169 .24842 L .36411 .25634 L p .702 .61 .823 r
F P
0 g
s
.36411 .25634 m .40892 .24138 L .40665 .23393 L p .702 .61 .823 r
F P
0 g
s
.40892 .24138 m .36411 .25634 L .36594 .26427 L p .688 .57 .785 r
F P
0 g
s
.40075 .21984 m .35538 .23344 L .35875 .24071 L p .722 .687 .893 r
F P
0 g
s
.35875 .24071 m .4039 .22669 L .40075 .21984 L p .722 .687 .893 r
F P
0 g
s
.36594 .26427 m .41066 .24884 L .40892 .24138 L p .688 .57 .785 r
F P
0 g
s
.41066 .24884 m .36594 .26427 L .36714 .27201 L p .672 .525 .743 r
F P
0 g
s
.39728 .21357 m .35164 .22678 L .35538 .23344 L p .729 .726 .928 r
F P
0 g
s
.35538 .23344 m .40075 .21984 L .39728 .21357 L p .729 .726 .928 r
F P
0 g
s
.32073 .31441 m .36523 .2971 L .36672 .29207 L p .512 .198 .421 r
F P
0 g
s
.36523 .2971 m .32073 .31441 L .3184 .31863 L p .399 .1 .227 r
F P
0 g
s
.15757 .36965 m .20292 .35039 L .20821 .35469 L p .796 .892 1 r
F P
0 g
s
.36714 .27201 m .41182 .2561 L .41066 .24884 L p .672 .525 .743 r
F P
0 g
s
.41182 .2561 m .36714 .27201 L .36768 .27936 L p .65 .472 .693 r
F P
0 g
s
.2134 .3575 m .25836 .33995 L .26297 .34118 L p .723 1 1 r
F P
0 g
s
.25836 .33995 m .2134 .3575 L .20821 .35469 L p .768 .951 1 r
F P
0 g
s
.39358 .20804 m .34765 .22091 L .35164 .22678 L p .735 .768 .965 r
F P
0 g
s
.35164 .22678 m .39728 .21357 L .39358 .20804 L p .735 .768 .965 r
F P
0 g
s
.26721 .34092 m .31198 .32321 L .31546 .32159 L p .1 .1 .1 r
F P
0 g
s
.31198 .32321 m .26721 .34092 L .26297 .34118 L p .59 1 .998 r
F P
0 g
s
.36768 .27936 m .41237 .26299 L .41182 .2561 L p .65 .472 .693 r
F P
0 g
s
.41237 .26299 m .36768 .27936 L .36754 .2861 L p .621 .407 .629 r
F P
0 g
s
.38975 .2034 m .34351 .21599 L .34765 .22091 L p .736 .814 1 r
F P
0 g
s
.34765 .22091 m .39358 .20804 L .38975 .2034 L p .736 .814 1 r
F P
0 g
s
.3184 .31863 m .36312 .30105 L .36523 .2971 L p .399 .1 .227 r
F P
0 g
s
.36312 .30105 m .3184 .31863 L .31546 .32159 L p .197 .1 .1 r
F P
0 g
s
.36754 .2861 m .4123 .26931 L .41237 .26299 L p .621 .407 .629 r
F P
0 g
s
.4123 .26931 m .36754 .2861 L .36672 .29207 L p .579 .32 .544 r
F P
0 g
s
.20821 .35469 m .25353 .33724 L .25836 .33995 L p .768 .951 1 r
F P
0 g
s
.25353 .33724 m .20821 .35469 L .20292 .35039 L p .778 .888 1 r
F P
0 g
s
.26297 .34118 m .30806 .32342 L .31198 .32321 L p .59 1 .998 r
F P
0 g
s
.30806 .32342 m .26297 .34118 L .25836 .33995 L p .723 1 1 r
F P
0 g
s
.36672 .29207 m .41159 .27489 L .4123 .26931 L p .579 .32 .544 r
F P
0 g
s
.41159 .27489 m .36672 .29207 L .36523 .2971 L p .512 .198 .421 r
F P
0 g
s
.31546 .32159 m .36044 .30381 L .36312 .30105 L p .197 .1 .1 r
F P
0 g
s
.36044 .30381 m .31546 .32159 L .31198 .32321 L p .1 .1 .1 r
F P
0 g
s
.20292 .35039 m .2486 .33313 L .25353 .33724 L p .778 .888 1 r
F P
0 g
s
.44962 .21249 m .4039 .22669 L .40665 .23393 L p .713 .649 .859 r
F P
0 g
s
.40665 .23393 m .45218 .21927 L .44962 .21249 L p .713 .649 .859 r
F P
0 g
s
.45218 .21927 m .40665 .23393 L .40892 .24138 L p .702 .61 .823 r
F P
0 g
s
.40892 .24138 m .45431 .22623 L .45218 .21927 L p .702 .61 .823 r
F P
0 g
s
.45431 .22623 m .40892 .24138 L .41066 .24884 L p .688 .57 .785 r
F P
0 g
s
.25836 .33995 m .3038 .32221 L .30806 .32342 L p .723 1 1 r
F P
0 g
s
.3038 .32221 m .25836 .33995 L .25353 .33724 L p .768 .951 1 r
F P
0 g
s
.4467 .20607 m .40075 .21984 L .4039 .22669 L p .722 .687 .893 r
F P
0 g
s
.4039 .22669 m .44962 .21249 L .4467 .20607 L p .722 .687 .893 r
F P
0 g
s
.41066 .24884 m .45596 .2332 L .45431 .22623 L p .688 .57 .785 r
F P
0 g
s
.45596 .2332 m .41066 .24884 L .41182 .2561 L p .672 .525 .743 r
F P
0 g
s
.36523 .2971 m .41027 .27959 L .41159 .27489 L p .512 .198 .421 r
F P
0 g
s
.41027 .27959 m .36523 .2971 L .36312 .30105 L p .399 .1 .227 r
F P
0 g
s
.44349 .2002 m .39728 .21357 L .40075 .21984 L p .729 .726 .928 r
F P
0 g
s
.40075 .21984 m .4467 .20607 L .44349 .2002 L p .729 .726 .928 r
F P
0 g
s
.41182 .2561 m .45708 .23999 L .45596 .2332 L p .672 .525 .743 r
F P
0 g
s
.45708 .23999 m .41182 .2561 L .41237 .26299 L p .65 .472 .693 r
F P
0 g
s
.31198 .32321 m .35726 .3053 L .36044 .30381 L p .1 .1 .1 r
F P
0 g
s
.35726 .3053 m .31198 .32321 L .30806 .32342 L p .59 1 .998 r
F P
0 g
s
.44009 .19501 m .39358 .20804 L .39728 .21357 L p .735 .768 .965 r
F P
0 g
s
.39728 .21357 m .44349 .2002 L .44009 .19501 L p .735 .768 .965 r
F P
0 g
s
.41237 .26299 m .45763 .24641 L .45708 .23999 L p .65 .472 .693 r
F P
0 g
s
.45763 .24641 m .41237 .26299 L .4123 .26931 L p .621 .407 .629 r
F P
0 g
s
.25353 .33724 m .29932 .31961 L .3038 .32221 L p .768 .951 1 r
F P
0 g
s
.29932 .31961 m .25353 .33724 L .2486 .33313 L p .778 .888 1 r
F P
0 g
s
.36312 .30105 m .40837 .28327 L .41027 .27959 L p .399 .1 .227 r
F P
0 g
s
.40837 .28327 m .36312 .30105 L .36044 .30381 L p .197 .1 .1 r
F P
0 g
s
.43657 .19065 m .38975 .2034 L .39358 .20804 L p .736 .814 1 r
F P
0 g
s
.39358 .20804 m .44009 .19501 L .43657 .19065 L p .736 .814 1 r
F P
0 g
s
.30806 .32342 m .35366 .30546 L .35726 .3053 L p .59 1 .998 r
F P
0 g
s
.35366 .30546 m .30806 .32342 L .3038 .32221 L p .723 1 1 r
F P
0 g
s
.4123 .26931 m .45761 .2523 L .45763 .24641 L p .621 .407 .629 r
F P
0 g
s
.45761 .2523 m .4123 .26931 L .41159 .27489 L p .579 .32 .544 r
F P
0 g
s
.2486 .33313 m .29474 .31569 L .29932 .31961 L p .778 .888 1 r
F P
0 g
s
.36044 .30381 m .40594 .28583 L .40837 .28327 L p .197 .1 .1 r
F P
0 g
s
.40594 .28583 m .36044 .30381 L .35726 .3053 L p .1 .1 .1 r
F P
0 g
s
.41159 .27489 m .45701 .2575 L .45761 .2523 L p .579 .32 .544 r
F P
0 g
s
.45701 .2575 m .41159 .27489 L .41027 .27959 L p .512 .198 .421 r
F P
0 g
s
.3038 .32221 m .34973 .30428 L .35366 .30546 L p .723 1 1 r
F P
0 g
s
.34973 .30428 m .3038 .32221 L .29932 .31961 L p .768 .951 1 r
F P
0 g
s
.49594 .1981 m .44962 .21249 L .45218 .21927 L p .713 .649 .859 r
F P
0 g
s
.45218 .21927 m .49831 .20441 L .49594 .1981 L p .713 .649 .859 r
F P
0 g
s
.49831 .20441 m .45218 .21927 L .45431 .22623 L p .702 .61 .823 r
F P
0 g
s
.45431 .22623 m .5003 .21089 L .49831 .20441 L p .702 .61 .823 r
F P
0 g
s
.5003 .21089 m .45431 .22623 L .45596 .2332 L p .688 .57 .785 r
F P
0 g
s
.41027 .27959 m .45585 .26186 L .45701 .2575 L p .512 .198 .421 r
F P
0 g
s
.45585 .26186 m .41027 .27959 L .40837 .28327 L p .399 .1 .227 r
F P
0 g
s
.49325 .19213 m .4467 .20607 L .44962 .21249 L p .722 .687 .893 r
F P
0 g
s
.44962 .21249 m .49594 .1981 L .49325 .19213 L p .722 .687 .893 r
F P
0 g
s
.35726 .3053 m .40304 .28718 L .40594 .28583 L p .1 .1 .1 r
F P
0 g
s
.40304 .28718 m .35726 .3053 L .35366 .30546 L p .59 1 .998 r
F P
0 g
s
.45596 .2332 m .50184 .21736 L .5003 .21089 L p .688 .57 .785 r
F P
0 g
s
.50184 .21736 m .45596 .2332 L .45708 .23999 L p .672 .525 .743 r
F P
0 g
s
.4903 .18665 m .44349 .2002 L .4467 .20607 L p .729 .726 .928 r
F P
0 g
s
.4467 .20607 m .49325 .19213 L .4903 .18665 L p .729 .726 .928 r
F P
0 g
s
.45708 .23999 m .50291 .22366 L .50184 .21736 L p .672 .525 .743 r
F P
0 g
s
.50291 .22366 m .45708 .23999 L .45763 .24641 L p .65 .472 .693 r
F P
0 g
s
.29932 .31961 m .34559 .3018 L .34973 .30428 L p .768 .951 1 r
F P
0 g
s
.34559 .3018 m .29932 .31961 L .29474 .31569 L p .778 .888 1 r
F P
0 g
s
.48718 .18181 m .44009 .19501 L .44349 .2002 L p .735 .768 .965 r
F P
0 g
s
.44349 .2002 m .4903 .18665 L .48718 .18181 L p .735 .768 .965 r
F P
0 g
s
.40837 .28327 m .45416 .26527 L .45585 .26186 L p .399 .1 .227 r
F P
0 g
s
.45416 .26527 m .40837 .28327 L .40594 .28583 L p .197 .1 .1 r
F P
0 g
s
.45763 .24641 m .50347 .22962 L .50291 .22366 L p .65 .472 .693 r
F P
0 g
s
.50347 .22962 m .45763 .24641 L .45761 .2523 L p .621 .407 .629 r
F P
0 g
s
.35366 .30546 m .39975 .2873 L .40304 .28718 L p .59 1 .998 r
F P
0 g
s
.39975 .2873 m .35366 .30546 L .34973 .30428 L p .723 1 1 r
F P
0 g
s
.48396 .17775 m .43657 .19065 L .44009 .19501 L p .736 .814 1 r
F P
0 g
s
.44009 .19501 m .48718 .18181 L .48396 .17775 L p .736 .814 1 r
F P
0 g
s
.29474 .31569 m .34135 .29806 L .34559 .3018 L p .778 .888 1 r
F P
0 g
s
.45761 .2523 m .5035 .23508 L .50347 .22962 L p .621 .407 .629 r
F P
0 g
s
.5035 .23508 m .45761 .2523 L .45701 .2575 L p .579 .32 .544 r
F P
0 g
s
.40594 .28583 m .45197 .26763 L .45416 .26527 L p .197 .1 .1 r
F P
0 g
s
.45197 .26763 m .40594 .28583 L .40304 .28718 L p .1 .1 .1 r
F P
0 g
s
.34973 .30428 m .39615 .28616 L .39975 .2873 L p .723 1 1 r
F P
0 g
s
.39615 .28616 m .34973 .30428 L .34559 .3018 L p .768 .951 1 r
F P
0 g
s
.45701 .2575 m .503 .23989 L .5035 .23508 L p .579 .32 .544 r
F P
0 g
s
.503 .23989 m .45701 .2575 L .45585 .26186 L p .512 .198 .421 r
F P
0 g
s
.40304 .28718 m .44934 .26887 L .45197 .26763 L p .1 .1 .1 r
F P
0 g
s
.44934 .26887 m .40304 .28718 L .39975 .2873 L p .59 1 .998 r
F P
0 g
s
.45585 .26186 m .50199 .24392 L .503 .23989 L p .512 .198 .421 r
F P
0 g
s
.50199 .24392 m .45585 .26186 L .45416 .26527 L p .399 .1 .227 r
F P
0 g
s
.34559 .3018 m .39235 .2838 L .39615 .28616 L p .768 .951 1 r
F P
0 g
s
.39235 .2838 m .34559 .3018 L .34135 .29806 L p .778 .888 1 r
F P
0 g
s
.54286 .18352 m .49594 .1981 L .49831 .20441 L p .713 .649 .859 r
F P
0 g
s
.49831 .20441 m .54504 .18936 L .54286 .18352 L p .713 .649 .859 r
F P
0 g
s
.54504 .18936 m .49831 .20441 L .5003 .21089 L p .702 .61 .823 r
F P
0 g
s
.5003 .21089 m .54688 .19534 L .54504 .18936 L p .702 .61 .823 r
F P
0 g
s
.54688 .19534 m .5003 .21089 L .50184 .21736 L p .688 .57 .785 r
F P
0 g
s
.5404 .178 m .49325 .19213 L .49594 .1981 L p .722 .687 .893 r
F P
0 g
s
.49594 .1981 m .54286 .18352 L .5404 .178 L p .722 .687 .893 r
F P
0 g
s
.50184 .21736 m .54832 .20132 L .54688 .19534 L p .688 .57 .785 r
F P
0 g
s
.54832 .20132 m .50184 .21736 L .50291 .22366 L p .672 .525 .743 r
F P
0 g
s
.53771 .17293 m .4903 .18665 L .49325 .19213 L p .729 .726 .928 r
F P
0 g
s
.49325 .19213 m .5404 .178 L .53771 .17293 L p .729 .726 .928 r
F P
0 g
s
.50291 .22366 m .54933 .20713 L .54832 .20132 L p .672 .525 .743 r
F P
0 g
s
.54933 .20713 m .50291 .22366 L .50347 .22962 L p .65 .472 .693 r
F P
0 g
s
.39975 .2873 m .44635 .26894 L .44934 .26887 L p .59 1 .998 r
F P
0 g
s
.44635 .26894 m .39975 .2873 L .39615 .28616 L p .723 1 1 r
F P
0 g
s
.45416 .26527 m .50049 .24707 L .50199 .24392 L p .399 .1 .227 r
F P
0 g
s
.50049 .24707 m .45416 .26527 L .45197 .26763 L p .197 .1 .1 r
F P
0 g
s
.53487 .16845 m .48718 .18181 L .4903 .18665 L p .735 .768 .965 r
F P
0 g
s
.4903 .18665 m .53771 .17293 L .53487 .16845 L p .735 .768 .965 r
F P
0 g
s
.34135 .29806 m .38844 .28026 L .39235 .2838 L p .778 .888 1 r
F P
0 g
s
.50347 .22962 m .54989 .21262 L .54933 .20713 L p .65 .472 .693 r
F P
0 g
s
.54989 .21262 m .50347 .22962 L .5035 .23508 L p .621 .407 .629 r
F P
0 g
s
.53195 .16468 m .48396 .17775 L .48718 .18181 L p .736 .814 1 r
F P
0 g
s
.48718 .18181 m .53487 .16845 L .53195 .16468 L p .736 .814 1 r
F P
0 g
s
.5035 .23508 m .54996 .21764 L .54989 .21262 L p .621 .407 .629 r
F P
0 g
s
.54996 .21764 m .5035 .23508 L .503 .23989 L p .579 .32 .544 r
F P
0 g
s
.45197 .26763 m .49853 .24923 L .50049 .24707 L p .197 .1 .1 r
F P
0 g
s
.49853 .24923 m .45197 .26763 L .44934 .26887 L p .1 .1 .1 r
F P
0 g
s
.39615 .28616 m .44307 .26785 L .44635 .26894 L p .723 1 1 r
F P
0 g
s
.44307 .26785 m .39615 .28616 L .39235 .2838 L p .768 .951 1 r
F P
0 g
s
.503 .23989 m .54956 .22207 L .54996 .21764 L p .579 .32 .544 r
F P
0 g
s
.54956 .22207 m .503 .23989 L .50199 .24392 L p .512 .198 .421 r
F P
0 g
s
.44934 .26887 m .49617 .25034 L .49853 .24923 L p .1 .1 .1 r
F P
0 g
s
.49617 .25034 m .44934 .26887 L .44635 .26894 L p .59 1 .998 r
F P
0 g
s
.39235 .2838 m .4396 .26561 L .44307 .26785 L p .768 .951 1 r
F P
0 g
s
.4396 .26561 m .39235 .2838 L .38844 .28026 L p .778 .888 1 r
F P
0 g
s
.50199 .24392 m .54869 .22577 L .54956 .22207 L p .512 .198 .421 r
F P
0 g
s
.54869 .22577 m .50199 .24392 L .50049 .24707 L p .399 .1 .227 r
F P
0 g
s
.5904 .16876 m .54286 .18352 L .54504 .18936 L p .713 .649 .859 r
F P
0 g
s
.54504 .18936 m .59239 .1741 L .5904 .16876 L p .713 .649 .859 r
F P
0 g
s
.59239 .1741 m .54504 .18936 L .54688 .19534 L p .702 .61 .823 r
F P
0 g
s
.54688 .19534 m .59408 .17959 L .59239 .1741 L p .702 .61 .823 r
F P
0 g
s
.59408 .17959 m .54688 .19534 L .54832 .20132 L p .688 .57 .785 r
F P
0 g
s
.58816 .16369 m .5404 .178 L .54286 .18352 L p .722 .687 .893 r
F P
0 g
s
.54286 .18352 m .5904 .16876 L .58816 .16369 L p .722 .687 .893 r
F P
0 g
s
.38844 .28026 m .43602 .26228 L .4396 .26561 L p .778 .888 1 r
F P
0 g
s
.54832 .20132 m .59541 .18506 L .59408 .17959 L p .688 .57 .785 r
F P
0 g
s
.59541 .18506 m .54832 .20132 L .54933 .20713 L p .672 .525 .743 r
F P
0 g
s
.44635 .26894 m .49347 .25038 L .49617 .25034 L p .59 1 .998 r
F P
0 g
s
.49347 .25038 m .44635 .26894 L .44307 .26785 L p .723 1 1 r
F P
0 g
s
.58573 .15903 m .53771 .17293 L .5404 .178 L p .729 .726 .928 r
F P
0 g
s
.5404 .178 m .58816 .16369 L .58573 .15903 L p .729 .726 .928 r
F P
0 g
s
.50049 .24707 m .54737 .22864 L .54869 .22577 L p .399 .1 .227 r
F P
0 g
s
.54737 .22864 m .50049 .24707 L .49853 .24923 L p .197 .1 .1 r
F P
0 g
s
.54933 .20713 m .59637 .19038 L .59541 .18506 L p .672 .525 .743 r
F P
0 g
s
.59637 .19038 m .54933 .20713 L .54989 .21262 L p .65 .472 .693 r
F P
0 g
s
.58317 .15492 m .53487 .16845 L .53771 .17293 L p .735 .768 .965 r
F P
0 g
s
.53771 .17293 m .58573 .15903 L .58317 .15492 L p .735 .768 .965 r
F P
0 g
s
.54989 .21262 m .59691 .1954 L .59637 .19038 L p .65 .472 .693 r
F P
0 g
s
.59691 .1954 m .54989 .21262 L .54996 .21764 L p .621 .407 .629 r
F P
0 g
s
.44307 .26785 m .4905 .24933 L .49347 .25038 L p .723 1 1 r
F P
0 g
s
.4905 .24933 m .44307 .26785 L .4396 .26561 L p .768 .951 1 r
F P
0 g
s
.58055 .15145 m .53195 .16468 L .53487 .16845 L p .736 .814 1 r
F P
0 g
s
.53487 .16845 m .58317 .15492 L .58055 .15145 L p .736 .814 1 r
F P
0 g
s
.49853 .24923 m .54563 .23061 L .54737 .22864 L p .197 .1 .1 r
F P
0 g
s
.54563 .23061 m .49853 .24923 L .49617 .25034 L p .1 .1 .1 r
F P
0 g
s
.54996 .21764 m .59702 .19999 L .59691 .1954 L p .621 .407 .629 r
F P
0 g
s
.59702 .19999 m .54996 .21764 L .54956 .22207 L p .579 .32 .544 r
F P
0 g
s
.4396 .26561 m .48735 .24722 L .4905 .24933 L p .768 .951 1 r
F P
0 g
s
.48735 .24722 m .4396 .26561 L .43602 .26228 L p .778 .888 1 r
F P
0 g
s
.54956 .22207 m .5967 .20402 L .59702 .19999 L p .579 .32 .544 r
F P
0 g
s
.5967 .20402 m .54956 .22207 L .54869 .22577 L p .512 .198 .421 r
F P
0 g
s
.49617 .25034 m .54353 .23161 L .54563 .23061 L p .1 .1 .1 r
F P
0 g
s
.54353 .23161 m .49617 .25034 L .49347 .25038 L p .59 1 .998 r
F P
0 g
s
.54869 .22577 m .59596 .20738 L .5967 .20402 L p .512 .198 .421 r
F P
0 g
s
.59596 .20738 m .54869 .22577 L .54737 .22864 L p .399 .1 .227 r
F P
0 g
s
.43602 .26228 m .48409 .2441 L .48735 .24722 L p .778 .888 1 r
F P
0 g
s
.49347 .25038 m .54112 .23161 L .54353 .23161 L p .59 1 .998 r
F P
0 g
s
.54112 .23161 m .49347 .25038 L .4905 .24933 L p .723 1 1 r
F P
0 g
s
.63857 .15379 m .5904 .16876 L .59239 .1741 L p .713 .649 .859 r
F P
0 g
s
.59239 .1741 m .64037 .15865 L .63857 .15379 L p .713 .649 .859 r
F P
0 g
s
.64037 .15865 m .59239 .1741 L .59408 .17959 L p .702 .61 .823 r
F P
0 g
s
.59408 .17959 m .6419 .16363 L .64037 .15865 L p .702 .61 .823 r
F P
0 g
s
.6419 .16363 m .59408 .17959 L .59541 .18506 L p .688 .57 .785 r
F P
0 g
s
.63656 .14919 m .58816 .16369 L .5904 .16876 L p .722 .687 .893 r
F P
0 g
s
.5904 .16876 m .63857 .15379 L .63656 .14919 L p .722 .687 .893 r
F P
0 g
s
.54737 .22864 m .59482 .20999 L .59596 .20738 L p .399 .1 .227 r
F P
0 g
s
.59482 .20999 m .54737 .22864 L .54563 .23061 L p .197 .1 .1 r
F P
0 g
s
.59541 .18506 m .64313 .16859 L .6419 .16363 L p .688 .57 .785 r
F P
0 g
s
.64313 .16859 m .59541 .18506 L .59637 .19038 L p .672 .525 .743 r
F P
0 g
s
.63438 .14496 m .58573 .15903 L .58816 .16369 L p .729 .726 .928 r
F P
0 g
s
.58816 .16369 m .63656 .14919 L .63438 .14496 L p .729 .726 .928 r
F P
0 g
s
.59637 .19038 m .64401 .17341 L .64313 .16859 L p .672 .525 .743 r
F P
0 g
s
.64401 .17341 m .59637 .19038 L .59691 .1954 L p .65 .472 .693 r
F P
0 g
s
.4905 .24933 m .53845 .23062 L .54112 .23161 L p .723 1 1 r
F P
0 g
s
.53845 .23062 m .4905 .24933 L .48735 .24722 L p .768 .951 1 r
F P
0 g
s
.63209 .14121 m .58317 .15492 L .58573 .15903 L p .735 .768 .965 r
F P
0 g
s
.58573 .15903 m .63438 .14496 L .63209 .14121 L p .735 .768 .965 r
F P
0 g
s
.59691 .1954 m .64453 .17795 L .64401 .17341 L p .65 .472 .693 r
F P
0 g
s
.64453 .17795 m .59691 .1954 L .59702 .19999 L p .621 .407 .629 r
F P
0 g
s
.54563 .23061 m .5933 .21177 L .59482 .20999 L p .197 .1 .1 r
F P
0 g
s
.5933 .21177 m .54563 .23061 L .54353 .23161 L p .1 .1 .1 r
F P
0 g
s
.62976 .13805 m .58055 .15145 L .58317 .15492 L p .736 .814 1 r
F P
0 g
s
.58317 .15492 m .63209 .14121 L .62976 .13805 L p .736 .814 1 r
F P
0 g
s
.59702 .19999 m .64468 .1821 L .64453 .17795 L p .621 .407 .629 r
F P
0 g
s
.64468 .1821 m .59702 .19999 L .5967 .20402 L p .579 .32 .544 r
F P
0 g
s
.48735 .24722 m .53561 .22864 L .53845 .23062 L p .768 .951 1 r
F P
0 g
s
.53561 .22864 m .48735 .24722 L .48409 .2441 L p .778 .888 1 r
F P
0 g
s
.54353 .23161 m .59144 .21266 L .5933 .21177 L p .1 .1 .1 r
F P
0 g
s
.59144 .21266 m .54353 .23161 L .54112 .23161 L p .59 1 .998 r
F P
0 g
s
.5967 .20402 m .64444 .18574 L .64468 .1821 L p .579 .32 .544 r
F P
0 g
s
.64444 .18574 m .5967 .20402 L .59596 .20738 L p .512 .198 .421 r
F P
0 g
s
.48409 .2441 m .53267 .22574 L .53561 .22864 L p .778 .888 1 r
F P
0 g
s
.59596 .20738 m .64382 .18877 L .64444 .18574 L p .512 .198 .421 r
F P
0 g
s
.64382 .18877 m .59596 .20738 L .59482 .20999 L p .399 .1 .227 r
F P
0 g
s
.54112 .23161 m .5893 .21263 L .59144 .21266 L p .59 1 .998 r
F P
0 g
s
.5893 .21263 m .54112 .23161 L .53845 .23062 L p .723 1 1 r
F P
0 g
s
.59482 .20999 m .64284 .19112 L .64382 .18877 L p .399 .1 .227 r
F P
0 g
s
.64284 .19112 m .59482 .20999 L .5933 .21177 L p .197 .1 .1 r
F P
0 g
s
.53845 .23062 m .58693 .21169 L .5893 .21263 L p .723 1 1 r
F P
0 g
s
.58693 .21169 m .53845 .23062 L .53561 .22864 L p .768 .951 1 r
F P
0 g
s
.68739 .13863 m .63857 .15379 L .64037 .15865 L p .713 .649 .859 r
F P
0 g
s
.64037 .15865 m .68899 .14299 L .68739 .13863 L p .713 .649 .859 r
F P
0 g
s
.68899 .14299 m .64037 .15865 L .6419 .16363 L p .702 .61 .823 r
F P
0 g
s
.6419 .16363 m .69037 .14745 L .68899 .14299 L p .702 .61 .823 r
F P
0 g
s
.69037 .14745 m .6419 .16363 L .64313 .16859 L p .688 .57 .785 r
F P
0 g
s
.68559 .1345 m .63656 .14919 L .63857 .15379 L p .722 .687 .893 r
F P
0 g
s
.63857 .15379 m .68739 .13863 L .68559 .1345 L p .722 .687 .893 r
F P
0 g
s
.64313 .16859 m .69147 .1519 L .69037 .14745 L p .688 .57 .785 r
F P
0 g
s
.69147 .1519 m .64313 .16859 L .64401 .17341 L p .672 .525 .743 r
F P
0 g
s
.68367 .13069 m .63438 .14496 L .63656 .14919 L p .729 .726 .928 r
F P
0 g
s
.63656 .14919 m .68559 .1345 L .68367 .13069 L p .729 .726 .928 r
F P
0 g
s
.64401 .17341 m .69229 .15621 L .69147 .1519 L p .672 .525 .743 r
F P
0 g
s
.69229 .15621 m .64401 .17341 L .64453 .17795 L p .65 .472 .693 r
F P
0 g
s
.5933 .21177 m .64152 .1927 L .64284 .19112 L p .197 .1 .1 r
F P
0 g
s
.64152 .1927 m .5933 .21177 L .59144 .21266 L p .1 .1 .1 r
F P
0 g
s
.68165 .12732 m .63209 .14121 L .63438 .14496 L p .735 .768 .965 r
F P
0 g
s
.63438 .14496 m .68367 .13069 L .68165 .12732 L p .735 .768 .965 r
F P
0 g
s
.53561 .22864 m .58439 .20986 L .58693 .21169 L p .768 .951 1 r
F P
0 g
s
.58439 .20986 m .53561 .22864 L .53267 .22574 L p .778 .888 1 r
F P
0 g
s
.64453 .17795 m .69279 .16028 L .69229 .15621 L p .65 .472 .693 r
F P
0 g
s
.69279 .16028 m .64453 .17795 L .64468 .1821 L p .621 .407 .629 r
F P
0 g
s
.6796 .12448 m .62976 .13805 L .63209 .14121 L p .736 .814 1 r
F P
0 g
s
.63209 .14121 m .68165 .12732 L .6796 .12448 L p .736 .814 1 r
F P
0 g
s
.64468 .1821 m .69295 .16399 L .69279 .16028 L p .621 .407 .629 r
F P
0 g
s
.69295 .16399 m .64468 .1821 L .64444 .18574 L p .579 .32 .544 r
F P
0 g
s
.59144 .21266 m .6399 .19348 L .64152 .1927 L p .1 .1 .1 r
F P
0 g
s
.6399 .19348 m .59144 .21266 L .5893 .21263 L p .59 1 .998 r
F P
0 g
s
.53267 .22574 m .58176 .20718 L .58439 .20986 L p .778 .888 1 r
F P
0 g
s
.64444 .18574 m .69278 .16723 L .69295 .16399 L p .579 .32 .544 r
F P
0 g
s
.69278 .16723 m .64444 .18574 L .64382 .18877 L p .512 .198 .421 r
F P
0 g
s
.5893 .21263 m .63802 .19344 L .6399 .19348 L p .59 1 .998 r
F P
0 g
s
.63802 .19344 m .5893 .21263 L .58693 .21169 L p .723 1 1 r
F P
0 g
s
.64382 .18877 m .69227 .16993 L .69278 .16723 L p .512 .198 .421 r
F P
0 g
s
.69227 .16993 m .64382 .18877 L .64284 .19112 L p .399 .1 .227 r
F P
0 g
s
.58693 .21169 m .63594 .19256 L .63802 .19344 L p .723 1 1 r
F P
0 g
s
.63594 .19256 m .58693 .21169 L .58439 .20986 L p .768 .951 1 r
F P
0 g
s
.64284 .19112 m .69145 .17201 L .69227 .16993 L p .399 .1 .227 r
F P
0 g
s
.69145 .17201 m .64284 .19112 L .64152 .1927 L p .197 .1 .1 r
F P
0 g
s
.58439 .20986 m .63369 .19088 L .63594 .19256 L p .768 .951 1 r
F P
0 g
s
.63369 .19088 m .58439 .20986 L .58176 .20718 L p .778 .888 1 r
F P
0 g
s
.73685 .12327 m .68739 .13863 L .68899 .14299 L p .713 .649 .859 r
F P
0 g
s
.68899 .14299 m .73826 .12712 L .73685 .12327 L p .713 .649 .859 r
F P
0 g
s
.73826 .12712 m .68899 .14299 L .69037 .14745 L p .702 .61 .823 r
F P
0 g
s
.69037 .14745 m .73948 .13106 L .73826 .12712 L p .702 .61 .823 r
F P
0 g
s
.73948 .13106 m .69037 .14745 L .69147 .1519 L p .688 .57 .785 r
F P
0 g
s
.73528 .11961 m .68559 .1345 L .68739 .13863 L p .722 .687 .893 r
F P
0 g
s
.68739 .13863 m .73685 .12327 L .73528 .11961 L p .722 .687 .893 r
F P
0 g
s
.64152 .1927 m .69032 .17341 L .69145 .17201 L p .197 .1 .1 r
F P
0 g
s
.69032 .17341 m .64152 .1927 L .6399 .19348 L p .1 .1 .1 r
F P
0 g
s
.69147 .1519 m .74047 .13499 L .73948 .13106 L p .688 .57 .785 r
F P
0 g
s
.74047 .13499 m .69147 .1519 L .69229 .15621 L p .672 .525 .743 r
F P
0 g
s
.7336 .11624 m .68367 .13069 L .68559 .1345 L p .729 .726 .928 r
F P
0 g
s
.68559 .1345 m .73528 .11961 L .7336 .11624 L p .729 .726 .928 r
F P
0 g
s
.69229 .15621 m .74121 .13879 L .74047 .13499 L p .672 .525 .743 r
F P
0 g
s
.74121 .13879 m .69229 .15621 L .69279 .16028 L p .65 .472 .693 r
F P
0 g
s
.73185 .11326 m .68165 .12732 L .68367 .13069 L p .735 .768 .965 r
F P
0 g
s
.68367 .13069 m .7336 .11624 L .73185 .11326 L p .735 .768 .965 r
F P
0 g
s
.58176 .20718 m .63136 .18843 L .63369 .19088 L p .778 .888 1 r
F P
0 g
s
.69279 .16028 m .74167 .14237 L .74121 .13879 L p .65 .472 .693 r
F P
0 g
s
.74167 .14237 m .69279 .16028 L .69295 .16399 L p .621 .407 .629 r
F P
0 g
s
.6399 .19348 m .68893 .17409 L .69032 .17341 L p .1 .1 .1 r
F P
0 g
s
.68893 .17409 m .6399 .19348 L .63802 .19344 L p .59 1 .998 r
F P
0 g
s
.73008 .11073 m .6796 .12448 L .68165 .12732 L p .736 .814 1 r
F P
0 g
s
.68165 .12732 m .73185 .11326 L .73008 .11073 L p .736 .814 1 r
F P
0 g
s
.69295 .16399 m .74185 .14563 L .74167 .14237 L p .621 .407 .629 r
F P
0 g
s
.74185 .14563 m .69295 .16399 L .69278 .16723 L p .579 .32 .544 r
F P
0 g
s
.63802 .19344 m .6873 .17403 L .68893 .17409 L p .59 1 .998 r
F P
0 g
s
.6873 .17403 m .63802 .19344 L .63594 .19256 L p .723 1 1 r
F P
0 g
s
.69278 .16723 m .74174 .14849 L .74185 .14563 L p .579 .32 .544 r
F P
0 g
s
.74174 .14849 m .69278 .16723 L .69227 .16993 L p .512 .198 .421 r
F P
0 g
s
.69227 .16993 m .74133 .15086 L .74174 .14849 L p .512 .198 .421 r
F P
0 g
s
.74133 .15086 m .69227 .16993 L .69145 .17201 L p .399 .1 .227 r
F P
0 g
s
.63594 .19256 m .68549 .17322 L .6873 .17403 L p .723 1 1 r
F P
0 g
s
.68549 .17322 m .63594 .19256 L .63369 .19088 L p .768 .951 1 r
F P
0 g
s
.69145 .17201 m .74065 .15267 L .74133 .15086 L p .399 .1 .227 r
F P
0 g
s
.74065 .15267 m .69145 .17201 L .69032 .17341 L p .197 .1 .1 r
F P
0 g
s
.63369 .19088 m .68354 .17169 L .68549 .17322 L p .768 .951 1 r
F P
0 g
s
.68354 .17169 m .63369 .19088 L .63136 .18843 L p .778 .888 1 r
F P
0 g
s
.69032 .17341 m .73971 .15389 L .74065 .15267 L p .197 .1 .1 r
F P
0 g
s
.73971 .15389 m .69032 .17341 L .68893 .17409 L p .1 .1 .1 r
F P
0 g
s
.63136 .18843 m .6815 .16948 L .68354 .17169 L p .778 .888 1 r
F P
0 g
s
.78692 .10778 m .73685 .12327 L .73826 .12712 L p .711 .648 .859 r
F P
0 g
s
.73826 .12712 m .78812 .11107 L .78692 .10778 L p .711 .648 .859 r
F P
0 g
s
.78812 .11107 m .73826 .12712 L .73948 .13106 L p .7 .61 .824 r
F P
0 g
s
.73948 .13106 m .78916 .11444 L .78812 .11107 L p .7 .61 .824 r
F P
0 g
s
.78916 .11444 m .73948 .13106 L .74047 .13499 L p .687 .569 .786 r
F P
0 g
s
.78559 .10465 m .73528 .11961 L .73685 .12327 L p .721 .686 .893 r
F P
0 g
s
.73685 .12327 m .78692 .10778 L .78559 .10465 L p .721 .686 .893 r
F P
0 g
s
.74047 .13499 m .79001 .11779 L .78916 .11444 L p .687 .569 .786 r
F P
0 g
s
.79001 .11779 m .74047 .13499 L .74121 .13879 L p .67 .524 .743 r
F P
0 g
s
.78418 .10177 m .7336 .11624 L .73528 .11961 L p .728 .725 .928 r
F P
0 g
s
.73528 .11961 m .78559 .10465 L .78418 .10177 L p .728 .725 .928 r
F P
0 g
s
.74121 .13879 m .79066 .12103 L .79001 .11779 L p .67 .524 .743 r
F P
0 g
s
.79066 .12103 m .74121 .13879 L .74167 .14237 L p .648 .471 .693 r
F P
0 g
s
.68893 .17409 m .73853 .15447 L .73971 .15389 L p .1 .1 .1 r
F P
0 g
s
.73853 .15447 m .68893 .17409 L .6873 .17403 L p .59 1 .998 r
F P
0 g
s
.78271 .09921 m .73185 .11326 L .7336 .11624 L p .733 .767 .965 r
F P
0 g
s
.7336 .11624 m .78418 .10177 L .78271 .09921 L p .733 .767 .965 r
F P
0 g
s
.74167 .14237 m .79108 .12408 L .79066 .12103 L p .648 .471 .693 r
F P
0 g
s
.79108 .12408 m .74167 .14237 L .74185 .14563 L p .619 .405 .629 r
F P
0 g
s
.78123 .09705 m .73008 .11073 L .73185 .11326 L p .734 .813 1 r
F P
0 g
s
.73185 .11326 m .78271 .09921 L .78123 .09705 L p .734 .813 1 r
F P
0 g
s
.74185 .14563 m .79126 .12686 L .79108 .12408 L p .619 .405 .629 r
F P
0 g
s
.79126 .12686 m .74185 .14563 L .74174 .14849 L p .576 .319 .544 r
F P
0 g
s
.6873 .17403 m .73715 .15439 L .73853 .15447 L p .59 1 .998 r
F P
0 g
s
.73715 .15439 m .6873 .17403 L .68549 .17322 L p .723 1 1 r
F P
0 g
s
.74174 .14849 m .7912 .12928 L .79126 .12686 L p .576 .319 .544 r
F P
0 g
s
.7912 .12928 m .74174 .14849 L .74133 .15086 L p .509 .196 .42 r
F P
0 g
s
.68549 .17322 m .7356 .15366 L .73715 .15439 L p .723 1 1 r
F P
0 g
s
.7356 .15366 m .68549 .17322 L .68354 .17169 L p .768 .951 1 r
F P
0 g
s
.74133 .15086 m .79089 .13129 L .7912 .12928 L p .509 .196 .42 r
F P
0 g
s
.79089 .13129 m .74133 .15086 L .74065 .15267 L p .395 .1 .226 r
F P
0 g
s
.68354 .17169 m .73392 .15229 L .7356 .15366 L p .768 .951 1 r
F P
0 g
s
.73392 .15229 m .68354 .17169 L .6815 .16948 L p .778 .888 1 r
F P
0 g
s
.74065 .15267 m .79035 .13283 L .79089 .13129 L p .395 .1 .226 r
F P
0 g
s
.79035 .13283 m .74065 .15267 L .73971 .15389 L p .191 .1 .1 r
F P
0 g
s
.6815 .16948 m .73216 .15032 L .73392 .15229 L p .778 .888 1 r
F P
0 g
s
.73971 .15389 m .78959 .13385 L .79035 .13283 L p .191 .1 .1 r
F P
0 g
s
.78959 .13385 m .73971 .15389 L .73853 .15447 L p .1 .1 .1 r
F P
0 g
s
.73853 .15447 m .78863 .13432 L .78959 .13385 L p .1 .1 .1 r
F P
0 g
s
.78863 .13432 m .73853 .15447 L .73715 .15439 L p .595 1 .999 r
F P
0 g
s
.83754 .09226 m .78692 .10778 L .78812 .11107 L p .708 .646 .859 r
F P
0 g
s
.78812 .11107 m .83849 .09488 L .83754 .09226 L p .708 .646 .859 r
F P
0 g
s
.83849 .09488 m .78812 .11107 L .78916 .11444 L p .697 .608 .824 r
F P
0 g
s
.78916 .11444 m .83932 .09757 L .83849 .09488 L p .697 .608 .824 r
F P
0 g
s
.83932 .09757 m .78916 .11444 L .79001 .11779 L p .684 .567 .786 r
F P
0 g
s
.8365 .08976 m .78559 .10465 L .78692 .10778 L p .717 .684 .894 r
F P
0 g
s
.78692 .10778 m .83754 .09226 L .8365 .08976 L p .717 .684 .894 r
F P
0 g
s
.79001 .11779 m .84001 .10023 L .83932 .09757 L p .684 .567 .786 r
F P
0 g
s
.84001 .10023 m .79001 .11779 L .79066 .12103 L p .667 .521 .743 r
F P
0 g
s
.83539 .08746 m .78418 .10177 L .78559 .10465 L p .725 .723 .929 r
F P
0 g
s
.78559 .10465 m .8365 .08976 L .83539 .08746 L p .725 .723 .929 r
F P
0 g
s
.79066 .12103 m .84054 .10281 L .84001 .10023 L p .667 .521 .743 r
F P
0 g
s
.84054 .10281 m .79066 .12103 L .79108 .12408 L p .645 .468 .692 r
F P
0 g
s
.73715 .15439 m .78749 .13424 L .78863 .13432 L p .595 1 .999 r
F P
0 g
s
.78749 .13424 m .73715 .15439 L .7356 .15366 L p .727 1 1 r
F P
0 g
s
.83424 .08542 m .78271 .09921 L .78418 .10177 L p .729 .765 .966 r
F P
0 g
s
.78418 .10177 m .83539 .08746 L .83424 .08542 L p .729 .765 .966 r
F P
0 g
s
.79108 .12408 m .84089 .10524 L .84054 .10281 L p .645 .468 .692 r
F P
0 g
s
.84089 .10524 m .79108 .12408 L .79126 .12686 L p .615 .402 .629 r
F P
0 g
s
.83309 .08369 m .78123 .09705 L .78271 .09921 L p .73 .812 1 r
F P
0 g
s
.78271 .09921 m .83424 .08542 L .83309 .08369 L p .73 .812 1 r
F P
0 g
s
.79126 .12686 m .84106 .10744 L .84089 .10524 L p .615 .402 .629 r
F P
0 g
s
.84106 .10744 m .79126 .12686 L .7912 .12928 L p .571 .315 .543 r
F P
0 g
s
.7356 .15366 m .78621 .1336 L .78749 .13424 L p .727 1 1 r
F P
0 g
s
.78621 .1336 m .7356 .15366 L .73392 .15229 L p .77 .952 1 r
F P
0 g
s
.7912 .12928 m .84103 .10936 L .84106 .10744 L p .571 .315 .543 r
F P
0 g
s
.84103 .10936 m .7912 .12928 L .79089 .13129 L p .502 .19 .418 r
F P
0 g
s
.73392 .15229 m .78482 .13241 L .78621 .1336 L p .77 .952 1 r
F P
0 g
s
.78482 .13241 m .73392 .15229 L .73216 .15032 L p .78 .889 1 r
F P
0 g
s
.79089 .13129 m .84082 .11094 L .84103 .10936 L p .502 .19 .418 r
F P
0 g
s
.84082 .11094 m .79089 .13129 L .79035 .13283 L p .385 .1 .222 r
F P
0 g
s
.73216 .15032 m .78336 .13072 L .78482 .13241 L p .78 .889 1 r
F P
0 g
s
.79035 .13283 m .84042 .11215 L .84082 .11094 L p .385 .1 .222 r
F P
0 g
s
.84042 .11215 m .79035 .13283 L .78959 .13385 L p .178 .1 .1 r
F P
0 g
s
.78959 .13385 m .83985 .11295 L .84042 .11215 L p .178 .1 .1 r
F P
0 g
s
.83985 .11295 m .78959 .13385 L .78863 .13432 L p .1 .1 .1 r
F P
0 g
s
.83912 .11331 m .83985 .11295 L p .82493 .11918 L .1 .1 .1 r
F P
0 g
s
.89155 .09223 m .84042 .11215 L .83985 .11295 L p .2 .1 .1 r
F P
0 g
s
.83985 .11295 m .89112 .09287 L .89155 .09223 L p .2 .1 .1 r
F P
0 g
s
.89112 .09287 m .83985 .11295 L .83912 .11331 L p .1 .1 .1 r
F P
0 g
s
.83912 .11331 m .83825 .11323 L p .85338 .10742 L .587 1 .998 r
F P
0 g
s
.89056 .09315 m .89112 .09287 L p .87461 .09936 L .1 .1 .1 r
F P
0 g
s
.83912 .11331 m .89056 .09315 L p .87461 .09936 L .1 .1 .1 r
F P
0 g
s
.94362 .07279 m .89155 .09223 L .89112 .09287 L p .215 .1 .1 r
F P
0 g
s
.89112 .09287 m .94329 .0733 L .94362 .07279 L p .215 .1 .1 r
F P
0 g
s
.94329 .0733 m .89112 .09287 L .89056 .09315 L p .1 .1 .1 r
F P
0 g
s
.89056 .09315 m .88989 .09307 L p .90658 .08691 L .573 .993 .997 r
F P
0 g
s
.89056 .09315 m .94285 .07352 L .94329 .0733 L p .1 .1 .1 r
F P
0 g
s
.94285 .07352 m .89056 .09315 L p .90658 .08691 L .573 .993 .997 r
F P
0 g
s
.89056 .09315 m .83912 .11331 L p .85338 .10742 L .587 1 .998 r
F P
0 g
s
.78863 .13432 m .83912 .11331 L p .82493 .11918 L .1 .1 .1 r
F P
0 g
s
.83912 .11331 m .78863 .13432 L .78749 .13424 L p .607 1 1 r
F P
0 g
s
.78749 .13424 m .83825 .11323 L .83912 .11331 L p .607 1 1 r
F P
0 g
s
.83825 .11323 m .78749 .13424 L .78621 .1336 L p .735 1 1 r
F P
0 g
s
.8891 .07621 m .83754 .09226 L .83849 .09488 L p .714 .649 .859 r
F P
0 g
s
.83849 .09488 m .88987 .07833 L .8891 .07621 L p .714 .649 .859 r
F P
0 g
s
.88987 .07833 m .83849 .09488 L .83932 .09757 L p .703 .611 .823 r
F P
0 g
s
.83932 .09757 m .89054 .0805 L .88987 .07833 L p .703 .611 .823 r
F P
0 g
s
.89054 .0805 m .83932 .09757 L .84001 .10023 L p .689 .57 .785 r
F P
0 g
s
.88827 .07419 m .8365 .08976 L .83754 .09226 L p .723 .687 .893 r
F P
0 g
s
.83754 .09226 m .8891 .07621 L .88827 .07419 L p .723 .687 .893 r
F P
0 g
s
.84001 .10023 m .8911 .08265 L .89054 .0805 L p .689 .57 .785 r
F P
0 g
s
.8911 .08265 m .84001 .10023 L .84054 .10281 L p .673 .525 .743 r
F P
0 g
s
.78621 .1336 m .83727 .11271 L .83825 .11323 L p .735 1 1 r
F P
0 g
s
.83727 .11271 m .78621 .1336 L .78482 .13241 L p .775 .953 1 r
F P
0 g
s
.88738 .07233 m .83539 .08746 L .8365 .08976 L p .73 .726 .928 r
F P
0 g
s
.8365 .08976 m .88827 .07419 L .88738 .07233 L p .73 .726 .928 r
F P
0 g
s
.84054 .10281 m .89154 .08473 L .8911 .08265 L p .673 .525 .743 r
F P
0 g
s
.89154 .08473 m .84054 .10281 L .84089 .10524 L p .651 .473 .693 r
F P
0 g
s
.88647 .07067 m .83424 .08542 L .83539 .08746 L p .735 .768 .964 r
F P
0 g
s
.83539 .08746 m .88738 .07233 L .88647 .07067 L p .735 .768 .964 r
F P
0 g
s
.84089 .10524 m .89184 .08668 L .89154 .08473 L p .651 .473 .693 r
F P
0 g
s
.89184 .08668 m .84089 .10524 L .84106 .10744 L p .622 .407 .629 r
F P
0 g
s
.88556 .06927 m .83309 .08369 L .83424 .08542 L p .737 .815 1 r
F P
0 g
s
.83424 .08542 m .88647 .07067 L .88556 .06927 L p .737 .815 1 r
F P
0 g
s
.78482 .13241 m .8362 .11175 L .83727 .11271 L p .775 .953 1 r
F P
0 g
s
.8362 .11175 m .78482 .13241 L .78336 .13072 L p .784 .89 1 r
F P
0 g
s
.84106 .10744 m .89199 .08845 L .89184 .08668 L p .622 .407 .629 r
F P
0 g
s
.89199 .08845 m .84106 .10744 L .84103 .10936 L p .58 .321 .544 r
F P
0 g
s
.84103 .10936 m .89199 .09 L .89199 .08845 L p .58 .321 .544 r
F P
0 g
s
.89199 .09 m .84103 .10936 L .84082 .11094 L p .514 .2 .421 r
F P
0 g
s
.78336 .13072 m .83507 .11039 L .8362 .11175 L p .784 .89 1 r
F P
0 g
s
.84082 .11094 m .89184 .09127 L .89199 .09 L p .514 .2 .421 r
F P
0 g
s
.89184 .09127 m .84082 .11094 L .84042 .11215 L p .401 .1 .228 r
F P
0 g
s
.84042 .11215 m .89155 .09223 L .89184 .09127 L p .401 .1 .228 r
F P
0 g
s
.83825 .11323 m .88989 .09307 L .89056 .09315 L p .587 1 .998 r
F P
0 g
s
.88989 .09307 m .83825 .11323 L .83727 .11271 L p .722 1 1 r
F P
0 g
s
.83727 .11271 m .88912 .09264 L .88989 .09307 L p .722 1 1 r
F P
0 g
s
.88912 .09264 m .83727 .11271 L .8362 .11175 L p .766 .95 1 r
F P
0 g
s
.8362 .11175 m .88828 .09185 L .88912 .09264 L p .766 .95 1 r
F P
0 g
s
.88828 .09185 m .8362 .11175 L .83507 .11039 L p .777 .888 1 r
F P
0 g
s
.94151 .05972 m .8891 .07621 L .88987 .07833 L p .717 .651 .858 r
F P
0 g
s
.88987 .07833 m .94214 .06145 L .94151 .05972 L p .717 .651 .858 r
F P
0 g
s
.94214 .06145 m .88987 .07833 L .89054 .0805 L p .706 .613 .823 r
F P
0 g
s
.89054 .0805 m .94269 .06323 L .94214 .06145 L p .706 .613 .823 r
F P
0 g
s
.94269 .06323 m .89054 .0805 L .8911 .08265 L p .693 .573 .785 r
F P
0 g
s
.94084 .05806 m .88827 .07419 L .8891 .07621 L p .726 .689 .892 r
F P
0 g
s
.8891 .07621 m .94151 .05972 L .94084 .05806 L p .726 .689 .892 r
F P
0 g
s
.8911 .08265 m .94315 .06499 L .94269 .06323 L p .693 .573 .785 r
F P
0 g
s
.94315 .06499 m .8911 .08265 L .89154 .08473 L p .677 .528 .743 r
F P
0 g
s
.94012 .05654 m .88738 .07233 L .88827 .07419 L p .734 .728 .927 r
F P
0 g
s
.88827 .07419 m .94084 .05806 L .94012 .05654 L p .734 .728 .927 r
F P
0 g
s
.83507 .11039 m .88739 .09075 L .88828 .09185 L p .777 .888 1 r
F P
0 g
s
.89154 .08473 m .94352 .06668 L .94315 .06499 L p .677 .528 .743 r
F P
0 g
s
.94352 .06668 m .89154 .08473 L .89184 .08668 L p .655 .476 .692 r
F P
0 g
s
.93939 .05518 m .88647 .07067 L .88738 .07233 L p .74 .77 .963 r
F P
0 g
s
.88738 .07233 m .94012 .05654 L .93939 .05518 L p .74 .77 .963 r
F P
0 g
s
.89184 .08668 m .94377 .06828 L .94352 .06668 L p .655 .476 .692 r
F P
0 g
s
.94377 .06828 m .89184 .08668 L .89199 .08845 L p .627 .411 .63 r
F P
0 g
s
.93867 .05402 m .88556 .06927 L .88647 .07067 L p .742 .817 1 r
F P
0 g
s
.88647 .07067 m .93939 .05518 L .93867 .05402 L p .742 .817 1 r
F P
0 g
s
.89199 .08845 m .94391 .06972 L .94377 .06828 L p .627 .411 .63 r
F P
0 g
s
.94391 .06972 m .89199 .08845 L .89199 .09 L p .586 .326 .545 r
F P
0 g
s
.89199 .09 m .94393 .07098 L .94391 .06972 L p .586 .326 .545 r
F P
0 g
s
.94393 .07098 m .89199 .09 L .89184 .09127 L p .521 .206 .423 r
F P
0 g
s
.89184 .09127 m .94383 .07201 L .94393 .07098 L p .521 .206 .423 r
F P
0 g
s
.94383 .07201 m .89184 .09127 L .89155 .09223 L p .412 .1 .231 r
F P
0 g
s
.89155 .09223 m .94362 .07279 L .94383 .07201 L p .412 .1 .231 r
F P
0 g
s
.88989 .09307 m .94233 .07345 L .94285 .07352 L p .573 .993 .997 r
F P
0 g
s
.94233 .07345 m .88989 .09307 L .88912 .09264 L p .712 1 1 r
F P
0 g
s
.88912 .09264 m .94173 .07308 L .94233 .07345 L p .712 1 1 r
F P
0 g
s
.94173 .07308 m .88912 .09264 L .88828 .09185 L p .759 .948 1 r
F P
0 g
s
.88828 .09185 m .94106 .07244 L .94173 .07308 L p .759 .948 1 r
F P
0 g
s
.94106 .07244 m .88828 .09185 L .88739 .09075 L p .771 .886 1 r
F P
0 g
s
.88739 .09075 m .94036 .07153 L .94106 .07244 L p .771 .886 1 r
F P
0 g
s
.25 Mabswid
.95886 0 m
.98445 .03804 L
s
.98445 .03804 m
1 .18401 L
s
1 .18401 m
.97383 .14677 L
s
.97383 .14677 m
.95886 0 L
s
.0127 .31334 m
0 .45265 L
s
0 .45265 m
.97383 .14677 L
s
.97383 .14677 m
.95886 0 L
s
.95886 0 m
.0127 .31334 L
s
.98445 .03804 m
1 .18401 L
s
.9864 .0564 m
.98015 .05842 L
s
[(-0.5)] .99891 .05235 -1 .32381 Mshowa
.99013 .09135 m
.98387 .09335 L
s
[(0)] 1.00264 .08734 -1 .31975 Mshowa
.99391 .12681 m
.98764 .12879 L
s
[(0.5)] 1.00644 .12285 -1 .31563 Mshowa
.99774 .16279 m
.99147 .16475 L
s
[(1)] 1.01029 .15888 -1 .31146 Mshowa
.125 Mabswid
.98714 .06335 m
.98339 .06456 L
s
.98789 .07032 m
.98413 .07152 L
s
.98863 .07731 m
.98488 .07851 L
s
.98938 .08432 m
.98562 .08552 L
s
.99088 .0984 m
.98712 .09959 L
s
.99163 .10547 m
.98788 .10666 L
s
.99239 .11256 m
.98863 .11375 L
s
.99315 .11967 m
.98939 .12086 L
s
.99467 .13396 m
.99091 .13515 L
s
.99543 .14114 m
.99167 .14232 L
s
.9962 .14833 m
.99244 .14951 L
s
.99697 .15555 m
.99321 .15673 L
s
.98567 .04947 m
.98192 .05068 L
s
.98493 .04256 m
.98118 .04378 L
s
.99851 .17005 m
.99475 .17122 L
s
.99929 .17734 m
.99552 .1785 L
s
gsave
1.06725 .08606 -61 -13.8819 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{325.938, 157.125},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000550000W@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl0AP000?l0AP000?l0AP000?l0AP000?l0
AP000?l0;`020@P00P4;0000o`0_00030@0000P000<100002@000?l080050@0500400@000P430004
0@0000@10`000`4000090000o`0R00090@0000400@0100@000D100000@0200050@0000402`000?l0
8P002@40000100400@0400030@00008000<100000`000`4000090000o`0;00815@002@4000010040
0@0400030@0000<000<100000P000`4000090000o`0:00040@000A@000@100000P400`0100040004
0@0000@10`000`4000090000o`0:00030@00008110000`40000200@110050@X00P4800812`000?l0
1@020@<000H10041004800030@0002/0003o00800`400`010@0200060@010@012@000`40000Z0000
oP030@@00P4200050@010@000P4400030@00008000<10040:@000?/00`4700<10P020@0500410040
2@020BT0003h00<12P040@8000H10041004600H1:0000?D00`4>00@10P020@800P4400050@010@00
0`4W0000lP030A40104300060@010@011P000`400@0200030@0002@0003_00<150000`400@020@80
00H10041004700810P000`40000T0000k0030AL000<100000`4200810P000`40000600<19`000>T0
0`4J00030@0000<10`001040004800030@0002L0003V00<17@00104000030@8000@1000120000`40
000W0000h`030B0000<100400P400`01000200812P030BH0003P00<18P030@<00P4;008110000`40
000200<17`000=d00`4R00<100<00@0010000`40000800040@000@/000<1000070000=T0104R00<1
10000`40000200812P001040004;00030@0001`0003F00<18`050@H000<100000P020@<010420004
0@000@P00`4O0000d`030B<0204600070@0000400@0900040@000@P000<100007`000=000`4S00/1
1P000`40000300030@0000L000@1000120000`40000=008120020@H0003=00<18`0;0@04K8D10@D0
00<100000P020@X00P4900@130000`40000800030@0000@0003:00<18`0<0@04R`41Q@811@000`40
000300030@0001@00`4200030@0000<11000104000040@<000<1000010000<L00`4R00d100>50@40
1P4500030@0000800P4G00040@000@8000<100001@001@400001008000D100000@060000a0030B80
5`4700030@0000800P45008140030@<00`4500030@00008000<100000`000`4000040000`@030B40
604:00030@0000<000<100000P001040004?00040@000@8000@1000110000`40000300030@000080
00<1000010000;h00`4P00l100>;0@401`4=00030@0000800P4400040@000@l000@100010`020@D0
00@10000104300030@0000@0002k00<17P0E0@03Q@4100@13`000`40000300030@00008000@10001
3P040@/00P4800811P000;P00`4L01h14`00104000030@<000@10001;`000;D00`4K01l15P000`40
000300050@0000400P000`40000]0000/P030A/00P400i2;0@0=0@05R`410HD02@4H00030@000080
0P440081<0000:l00`4J00<100>;0@401@400h/10@050@05R`41K8D02P4J00030@0000<000<10000
=0000:`00`4J00<100>@0@401@400h/10@090@04Q@41K0L17@000`4000020081=P000:T00`4J00X1
00>:0@401@400f`10@0:0@04T0410Al000<100000`000`40000d0000YP030AT00`400h/10@050@03
R`41008100B;0H^51@400hD10@040@03090100818@000`40000200811@020@@000<100000P030B<0
002S00<16@0:0@03R`4100<100B;0@5/0P400hD10@080@04T0410B@000<100000`000`4000020004
0@000@/000<1000080000:000`4H00L100F@0@41R`0>0@03Q@4100T19`000`400002008110001040
004;00030@000200002M00<1600B0@04Q@41R`D100>50HD02@4Y00030@0000<000<100000P001040
004800<18`0009X00`4G00<100F;0@41R`030@05R`410H/00`400hX10@060@06Q@41Q@5V1`4/0004
0@0000<10`001040004800030@0002<0002G00<15`030@05R`410H/03P400hX10@090@03Q@4100D1
;P000`40000300050@0000400P000`40000600030@0002<0002D00<15P0?0@05R`410H/00`4018/1
0HD30@03Q@4100L100B@0@41<0000`400002008110020@T0104R0000T@030AD010401H/10@6;00<1
00F;0@41R`0:0@05K041QF`02P400i010@0c00030@0000<000<10000<`0008h00`4D00<100>@R`40
2`400h/10@040@03Q@4100D100B50@652P4f00030@0000800P4e0000R`030A<03@401HX10@6500H1
00>:K0401@400hE/0@040@03IP4100D1>0020@D000<10000<`0008P00`4B00P100>;0@400P42R`<1
00>;0@401@4018D10H/30@>520400i010@020CL00`4200040@00008110040Bd0002400@14P020@03
R`41008100>;0@400P400hZ;RP030@03Q@4100@100>:K0400P401V`1K8D1Q@P100>;T0400P4g00<1
1@000`40000200030@0000@000<10000;00008400`4B00<100F;0@41R`040@>;2`400h^5R`020@03
Q@4100>520400i000@020CH0104900030@0000811P000`40000/0000OP030A801`4028/10@6:R`6;
0`401H/10@6;008100A/RP6:1P401Fb50@5[00D100>;0@400`4e00<13P001040004600030@0002`0
001k00<14@070@03T041008118/;0@03RhF;00<100B;0@6510400hD1Q@030@0300410081=@030A80
0P400`01000400030@0002`0001h00<13`050@05R`410H/00`401H^:RhZ;00H100R50@41K8Y/RP81
00=/Q@401@400f^50@040@03T0410081=@030AD000@100010`030Bh0001e00<13P040@03T8^@00<1
00B@0@4118/70@07RP410HF;Q@040@04R`41Q@<10XD50@03090100<1=0030AT00`4500030@0002`0
001b00<13@040@>;104028/10@6;RP6:1`401hF;0HY/0F`00`400f`10@020@06Q@65J`5[1@400h/1
0@020C<0104M0081=00006l00`4=008100F;T8/1R`030@03R`4100810h/70@03RhZ;008100>5RhD0
0`4018D10@42Q@810hD;0C800`4Q0081=00006`00`4=00<10h/40@07R`410HZ;RP040@07R`410H^5
R`030@06RV`10HE/0`400fb50@020@05QF/10HD00P400h/10@040C400`4T0081=00006T00`4=00<1
00>@R`400`400i2;0@020@:;1042R`<100B;RP6:0P400h^50@020@03QH/100:500D1Q@41Q@0=0C40
0`4Q00030@0000<1=P0006H00`4<00@100>;0@400P43R`<100>:0@400P400h^5R`030@04RhF;Q@<1
00I/0@41K8D20@03K8E/00H100F50@41T0020@04T0410C000`4S00@1>@0006<00`4;00@100>;0@40
0P400i2;T0040@04R`410@:;00<1R`400P4018^:RhX20@03RhD1008100>5R`400P42Q@8100F50@41
Q@030@050041T0000`4^00@18`030Cd0001P00<12P040@:;0`42R`030H/1008100>;0@400P4018F;
QH/30@03RhF;00@100M/RP65K05/008100>50@400P4016/1IXD30@05T0410I000`4^00<190030D00
001M00<12@040@03Ri2;00<100>;T8/0104018/10@45R`<100>;0H/00`400h^5R`020@03QH^500H1
0XD20@03QFH100D100<00@400P4]00<190030D<0001J00<12@020@>;00<1R`400P42R`@100>:R`40
0P4018^5RhD40@03RhF;00<100>:K8X00`401HE/0@6500<100>5J`400P401FH10@6;008100>@0@40
0P4]00<18`040DH0001G00<120040@03T8^@00@100>;0@400P43R`<10X/00`6;0@020@04RX^:R`81
00F5RhD1Q@020@03R`41008100J50@650HD20@06IP410I000P4019010@4]00<18`030DX0001D00<1
20030@>;1042R`<100B:RhZ;0`4018^5RhD30@04QH^5R`<100B:K8Y/1P400hE/0@020@03JhD100H1
00>@R`400Y030B/0104P00030@0000<1C@0005400`4700@100>;T0400`400h^@R`030@:;00<1R`40
0P44R`<100B:R`6;0`4018^5RhD20@04R`410@>50P43Q@030FH100@100B@0@400`4[00<190030E00
001>00<11P040@:;1043R`@100B:RhZ;0`4018^5RhD30@04QH^5R`<100>:K8X010401f`1K8E/0F`0
0P401XE[0@650@9V2@4Z00<18`040E<0001;00<11@040@03Ri2;00<100B;T06@0`45R`<10X/00`6;
0@020@04RX^:R`8100B5R`6;0`4018^50@44Q@<10XD40@03IXD1008100<00@400P4Z00<18`030EL0
001800<11@030@B;0`43R`@100B;RX^:104018^5RhD30@03QH^500@100>:K8X00`400hE/Q@020@03
K8E/008100B5J`5[10400fH10@060BX00`4S00<1FP0004D00`45008100F;T8/1R`030@03Ri0100<1
0h/00`6;0@020@F;0`400hX1RP030@04QH^5R`8100F5RhD1Q@020@03Q@4100810hD90@03T0410081
:0040B<00`5M0000@P030@@01043R`@100>;0@400P401H^:RhZ;00@100B5RhF;10400hF;Q@030@04
K8Y/RP<100B5K8E/1P4018E[QF/20@03Q@4100@100>@0@400P4X00<18`040F00000o00<110030@03
Ri2;00@100>@R`400P45R`@10h/00`6;0@020@04RhZ;RP<100B5R`6;0`4018^5RhD20@04Q@410@>5
10400fH1IP030@03T04100<19`030B0000<100000`5T0000?0030@<01042R`@10h/30@04RX/1R`@1
00F5RhF;Q@040@04RhF;Q@<100A/RVb:0`400hE/Q@040@06K041Q@650`400hEV0@0:0BL00`4S00<1
I`0003T00`4200@100>@Ri000`40192;T8/30@B;0`45R`@100F:RhX1RP030@04QH^5R`8100B5R`6;
0`42Q@<10XD30@03QFH100D100<00@400P4W00<18`030FX0000f00<100<00@400P44R`<10X/00`6;
0@020@04RX^:R`<100B;Q@6510401X^5RhF;Q@<100A/RVb:0`400hE/Q@030@03QFb5008100Q[QF/1
IXD1Q@D10Y040BH00`4R00@1K@0003<00`402@010@6@Ri2;T0030@03T8^@00@118/30@B;0`401hZ;
RX^:RhX00`4018F;0H/30@03RhF;00<10XD010650@42Q@8100>5IXD01P400i010@020BD0104R00<1
L@0003000`42008100>@R`400X/30@:;10401H^:R`6;00<100B;QH^50`401hD1QH^5RhD00`4016b:
K8X30@03QFb500<100B5K8E/0`401V^50@5VQ@H100>@0@400`4T00<18`030G@0000/00@10P030@04
R`2;00<100>@0@400P45R`@118/30@07RX^:RhX1RP030@04QH^5R`8100B5R`6;0`44Q@810XD01@65
0@6500H100>@0@400`4T00<18`030GL0000Y00<110020@05Ri2;T8/00P402H/10@6;RX^:R`040@05
QH^50HD00`401XF;QH^5R`@100A/RVb:0`400hE/Q@030@04QFb5K0<100=[QF/02P400i010@030B<0
0`4P00030@0100<1NP0002H00`45008100@0R`2;0`401h/10J:;0H/01046R`<100F:RhZ;RP040@05
RhF;0H/00`400h^5R`030@:500<1Q@400P43Q@8100>50@401@400`2@0@040B800`4R00<140040@<0
0P5U00008`030@H00P400h^@R`030@:;0P400hZ;RP030@07QH^5RhF;Q@030@04QH/1R`@100I/RVb:
K8X30@03QFb500<100B5K8E/0`400f^5J`030@03IP4100D100>@0@400`4Q00@18P030AD000D10000
0@0200030@000680000P00<1200010400@020@03R`2;008100>RRj800P400hX1R`04R`<100B:RhZ;
0`401h^5RhF;QH/00P4018F;0H/30@B50`400hD1Q@030@03IP4100D100>@0@40104P00<18`030AP0
00D100000@0200030@000680000M00<12`00106Q0@43R`8100>:RhX00P401X^:RhX1Q@<100B5RhF;
0`401hX1RVb:K8X00`400hE/Q@030@04QFb5K0<100=[QF/00`400fJ50@020@04T041T0H180030B80
104K00050@0000400P000`40001R00006P030@d000H1X@6Q0H/20@03XX^R008100F:XXZRRP030@04
RX^:R`<100N;QH^5R`6;00<100>;QH/00`42Q@030HD1008118D20@03IP5V008100D0T04100050B00
0`4R00<17`001@400001008000<10000HP0001L00`4?000:0J41RZ::0HZ;RP8100F;RX^:R`020@04
QH/1R`<100N:K8Y/RVb:00<100>5K8D00`4018E/QF`30@04JhE[Q@<100>5IP400P400h/10@02T0D1
7`030B800`4P00<10`001040005T000050030A4000P1X@010:41X@<100B:XXZR0`400h^:R`030@07
RhF;QH^5R`030@03R`6;00<11HD20@03Q@6500:50P400hEVQ@020@05T0410I001@4N00@180000`40
0@020BD000<100000P020FD0000A00<150003@400J::0HZR0H^:RhX00P4018^5RhD20@07RP6:K8Y/
RP030@03QFb500<100F5K8E/Q@030@05QF^50HD00P400fJ50@020@04T041T0H17@030B80106D0000
3P030AH000`10040X@6Q006R0J820@04RX^:R`8100J5RhF;0H/30@03RhF;00<100>50HD00hD20@F5
0`400hEV0@020@04T04100D17@030B800`6H00002`030AP000/1006RRZ81XP6QRP020@04RhF;Q@81
00F:QHZ5RP030@04QFb5K08100N5K8E/Q@65008100A[QF^50P4016J50HD20@05Ri010I001@4L00<1
8P030I/0000800<16`002P6Q0J400@01XZ420@03RhZ;008100F5RhF;Q@020@04RhD1Q@<11XD30@>5
0`400hEVQ@030@050041T0001@4K00<18P030Ih0000500<17@003@6Q0HZR0J81XHX1RZ400`4018Z5
RXD30@03QHZ500<100I/QFb5K8D20@04JhD1Q@<100>5IXD00`401I010@6@00D16P040B40106Q0000
0`020Al001<1X@6Q00400J:Q0HZRRP65RhF;008100B5RhF;0P400hD1Q@03Q@<10hD30@03QFJ500<1
00D00@4100060AT00`4R00<1Y@0000<00P4N000B0J41RZ81XP6QRP6:XHX1Q@650P4018Z5RXD20@05
QFb50HD00`400hE[Q@020@04IXD1Q@@100B;0@6@1P4I00<18P030JP0000300030@0101X010403`01
00400J:Q0HZRRP6:XP020@04QH^5R`8100B5RXF:0`43Q@<100>5IXD010401I000@6@00D16@030B80
106Z00000`001040004F00<10P0030400J81XXX1RP6:X@8100N5XHD1QHZ5008100B5K8E/0P400hD1
Q@030@03QFJ500D100>@0@400Y050AP00`4Q00@1[P0000<000D100000@0B00<1100050400@010:41
X@6:XP6RRZ::X@6;0P400hF:Q@030@:50`400hEVQ@050@0500410@000`400i01000F00@18@030K80
000300030@00008000<1000030030@H001<1006RRP6:0HZQ0J41XHD1QJ65008100>5K8D00P400f^5
J`020@03Q@6500D100F@R`41T0060AH00`4R00<1]@0000<000<100000`000`40000800<12@005P6Q
0J41X@6QXP6R0J::0HZQRZ650HD20@:50`400hB50@050@05T0010I001@4F00<18P030KP000020003
0@0000D000<1000010030@/001D1X@6:0HX1RZ41X@6QQ@65X@6:QHX00P401f^50@5VQFH00P401Wl1
0H^@R`8100>@0@400P4F00<180000`400@020K/0000200030@0000H000<100000`4=000H0J41X@6Q
0J6R0J81XXX1RZ41XHD1Q@650P400hF4Q@050@0609010@2@104E00<18@040Kh0000200030@0000L0
0P4?000G0HX1RP6:0HZQ0J41XHD1QJ41RXD1Q8D00P400fJ5IP040@03T8/1008100>@0@400P4D00@1
8@030L80000200030@0000L000<100003P020@06X@6Q0J6R0`401j::0HZQ0J400P401XD1Q@64Q@@1
00>@0@401`4C00<18P030LD0000200030@0000L000<100003@020@05RP6:0HX00`401HFQQ@650081
00N:0HF40HB500D100>@0@401P4C00<18P030LP0000200030@0000L000<10000300020400@6Q0J6R
0P401hZR0@6:0HD00P400j41Q@03Q@8100<0T0400`4019010@4B00<18@040L/0000200030@0000L0
00<100002`0020400@6:0HZQ0P403HFQ0J650HD1Q8D1QH@00`401H/10I0100:@0`4A00<17P000`40
00030Ll0000200030@0000L000<100002P002`400@6Q0J6R0J::008100JQRP650HD50@03T04100<1
0Y020A400`4Q00<1dP00008000<100001`000`40000:00<100Z:0HZQ0J65X@6Q10401X@10@6;T081
00B@0@6@0`4?00@18@040M@0000200030@0000H000<100002P030@09X@6QXP6RRZ::008100BQ0J65
0P401@2@0@4000D13`030B40107H00000P000`40000600030@0000T00P402j810HZQ0J650HFQ00<1
00F50@6;00020@03T04100<13P030B400`7L00000P000`40000600030@0000P00P404P01006R0J::
0HX1Q@65X@40T08100<00@400`4=00<18@030Ml0000200030@0000H000<100001`020@03XP6R00<1
00BQQ@650P401HD10@40008100>@0@400`4<00<18@030N8000000`01000800030@0000H000D1X@41
00030@04XXX1RP8100NQQ@6@004000D12P040B00107U000000<00@0020000`40000600<100FR0@41
X@030@03RP65008100>;0@401@4900<18@040A400`7D000000<00@0020000`40000500<100L00@41
XP6Q008100F5X@40T0030@03T04100P00`4Q00<160000`40003A000000<00@0020000`40000400<1
00NR0@41X@6Q008100B40@400`4010010@4600<18@030A/000<10000d@000003004000P000<10000
0P040@050041RZ8010400hD1T0030@04T0410@D00`4P00@16`030M@000000`01000800040@0000L1
00>50@401@4018/10@030@<0104P00<17`000`40003D000000<00@0020000`400@070@03RP6:0081
00BQ0@2@0P400`010@0200<180040B8000<10000e0000003004000P020400hX10@050@040041T081
00@00@418@030BD0107C000000<00@001`080@03XP4100@100B@0@401@4P00@1o`000003004000H0
1`400hF:0@0<0B000`7o00@000000`010004010100>@0@400P4P00<1o`07000000<00@000`0C0B00
0`7o00X000000`40000700d17P040Ol03@0000030@0000L000<100000P060Ad0107o010000000`40
000700030@01008180030Ol0500000030@0000D0104P00<1E0000`400@2o000000<100000P030@80
00<100006`040EL000<10040_`0000030@0100811@000`40000H00<1FP001@400@010;h000020@P0
00<100005@030Ed000D100400@2n000000<00@001`000`40000B00<1H00010400@020Kd000020003
0@0000@000<100003`040Ol09`0000<000<100000`000`40000<00<1o`0[000010020@<000<10000
2@030Ol0;P0000H000@100011@000`4000030Ol0<@0000L000<1004010040Ol0=00000P00P401P01
0@400Ol0=P0000T00P7o03/0003o04H0003o04H0003o04H0003o04H0000=0081o`0g000030001040
007o03H0000<00040@000Ol0=P0000`000@10001o`0f000030001040007o03H0000<00040@000Ol0
=P0000d00P7o03L0003o04H00000\
\>"],
  ImageRangeCache->{{{0, 324.938}, {156.125, 0}} -> {0.0487576, 0.215351, 
  0.00348948, 0.00348948}}]
}, Open  ]]
}, Open  ]],

Cell["Optimization is next", "Text"],

Cell[CellGroupData[{

Cell["\<\

OPTDesign[p,Center->k,Radius->r,Ngrid->256, Nsmth -> 30 ];\
\>", "Input"],

Cell[BoxData[
    \("\nParameterization for internal stability:"\)], "Print"],

Cell[BoxData[
    \("  T = A + B * T1, where T1 is RHP stable and"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("A[s]="\[InvisibleSpace]\(\(16.0000000000000008`\[InvisibleSpace]\) + 
            32.0000000000000017`\ s\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4
            \),
      SequenceForm[ "A[s]=", 
        Times[ 
          Power[ 
            Plus[ 2.0, s], -4], 
          Plus[ 16.0, 
            Times[ 32.0, s]]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("B[s]="\[InvisibleSpace]s\^2\/\((\(2.`\[InvisibleSpace]\) + s)\)\^5\),
      
      SequenceForm[ "B[s]=", 
        Times[ 
          Power[ s, 2], 
          Power[ 
            Plus[ 2.0, s], -5]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("  "\)], "Print"],

Cell[BoxData[
    \("Processing performance function..."\)], "Print"],

Cell[BoxData[
    \("Possible error in input to OPTDesign: \n The radius function does not \
appear to have the asymptotic form r~const/w^n for w near Infinity.  Here n = \
dp + dc, the sum of the relative degrees of the plant and the compensator (or \
roll-off)."\)], "Print"],

Cell[BoxData[
    \("Sampling Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Smoothing Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Optimization routine Anopt now number-crunching ..."\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 7.0149724408267E+00 :   N/A   : 9.9E-01 : 0  E+00 :   N/A   : NON \
: 256"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.4607375994092E+00 : 1.8E+02 : 2.8E-05 : 0  E+00 : 5.6E-05 : NON \
: 256"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.460737599409172E+00"\)], "Print"],

Cell[BoxData[
    \("flat      = 2.7645066167E-05"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 5.64E-05"\)], "Print"],

Cell[BoxData[
    \("Calculating output functions ..."\)], "Print"],

Cell[BoxData[
    \("Resetting options ..."\)], "Print"],

Cell[BoxData[
    \("Done! "\)], "Print"]
}, Open  ]],

Cell["\<\
The Bode plots of the calculated closed loop are now produced:\
\>",
   "Text"],

Cell[CellGroupData[{

Cell["\<\
FigBodeMag  = BodeMagnitude[T];
FigBodePha  = BodePhase[T];\
\>", "Input"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.25 0.416584 0.192316 [
[.25 -0.0125 -11 -15 ]
[.25 -0.0125 11 0 ]
[.5 -0.0125 -5 -15 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -8 -15 ]
[.75 -0.0125 8 0 ]
[1 -0.0125 -11 -15 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -30 ]
[.5 -0.0125 29 -15 ]
[-0.0125 .03195 -12 -4.5 ]
[-0.0125 .03195 0 4.5 ]
[-0.0125 .12811 -24 -4.5 ]
[-0.0125 .12811 0 4.5 ]
[-0.0125 .22427 -12 -4.5 ]
[-0.0125 .22427 0 4.5 ]
[-0.0125 .32043 -24 -4.5 ]
[-0.0125 .32043 0 4.5 ]
[-0.0125 .41658 -6 -4.5 ]
[-0.0125 .41658 0 4.5 ]
[-0.0125 .51274 -18 -4.5 ]
[-0.0125 .51274 0 4.5 ]
[-0.0125 .6089 -6 -4.5 ]
[-0.0125 .6089 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40 -7.5 ]
[-0.01875 .30902 -24 7.5 ]
[0 .63053 -14 0 ]
[0 .63053 14 15 ]
[.25 .63053 -11 0 ]
[.25 .63053 11 15 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 15 ]
[.75 .63053 -8 0 ]
[.75 .63053 8 15 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.25 0 m
.25 .61803 L
s
.5 0 m
.5 .61803 L
s
.75 0 m
.75 .61803 L
s
1 0 m
1 .61803 L
s
0 .03195 m
1 .03195 L
s
0 .12811 m
1 .12811 L
s
0 .22427 m
1 .22427 L
s
0 .32043 m
1 .32043 L
s
0 .41658 m
1 .41658 L
s
0 .51274 m
1 .51274 L
s
0 g
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(100) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 0 m
.07526 .00375 L
s
.11928 0 m
.11928 .00375 L
s
.15051 0 m
.15051 .00375 L
s
.17474 0 m
.17474 .00375 L
s
.19454 0 m
.19454 .00375 L
s
.21127 0 m
.21127 .00375 L
s
.22577 0 m
.22577 .00375 L
s
.23856 0 m
.23856 .00375 L
s
.32526 0 m
.32526 .00375 L
s
.36928 0 m
.36928 .00375 L
s
.40051 0 m
.40051 .00375 L
s
.42474 0 m
.42474 .00375 L
s
.44454 0 m
.44454 .00375 L
s
.46127 0 m
.46127 .00375 L
s
.47577 0 m
.47577 .00375 L
s
.48856 0 m
.48856 .00375 L
s
.57526 0 m
.57526 .00375 L
s
.61928 0 m
.61928 .00375 L
s
.65051 0 m
.65051 .00375 L
s
.67474 0 m
.67474 .00375 L
s
.69454 0 m
.69454 .00375 L
s
.71127 0 m
.71127 .00375 L
s
.72577 0 m
.72577 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.82526 0 m
.82526 .00375 L
s
.86928 0 m
.86928 .00375 L
s
.90051 0 m
.90051 .00375 L
s
.92474 0 m
.92474 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.96127 0 m
.96127 .00375 L
s
.97577 0 m
.97577 .00375 L
s
.98856 0 m
.98856 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -34 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(H) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 15.000000 moveto
(\\220) show
105.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 15.000000 moveto
(L) show
117.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
0 .03195 m
.00625 .03195 L
s
[(-2)] -0.0125 .03195 1 0 Mshowa
0 .12811 m
.00625 .12811 L
s
[(-1.5)] -0.0125 .12811 1 0 Mshowa
0 .22427 m
.00625 .22427 L
s
[(-1)] -0.0125 .22427 1 0 Mshowa
0 .32043 m
.00625 .32043 L
s
[(-0.5)] -0.0125 .32043 1 0 Mshowa
0 .41658 m
.00625 .41658 L
s
[(0)] -0.0125 .41658 1 0 Mshowa
0 .51274 m
.00625 .51274 L
s
[(0.5)] -0.0125 .51274 1 0 Mshowa
0 .6089 m
.00625 .6089 L
s
[(1)] -0.0125 .6089 1 0 Mshowa
.125 Mabswid
0 .05118 m
.00375 .05118 L
s
0 .07042 m
.00375 .07042 L
s
0 .08965 m
.00375 .08965 L
s
0 .10888 m
.00375 .10888 L
s
0 .14734 m
.00375 .14734 L
s
0 .16657 m
.00375 .16657 L
s
0 .18581 m
.00375 .18581 L
s
0 .20504 m
.00375 .20504 L
s
0 .2435 m
.00375 .2435 L
s
0 .26273 m
.00375 .26273 L
s
0 .28196 m
.00375 .28196 L
s
0 .30119 m
.00375 .30119 L
s
0 .33966 m
.00375 .33966 L
s
0 .35889 m
.00375 .35889 L
s
0 .37812 m
.00375 .37812 L
s
0 .39735 m
.00375 .39735 L
s
0 .43582 m
.00375 .43582 L
s
0 .45505 m
.00375 .45505 L
s
0 .47428 m
.00375 .47428 L
s
0 .49351 m
.00375 .49351 L
s
0 .53197 m
.00375 .53197 L
s
0 .55121 m
.00375 .55121 L
s
0 .57044 m
.00375 .57044 L
s
0 .58967 m
.00375 .58967 L
s
0 .01272 m
.00375 .01272 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -101 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(dB) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
87.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 .61178 m
.25 .61803 L
s
gsave
.25 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 .61178 m
.75 .61803 L
s
gsave
.75 .63053 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 .61428 m
.07526 .61803 L
s
.11928 .61428 m
.11928 .61803 L
s
.15051 .61428 m
.15051 .61803 L
s
.17474 .61428 m
.17474 .61803 L
s
.19454 .61428 m
.19454 .61803 L
s
.21127 .61428 m
.21127 .61803 L
s
.22577 .61428 m
.22577 .61803 L
s
.23856 .61428 m
.23856 .61803 L
s
.32526 .61428 m
.32526 .61803 L
s
.36928 .61428 m
.36928 .61803 L
s
.40051 .61428 m
.40051 .61803 L
s
.42474 .61428 m
.42474 .61803 L
s
.44454 .61428 m
.44454 .61803 L
s
.46127 .61428 m
.46127 .61803 L
s
.47577 .61428 m
.47577 .61803 L
s
.48856 .61428 m
.48856 .61803 L
s
.57526 .61428 m
.57526 .61803 L
s
.61928 .61428 m
.61928 .61803 L
s
.65051 .61428 m
.65051 .61803 L
s
.67474 .61428 m
.67474 .61803 L
s
.69454 .61428 m
.69454 .61803 L
s
.71127 .61428 m
.71127 .61803 L
s
.72577 .61428 m
.72577 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.82526 .61428 m
.82526 .61803 L
s
.86928 .61428 m
.86928 .61803 L
s
.90051 .61428 m
.90051 .61803 L
s
.92474 .61428 m
.92474 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.96127 .61428 m
.96127 .61803 L
s
.97577 .61428 m
.97577 .61803 L
s
.98856 .61428 m
.98856 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .03195 m
1 .03195 L
s
.99375 .12811 m
1 .12811 L
s
.99375 .22427 m
1 .22427 L
s
.99375 .32043 m
1 .32043 L
s
.99375 .41658 m
1 .41658 L
s
.99375 .51274 m
1 .51274 L
s
.99375 .6089 m
1 .6089 L
s
.125 Mabswid
.99625 .05118 m
1 .05118 L
s
.99625 .07042 m
1 .07042 L
s
.99625 .08965 m
1 .08965 L
s
.99625 .10888 m
1 .10888 L
s
.99625 .14734 m
1 .14734 L
s
.99625 .16657 m
1 .16657 L
s
.99625 .18581 m
1 .18581 L
s
.99625 .20504 m
1 .20504 L
s
.99625 .2435 m
1 .2435 L
s
.99625 .26273 m
1 .26273 L
s
.99625 .28196 m
1 .28196 L
s
.99625 .30119 m
1 .30119 L
s
.99625 .33966 m
1 .33966 L
s
.99625 .35889 m
1 .35889 L
s
.99625 .37812 m
1 .37812 L
s
.99625 .39735 m
1 .39735 L
s
.99625 .43582 m
1 .43582 L
s
.99625 .45505 m
1 .45505 L
s
.99625 .47428 m
1 .47428 L
s
.99625 .49351 m
1 .49351 L
s
.99625 .53197 m
1 .53197 L
s
.99625 .55121 m
1 .55121 L
s
.99625 .57044 m
1 .57044 L
s
.99625 .58967 m
1 .58967 L
s
.99625 .01272 m
1 .01272 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.02223 .41689 Mdot
.09751 .4178 Mdot
.14156 .41931 Mdot
.17283 .42141 Mdot
.19711 .42408 Mdot
.21696 .42732 Mdot
.23377 .4311 Mdot
.24835 .43539 Mdot
.26123 .44017 Mdot
.27277 .44541 Mdot
.28324 .45107 Mdot
.29281 .45713 Mdot
.30164 .46354 Mdot
.30984 .47027 Mdot
.31749 .47726 Mdot
.32467 .48449 Mdot
.33143 .4919 Mdot
.33783 .49944 Mdot
.34391 .50708 Mdot
.3497 .51476 Mdot
.35522 .52243 Mdot
.36052 .53004 Mdot
.3656 .53754 Mdot
.37048 .54489 Mdot
.3752 .55202 Mdot
.37975 .55889 Mdot
.38415 .56544 Mdot
.38841 .57163 Mdot
.39255 .5774 Mdot
.39657 .5827 Mdot
.40049 .58748 Mdot
.40431 .59171 Mdot
.40803 .59533 Mdot
.41167 .5983 Mdot
.41522 .60061 Mdot
.41871 .60221 Mdot
.42212 .60312 Mdot
.42547 .60332 Mdot
.42876 .60284 Mdot
.43199 .60171 Mdot
.43517 .59999 Mdot
.4383 .59774 Mdot
.44138 .59506 Mdot
.44443 .59202 Mdot
.44743 .58873 Mdot
.45039 .58526 Mdot
.45332 .58167 Mdot
.45622 .57803 Mdot
.45909 .57436 Mdot
.46194 .57066 Mdot
.46475 .56693 Mdot
.46755 .56316 Mdot
.47032 .5593 Mdot
.47308 .55536 Mdot
.47582 .55131 Mdot
.47854 .54714 Mdot
.48125 .54288 Mdot
.48395 .53852 Mdot
.48664 .53409 Mdot
.48932 .52962 Mdot
.492 .52513 Mdot
.49467 .52065 Mdot
.49733 .51619 Mdot
.5 .51178 Mdot
.50267 .50742 Mdot
.50533 .50313 Mdot
.508 .49892 Mdot
.51068 .49479 Mdot
.51336 .49076 Mdot
.51605 .48682 Mdot
.51875 .48299 Mdot
.52146 .47926 Mdot
.52418 .47564 Mdot
.52692 .47211 Mdot
.52968 .46865 Mdot
.53245 .46524 Mdot
.53525 .46181 Mdot
.53806 .45831 Mdot
.54091 .45465 Mdot
.54378 .45073 Mdot
.54668 .44643 Mdot
.54961 .44166 Mdot
.55257 .4363 Mdot
.55557 .43031 Mdot
.55862 .42364 Mdot
.5617 .41633 Mdot
.56483 .40847 Mdot
.56801 .4002 Mdot
.57124 .39173 Mdot
.57453 .38331 Mdot
.57788 .37522 Mdot
.58129 .36775 Mdot
.58478 .36115 Mdot
.58833 .35566 Mdot
.59197 .35148 Mdot
.59569 .34873 Mdot
.59951 .34751 Mdot
.60343 .34781 Mdot
.60745 .34958 Mdot
.61159 .35265 Mdot
.61585 .35669 Mdot
.62025 .36118 Mdot
.6248 .36532 Mdot
.62952 .36794 Mdot
.6344 .36745 Mdot
.63948 .36176 Mdot
.64478 .34828 Mdot
.6503 .3239 Mdot
.65609 .28504 Mdot
.66217 .22769 Mdot
.66857 .14753 Mdot
.67533 .03992 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@0009@000<10000
7P000`40001X0000T`000`40000C00030@0000X000<10000I`0008D000<100402`00104000040@<0
0P40100100020@050040004010030@@000<10000I`0008D000<100402`000`40000200030@000080
00@100010P001040004300030@0000D000D100000@1Y0000Q0001@400@0100X000<100000P000`40
000300<10P0010400043008110020@@000<10000I`0008@000D100400@0:00030@00008000<10000
1@001040004200030@00008000D100000@0600030@0006L0002400040@0100812@001P40000100<1
0P020@@00`4400030@0000800`4300030@0006L0002D00030@0000l000<100000`000`4000060003
0@0006P0002U0081N@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@0006D00P4400030@000080104Y00@1;@040@<00P4W00@10`020@@00P4C0000I0001040004:0003
0@0002X000<10000;P001@400001008000<100009P001@400001008000@100010P000`40000@0000
I0001040004:00030@0002X000<10000;P001@400001008000<100009P001@400001008000@10001
0P000`40000@0000I0001040004:00030@0002X000<10000;P001@400001008000<100009P001@40
0001008000@100010P000`40000@0000I0001040004:00030@0002X000<10000;P001@4000010080
00<100009P001@400001008000@100010P000`40000@0000I0001040004800<1:P030Bh00`430004
0@000BH00`4300040@000@8000@100014P0006D00P4;00030@0002X000<10000;P000`4000020081
:@000`400002008110020A<0003o0240003o0240003o0240003o0240003o0240000i0<h16@0003T0
00<1000030000`40000600030@0000@000<100000P001@400001008000@100010P000`400@0?0003
0@0000H000<100000`000`40000200050@0000400P001040004200030@0100l000<100001P000`40
000300030@00008000D100000@0300060@000@010P000`40000<00030@0000H000<1000010000`40
000200050@0000400P001040004200030@0101T0000i00030@00030000<b0000<0000c80000a0003
<P00030000<100005`0003T00P4a0003<P00030000<b0000<@000c80000_00816@0003T000<10000
<0000c80000`0003<P00034000<b0000<0000`40000G0000<@040@@000<10000<0000c80000`0003
<P00034000<b0000<0000`40000G0000<@000`40000500030@00030000<b0000<0000c80000a0003
<P00030000<100005`00038000<1000010020LXb0P4I0000:`040@@000<100000`000`40000`0003
<P00030000<b00008@020@h000<b0000<0000`40000G0000=0000`40000200030@00030000<b0000
<0000c80000R00030@0000`000<b0000<0000`40000G0000<@001040004400030@00030000<b0000
<0000c80000a0003<P00030000<100005`0003800P450081<@000c80000`0003<P00034000<b0000
;`020AT0000i00030@00030000<b0000<0000c80000a0003<P00030000<100005`0003T000<10000
<0000c80000`0003<P00034000<b0000<0000`40000G0000>@000`40000`0003<P00030000<b0000
<@000c80000`00030@0001L0000i0081<@000c80000`0003<P00034000<b0000;`020AT0000i0003
0@00030000<b0000<0000c80000a0003<P00030000<100005`0003T000<10000<0000c80000`0003
<P00034000<b0000<0000`40000G0000>@000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001L0000i0081<@000c80000`0003<P00034000<b0000;`020AT0000i00030@00030000<b0000
<0000c80000a0003<P00030000<100005`0003T000<10000<0000c80000`0003<P00034000<b0000
<0000`40000G0000>@020C4000<b0000<0000c80000a0003<P0002l00P4I0000>@000`40000`0003
<P00030000<b0000<@000c80000`00030@0001L0000i00030@00030000<b0000<0000c80000a0003
<P00030000<100005`0002D0104300030@0000800`4500030@00030000<b0000<0000c80000a0003
<P00030000<100005`0002L000<100002P000`4000020081bS820AT0000W00030@0000X000<10000
0P000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000O00@110000`40000700<1
1@000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000W00030@0000L000<10000
1@000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000U00<12@000`4000050081
<@000c80000`0003<P0002000P4?0003<P0002l00P4I00009`000`40000700@110000`40000`0003
<P00030000<b00008@000`40000=0003<P00030000<100005`0003T000<10000<0000c80000`0003
<P00034000<b0000<0000`40000G0000>@000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001L0000i0081<@000c80000`0003<P00034000<b0000;`020AT0000i00030@00030000<b0000
<0000c80000a0003<P00030000<100005`0003T000<10000<0000c80000`0003<P00034000<b0000
<0000`40000G0000>@000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000i0081
<@000c80000`0003<P00034000<b0000;`020AT0000i00030@00030000<b0000<0000c80000a0003
<P00030000<100005`0003T000<10000<0000c80000`0003<P00034000<b0000<0000`40000G0000
>@000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000i0081<@000c80000`0003
<P00034000<b0000;`020AT0000i00030@00030000<b0000<0000c80000a0003<P00030000<10000
5`0003T000<10000<0000c80000`0003<P00034000<b0000<0000`40000G0000<@040@@000<10000
<0000c80000`0003<P00034000<b0000<0000`40000G0000<`000`4000030081QC820D<b0P4I0000
<`000`40000300030@00030000<b0000<0000c80000O00030@0000l000<b0000<0000`40000G0000
:`040@@000<100000`000`40000`0003<P00030000<b0000<@000c80000`00030@0001L0000c0003
0@0000<000<10000<0000c80000`0003<P00034000<b0000<0000`40000G0000<@030@D00P4a0003
<P00030000<b0000<@000c80000_00816@0003<000<100000`000`40000`0003<P00030000<b0000
<@000c80000`00030@0001L0000i00030@00030000<b0000<0000c80000a0003<P00030000<10000
5`0003T000<10000<0000c80000`0003<P00034000<b0000<0000`40000G0000>@020C4000<b0000
<0000c80000a0003<P0002l00P4I0000>@000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001L0000i00030@00030000<b0000<0000c80000a0003<P00030000<100005`0003T000<10000
<0000c80000`0003<P00034000<b0000<0000`40000G0000>@020C4000<b0000<0000c80000M0081
4P000c80000_00816@0003T000<10000<0000c80000`0003<P0001h000<1000040000c80000`0003
0@0001L0000i00030@00030000<b0000<0000c80000a0003<P00030000<100005`0001400P400`01
0@030B0000<10000<0000c80000`0003<P00034000<b0000<0000`40000G00004000104000420004
0@000Al00P4a0003<P00030000<b0000<@000c80000_00816@00010000@100010P001040004O0003
0@00030000<b0000<0000c80000a0003<P00030000<100005`00010000@100010P030B0000<10000
<0000c80000`0003<P00034000<b0000<0000`40000G00004@030@8000@1000130020@@000<10000
0P030@D000<10000<0000c80000`0003<P00034000<b0000<0000`40000G00004`000`400@030@`0
00@100012`000`4000020081Pc820DDb0P4I00004P020A4000@100012`000`40000200030@000300
00<b0000<0000c80000M00030@00014000<b0000<0000`40000G00007`040@8000@1000120030@D0
00<10000<0000c80000`0003<P00034000<b0000<0000`40000G00009@001040004800030@0000D0
00<10000<0000c80000`0003<P00034000<b0000<0000`40000G00009@001040004800030@0000D0
0P4a0003<P00030000<b0000<@000c80000_00816@0002H00P4900@110000`40000`0003<P000300
00<b00004@040@H00P4D0003<P00030000<100005`0003T000<10000<0000c80000`0003<P000100
1P4600030@00018000<b0000<0000`40000G0000>@000`40000`0003<P00030000<b00003`030@<0
0P4J0003<P00030000<100005`0003T00P4a0003<P00030000<b00003P030@D00P4200815@000c80
000_00816@0003T000<10000<0000c80000`0003<P0000h00P4700D15@000c80000`00030@0001L0
000i00030@00030000<b0000<0000c80000?00030@0000H00`4F0003<P00030000<100005`0003T0
00<10000<0000c80000`0003<P0000d00P4R0003<P00030000<100005`0003T00P4a0003<P000300
00<b000030030B8000<b0000;`020AT0000i00030@00030000<b0000<0000c80000=00030@000240
00<b0000<0000`40000G0000>@000`40000`0003<P00030000<b000030020B<000<b0000<0000`40
000G0000>@000`40000`0003<P00030000<b00003@000`40000Q0003<P00030000<100005`0003T0
0P4a0003<P00030000<b00002`020B@000<b0000;`020AT0000i00030@00030000<b0000<0000c80
000:00<190000c80000`00030@0001L0000b00811@000`40000`0003<P00030000<b00002`000`40
000S0003<P00030000<100005`00034000@100011000104000020@d00P4N0003<P00030000<b0000
2P020BD000<b0000<0000`40000G0000<@001040004400810c800`4b<P0<<P030C8b00Db0P45<P81
Cc800`4b<P1E<P816@00034000@1000110000`40000J00030@0000@000@100000P4200811@000c80
000`0003<P0000T00P4V0003<P00030000<100005`00034000@1000110000`40000V00060@000040
0P420003<P00030000<b000020030BH000<b0000<0000`40000G0000<@001040004400030@0002d0
00@10041<P000c80000800819`000c80000`00030@0001L0000b00811@020C4000@10041;`000c80
000700<19`000c80000_00816@0003T000<10000<0000c8000030Bd000<b00001`020BP000<b0000
<0000`40000G0000>@000`40000`0003<P0000800`4[0003<P0000H00`4X0003<P00030000<10000
5`0003T00P4a0003<P0000@00`4Y0003<P0000D00`4Y0003<P0002l00P4I0000>@000`40000`0003
<P0000H000<100009`000c80000400<1:P000c80000`00030@0001L0000i00030@00030000<b0000
1`020BL000<b000010030BX000<b0000<0000`40000G0000>@000`40000`0003<P0000P00`4U0003
<P000080104[0003<P00030000<100005`0003T00P4a0003<P0000X000<100008`000c80000200<1
;0000c80000_00816@0003T000<10000<0000c80000:00819000138000030Bd000<b0000<0000`40
000G0000>@000`40000`0003<P0000/00`4R0004<P000081;P000c80000`00030@0001L0000i0003
0@00030000<b00003@000`40000P0003<P0000<1;P000c80000`00030@0001L0000i0081<@000c80
000=00818@000c800@020Bl000<b0000;`020AT0000i00030@00030000<b00003P030Al000@b0041
<0000c80000`00030@0001L0000i00030@00030000<b000040000`40000M0004<P410C0000<b0000
<0000`40000G00009P020@@000<100000P030@D000<10000<0000c80000@00817P030C4000<b0000
<0000`40000G00009@001040004;00030@0000800P55<P817C820F@b0P4I00009@001040004;0003
0@00008000<10000<0000c80000B00030@0001X00`4b0003<P00030000<100005`0002D000@10001
20030@D000<10000<0000c80000B00816P030C<000<b0000<0000`40000G00009@00104000480003
0@0000D000<10000<0000c80000C00816@020@03<P00034000<b0000<0000`40000G00009@001040
004800030@0000D00P4a0003<P0001@000<100005P030@03<P00034000<b0000;`020AT0000V0081
2@040@@000<10000<0000c80000D00815`020@030380038000<b0000<0000`40000G0000>@000`40
000`0003<P0001D00P4E00<100<0<P00<P000c80000`00030@0001L0000i00030@00030000<b0000
5P000`40000C00810P000c80000a0003<P00030000<100005`0003T00P4a0003<P0001H00P4C00<1
0P000c80000a0003<P0002l00P4I0000>@000`40000`0003<P0001L00P4A00<10`000c80000a0003
<P00030000<100005`0003T000<10000<0000c80000H00030@0000h00`440003<P00034000<b0000
<0000`40000G0000>@000`40000`0003<P0001P00P4?00811@000c80000a0003<P00030000<10000
5`0003T00P4a0003<P0001T00P4=00<11@000c80000a0003<P0002l00P4I0000>@000`40000`0003
<P0001X00P4;00<11P000c80000a0003<P00030000<100005`0003T000<10000<0000c80000K0081
2@030@L000<b0000<@000c80000`00030@0001L0000i00030@00030000<b000070020@L00`480003
<P00034000<b0000<0000`40000G0000>@020C4000<b000070020@L00`480003<P00034000<b0000
;`020AT0000i00030@00030000<b00007@030@<010490003<P00034000<b0000<0000`40000G0000
>@000`40000`0003<P0001h0204:0003<P00034000<b0000<0000`40000G0000<@040@@000<10000
<0000c80000P00D12`000c80000a0003<P00030000<100005`0003<000<100000`020C4000<b0000
<0000c80000a0003<P0002l00P4I0000<`000`40000300030@0000`000<100001P000`4000040003
0@00008000D100000@0200040@000@8000<100403`000`40000600030@0000<000<100000P001@40
0001008000@100010P000`400@0?00030@0000H000<100000`000`40000200050@0000400`001P40
00400@8000<1000030000`40000600030@0000@000<100000P001@400001008000@100010P000`40
0@0I0000<`000`4000030<h16@0003<000<10000jP0003400`7/0000<`000`40003Z0000o`0Q0000
o`0Q0000o`0Q0000;`020@@000<100000`020@<0104Q008110000`40000200@1:@040Bd010430081
B@0002h000@1000120001040004400030@0001l000@100012P000`40000Z00030@0002h000D10000
0@0200030@0004H0000^00040@000@P000@1000110000`40000O00040@000@X000<10000:P000`40
000^00050@0000400P000`4000160000;P001040004800040@000@@000<100007`001040004:0003
0@0002X000<10000;P001@400001008000<10000AP0002h000@1000120001040004400030@0001l0
00@100012P000`40000Z00030@0002h000D100000@0200030@0004H0000^00040@000@P000@10001
0P030B4000@1000120030BX00`4^00<10`00104000580000;`020@X00P4500030@0002000P4;0003
0@0002X000<10000;P000`4000020081B@000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-3.11774, -2.99144, 
  0.0194952, 0.0253426}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.25 0.603319 0.0032227 [
[.25 -0.0125 -11 -15 ]
[.25 -0.0125 11 0 ]
[.5 -0.0125 -5 -15 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -8 -15 ]
[.75 -0.0125 8 0 ]
[1 -0.0125 -11 -15 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -30 ]
[.5 -0.0125 29 -15 ]
[-0.0125 .03935 -24 -4.5 ]
[-0.0125 .03935 0 4.5 ]
[-0.0125 .11991 -24 -4.5 ]
[-0.0125 .11991 0 4.5 ]
[-0.0125 .20048 -24 -4.5 ]
[-0.0125 .20048 0 4.5 ]
[-0.0125 .28105 -24 -4.5 ]
[-0.0125 .28105 0 4.5 ]
[-0.0125 .36162 -18 -4.5 ]
[-0.0125 .36162 0 4.5 ]
[-0.0125 .44218 -18 -4.5 ]
[-0.0125 .44218 0 4.5 ]
[-0.0125 .52275 -18 -4.5 ]
[-0.0125 .52275 0 4.5 ]
[-0.0125 .60332 -6 -4.5 ]
[-0.0125 .60332 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -46 -7.5 ]
[-0.01875 .30902 -24 7.5 ]
[0 .63053 -14 0 ]
[0 .63053 14 15 ]
[.25 .63053 -11 0 ]
[.25 .63053 11 15 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 15 ]
[.75 .63053 -8 0 ]
[.75 .63053 8 15 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.25 0 m
.25 .61803 L
s
.5 0 m
.5 .61803 L
s
.75 0 m
.75 .61803 L
s
1 0 m
1 .61803 L
s
0 .03935 m
1 .03935 L
s
0 .11991 m
1 .11991 L
s
0 .20048 m
1 .20048 L
s
0 .28105 m
1 .28105 L
s
0 .36162 m
1 .36162 L
s
0 .44218 m
1 .44218 L
s
0 .52275 m
1 .52275 L
s
0 g
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(100) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 0 m
.07526 .00375 L
s
.11928 0 m
.11928 .00375 L
s
.15051 0 m
.15051 .00375 L
s
.17474 0 m
.17474 .00375 L
s
.19454 0 m
.19454 .00375 L
s
.21127 0 m
.21127 .00375 L
s
.22577 0 m
.22577 .00375 L
s
.23856 0 m
.23856 .00375 L
s
.32526 0 m
.32526 .00375 L
s
.36928 0 m
.36928 .00375 L
s
.40051 0 m
.40051 .00375 L
s
.42474 0 m
.42474 .00375 L
s
.44454 0 m
.44454 .00375 L
s
.46127 0 m
.46127 .00375 L
s
.47577 0 m
.47577 .00375 L
s
.48856 0 m
.48856 .00375 L
s
.57526 0 m
.57526 .00375 L
s
.61928 0 m
.61928 .00375 L
s
.65051 0 m
.65051 .00375 L
s
.67474 0 m
.67474 .00375 L
s
.69454 0 m
.69454 .00375 L
s
.71127 0 m
.71127 .00375 L
s
.72577 0 m
.72577 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.82526 0 m
.82526 .00375 L
s
.86928 0 m
.86928 .00375 L
s
.90051 0 m
.90051 .00375 L
s
.92474 0 m
.92474 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.96127 0 m
.96127 .00375 L
s
.97577 0 m
.97577 .00375 L
s
.98856 0 m
.98856 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -34 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(H) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 15.000000 moveto
(\\220) show
105.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 15.000000 moveto
(L) show
117.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
0 .03935 m
.00625 .03935 L
s
[(-175)] -0.0125 .03935 1 0 Mshowa
0 .11991 m
.00625 .11991 L
s
[(-150)] -0.0125 .11991 1 0 Mshowa
0 .20048 m
.00625 .20048 L
s
[(-125)] -0.0125 .20048 1 0 Mshowa
0 .28105 m
.00625 .28105 L
s
[(-100)] -0.0125 .28105 1 0 Mshowa
0 .36162 m
.00625 .36162 L
s
[(-75)] -0.0125 .36162 1 0 Mshowa
0 .44218 m
.00625 .44218 L
s
[(-50)] -0.0125 .44218 1 0 Mshowa
0 .52275 m
.00625 .52275 L
s
[(-25)] -0.0125 .52275 1 0 Mshowa
0 .60332 m
.00625 .60332 L
s
[(0)] -0.0125 .60332 1 0 Mshowa
.125 Mabswid
0 .05546 m
.00375 .05546 L
s
0 .07157 m
.00375 .07157 L
s
0 .08769 m
.00375 .08769 L
s
0 .1038 m
.00375 .1038 L
s
0 .13603 m
.00375 .13603 L
s
0 .15214 m
.00375 .15214 L
s
0 .16825 m
.00375 .16825 L
s
0 .18437 m
.00375 .18437 L
s
0 .21659 m
.00375 .21659 L
s
0 .23271 m
.00375 .23271 L
s
0 .24882 m
.00375 .24882 L
s
0 .26494 m
.00375 .26494 L
s
0 .29716 m
.00375 .29716 L
s
0 .31328 m
.00375 .31328 L
s
0 .32939 m
.00375 .32939 L
s
0 .3455 m
.00375 .3455 L
s
0 .37773 m
.00375 .37773 L
s
0 .39384 m
.00375 .39384 L
s
0 .40996 m
.00375 .40996 L
s
0 .42607 m
.00375 .42607 L
s
0 .4583 m
.00375 .4583 L
s
0 .47441 m
.00375 .47441 L
s
0 .49052 m
.00375 .49052 L
s
0 .50664 m
.00375 .50664 L
s
0 .53886 m
.00375 .53886 L
s
0 .55498 m
.00375 .55498 L
s
0 .57109 m
.00375 .57109 L
s
0 .58721 m
.00375 .58721 L
s
0 .02323 m
.00375 .02323 L
s
0 .00712 m
.00375 .00712 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -107 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(deg) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
87.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 .61178 m
.25 .61803 L
s
gsave
.25 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 .61178 m
.75 .61803 L
s
gsave
.75 .63053 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 .61428 m
.07526 .61803 L
s
.11928 .61428 m
.11928 .61803 L
s
.15051 .61428 m
.15051 .61803 L
s
.17474 .61428 m
.17474 .61803 L
s
.19454 .61428 m
.19454 .61803 L
s
.21127 .61428 m
.21127 .61803 L
s
.22577 .61428 m
.22577 .61803 L
s
.23856 .61428 m
.23856 .61803 L
s
.32526 .61428 m
.32526 .61803 L
s
.36928 .61428 m
.36928 .61803 L
s
.40051 .61428 m
.40051 .61803 L
s
.42474 .61428 m
.42474 .61803 L
s
.44454 .61428 m
.44454 .61803 L
s
.46127 .61428 m
.46127 .61803 L
s
.47577 .61428 m
.47577 .61803 L
s
.48856 .61428 m
.48856 .61803 L
s
.57526 .61428 m
.57526 .61803 L
s
.61928 .61428 m
.61928 .61803 L
s
.65051 .61428 m
.65051 .61803 L
s
.67474 .61428 m
.67474 .61803 L
s
.69454 .61428 m
.69454 .61803 L
s
.71127 .61428 m
.71127 .61803 L
s
.72577 .61428 m
.72577 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.82526 .61428 m
.82526 .61803 L
s
.86928 .61428 m
.86928 .61803 L
s
.90051 .61428 m
.90051 .61803 L
s
.92474 .61428 m
.92474 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.96127 .61428 m
.96127 .61803 L
s
.97577 .61428 m
.97577 .61803 L
s
.98856 .61428 m
.98856 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .03935 m
1 .03935 L
s
.99375 .11991 m
1 .11991 L
s
.99375 .20048 m
1 .20048 L
s
.99375 .28105 m
1 .28105 L
s
.99375 .36162 m
1 .36162 L
s
.99375 .44218 m
1 .44218 L
s
.99375 .52275 m
1 .52275 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .05546 m
1 .05546 L
s
.99625 .07157 m
1 .07157 L
s
.99625 .08769 m
1 .08769 L
s
.99625 .1038 m
1 .1038 L
s
.99625 .13603 m
1 .13603 L
s
.99625 .15214 m
1 .15214 L
s
.99625 .16825 m
1 .16825 L
s
.99625 .18437 m
1 .18437 L
s
.99625 .21659 m
1 .21659 L
s
.99625 .23271 m
1 .23271 L
s
.99625 .24882 m
1 .24882 L
s
.99625 .26494 m
1 .26494 L
s
.99625 .29716 m
1 .29716 L
s
.99625 .31328 m
1 .31328 L
s
.99625 .32939 m
1 .32939 L
s
.99625 .3455 m
1 .3455 L
s
.99625 .37773 m
1 .37773 L
s
.99625 .39384 m
1 .39384 L
s
.99625 .40996 m
1 .40996 L
s
.99625 .42607 m
1 .42607 L
s
.99625 .4583 m
1 .4583 L
s
.99625 .47441 m
1 .47441 L
s
.99625 .49052 m
1 .49052 L
s
.99625 .50664 m
1 .50664 L
s
.99625 .53886 m
1 .53886 L
s
.99625 .55498 m
1 .55498 L
s
.99625 .57109 m
1 .57109 L
s
.99625 .58721 m
1 .58721 L
s
.99625 .02323 m
1 .02323 L
s
.99625 .00712 m
1 .00712 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.02223 .60332 Mdot
.09751 .60331 Mdot
.14156 .60329 Mdot
.17283 .60326 Mdot
.19711 .6032 Mdot
.21696 .60312 Mdot
.23377 .603 Mdot
.24835 .60285 Mdot
.26123 .60265 Mdot
.27277 .60241 Mdot
.28324 .60212 Mdot
.29281 .60177 Mdot
.30164 .60137 Mdot
.30984 .6009 Mdot
.31749 .60037 Mdot
.32467 .59978 Mdot
.33143 .59911 Mdot
.33783 .59838 Mdot
.34391 .59757 Mdot
.3497 .59668 Mdot
.35522 .59573 Mdot
.36052 .5947 Mdot
.3656 .59359 Mdot
.37048 .59241 Mdot
.3752 .59116 Mdot
.37975 .58983 Mdot
.38415 .58844 Mdot
.38841 .58697 Mdot
.39255 .58544 Mdot
.39657 .58385 Mdot
.40049 .58221 Mdot
.40431 .5805 Mdot
.40803 .57876 Mdot
.41167 .57697 Mdot
.41522 .57515 Mdot
.41871 .5733 Mdot
.42212 .57144 Mdot
.42547 .56957 Mdot
.42876 .5677 Mdot
.43199 .56585 Mdot
.43517 .56401 Mdot
.4383 .5622 Mdot
.44138 .56042 Mdot
.44443 .55867 Mdot
.44743 .55696 Mdot
.45039 .55528 Mdot
.45332 .55362 Mdot
.45622 .55198 Mdot
.45909 .55036 Mdot
.46194 .54874 Mdot
.46475 .54712 Mdot
.46755 .54549 Mdot
.47032 .54386 Mdot
.47308 .54223 Mdot
.47582 .54058 Mdot
.47854 .53894 Mdot
.48125 .53728 Mdot
.48395 .53563 Mdot
.48664 .53397 Mdot
.48932 .5323 Mdot
.492 .53063 Mdot
.49467 .52895 Mdot
.49733 .52725 Mdot
.5 .52554 Mdot
.50267 .52381 Mdot
.50533 .52206 Mdot
.508 .52028 Mdot
.51068 .51847 Mdot
.51336 .51662 Mdot
.51605 .51473 Mdot
.51875 .51279 Mdot
.52146 .5108 Mdot
.52418 .50875 Mdot
.52692 .50664 Mdot
.52968 .50444 Mdot
.53245 .50216 Mdot
.53525 .49979 Mdot
.53806 .49731 Mdot
.54091 .49473 Mdot
.54378 .49202 Mdot
.54668 .4892 Mdot
.54961 .48627 Mdot
.55257 .48322 Mdot
.55557 .48007 Mdot
.55862 .47683 Mdot
.5617 .47351 Mdot
.56483 .47013 Mdot
.56801 .46668 Mdot
.57124 .46317 Mdot
.57453 .45959 Mdot
.57788 .45593 Mdot
.58129 .45217 Mdot
.58478 .44826 Mdot
.58833 .44416 Mdot
.59197 .43982 Mdot
.59569 .43515 Mdot
.59951 .4301 Mdot
.60343 .42457 Mdot
.60745 .41846 Mdot
.61159 .41165 Mdot
.61585 .40402 Mdot
.62025 .39541 Mdot
.6248 .38567 Mdot
.62952 .37461 Mdot
.6344 .36204 Mdot
.63948 .3478 Mdot
.64478 .33171 Mdot
.6503 .31366 Mdot
.65609 .29354 Mdot
.66217 .27134 Mdot
.66857 .24706 Mdot
.67533 .22078 Mdot
.68251 .19266 Mdot
.69016 .16287 Mdot
.69836 .13167 Mdot
.70719 .09933 Mdot
.71676 .06618 Mdot
.72723 .03257 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@0009L000<10000
7P000`40001U0000UP000`40000C00030@0000X000<10000I00008P000<100402`00104000040@<0
0P40100100020@050040004010030@@000<10000I00008P000<100402`000`40000200030@000080
00@100010P001040004300030@0000D000D100000@1V0000Q`001@400@0100X000<100000P000`40
000300<10P0010400043008110020@@000<10000I00008L000D100400@0:00030@00008000<10000
1@001040004200030@00008000D100000@0600030@0006@0002700040@0100812@001P40000100<1
0P020@@00`4400030@0000800`4300030@0006@0002G00030@0000l000<100000`000`4000060003
0@0006D0002X0081MP000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@0006P00P4400030@000080104Y00@1;@040@<00P4W00@10`020@@00P4@0000I`001040004:0003
0@0002X000<10000;P001@400001008000<100009P001@400001008000@100010P000`40000=0000
I`001040004:00030@0002X000<10000;P001@400001008000<100009P001@400001008000@10001
0P000`40000=0000I`001040004:00030@0002X000<10000;P001@400001008000<100009P001@40
0001008000@100010P000`40000=0000I`001040004:00030@0002X000<10000;P001@4000010080
00<100009P001@400001008000@100010P000`40000=0000I`001040004800<1:P030Bh00`430004
0@000BH00`4300040@000@8000@100013`0006P00P4;00030@0002X000<10000;P000`4000020081
:@000`400002008110020A00003o0240003o0240003o0240003o0240003o0240000l0<h15P0003`0
00<1000030000`40000600030@0000@000<100000P001@400001008000@100010P000`400@0?0003
0@0000H000<100000`000`40000200050@0000400P001040004200030@0100l000<100001P000`40
000300030@00008000D100000@0300060@000@010P000`40000<00030@0000H000<1000010000`40
000200050@0000400P001040004200030@0101H0000l0081<@000c80000`0003<P00034000<b0000
;`020AH0000l00030@00030000<b0000<0000c80000a0003<P00030000<10000500003`000<10000
<0000c80000`0003<P00034000<b0000<0000`40000D0000?0020C4000<b0000<0000c80000a0003
<P0002l00P4F0000?0000`40000`0003<P00030000<b0000;0020@<000<b0000<0000`40000D0000
:0040@<000<100000P030@D000<10000<0000c80000`0003<P0002d000D10000<P0b00030@0001@0
000Z00030@00008000<100001@000`4000020081bS820AH0000Z00030@0000<000<1000010000`40
000200030@00030000<b0000<0000c80000a0003<P00030000<1000050000280104400030@0000<0
00@100000`4500030@00030000<b0000<0000c80000a0003<P00030000<10000500002X000<10000
10001040004700030@00030000<b0000<0000c80000a0003<P00030000<10000500002P00`430004
0@000@8000<100001@020C4000<b0000<0000c80000a0003<P0002l00P4F0000:P00104000040@80
104400030@00030000<b0000<0000c80000Y00811P000c80000`00030@0001@0000l00030@000300
00<b0000<0000c80000Z00030@0000@000<b0000<0000`40000D0000?0020C4000<b0000<0000c80
000a0003<P0002l00P4F0000?0000`40000`0003<P00030000<b0000<@000c80000`00030@0001@0
000l00030@00030000<b0000<0000c80000a0003<P00030000<10000500003`00P4a0003<P000300
00<b0000<@000c80000_00815P0003`000<10000<0000c80000`0003<P00034000<b0000<0000`40
000D0000?0000`40000`0003<P00030000<b00009`020@P000<b0000<0000`40000D0000?0020C40
00<b0000<0000c80000X00030@0000H000<b0000;`020AH0000l00030@00030000<b0000<0000c80
000a0003<P00030000<10000500002P0104200<110020@D000<10000<0000c80000`0003<P000340
00<b0000<0000`40000D0000:P000`40000400040@000@8000<100000P000`40000`0003<P000300
00<b0000<@000c80000`00030@0001@0000Z00030@0000@000@100010P000`4000020081bS820AH0
000R00@11000104000030@<000@1000110000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001@0000Z00050@0000401@001040004400030@00030000<b0000<0000c80000V00812@000c80
000`00030@0001@0000X00<10`000`40000300040@000@@00P4a0003<P00030000<b00009`000`40
00070003<P0002l00P4F0000:P00104000040@<00P4500030@00030000<b0000<0000c80000a0003
<P00030000<10000500003`000<10000<0000c80000`0003<P00034000<b0000<0000`40000D0000
?0020C4000<b0000<0000c80000a0003<P0002l00P4F0000?0000`40000`0003<P00030000<b0000
<@000c80000`00030@0001@0000l00030@00030000<b0000<0000c80000T00812`000c80000`0003
0@0001@0000l00030@00030000<b0000<0000c80000U00030@0000T000<b0000<0000`40000D0000
?0020C4000<b0000<0000c80000a0003<P0002l00P4F0000?0000`40000`0003<P00030000<b0000
<@000c80000`00030@0001@0000l00030@00030000<b0000<0000c80000a0003<P00030000<10000
500003`00P4a0003<P00030000<b0000<@000c80000_00815P0003`000<10000<0000c80000`0003
<P0002800P4=0003<P00030000<10000500002P0104200@10P030@D000<10000<0000c80000`0003
<P0002<000<100002`000c80000`00030@0001@0000Z00050@00004020000`4000020081bS820AH0
000Z00030@00008000<100001@000`40000200030@00030000<b0000<0000c80000a0003<P000300
00<1000050000280104400030@0000<000@100000`4500030@00030000<b0000<0000c80000a0003
<P00030000<10000500002X000<1000010001040004700030@00030000<b0000<0000c80000a0003
<P00030000<10000500002P00`4300040@000@8000<100001@020C4000<b0000<0000c80000Q0081
3P000c80000_00815P0002X000<100000P020@<0104400030@00030000<b0000<0000c80000R0003
0@0000`000<b0000<0000`40000D0000?0000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001@0000l0081<@000c80000`0003<P00034000<b0000;`020AH0000l00030@00030000<b0000
<0000c80000a0003<P00030000<10000500003`000<10000<0000c80000`0003<P0002000P4?0003
<P00030000<10000500003`00P4a0003<P00030000<b00008@000`40000=0003<P0002l00P4F0000
?0000`40000`0003<P00030000<b0000<@000c80000`00030@0001@0000l00030@00030000<b0000
<0000c80000a0003<P00030000<10000500003`00P4a0003<P00030000<b0000<@000c80000_0081
5P0003`000<10000<0000c80000`0003<P0001h00P4A0003<P00030000<10000500002P010430081
10020@D000<10000<0000c80000`0003<P0001l000<100003`000c80000`00030@0001@0000Z0005
0@0000400P001040004200030@00008000<10000<0000c80000`0003<P00034000<b0000<0000`40
000D0000:P001@400001008000@100010P000`4000020081bS820AH0000I00<11P040@@000D10000
0@0200040@000@8000<100000P000`40000`0003<P00030000<b0000<@000c80000`00030@0001@0
000L00030@0000/000D100000@0200040@000@8000<100000P000`40000`0003<P00030000<b0000
7@020A8000<b0000<0000`40000D00003P020@04004000<10`030@/00`4300040@000@8000@10001
10020C4000<b0000<0000c80000N00030@00010000<b0000;`020AH0000=00040@000@8000<10000
0`001040004=00030@0000800P4400811@000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001@0000=00040@000@800`4300040@000Al000<10000<0000c80000`0003<P00034000<b0000
<0000`40000D00003@001040004200040@000@8000@100017`020C4000<b0000<0000c80000L0081
4`000c80000_00815P0000h00`43008110020@03004001d000<10000<0000c80000`0003<P0001d0
00<100004@000c80000`00030@0001@0000@00030@0002T000<10000<0000c80000`0003<P000340
00<b0000<0000`40000D00003`020B/000<10000<0000c80000`0003<P00034000<b0000<0000`40
000D0000?0020C4000<b0000<0000c80000K008150000c80000_00815P0003`000<10000<0000c80
000`0003<P0001`000<100004P000c80000`00030@0001@0000l00030@00030000<b0000<0000c80
000a0003<P00030000<10000500003`00P4a0003<P00030000<b00006P020AD000<b0000;`020AH0
000l00030@00030000<b0000<0000c80000K00030@0001<000<b0000<0000`40000D0000;`000`40
000200<11@000`40000`0003<P00030000<b0000<@000c80000`00030@0001@0000_00030@0000D0
00<100000P020H0b0P58<P815P00030000<1000010000`40000200030@00030000<b0000<0000c80
000J00030@0001@000<b0000<0000`40000D0000:0040@@000@100000`4500030@00030000<b0000
<0000c80000H00815`000c80000`00030@0001@0000a00040@000@L000<10000<0000c80000`0003
<P0001T000<100005@000c80000`00030@0001@0000^00040@000@8000<100001@020C4000<b0000
<0000c80000a0003<P0002l00P4F0000;P040@80104400030@00030000<b0000<0000c80000G0081
60000c80000`00030@0001@0000l00030@00030000<b0000<0000c80000H00030@0001H000<b0000
<0000`40000D0000?0020C4000<b0000<0000c80000F00816@000c80000_00815P0003`000<10000
<0000c80000`0003<P0001L000<100005`000c80000`00030@0001@0000l00030@00030000<b0000
<0000c80000E00816P000c80000`00030@0001@0000l0081<@000c80000`0003<P0001@00`4J0003
<P0002l00P4F0000?0000`40000`0003<P00030000<b00004`030A/000<b0000<0000`40000D0000
?0000`40000`0003<P00030000<b000050000`40000J0003<P00030000<10000500003`000<10000
<0000c80000`0003<P0001800P4M0003<P00030000<10000500003`00P4a0003<P00030000<b0000
4@030Ad000<b0000;`020AH0000l00030@00030000<b0000<0000c80000A00817P000c80000`0003
0@0001@0000^00<110020@D000<10000<0000c80000`0003<P0001000`4N0003<P00030000<10000
5000034000@100010P000`4000020081MS830E4b0P4F0000<@001040004200030@00008000<10000
<0000c80000`0003<P0000h00`4P0003<P00030000<10000500002P0104200<10`00104000440003
0@00030000<b0000<0000c80000=00<18@000c80000`00030@0001@0000^00030@0000<000@10001
10020C4000<b0000<0000c80000<00<18P000c80000_00815P0002h000<100000`00104000440003
0@00030000<b0000<0000c80000;00<18`000c80000`00030@0001@0000^00@10`020@D000<10000
<0000c80000`0003<P0000X0104S0003<P00030000<10000500003`00P4a0003<P00030000<b0000
2@030BD000<b0000;`020AH0000l00030@00030000<b0000<0000c80000800@19@000c80000`0003
0@0001@0000l00030@00030000<b0000<0000c80000700<19`000c80000`00030@0001@0000l0003
0@00030000<b0000<0000c80000700<19`000c80000`00030@0001@0000l0081<@000c80000`0003
<P0000D0104X0003<P0002l00P4F0000?0000`40000`0003<P00030000<b000010040BT000<b0000
<0000`40000D0000?0000`40000`0003<P00030000<b00000`040BX000<b0000<0000`40000D0000
?0020C4000<b0000<000138000050B/000<b0000;`020AH0000l00030@00030000<b0000<0000c80
00040Bd000<b0000<0000`40000D0000;P040@800`4500030@00030000<b0000<0000c800@030Bh0
00<b0000<0000`40000D0000;P000`40000600030@0000800P5T<PD1HC820AH0000_00030@0000D0
00<100000P000`40000`0003<P0002l0104a0003<P00030000<10000500002P0104400040@0000<1
1@000`40000`0003<P0002d0104c0003<P00030000<100005000034000@100011`000`40000`0003
<P0002`010400c80000a0003<P00030000<10000500002h000@100010P000`4000050081<@000c80
000Z00D100<0<P00<P000c80000_00815P0002l00P4300@110000`40000`0003<P0002P01@430003
<P00034000<b0000<0000`40000D0000?0000`40000`0003<P0002H01@450003<P00034000<b0000
<0000`40000D0000?0020C4000<b00009@050@H000<b0000<@000c80000_00815P0003`000<10000
<0000c80000S00D120000c80000a0003<P00030000<10000500003`000<10000<0000c80000Q00D1
2P000c80000a0003<P00030000<10000500003`00P4a0003<P0001l01@4<0003<P00034000<b0000
;`020AH0000l00030@00030000<b00007@050@h000<b0000<@000c80000`00030@0001@0000l0003
0@00030000<b00006`060@l000<b0000<@000c80000`00030@0001@0000l00030@00030000<b0000
60060A8000<b0000<@000c80000`00030@0001@0000l0081<@000c80000D00P150000c80000a0003
<P0002l00P4F0000?0000`40000`0003<P0000l02P4G0003<P00034000<b0000<0000`40000D0000
=@020@D000@100000P4=00811`020@D00P4300810P020@040041008100<00@403P400`010@030A/0
00<b0000<@000c80000`00030@0001@0000d00040@000@@00P4300030@0000`000<100001P000`40
000400030@00008000D100000@0200040@000@8000X100400@0100400P400`010@0Q0003<P000340
00<b0000;`020AH0000d00040@000@@000<10000<0000c80000`0003<P00034000<b0000<0000`40
000D0000=0001040004400030@0000`000<100001P000`40000400030@00008000D100000@020004
0@000@8000<100403`000`40000600030@0000<000<100000P001@400001008000@100010P000`40
0@0?00030@0000H000<100000`000`40000200050@0000400`001P4000400@8000<1000030000`40
000600030@0000@000<100000P001@400001008000@100010P000`400@0F0000=000104000440<h1
5P0003@000@10001j00003D00P7Y0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000<P020@@000<10000
0`020@<0104Q008110000`40000200@1:@040Bd010430081AP00034000@100012000104000440003
0@0001l000@100012P000`40000Z00030@0002h000D100000@0200030@0004<0000a00040@000@P0
00@1000110000`40000O00040@000@X000<10000:P000`40000^00050@0000400P000`4000130000
<@001040004800040@000@@000<100007`001040004:00030@0002X000<10000;P001@4000010080
00<10000@`00034000@1000120001040004400030@0001l000@100012P000`40000Z00030@0002h0
00D100000@0200030@0004<0000a00040@000@P000@100010P030B4000@1000120030BX00`4^00<1
0`00104000550000<P020@X00P4500030@0002000P4;00030@0002X000<10000;P000`4000020081
AP000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-3.17622, -236.459, 
  0.0194952, 1.51233}}]
}, Open  ]]
}, Open  ]],

Cell["The rational approximation of T is ", "Text"],

Cell[CellGroupData[{

Cell["\<\

Trat[s_]    = RationalModel[s,DegreeOfDenominator -> 7]\
\>", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]5.73172810635601859`\),
      SequenceForm[ "Error = ", 5.7317281063560186],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(16.0000000000000008`\[InvisibleSpace]\) + 
          32.0000000000000017`\ s\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4 + 
      \(s\^2\ \((
          \(45.3564605093607831`\[InvisibleSpace]\) + 
            78.8229561794683242`\ s + 11.0015775334985876`\ s\^2)
          \)\)\/\(\((\(2.`\[InvisibleSpace]\) + s)\)\^5\ 
          \((\(3.31431116956753291`\[InvisibleSpace]\) + 
              0.614396497166845456`\ s + 0.0712923332656223429`\ s\^2)\)\)\)],
   "Output"]
}, Open  ]],

Cell["The corresponding step response is produced with", "Text"],

Cell["\<\

step1[t_]   = Chop[Simplify[InverseLaplaceTransform[Trat[s]/s,s,t]]];
\
\>", "Input"],

Cell[CellGroupData[{

Cell["FigStepT1a  = Plot[ step1[t] ,{t,0,14},PlotRange -> All];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0680272 0.0147151 0.522455 [
[.15986 .00222 -3 -9 ]
[.15986 .00222 3 0 ]
[.29592 .00222 -3 -9 ]
[.29592 .00222 3 0 ]
[.43197 .00222 -3 -9 ]
[.43197 .00222 3 0 ]
[.56803 .00222 -3 -9 ]
[.56803 .00222 3 0 ]
[.70408 .00222 -6 -9 ]
[.70408 .00222 6 0 ]
[.84014 .00222 -6 -9 ]
[.84014 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .11921 -18 -4.5 ]
[.01131 .11921 0 4.5 ]
[.01131 .2237 -18 -4.5 ]
[.01131 .2237 0 4.5 ]
[.01131 .32819 -18 -4.5 ]
[.01131 .32819 0 4.5 ]
[.01131 .43268 -18 -4.5 ]
[.01131 .43268 0 4.5 ]
[.01131 .53717 -6 -4.5 ]
[.01131 .53717 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15986 .01472 m
.15986 .02097 L
s
[(2)] .15986 .00222 0 1 Mshowa
.29592 .01472 m
.29592 .02097 L
s
[(4)] .29592 .00222 0 1 Mshowa
.43197 .01472 m
.43197 .02097 L
s
[(6)] .43197 .00222 0 1 Mshowa
.56803 .01472 m
.56803 .02097 L
s
[(8)] .56803 .00222 0 1 Mshowa
.70408 .01472 m
.70408 .02097 L
s
[(10)] .70408 .00222 0 1 Mshowa
.84014 .01472 m
.84014 .02097 L
s
[(12)] .84014 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(14)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05782 .01472 m
.05782 .01847 L
s
.09184 .01472 m
.09184 .01847 L
s
.12585 .01472 m
.12585 .01847 L
s
.19388 .01472 m
.19388 .01847 L
s
.22789 .01472 m
.22789 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.32993 .01472 m
.32993 .01847 L
s
.36395 .01472 m
.36395 .01847 L
s
.39796 .01472 m
.39796 .01847 L
s
.46599 .01472 m
.46599 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.53401 .01472 m
.53401 .01847 L
s
.60204 .01472 m
.60204 .01847 L
s
.63605 .01472 m
.63605 .01847 L
s
.67007 .01472 m
.67007 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.77211 .01472 m
.77211 .01847 L
s
.80612 .01472 m
.80612 .01847 L
s
.87415 .01472 m
.87415 .01847 L
s
.90816 .01472 m
.90816 .01847 L
s
.94218 .01472 m
.94218 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11921 m
.03006 .11921 L
s
[(0.2)] .01131 .11921 1 0 Mshowa
.02381 .2237 m
.03006 .2237 L
s
[(0.4)] .01131 .2237 1 0 Mshowa
.02381 .32819 m
.03006 .32819 L
s
[(0.6)] .01131 .32819 1 0 Mshowa
.02381 .43268 m
.03006 .43268 L
s
[(0.8)] .01131 .43268 1 0 Mshowa
.02381 .53717 m
.03006 .53717 L
s
[(1)] .01131 .53717 1 0 Mshowa
.125 Mabswid
.02381 .04084 m
.02756 .04084 L
s
.02381 .06696 m
.02756 .06696 L
s
.02381 .09308 m
.02756 .09308 L
s
.02381 .14533 m
.02756 .14533 L
s
.02381 .17145 m
.02756 .17145 L
s
.02381 .19757 m
.02756 .19757 L
s
.02381 .24982 m
.02756 .24982 L
s
.02381 .27594 m
.02756 .27594 L
s
.02381 .30207 m
.02756 .30207 L
s
.02381 .35431 m
.02756 .35431 L
s
.02381 .38043 m
.02756 .38043 L
s
.02381 .40656 m
.02756 .40656 L
s
.02381 .4588 m
.02756 .4588 L
s
.02381 .48492 m
.02756 .48492 L
s
.02381 .51105 m
.02756 .51105 L
s
.02381 .56329 m
.02756 .56329 L
s
.02381 .58942 m
.02756 .58942 L
s
.02381 .61554 m
.02756 .61554 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
.5 Mabswid
.02381 .01472 m
.02499 .01479 L
.02605 .01525 L
.02729 .0166 L
.02846 .01899 L
.02954 .02235 L
.03053 .02655 L
.03163 .03245 L
.03279 .04023 L
.0352 .06146 L
.03746 .08721 L
.04262 .16275 L
.05278 .33889 L
.05779 .41759 L
.06244 .47802 L
.06751 .52775 L
.07019 .54729 L
.07305 .56338 L
.0743 .56902 L
.07566 .57422 L
.07686 .57806 L
.07813 .58141 L
.07928 .58385 L
.08053 .58595 L
.08168 .58737 L
.08274 .58832 L
.08345 .58875 L
.0841 .58902 L
.08481 .5892 L
.08557 .58924 L
.08685 .58902 L
.08757 .58875 L
.08826 .5884 L
.08952 .58755 L
.09067 .58656 L
.09328 .58376 L
.10458 .56817 L
.10932 .56232 L
.11192 .55962 L
.11438 .55745 L
.11668 .55578 L
.11916 .55436 L
.12026 .55386 L
.12142 .55342 L
.12249 .55308 L
.1235 .55283 L
.12479 .5526 L
.12596 .55247 L
.12722 .55242 L
.12793 .55243 L
Mistroke
.1286 .55246 L
.12981 .55258 L
.1309 .55275 L
.13217 .55302 L
.13336 .55334 L
.13608 .55429 L
.13857 .55542 L
.1442 .55868 L
.16467 .57498 L
.17465 .58321 L
.18374 .58987 L
.19255 .59518 L
.20211 .59948 L
.20736 .60115 L
.20967 .60173 L
.21221 .60226 L
.21444 .60263 L
.21685 .60294 L
.21817 .60307 L
.21939 .60317 L
.22055 .60324 L
.22177 .60329 L
.22301 .60331 L
.22371 .60332 L
.22435 .60332 L
.22553 .6033 L
.22679 .60325 L
.22791 .6032 L
.22895 .60313 L
.23134 .60291 L
.23391 .6026 L
.23626 .60224 L
.24155 .6012 L
.25098 .59864 L
.25993 .59556 L
.29823 .57904 L
.33898 .56253 L
.3592 .55613 L
.37821 .55132 L
.39635 .5477 L
.41593 .54471 L
.43594 .54245 L
.44614 .54155 L
.45704 .54075 L
.47675 .53963 L
.49511 .53889 L
.50534 .53857 L
.51643 .53829 L
.52602 .53809 L
.53658 .53791 L
Mistroke
.55643 .53766 L
.56729 .53756 L
.57746 .53748 L
.58785 .53742 L
.59773 .53737 L
.61626 .53731 L
.62585 .53728 L
.63608 .53726 L
.65447 .53723 L
.66476 .53722 L
.6759 .53721 L
.68552 .5372 L
.69607 .53719 L
.7057 .53719 L
.71602 .53718 L
.72692 .53718 L
.73709 .53718 L
.74728 .53718 L
.75835 .53718 L
.76881 .53717 L
.77847 .53717 L
.79715 .53717 L
.80778 .53717 L
.81776 .53717 L
.8276 .53717 L
.83796 .53717 L
.85647 .53717 L
.86602 .53717 L
.87622 .53717 L
.88579 .53717 L
.8946 .53717 L
.90485 .53717 L
.91596 .53717 L
.92575 .53717 L
.93611 .53717 L
.94556 .53717 L
.9557 .53717 L
.97402 .53717 L
.97619 .53717 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@0003P0
104S00030@0002400P4S00817`040@<00P4L00@10P040A/0104400030@000080000h00030@0002@0
00<1000080001040004Q00040@000B0000D100000@0200030@0001/000D100000@0P00030@0000<0
00<100000P0003T000<1000080050B4000@100018@001040004P00050@0000400P000`40000K0003
0@00008000<100007@000`4000050@<0000j00030@0001l000@100018P030B<00P4Q00050@000040
0P000`40000K00030@0000<000<1000070001040004200030@000080000E00030@0002<000<10000
7`000`400@0R00030@00028000@1000180001@400001008000<100006`000`40000400030@0001/0
00L100000@0100@0000E00030@00020000@100018P020B8000<100008P001040004N00<10`001040
004K00<10`001040004K00<110020@@0000E00030@0002400P4T00030@0002400`4R00818@000`40
000200817P000`40000200817P000`40000300030@000080000E00030@000?l02@0000h0o`4C0@00
5@020@L000<100001P000`40000700030@0000H000<100001P000`40000600030@0000L000<10000
1P000`40000600030@0000H000<100001`000`40000600030@0000H000<100001P000`4000070003
0@0000H000<100001P000`40000600030@0000L000<100001P000`40000600030@0000H000<10000
1`000`40000600030@0000H000<100001`000`40000600030@0000H000<100001@0001D00P4S0003
0@00028000<100008P000`40000R00030@00028000<100008P000`40000R00030@0000D0000E0003
0@010?l02@0001D000<10040o`0900005@000`400@3o00T0000E00030@010?l02@0001D00`7o00T0
000E00030@010?l02@0001D000<10040o`0900005@001040007o00P0000E00040@000Ol0200001D0
00@10001o`0800005@001040007o00P0000E008100<00@00o`0700005@001040007o00P0000E0004
0@000Ol0200001D000D100000@3o00L0000E00050@000040o`0700005@001@4000010?l01`0001D0
00D100000@3o00L0000E00810P000`40003o00D0000E00050@000040o`0700005@001@4000010?l0
1`0001D000D100000@3o00L0000E00050@000040o`0700005@001@4000010?l01`0000800P440003
0@000080104400050@000040o`07000000D00@000@0800030@0000D000D100000@3o00L000001@01
000100T000<1000010030@0300400?l01P000005004000402P000`40000300030@00008000<10000
o`04000000D00@000@0;00030@00008000<100000P000`40003o00@000001@01000100P000@10001
10000`40000200030@000?l0100000800P4:00811@000`40000200030@000?l0100001D000<10000
0P000`40003o00@0000E00030@00008000<10000o`0400005@020@<000<10000o`0400005@000`40
000200030@000?l0100001D000<100000P000`40003o00@0000E00030@00008000<10000o`040000
5@000`40000200030@000?l0100001D000<100000P000`40003o00@0000E00030@00008000<10000
o`0400005@020@<000<10000o`0400005@000`40000200030@000?l0100001D000<100000P000`40
003o00@0000E00030@00008000<10000o`0400005@000`40000200030@000?l0100001D000<10000
0`000`40003o00<0000E00030@0000<000<10000o`0300005@020@@000<10000o`0300005@000`40
000300030@000?l00`0001D000<100000`000`40003o00<0000E00030@0000<000<10000o`030000
5@000`40000300030@000?l00`0001D000<100000`000`40003o00<00002008110000`4000040003
0@0000<000<100000`000`40003o00<000001@01000100X000<100000`030@<000<10000o`030000
00D00@000@0700D110000`40000300030@000?l00`000005004000401`001040004500030@0000<0
00<10000o`03000000D00@000@0800030@0100D000<100000`000`40003o00<000001@01000100T0
0P4500030@0000<000<10000o`0300000P020@/000<100000`000`40000300030@000?l00`0001D0
00<100000`000`40003o00<0000E00811@000`40003o0080000E00030@0000@000<10000o`020000
5@000`40000400030@000?l00P0001D000<1000010000`40003o0080000E00030@0000@000<10000
o`0200005@000`40000400030@000?l00P0001D000<1000010000`40003o0080000E00811@000`40
003o0080000E00030@0000@000<10000o`0200005@000`40000400030@000?l00P0001D000<10000
10000`40003o0080000E00030@0000@000<10000o`0200005@000`40000400030@000?l00P0001D0
00<1000010000`40003o0080000E00811@000`40003o0080000E00030@0000@000<10000o`020000
5@000`40000500030@000?l00@0001D000<100001@000`40003o0040000E00030@0000D000<10000
o`0100005@000`40000500030@000?l00@0000800P4400030@0000<00P4500030@0000D000<10000
o`01000000D00@000@0800040@000@@00`4500030@000?l00@000005004000402000104000440003
0@0000D000<10000o`01000000D00@000@0800<11@000`40000500030@000?l00@00000500400040
20000`40000500030@0000D000<10000o`01000000D00@000@0800030@0000D000<100001@000`40
003o0040000200812P030@@000<100001@000`40003o0040000E00030@0000D000<10000o`010000
5@000`40000500030@000?l00@0001D00P4600030@000?l00@0001D000<100001@000`40003o0040
000E00030@0000D000<10000o`0100005@000`40000500030@000?l00@0001D000<100001@000`40
003o0040000E00030@0000D000<10000o`0100005@000`40000600030@000?l0000E00811`000`40
003o00005@000`40000600030@000?l0000E00030@0000H000<10000o`0001D000<100001P000`40
003o00005@000`40000600030@000?l0000E00030@0000H000<10000o`0001D000<100001P000`40
003o00005@020@L000<10000o`0001D000<100001P000`40003o00005@000`40000600030@000?l0
000E00030@0000H000<10000o`0001D000<100001P000`40003o00005@000`40000600030@000?l0
0002008110000`40000300811@000`40000600030@000?l000001@01000100P000@1000110030@H0
00<10000o`0000050040004020001040004400030@0000H000<10000o`000005004000402@020@D0
00<100001P000`40003o000000D00@000@0800040@000@@000<100001P000`40003o000000D00@00
0@0800040@000@@000<100001`000`40003n00000P020@X00P4500030@0000L000<10000oP0001D0
00<100001`000`40003n00005@020@P000<10000oP0001D000<100001`000`40003n00005@000`40
000700030@000?h0000E00030@0000L000<10000oP0001D000<100001`000`40003n00005@000`40
000700030@000?h0000E00030@0000L000<10000oP0001D00P4800030@000?h0000E00030@0000L0
00<10000oP0001D000<1000020000`40003m00005@000`40000800030@000?d0000E00030@0000P0
00<10000o@0001D000<1000020000`40003m00005@000`40000800030@000?d0000E00812@000`40
003m00005@000`40000800030@000?d0000E00030@0000T000<10000o00001D000<100002@000`40
003l00005@000`40000900030@000?`0000E00030@0000T000<10000o00000d0104400030@0000T0
00<10000o00000l000<100000`030@T000<10000O`1f0@L0000?00030@0000<000<100002P000`40
001R01`1O@0000l000<100000`000`40000:00030@0005P02P6I00003`000`40000300030@0000X0
00<10000D`050J<0000=00<11@000`40000:00030@0000`00P5000D1Z00000l000<100000`000`40
000:00030@0000T00`400`010@020CT0106]00005@000`40000:00030@0000P00P4700030@0003@0
0P6a00005@020@/000<100001`000`40000800030@0003000`6c00005@000`40000;00030@0000H0
00<100002@000`40000]0081]P0001D000<100002`000`40000500030@0000/000<10000:P020KP0
000E00030@0000/000<1000010000`40000=00030@0002L00P6j00005@000`40000;00030@0000@0
00<100003P000`40000T0081_00001D000<1000030000`40000200030@0001000P4Q00<1_P0001D0
00<1000030001@40000101D000<100006`030L40000E00030@0000`0104G00030@0001P00P740000
5@020BT00P4E00<1aP0001D000<10000:P030@l00`7900005@000`40000]00812`020L`0000E0003
0@0002l02`7>00005@000`40003o00T0000E00030@000?l02@0001D000<10000o`0900005@020Ol0
2P000001\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.15634, -0.0886075, 
  0.0540293, 0.00703499}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

FigStepT1b  = Plot[ step1[t] ,{t,0,14},PlotRange -> {0.9,1.2}];\
\>", "Input"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(FigStepT1b\)\" is \
similar to existing symbol \"\!\(FigStepT1a\)\"."\)], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0680272 -1.8541 2.06011 [
[.15986 .19351 -3 -9 ]
[.15986 .19351 3 0 ]
[.29592 .19351 -3 -9 ]
[.29592 .19351 3 0 ]
[.43197 .19351 -3 -9 ]
[.43197 .19351 3 0 ]
[.56803 .19351 -3 -9 ]
[.56803 .19351 3 0 ]
[.70408 .19351 -6 -9 ]
[.70408 .19351 6 0 ]
[.84014 .19351 -6 -9 ]
[.84014 .19351 6 0 ]
[.97619 .19351 -6 -9 ]
[.97619 .19351 6 0 ]
[.01131 0 -18 -4.5 ]
[.01131 0 0 4.5 ]
[.01131 .10301 -24 -4.5 ]
[.01131 .10301 0 4.5 ]
[.01131 .30902 -24 -4.5 ]
[.01131 .30902 0 4.5 ]
[.01131 .41202 -18 -4.5 ]
[.01131 .41202 0 4.5 ]
[.01131 .51503 -24 -4.5 ]
[.01131 .51503 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15986 .20601 m
.15986 .21226 L
s
[(2)] .15986 .19351 0 1 Mshowa
.29592 .20601 m
.29592 .21226 L
s
[(4)] .29592 .19351 0 1 Mshowa
.43197 .20601 m
.43197 .21226 L
s
[(6)] .43197 .19351 0 1 Mshowa
.56803 .20601 m
.56803 .21226 L
s
[(8)] .56803 .19351 0 1 Mshowa
.70408 .20601 m
.70408 .21226 L
s
[(10)] .70408 .19351 0 1 Mshowa
.84014 .20601 m
.84014 .21226 L
s
[(12)] .84014 .19351 0 1 Mshowa
.97619 .20601 m
.97619 .21226 L
s
[(14)] .97619 .19351 0 1 Mshowa
.125 Mabswid
.05782 .20601 m
.05782 .20976 L
s
.09184 .20601 m
.09184 .20976 L
s
.12585 .20601 m
.12585 .20976 L
s
.19388 .20601 m
.19388 .20976 L
s
.22789 .20601 m
.22789 .20976 L
s
.2619 .20601 m
.2619 .20976 L
s
.32993 .20601 m
.32993 .20976 L
s
.36395 .20601 m
.36395 .20976 L
s
.39796 .20601 m
.39796 .20976 L
s
.46599 .20601 m
.46599 .20976 L
s
.5 .20601 m
.5 .20976 L
s
.53401 .20601 m
.53401 .20976 L
s
.60204 .20601 m
.60204 .20976 L
s
.63605 .20601 m
.63605 .20976 L
s
.67007 .20601 m
.67007 .20976 L
s
.7381 .20601 m
.7381 .20976 L
s
.77211 .20601 m
.77211 .20976 L
s
.80612 .20601 m
.80612 .20976 L
s
.87415 .20601 m
.87415 .20976 L
s
.90816 .20601 m
.90816 .20976 L
s
.94218 .20601 m
.94218 .20976 L
s
.25 Mabswid
0 .20601 m
1 .20601 L
s
.02381 0 m
.03006 0 L
s
[(0.9)] .01131 0 1 0 Mshowa
.02381 .10301 m
.03006 .10301 L
s
[(0.95)] .01131 .10301 1 0 Mshowa
.02381 .30902 m
.03006 .30902 L
s
[(1.05)] .01131 .30902 1 0 Mshowa
.02381 .41202 m
.03006 .41202 L
s
[(1.1)] .01131 .41202 1 0 Mshowa
.02381 .51503 m
.03006 .51503 L
s
[(1.15)] .01131 .51503 1 0 Mshowa
.125 Mabswid
.02381 .0206 m
.02756 .0206 L
s
.02381 .0412 m
.02756 .0412 L
s
.02381 .0618 m
.02756 .0618 L
s
.02381 .0824 m
.02756 .0824 L
s
.02381 .12361 m
.02756 .12361 L
s
.02381 .14421 m
.02756 .14421 L
s
.02381 .16481 m
.02756 .16481 L
s
.02381 .18541 m
.02756 .18541 L
s
.02381 .22661 m
.02756 .22661 L
s
.02381 .24721 m
.02756 .24721 L
s
.02381 .26781 m
.02756 .26781 L
s
.02381 .28842 m
.02756 .28842 L
s
.02381 .32962 m
.02756 .32962 L
s
.02381 .35022 m
.02756 .35022 L
s
.02381 .37082 m
.02756 .37082 L
s
.02381 .39142 m
.02756 .39142 L
s
.02381 .43262 m
.02756 .43262 L
s
.02381 .45322 m
.02756 .45322 L
s
.02381 .47383 m
.02756 .47383 L
s
.02381 .49443 m
.02756 .49443 L
s
.02381 .53563 m
.02756 .53563 L
s
.02381 .55623 m
.02756 .55623 L
s
.02381 .57683 m
.02756 .57683 L
s
.02381 .59743 m
.02756 .59743 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.06315 0 m
.06751 .16887 L
.07019 .24593 L
.07305 .30934 L
.0743 .33162 L
.07566 .3521 L
.07686 .36723 L
.07813 .38044 L
.07928 .39008 L
.08053 .39837 L
.08168 .40397 L
.08274 .40769 L
.08345 .40939 L
.0841 .41048 L
.08481 .41116 L
.08557 .41134 L
.08685 .41047 L
.08758 .40938 L
.08826 .40801 L
.08952 .40465 L
.09068 .40075 L
.09328 .38971 L
.10458 .32826 L
.10932 .30518 L
.11192 .29452 L
.11438 .28598 L
.11668 .2794 L
.11916 .2738 L
.12026 .27183 L
.12142 .27008 L
.12249 .26876 L
.1235 .26776 L
.12479 .26684 L
.12596 .26634 L
.12722 .26614 L
.12793 .26618 L
.1286 .26631 L
.12981 .26678 L
.1309 .26745 L
.13217 .26851 L
.13336 .26977 L
.13608 .27352 L
.13857 .27795 L
.1442 .29082 L
.16467 .35511 L
.17465 .38755 L
.18374 .4138 L
.19255 .43476 L
.20211 .45169 L
.20736 .45828 L
Mistroke
.20967 .46056 L
.21221 .46266 L
.21444 .46414 L
.21685 .46536 L
.21817 .46588 L
.21939 .46625 L
.22055 .46652 L
.22177 .46672 L
.22301 .46683 L
.22371 .46684 L
.22435 .46684 L
.22553 .46676 L
.22679 .46659 L
.22791 .46636 L
.22895 .46609 L
.23134 .46525 L
.23391 .464 L
.23626 .46259 L
.24155 .45848 L
.25098 .44839 L
.25993 .43625 L
.29823 .37112 L
.33898 .30599 L
.3592 .28079 L
.37821 .2618 L
.39635 .24752 L
.41593 .23573 L
.43594 .22683 L
.44614 .22329 L
.45704 .22012 L
.47675 .2157 L
.49511 .21278 L
.50534 .21154 L
.51643 .21043 L
.52602 .20965 L
.53658 .20894 L
.55643 .20795 L
.56729 .20755 L
.57746 .20725 L
.58785 .207 L
.59773 .20681 L
.61626 .20655 L
.62585 .20644 L
.63608 .20636 L
.65447 .20624 L
.66476 .20619 L
.6759 .20615 L
.68552 .20612 L
.69607 .2061 L
.7057 .20608 L
Mistroke
.71602 .20607 L
.72692 .20606 L
.73709 .20605 L
.74728 .20604 L
.75835 .20603 L
.76881 .20603 L
.77847 .20602 L
.79715 .20602 L
.80778 .20602 L
.81776 .20602 L
.8276 .20602 L
.83796 .20601 L
.85647 .20601 L
.86602 .20601 L
.87622 .20601 L
.88579 .20601 L
.8946 .20601 L
.90485 .20601 L
.91596 .20601 L
.92575 .20601 L
.93611 .20601 L
.94556 .20601 L
.9557 .20601 L
.97402 .20601 L
.97619 .20601 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@0000P00P4400030@0000800`7o00/0000700040@000@/000<100000P020@P0
00<10000n00000L000@100012`000`40000200030@0000L000<10000n00000L000@100012@030@@0
00<100001`000`40003h00001`001040004800040@000@@000<100001`000`40003h00001`001040
004800040@000@@000<100001`000`40003h000020020@X00P4500030@0000L000<10000n00001/0
0P4800030@000?P0000K00030@0000L000<10000n00001/000<100001`000`40003h00006`000`40
000700030@000?P0000K00030@0000L000<10000n00001/00P4800030@000?P0000K00030@0000L0
00<10000n00001/000<100001`000`40003h00006`000`40000700030@000?P0000K00030@0000L0
00<10000n00001/000<100001`000`40003h00006`020@P000<10000n00001/000<100001`000`40
003h00006`000`40000700030@000?P0000K00030@0000L000<10000n00001/000<100001`000`40
003h00006`020@P000<10000n00001/000<1000020000`40003g00006`000`40000800030@000?L0
000K00030@0000P000<10000m`0000800P4400030@0000800`4300<11@000`40000800030@000?L0
00001@01000100/000<100000`000`40000200812@000`40003g000000D00@000@0;00030@0000<0
00<100000P000`40000800030@000?L000001@01000100T00`4200<11@000`40000800030@000?L0
00001@01000100P000@100010P000`40000500030@0000P000<10000m`0000050040004020001040
004200030@0000D000<1000020000`40003g00000P020@X00P4300@110000`40000800030@000?L0
000K00812@000`40003g00006`000`40000800030@000?L0000K00030@0000P000<10000m`0001/0
00<1000020000`40003g00006`000`40000800030@000?L0000K00812@000`40003g00006`000`40
000800030@000?L0000K00030@0000P000<10000m`0001/000<1000020000`40003g00006`000`40
000800030@000?L0000K00030@0000P000<10000m`0001/00P4900030@000?L0000K00030@0000P0
00<10000m`0001/000<1000020000`40000D00@18P000`40000P00818`020Ah0104300816`040@80
104J00@110000`40000200006`000`40000800030@0001@000<100008`000`40000O00040@000B40
00@100017`001@400001008000<100006P001@40000101l000<100000`000`40000200006`000`40
000800030@0001D000<100007`050B0000@100018@001040004O00050@0000400P000`40000J0003
0@00008000<1000070000`4000050@<0000K00812@000`40000F00030@0001h000@100018@030B<0
0P4P00050@0000400P000`40000J00030@0000<000<100006`001040004200030@000080000K0003
0@0000P000<100005`000`40000N00030@01024000<100008P001040004O00050@0000400P000`40
000J00030@0000@000<100006P001`4000010040100001/000<1000020000`40000D00040@000B40
0P4Q00030@00028000@100017@030@<000@100016P030@<000@100016P030@@00P4400006`000`40
000800030@0001D00P4S00030@0002000`4R008180000`40000200817@000`40000200817@000`40
000300030@000080000K00030@0000P000<10000m`0001/000<1000020000`40003g0000503o0@d1
000K00030@0000H000@100011@000`40000600030@0000H000<100001P000`40000600030@0000H0
00<100001P000`40000600030@0000H000<100001P000`40000600030@0000H000<100001P0=0@H0
00<100001P000`40000600030@0000H000<100001P000`40000600030@0000H000<100001P000`40
000600030@0000H000<100001P000`40000600030@0000H000<100001@0001/000<100002@000`40
000E00030@00024000<100008@000`40000800P14P000`40000Q00030@00024000<100008@000`40
000500006`000`40000900030@0006H00P6>00006`000`40000900030@0006001P6@00006`020@X0
00<10000GP020IH0000K00030@0000T000<10000G0020IP0000K00030@0000T000<10000FP020IX0
000K00030@0000T000<10000F@000`40002J00006`000`40000900030@0005L00P6M00006`000`40
000900030@0005D00P6O00006`020@X000<10000E0000`40002O00006`000`40000900030@0005<0
00<10000X00001/000<100002@000`40001A0081X`0001/000<100002@000`40001@00030@000:<0
000K00030@0000T000<10000C`000`40002T00006`020@X000<100002`030D0000<10000Y@0001/0
00<100002@000`40000:00810P000`40000l00030@000:H0000K00030@0000T000<100002P000`40
000200030@0003X000<10000Y`0001/000<100002P000`40000900030@00008000<10000>@000`40
002X00006`000`40000:00030@0000T000<100000`000`40000g00030@000:T0000K00030@0000X0
00<1000020000`40000400030@0003H000<10000ZP0001/00P4;00030@0000P000<100001@000`40
000d00030@000:/0000K00030@0000X000<100001`000`40000600030@0003@000<10000Z`0001/0
00<100002P000`40000700030@0000L000<10000<P000`40002/00006`000`40000:00030@0000L0
00<100001`000`40000a00030@000:d000000`010@020@<000<100000`020@<00`4500030@0000X0
00<100001`000`40000700030@00030000<10000[P0000<000<100001`001040004500030@000080
0P4;00030@0000L000<100001`000`40000_00030@000:l0000300030@0000L000@100011@000`40
000200030@0000X000<100001P000`40000900030@0002h000<10000[`0000<000<100001`001040
004200<11@000`40000:00030@0000H000<100002@000`40000]00030@000;00000300030@0000L0
00@100010P000`40000500030@0000X000<100001P000`40000900030@0002d000<10000/0000004
00410@T000@100010P000`40000500030@0000X000<100001@000`40000;00030@0002/000<10000
/@0000<000<1000020020@<0104400030@0000X000<100001@000`40000;00030@0002X000<10000
/P0001/00P4;00030@0000D000<100002`000`40000Z00030@000;80000K00030@0000X000<10000
1@000`40000;00030@0002T000<10000/`0001/000<100002P000`40000400030@0000d000<10000
:0000`40002c00006`000`40000;00030@0000<000<100003@000`40000W00030@000;@0000K0003
0@0000/000<100000`000`40000=00030@0002H000<10000]@0001/00P4<00030@0000<000<10000
3@000`40000V00030@000;D0000K00030@0000/000<100000`000`40000>00030@0002@000<10000
]P0001/000<100002`000`40000300030@0000h000<100008`000`40002g00006`000`40000;0003
0@00008000<100003`000`40000S00030@000;L0000K00030@0000/000<100000P000`40000@0003
0@00024000<10000^00001/000<100002`000`40000200030@00010000<100008@000`40002h0000
6`020@`000<100000P000`40000A00030@0001l000<10000^@0001/000<100002`000`4000020003
0@00014000<100007P000`40002j00006`000`40000;00030@00008000<100004@000`40000N0003
0@000;X0000K00030@0000/000D100000@0E00030@0001`000<10000^`0001/000<1000030001040
004E00030@0001`000<10000^`0001/00P4=00040@000AD000<100006`000`40002l00006`000`40
000<00040@000AH000<100006@000`40002m00006`000`40000<00040@000AH000<100006@000`40
002m00006`000`40000<00040@000AH000<1000060000`40002n00001`040@<000<100000P040@@0
00<1000030000`400@0G00030@0001P000<10000_P0000T000<100002@000`40000300030@0000d0
0P4H00030@0001H000<10000_`0000T000<100002@000`4000030081:0000`40000F00030@000;l0
000900030@0000T000<100000`000`40000X00030@0001@000<10000`00000T000<100002@000`40
000300030@0002P000<100004`000`40003100001`030@T00`4500030@0002T000<100004P000`40
003100002@000`40000900030@0000<000<10000:@000`40000A00030@000<80000K0081:`000`40
000@00030@000<80000K00030@0002X000<100003`000`40003300006`000`40000[00030@0000d0
00<10000a00001/000<10000:`000`40000<00030@000<D0000K00030@0002`000<100002P000`40
003600006`000`40000/00030@0000T000<10000a`0001/00P4^00030@0000P000<10000a`0001/0
00<10000;P000`4000050081bP0001/000<10000;P020@D000<10000bP0001/000<10000<0060L`0
000K00030@000?l00`0001/00P7o00@0000K00030@000?l00`0001/000<10000o`0300006`000`40
003o00<0000K00030@000?l00`0001/000<10000o`0300006`020Ol0100001/000<10000o`030000
6`000`40003o00<0000K00030@000?l00`000003004100810`000`40000200@10P030@D000<10000
o`0300000`000`40000900030@0000@000<100000P020Ol0100000<000<100002@000`4000040003
0@00008000<10000o`0300000`000`40000900040@0000<11@000`40003o00<0000300030@0000T0
00D100000@0700030@000?l00`00000400410@T00`4300030@0000D000<10000o`0300000`000`40
000900040@0000@110000`40003o00<0000K0081o`0400006`000`40003o00<0000K00030@000?l0
0`0001/000<10000o`0300006`000`40003o00<0000K0081o`0400006`000`40003o00<0000K0003
0@000?l00`0001/000<10000o`0300006`000`40003o00<0000K00030@000?l00`0001/00P7o00@0
000K00030@000?l00`0001/000<10000o`0300006`000`40003o00<0000K00030@000?l00`0001/0
0P7o00@0000K00030@000?l00`0001/000<10000o`0300006`000`40003o00<0000K00030@000?l0
0`0001/000<10000o`030000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.5101, 0.884455, 0.0552619,
   0.00182481}}]
}, Open  ]],

Cell["The compensator we obtained is ", "Text"],

Cell[CellGroupData[{

Cell["\<\

Crat1[s_]   = CancelZP[ pinv[s] Trat[s]/(1-Trat[s]),s,s]\
\>", "Input"],

Cell[BoxData[
    \(\(\(2947.23279469809806`\[InvisibleSpace]\) + 8032.27283345494669`\ s + 
        8899.41144598526356`\ s\^2 + 11031.8940758898254`\ s\^3 + 
        6119.53386039581559`\ s\^4 + 2908.84300105491977`\ s\^5 + 
        368.970338833291711`\ s\^6\)\/\(\(2843.26189074701964`
        \[InvisibleSpace]\) + 2194.74353731092097`\ s + 
        1336.72617360725243`\ s\^2 + 357.888662842122462`\ s\^3 + 
        45.492965217437602`\ s\^4 + 3.10962816146362719`\ s\^5 + 
        0.0712923332656223429`\ s\^6\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

FigCrat1ZPa = PlotZP[Crat1[s],s];\
\>", "Input"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(Crat1\)\" is similar to \
existing symbol \"\!\(crat1\)\"."\)], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.961538 0.034965 0.5 0.0671487 [
[.08741 .4875 -9 -9 ]
[.08741 .4875 9 0 ]
[.26224 .4875 -9 -9 ]
[.26224 .4875 9 0 ]
[.43706 .4875 -9 -9 ]
[.43706 .4875 9 0 ]
[.61189 .4875 -9 -9 ]
[.61189 .4875 9 0 ]
[.78671 .4875 -6 -9 ]
[.78671 .4875 6 0 ]
[.94904 .09711 -12 -4.5 ]
[.94904 .09711 0 4.5 ]
[.94904 .23141 -12 -4.5 ]
[.94904 .23141 0 4.5 ]
[.94904 .3657 -12 -4.5 ]
[.94904 .3657 0 4.5 ]
[.94904 .6343 -6 -4.5 ]
[.94904 .6343 0 4.5 ]
[.94904 .76859 -6 -4.5 ]
[.94904 .76859 0 4.5 ]
[.94904 .90289 -6 -4.5 ]
[.94904 .90289 0 4.5 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.08741 .5 m
.08741 .50625 L
s
[(-25)] .08741 .4875 0 1 Mshowa
.26224 .5 m
.26224 .50625 L
s
[(-20)] .26224 .4875 0 1 Mshowa
.43706 .5 m
.43706 .50625 L
s
[(-15)] .43706 .4875 0 1 Mshowa
.61189 .5 m
.61189 .50625 L
s
[(-10)] .61189 .4875 0 1 Mshowa
.78671 .5 m
.78671 .50625 L
s
[(-5)] .78671 .4875 0 1 Mshowa
.125 Mabswid
.12238 .5 m
.12238 .50375 L
s
.15734 .5 m
.15734 .50375 L
s
.19231 .5 m
.19231 .50375 L
s
.22727 .5 m
.22727 .50375 L
s
.2972 .5 m
.2972 .50375 L
s
.33217 .5 m
.33217 .50375 L
s
.36713 .5 m
.36713 .50375 L
s
.4021 .5 m
.4021 .50375 L
s
.47203 .5 m
.47203 .50375 L
s
.50699 .5 m
.50699 .50375 L
s
.54196 .5 m
.54196 .50375 L
s
.57692 .5 m
.57692 .50375 L
s
.64685 .5 m
.64685 .50375 L
s
.68182 .5 m
.68182 .50375 L
s
.71678 .5 m
.71678 .50375 L
s
.75175 .5 m
.75175 .50375 L
s
.82168 .5 m
.82168 .50375 L
s
.85664 .5 m
.85664 .50375 L
s
.89161 .5 m
.89161 .50375 L
s
.92657 .5 m
.92657 .50375 L
s
.05245 .5 m
.05245 .50375 L
s
.01748 .5 m
.01748 .50375 L
s
.9965 .5 m
.9965 .50375 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.96154 .09711 m
.96779 .09711 L
s
[(-6)] .94904 .09711 1 0 Mshowa
.96154 .23141 m
.96779 .23141 L
s
[(-4)] .94904 .23141 1 0 Mshowa
.96154 .3657 m
.96779 .3657 L
s
[(-2)] .94904 .3657 1 0 Mshowa
.96154 .6343 m
.96779 .6343 L
s
[(2)] .94904 .6343 1 0 Mshowa
.96154 .76859 m
.96779 .76859 L
s
[(4)] .94904 .76859 1 0 Mshowa
.96154 .90289 m
.96779 .90289 L
s
[(6)] .94904 .90289 1 0 Mshowa
.125 Mabswid
.96154 .13068 m
.96529 .13068 L
s
.96154 .16426 m
.96529 .16426 L
s
.96154 .19783 m
.96529 .19783 L
s
.96154 .26498 m
.96529 .26498 L
s
.96154 .29855 m
.96529 .29855 L
s
.96154 .33213 m
.96529 .33213 L
s
.96154 .39928 m
.96529 .39928 L
s
.96154 .43285 m
.96529 .43285 L
s
.96154 .46643 m
.96529 .46643 L
s
.96154 .53357 m
.96529 .53357 L
s
.96154 .56715 m
.96529 .56715 L
s
.96154 .60072 m
.96529 .60072 L
s
.96154 .66787 m
.96529 .66787 L
s
.96154 .70145 m
.96529 .70145 L
s
.96154 .73502 m
.96529 .73502 L
s
.96154 .80217 m
.96529 .80217 L
s
.96154 .83574 m
.96529 .83574 L
s
.96154 .86932 m
.96529 .86932 L
s
.96154 .06353 m
.96529 .06353 L
s
.96154 .02996 m
.96529 .02996 L
s
.96154 .93647 m
.96529 .93647 L
s
.96154 .97004 m
.96529 .97004 L
s
.25 Mabswid
.96154 0 m
.96154 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
54.8593 18.8327 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
67.6307 15.172 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
67.6307 22.4934 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
68.252 18.8327 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
69.502 16.3035 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
69.502 21.3619 1 0 365.73 arc
setmatrix s
.06923 .48182 m
.10559 .51818 L
s
.06923 .51818 m
.10559 .48182 L
s
.73581 .02727 m
.77217 .06364 L
s
.73581 .06364 m
.77217 .02727 L
s
.73581 .93636 m
.77217 .97273 L
s
.73581 .97273 m
.77217 .93636 L
s
.75916 .48182 m
.79552 .51818 L
s
.75916 .51818 m
.79552 .48182 L
s
.91752 .36189 m
.95388 .39826 L
s
.91752 .39826 m
.95388 .36189 L
s
.91752 .60174 m
.95388 .63811 L
s
.91752 .63811 m
.95388 .60174 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003C00030@0000L000<10000<`000`40000:0000e0000`40
000500030@0003@00`4:0000e@000`40000300030@0003D000<100002P000=H000D100000@0h0003
0@0000X0003G00030@0103T000<100002P000=P000<10000>0000`40000:0000f0000`40000h0003
0@0000X0003G00030@0103T000<100002P000=H000D100000@0h00030@0000X0003E00030@0000<0
00<10000=@000`40000:0000e0000`40000500030@0003@000<100002P000=<000<100001`000`40
000c00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0=00811@000`40
000:0000o`0<00040@000@@00`4:0000o`0<00040@000@@000<100002P000?l01P040@800`450003
0@0000X0003o00`000<100001@000`40000:0000o`0<00030@0000D000<100002P000?l03@030@@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050030@X0003o01@000<100002P000?l050000`40000:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o00h000<100000`000`40000:0000o`0>00030@0000<0
0`4:0000o`0;00D110000`40000:0000o`0600@100D00@000@0500030@0000X0003o00`000<10040
1@000`40000:0000o`0=00811@000`40000:0000o`0>00030@0000<000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00<12P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@00`4:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`080004
0@0000@10P000`400@0<0000o`0900040@000@@000<100000`4:0000o`0:00040@000@8000D10000
0@0<0000o`0600@100@00@000P4400030@0000X0003o00`000@1004110000`40000:0000o`0<0081
00<00@000`000`40000:0000o`0:008100@00@4110000`40000:0000o`0:00070@0100400@030003
0@0000X0003o00T000<100400`020@<000<100002P000?l02@020@D00P4200030@0000X0003o00T0
00<100001@001040004<0000o`0800811`020@0400410@X0003o00T000<100001@001040004<0000
o`0900030@0000D000<100000P4;0000o`0900030@0000@000<100000P400`0100090000o`090003
0@0000@000L100400@0100X0003o00X000<100000`001@400@01008000<100001`000?l02P000`40
000300810P001@40000100P0003o00/000<100000P020@8000D100000@080000o`0<00030@010081
0`001@40000100P0003o00d000@100010`001@40000100P0003o010000@100000`400`0100070000
o`0A00040@000@<000<100001P000?l04@001040004300030@0000H0003o014000@100010`000`40
00060000o`0A00040@000@<000<100001P000?l04@001040004200030@0000L0003o01801@4:0000
o`0D00030@0000X0003o01@000<100002P000?l050030@X0000G00@10P030BT010430081:@040@80
0`4Y00@10`020B`00`4^00030@0000X0000G00030@0000H000<100009P000`40000300040@000BX0
00<1000010000`40000X00050@0000400P000`40000P00030@0000T000<100009`001@40000100`0
000H00030@0000D000<100009`000`40000200040@000BX000<1000010000`40000X00050@000040
0P000`40000O00030@0100X000<100009@020@050040004030000140104400040@0000<18`040@@0
00D100000@0200030@000200104400040@0000<18`040@@000D100000@0200030@0001h000@10000
1@4200<19`001@400001008000<100002P0001<000<1000010000`4000020Bd000@100010P000`40
000X00050@000040;@001@400001008000<100007P000`400@0200030@0000<000<100409`000`40
000200030@0100`0000D00040@000@8000@10001:`001040004200040@000BP00`4300030@0002T0
0`4300040@000Al000L100000@0100D00P4W00030@0000<000<10040300001D000<100000P4200D1
:@020@@00P4[00040@0000@1:P000`400002008180000`40000200030@0100@0104U00030@0000@0
0P4<00005P000`40000200030@000;P000<100000`020@<000<100009`000`4000040081300001L0
00@10001^P000`40000500040@000BT000<100001@020@`0003o0B41000500030@0000L000<10000
1@0010400@4800030@0000L000<100001`000`40000700030@0000L000<100001`000`4000070003
0@0000L000<100001`000`40000700030@0000L000<100001`000`40000700030@0000L000<10000
1`000`40000700030@0000L000<100001`000`40000500030@0100D000<100000P4900030@0000L0
00<100001`000`40000700811`020@T00@410040000F00060@000@01;`000`40000_00030@0002l0
00<100008`000`40000400810`000`40000W00030@0000@00P4<00005@000`40000400030@000;H0
00<100000`000`400@0400030@0002H000<1000010020@`0000D00030@0000H000<10000]P001@40
0001008000<100000`000`40000U00030@0000@00P4<00004`000`40000800030@000;D000@10001
0P000`40000500030@0002@000<100000`000`400@0<0000e`001@400001030000D100000@020003
0@0000X0003H00030@01038000<100400`000`40000:0000f@000`40000b00030@00008000<10000
2P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D00812`000?l04`020@03
004000T0003o018000<100400P000`4000070000o`0B00030@01008000<100001`000?l04@001040
004300030@0000H0003o014000@100010`000`4000060000o`0@00050@0000400`000`4000060000
o`0@00040@0000<100<00@001`000?l03@001040004300050@00004020000?l02`020@0500400040
0P001@40000100P0003o00X000<100000P000`400@0200050@00004020000?l02P000`4000030081
0P001@40000100P0003o00T000<1000010020@8000@100012@000?l02@000`400005008100D00@00
0@090000o`0900030@0000D000<100400`4:0000o`0900030@0000D000@1000130000?l02@000`40
000500040@000@`0003o00P00P47008100@00@412P000?l02@000`40000400810P000`40000:0000
o`0900811@001@40000100`0003o00X00P4300810`000`40000:0000o`0:00070@0100400@030003
0@0000X0003o00/000D100400@0400030@0000X0003o00`00`4500030@0000X0003o00`00`450003
0@0000X0003o00/01@4400030@0000X0003o00X000<100400`001@40000100`0003o00T000D10000
0@0300030@0000<12P000?l020000`40000300070@0000400@0<0000o`0?00030@00008000<10000
2P000?l030001040004400030@0000X0003o00d00P4500030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050030@X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
0`4:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50030@X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000
o`0>00030@0000<000<100002P000?l03P000`40000300<12P000?l02`050@@000<100002P000?l0
2`001040004500030@0000X0003o00`000<100401@000`40000:0000o`0=00811@000`40000:0000
o`0>00030@0000<000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@00`4:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0=00811@000`40000:0000o`0<00040@000@@000<10000
2P000?l030001040004400<12P000?l030030@D000<100002P000?l030000`40000500030@0000X0
003o00`000<100001@000`40000:0000o`0=00<110000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003C00030@0000L000<10000<`030@X0
003D00030@0000D000<10000=0000`40000:0000e@000`40000300030@0003D000<100002P000=H0
00D100000@0h00030@0000X0003G00030@0103T000<100002P000=P000<10000>0000`40000:0000
f0000`40000h00030@0000X0003G00030@0103T000<100002P000=H000D100000@0h00030@0000X0
003E00030@0000<000<10000=@000`40000:0000e0000`40000500030@0003@00`4:0000d`000`40
000700030@0003<000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000
\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-27.5002, -7.44624, 0.0996527, 
  0.0518901}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

FigCrat1ZPb = Show[FigCrat1ZPa,PlotRange -> {{-3,3},{-3,3}}];
 \
\>", "Input"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(FigCrat1ZPb\)\" is \
similar to existing symbol \"\!\(FigCrat1ZPa\)\"."\)], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.166667 0.5 0.166667 [
[0 .4875 -6 -9 ]
[0 .4875 6 0 ]
[.16667 .4875 -6 -9 ]
[.16667 .4875 6 0 ]
[.33333 .4875 -6 -9 ]
[.33333 .4875 6 0 ]
[.66667 .4875 -3 -9 ]
[.66667 .4875 3 0 ]
[.83333 .4875 -3 -9 ]
[.83333 .4875 3 0 ]
[1 .4875 -3 -9 ]
[1 .4875 3 0 ]
[.4875 0 -12 -4.5 ]
[.4875 0 0 4.5 ]
[.4875 .16667 -12 -4.5 ]
[.4875 .16667 0 4.5 ]
[.4875 .33333 -12 -4.5 ]
[.4875 .33333 0 4.5 ]
[.4875 .66667 -6 -4.5 ]
[.4875 .66667 0 4.5 ]
[.4875 .83333 -6 -4.5 ]
[.4875 .83333 0 4.5 ]
[.4875 1 -6 -4.5 ]
[.4875 1 0 4.5 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
0 .5 m
0 .50625 L
s
[(-3)] 0 .4875 0 1 Mshowa
.16667 .5 m
.16667 .50625 L
s
[(-2)] .16667 .4875 0 1 Mshowa
.33333 .5 m
.33333 .50625 L
s
[(-1)] .33333 .4875 0 1 Mshowa
.66667 .5 m
.66667 .50625 L
s
[(1)] .66667 .4875 0 1 Mshowa
.83333 .5 m
.83333 .50625 L
s
[(2)] .83333 .4875 0 1 Mshowa
1 .5 m
1 .50625 L
s
[(3)] 1 .4875 0 1 Mshowa
.125 Mabswid
.03333 .5 m
.03333 .50375 L
s
.06667 .5 m
.06667 .50375 L
s
.1 .5 m
.1 .50375 L
s
.13333 .5 m
.13333 .50375 L
s
.2 .5 m
.2 .50375 L
s
.23333 .5 m
.23333 .50375 L
s
.26667 .5 m
.26667 .50375 L
s
.3 .5 m
.3 .50375 L
s
.36667 .5 m
.36667 .50375 L
s
.4 .5 m
.4 .50375 L
s
.43333 .5 m
.43333 .50375 L
s
.46667 .5 m
.46667 .50375 L
s
.53333 .5 m
.53333 .50375 L
s
.56667 .5 m
.56667 .50375 L
s
.6 .5 m
.6 .50375 L
s
.63333 .5 m
.63333 .50375 L
s
.7 .5 m
.7 .50375 L
s
.73333 .5 m
.73333 .50375 L
s
.76667 .5 m
.76667 .50375 L
s
.8 .5 m
.8 .50375 L
s
.86667 .5 m
.86667 .50375 L
s
.9 .5 m
.9 .50375 L
s
.93333 .5 m
.93333 .50375 L
s
.96667 .5 m
.96667 .50375 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.5 0 m
.50625 0 L
s
[(-3)] .4875 0 1 0 Mshowa
.5 .16667 m
.50625 .16667 L
s
[(-2)] .4875 .16667 1 0 Mshowa
.5 .33333 m
.50625 .33333 L
s
[(-1)] .4875 .33333 1 0 Mshowa
.5 .66667 m
.50625 .66667 L
s
[(1)] .4875 .66667 1 0 Mshowa
.5 .83333 m
.50625 .83333 L
s
[(2)] .4875 .83333 1 0 Mshowa
.5 1 m
.50625 1 L
s
[(3)] .4875 1 1 0 Mshowa
.125 Mabswid
.5 .03333 m
.50375 .03333 L
s
.5 .06667 m
.50375 .06667 L
s
.5 .1 m
.50375 .1 L
s
.5 .13333 m
.50375 .13333 L
s
.5 .2 m
.50375 .2 L
s
.5 .23333 m
.50375 .23333 L
s
.5 .26667 m
.50375 .26667 L
s
.5 .3 m
.50375 .3 L
s
.5 .36667 m
.50375 .36667 L
s
.5 .4 m
.50375 .4 L
s
.5 .43333 m
.50375 .43333 L
s
.5 .46667 m
.50375 .46667 L
s
.5 .53333 m
.50375 .53333 L
s
.5 .56667 m
.50375 .56667 L
s
.5 .6 m
.50375 .6 L
s
.5 .63333 m
.50375 .63333 L
s
.5 .7 m
.50375 .7 L
s
.5 .73333 m
.50375 .73333 L
s
.5 .76667 m
.50375 .76667 L
s
.5 .8 m
.50375 .8 L
s
.5 .86667 m
.50375 .86667 L
s
.5 .9 m
.50375 .9 L
s
.5 .93333 m
.50375 .93333 L
s
.5 .96667 m
.50375 .96667 L
s
.25 Mabswid
.5 0 m
.5 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
-0.46827 .5 .0659 0 365.73 arc
s
newpath
.37335 .25876 .0659 0 365.73 arc
s
newpath
.37335 .74124 .0659 0 365.73 arc
s
newpath
.41429 .5 .0659 0 365.73 arc
s
newpath
.49667 .33333 .0659 0 365.73 arc
s
newpath
.49667 .66667 .0659 0 365.73 arc
s
.29017 .15721 m
.4635 .24747 L
s
.29017 .24747 m
.4635 .15721 L
s
.29017 .75253 m
.4635 .84279 L
s
.29017 .84279 m
.4635 .75253 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@0008T0
0P6E0000R000104000450081S@0008/000<100000`000`40002<0000PP040@<00P4600030@0008`0
002;00030@0000<000<10000S00008P000@100011@000`40002<0000R@020@H000<10000S0000940
00<10000S000094000<10000S000094000<10000S00009400P6=0000T@000`40002<0000T@000`40
002<0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40
002<0000T@000`40002<0000T@000`40002<0000T@020Hd0002A00030@0008`0002A00030@0008`0
002A00030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0
002A00030@0008`0002A0081S@00094000<10000S000094000<10000S000094000<10000S0000940
00<10000S000094000<10000S000094000<10000S000094000<10000S000094000<10000S0000940
0P6=0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40
002<0000T@000`40002<0000EP000`40000]00030@0000P000<10000S00005L00P4[00810P040@D0
00<10000S00005T00P4W008110000`4000060081S@0005/00P4S00811`000`40000500030@0008`0
001M00817`020@80104400030@0000@000<10000S00005l00P4K00813@000`40000300030@0008`0
001Q00815`020@`000@100011@000`40002<0000H`020A<00P4?00811P000`40002<0000I@020@l0
0P4I00030@0008`0001W00812`020A/000<10000S00006P0304M00030@0008`0001U00<10`020@<0
0P4200816`000`40002<0000I0000`40000600<11P000`40000H0081S@0006800P4900030@0100L0
0P4H00030@0008`0001Q00030@0000L00P4300811`000`40000E00030@0008`0001P00030@0000H0
0P4700811P000`40000D00030@0008`0001P00030@0000@00P4;00811@000`40000C00030@0008`0
001O00030@0000<00P4?008110000`40000B00030@0008`0001N00030@0000800P4C00810P000`40
000B00030@0008`0001N00030@0000815`020@030040018000<10000S00005d000@100416`030A80
00<10000S00005d00P4O00814@020Hd0001K00818P030@l000<10000S00005T00P400`01000Q0003
0@0000813@000`40002<0000E`020@<000<1000080000`40000200812`000`40002<0000EP000`40
000200030@00028000<100000`000`40000800030@0008`0001K00030@00028000<100003P000`40
002<0000F`000`40000R00030@0000h000<10000S00005/000<100008P000`40000>00030@0008`0
001K00030@00028000<100003P000`40002<0000F`000`40000R00030@0000X01P6=0000F`000`40
000R00030@0000H0104400030@010081RP0005`000<1000080000`400005008120000`40000200@1
QP0005`000<1000080000`40000400030@0000P000<100001P000`4000230000G0000`40000P0003
0@0000800P4;00030@0000L000<10000PP0005d000<100007P000`40000200030@0000/000<10000
20000`4000210000G@000`40000N00050@0000403P000`40000900030@000800001N00030@0001`0
00<100000P000`40000<00030@0000X000<10000O`0005l000<100006P000`40000200030@0000d0
00<100002`000`40001n0000G`000`40000I00030@00008000<100003P020@`000<10000OP000600
00<1000060000`40000200030@0000h000<1000030000`40001m0000H@000`40000F00030@000080
00<100003`000`40000<00030@0007d0001R00030@0001<00P4400030@00010000<1000030000`40
001m0000H`000`40000A00030@0000@000<1000040000`40000=00030@0007`0001T00813`020@L0
00<1000040000`40000>00030@0007/0001V00<12P020@T000<1000040000`40000>00030@0007/0
001Y00X12P000`40000A00030@0000h000<10000N`0007d000<1000020040@D000<100003P000`40
001k0000O@000`40000:00030@0000@000<100003P000`40001k0000O@000`40000:00030@0000@0
0P4?00030@0007/0001m00030@000080104400030@0000@000<100003P000`40001k0000O@000`40
000:00030@0000@000<100003P000`40001k0000OP000`40000700<11P000`40000>00030@0007/0
001n00030@0000T000<1000010000`40000>00030@0007/0001n00030@00010000<100003@000`40
001l0000OP000`40000@00030@0000d000<10000O00007l000<100003`000`40000<00030@0007d0
001o00030@0000l000<1000030000`40001m0000P0000`40000>00813@000`40001m0000P@000`40
000=00030@0000/000<10000OP00084000<100003@000`40000:00030@0007l0002200030@0000`0
00<100002@000`4000200000P`000`40000;00030@0000P000<10000P@0008@000<100002P000`40
000700030@000880002500030@0000T000<100001@020HD0002600@11`000`4000030081Q`0008X0
00<1000010000`4000030HT0002;00T1S000094000<10000S000094000<10000S000094000<10000
S000094000<10000S000094000<10000S000094000<10000S000094000<10000S000094000<10000
S00009400P6=0000M00;0A8000<10000S00007400`4;008140000`40002<0000L0000`40000>0081
3P000`40002<0000KP020A<000<100002`000`40002<0000K@000`40000D00030@0000X000<10000
S00006`000<100005P020@X000<10000S00006/000<100006@000`40000700030@0008`0001Z0003
0@0001/000<100001P000`40002<0000J@000`40000L00030@0000H000<10000S00000P00P4[00@1
:P040@8000<100007@000`4000050081:P040BX0104/00810`0000L000@10001:P000`40000]0004
0@000B4000<1000010000`40000[00030@0002T000<10000;0001040004200002P000`40000Y0003
0@0002`000@100018@000`40000400030@0002/000<10000:P000`40000^00410@01000000<00@40
0P4300819@040@@000<100008`040@@000<100408P000`40000400030@0002/000<10000:`000`40
000[00810`0000X000<10000:`000`40000Z00030@01028000<1000010000`40000[00030@0002`0
00<10000;0010@400@0000L000@10001:P001040004Z00<100<00@008P000`40000300030@0002T0
0`4[00040@000B/000@100010P0000P00P4/0081;@000`400@0S00030@0000<000<10000:`000`40
000Z0081;@020@<0001W00030@00024000<100000`000`40002<0000IP000`40000R00030@0000<0
00<10000S00000H0o`4H0@<0000600030@0000H000<100001P000`40000600030@0000L000<10000
1P000`40000600030@0000H000<100001`000`40000600030@0000H000D100000@0400030@0000L0
00<100001P000`40000600050@0000401@000`40000600030@0000H000<100001P000`4000070003
0@0000H000<100001P000`40000600030@0000L000<100001P000`40000600030@0000H000<10000
1`000`40000600030@0000H000<100001`000`40000100001P000`40000[00030@0002/000<10000
0P000`40000Q00030@0000<000<10000:`000`40000[00030@0002`000<100000@0006L000<10000
8@000`40000300030@0008`0001W00030@00024000<100000`000`40002<0000I`000`40000P0003
0@0000@000<10000S00006L000<1000080000`40000400030@0008`0001X00030@0001l000<10000
10000`40002<0000J0000`40000O00030@0000@000<10000S00006T000<100007@000`4000050081
S@0006T000<1000070000`40000600030@0008`0001Z00030@0001X000<100001`000`40002<0000
J`000`40000H00030@0000P000<10000S00006`000<100005P000`40000900030@0008`0001]0003
0@0001@000<100002P000`40002<0000KP020A800P4=00030@0008`0001`00030@0000d00P4?0003
0@0008`0001a00@12P000`40000?00030@0008`0001e00X14P000`40002<0000T@020Hd0002A0003
0@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0002A0003
0@0008`0002A00030@0008`0002A00030@0008`0002;00X1R`0008T00P4600040@0000<1R00008H0
0`4800030@0000@00P660000Q@000`40000900030@0000H000<10000P`0008@000<100002P000`40
000700030@000880002300030@0000/000<1000020000`4000210000PP000`40000<00030@0000T0
00<10000P000084000<100003@000`40000:00030@0007l0002100030@0000d000<100002`000`40
001n0000P0000`40000>008130000`40001n0000O`000`40000?00030@0000`000<10000O@0007l0
00<100003`000`40000<00030@0007d0001n00030@00010000<100003@000`40001l0000OP000`40
000@00030@0000d000<10000O00007h000<1000040000`40000>00030@0007/0001n00030@000100
00<100003P000`40001k0000O@000`40000A00030@0000h000<10000N`0007d000<1000020040@D0
00<100003P000`40001k0000O@000`40000:00030@0000@00P4?00030@0007/0001m00030@0000X0
00<1000010000`40000>00030@0007/0001m00030@0000X000<1000010000`40000>00030@0007/0
001X00`12@000`40000:00030@0000@000<100003P000`40001k0000I@030@`00P4800030@0000L0
0`4600030@0000h000<10000N`0006@000<100003`000`40000500030@0000T000<1000010000`40
000>00030@0007/0001R00814`020@D000<1000040000`40000=00030@0007`0001Q00030@0001D0
00<100000P000`40000@00030@0000`000<10000O@00060000<100005`000`40000200030@0000l0
00<1000030000`40001m0000H0000`40000H00030@00008000<100003P000`40000<00030@0007d0
001O00030@0001X000D100000@0@00813@000`40001m0000G`000`40000J00030@00008000<10000
3@000`40000;00030@0007h0001N00030@0001`000D100000@0?00030@0000T00P610000G@000`40
000N00050@0000403P000`40000800030@000840001M00030@0001h000<100000P020@`000<10000
1`000`4000220000G0000`40000P00030@0000<000<100002@000`40000700030@000880001L0003
0@00020000<1000010020@T000<100001@020HD0001L00030@00020000<100001P020@L000<10000
0P030HL0001K00030@00028000<100001`030@@01@6:0000F`000`40000R00030@0000X01P6=0000
F`000`40000R00030@0000h000<10000S00005/000<100008P000`40000>00030@0008`0001K0003
0@00028000<100003P000`40002<0000F`000`40000R00030@0000h000<10000S00005H000<10000
0P000`40000R00030@0000<000<1000020000`40002<0000E`020@<000<1000080000`4000020081
2`000`40002<0000F@020@030040024000<100000P4=00030@0008`0001K00818P030@l000<10000
S00005d00P4O00814@020Hd0001M00040@010A/00`4B00030@0008`0001N00030@0000815`020@03
0040018000<10000S00005h000<100000P020A<00P4200030@00018000<10000S00005l000<10000
0`020@l00P4300030@0001<000<10000S00005l000<100001@020@/00P4500030@0001<000<10000
S000060000<100001P020@L00P4600030@0001@000<10000S00006400P4800810`020@H00P4G0003
0@0008`0001S00030@0000L00`4700030@0001L000<10000S00006@00P4700030@0100D00P4J0081
S@0006H00`4200810`020@03004101`000<10000S00006T02`4M00030@0008`0001W00812`020A/0
00<10000S00006D00P4?00816@000`40002<0000H`020A<00P4G00030@0008`0001Q00815`020AD0
00<10000S00005l00P4K00814`000`40002<0000G@020Al00P4800@11@000`40002<0000F`020B<0
0P4600030@0000H000<10000S00005T00P4W00811@000`4000050081S@0005L00P4[008110000`40
000400030@0008`0001F00030@0002d000<100000P000`40000300030@0008`0002800040@000@D0
00<10000S00008T00P4600030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0
002A00030@0008`0002A0081S@00094000<10000S000094000<10000S000094000<10000S0000940
00<10000S000094000<10000S000094000<10000S000094000<10000S000094000<10000S0000940
0P6=0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@000`40
002<0000T@000`40002<0000T@000`40002<0000T@000`40002<0000T@020Hd0002A00030@0008`0
002A00030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0002A00030@0008`0
002A00030@0008`0002A00030@0008`0002A00030@0008`0002A0081S@00094000<10000S0000940
00<10000S000094000<10000S000094000<10000S000094000<10000S000094000<10000S0000940
00<10000S00008T00P4600030@0008`0002800040@000@D00P6=0000R`000`40002B0000R@020ID0
002;00030@000980002800040@000I@00000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-3.12952, -3.09715, 0.0215829, 
  0.0215829}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\

{num,den}   = {Numerator[Crat1[s]],Denominator[Crat1[s]]};
zeros       = s /. Solve[num==0,s]

poles       = s /. Solve[den==0,s]\
\>", "Input"],

Cell[BoxData[
    \({\(-5.80960307253229402`\), 
      \(-0.759910809019448585`\) - 1.44744250580916933`\ I, 
      \(-0.759910809019448585`\) + 1.44744250580916933`\ I, 
      \(-0.514252382910841276`\), \(-0.0200000000000000266`\) - 1.`\ I, 
      \(-0.0200000000000000266`\) + 1.`\ I}\)], "Output"],

Cell[BoxData[
    \({\(-25.0000000000000088`\), 
      \(-5.93594589869387068`\) - 6.7692356211391127`\ I, 
      \(-5.93594589869387068`\) + 6.7692356211391127`\ I, 
      \(-5.26811350579824289`\), 
      \(-0.738991520877942953`\) - 1.78596925273950546`\ I, 
      \(-0.738991520877942953`\) + 1.78596925273950546`\ I}\)], "Output"]
}, Open  ]],

Cell["\<\
We now attempt a simple elimination of pole-zero pairs in the \
compensator to obtain a low order compensator.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\

Crat2[s_]   = Crat1[0] *
              (1 - s/zeros[[4]])(1 - s/zeros[[5]])(1 - s/zeros[[6]])/
                   (  (1 - s/poles[[1]])(1 - s/poles[[2]])(1 - s/poles[[3]]))
                   \
\>", "Input"],

Cell[BoxData[
    \(\(1.03656747353785316`\ 
        \((1 + \((
                \(0.0199920031987205426`\[InvisibleSpace]\) - 
                  0.999600159936025534`\ I)\)\ s)\)\ 
        \((1 + \((
                \(0.0199920031987205426`\[InvisibleSpace]\) + 
                  0.999600159936025534`\ I)\)\ s)\)\ 
        \((1 + 1.94457047401445954`\ s)\)\)\/\(\((
          1 + 0.0399999999999999857`\ s)\)\ 
        \((1 + \((
                \(0.0732308416361117586`\[InvisibleSpace]\) - 
                  0.0835110073827056176`\ I)\)\ s)\)\ 
        \((1 + \((
                \(0.0732308416361117586`\[InvisibleSpace]\) + 
                  0.0835110073827056176`\ I)\)\ s)\)\)\)], "Output"]
}, Open  ]],

Cell["\<\
At low frequencies, the magnitude of the compensators is nearly \
identical\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\

FigCrat1Crat2 = Plot[{Abs[Crat1[I w]],Abs[Crat2[I w]]},{w,0,2}];\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0147151 0.233553 [
[.2619 .00222 -9 -9 ]
[.2619 .00222 9 0 ]
[.5 .00222 -3 -9 ]
[.5 .00222 3 0 ]
[.7381 .00222 -9 -9 ]
[.7381 .00222 9 0 ]
[.97619 .00222 -3 -9 ]
[.97619 .00222 3 0 ]
[.01131 .13149 -18 -4.5 ]
[.01131 .13149 0 4.5 ]
[.01131 .24827 -6 -4.5 ]
[.01131 .24827 0 4.5 ]
[.01131 .36504 -18 -4.5 ]
[.01131 .36504 0 4.5 ]
[.01131 .48182 -6 -4.5 ]
[.01131 .48182 0 4.5 ]
[.01131 .5986 -18 -4.5 ]
[.01131 .5986 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .01472 m
.2619 .02097 L
s
[(0.5)] .2619 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(1)] .5 .00222 0 1 Mshowa
.7381 .01472 m
.7381 .02097 L
s
[(1.5)] .7381 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(2)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13149 m
.03006 .13149 L
s
[(0.5)] .01131 .13149 1 0 Mshowa
.02381 .24827 m
.03006 .24827 L
s
[(1)] .01131 .24827 1 0 Mshowa
.02381 .36504 m
.03006 .36504 L
s
[(1.5)] .01131 .36504 1 0 Mshowa
.02381 .48182 m
.03006 .48182 L
s
[(2)] .01131 .48182 1 0 Mshowa
.02381 .5986 m
.03006 .5986 L
s
[(2.5)] .01131 .5986 1 0 Mshowa
.125 Mabswid
.02381 .03807 m
.02756 .03807 L
s
.02381 .06143 m
.02756 .06143 L
s
.02381 .08478 m
.02756 .08478 L
s
.02381 .10814 m
.02756 .10814 L
s
.02381 .15485 m
.02756 .15485 L
s
.02381 .1782 m
.02756 .1782 L
s
.02381 .20156 m
.02756 .20156 L
s
.02381 .22491 m
.02756 .22491 L
s
.02381 .27162 m
.02756 .27162 L
s
.02381 .29498 m
.02756 .29498 L
s
.02381 .31833 m
.02756 .31833 L
s
.02381 .34169 m
.02756 .34169 L
s
.02381 .3884 m
.02756 .3884 L
s
.02381 .41175 m
.02756 .41175 L
s
.02381 .43511 m
.02756 .43511 L
s
.02381 .45847 m
.02756 .45847 L
s
.02381 .50518 m
.02756 .50518 L
s
.02381 .52853 m
.02756 .52853 L
s
.02381 .55189 m
.02756 .55189 L
s
.02381 .57524 m
.02756 .57524 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .25681 m
.02499 .25681 L
.02605 .25681 L
.02729 .25682 L
.02846 .25683 L
.03053 .25685 L
.03279 .25688 L
.03527 .25693 L
.0379 .25699 L
.04262 .25713 L
.05205 .25754 L
.06244 .25815 L
.0842 .25996 L
.10458 .26214 L
.14335 .26678 L
.16306 .26892 L
.17348 .26987 L
.17919 .27033 L
.18458 .2707 L
.18976 .271 L
.19538 .27126 L
.19786 .27135 L
.20048 .27143 L
.20271 .27148 L
.20517 .27153 L
.20645 .27154 L
.20766 .27156 L
.20875 .27156 L
.20994 .27157 L
.21123 .27157 L
.21188 .27156 L
.2126 .27156 L
.21386 .27155 L
.21505 .27154 L
.21618 .27152 L
.21737 .27149 L
.2195 .27144 L
.22178 .27137 L
.22428 .27127 L
.22957 .27099 L
.23456 .27064 L
.2439 .26972 L
.25395 .26833 L
.26485 .26633 L
.28472 .26119 L
.29474 .25781 L
.30576 .25342 L
.32684 .24296 L
.3461 .2308 L
.38493 .19787 L
Mistroke
.42469 .15094 L
.46293 .09185 L
.47394 .07253 L
.48429 .05419 L
.48882 .04672 L
.49359 .04009 L
.49479 .03879 L
.49609 .03764 L
.49683 .03711 L
.4975 .03673 L
.49882 .03626 L
.50001 .03617 L
.50129 .03645 L
.50251 .03707 L
.50363 .03793 L
.50481 .03909 L
.50588 .04037 L
.5083 .04389 L
.51267 .05181 L
.52264 .07392 L
.54281 .12701 L
.58249 .25436 L
.62065 .40744 L
.66126 .61041 L
Mfstroke
.66126 .61041 m
.66249 .61803 L
s
.02381 .25681 m
.02499 .25681 L
.02605 .25681 L
.02729 .25682 L
.02846 .25683 L
.03053 .25685 L
.03279 .25689 L
.03527 .25693 L
.0379 .257 L
.04262 .25714 L
.05205 .25756 L
.06244 .25819 L
.0842 .26006 L
.10458 .26232 L
.14335 .26717 L
.16306 .26945 L
.17348 .27047 L
.18458 .27138 L
.19014 .27174 L
.19538 .27202 L
.20008 .27221 L
.20273 .2723 L
.20517 .27236 L
.20627 .27239 L
.20746 .27241 L
.20848 .27242 L
.2096 .27243 L
.21091 .27244 L
.21214 .27245 L
.21335 .27245 L
.21446 .27245 L
.21563 .27244 L
.21673 .27243 L
.21793 .27241 L
.21923 .27239 L
.22186 .27233 L
.22428 .27225 L
.22902 .27203 L
.23409 .27172 L
.23887 .27133 L
.24321 .2709 L
.25386 .26952 L
.26391 .26774 L
.28392 .26274 L
.30238 .25618 L
.32114 .24738 L
.34178 .23498 L
.38212 .20147 L
.42094 .1562 L
.46222 .09276 L
Mistroke
.47221 .07523 L
.48285 .05634 L
.48802 .04765 L
.49056 .04378 L
.49289 .04064 L
.49502 .0383 L
.49609 .03737 L
.49726 .03659 L
.49844 .03609 L
.4997 .03589 L
.50089 .03604 L
.50198 .03647 L
.50311 .03719 L
.50435 .03828 L
.50688 .04133 L
.5114 .04878 L
.52163 .07042 L
.54022 .11701 L
.58048 .23679 L
.61922 .37471 L
.66041 .54692 L
Mfstroke
.66041 .54692 m
.67508 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@000?l08@0004l0
0P4400030@0000800`4g00@1=`040@<000<100000P030CP010460000CP001040004;00030@0003H0
00<10000>0000`40000:00030@0003D000<100001`0004h000@100012`000`40000f00030@0003P0
00<100002P000`40000f00030@0000H0001>00040@000@P00`4i00030@0003P000<100001`030CX0
00<100001@0001D000<10000=P001040004800030@0003T000<10000>0000`40000700030@0003/0
00<10000100001D000<10000=P001040004800030@0003L00`4h00<12@000`40000h00040@000@H0
000E00030@0003L00P4900@1>0000`40000h00030@0000L0104h00811`0001D000<10000o`090000
3P3o0A<1000E00030@0000X000<100002P000`40000:00030@0000X000<100002P000`40000:0003
0@0000X000<100002P000`40000:00030@0000T000<100002P000`40000:00030@0000X000<10000
2P000`40000:00030@0000X000<100002P000`40000:00030@0000X000<100002P000`4000050000
5@000`40000n00030@0003d000<10000?P000`40000n00030@0000D0000E00030@000?l02@0001D0
00<10000o`0900005@000`40003o00T0000E00030@0007d00`6800005@020Gh000@10001Q`0001D0
00<10000O0000`40000200030@0008@0000E00030@0007/000<1000010000`40002300005@000`40
001k00030@0000@000<10000P`0001D000<10000NP000`40000600030@000880000E00030@0007T0
0P4800030@000880000E0081NP000`4000070081P`0001D000<10000N0020@X000<10000P@0001D0
00<10000N0000`4000090081PP0001D000<10000M`020@/00P6200005@000`40001f00813@000`40
002000005@000`40001f00030@0000`00P6100005@020GH00P4?00030@0007l0000E00030@0007D0
00<100003P000`40001o00005@000`40001d00030@0000l00P6000005@000`40001c00030@000140
00<10000OP0001D000<10000L`000`40000A00030@0007h0000E00030@00078000<100004P020Gl0
000E00030@00078000<100004`000`40001m00005@020G8000<1000050000`40001m00005@000`40
001`00030@0001H000<10000O00001D000<10000L0000`40000F00030@0007`0000E00030@0006l0
00<100005`000`40001l00005@000`40001^00816P000`40001k00000P020@@000<100000P030@D0
00<10000KP000`40000I00030@0007/000001@01000100/000<100000P030Fd000<100006P000`40
001k000000D00@000@0;00030@00008000<10000K@000`40000K00030@0007X000001@01000100P0
0`4500030@0006`000<1000070000`40001j000000D00@000@0800030@0000D000<10000J`000`40
000M00030@0007X000001@01000100P000<100001@000`40001[00030@0001h000<10000N@000080
0P4900@110000`40001Z00030@0001l000<10000N@0001D00P5Z00818@000`40001i00005@000`40
001X00818`000`40001h00005@000`40001W008190000`40001h00005@000`40001W00030@0002<0
00<10000N00001D000<10000IP000`40000U00030@0007L0000E00030@0006D000<100009P000`40
001g00005@000`40001T00030@0002L000<10000M`0001D00P5T0081:P000`40001f00005@000`40
001R0081:`000`40001f00005@000`40001R00030@0002X000<10000MP0001D000<10000H@000`40
000/00030@0007D0000E00030@00060000<10000;@000`40001e00005@000`40001O00030@0002h0
00<10000M@0001D00P5O00030@0002l00P5f00005@000`40001M00030@00034000<10000M00001D0
00<10000F`020C@000<10000M00001D000<10000FP020CD00P5e00005@000`40001I00030@0003H0
00<10000L`0001D000<10000F0000`40000g00030@0007<0000E0081F0000`40000h0081M00001D0
00<10000E@020C`000<10000LP0001D000<10000E0020Cd000<10000LP0001D000<10000DP020Cl0
0P5c00005@000`40001@00<1@@000`40001a00003@040@@000<10000C`020D<000<10000L@0000l0
00<100000`000`40001=0081A@020G80000?00030@0000<00`5;00<1A`000`40001`00003`000`40
000300030@0004T00`5900030@000700000?00030@0000<03P4k00@1B`000`40001`00003@030@D0
00<100002`0;0Bd01@5=0081L@0000l000<100000`000`40000A00X19@060E0000<10000K`0001D0
00<100006@0<0AD01`5E00030@0006l0000E00819P0F0E/00P5`00005@000`40002F0081L00001D0
00<10000U`000`40001^00005@000`40002G0081K`0001D000<10000U`020Fl0000E00030@0009L0
0P5_00005@020IT000<10000K@0001D000<10000V0020Fh0000E00030@0009P00P5^00005@000`40
002H0081KP0001D000<10000V@020Fd0000E00030@0009T00P5]00005@000`40002I0081K@0001D0
0P6J00030@0106`0000E00030@0009X00P5/00005@000`40002J0081K00001D000<10000VP020F`0
000E00030@0009X000<10040J`0001D000<10000VP000`400@1[00005@020I`00P5[00005@000`40
002K00030@0106X0000E00030@0009/000<10040JP0001D000<10000V`000`400@1Z00005@000`40
002L00030@0106T000000`010@020@<000<100000P030@D000<10000W0000`400@1Y00000`000`40
000:00030@0000800`6L00030@0106T0000300030@0000X000<100000P000`40002L00030@0106T0
000300030@0000L00`4500030@0009d000<10040J00000<000<100001`000`40000500030@0009d0
00<10040J000000400410@T000<100001@000`40002M00030@0106P0000300030@0000L010440003
0@0009d000<10040J00001D000<10000WP000`400@1W00005@020Il000<10040I`0001D000<10000
WP000`400@1W00005@000`40002N00030@0106L0000E00030@0009l000<10040IP0001D000<10000
W`000`400@1V00005@000`40002O00030@0106H0000E0081X0000`400@1V00005@000`40002O0004
0@000FD0000E00030@000:0000<10040I@0001D000<10000X0000`400@1U00005@000`40002P0003
0@0106D0000E00030@000:0000@10001I00001D000<10000X0001040005T00005@020J8000<10040
I00001D000<10000X@000`400@1T00005@000`40002Q00030@0106@0000E00030@000:4000@10001
H`0001D000<10000X@001040005S00005@000`40002R00030@0106<0000E0081X`000`400@1S0000
5@000`40002R00040@000F80000E00030@000:8000@10001HP0001D000<10000XP001040005R0000
5@000`40002S00030@010680000=00@110000`40002S00040@000F40000=00030@0000D00`6S0004
0@000F40000>00030@0000@000<10000X`001040005Q00003`000`40000300030@000:<000@10001
H@00010000<100000P000`40002T00030@010640000=00040@000@@000<10000Y0001040005P0000
3P020@D000<10000Y0001040005P00005@000`40002T00040@000F00000E0081Y@001040005P0000
5@000`40002U00040@000El0000E00030@000:D000@10001G`0001D000<10000Y@001040005O0000
5@000`40002U00040@000El0000E00030@000:D000D100000@1N00005@020JL000@10001GP0001D0
00<10000YP001040005N00005@000`40002V00040@000Eh0000E00030@000:H000D100000@1M0000
5@000`40002V00050@000040G@0001D000<10000Y`001040005M00005@020JP000@10001G@0001D0
00<10000Y`001040005M00005@000`40002W00050@000040G00001D000<10000Y`001@40000105`0
000E00030@000:P000@10001G00001D000<10000Z0001040005L00005@000`40002X00040@000E`0
000E0081Z@001@40000105/0000E00030@000:P000D100000@1K00005@000`40002Y00040@000E/0
000E00030@000:T000@10001F`0001D000<10000Z@001@40000105X000000`010@020@<000<10000
0P030@D000<10000Z@001@40000105X000000`01000=00030@0000800`6Y00050@000040FP000080
00<100002`000`40000200030@000:X000@10001FP0000<000<100001`030@D000<10000ZP001040
005J000010000`40000600030@0000D000<10000ZP001@40000105T000001@01000100P000<10000
1@000`40002[00040@000ET0000200812@040@@000<10000Z`001040005I0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.165192, -0.198214, 
  0.00771848, 0.0157372}}]
}, Open  ]],

Cell["\<\
The low order compensator is used to produce a corresponding closed \
loop function\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\

Trat2[s_]   = Together[ p[s] Crat2[s]/(1 + p[s] Crat2[s])] //Chop 

FigCrat2ZP  = PlotZP[Crat2[s],s];\
\>", "Input"],

Cell[BoxData[
    \(\((0.0373164290473627069`\ 
          \((1 + \((
                  \(0.0199920031987205426`\[InvisibleSpace]\) - 
                    0.999600159936025534`\ I)\)\ s)\)\ 
          \((1 + \((
                  \(0.0199920031987205426`\[InvisibleSpace]\) + 
                    0.999600159936025534`\ I)\)\ s)\)\ \((25 + s)\)\ 
          \((1 + 1.94457047401445954`\ s)\))\)/
      \((\(0.932910726184067584`\[InvisibleSpace]\) + 
          1.88872859052027397`\ s + 2.07952960951491894`\ s\^2 + 
          2.08012449177713465`\ s\^3 + 1.09819646926370072`\ s\^4 + 
          0.187683166916635997`\ s\^5 + 0.0182150508029298219`\ s\^6 + 
          0.000493473780832303532`\ s\^7)\)\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.961538 0.034965 0.5 0.0671487 [
[.08741 .4875 -9 -9 ]
[.08741 .4875 9 0 ]
[.26224 .4875 -9 -9 ]
[.26224 .4875 9 0 ]
[.43706 .4875 -9 -9 ]
[.43706 .4875 9 0 ]
[.61189 .4875 -9 -9 ]
[.61189 .4875 9 0 ]
[.78671 .4875 -6 -9 ]
[.78671 .4875 6 0 ]
[.94904 .09711 -12 -4.5 ]
[.94904 .09711 0 4.5 ]
[.94904 .23141 -12 -4.5 ]
[.94904 .23141 0 4.5 ]
[.94904 .3657 -12 -4.5 ]
[.94904 .3657 0 4.5 ]
[.94904 .6343 -6 -4.5 ]
[.94904 .6343 0 4.5 ]
[.94904 .76859 -6 -4.5 ]
[.94904 .76859 0 4.5 ]
[.94904 .90289 -6 -4.5 ]
[.94904 .90289 0 4.5 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.08741 .5 m
.08741 .50625 L
s
[(-25)] .08741 .4875 0 1 Mshowa
.26224 .5 m
.26224 .50625 L
s
[(-20)] .26224 .4875 0 1 Mshowa
.43706 .5 m
.43706 .50625 L
s
[(-15)] .43706 .4875 0 1 Mshowa
.61189 .5 m
.61189 .50625 L
s
[(-10)] .61189 .4875 0 1 Mshowa
.78671 .5 m
.78671 .50625 L
s
[(-5)] .78671 .4875 0 1 Mshowa
.125 Mabswid
.12238 .5 m
.12238 .50375 L
s
.15734 .5 m
.15734 .50375 L
s
.19231 .5 m
.19231 .50375 L
s
.22727 .5 m
.22727 .50375 L
s
.2972 .5 m
.2972 .50375 L
s
.33217 .5 m
.33217 .50375 L
s
.36713 .5 m
.36713 .50375 L
s
.4021 .5 m
.4021 .50375 L
s
.47203 .5 m
.47203 .50375 L
s
.50699 .5 m
.50699 .50375 L
s
.54196 .5 m
.54196 .50375 L
s
.57692 .5 m
.57692 .50375 L
s
.64685 .5 m
.64685 .50375 L
s
.68182 .5 m
.68182 .50375 L
s
.71678 .5 m
.71678 .50375 L
s
.75175 .5 m
.75175 .50375 L
s
.82168 .5 m
.82168 .50375 L
s
.85664 .5 m
.85664 .50375 L
s
.89161 .5 m
.89161 .50375 L
s
.92657 .5 m
.92657 .50375 L
s
.05245 .5 m
.05245 .50375 L
s
.01748 .5 m
.01748 .50375 L
s
.9965 .5 m
.9965 .50375 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.96154 .09711 m
.96779 .09711 L
s
[(-6)] .94904 .09711 1 0 Mshowa
.96154 .23141 m
.96779 .23141 L
s
[(-4)] .94904 .23141 1 0 Mshowa
.96154 .3657 m
.96779 .3657 L
s
[(-2)] .94904 .3657 1 0 Mshowa
.96154 .6343 m
.96779 .6343 L
s
[(2)] .94904 .6343 1 0 Mshowa
.96154 .76859 m
.96779 .76859 L
s
[(4)] .94904 .76859 1 0 Mshowa
.96154 .90289 m
.96779 .90289 L
s
[(6)] .94904 .90289 1 0 Mshowa
.125 Mabswid
.96154 .13068 m
.96529 .13068 L
s
.96154 .16426 m
.96529 .16426 L
s
.96154 .19783 m
.96529 .19783 L
s
.96154 .26498 m
.96529 .26498 L
s
.96154 .29855 m
.96529 .29855 L
s
.96154 .33213 m
.96529 .33213 L
s
.96154 .39928 m
.96529 .39928 L
s
.96154 .43285 m
.96529 .43285 L
s
.96154 .46643 m
.96529 .46643 L
s
.96154 .53357 m
.96529 .53357 L
s
.96154 .56715 m
.96529 .56715 L
s
.96154 .60072 m
.96529 .60072 L
s
.96154 .66787 m
.96529 .66787 L
s
.96154 .70145 m
.96529 .70145 L
s
.96154 .73502 m
.96529 .73502 L
s
.96154 .80217 m
.96529 .80217 L
s
.96154 .83574 m
.96529 .83574 L
s
.96154 .86932 m
.96529 .86932 L
s
.96154 .06353 m
.96529 .06353 L
s
.96154 .02996 m
.96529 .02996 L
s
.96154 .93647 m
.96529 .93647 L
s
.96154 .97004 m
.96529 .97004 L
s
.25 Mabswid
.96154 0 m
.96154 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
68.252 18.8327 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
69.502 16.3035 1 0 365.73 arc
setmatrix s
newpath
matrix currentmatrix
0.0138246 0.0265496 scale
69.502 21.3619 1 0 365.73 arc
setmatrix s
.06923 .48182 m
.10559 .51818 L
s
.06923 .51818 m
.10559 .48182 L
s
.73581 .02727 m
.77217 .06364 L
s
.73581 .06364 m
.77217 .02727 L
s
.73581 .93636 m
.77217 .97273 L
s
.73581 .97273 m
.77217 .93636 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000180P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003C00030@0000L000<10000<`000`40000:0000e0000`40
000500030@0003@00`4:0000e@000`40000300030@0003D000<100002P000=H000D100000@0h0003
0@0000X0003G00030@0103T000<100002P000=P000<10000>0000`40000:0000f0000`40000h0003
0@0000X0003G00030@0103T000<100002P000=H000D100000@0h00030@0000X0003E00030@0000<0
00<10000=@000`40000:0000e0000`40000500030@0003@000<100002P000=<000<100001`000`40
000c00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0=00811@000`40
000:0000o`0<00040@000@@00`4:0000o`0<00040@000@@000<100002P000?l01P040@800`450003
0@0000X0003o00`000<100001@000`40000:0000o`0<00030@0000D000<100002P000?l03@030@@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050030@X0003o01@000<100002P000?l050000`40000:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o00h000<100000`000`40000:0000o`0>00030@0000<0
0`4:0000o`0;00D110000`40000:0000o`0600@100D00@000@0500030@0000X0003o00`000<10040
1@000`40000:0000o`0=00811@000`40000:0000o`0>00030@0000<000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00<12P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@00`4:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0<00@1
10000`40000:0000o`0<00030@0000D00`4:0000o`0=00030@0000@000<100002P000?l01P040@@0
00<100000`000`40000:0000o`0?00030@00008000<100002P000?l030001040004400030@0000X0
003o00d00P4500030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050030@X0003o01@000<100002P000?l050020@/0003o01<00P400`01
00090000o`0B00050@0100402P000?l04P000`400@0200030@0000L0003o014000@100010`000`40
00060000o`0A00040@000@<000<100001P000?l040001@40000100<000<100001P000?l040001@40
000100<000<100001P000?l04000104000030@03004000L0003o014000@100010`000`4000060000
o`0A00040@000@<000<100001P000?l04@001040004300030@0000H0003o014000@100010`000`40
00060000o`0A00040@000@8000<100001`000?l04P050@X0003o01@000<100002P000?l050000`40
000:0000o`0D00<12P0001L0104200<1:@040@<00P4Y00@10P030BT010430081;0030Bh000<10000
2P0001L000<100001P000`40000V00030@0000<000@10001:P000`40000400030@0002P000D10000
0@0200030@0002`000<100009`001@40000100`0000H00030@0000D000<100009`000`4000020004
0@000BX000<1000010000`40000X00050@0000400P000`40000/00030@0002D00P401@01000100`0
000A00@11000104000030B<0104400050@0000400P000`40000P00@11000104000030B<010440005
0@0000400P000`40000S00@10P030BL000D100000@0200030@0000X0000C00030@0000@000<10000
0P4]00040@000@8000<10000:0001@40000102d000D100000@0200030@0002T000<100009`000`40
000200030@0100`0000D00040@000@8000@10001:`001040004200040@000BP00`4300030@0002T0
0`4300040@000B/000<100009P000`40000300030@0100`0000E00030@0000810P050BT00P440081
:`00104000040BX000<100000P020B`0104U00030@0000@00P4<00005P000`40000200030@000>d0
00<1000010020@`0000G00040@000Nl000<100001@020@`0003o0B41000500030@0000L000<10000
1@0010400@4800030@0000L000<100001`000`40000700030@0000L000<100001`000`4000070003
0@0000L000<100001`000`40000700030@0000L000<100001`000`40000700030@0000L000<10000
1`000`40000700030@0000L000<100001`000`40000700030@0000L000<100001`000`4000070003
0@0000L000<100001`020@L00P4900410@0100005P001P4000400Bl000<10000;`000`40000_0003
0@0002l000<100009`000`4000040081300001D000<1000010000`40003/00030@0000@00P4<0000
50000`40000600030@000>/000<1000010020@`0000C00030@0000P000<10000jP000`4000030003
0@0100`0003o00d000D100000@0200030@0000X0003o00h000<100400`000`40000:0000o`0?0003
0@00008000<100002P000?l050000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D0081
2`000?l04`020@03004000T0003o018000<100400P000`4000070000o`0B00030@01008000<10000
1`000?l04@001040004300030@0000H0003o014000@100010`000`4000060000o`0@00050@000040
0`000`4000060000o`0@00040@0000<100<00@001`000?l040001@40000100<000<100001P000?l0
4@001040004300030@0000H0003o014000@100010`000`4000060000o`0A00040@000@<000<10000
1P000?l04@001040004200030@0000L0003o018000<100400P000`4000070000o`0C00@12P000?l0
50000`40000:0000o`0D00030@0000X0003o01@00`4:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000
o`0D00030@0000X0003o00`0104400030@0000X0003o00`000<100001@000`40000:0000o`0=0003
0@0000@00`4:0000o`0>00030@0000<000<100002P000?l03`000`40000200030@0000X0003o00`0
00@1000110000`40000:0000o`0=00811@000`40000:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@00`4:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00<12P000?l0
50000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D0003
0@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@00`4:0000
o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l03P000`40
000300030@0000X0003o00h000<100000`030@X0003o00/01@4400030@0000X0003o00/000@10001
1@000`40000:0000o`0<00030@0100D000<100002P000?l03@020@D000<100002P000?l03P000`40
000300030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050030@X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l0
50000`40000:0000o`0D00<12P000?l050000`40000:0000o`0D00030@0000X0003o01@000<10000
2P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000
o`0D00030@0000X0003o01@000<100002P000?l050030@X0003o01@000<100002P000?l050000`40
000:0000o`0D00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l03@020@D000<100002P000?l030001040004400030@0000X0003o00`0
00@1000110030@X0003o00`00`4500030@0000X0003o00`000<100001@000`40000:0000o`0<0003
0@0000D000<100002P000?l03@030@@000<100002P000?l050000`40000:0000o`0D00030@0000X0
003o01@000<100002P000?l050000`40000:0000d`000`40000700030@0003<00`4:0000e0000`40
000500030@0003@000<100002P000=D000<100000`000`40000e00030@0000X0003F00050@000040
>0000`40000:0000e`000`400@0i00030@0000X0003H00030@0003P000<100002P000=P000<10000
>0000`40000:0000e`000`400@0i00030@0000X0003F00050@000040>0000`40000:0000e@000`40
000300030@0003D000<100002P000=@000<100001@000`40000d00<12P000=<000<100001`000`40
000c00030@0000X0003o01@000<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@0
00<100002P000?l050000`40000:0000o`0D00030@0000X0003o01@000<100002P000001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-27.5002, -7.44624, 0.0996527, 
  0.0518901}}]
}, Open  ]],

Cell["\<\
The last thing to do is to verify that the step response has the \
desired form\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\

step2[t_]   = Chop[Simplify[InverseLaplaceTransform[ Trat2[s]/s,s,t] ]];
FigStepT2   = Plot[step2[t],{t,0,14}];

\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0680272 0.0147151 0.495946 [
[.15986 .00222 -3 -9 ]
[.15986 .00222 3 0 ]
[.29592 .00222 -3 -9 ]
[.29592 .00222 3 0 ]
[.43197 .00222 -3 -9 ]
[.43197 .00222 3 0 ]
[.56803 .00222 -3 -9 ]
[.56803 .00222 3 0 ]
[.70408 .00222 -6 -9 ]
[.70408 .00222 6 0 ]
[.84014 .00222 -6 -9 ]
[.84014 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[.01131 .1139 -18 -4.5 ]
[.01131 .1139 0 4.5 ]
[.01131 .21309 -18 -4.5 ]
[.01131 .21309 0 4.5 ]
[.01131 .31228 -18 -4.5 ]
[.01131 .31228 0 4.5 ]
[.01131 .41147 -18 -4.5 ]
[.01131 .41147 0 4.5 ]
[.01131 .51066 -6 -4.5 ]
[.01131 .51066 0 4.5 ]
[.01131 .60985 -18 -4.5 ]
[.01131 .60985 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15986 .01472 m
.15986 .02097 L
s
[(2)] .15986 .00222 0 1 Mshowa
.29592 .01472 m
.29592 .02097 L
s
[(4)] .29592 .00222 0 1 Mshowa
.43197 .01472 m
.43197 .02097 L
s
[(6)] .43197 .00222 0 1 Mshowa
.56803 .01472 m
.56803 .02097 L
s
[(8)] .56803 .00222 0 1 Mshowa
.70408 .01472 m
.70408 .02097 L
s
[(10)] .70408 .00222 0 1 Mshowa
.84014 .01472 m
.84014 .02097 L
s
[(12)] .84014 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(14)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05782 .01472 m
.05782 .01847 L
s
.09184 .01472 m
.09184 .01847 L
s
.12585 .01472 m
.12585 .01847 L
s
.19388 .01472 m
.19388 .01847 L
s
.22789 .01472 m
.22789 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.32993 .01472 m
.32993 .01847 L
s
.36395 .01472 m
.36395 .01847 L
s
.39796 .01472 m
.39796 .01847 L
s
.46599 .01472 m
.46599 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.53401 .01472 m
.53401 .01847 L
s
.60204 .01472 m
.60204 .01847 L
s
.63605 .01472 m
.63605 .01847 L
s
.67007 .01472 m
.67007 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.77211 .01472 m
.77211 .01847 L
s
.80612 .01472 m
.80612 .01847 L
s
.87415 .01472 m
.87415 .01847 L
s
.90816 .01472 m
.90816 .01847 L
s
.94218 .01472 m
.94218 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .1139 m
.03006 .1139 L
s
[(0.2)] .01131 .1139 1 0 Mshowa
.02381 .21309 m
.03006 .21309 L
s
[(0.4)] .01131 .21309 1 0 Mshowa
.02381 .31228 m
.03006 .31228 L
s
[(0.6)] .01131 .31228 1 0 Mshowa
.02381 .41147 m
.03006 .41147 L
s
[(0.8)] .01131 .41147 1 0 Mshowa
.02381 .51066 m
.03006 .51066 L
s
[(1)] .01131 .51066 1 0 Mshowa
.02381 .60985 m
.03006 .60985 L
s
[(1.2)] .01131 .60985 1 0 Mshowa
.125 Mabswid
.02381 .03951 m
.02756 .03951 L
s
.02381 .06431 m
.02756 .06431 L
s
.02381 .08911 m
.02756 .08911 L
s
.02381 .1387 m
.02756 .1387 L
s
.02381 .1635 m
.02756 .1635 L
s
.02381 .1883 m
.02756 .1883 L
s
.02381 .23789 m
.02756 .23789 L
s
.02381 .26269 m
.02756 .26269 L
s
.02381 .28749 m
.02756 .28749 L
s
.02381 .33708 m
.02756 .33708 L
s
.02381 .36188 m
.02756 .36188 L
s
.02381 .38667 m
.02756 .38667 L
s
.02381 .43627 m
.02756 .43627 L
s
.02381 .46107 m
.02756 .46107 L
s
.02381 .48586 m
.02756 .48586 L
s
.02381 .53546 m
.02756 .53546 L
s
.02381 .56026 m
.02756 .56026 L
s
.02381 .58505 m
.02756 .58505 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01477 L
.02605 .01511 L
.02729 .01612 L
.02846 .01789 L
.02954 .02037 L
.03053 .02347 L
.03163 .0278 L
.03279 .03351 L
.0352 .04905 L
.03746 .06783 L
.04262 .1229 L
.05278 .25378 L
.06244 .36676 L
.06751 .41427 L
.07305 .45649 L
.08274 .51023 L
.08822 .53201 L
.09407 .55035 L
.09954 .56403 L
.10458 .57422 L
.10948 .5823 L
.11487 .58934 L
.12028 .59475 L
.12327 .59708 L
.12599 .59885 L
.1284 .60013 L
.13103 .60126 L
.1338 .60216 L
.13505 .60248 L
.13642 .60276 L
.13769 .60297 L
.13888 .60312 L
.14013 .60323 L
.14084 .60327 L
.14148 .6033 L
.14257 .60332 L
.14376 .6033 L
.14501 .60325 L
.14616 .60316 L
.14744 .60303 L
.14879 .60285 L
.15121 .60243 L
.15382 .60185 L
.15668 .60107 L
.16662 .59746 L
.18521 .5883 L
.22555 .56586 L
.26437 .54757 L
.28412 .54022 L
Mistroke
.30565 .53365 L
.34541 .52483 L
.36512 .52173 L
.38611 .51914 L
.40756 .51709 L
.42774 .51561 L
.44786 .51446 L
.46634 .51364 L
.48521 .51298 L
.50588 .51242 L
.52543 .51201 L
.54636 .51168 L
.56764 .51143 L
.58777 .51124 L
.60772 .51111 L
.62615 .51101 L
.64486 .51093 L
.66547 .51086 L
.68486 .51082 L
.70573 .51078 L
.72581 .51075 L
.74692 .51073 L
.7667 .51071 L
.78508 .5107 L
.80558 .51069 L
.82418 .51068 L
.8434 .51068 L
.86421 .51067 L
.88413 .51067 L
.90518 .51067 L
.9248 .51067 L
.94312 .51067 L
.97619 .51066 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl9UVoiVI09VI<iVIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<oiWo09Wo
<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV<laVI/aV
VLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/coVLcoc<co
ool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI<onII_nI
VOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<oooI_ooVOl08@000?l08@000?l08@0003T0104S0003
0@0002000P4S00817P040@<00P4L00@10P040A/0104400030@0000<0000i00030@0002@000<10000
7`001040004Q00040@000Al000D100000@0200030@0001/000D100000@0P00030@0000<000<10000
0`0003X000<1000080050B0000@100018@001040004O00050@0000400P000`40000K00030@000080
00<100007@000`4000050@@0000k00030@0001l000@100018@030B<00P4P00050@0000400P000`40
000K00030@0000<000<1000070001040004200030@0000<0000l00030@0001l000<100408@000`40
000R00040@000Al000D100000@0200030@0001/000<1000010000`40000K00070@0000400@050000
5P000`40000P00040@000B800P4Q00030@00028000@100017@030@<000@100016`030@<000@10001
6`030@@00P4500005P000`40000Q008190000`40000P00<18P020B0000<100000P020Ah000<10000
0P020Ah000<100000`000`40000300005P000`40003o00P0000F00030@000?l020000100o`4?0@80
000F00030@0100L000<100001P000`40000600030@0000H000<100001P000`40000600030@0000H0
00<100001`000`40000600030@0000H000<100001P000`40000600030@0000H000<100001P000`40
000700030@0000H000<100001P000`40000600030@0000H000<100001P000`40000700030@0000H0
00<100001P000`40000600030@0000H000<100001P000`40000600030@0000L000<100001P0001H0
00<100408P000`40000R00030@00024000<100008P000`40000Q00030@00028000<100008P000`40
000600005P001040007o00L0000F00040@000Ol01`0001H000@10001o`0700005P001040007o00L0
000F008100<00@00o`0600005P001040007o00L0000F00040@000Ol01`0001H000@10001o`070000
5P001040007o00L0000F00050@000040o`0600005P020@8000<10000o`0400005P001@4000010?l0
1P0001H000D100000@3o00H0000F00050@000040o`0600005P001@4000010?l01P0001H000D10000
0@3o00H0000F00050@000040o`0600005P020@8000<10000o`0400005P001@4000010?l01P0001H0
00<100000P000`40003o00<0000F00030@00008000<10000o`0300005P000`40000200030@000?l0
0`0000<00P4400030@000080104400030@00008000<10000o`0300000P001040004800030@0000D0
00<100000P000`40003o00<0000200040@000@T000<1000010030@8000<10000o`0300000P001040
004:00030@0000<000<100000P000`40003o00<0000200040@000@/000<100000P000`4000020003
0@000?l00`00008000@1000120001040004400030@00008000<10000o`0300000`020@X00P450003
0@00008000<10000o`0300005P000`40000200030@000?l00`0001H00P4300030@000?l00`0001H0
00<100000P000`40003o00<0000F00030@0000<000<10000o`0200005P000`40000300030@000?l0
0P0001H000<100000`000`40003o0080000F00030@0000<000<10000o`0200005P000`4000030003
0@000?l00P0001H00P4400030@000?l00P0001H000<100000`000`40003o0080000F00030@0000<0
00<10000o`0200005P000`40000300030@000?l00P0001H000<100000`000`40003o0080000F0003
0@0000<000<10000o`0200005P000`40000300030@000?l00P0001H00P4500030@000?l00@0001H0
00<1000010000`40003o0040000F00030@0000@000<10000o`0100005P000`40000400030@000?l0
0@0001H000<1000010000`40003o00400003008110000`40000400030@0000<000<1000010000`40
003o0040000200040@000@X000<100000`030@@000<10000o`0100000P001040004700D110000`40
000400030@000?l00@00008000@100011`001040004500030@0000@000<10000o`0100000P001040
004800030@0100D000<1000010000`40003o0040000200040@000@T00P4500030@0000@000<10000
o`0100000`020@/000<100000`000`40000400030@000?l00@0001H000<100001@000`40003o0000
5P020@H000<10000o`0001H000<100001@000`40003o00005P000`40000500030@000?l0000F0003
0@0000D000<10000o`0001H000<100001@000`40003o00005P000`40000500030@000?l0000F0003
0@0000D000<10000o`0001H00P4600030@000?l0000F00030@0000D000<10000o`0001H000<10000
1@000`40003o00005P000`40000600030@000?h0000F00030@0000H000<10000oP0001H000<10000
1P000`40003n00005P020@L000<10000oP0001H000<100001P000`40003n00005P000`4000060003
0@000?h0000F00030@0000H000<10000oP0001H000<100001P000`40003n00005P000`4000060003
0@000?h00003008110000`40000300811@000`40000600030@000?h0000200040@000@P000@10001
10030@L000<10000o@00008000@1000120001040004400030@0000L000<10000o@00008000@10001
20030@D000<100001`000`40003m00000P001040004800030@0000D000<100001`000`40003m0000
0P001040004800030@0000D000<100001`000`40003m00000`020@X00`4400030@0000L000<10000
o@0001H000<100001`000`40003m00005P020@P000<10000o@0001H000<100001`000`40003m0000
5P000`40000700030@000?d0000F00030@0000P000<10000o00001H000<1000020000`40003l0000
5P000`40000800030@000?`0000F00812@000`40003l00005P000`40000800030@000?`0000F0003
0@0000P000<10000o00001H000<1000020000`40003l00005P000`40000800030@000?`0000F0003
0@0000P000<10000o00001H000<1000020000`40003l00005P020@T000<10000o00001H000<10000
2@000`40003k00005P000`40000900030@000?/0000F00030@0000T000<10000n`0001H000<10000
2@000`40003k00000`020@@000<100000`020@D000<100002@000`40003k00000P00104000480004
0@000@@000<100002@000`40003k00000P001040004800040@000@@00`4900030@000?/000020004
0@000@T00P4500030@0000T000<10000n`00008000@1000120001040004400030@0000T000<10000
n`00008000@1000120001040004400030@0000X000<10000nP0000<00P4:00811@000`40000:0003
0@000?X0000F00030@0000X000<10000nP0001H00P4;00030@000?X0000F00030@0000X000<10000
nP0001H000<100002P000`40003j00005P000`40000;00030@000?T0000F00030@0000/000<10000
n@0001H000<100002`000`40003i00005P000`40000;00030@000?T0000F008130000`40003i0000
5P000`40000;00030@000?T0000F00030@0000/000<10000n@0001H000<1000030000`40003h0000
5P000`40000<00030@000?P0000F00030@0000`000<10000n00001H000<1000030000`40003h0000
5P020@d000<10000n00001H000<1000030000`40003h00005P000`40000<00030@000?P0000F0003
0@0000d000<10000m`0001H000<100003@000`40003g00003P040@@000<100003@000`40003g0000
40000`40000300<13@000`40002306`12000010000<100000`000`40000=00030@0005d09P5d0000
40000`40000300030@0000h000<10000D@0;0IX0000@00030@0000<000<100003P000`40001600/1
Y@0000h00`4500030@0000h000<10000@P040K00000@00030@0000<000<100003`000`40000m00@1
]00001H000<100003`000`40000i00@1^00001H00P4@00030@0003@01@6l00005P000`40000?0003
0@000300107100005P000`40000@00030@0002d00P7500005P000`40000@00030@0002X00`770000
5P000`40000@00030@0002P00P7:00005P000`40000A00030@0002D00P7<00005P000`40000A0003
0@0002<00P7>00005P020A<000<1000080020M00000F00030@00018000<100007`000`40003@0000
5P000`40000C00030@0001`00P7C00005P000`40000C00030@0001X00P7E00005P000`40000C0003
0@0001P00P7G00005P000`40000D00030@0001D00P7I00005P020AD00P4D0081f`0001H000<10000
5P000`40000@00030@000=/0000F00030@0001L000<100003@020Mh0000F00030@0001L000<10000
2`020N00000F00030@0001P00`480081hP0001H000<100006P090N@0000200@10`000`40000200@1
10000`40003o00P0000400030@0000L000<100001@030Ol0200000@000<1000020000`4000040003
0@000?l0200000@000<100002@000`40000300030@000?l0200000@000<100002P000`40003o00d0
000200<12@001040007o00l00000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.25136, -0.0926689, 
  0.0546524, 0.00749648}}]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
WindowSize->{684, 652},
WindowMargins->{{Automatic, 145}, {0, Automatic}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 33, 0, 107, "Title"],
Cell[1767, 53, 271, 6, 68, "Text"],

Cell[CellGroupData[{
Cell[2063, 63, 68, 3, 42, "Input"],
Cell[2134, 68, 115, 2, 39, "Print"]
}, Open  ]],
Cell[2264, 73, 54, 0, 32, "Text"],
Cell[2321, 75, 106, 4, 57, "Input"],
Cell[2430, 81, 113, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[2568, 88, 89, 3, 27, "Input"],
Cell[2660, 93, 62, 1, 46, "Output"]
}, Open  ]],
Cell[2737, 97, 118, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[2880, 104, 134, 6, 87, "Input"],
Cell[3017, 112, 12137, 417, 186, 4482, 318, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[15169, 532, 197, 4, 50, "Text"],

Cell[CellGroupData[{
Cell[15391, 540, 93, 3, 27, "Input"],
Cell[15487, 545, 12366, 414, 186, 4524, 313, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[27868, 962, 73, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[27966, 966, 46, 0, 27, "Input"],
Cell[28015, 968, 10946, 369, 186, 4001, 279, "GraphicsData", 
"PostScript", "Graphics"],
Cell[38964, 1339, 130, 3, 27, "Output"]
}, Open  ]],
Cell[39109, 1345, 62, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[39196, 1349, 48, 0, 27, "Input"],
Cell[39247, 1351, 10722, 347, 186, 3664, 255, "GraphicsData", 
"PostScript", "Graphics"],
Cell[49972, 1700, 130, 3, 27, "Output"]
}, Open  ]],
Cell[50117, 1706, 136, 3, 32, "Text"],
Cell[50256, 1711, 683, 23, 342, "Input"],
Cell[50942, 1736, 49, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[51016, 1740, 149, 6, 87, "Input"],

Cell[CellGroupData[{
Cell[51190, 1750, 13643, 545, 186, 6697, 455, "GraphicsData", 
"PostScript", "Graphics"],
Cell[64836, 2297, 79280, 3687, 166, 65552, 3513, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],
Cell[144143, 5988, 36, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[144204, 5992, 84, 3, 42, "Input"],
Cell[144291, 5997, 77, 1, 39, "Print"],
Cell[144371, 6000, 79, 1, 23, "Print"],
Cell[144453, 6003, 394, 11, 42, "Print"],
Cell[144850, 6016, 279, 9, 44, "Print"],
Cell[145132, 6027, 37, 1, 23, "Print"],
Cell[145172, 6030, 69, 1, 23, "Print"],
Cell[145244, 6033, 277, 4, 71, "Print"],
Cell[145524, 6039, 65, 1, 23, "Print"],
Cell[145592, 6042, 66, 1, 23, "Print"],
Cell[145661, 6045, 86, 1, 23, "Print"],
Cell[145750, 6048, 36, 1, 23, "Print"],
Cell[145789, 6051, 114, 2, 23, "Print"],
Cell[145906, 6055, 109, 2, 23, "Print"],
Cell[146018, 6059, 114, 2, 23, "Print"],
Cell[146135, 6063, 113, 2, 23, "Print"],
Cell[146251, 6067, 113, 2, 23, "Print"],
Cell[146367, 6071, 36, 1, 23, "Print"],
Cell[146406, 6074, 36, 1, 23, "Print"],
Cell[146445, 6077, 42, 1, 23, "Print"],
Cell[146490, 6080, 42, 1, 23, "Print"],
Cell[146535, 6083, 68, 1, 23, "Print"],
Cell[146606, 6086, 63, 1, 23, "Print"],
Cell[146672, 6089, 63, 1, 23, "Print"],
Cell[146738, 6092, 54, 1, 23, "Print"],
Cell[146795, 6095, 67, 1, 23, "Print"],
Cell[146865, 6098, 56, 1, 23, "Print"],
Cell[146924, 6101, 41, 1, 23, "Print"]
}, Open  ]],
Cell[146980, 6105, 89, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[147094, 6112, 84, 3, 42, "Input"],

Cell[CellGroupData[{
Cell[147203, 6119, 30678, 1179, 186, 18537, 1025, "GraphicsData", 
"PostScript", "Graphics"],
Cell[177884, 7300, 31624, 1232, 186, 19112, 1073, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],
Cell[209535, 8536, 51, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[209611, 8540, 81, 3, 42, "Input"],
Cell[209695, 8545, 175, 3, 23, "Print"],
Cell[209873, 8550, 491, 9, 48, "Output"]
}, Open  ]],
Cell[210379, 8562, 64, 0, 32, "Text"],
Cell[210446, 8564, 96, 4, 57, "Input"],

Cell[CellGroupData[{
Cell[210567, 8572, 74, 0, 27, "Input"],
Cell[210644, 8574, 13191, 478, 186, 5566, 379, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[223872, 9057, 89, 3, 42, "Input"],
Cell[223964, 9062, 187, 3, 39, "Message"],
Cell[224154, 9067, 14160, 494, 186, 5518, 383, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[238329, 9564, 47, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[238401, 9568, 82, 3, 42, "Input"],
Cell[238486, 9573, 532, 8, 45, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[239055, 9586, 59, 3, 42, "Input"],
Cell[239117, 9591, 177, 3, 39, "Message"],
Cell[239297, 9596, 13653, 436, 296, 4265, 315, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[252987, 10037, 89, 4, 57, "Input"],
Cell[253079, 10043, 189, 3, 39, "Message"],
Cell[253271, 10048, 14288, 432, 296, 3487, 294, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[267596, 10485, 155, 6, 87, "Input"],
Cell[267754, 10493, 301, 5, 43, "Output"],
Cell[268058, 10500, 335, 6, 43, "Output"]
}, Open  ]],
Cell[268408, 10509, 134, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[268567, 10516, 218, 6, 87, "Input"],
Cell[268788, 10524, 700, 15, 46, "Output"]
}, Open  ]],
Cell[269503, 10542, 99, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[269627, 10549, 90, 3, 42, "Input"],
Cell[269720, 10554, 12643, 460, 186, 5286, 365, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[282378, 11017, 107, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[282510, 11024, 127, 5, 72, "Input"],
Cell[282640, 11031, 713, 13, 65, "Output"],
Cell[283356, 11046, 12499, 394, 296, 3768, 282, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[295870, 11443, 103, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[295998, 11450, 139, 6, 87, "Input"],
Cell[296140, 11458, 12347, 426, 186, 4783, 328, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

