(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    364160,      12673]*)
(*NotebookOutlinePosition[    388179,      13547]*)
(*  CellTagsIndexPosition[    388135,      13543]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["OPTDesign Chapter 5 Example", "Title"],

Cell["\<\
This demo corresponds to the example in Chapter 5 of the book \
Classical Control Using Hinfinity Methods, by J. W. Helton and O. Merino,
NOTE: Depending upon the settings of the programs, some numbers  may differ \
from those in the book.\
\>", "Text"],

Cell[CellGroupData[{

Cell["5.1 Loading the software", "Section"],

Cell[CellGroupData[{

Cell["<<OPTDesign`", "Input"],

Cell[BoxData[
    \(" \n Anopt 4.0 (c) Copyright 1991-2001 \n J.W. Helton and O. Merino.  \
All rights reserved."\)], "Print"],

Cell[BoxData[
    \("OPTDesign 2.1 (c) 1991-2001 \n   J.W. Helton and O. Merino.  All \
rights reserved."\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5.2 Setting up the problem", "Section"],

Cell[CellGroupData[{

Cell["\<\
     p[s_]= (1 - s/5)/(1+s/2)^2;\t\t

      wt  = 0.3;
  alphat  = 1.5;

 alphagpm = 2.0;

       wb = 2.0;
   alphab = 0.7;
              
       wr = 4.0;
   alphar = 2.5;
 
    k0[w_]= Which[ Abs[w] <= wb, 1.0,
                     True, 0.0];
 
   r0[w_] = Which[Abs[w] < wt,  alphat,
               Abs[w] < wb,  alphagpm,
               Abs[w] < wr,  alphab,
                     True ,  alphar*Abs[p[I w]]
                ];

     wpts = Grid[Ngrid ->128,GridSpread -> 1.0];

EnvelopePlot[Radius ->r0,Center -> k0,
\t\t\tFrequencyBand -> {0,wr+1} 
\t\t \t(* ,GridPoints -> wpts *) ];

EnvelopePlot3D[Radius ->r0,Center -> k0,
\t\t\tFrequencyBand -> {0,wr+1} 
\t\t \t(* ,GridPoints -> wpts *) ];

\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(alphab\)\" is similar \
to existing symbol \"\!\(alphat\)\"."\)], "Message"],

Cell[BoxData[
    \(General::"spell" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(alphar\)\" is similar \
to existing symbols \!\({alphab, alphat}\)."\)], "Message"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.77556e-17 0.2 0.161866 0.147151 [
[.2 .14937 -3 -9 ]
[.2 .14937 3 0 ]
[.4 .14937 -3 -9 ]
[.4 .14937 3 0 ]
[.6 .14937 -3 -9 ]
[.6 .14937 3 0 ]
[.8 .14937 -3 -9 ]
[.8 .14937 3 0 ]
[1 .14937 -3 -9 ]
[1 .14937 3 0 ]
[-0.0125 .01472 -12 -4.5 ]
[-0.0125 .01472 0 4.5 ]
[-0.0125 .30902 -6 -4.5 ]
[-0.0125 .30902 0 4.5 ]
[-0.0125 .45617 -6 -4.5 ]
[-0.0125 .45617 0 4.5 ]
[-0.0125 .60332 -6 -4.5 ]
[-0.0125 .60332 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 .16187 m
.2 .16812 L
s
[(1)] .2 .14937 0 1 Mshowa
.4 .16187 m
.4 .16812 L
s
[(2)] .4 .14937 0 1 Mshowa
.6 .16187 m
.6 .16812 L
s
[(3)] .6 .14937 0 1 Mshowa
.8 .16187 m
.8 .16812 L
s
[(4)] .8 .14937 0 1 Mshowa
1 .16187 m
1 .16812 L
s
[(5)] 1 .14937 0 1 Mshowa
.125 Mabswid
.04 .16187 m
.04 .16562 L
s
.08 .16187 m
.08 .16562 L
s
.12 .16187 m
.12 .16562 L
s
.16 .16187 m
.16 .16562 L
s
.24 .16187 m
.24 .16562 L
s
.28 .16187 m
.28 .16562 L
s
.32 .16187 m
.32 .16562 L
s
.36 .16187 m
.36 .16562 L
s
.44 .16187 m
.44 .16562 L
s
.48 .16187 m
.48 .16562 L
s
.52 .16187 m
.52 .16562 L
s
.56 .16187 m
.56 .16562 L
s
.64 .16187 m
.64 .16562 L
s
.68 .16187 m
.68 .16562 L
s
.72 .16187 m
.72 .16562 L
s
.76 .16187 m
.76 .16562 L
s
.84 .16187 m
.84 .16562 L
s
.88 .16187 m
.88 .16562 L
s
.92 .16187 m
.92 .16562 L
s
.96 .16187 m
.96 .16562 L
s
.25 Mabswid
0 .16187 m
1 .16187 L
s
0 .01472 m
.00625 .01472 L
s
[(-1)] -0.0125 .01472 1 0 Mshowa
0 .30902 m
.00625 .30902 L
s
[(1)] -0.0125 .30902 1 0 Mshowa
0 .45617 m
.00625 .45617 L
s
[(2)] -0.0125 .45617 1 0 Mshowa
0 .60332 m
.00625 .60332 L
s
[(3)] -0.0125 .60332 1 0 Mshowa
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .13244 m
.00375 .13244 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .25016 m
.00375 .25016 L
s
0 .27959 m
.00375 .27959 L
s
0 .33845 m
.00375 .33845 L
s
0 .36788 m
.00375 .36788 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .4856 m
.00375 .4856 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
1 .08914 m
.89064 .07722 L
.79844 .06631 L
.72271 .06048 L
.65931 .05898 L
.60541 .05886 L
.55896 .05886 L
.51848 .05886 L
.48284 .0586 L
.45119 .05548 L
.42286 .04471 L
.39733 .02887 L
.37417 .01809 L
.35305 .01497 L
.33368 .01472 L
.31583 .01472 L
.29932 .01472 L
.28398 .01472 L
.26967 .01472 L
.25628 .01472 L
.2437 .01472 L
.23186 .01472 L
.22067 .01472 L
.21007 .01472 L
.2 .01472 L
.19042 .01472 L
.18127 .01472 L
.17252 .01472 L
.16414 .01472 L
.15608 .01472 L
.14833 .01472 L
.14086 .01472 L
.13364 .01472 L
.12665 .01472 L
.11988 .01472 L
.1133 .01472 L
.1069 .01472 L
.10067 .01472 L
.09459 .01472 L
.08865 .01472 L
.08284 .01472 L
.07715 .01472 L
.07156 .01514 L
.06607 .02034 L
.06067 .03831 L
.05535 .0647 L
.0501 .08266 L
.04491 .08786 L
.03978 .08829 L
.0347 .08829 L
Mistroke
.02967 .08829 L
.02467 .08829 L
.0197 .08829 L
.01475 .08829 L
.00983 .08829 L
.00491 .08829 L
0 .08829 L
Mfstroke
1 .2346 m
.89064 .24651 L
.79844 .25742 L
.72271 .26325 L
.65931 .26475 L
.60541 .26487 L
.55896 .26487 L
.51848 .26487 L
.48284 .26685 L
.45119 .29076 L
.42286 .37339 L
.39733 .4948 L
.37417 .57743 L
.35305 .60134 L
.33368 .60332 L
.31583 .60332 L
.29932 .60332 L
.28398 .60332 L
.26967 .60332 L
.25628 .60332 L
.2437 .60332 L
.23186 .60332 L
.22067 .60332 L
.21007 .60332 L
.2 .60332 L
.19042 .60332 L
.18127 .60332 L
.17252 .60332 L
.16414 .60332 L
.15608 .60332 L
.14833 .60332 L
.14086 .60332 L
.13364 .60332 L
.12665 .60332 L
.11988 .60332 L
.1133 .60332 L
.1069 .60332 L
.10067 .60332 L
.09459 .60332 L
.08865 .60332 L
.08284 .60332 L
.07715 .60332 L
.07156 .60289 L
.06607 .59769 L
.06067 .57973 L
.05535 .55333 L
.0501 .53537 L
.04491 .53017 L
.03978 .52974 L
.0347 .52974 L
Mistroke
.02967 .52974 L
.02467 .52974 L
.0197 .52974 L
.01475 .52974 L
.00983 .52974 L
.00491 .52974 L
0 .52974 L
Mfstroke
.5 g
.01 w
0 .52974 m
.01975 .52974 L
.04057 .52974 L
.04585 .52974 L
.0515 .52974 L
.05424 .52974 L
.05683 .52974 L
.05796 .52974 L
.05916 .52974 L
.06047 .60332 L
.06168 .60332 L
.08481 .60332 L
.12454 .60332 L
.16684 .60332 L
.20755 .60332 L
.25083 .60332 L
.29253 .60332 L
.33263 .60332 L
.35501 .60332 L
.37531 .60332 L
.38547 .60332 L
.39082 .60332 L
.39374 .60332 L
.39646 .60332 L
.39769 .60332 L
.39885 .60332 L
.40019 .26487 L
.40141 .26487 L
.40684 .26487 L
.41639 .26487 L
.45664 .26487 L
.49946 .26487 L
.54069 .26487 L
.58033 .26487 L
.62254 .26487 L
.66316 .26487 L
.70636 .26487 L
.74797 .26487 L
.76893 .26487 L
.78798 .26487 L
.79306 .26487 L
.79574 .26487 L
.7972 .26487 L
.79857 .26487 L
.79977 .26487 L
.80105 .25594 L
.80178 .25584 L
.80247 .25574 L
.80377 .25555 L
.80857 .25488 L
Mistroke
.83057 .2519 L
.87073 .24687 L
.91346 .24204 L
.9546 .23783 L
.99832 .23377 L
1 .23363 L
Mfstroke
0 .08829 m
.01975 .08829 L
.04057 .08829 L
.04585 .08829 L
.0515 .08829 L
.05424 .08829 L
.05683 .08829 L
.05796 .08829 L
.05916 .08829 L
.06047 .01472 L
.06168 .01472 L
.08481 .01472 L
.12454 .01472 L
.16684 .01472 L
.20755 .01472 L
.25083 .01472 L
.29253 .01472 L
.33263 .01472 L
.35501 .01472 L
.37531 .01472 L
.38547 .01472 L
.39082 .01472 L
.39374 .01472 L
.39646 .01472 L
.39769 .01472 L
.39885 .01472 L
.40019 .05886 L
.40141 .05886 L
.40684 .05886 L
.41639 .05886 L
.45664 .05886 L
.49946 .05886 L
.54069 .05886 L
.58033 .05886 L
.62254 .05886 L
.66316 .05886 L
.70636 .05886 L
.74797 .05886 L
.76893 .05886 L
.78798 .05886 L
.79306 .05886 L
.79574 .05886 L
.7972 .05886 L
.79857 .05886 L
.79977 .05886 L
.80105 .06779 L
.80178 .0679 L
.80247 .06799 L
.80377 .06818 L
.80857 .06885 L
Mistroke
.83057 .07183 L
.87073 .07686 L
.91346 .0817 L
.9546 .08591 L
.99832 .08996 L
1 .09011 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool000ooo`03003ooooo0?ooo`ooo`003ooo00<00?ooool0
oooo3ooo000?ool00`00ooooo`3oool?ool000Ooo`@000Coo`03003ooooo0?ooo`ooo`002Ooo00<0
0?ooool00ooo0`003?oo05ekkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnl0YOoo0009ool00`00ooooo`03ool00`00ooooo`0<ool0GH@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn4402Uool00003ool00000008000Coo`03003ooooo00?oo`03003ooooo00coo`04Nnn44?oo001C
ool20002ool00h@@Nnooo`2Tool000Woo`03003ooooo00?oo`03003ooooo00coo`03Q11kk`0005Ko
o`800003Nnn44?oo0:Coo`001ooo0`001Ooo00<00?ooool03?oo00=kkh@@0000F?oo00>447__ool0
Y?oo0009ool00`00ooooo`03ool00`00ooooo`0<ool00h@@Nnooo`1Hool017__Q1000000Xooo000?
ool00`00ooooo`0<ool00g__Q13oo`1Hool01H@@Nnoooooo0000X_oo000?ool00`00ooooo`0<ool0
0h@@Nnooo`1Hool00g__Q13oo`02ool2002Pool000ooo`8000goo`03Nnn44?oo05Soo`03Q11kkooo
00Coo`8009koo`003ooo00<00?ooool03?oo00>447__ool0F?oo00=kkh@@ool01_oo0P00W?oo000?
ool00`00ooooo`0<ool00g__Q13oo`1Hool00h@@Nnooo`08ool3002Iool000ooo`03003ooooo00co
o`03Q11kkooo05Soo`03Nnn44?oo00_oo`H009?oo`003ooo00<00?ooool03?oo00=kkh@@ool0F?oo
06j447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__>Ooo000?ool00`00ooooo`0<
ool00h@@Nnooo`1Hool0Kg__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`0h
ool000ooo`03003ooooo00coo`03Nnn44?oo0;ooo`D00005Q11kkh@@Nnl00004ool00g__ooooo`0_
ool000ooo`03003ooooo00_oo`03002447__0<Goo`0?Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__02coo`003ooo0P003?oo00<007__Q100aooo01>447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__
Q11kkh@@Nnn4400Vool000ooo`03003ooooo00[oo`04003ooh@@NnoDool04H@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q1006ooo000?ool00`00ooooo`0:ool01000oomkkh@@fOoo01N447__
Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@013oo`003ooo00<00?ooool0
2_oo00@00?ooQ11kknCoo`0HQ11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__1?oo000>ool04W__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn44>ooo`0>Q11k
kh@@Nnn447__Q11kkh@@Nnn447__Q11kk`?oo`003_oo01:447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnojool00h@@Nnn44003ool000ooo`03003ooooo0?ooo`ooo`003ooo00<00?ooool0
oooo3ooo000?ool2003oool@ool000ooo`03003ooooo0?ooo`ooo`003ooo00<00?ooool0oooo3ooo
000?ool00`00ooooo`3oool?ool000ooo`03003ooooo0?ooo`ooo`003ooo00<00?ooool0oooo3ooo
000?ool00`00ooooo`3oool?ool000ooo`03003ooooo037oo`@0037oo`@003?oo`8003Goo`03003o
oooo037oo`<000?oo`003ooo0P00=?oo00<00?ooool0<?oo00<00?ooool0<ooo00@00?ooool003Co
o`03003ooooo03Coo`40007oo`7oo`003ooo00<00?ooool0<ooo00<00?ooool0<Ooo00<00?ooool0
=Ooo00<00?ooool0;ooo1@00=Ooo0@000Ooo0Ooo000?ool00`00ooooo`0cool00`00ooooo`0bool0
0`00ooooo`0bool2000bool01000ooooo`00<ooo0`000ooo000?ool00`00ooooo`0cool00`00oooo
o`0cool00`00ooooo`0cool00`00ooooo`0`ool00`00ool0000cool00`00ooooo`03ool000ooo`03
003ooooo037oo`<003;oo`04003ooooo000bool01000ooooo`00<ooo0P00<ooo00<00?ooool00ooo
000?ool00`00ooooo`0cool00`00ooooo`0aool2000dool2000eool00`00ooooo`0aool40002ool0
00ooo`03003ooooo0?ooo`ooo`003oooo`003`000ooo000?ool00`00ooooo`08ool00`00ooooo`07
ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`07ool00`00ooooo`08
ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`08
ool00`00ooooo`08ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08
ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`07
ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`01ool000ooo`03003o
oooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003o
oooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03003o
oooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003o
oooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003o
oooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003o
oooo007oo`003ooo00<00?ooool0oooo3ooo000?ool00`00ooooo`3oool?ool000ooo`03003ooooo
0?ooo`ooo`003ooo00<00?ooool0oooo3ooo000?ool00`00ooooo`3oool?ool000ooo`800?oooa3o
o`003ooo00<00?ooool0oooo3ooo000?ool00`00ooooo`3oool?ool000ooo`03003ooooo0?ooo`oo
o`003ooo00<00?ooool0oooo3ooo000?ool00`00ooooo`3oool?ool000ooo`03003ooooo0?ooo`oo
o`003ooo00<00?ooool0oooo3ooo000?ool2003oool@ool000ooo`03003ooooo0?ooo`ooo`003ooo
00<00?ooool0oooo3ooo000?ool00`00ooooo`3oool:ool00h@@ooooo`02ool000ooo`03003ooooo
0?goo`0>Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn440?oo`003ooo00<00?ooool0l_oo01Ukkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`03ool000ooo`03003o
oooo0>Ooo`0GNnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`0@ool0
00ooo`03003ooooo0>;oo`0ANnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`0Kool000oo
o`800=Ooo`03Q13ooooo00?oo`0<Q11kkh@@Nnn447__Q11kkh@@Nnn447__9_oo000?ool00`00oooo
o`3Eool03H@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@02coo`003ooo00<00?ooool0c_oo1@0000Yk
kh@@Nnn447__Q11kkh@@Nnn4437oo`003ooo00<00?ooool0a?oo2P001Ooo00>447__Q100>?oo000?
ool00`00ooooo`1Wool0KW__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q10iool0
00ooo`03003ooooo06Ooo`1^Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkcWo
o`003ooo00<00?ooool0Iooo00=kkh@@ool04ooo0P00Sooo000?ool00`00ooooo`1Wool00h@@Nnoo
o`0Bool00`00ooooo`2?ool000ooo`8006Soo`03Nnn44?oo013oo`8009;oo`003ooo00<00?ooool0
Iooo00>447__ool03ooo00<00?ooool0T_oo000?ool00`00ooooo`1Wool00g__Q13oo`0=ool2002E
ool000ooo`03003ooooo06Ooo`03Q11kkooo00coo`03003ooooo09Goo`003ooo00<00?ooool0Iooo
00=kkh@@ool03?oo00<00?ooool0UOoo000?ool00`00ooooo`1Wool00h@@Nnooo`0;ool00`00oooo
o`2Fool000ooo`03003ooooo06Ooo`03Nnn44?oo00_oo`03003ooooo09Koo`001ooo10001?oo00<0
0?ooool0Iooo00>447__ool02ooo00<00?ooool0U_oo0009ool00`00ooooo`03ool3001Wool00g__
Q13oo`0:ool00`00ooooo`2Gool000Woo`03003ooooo00?oo`03003ooooo06Ooo`03Q11kkooo00[o
o`03003ooooo09Ooo`002Ooo00<00?ooool00ooo00<00?ooool0Iooo00=kkh@@ool02Ooo00<00?oo
ool0V?oo0009ool00`00ooooo`03ool00`00ooooo`1Wool00h@@Nnooo`09ool00`00ooooo`2Hool0
00Ooo`<000Goo`03003ooooo06Ooo`03Nnn44?oo00Woo`03003ooooo09Soo`002Ooo00<00?ooool0
0ooo00<00?ooool0Iooo00>447__ool02?oo00<00?ooool0VOoo000?ool00`00ooooo`1Wool00g__
Q13oo`08ool00`00ooooo`2Iool000ooo`03003ooooo06Ooo`03Q11kkooo00Soo`03003ooooo09Wo
o`003ooo0P00J?oo00=kkh@@ool01ooo00<00?ooool0V_oo000?ool00`00ooooo`1Wool00h@@Nnoo
o`07ool00`00ooooo`2Jool000ooo`03003ooooo06Ooo`03Nnn44?oo00Ooo`03003ooooo09[oo`00
3ooo00<00?ooool0Iooo00>447__ool01_oo00<00?ooool0Vooo000?ool00`00ooooo`1Wool00g__
Q13oo`06ool00`00ooooo`2Kool000ooo`03003ooooo06Ooo`03Q11kkooo00Goo`03003ooooo09co
o`003ooo00<00?ooool0Iooo00=kkh@@ool01Ooo00<00?ooool0W?oo000?ool00`00ooooo`1Wool0
0h@@Nnooo`05ool00`00ooooo`2Lool000ooo`8006Soo`03Nnn44?oo00Coo`03003ooooo09goo`00
3ooo00<00?ooool0Iooo00>447__ool01?oo00<00?ooool0WOoo000?ool00`00ooooo`1Wool00g__
Q13oo`04ool00`00ooooo`2Mool000ooo`03003ooooo06Ooo`03Q11kkooo00Coo`03003ooooo09go
o`003ooo00<00?ooool0Iooo00=kkh@@ool00ooo00<00?ooool0W_oo000?ool00`00ooooo`1Wool0
0h@@Nnooo`03ool00`00ooooo`2Nool000ooo`03003ooooo06Ooo`03Nnn44?oo00?oo`03003ooooo
09koo`003ooo00<00?ooool0Iooo00>447__ool00ooo00<00?ooool0W_oo000?ool2001Xool00g__
Q13oo`03ool00`00ooooo`2Nool000ooo`03003ooooo06Ooo`03Q11kkooo00?oo`03003ooooo09ko
o`003ooo00<00?ooool0Iooo00=kkh@@ool00_oo00<00?ooool0Wooo000?ool00`00ooooo`1Wool0
0h@@Nnooo`02ool00`00ooooo`2Oool000ooo`03003ooooo06Ooo`03Nnn44?oo00;oo`03003ooooo
09ooo`003ooo00<00?ooool0Iooo00>447__ool00_oo00<00?ooool0Wooo000?ool00`00ooooo`1W
ool00g__Q13oo`02ool00`00ooooo`2Oool000ooo`03003ooooo06Ooo`05Q11kkoooool0002Rool0
00ooo`8006Soo`05Nnn44?ooool0002Rool000ooo`03003ooooo06Ooo`05Q11kkoooool0002Rool0
00ooo`03003ooooo06Ooo`05Nnn44?ooool0002Rool000ooo`03003ooooo06Ooo`05Q11kkoooool0
002Rool000ooo`03003ooooo06Ooo`05Nnn44?ooool0002Rool000ooo`03003ooooo06Ooo`04Q11k
kooo002Sool000Ooo`@000Coo`03003ooooo06Ooo`04Nnn44?oo002Sool000Ooo`03003ooooo00Go
o`03003ooooo06Ooo`04Q11kkooo002Sool000Soo`03003ooooo00Coo`<006Ooo`04Nnn44?oo002S
ool000Woo`03003ooooo00?oo`03003ooooo06Ooo`04Q11kkooo002Sool000[oo`03003ooooo00;o
o`03003ooooo06Ooo`03Nnn440000:Coo`001ooo00@00?ooool000Coo`03003ooooo06Ooo`03Q11k
k`000:Coo`002?oo0P001Ooo00<00?ooool0Iooo00=kkh@@0000Y?oo000?ool00`00ooooo`1Wool0
0h@@Nnl0002Tool000ooo`03003ooooo06Ooo`03Nnn440000:Coo`003ooo0P00J?oo00>447__0000
Y?oo000?ool00`00ooooo`1Wool00g__Q13oo`2Tool000ooo`03003ooooo06Ooo`03Q11kkooo0:Co
o`003ooo00<00?ooool0Iooo00=kkh@@ool0Y?oo000?ool00`00ooooo`1Wool00h@@Nnooo`2Tool0
00ooo`03003ooooo06Ooo`03Nnn44?oo0:Coo`003ooo00<00?ooool0Iooo00>447__ool0Y?oo000?
ool00`00ooooo`1Wool00g__Q13oo`2Tool000ooo`8006Ooo`03002447__0:Goo`003ooo00<00?oo
ool0I_oo00<007__Q100YOoo000?ool00`00ooooo`1Vool00`00Q11kk`2Uool000ooo`03003ooooo
06Goo`04003oog__Q12Uool000koo`0BNnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Eooo00@00?ooQ11kkjGoo`003_oo01:447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@NnmG
ool01000oomkkh@@YOoo000?ool00`00ooooo`0:ool01000oomkkh@@Eooo00@00?ooQ11kkjGoo`00
3ooo00<00?ooool02_oo00@00?ooQ11kkeKoo`05003oooooNnn4402Uool000ooo`8000coo`03001k
kh@@05Koo`05003oooooQ11kk`2Uool000ooo`03003ooooo00_oo`03002447__05Koo`05003ooooo
Nnn4402Uool000ooo`03003ooooo00coo`03Nnn44?oo05Coo`06003ooooooon447__YOoo000?ool0
0`00ooooo`0<ool00h@@Nnooo`1Dool01P00ooooooooNnn44:Goo`003ooo00<00?ooool03?oo00=k
kh@@ool0E?oo00H00?oooooooh@@NnnUool000ooo`03003ooooo00coo`03Q11kkooo05?oo`03003o
oooo00;oo`03Nnn44?oo0:Coo`003ooo00<00?ooool03?oo00=kkh@@ool0Dooo00<00?ooool00_oo
00>447__ool0Y?oo000?ool00`00ooooo`0<ool00h@@Nnooo`1Cool00`00ooooo`02ool00g__Q13o
o`2Tool000ooo`8000goo`03Nnn44?oo05;oo`03003ooooo00?oo`03Q11kkooo0:Coo`003ooo00<0
0?ooool03?oo00>447__ool0D_oo00<00?ooool00ooo00=kkh@@ool0Y?oo000?ool00`00ooooo`0<
ool00g__Q13oo`1Aool00`00ooooo`04ool00h@@Nnooo`2Tool000ooo`03003ooooo00coo`03Q11k
kooo057oo`03003ooooo00Coo`03Nnn44?oo0:Coo`003ooo00<00?ooool03?oo00=kkh@@0000D?oo
00<00?ooool01Ooo00>447__ool0Y?oo000?ool00`00ooooo`0<ool00h@@Nnl0001?ool00`00oooo
o`06ool00g__Q13oo`2Tool000ooo`03003ooooo00coo`04Nnn44?oo001=ool00`00ooooo`07ool0
0h@@Nnooo`2Tool000Soo`8000Goo`03003ooooo00coo`04Q11kkooo001=ool00`00ooooo`07ool0
0g__Q13oo`2Tool000Ooo`04003ooooo0004ool3000<ool0GG__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`2Uool000[oo`03003ooooo
00;oo`03003ooooo00coo`1MQ11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@0:Goo`002?oo0P001Ooo00<00?ooool0oooo3ooo000:ool0
0`00ooooo`02ool00`00ooooo`3oool?ool000Ooo`04003ooooo0004ool00`00ooooo`3oool?ool0
00Soo`800?oooaOoo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.285873, -1.23796, 
  0.0186123, 0.0252969}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .98709 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0161164 1.01612 -0.000359473 1.01612 [
[.21669 .99535 -3.18333 0 ]
[.21669 .99535 2.81667 9 ]
[.50187 .92136 -2.99327 0 ]
[.50187 .92136 3.00673 9 ]
[.81334 .84051 -2.77831 0 ]
[.81334 .84051 3.22169 9 ]
[.57837 .96533 -4.90342 0 ]
[.57837 .96533 5.09658 9.8125 ]
[.00377 .85528 -3.33559 0 ]
[.00377 .85528 2.66441 9 ]
[.05477 .8929 -9.89388 0 ]
[.05477 .8929 8.10612 9 ]
[.10371 .92899 -3.26245 0 ]
[.10371 .92899 2.73755 9 ]
[.1507 .96365 -9.6867 0 ]
[.1507 .96365 8.3133 9 ]
[.19586 .99696 -3.19714 0 ]
[.19586 .99696 2.80286 9 ]
[.09784 .99201 -18.4886 0 ]
[.09784 .99201 15.5114 10.125 ]
[.94612 .21871 0 -6.03056 ]
[.94612 .21871 12 2.96944 ]
[.96063 .34082 0 -5.9512 ]
[.96063 .34082 6 3.0488 ]
[.97612 .47119 0 -5.86684 ]
[.97612 .47119 6 3.13316 ]
[.99268 .61067 0 -5.77697 ]
[.99268 .61067 6 3.22303 ]
[1.01044 .76026 0 -5.68105 ]
[1.01044 .76026 6 3.31895 ]
[1.03688 .45273 0 -6.60019 ]
[1.03688 .45273 34 3.52481 ]
[ 0 0 0 0 ]
[ 1 .98709 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.20049 .98709 m
1 .77902 L
s
.21747 .98267 m
.21785 .97633 L
s
[(0)] .21669 .99535 .06111 -1 Mshowa
.50184 .90866 m
.50183 .90231 L
s
[(2)] .50187 .92136 -0.00224 -1 Mshowa
.81241 .82784 m
.81194 .82151 L
s
[(4)] .81334 .84051 -0.0739 -1 Mshowa
.125 Mabswid
.28629 .96476 m
.28646 .96095 L
s
.35659 .94646 m
.35671 .94266 L
s
.42843 .92777 m
.42848 .92396 L
s
.57689 .88913 m
.57682 .88532 L
s
.65363 .86916 m
.65349 .86535 L
s
.73212 .84874 m
.73191 .84493 L
s
.89457 .80646 m
.89421 .80267 L
s
.97867 .78457 m
.97824 .78079 L
s
gsave
.57837 .96533 -65.9034 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 Mabswid
0 .83882 m
.20049 .98709 L
s
.00518 .84266 m
.00589 .83634 L
s
[(0)] .00377 .85528 .11186 -1 Mshowa
.05603 .88026 m
.05666 .87394 L
s
[(0.5)] .05477 .8929 .09932 -1 Mshowa
.10481 .91634 m
.10537 .91001 L
s
[(1)] .10371 .92899 .08748 -1 Mshowa
.15167 .95098 m
.15215 .94465 L
s
[(1.5)] .1507 .96365 .0763 -1 Mshowa
.1967 .98429 m
.19711 .97795 L
s
[(2)] .19586 .99696 .06571 -1 Mshowa
.125 Mabswid
.01552 .8503 m
.01594 .84651 L
s
.02578 .85788 m
.02618 .8541 L
s
.03594 .8654 m
.03634 .86161 L
s
.04603 .87286 m
.04641 .86907 L
s
.06595 .88759 m
.06631 .8838 L
s
.07578 .89487 m
.07614 .89107 L
s
.08554 .90208 m
.08589 .89829 L
s
.09521 .90924 m
.09556 .90544 L
s
.11434 .92338 m
.11466 .91958 L
s
.12378 .93036 m
.12409 .92656 L
s
.13315 .93729 m
.13346 .93349 L
s
.14244 .94417 m
.14274 .94037 L
s
.16081 .95775 m
.1611 .95395 L
s
.16989 .96446 m
.17016 .96066 L
s
.1789 .97112 m
.17916 .96732 L
s
.18783 .97773 m
.18809 .97393 L
s
gsave
.09784 .99201 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(@) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 11.250000 moveto
(D) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
.25 Mabswid
.9327 .21103 m
1 .77902 L
s
.93409 .2228 m
.92808 .22484 L
s
[(-1)] .94612 .21871 -1 .34012 Mshowa
.94854 .34472 m
.9425 .34667 L
s
[(0)] .96063 .34082 -1 .32249 Mshowa
.96396 .47488 m
.95789 .47673 L
s
[(1)] .97612 .47119 -1 .30374 Mshowa
.98046 .61414 m
.97435 .61587 L
s
[(2)] .99268 .61067 -1 .28377 Mshowa
.99816 .76349 m
.99202 .7651 L
s
[(3)] 1.01044 .76026 -1 .26245 Mshowa
.125 Mabswid
.93691 .24656 m
.9333 .24778 L
s
.93976 .27063 m
.93615 .27183 L
s
.94265 .29501 m
.93903 .2962 L
s
.94558 .3197 m
.94195 .32088 L
s
.95154 .37007 m
.94791 .37123 L
s
.95459 .39575 m
.95095 .3969 L
s
.95767 .42178 m
.95403 .42291 L
s
.96079 .44815 m
.95715 .44927 L
s
.96717 .50197 m
.96352 .50307 L
s
.97043 .52944 m
.96677 .53052 L
s
.97373 .55728 m
.97007 .55835 L
s
.97707 .58551 m
.97341 .58657 L
s
.9839 .64317 m
.98023 .64419 L
s
.98739 .67261 m
.98372 .67362 L
s
.99093 .70247 m
.98725 .70347 L
s
.99452 .73276 m
.99084 .73374 L
s
gsave
1.03688 .45273 -61 -10.6002 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(@) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 11.250000 moveto
(D) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
.25 Mabswid
.06337 .27882 m
0 .83882 L
s
0 .83882 m
.20049 .98709 L
s
.20049 .98709 m
.23534 .44941 L
s
.23534 .44941 m
.06337 .27882 L
s
.80061 0 m
.9327 .21103 L
s
.9327 .21103 m
1 .77902 L
s
1 .77902 m
.85247 .5896 L
s
.85247 .5896 m
.80061 0 L
s
.06337 .27882 m
0 .83882 L
s
0 .83882 m
.85247 .5896 L
s
.85247 .5896 m
.80061 0 L
s
.80061 0 m
.06337 .27882 L
s
.23534 .44941 m
.9327 .21103 L
s
.9327 .21103 m
1 .77902 L
s
1 .77902 m
.20049 .98709 L
s
.20049 .98709 m
.23534 .44941 L
s
0 0 m
1 0 L
1 .98709 L
0 .98709 L
closepath
clip
newpath
.5 Mabswid
.28182 .40652 m .24819 .41815 L .26602 .46425 L p .734 .659 .853 r
F P
0 g
s
.26602 .46425 m .29966 .45288 L .28182 .40652 L p .734 .659 .853 r
F P
0 g
s
.1998 .48724 m .26602 .46425 L .24819 .41815 L p .934 .756 .761 r
F P
0 g
s
.29966 .45288 m .26602 .46425 L .28037 .51151 L p .724 .624 .82 r
F P
0 g
s
.26602 .46425 m .1998 .48724 L .21074 .52335 L p .929 .738 .745 r
F P
0 g
s
.21074 .52335 m .28037 .51151 L .26602 .46425 L p .929 .738 .745 r
F P
0 g
s
.28037 .51151 m .31411 .50038 L .29966 .45288 L p .724 .624 .82 r
F P
0 g
s
.31411 .50038 m .28037 .51151 L .29097 .559 L p .712 .587 .786 r
F P
0 g
s
.28037 .51151 m .21074 .52335 L .21867 .55963 L p .925 .72 .726 r
F P
0 g
s
.21867 .55963 m .29097 .559 L .28037 .51151 L p .925 .72 .726 r
F P
0 g
s
.2609 .36221 m .22723 .37413 L .24819 .41815 L p .744 .694 .884 r
F P
0 g
s
.24819 .41815 m .28182 .40652 L .2609 .36221 L p .744 .694 .884 r
F P
0 g
s
.18609 .45207 m .24819 .41815 L .22723 .37413 L p .941 .775 .775 r
F P
0 g
s
.24819 .41815 m .18609 .45207 L .1998 .48724 L p .934 .756 .761 r
F P
0 g
s
.29097 .559 m .32487 .54808 L .31411 .50038 L p .712 .587 .786 r
F P
0 g
s
.32487 .54808 m .29097 .559 L .29757 .60578 L p .697 .546 .747 r
F P
0 g
s
.29097 .559 m .21867 .55963 L .22339 .59529 L p .922 .703 .706 r
F P
0 g
s
.22339 .59529 m .29757 .60578 L .29097 .559 L p .922 .703 .706 r
F P
0 g
s
.31579 .39476 m .28182 .40652 L .29966 .45288 L p .734 .659 .853 r
F P
0 g
s
.29966 .45288 m .33366 .44139 L .31579 .39476 L p .734 .659 .853 r
F P
0 g
s
.33366 .44139 m .29966 .45288 L .31411 .50038 L p .724 .624 .82 r
F P
0 g
s
.31411 .50038 m .3482 .48913 L .33366 .44139 L p .724 .624 .82 r
F P
0 g
s
.3482 .48913 m .31411 .50038 L .32487 .54808 L p .712 .587 .786 r
F P
0 g
s
.2373 .32088 m .20351 .33309 L .22723 .37413 L p .753 .73 .915 r
F P
0 g
s
.22723 .37413 m .2609 .36221 L .2373 .32088 L p .753 .73 .915 r
F P
0 g
s
.16988 .41861 m .22723 .37413 L .20351 .33309 L p .95 .794 .788 r
F P
0 g
s
.22723 .37413 m .16988 .41861 L .18609 .45207 L p .941 .775 .775 r
F P
0 g
s
.29493 .35017 m .2609 .36221 L .28182 .40652 L p .744 .694 .884 r
F P
0 g
s
.28182 .40652 m .31579 .39476 L .29493 .35017 L p .744 .694 .884 r
F P
0 g
s
.32487 .54808 m .35912 .53705 L .3482 .48913 L p .712 .587 .786 r
F P
0 g
s
.35912 .53705 m .32487 .54808 L .33169 .59504 L p .697 .546 .747 r
F P
0 g
s
.29757 .60578 m .33169 .59504 L .32487 .54808 L p .697 .546 .747 r
F P
0 g
s
.33169 .59504 m .29757 .60578 L .29996 .65088 L p .68 .499 .701 r
F P
0 g
s
.29757 .60578 m .22339 .59529 L .22476 .62955 L p .922 .686 .683 r
F P
0 g
s
.22476 .62955 m .29996 .65088 L .29757 .60578 L p .922 .686 .683 r
F P
0 g
s
.35012 .38289 m .31579 .39476 L .33366 .44139 L p .734 .659 .853 r
F P
0 g
s
.33366 .44139 m .36802 .42978 L .35012 .38289 L p .734 .659 .853 r
F P
0 g
s
.36802 .42978 m .33366 .44139 L .3482 .48913 L p .724 .624 .82 r
F P
0 g
s
.3482 .48913 m .38265 .47777 L .36802 .42978 L p .724 .624 .82 r
F P
0 g
s
.38265 .47777 m .3482 .48913 L .35912 .53705 L p .712 .587 .786 r
F P
0 g
s
.27144 .30853 m .2373 .32088 L .2609 .36221 L p .753 .73 .915 r
F P
0 g
s
.2609 .36221 m .29493 .35017 L .27144 .30853 L p .753 .73 .915 r
F P
0 g
s
.32931 .338 m .29493 .35017 L .31579 .39476 L p .744 .694 .884 r
F P
0 g
s
.31579 .39476 m .35012 .38289 L .32931 .338 L p .744 .694 .884 r
F P
0 g
s
.18609 .45207 m .15253 .46362 L .16623 .49858 L p .732 .659 .854 r
F P
0 g
s
.16623 .49858 m .1998 .48724 L .18609 .45207 L p .732 .659 .854 r
F P
0 g
s
.1998 .48724 m .16623 .49858 L .17709 .53451 L p .721 .622 .821 r
F P
0 g
s
.21143 .28343 m .17746 .29596 L .20351 .33309 L p .761 .768 .946 r
F P
0 g
s
.20351 .33309 m .2373 .32088 L .21143 .28343 L p .761 .768 .946 r
F P
0 g
s
.15151 .3876 m .20351 .33309 L .17746 .29596 L p .96 .814 .799 r
F P
0 g
s
.20351 .33309 m .15151 .3876 L .16988 .41861 L p .95 .794 .788 r
F P
0 g
s
.17709 .53451 m .21074 .52335 L .1998 .48724 L p .721 .622 .821 r
F P
0 g
s
.21074 .52335 m .17709 .53451 L .1849 .57063 L p .709 .584 .785 r
F P
0 g
s
.35912 .53705 m .39374 .5259 L .38265 .47777 L p .712 .587 .786 r
F P
0 g
s
.39374 .5259 m .35912 .53705 L .36617 .58419 L p .697 .546 .747 r
F P
0 g
s
.33169 .59504 m .36617 .58419 L .35912 .53705 L p .697 .546 .747 r
F P
0 g
s
.36617 .58419 m .33169 .59504 L .33437 .6403 L p .68 .499 .701 r
F P
0 g
s
.29996 .65088 m .33437 .6403 L .33169 .59504 L p .68 .499 .701 r
F P
0 g
s
.16988 .41861 m .13629 .43037 L .15253 .46362 L p .742 .695 .886 r
F P
0 g
s
.15253 .46362 m .18609 .45207 L .16988 .41861 L p .742 .695 .886 r
F P
0 g
s
.38482 .37088 m .35012 .38289 L .36802 .42978 L p .734 .659 .853 r
F P
0 g
s
.36802 .42978 m .40273 .41805 L .38482 .37088 L p .734 .659 .853 r
F P
0 g
s
.40273 .41805 m .36802 .42978 L .38265 .47777 L p .724 .624 .82 r
F P
0 g
s
.1849 .57063 m .21867 .55963 L .21074 .52335 L p .709 .584 .785 r
F P
0 g
s
.21867 .55963 m .1849 .57063 L .18946 .60615 L p .694 .541 .744 r
F P
0 g
s
.38265 .47777 m .41746 .46628 L .40273 .41805 L p .724 .624 .82 r
F P
0 g
s
.41746 .46628 m .38265 .47777 L .39374 .5259 L p .712 .587 .786 r
F P
0 g
s
.30594 .29606 m .27144 .30853 L .29493 .35017 L p .753 .73 .915 r
F P
0 g
s
.29493 .35017 m .32931 .338 L .30594 .29606 L p .753 .73 .915 r
F P
0 g
s
.33437 .6403 m .29996 .65088 L .298 .69333 L p .657 .443 .645 r
F P
0 g
s
.29996 .65088 m .22476 .62955 L .22269 .6616 L p .922 .67 .657 r
F P
0 g
s
.22269 .6616 m .298 .69333 L .29996 .65088 L p .922 .67 .657 r
F P
0 g
s
.36405 .32571 m .32931 .338 L .35012 .38289 L p .744 .694 .884 r
F P
0 g
s
.35012 .38289 m .38482 .37088 L .36405 .32571 L p .744 .694 .884 r
F P
0 g
s
.15151 .3876 m .11784 .39958 L .13629 .43037 L p .751 .731 .918 r
F P
0 g
s
.13629 .43037 m .16988 .41861 L .15151 .3876 L p .751 .731 .918 r
F P
0 g
s
.24575 .27077 m .21143 .28343 L .2373 .32088 L p .761 .768 .946 r
F P
0 g
s
.2373 .32088 m .27144 .30853 L .24575 .27077 L p .761 .768 .946 r
F P
0 g
s
.39374 .5259 m .42871 .51463 L .41746 .46628 L p .712 .587 .786 r
F P
0 g
s
.42871 .51463 m .39374 .5259 L .40102 .57323 L p .697 .546 .747 r
F P
0 g
s
.36617 .58419 m .40102 .57323 L .39374 .5259 L p .697 .546 .747 r
F P
0 g
s
.40102 .57323 m .36617 .58419 L .36914 .6296 L p .68 .499 .701 r
F P
0 g
s
.33437 .6403 m .36914 .6296 L .36617 .58419 L p .68 .499 .701 r
F P
0 g
s
.18946 .60615 m .22339 .59529 L .21867 .55963 L p .694 .541 .744 r
F P
0 g
s
.22339 .59529 m .18946 .60615 L .19062 .64029 L p .675 .493 .697 r
F P
0 g
s
.41988 .35875 m .38482 .37088 L .40273 .41805 L p .734 .659 .853 r
F P
0 g
s
.40273 .41805 m .43781 .4062 L .41988 .35875 L p .734 .659 .853 r
F P
0 g
s
.43781 .4062 m .40273 .41805 L .41746 .46628 L p .724 .624 .82 r
F P
0 g
s
.41746 .46628 m .45264 .45467 L .43781 .4062 L p .724 .624 .82 r
F P
0 g
s
.45264 .45467 m .41746 .46628 L .42871 .51463 L p .712 .587 .786 r
F P
0 g
s
.34081 .28345 m .30594 .29606 L .32931 .338 L p .753 .73 .915 r
F P
0 g
s
.32931 .338 m .36405 .32571 L .34081 .28345 L p .753 .73 .915 r
F P
0 g
s
.36914 .6296 m .33437 .6403 L .33275 .68286 L p .657 .443 .645 r
F P
0 g
s
.298 .69333 m .33275 .68286 L .33437 .6403 L p .657 .443 .645 r
F P
0 g
s
.39917 .31328 m .36405 .32571 L .38482 .37088 L p .744 .694 .884 r
F P
0 g
s
.38482 .37088 m .41988 .35875 L .39917 .31328 L p .744 .694 .884 r
F P
0 g
s
.18378 .25078 m .14958 .26363 L .17746 .29596 L p .767 .81 .979 r
F P
0 g
s
.17746 .29596 m .21143 .28343 L .18378 .25078 L p .767 .81 .979 r
F P
0 g
s
.13135 .35977 m .17746 .29596 L .14958 .26363 L p .973 .835 .807 r
F P
0 g
s
.17746 .29596 m .13135 .35977 L .15151 .3876 L p .96 .814 .799 r
F P
0 g
s
.13135 .35977 m .09755 .37199 L .11784 .39958 L p .758 .77 .95 r
F P
0 g
s
.11784 .39958 m .15151 .3876 L .13135 .35977 L p .758 .77 .95 r
F P
0 g
s
.28044 .25798 m .24575 .27077 L .27144 .30853 L p .761 .768 .946 r
F P
0 g
s
.27144 .30853 m .30594 .29606 L .28044 .25798 L p .761 .768 .946 r
F P
0 g
s
.42871 .51463 m .46406 .50324 L .45264 .45467 L p .712 .587 .786 r
F P
0 g
s
.46406 .50324 m .42871 .51463 L .43623 .56214 L p .697 .546 .747 r
F P
0 g
s
.40102 .57323 m .43623 .56214 L .42871 .51463 L p .697 .546 .747 r
F P
0 g
s
.43623 .56214 m .40102 .57323 L .40429 .61878 L p .68 .499 .701 r
F P
0 g
s
.36914 .6296 m .40429 .61878 L .40102 .57323 L p .68 .499 .701 r
F P
0 g
s
.19062 .64029 m .22476 .62955 L .22339 .59529 L p .675 .493 .697 r
F P
0 g
s
.41988 .35875 m .43781 .4062 L p .39468 .28209 L .1 .1 .445 r
F P
0 g
s
.39997 .29579 m .39343 .27848 L p .39468 .28209 L .1 .1 .445 r
F P
0 g
s
.43781 .4062 m .39997 .29579 L p .39468 .28209 L .1 .1 .445 r
F P
0 g
s
.442 .30042 m .43658 .28274 L p .40637 .31238 L .72 .622 .821 r
F P
0 g
s
.39997 .29579 m .43781 .4062 L .45264 .45467 L p .857 .55 .1 r
F P
0 g
s
.40535 .31338 m .39997 .29579 L p .40121 .29953 L .857 .55 .1 r
F P
0 g
s
.45264 .45467 m .40535 .31338 L p .40121 .29953 L .857 .55 .1 r
F P
0 g
s
.39424 .28093 m .41988 .35875 L p .39468 .28209 L .1 .1 .445 r
F P
0 g
s
.39917 .31328 m .41988 .35875 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.38589 .26189 m .38838 .27151 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.38838 .27151 m .39917 .31328 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.44615 .31795 m .442 .30042 L p .41649 .32424 L .707 .582 .784 r
F P
0 g
s
.39343 .27848 m .38589 .26189 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.41988 .35875 m .3956 .28505 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.3956 .28505 m .39343 .27848 L p .39098 .2764 L .1 .1 .648 r
F P
0 g
s
.45264 .45467 m .46406 .50324 L p .40535 .31338 L .983 .687 .174 r
F P
0 g
s
.40944 .33083 m .40535 .31338 L p .40652 .31715 L .983 .687 .174 r
F P
0 g
s
.46406 .50324 m .40944 .33083 L p .40652 .31715 L .983 .687 .174 r
F P
0 g
s
.43623 .56214 m .47182 .55095 L .46406 .50324 L p .697 .546 .747 r
F P
0 g
s
.41213 .34771 m .40944 .33083 L p .4105 .33458 L 1 .768 .316 r
F P
0 g
s
.47182 .55095 m .41213 .34771 L p .4105 .33458 L 1 .768 .316 r
F P
0 g
s
.47182 .55095 m .43623 .56214 L .43981 .60786 L p .68 .499 .701 r
F P
0 g
s
.43981 .60786 m .47571 .59681 L .47182 .55095 L p .68 .499 .701 r
F P
0 g
s
.41335 .36359 m .41213 .34771 L p .41307 .3514 L 1 .811 .415 r
F P
0 g
s
.47571 .59681 m .41335 .36359 L p .41307 .3514 L 1 .811 .415 r
F P
0 g
s
.47571 .59681 m .43981 .60786 L .43926 .65077 L p .657 .443 .645 r
F P
0 g
s
.43926 .65077 m .47554 .63984 L .47571 .59681 L p .657 .443 .645 r
F P
0 g
s
.41306 .37808 m .41335 .36359 L p .41416 .3672 L 1 .833 .482 r
F P
0 g
s
.47554 .63984 m .41306 .37808 L p .41416 .3672 L 1 .833 .482 r
F P
0 g
s
.47554 .63984 m .43926 .65077 L .43449 .68988 L p .626 .372 .573 r
F P
0 g
s
.43449 .68988 m .47119 .67903 L .47554 .63984 L p .626 .372 .573 r
F P
0 g
s
.41124 .3908 m .41306 .37808 L p .41374 .3816 L 1 .843 .529 r
F P
0 g
s
.47119 .67903 m .41124 .3908 L p .41374 .3816 L 1 .843 .529 r
F P
0 g
s
.40429 .61878 m .43981 .60786 L .43623 .56214 L p .68 .499 .701 r
F P
0 g
s
.47119 .67903 m .43449 .68988 L .42542 .72416 L p .579 .276 .473 r
F P
0 g
s
.42542 .72416 m .4626 .71334 L .47119 .67903 L p .579 .276 .473 r
F P
0 g
s
.40792 .40138 m .41124 .3908 L p .41178 .39421 L 1 .847 .562 r
F P
0 g
s
.4626 .71334 m .40792 .40138 L p .41178 .39421 L 1 .847 .562 r
F P
0 g
s
.43981 .60786 m .40429 .61878 L .40338 .66158 L p .657 .443 .645 r
F P
0 g
s
.40338 .66158 m .43926 .65077 L .43981 .60786 L p .657 .443 .645 r
F P
0 g
s
.43926 .65077 m .40338 .66158 L .39818 .70061 L p .626 .372 .573 r
F P
0 g
s
.39818 .70061 m .43449 .68988 L .43926 .65077 L p .626 .372 .573 r
F P
0 g
s
.4626 .71334 m .42542 .72416 L .4121 .75259 L p .505 .139 .326 r
F P
0 g
s
.4121 .75259 m .44979 .74177 L .4626 .71334 L p .505 .139 .326 r
F P
0 g
s
.40317 .40953 m .40792 .40138 L p .40833 .4047 L 1 .847 .586 r
F P
0 g
s
.40429 .61878 m .36914 .6296 L .36788 .67228 L p .657 .443 .645 r
F P
0 g
s
.36788 .67228 m .40338 .66158 L .40429 .61878 L p .657 .443 .645 r
F P
0 g
s
.43449 .68988 m .39818 .70061 L .38865 .73485 L p .579 .276 .473 r
F P
0 g
s
.38865 .73485 m .42542 .72416 L .43449 .68988 L p .579 .276 .473 r
F P
0 g
s
.42542 .72416 m .38865 .73485 L .37482 .7633 L p .505 .139 .326 r
F P
0 g
s
.37482 .7633 m .4121 .75259 L .42542 .72416 L p .505 .139 .326 r
F P
0 g
s
.4121 .75259 m .37482 .7633 L .35684 .78496 L p .377 .1 .1 r
F P
0 g
s
.44979 .74177 m .4121 .75259 L .39464 .7742 L p .377 .1 .1 r
F P
0 g
s
.35684 .78496 m .39464 .7742 L .4121 .75259 L p .377 .1 .1 r
F P
0 g
s
.39464 .7742 m .43287 .76331 L .44979 .74177 L p .377 .1 .1 r
F P
0 g
s
.38981 .41758 m .39709 .415 L p .39722 .41809 L 1 .842 .617 r
F P
0 g
s
.43287 .76331 m .39464 .7742 L .3733 .78806 L p .154 .1 .1 r
F P
0 g
s
.3733 .78806 m .41208 .77707 L .43287 .76331 L p .154 .1 .1 r
F P
0 g
s
.41208 .77707 m .38981 .41758 L p .39722 .41809 L 1 .842 .617 r
F P
0 g
s
.38981 .41758 m .38152 .41716 L p .38195 .41704 L .58 .991 1 r
F P
0 g
s
.39709 .415 m .38981 .41758 L p .39122 .4171 L .1 .1 .1 r
F P
0 g
s
.44979 .74177 m .40361 .41268 L p .40907 .41073 L 1 .847 .586 r
F P
0 g
s
.40944 .33083 m .46406 .50324 L .47182 .55095 L p 1 .768 .316 r
F P
0 g
s
.41213 .34771 m .47182 .55095 L .47571 .59681 L p 1 .811 .415 r
F P
0 g
s
.41335 .36359 m .47571 .59681 L .47554 .63984 L p 1 .833 .482 r
F P
0 g
s
.41306 .37808 m .47554 .63984 L .47119 .67903 L p 1 .843 .529 r
F P
0 g
s
.41124 .3908 m .47119 .67903 L .4626 .71334 L p 1 .847 .562 r
F P
0 g
s
.4626 .71334 m .44979 .74177 L p .40833 .4047 L 1 .847 .586 r
F P
0 g
s
.40792 .40138 m .40848 .40456 L p .40833 .4047 L 1 .847 .586 r
F P
0 g
s
.40848 .40456 m .4626 .71334 L p .40833 .4047 L 1 .847 .586 r
F P
0 g
s
.40907 .41073 m .40317 .40953 L .40833 .4047 L 1 .847 .586 r
F
.40361 .41268 m .40317 .40953 L p .40907 .41073 L F P
0 g
s
.44979 .74177 m .43287 .76331 L p .40394 .41869 L 1 .845 .604 r
F P
0 g
s
.40317 .40953 m .40446 .41873 L p .40394 .41869 L 1 .845 .604 r
F P
0 g
s
.40446 .41873 m .44979 .74177 L p .40394 .41869 L 1 .845 .604 r
F P
0 g
s
.38195 .41704 m .38287 .4168 L .38981 .41758 L .58 .991 1 r
F
.39709 .415 m .40317 .40953 L p .40344 .41274 L 1 .845 .604 r
F P
0 g
s
.43287 .76331 m .39709 .415 L p .40344 .41274 L 1 .845 .604 r
F P
0 g
s
.39709 .415 m .43287 .76331 L .41208 .77707 L p 1 .842 .617 r
F P
0 g
s
.42775 .40473 m .43484 .40221 L p .39026 .41743 L .1 .1 .1 r
F P
0 g
s
.38981 .41758 m .39019 .41745 L p .39026 .41743 L .1 .1 .1 r
F P
0 g
s
.39019 .41745 m .42775 .40473 L p .39026 .41743 L .1 .1 .1 r
F P
0 g
s
.43484 .40221 m .44074 .39679 L p .39755 .41481 L .199 .1 .1 r
F P
0 g
s
.39709 .415 m .39749 .41486 L p .39755 .41481 L .199 .1 .1 r
F P
0 g
s
.39749 .41486 m .43484 .40221 L p .39755 .41481 L .199 .1 .1 r
F P
0 g
s
.43484 .40221 m .42605 .40519 L p .42637 .4051 L .1 .1 .1 r
F P
0 g
s
.42605 .40519 m .39749 .41486 L p .42637 .4051 L .1 .1 .1 r
F P
0 g
s
.39122 .4171 m .42637 .4051 L .39749 .41486 L .1 .1 .1 r
F
.39749 .41486 m .39709 .415 L p .39122 .4171 L F P
0 g
s
.40317 .40953 m .39709 .415 L p .39755 .41481 L .199 .1 .1 r
F P
0 g
s
.44074 .39679 m .40317 .40953 L p .39755 .41481 L .199 .1 .1 r
F P
0 g
s
.42775 .40473 m .39019 .41745 L p .38287 .4168 L .58 .991 1 r
F P
0 g
s
.39019 .41745 m .38981 .41758 L p .38287 .4168 L .58 .991 1 r
F P
0 g
s
.40792 .40138 m .40317 .40953 L p .40366 .40929 L .411 .1 .198 r
F P
0 g
s
.44531 .38866 m .40792 .40138 L p .40366 .40929 L .411 .1 .198 r
F P
0 g
s
.41124 .3908 m .40792 .40138 L p .40844 .40108 L .529 .202 .407 r
F P
0 g
s
.44845 .37808 m .41124 .3908 L p .40844 .40108 L .529 .202 .407 r
F P
0 g
s
.41306 .37808 m .41124 .3908 L p .4118 .39043 L .597 .329 .538 r
F P
0 g
s
.45012 .36536 m .41306 .37808 L p .4118 .39043 L .597 .329 .538 r
F P
0 g
s
.41335 .36359 m .41306 .37808 L p .41367 .37763 L .639 .418 .627 r
F P
0 g
s
.45027 .35083 m .41335 .36359 L p .41367 .37763 L .639 .418 .627 r
F P
0 g
s
.41213 .34771 m .41335 .36359 L p .41403 .36304 L .669 .484 .692 r
F P
0 g
s
.44894 .3349 m .41213 .34771 L p .41403 .36304 L .669 .484 .692 r
F P
0 g
s
.40944 .33083 m .41213 .34771 L p .4129 .34703 L .69 .537 .742 r
F P
0 g
s
.44615 .31795 m .40944 .33083 L p .4129 .34703 L .69 .537 .742 r
F P
0 g
s
.41032 .33001 m .41649 .32424 L .40535 .31338 L .707 .582 .784 r
F
.442 .30042 m .40985 .31179 L p .41649 .32424 L F P
0 g
s
.40985 .31179 m .40535 .31338 L p .41649 .32424 L .707 .582 .784 r
F P
0 g
s
.40944 .33083 m .41834 .32771 L p .41649 .32424 L .707 .582 .784 r
F P
0 g
s
.41834 .32771 m .44615 .31795 L p .41649 .32424 L .707 .582 .784 r
F P
0 g
s
.40535 .31338 m .40944 .33083 L p .41032 .33001 L .707 .582 .784 r
F P
0 g
s
.40535 .31338 m .40659 .31294 L p .40637 .31238 L .72 .622 .821 r
F P
0 g
s
.40659 .31294 m .442 .30042 L p .40637 .31238 L .72 .622 .821 r
F P
0 g
s
.39997 .29579 m .40535 .31338 L p .40637 .31238 L .72 .622 .821 r
F P
0 g
s
.43658 .28274 m .39997 .29579 L p .40637 .31238 L .72 .622 .821 r
F P
0 g
s
.39343 .27848 m .39424 .28093 L p .39468 .28209 L .1 .1 .445 r
F P
0 g
s
.22476 .62955 m .19062 .64029 L .18829 .67225 L p .651 .434 .639 r
F P
0 g
s
.33275 .68286 m .298 .69333 L .2916 .73212 L p .626 .372 .573 r
F P
0 g
s
.298 .69333 m .22269 .6616 L .21715 .69066 L p .925 .653 .628 r
F P
0 g
s
.21715 .69066 m .2916 .73212 L .298 .69333 L p .925 .653 .628 r
F P
0 g
s
.37605 .27071 m .34081 .28345 L .36405 .32571 L p .753 .73 .915 r
F P
0 g
s
.36405 .32571 m .39917 .31328 L .37605 .27071 L p .753 .73 .915 r
F P
0 g
s
.33275 .68286 m .36788 .67228 L .36914 .6296 L p .657 .443 .645 r
F P
0 g
s
.21836 .2378 m .18378 .25078 L .21143 .28343 L p .767 .81 .979 r
F P
0 g
s
.21143 .28343 m .24575 .27077 L .21836 .2378 L p .767 .81 .979 r
F P
0 g
s
.31551 .24505 m .28044 .25798 L .30594 .29606 L p .761 .768 .946 r
F P
0 g
s
.30594 .29606 m .34081 .28345 L .31551 .24505 L p .761 .768 .946 r
F P
0 g
s
.10982 .33582 m .07584 .34827 L .09755 .37199 L p .765 .813 .984 r
F P
0 g
s
.09755 .37199 m .13135 .35977 L .10982 .33582 L p .765 .813 .984 r
F P
0 g
s
.18829 .67225 m .22269 .6616 L .22476 .62955 L p .651 .434 .639 r
F P
0 g
s
.36788 .67228 m .33275 .68286 L .32674 .72173 L p .626 .372 .573 r
F P
0 g
s
.2916 .73212 m .32674 .72173 L .33275 .68286 L p .626 .372 .573 r
F P
0 g
s
.37753 .24642 m .37605 .27071 L .39917 .31328 L p .1 .248 .791 r
F P
0 g
s
.39917 .31328 m .38589 .26189 L .37753 .24642 L p .1 .248 .791 r
F P
0 g
s
.2533 .22468 m .21836 .2378 L .24575 .27077 L p .767 .81 .979 r
F P
0 g
s
.24575 .27077 m .28044 .25798 L .2533 .22468 L p .767 .81 .979 r
F P
0 g
s
.22269 .6616 m .18829 .67225 L .18246 .70126 L p .617 .358 .562 r
F P
0 g
s
.14958 .26363 m .10982 .33582 L .13135 .35977 L p .973 .835 .807 r
F P
0 g
s
.35095 .23198 m .31551 .24505 L .34081 .28345 L p .761 .768 .946 r
F P
0 g
s
.34081 .28345 m .37605 .27071 L .35095 .23198 L p .761 .768 .946 r
F P
0 g
s
.40338 .66158 m .36788 .67228 L .36227 .71123 L p .626 .372 .573 r
F P
0 g
s
.32674 .72173 m .36227 .71123 L .36788 .67228 L p .626 .372 .573 r
F P
0 g
s
.28863 .21142 m .2533 .22468 L .28044 .25798 L p .767 .81 .979 r
F P
0 g
s
.28044 .25798 m .31551 .24505 L .28863 .21142 L p .767 .81 .979 r
F P
0 g
s
.18246 .70126 m .21715 .69066 L .22269 .6616 L p .617 .358 .562 r
F P
0 g
s
.32674 .72173 m .2916 .73212 L .28073 .76625 L p .579 .276 .473 r
F P
0 g
s
.2916 .73212 m .21715 .69066 L .20817 .71597 L p .929 .636 .595 r
F P
0 g
s
.20817 .71597 m .28073 .76625 L .2916 .73212 L p .929 .636 .595 r
F P
0 g
s
.36855 .23247 m .35095 .23198 L .37605 .27071 L p .1 .369 .875 r
F P
0 g
s
.37605 .27071 m .37753 .24642 L .36855 .23247 L p .1 .369 .875 r
F P
0 g
s
.21715 .69066 m .18246 .70126 L .17317 .72653 L p .566 .256 .455 r
F P
0 g
s
.36227 .71123 m .39818 .70061 L .40338 .66158 L p .626 .372 .573 r
F P
0 g
s
.32434 .19801 m .28863 .21142 L .31551 .24505 L p .767 .81 .979 r
F P
0 g
s
.31551 .24505 m .35095 .23198 L .32434 .19801 L p .767 .81 .979 r
F P
0 g
s
.36227 .71123 m .32674 .72173 L .31631 .7559 L p .579 .276 .473 r
F P
0 g
s
.28073 .76625 m .31631 .7559 L .32674 .72173 L p .579 .276 .473 r
F P
0 g
s
.17317 .72653 m .20817 .71597 L .21715 .69066 L p .566 .256 .455 r
F P
0 g
s
.35917 .2204 m .32434 .19801 L .35095 .23198 L p .1 .42 .907 r
F P
0 g
s
.35095 .23198 m .36855 .23247 L .35917 .2204 L p .1 .42 .907 r
F P
0 g
s
.39818 .70061 m .36227 .71123 L .35228 .74543 L p .579 .276 .473 r
F P
0 g
s
.31631 .7559 m .35228 .74543 L .36227 .71123 L p .579 .276 .473 r
F P
0 g
s
.31631 .7559 m .28073 .76625 L .26547 .7947 L p .505 .139 .326 r
F P
0 g
s
.28073 .76625 m .20817 .71597 L .19586 .73677 L p .934 .619 .557 r
F P
0 g
s
.19586 .73677 m .26547 .7947 L .28073 .76625 L p .934 .619 .557 r
F P
0 g
s
.20817 .71597 m .17317 .72653 L .16052 .74734 L p .483 .106 .294 r
F P
0 g
s
.35228 .74543 m .38865 .73485 L .39818 .70061 L p .579 .276 .473 r
F P
0 g
s
.35228 .74543 m .31631 .7559 L .30152 .78435 L p .505 .139 .326 r
F P
0 g
s
.26547 .7947 m .30152 .78435 L .31631 .7559 L p .505 .139 .326 r
F P
0 g
s
.16052 .74734 m .19586 .73677 L .20817 .71597 L p .483 .106 .294 r
F P
0 g
s
.43004 .26533 m .39343 .27848 L .39997 .29579 L p .732 .659 .854 r
F P
0 g
s
.39997 .29579 m .43658 .28274 L .43004 .26533 L p .732 .659 .854 r
F P
0 g
s
.38865 .73485 m .35228 .74543 L .33797 .77388 L p .505 .139 .326 r
F P
0 g
s
.30152 .78435 m .33797 .77388 L .35228 .74543 L p .505 .139 .326 r
F P
0 g
s
.42252 .24862 m .38589 .26189 L .39343 .27848 L p .742 .695 .886 r
F P
0 g
s
.39343 .27848 m .43004 .26533 L .42252 .24862 L p .742 .695 .886 r
F P
0 g
s
.19586 .73677 m .16052 .74734 L .14473 .763 L p .337 .1 .1 r
F P
0 g
s
.41421 .23303 m .37753 .24642 L .38589 .26189 L p .751 .731 .917 r
F P
0 g
s
.38589 .26189 m .42252 .24862 L .41421 .23303 L p .751 .731 .917 r
F P
0 g
s
.41213 .34771 m .44894 .3349 L .44615 .31795 L p .69 .537 .742 r
F P
0 g
s
.30152 .78435 m .26547 .7947 L .24598 .81652 L p .377 .1 .1 r
F P
0 g
s
.26547 .7947 m .19586 .73677 L .18042 .75238 L p .941 .602 .513 r
F P
0 g
s
.18042 .75238 m .24598 .81652 L .26547 .7947 L p .941 .602 .513 r
F P
0 g
s
.4053 .21896 m .36855 .23247 L .37753 .24642 L p .759 .77 .949 r
F P
0 g
s
.37753 .24642 m .41421 .23303 L .4053 .21896 L p .759 .77 .949 r
F P
0 g
s
.46705 .25202 m .43004 .26533 L .43658 .28274 L p .732 .659 .854 r
F P
0 g
s
.43658 .28274 m .47361 .26954 L .46705 .25202 L p .732 .659 .854 r
F P
0 g
s
.47361 .26954 m .43658 .28274 L .442 .30042 L p .72 .622 .821 r
F P
0 g
s
.442 .30042 m .47906 .28732 L .47361 .26954 L p .72 .622 .821 r
F P
0 g
s
.47906 .28732 m .442 .30042 L .44615 .31795 L p .707 .582 .784 r
F P
0 g
s
.33797 .77388 m .37482 .7633 L .38865 .73485 L p .505 .139 .326 r
F P
0 g
s
.45955 .2352 m .42252 .24862 L .43004 .26533 L p .742 .695 .886 r
F P
0 g
s
.43004 .26533 m .46705 .25202 L .45955 .2352 L p .742 .695 .886 r
F P
0 g
s
.41335 .36359 m .45027 .35083 L .44894 .3349 L p .669 .484 .692 r
F P
0 g
s
.14473 .763 m .18042 .75238 L .19586 .73677 L p .337 .1 .1 r
F P
0 g
s
.44615 .31795 m .48328 .30493 L .47906 .28732 L p .707 .582 .784 r
F P
0 g
s
.48328 .30493 m .44615 .31795 L .44894 .3349 L p .69 .537 .742 r
F P
0 g
s
.45129 .21949 m .41421 .23303 L .42252 .24862 L p .751 .731 .917 r
F P
0 g
s
.42252 .24862 m .45955 .2352 L .45129 .21949 L p .751 .731 .917 r
F P
0 g
s
.39602 .20676 m .35917 .2204 L .36855 .23247 L p .766 .812 .982 r
F P
0 g
s
.36855 .23247 m .4053 .21896 L .39602 .20676 L p .766 .812 .982 r
F P
0 g
s
.44894 .3349 m .48615 .32194 L .48328 .30493 L p .69 .537 .742 r
F P
0 g
s
.48615 .32194 m .44894 .3349 L .45027 .35083 L p .669 .484 .692 r
F P
0 g
s
.41306 .37808 m .45012 .36536 L .45027 .35083 L p .639 .418 .627 r
F P
0 g
s
.33797 .77388 m .30152 .78435 L .28252 .80612 L p .377 .1 .1 r
F P
0 g
s
.24598 .81652 m .28252 .80612 L .30152 .78435 L p .377 .1 .1 r
F P
0 g
s
.44247 .20529 m .4053 .21896 L .41421 .23303 L p .759 .77 .949 r
F P
0 g
s
.41421 .23303 m .45129 .21949 L .44247 .20529 L p .759 .77 .949 r
F P
0 g
s
.50448 .23857 m .46705 .25202 L .47361 .26954 L p .732 .659 .854 r
F P
0 g
s
.47361 .26954 m .51104 .25619 L .50448 .23857 L p .732 .659 .854 r
F P
0 g
s
.51104 .25619 m .47361 .26954 L .47906 .28732 L p .72 .622 .821 r
F P
0 g
s
.47906 .28732 m .51653 .27407 L .51104 .25619 L p .72 .622 .821 r
F P
0 g
s
.51653 .27407 m .47906 .28732 L .48328 .30493 L p .707 .582 .784 r
F P
0 g
s
.497 .22163 m .45955 .2352 L .46705 .25202 L p .742 .695 .886 r
F P
0 g
s
.46705 .25202 m .50448 .23857 L .497 .22163 L p .742 .695 .886 r
F P
0 g
s
.45027 .35083 m .48761 .33793 L .48615 .32194 L p .669 .484 .692 r
F P
0 g
s
.48761 .33793 m .45027 .35083 L .45012 .36536 L p .639 .418 .627 r
F P
0 g
s
.48328 .30493 m .52081 .29177 L .51653 .27407 L p .707 .582 .784 r
F P
0 g
s
.52081 .29177 m .48328 .30493 L .48615 .32194 L p .69 .537 .742 r
F P
0 g
s
.18042 .75238 m .14473 .763 L .1261 .77289 L p .1 .1 .1 r
F P
0 g
s
.48879 .2058 m .45129 .21949 L .45955 .2352 L p .751 .731 .917 r
F P
0 g
s
.45955 .2352 m .497 .22163 L .48879 .2058 L p .751 .731 .917 r
F P
0 g
s
.43329 .19297 m .39602 .20676 L .4053 .21896 L p .766 .812 .982 r
F P
0 g
s
.4053 .21896 m .44247 .20529 L .43329 .19297 L p .766 .812 .982 r
F P
0 g
s
.48615 .32194 m .52379 .30885 L .52081 .29177 L p .69 .537 .742 r
F P
0 g
s
.52379 .30885 m .48615 .32194 L .48761 .33793 L p .669 .484 .692 r
F P
0 g
s
.41124 .3908 m .44845 .37808 L .45012 .36536 L p .597 .329 .538 r
F P
0 g
s
.45012 .36536 m .48759 .35249 L .48761 .33793 L p .639 .418 .627 r
F P
0 g
s
.48759 .35249 m .45012 .36536 L .44845 .37808 L p .597 .329 .538 r
F P
0 g
s
.37482 .7633 m .33797 .77388 L .31947 .7956 L p .377 .1 .1 r
F P
0 g
s
.28252 .80612 m .31947 .7956 L .33797 .77388 L p .377 .1 .1 r
F P
0 g
s
.48005 .19147 m .44247 .20529 L .45129 .21949 L p .759 .77 .949 r
F P
0 g
s
.45129 .21949 m .48879 .2058 L .48005 .19147 L p .759 .77 .949 r
F P
0 g
s
.54232 .22497 m .50448 .23857 L .51104 .25619 L p .732 .659 .854 r
F P
0 g
s
.51104 .25619 m .54889 .2427 L .54232 .22497 L p .732 .659 .854 r
F P
0 g
s
.54889 .2427 m .51104 .25619 L .51653 .27407 L p .72 .622 .821 r
F P
0 g
s
.51653 .27407 m .55442 .26068 L .54889 .2427 L p .72 .622 .821 r
F P
0 g
s
.55442 .26068 m .51653 .27407 L .52081 .29177 L p .707 .582 .784 r
F P
0 g
s
.53487 .20791 m .497 .22163 L .50448 .23857 L p .742 .695 .886 r
F P
0 g
s
.50448 .23857 m .54232 .22497 L .53487 .20791 L p .742 .695 .886 r
F P
0 g
s
.48761 .33793 m .52536 .32488 L .52379 .30885 L p .669 .484 .692 r
F P
0 g
s
.52536 .32488 m .48761 .33793 L .48759 .35249 L p .639 .418 .627 r
F P
0 g
s
.52081 .29177 m .55877 .27846 L .55442 .26068 L p .707 .582 .784 r
F P
0 g
s
.55877 .27846 m .52081 .29177 L .52379 .30885 L p .69 .537 .742 r
F P
0 g
s
.24598 .81652 m .18042 .75238 L .16214 .76219 L p .949 .585 .462 r
F P
0 g
s
.1261 .77289 m .16214 .76219 L p .17634 .75392 L .1 .1 .1 r
F P
0 g
s
.16214 .76219 m .1261 .77289 L .10504 .77654 L p .409 .9 .898 r
F P
0 g
s
.14142 .76572 m .16214 .76219 L p .15785 .76327 L .409 .9 .898 r
F P
0 g
s
.10504 .77654 m .14142 .76572 L p .15785 .76327 L .409 .9 .898 r
F P
0 g
s
.14142 .76572 m .10504 .77654 L .08204 .77359 L p .614 .987 1 r
F P
0 g
s
.11875 .76262 m .14142 .76572 L p .13695 .76631 L .614 .987 1 r
F P
0 g
s
.08204 .77359 m .11875 .76262 L p .13695 .76631 L .614 .987 1 r
F P
0 g
s
.09472 .75273 m .11875 .76262 L p .11416 .76271 L .703 .957 1 r
F P
0 g
s
.05772 .76388 m .09472 .75273 L p .11416 .76271 L .703 .957 1 r
F P
0 g
s
.03278 .74748 m .07004 .7361 L .09472 .75273 L p .734 .898 1 r
F P
0 g
s
.09472 .75273 m .07004 .7361 L p .07484 .7425 L .965 .517 .193 r
F P
0 g
s
.11875 .76262 m .08204 .77359 L .05772 .76388 L p .703 .957 1 r
F P
0 g
s
.09472 .75273 m .05772 .76388 L .03278 .74748 L p .734 .898 1 r
F P
0 g
s
.13403 .82146 m .09472 .75273 L p .07484 .7425 L .965 .517 .193 r
F P
0 g
s
.11875 .76262 m .09472 .75273 L p .10009 .75883 L .967 .532 .267 r
F P
0 g
s
.16592 .83376 m .11875 .76262 L p .10009 .75883 L .967 .532 .267 r
F P
0 g
s
.14142 .76572 m .11875 .76262 L p .12458 .76823 L .964 .55 .339 r
F P
0 g
s
.19569 .83674 m .14142 .76572 L p .12458 .76823 L .964 .55 .339 r
F P
0 g
s
.16214 .76219 m .14142 .76572 L p .14759 .77067 L .957 .567 .404 r
F P
0 g
s
.22258 .8308 m .16214 .76219 L p .14759 .77067 L .957 .567 .404 r
F P
0 g
s
.16214 .76219 m .18042 .75238 L p .17634 .75392 L .1 .1 .1 r
F P
0 g
s
.52672 .19195 m .48879 .2058 L .497 .22163 L p .751 .731 .917 r
F P
0 g
s
.497 .22163 m .53487 .20791 L .52672 .19195 L p .751 .731 .917 r
F P
0 g
s
.40792 .40138 m .44531 .38866 L .44845 .37808 L p .529 .202 .407 r
F P
0 g
s
.47098 .17902 m .43329 .19297 L .44247 .20529 L p .766 .812 .982 r
F P
0 g
s
.44247 .20529 m .48005 .19147 L .47098 .17902 L p .766 .812 .982 r
F P
0 g
s
.52379 .30885 m .56184 .2956 L .55877 .27846 L p .69 .537 .742 r
F P
0 g
s
.56184 .2956 m .52379 .30885 L .52536 .32488 L p .669 .484 .692 r
F P
0 g
s
.44845 .37808 m .48609 .36523 L .48759 .35249 L p .597 .329 .538 r
F P
0 g
s
.48609 .36523 m .44845 .37808 L .44531 .38866 L p .529 .202 .407 r
F P
0 g
s
.28252 .80612 m .24598 .81652 L .22258 .8308 L p .154 .1 .1 r
F P
0 g
s
.16214 .76219 m .22258 .8308 L .24598 .81652 L p .949 .585 .462 r
F P
0 g
s
.48759 .35249 m .52549 .33947 L .52536 .32488 L p .639 .418 .627 r
F P
0 g
s
.52549 .33947 m .48759 .35249 L .48609 .36523 L p .597 .329 .538 r
F P
0 g
s
.31947 .7956 m .35684 .78496 L .37482 .7633 L p .377 .1 .1 r
F P
0 g
s
.51805 .17749 m .48005 .19147 L .48879 .2058 L p .759 .77 .949 r
F P
0 g
s
.48879 .2058 m .52672 .19195 L .51805 .17749 L p .759 .77 .949 r
F P
0 g
s
.58059 .21121 m .54232 .22497 L .54889 .2427 L p .732 .659 .854 r
F P
0 g
s
.54889 .2427 m .58717 .22905 L .58059 .21121 L p .732 .659 .854 r
F P
0 g
s
.58717 .22905 m .54889 .2427 L .55442 .26068 L p .72 .622 .821 r
F P
0 g
s
.55442 .26068 m .59274 .24713 L .58717 .22905 L p .72 .622 .821 r
F P
0 g
s
.59274 .24713 m .55442 .26068 L .55877 .27846 L p .707 .582 .784 r
F P
0 g
s
.57317 .19403 m .53487 .20791 L .54232 .22497 L p .742 .695 .886 r
F P
0 g
s
.54232 .22497 m .58059 .21121 L .57317 .19403 L p .742 .695 .886 r
F P
0 g
s
.52536 .32488 m .56354 .31169 L .56184 .2956 L p .669 .484 .692 r
F P
0 g
s
.56354 .31169 m .52536 .32488 L .52549 .33947 L p .639 .418 .627 r
F P
0 g
s
.55877 .27846 m .59716 .26499 L .59274 .24713 L p .707 .582 .784 r
F P
0 g
s
.59716 .26499 m .55877 .27846 L .56184 .2956 L p .69 .537 .742 r
F P
0 g
s
.56507 .17794 m .52672 .19195 L .53487 .20791 L p .751 .731 .917 r
F P
0 g
s
.53487 .20791 m .57317 .19403 L .56507 .17794 L p .751 .731 .917 r
F P
0 g
s
.40317 .40953 m .44074 .39679 L .44531 .38866 L p .411 .1 .198 r
F P
0 g
s
.44531 .38866 m .48312 .3758 L .48609 .36523 L p .529 .202 .407 r
F P
0 g
s
.48312 .3758 m .44531 .38866 L .44074 .39679 L p .411 .1 .198 r
F P
0 g
s
.5091 .16491 m .47098 .17902 L .48005 .19147 L p .766 .812 .982 r
F P
0 g
s
.48005 .19147 m .51805 .17749 L .5091 .16491 L p .766 .812 .982 r
F P
0 g
s
.56184 .2956 m .60033 .28221 L .59716 .26499 L p .69 .537 .742 r
F P
0 g
s
.60033 .28221 m .56184 .2956 L .56354 .31169 L p .669 .484 .692 r
F P
0 g
s
.48609 .36523 m .52415 .35223 L .52549 .33947 L p .597 .329 .538 r
F P
0 g
s
.52415 .35223 m .48609 .36523 L .48312 .3758 L p .529 .202 .407 r
F P
0 g
s
.31947 .7956 m .28252 .80612 L .25961 .8203 L p .154 .1 .1 r
F P
0 g
s
.22258 .8308 m .25961 .8203 L .28252 .80612 L p .154 .1 .1 r
F P
0 g
s
.52549 .33947 m .56381 .32631 L .56354 .31169 L p .639 .418 .627 r
F P
0 g
s
.56381 .32631 m .52549 .33947 L .52415 .35223 L p .597 .329 .538 r
F P
0 g
s
.55649 .16335 m .51805 .17749 L .52672 .19195 L p .759 .77 .949 r
F P
0 g
s
.52672 .19195 m .56507 .17794 L .55649 .16335 L p .759 .77 .949 r
F P
0 g
s
.6193 .1973 m .58059 .21121 L .58717 .22905 L p .732 .659 .854 r
F P
0 g
s
.58717 .22905 m .62588 .21525 L .6193 .1973 L p .732 .659 .854 r
F P
0 g
s
.62588 .21525 m .58717 .22905 L .59274 .24713 L p .72 .622 .821 r
F P
0 g
s
.59274 .24713 m .63149 .23343 L .62588 .21525 L p .72 .622 .821 r
F P
0 g
s
.63149 .23343 m .59274 .24713 L .59716 .26499 L p .707 .582 .784 r
F P
0 g
s
.6119 .18 m .57317 .19403 L .58059 .21121 L p .742 .695 .886 r
F P
0 g
s
.58059 .21121 m .6193 .1973 L .6119 .18 L p .742 .695 .886 r
F P
0 g
s
.56354 .31169 m .60215 .29835 L .60033 .28221 L p .669 .484 .692 r
F P
0 g
s
.60215 .29835 m .56354 .31169 L .56381 .32631 L p .639 .418 .627 r
F P
0 g
s
.59716 .26499 m .63598 .25138 L .63149 .23343 L p .707 .582 .784 r
F P
0 g
s
.63598 .25138 m .59716 .26499 L .60033 .28221 L p .69 .537 .742 r
F P
0 g
s
.60386 .16378 m .56507 .17794 L .57317 .19403 L p .751 .731 .917 r
F P
0 g
s
.57317 .19403 m .6119 .18 L .60386 .16378 L p .751 .731 .917 r
F P
0 g
s
.44074 .39679 m .47873 .38391 L .48312 .3758 L p .411 .1 .198 r
F P
0 g
s
.47873 .38391 m .44074 .39679 L .43484 .40221 L p .199 .1 .1 r
F P
0 g
s
.48312 .3758 m .52135 .36279 L .52415 .35223 L p .529 .202 .407 r
F P
0 g
s
.52135 .36279 m .48312 .3758 L .47873 .38391 L p .411 .1 .198 r
F P
0 g
s
.54765 .15064 m .5091 .16491 L .51805 .17749 L p .766 .812 .982 r
F P
0 g
s
.51805 .17749 m .55649 .16335 L .54765 .15064 L p .766 .812 .982 r
F P
0 g
s
.60033 .28221 m .63925 .26866 L .63598 .25138 L p .69 .537 .742 r
F P
0 g
s
.63925 .26866 m .60033 .28221 L .60215 .29835 L p .669 .484 .692 r
F P
0 g
s
.52415 .35223 m .56264 .33908 L .56381 .32631 L p .597 .329 .538 r
F P
0 g
s
.56264 .33908 m .52415 .35223 L .52135 .36279 L p .529 .202 .407 r
F P
0 g
s
.35684 .78496 m .31947 .7956 L .29707 .80968 L p .154 .1 .1 r
F P
0 g
s
.25961 .8203 m .29707 .80968 L .31947 .7956 L p .154 .1 .1 r
F P
0 g
s
.56381 .32631 m .60258 .313 L .60215 .29835 L p .639 .418 .627 r
F P
0 g
s
.60258 .313 m .56381 .32631 L .56264 .33908 L p .597 .329 .538 r
F P
0 g
s
.59537 .14906 m .55649 .16335 L .56507 .17794 L p .759 .77 .949 r
F P
0 g
s
.56507 .17794 m .60386 .16378 L .59537 .14906 L p .759 .77 .949 r
F P
0 g
s
.65845 .18323 m .6193 .1973 L .62588 .21525 L p .732 .659 .854 r
F P
0 g
s
.62588 .21525 m .66504 .2013 L .65845 .18323 L p .732 .659 .854 r
F P
0 g
s
.66504 .2013 m .62588 .21525 L .63149 .23343 L p .72 .622 .821 r
F P
0 g
s
.63149 .23343 m .67068 .21957 L .66504 .2013 L p .72 .622 .821 r
F P
0 g
s
.67068 .21957 m .63149 .23343 L .63598 .25138 L p .707 .582 .784 r
F P
0 g
s
.65107 .1658 m .6119 .18 L .6193 .1973 L p .742 .695 .886 r
F P
0 g
s
.6193 .1973 m .65845 .18323 L .65107 .1658 L p .742 .695 .886 r
F P
0 g
s
.60215 .29835 m .6412 .28485 L .63925 .26866 L p .669 .484 .692 r
F P
0 g
s
.6412 .28485 m .60215 .29835 L .60258 .313 L p .639 .418 .627 r
F P
0 g
s
.63598 .25138 m .67525 .23761 L .67068 .21957 L p .707 .582 .784 r
F P
0 g
s
.67525 .23761 m .63598 .25138 L .63925 .26866 L p .69 .537 .742 r
F P
0 g
s
.64309 .14945 m .60386 .16378 L .6119 .18 L p .751 .731 .917 r
F P
0 g
s
.6119 .18 m .65107 .1658 L .64309 .14945 L p .751 .731 .917 r
F P
0 g
s
.47873 .38391 m .51716 .37087 L .52135 .36279 L p .411 .1 .198 r
F P
0 g
s
.51716 .37087 m .47873 .38391 L .47302 .38928 L p .199 .1 .1 r
F P
0 g
s
.43484 .40221 m .47302 .38928 L .47873 .38391 L p .199 .1 .1 r
F P
0 g
s
.52135 .36279 m .56003 .34964 L .56264 .33908 L p .529 .202 .407 r
F P
0 g
s
.56003 .34964 m .52135 .36279 L .51716 .37087 L p .411 .1 .198 r
F P
0 g
s
.58665 .13621 m .54765 .15064 L .55649 .16335 L p .766 .812 .982 r
F P
0 g
s
.55649 .16335 m .59537 .14906 L .58665 .13621 L p .766 .812 .982 r
F P
0 g
s
.47302 .38928 m .43484 .40221 L .42775 .40473 L p .1 .1 .1 r
F P
0 g
s
.39464 .7742 m .35684 .78496 L .33496 .79893 L p .154 .1 .1 r
F P
0 g
s
.33496 .79893 m .3733 .78806 L .39464 .7742 L p .154 .1 .1 r
F P
0 g
s
.38981 .41758 m .41208 .77707 L .38778 .78223 L p 1 .838 .627 r
F P
0 g
s
.38778 .78223 m .38152 .41716 L .38981 .41758 L p 1 .838 .627 r
F P
0 g
s
.63925 .26866 m .67862 .25496 L .67525 .23761 L p .69 .537 .742 r
F P
0 g
s
.67862 .25496 m .63925 .26866 L .6412 .28485 L p .669 .484 .692 r
F P
0 g
s
.56264 .33908 m .60157 .32578 L .60258 .313 L p .597 .329 .538 r
F P
0 g
s
.60157 .32578 m .56264 .33908 L .56003 .34964 L p .529 .202 .407 r
F P
0 g
s
.29707 .80968 m .33496 .79893 L .35684 .78496 L p .154 .1 .1 r
F P
0 g
s
.60258 .313 m .64178 .29953 L .6412 .28485 L p .639 .418 .627 r
F P
0 g
s
.64178 .29953 m .60258 .313 L .60157 .32578 L p .597 .329 .538 r
F P
0 g
s
.14142 .76572 m .19569 .83674 L .22258 .8308 L p .957 .567 .404 r
F P
0 g
s
.25961 .8203 m .22258 .8308 L .19569 .83674 L p .345 .855 .841 r
F P
0 g
s
.6347 .1346 m .59537 .14906 L .60386 .16378 L p .759 .77 .949 r
F P
0 g
s
.60386 .16378 m .64309 .14945 L .6347 .1346 L p .759 .77 .949 r
F P
0 g
s
.69426 .17252 m .65845 .18323 L .66504 .2013 L p .629 .6 .868 r
F P
0 g
s
.66504 .2013 m .69996 .18823 L .69426 .17252 L p .629 .6 .868 r
F P
0 g
s
.69996 .18823 m .66504 .2013 L .67068 .21957 L p .615 .558 .831 r
F P
0 g
s
.67068 .21957 m .70487 .2041 L .69996 .18823 L p .615 .558 .831 r
F P
0 g
s
.70487 .2041 m .67068 .21957 L .67525 .23761 L p .596 .512 .79 r
F P
0 g
s
.68791 .15737 m .65107 .1658 L .65845 .18323 L p .638 .638 .903 r
F P
0 g
s
.65845 .18323 m .69426 .17252 L .68791 .15737 L p .638 .638 .903 r
F P
0 g
s
.6412 .28485 m .68071 .2712 L .67862 .25496 L p .669 .484 .692 r
F P
0 g
s
.68071 .2712 m .6412 .28485 L .64178 .29953 L p .639 .418 .627 r
F P
0 g
s
.67525 .23761 m .70887 .21976 L .70487 .2041 L p .596 .512 .79 r
F P
0 g
s
.70887 .21976 m .67525 .23761 L .67862 .25496 L p .57 .457 .743 r
F P
0 g
s
.68106 .14315 m .64309 .14945 L .65107 .1658 L p .643 .676 .937 r
F P
0 g
s
.65107 .1658 m .68791 .15737 L .68106 .14315 L p .643 .676 .937 r
F P
0 g
s
.51716 .37087 m .55603 .35769 L .56003 .34964 L p .411 .1 .198 r
F P
0 g
s
.55603 .35769 m .51716 .37087 L .51165 .3762 L p .199 .1 .1 r
F P
0 g
s
.47302 .38928 m .51165 .3762 L .51716 .37087 L p .199 .1 .1 r
F P
0 g
s
.56003 .34964 m .59915 .33633 L .60157 .32578 L p .529 .202 .407 r
F P
0 g
s
.59915 .33633 m .56003 .34964 L .55603 .35769 L p .411 .1 .198 r
F P
0 g
s
.62609 .12161 m .58665 .13621 L .59537 .14906 L p .766 .812 .982 r
F P
0 g
s
.59537 .14906 m .6347 .1346 L .62609 .12161 L p .766 .812 .982 r
F P
0 g
s
.51165 .3762 m .47302 .38928 L .46613 .39174 L p .1 .1 .1 r
F P
0 g
s
.42775 .40473 m .46613 .39174 L .47302 .38928 L p .1 .1 .1 r
F P
0 g
s
.46613 .39174 m .42775 .40473 L .41964 .40423 L p .58 .991 1 r
F P
0 g
s
.38152 .41716 m .41964 .40423 L .42775 .40473 L p .58 .991 1 r
F P
0 g
s
.67862 .25496 m .71186 .23479 L .70887 .21976 L p .57 .457 .743 r
F P
0 g
s
.71186 .23479 m .67862 .25496 L .68071 .2712 L p .533 .388 .684 r
F P
0 g
s
.60157 .32578 m .64095 .31232 L .64178 .29953 L p .597 .329 .538 r
F P
0 g
s
.64095 .31232 m .60157 .32578 L .59915 .33633 L p .529 .202 .407 r
F P
0 g
s
.41964 .40423 m .38152 .41716 L .37243 .41369 L p .703 .988 1 r
F P
0 g
s
.38152 .41716 m .38778 .78223 L .36049 .77819 L p 1 .833 .634 r
F P
0 g
s
.36049 .77819 m .37243 .41369 L .38152 .41716 L p 1 .833 .634 r
F P
0 g
s
.64178 .29953 m .68144 .28591 L .68071 .2712 L p .639 .418 .627 r
F P
0 g
s
.68144 .28591 m .64178 .29953 L .64095 .31232 L p .597 .329 .538 r
F P
0 g
s
.29707 .80968 m .25961 .8203 L .23321 .8261 L p .345 .855 .841 r
F P
0 g
s
.19569 .83674 m .23321 .8261 L .25961 .8203 L p .345 .855 .841 r
F P
0 g
s
.6739 .13023 m .6347 .1346 L .64309 .14945 L p .642 .713 .973 r
F P
0 g
s
.64309 .14945 m .68106 .14315 L .6739 .13023 L p .642 .713 .973 r
F P
0 g
s
.68071 .2712 m .71376 .24884 L .71186 .23479 L p .533 .388 .684 r
F P
0 g
s
.71376 .24884 m .68071 .2712 L .68144 .28591 L p .479 .299 .606 r
F P
0 g
s
.55603 .35769 m .59535 .34435 L .59915 .33633 L p .411 .1 .198 r
F P
0 g
s
.59535 .34435 m .55603 .35769 L .55072 .36297 L p .199 .1 .1 r
F P
0 g
s
.51165 .3762 m .55072 .36297 L .55603 .35769 L p .199 .1 .1 r
F P
0 g
s
.59915 .33633 m .63872 .32287 L .64095 .31232 L p .529 .202 .407 r
F P
0 g
s
.63872 .32287 m .59915 .33633 L .59535 .34435 L p .411 .1 .198 r
F P
0 g
s
.66658 .11896 m .62609 .12161 L .6347 .1346 L p .634 .752 1 r
F P
0 g
s
.6347 .1346 m .6739 .13023 L .66658 .11896 L p .634 .752 1 r
F P
0 g
s
.55072 .36297 m .51165 .3762 L .50494 .37859 L p .1 .1 .1 r
F P
0 g
s
.46613 .39174 m .50494 .37859 L .51165 .3762 L p .1 .1 .1 r
F P
0 g
s
.50494 .37859 m .46613 .39174 L .4582 .39115 L p .58 .991 1 r
F P
0 g
s
.41964 .40423 m .4582 .39115 L .46613 .39174 L p .58 .991 1 r
F P
0 g
s
.64095 .31232 m .68079 .29872 L .68144 .28591 L p .597 .329 .538 r
F P
0 g
s
.68079 .29872 m .64095 .31232 L .63872 .32287 L p .529 .202 .407 r
F P
0 g
s
.4582 .39115 m .41964 .40423 L .41072 .40066 L p .703 .988 1 r
F P
0 g
s
.37243 .41369 m .41072 .40066 L .41964 .40423 L p .703 .988 1 r
F P
0 g
s
.41072 .40066 m .37243 .41369 L .36279 .40722 L p .746 .934 1 r
F P
0 g
s
.37243 .41369 m .36049 .77819 L .33088 .76459 L p 1 .828 .64 r
F P
0 g
s
.33088 .76459 m .36279 .40722 L .37243 .41369 L p 1 .828 .64 r
F P
0 g
s
.68144 .28591 m .71451 .26153 L .71376 .24884 L p .479 .299 .606 r
F P
0 g
s
.71451 .26153 m .68144 .28591 L .68079 .29872 L p .397 .173 .495 r
F P
0 g
s
.33496 .79893 m .29707 .80968 L .27118 .81532 L p .345 .855 .841 r
F P
0 g
s
.23321 .8261 m .27118 .81532 L .29707 .80968 L p .345 .855 .841 r
F P
0 g
s
.73291 .15959 m .69426 .17252 L .69996 .18823 L p .69 .635 .861 r
F P
0 g
s
.69996 .18823 m .73822 .17431 L .73291 .15959 L p .69 .635 .861 r
F P
0 g
s
.73822 .17431 m .69996 .18823 L .70487 .2041 L p .678 .596 .826 r
F P
0 g
s
.70487 .2041 m .74283 .18917 L .73822 .17431 L p .678 .596 .826 r
F P
0 g
s
.74283 .18917 m .70487 .2041 L .70887 .21976 L p .662 .554 .788 r
F P
0 g
s
.72702 .14539 m .68791 .15737 L .69426 .17252 L p .7 .673 .895 r
F P
0 g
s
.69426 .17252 m .73291 .15959 L .72702 .14539 L p .7 .673 .895 r
F P
0 g
s
.70887 .21976 m .74661 .20381 L .74283 .18917 L p .662 .554 .788 r
F P
0 g
s
.74661 .20381 m .70887 .21976 L .71186 .23479 L p .641 .504 .744 r
F P
0 g
s
.7207 .13206 m .68106 .14315 L .68791 .15737 L p .708 .71 .928 r
F P
0 g
s
.68791 .15737 m .72702 .14539 L .7207 .13206 L p .708 .71 .928 r
F P
0 g
s
.71186 .23479 m .74948 .21786 L .74661 .20381 L p .641 .504 .744 r
F P
0 g
s
.74948 .21786 m .71186 .23479 L .71376 .24884 L p .614 .445 .69 r
F P
0 g
s
.59535 .34435 m .63512 .33086 L .63872 .32287 L p .411 .1 .198 r
F P
0 g
s
.63512 .33086 m .59535 .34435 L .59024 .34958 L p .199 .1 .1 r
F P
0 g
s
.55072 .36297 m .59024 .34958 L .59535 .34435 L p .199 .1 .1 r
F P
0 g
s
.63872 .32287 m .67876 .30925 L .68079 .29872 L p .529 .202 .407 r
F P
0 g
s
.67876 .30925 m .63872 .32287 L .63512 .33086 L p .411 .1 .198 r
F P
0 g
s
.59024 .34958 m .55072 .36297 L .54421 .36529 L p .1 .1 .1 r
F P
0 g
s
.50494 .37859 m .54421 .36529 L .55072 .36297 L p .1 .1 .1 r
F P
0 g
s
.54421 .36529 m .50494 .37859 L .49721 .37792 L p .58 .991 1 r
F P
0 g
s
.4582 .39115 m .49721 .37792 L .50494 .37859 L p .58 .991 1 r
F P
0 g
s
.36279 .40722 m .40123 .39408 L .41072 .40066 L p .746 .934 1 r
F P
0 g
s
.40123 .39408 m .36279 .40722 L .35285 .39788 L p .757 .875 1 r
F P
0 g
s
.36279 .40722 m .33088 .76459 L .29977 .74136 L p 1 .823 .644 r
F P
0 g
s
.29977 .74136 m .35285 .39788 L .36279 .40722 L p 1 .823 .644 r
F P
0 g
s
.68079 .29872 m .71408 .27253 L .71451 .26153 L p .397 .173 .495 r
F P
0 g
s
.71408 .27253 m .68079 .29872 L .67876 .30925 L p .264 .1 .325 r
F P
0 g
s
.49721 .37792 m .4582 .39115 L .44946 .38749 L p .703 .988 1 r
F P
0 g
s
.41072 .40066 m .44946 .38749 L .4582 .39115 L p .703 .988 1 r
F P
0 g
s
.44946 .38749 m .41072 .40066 L .40123 .39408 L p .746 .934 1 r
F P
0 g
s
.71411 .11994 m .6739 .13023 L .68106 .14315 L p .713 .749 .962 r
F P
0 g
s
.68106 .14315 m .7207 .13206 L .71411 .11994 L p .713 .749 .962 r
F P
0 g
s
.11875 .76262 m .16592 .83376 L .19569 .83674 L p .964 .55 .339 r
F P
0 g
s
.23321 .8261 m .19569 .83674 L .16592 .83376 L p .577 .978 1 r
F P
0 g
s
.3733 .78806 m .33496 .79893 L .30959 .80442 L p .345 .855 .841 r
F P
0 g
s
.27118 .81532 m .30959 .80442 L .33496 .79893 L p .345 .855 .841 r
F P
0 g
s
.71376 .24884 m .75135 .23096 L .74948 .21786 L p .614 .445 .69 r
F P
0 g
s
.75135 .23096 m .71376 .24884 L .71451 .26153 L p .575 .369 .62 r
F P
0 g
s
.70741 .10936 m .66658 .11896 L .6739 .13023 L p .714 .79 .998 r
F P
0 g
s
.6739 .13023 m .71411 .11994 L .70741 .10936 L p .714 .79 .998 r
F P
0 g
s
.77237 .14628 m .73291 .15959 L .73822 .17431 L p .696 .638 .861 r
F P
0 g
s
.73822 .17431 m .77734 .16012 L .77237 .14628 L p .696 .638 .861 r
F P
0 g
s
.77734 .16012 m .73822 .17431 L .74283 .18917 L p .683 .599 .826 r
F P
0 g
s
.74283 .18917 m .78167 .17408 L .77734 .16012 L p .683 .599 .826 r
F P
0 g
s
.78167 .17408 m .74283 .18917 L .74661 .20381 L p .667 .557 .788 r
F P
0 g
s
.76689 .13291 m .72702 .14539 L .73291 .15959 L p .706 .675 .894 r
F P
0 g
s
.73291 .15959 m .77237 .14628 L .76689 .13291 L p .706 .675 .894 r
F P
0 g
s
.71451 .26153 m .75217 .24278 L .75135 .23096 L p .575 .369 .62 r
F P
0 g
s
.75217 .24278 m .71451 .26153 L .71408 .27253 L p .517 .265 .522 r
F P
0 g
s
.35285 .39788 m .39143 .38462 L .40123 .39408 L p .757 .875 1 r
F P
0 g
s
.74661 .20381 m .78526 .18782 L .78167 .17408 L p .667 .557 .788 r
F P
0 g
s
.78526 .18782 m .74661 .20381 L .74948 .21786 L p .647 .508 .744 r
F P
0 g
s
.63512 .33086 m .67536 .31721 L .67876 .30925 L p .411 .1 .198 r
F P
0 g
s
.67536 .31721 m .63512 .33086 L .63023 .33604 L p .199 .1 .1 r
F P
0 g
s
.59024 .34958 m .63023 .33604 L .63512 .33086 L p .199 .1 .1 r
F P
0 g
s
.67876 .30925 m .71247 .28154 L .71408 .27253 L p .264 .1 .325 r
F P
0 g
s
.71247 .28154 m .67876 .30925 L .67536 .31721 L p .1 .1 .1 r
F P
0 g
s
.63023 .33604 m .59024 .34958 L .58394 .35184 L p .1 .1 .1 r
F P
0 g
s
.54421 .36529 m .58394 .35184 L .59024 .34958 L p .1 .1 .1 r
F P
0 g
s
.58394 .35184 m .54421 .36529 L .53668 .36454 L p .58 .991 1 r
F P
0 g
s
.49721 .37792 m .53668 .36454 L .54421 .36529 L p .58 .991 1 r
F P
0 g
s
.40123 .39408 m .44013 .38079 L .44946 .38749 L p .746 .934 1 r
F P
0 g
s
.44013 .38079 m .40123 .39408 L .39143 .38462 L p .757 .875 1 r
F P
0 g
s
.76103 .12036 m .7207 .13206 L .72702 .14539 L p .713 .713 .927 r
F P
0 g
s
.72702 .14539 m .76689 .13291 L .76103 .12036 L p .713 .713 .927 r
F P
0 g
s
.53668 .36454 m .49721 .37792 L .48865 .37416 L p .703 .988 1 r
F P
0 g
s
.44946 .38749 m .48865 .37416 L .49721 .37792 L p .703 .988 1 r
F P
0 g
s
.48865 .37416 m .44946 .38749 L .44013 .38079 L p .746 .934 1 r
F P
0 g
s
.74948 .21786 m .78802 .201 L .78526 .18782 L p .647 .508 .744 r
F P
0 g
s
.78802 .201 m .74948 .21786 L .75135 .23096 L p .621 .45 .69 r
F P
0 g
s
.27118 .81532 m .23321 .8261 L .20391 .82291 L p .577 .978 1 r
F P
0 g
s
.16592 .83376 m .20391 .82291 L .23321 .8261 L p .577 .978 1 r
F P
0 g
s
.41208 .77707 m .3733 .78806 L .34845 .79339 L p .345 .855 .841 r
F P
0 g
s
.30959 .80442 m .34845 .79339 L .3733 .78806 L p .345 .855 .841 r
F P
0 g
s
.75495 .10895 m .71411 .11994 L .7207 .13206 L p .719 .751 .96 r
F P
0 g
s
.7207 .13206 m .76103 .12036 L .75495 .10895 L p .719 .751 .96 r
F P
0 g
s
.71408 .27253 m .75191 .253 L .75217 .24278 L p .517 .265 .522 r
F P
0 g
s
.75191 .253 m .71408 .27253 L .71247 .28154 L p .423 .112 .373 r
F P
0 g
s
.75135 .23096 m .78987 .21327 L .78802 .201 L p .621 .45 .69 r
F P
0 g
s
.78987 .21327 m .75135 .23096 L .75217 .24278 L p .583 .375 .621 r
F P
0 g
s
.39143 .38462 m .43047 .37121 L .44013 .38079 L p .757 .875 1 r
F P
0 g
s
.67536 .31721 m .7097 .2883 L .71247 .28154 L p .1 .1 .1 r
F P
0 g
s
.7097 .2883 m .67536 .31721 L .67068 .32234 L p .1 .1 .1 r
F P
0 g
s
.63023 .33604 m .67068 .32234 L .67536 .31721 L p .199 .1 .1 r
F P
0 g
s
.74879 .09897 m .70741 .10936 L .71411 .11994 L p .72 .793 .996 r
F P
0 g
s
.71411 .11994 m .75495 .10895 L .74879 .09897 L p .72 .793 .996 r
F P
0 g
s
.67068 .32234 m .63023 .33604 L .62414 .33822 L p .1 .1 .1 r
F P
0 g
s
.58394 .35184 m .62414 .33822 L .63023 .33604 L p .1 .1 .1 r
F P
0 g
s
.62414 .33822 m .58394 .35184 L .57661 .35099 L p .58 .991 1 r
F P
0 g
s
.53668 .36454 m .57661 .35099 L .58394 .35184 L p .58 .991 1 r
F P
0 g
s
.44013 .38079 m .47948 .36735 L .48865 .37416 L p .746 .934 1 r
F P
0 g
s
.47948 .36735 m .44013 .38079 L .43047 .37121 L p .757 .875 1 r
F P
0 g
s
.8126 .1326 m .77237 .14628 L .77734 .16012 L p .7 .641 .86 r
F P
0 g
s
.77734 .16012 m .81726 .14566 L .8126 .1326 L p .7 .641 .86 r
F P
0 g
s
.81726 .14566 m .77734 .16012 L .78167 .17408 L p .687 .602 .825 r
F P
0 g
s
.57661 .35099 m .53668 .36454 L .5283 .36067 L p .703 .988 1 r
F P
0 g
s
.48865 .37416 m .5283 .36067 L .53668 .36454 L p .703 .988 1 r
F P
0 g
s
.78167 .17408 m .82136 .15883 L .81726 .14566 L p .687 .602 .825 r
F P
0 g
s
.82136 .15883 m .78167 .17408 L .78526 .18782 L p .672 .56 .787 r
F P
0 g
s
.5283 .36067 m .48865 .37416 L .47948 .36735 L p .746 .934 1 r
F P
0 g
s
.80748 .11998 m .76689 .13291 L .77237 .14628 L p .71 .678 .893 r
F P
0 g
s
.77237 .14628 m .8126 .1326 L .80748 .11998 L p .71 .678 .893 r
F P
0 g
s
.75217 .24278 m .79075 .22432 L .78987 .21327 L p .583 .375 .621 r
F P
0 g
s
.79075 .22432 m .75217 .24278 L .75191 .253 L p .527 .272 .523 r
F P
0 g
s
.78526 .18782 m .82478 .17177 L .82136 .15883 L p .672 .56 .787 r
F P
0 g
s
.82478 .17177 m .78526 .18782 L .78802 .201 L p .652 .512 .743 r
F P
0 g
s
.30959 .80442 m .27118 .81532 L .24236 .81193 L p .577 .978 1 r
F P
0 g
s
.20391 .82291 m .24236 .81193 L .27118 .81532 L p .577 .978 1 r
F P
0 g
s
.34845 .79339 m .38778 .78223 L .41208 .77707 L p .345 .855 .841 r
F P
0 g
s
.71247 .28154 m .75056 .26134 L .75191 .253 L p .423 .112 .373 r
F P
0 g
s
.75056 .26134 m .71247 .28154 L .7097 .2883 L p .261 .1 .134 r
F P
0 g
s
.80204 .10812 m .76103 .12036 L .76689 .13291 L p .718 .715 .926 r
F P
0 g
s
.76689 .13291 m .80748 .11998 L .80204 .10812 L p .718 .715 .926 r
F P
0 g
s
.78802 .201 m .82744 .18417 L .82478 .17177 L p .652 .512 .743 r
F P
0 g
s
.82744 .18417 m .78802 .201 L .78987 .21327 L p .626 .454 .69 r
F P
0 g
s
.09472 .75273 m .13403 .82146 L .16592 .83376 L p .967 .532 .267 r
F P
0 g
s
.79641 .09733 m .75495 .10895 L .76103 .12036 L p .723 .754 .959 r
F P
0 g
s
.76103 .12036 m .80204 .10812 L .79641 .09733 L p .723 .754 .959 r
F P
0 g
s
.75191 .253 m .79064 .23385 L .79075 .22432 L p .527 .272 .523 r
F P
0 g
s
.79064 .23385 m .75191 .253 L .75056 .26134 L p .436 .123 .377 r
F P
0 g
s
.43047 .37121 m .46997 .35764 L .47948 .36735 L p .757 .875 1 r
F P
0 g
s
.78987 .21327 m .82926 .19571 L .82744 .18417 L p .626 .454 .69 r
F P
0 g
s
.82926 .19571 m .78987 .21327 L .79075 .22432 L p .59 .38 .621 r
F P
0 g
s
.70583 .29258 m .7097 .2883 L p .70078 .29608 L .1 .1 .1 r
F P
0 g
s
.74815 .26756 m .7097 .2883 L .70583 .29258 L p .1 .1 .1 r
F P
0 g
s
.70583 .29258 m .70095 .29425 L p .71094 .28905 L .53 .958 .83 r
F P
0 g
s
.74472 .27147 m .74815 .26756 L p .7165 .28628 L .1 .1 .1 r
F P
0 g
s
.70583 .29258 m .74472 .27147 L p .7165 .28628 L .1 .1 .1 r
F P
0 g
s
.78744 .24736 m .74815 .26756 L .74472 .27147 L p .1 .1 .1 r
F P
0 g
s
.7844 .25093 m .78744 .24736 L p .75475 .26581 L .1 .1 .1 r
F P
0 g
s
.74472 .27147 m .7844 .25093 L p .75475 .26581 L .1 .1 .1 r
F P
0 g
s
.82753 .22757 m .78744 .24736 L .7844 .25093 L p .1 .1 .1 r
F P
0 g
s
.7844 .25093 m .82485 .23085 L .82753 .22757 L p .1 .1 .1 r
F P
0 g
s
.7844 .25093 m .7805 .25219 L p .81441 .23587 L .49 .935 .821 r
F P
0 g
s
.82485 .23085 m .7844 .25093 L p .81441 .23587 L .49 .935 .821 r
F P
0 g
s
.74472 .27147 m .74035 .27291 L p .77327 .25648 L .508 .946 .825 r
F P
0 g
s
.7844 .25093 m .74472 .27147 L p .77327 .25648 L .508 .946 .825 r
F P
0 g
s
.74472 .27147 m .70583 .29258 L p .71094 .28905 L .53 .958 .83 r
F P
0 g
s
.67068 .32234 m .70583 .29258 L p .70078 .29608 L .1 .1 .1 r
F P
0 g
s
.62414 .33822 m .66482 .32445 L .67068 .32234 L p .1 .1 .1 r
F P
0 g
s
.70583 .29258 m .67068 .32234 L .66482 .32445 L p .738 1 .858 r
F P
0 g
s
.66482 .32445 m .62414 .33822 L .61701 .33729 L p .58 .991 1 r
F P
0 g
s
.57661 .35099 m .61701 .33729 L .62414 .33822 L p .58 .991 1 r
F P
0 g
s
.47948 .36735 m .5193 .35375 L .5283 .36067 L p .746 .934 1 r
F P
0 g
s
.5193 .35375 m .47948 .36735 L .46997 .35764 L p .757 .875 1 r
F P
0 g
s
.61701 .33729 m .57661 .35099 L .56842 .34702 L p .703 .988 1 r
F P
0 g
s
.5283 .36067 m .56842 .34702 L .57661 .35099 L p .703 .988 1 r
F P
0 g
s
.56842 .34702 m .5283 .36067 L .5193 .35375 L p .746 .934 1 r
F P
0 g
s
.20391 .82291 m .16592 .83376 L .13403 .82146 L p .685 .962 1 r
F P
0 g
s
.7097 .2883 m .74815 .26756 L .75056 .26134 L p .261 .1 .134 r
F P
0 g
s
.79073 .08788 m .74879 .09897 L .75495 .10895 L p .726 .796 .995 r
F P
0 g
s
.75495 .10895 m .79641 .09733 L .79073 .08788 L p .726 .796 .995 r
F P
0 g
s
.34845 .79339 m .30959 .80442 L .28126 .80082 L p .577 .978 1 r
F P
0 g
s
.24236 .81193 m .28126 .80082 L .30959 .80442 L p .577 .978 1 r
F P
0 g
s
.79075 .22432 m .83021 .20607 L .82926 .19571 L p .59 .38 .621 r
F P
0 g
s
.83021 .20607 m .79075 .22432 L .79064 .23385 L p .535 .279 .525 r
F P
0 g
s
.75056 .26134 m .78953 .24161 L .79064 .23385 L p .436 .123 .377 r
F P
0 g
s
.78953 .24161 m .75056 .26134 L .74815 .26756 L p .279 .1 .14 r
F P
0 g
s
.46997 .35764 m .50993 .34391 L .5193 .35375 L p .757 .875 1 r
F P
0 g
s
.79064 .23385 m .83024 .215 L .83021 .20607 L p .535 .279 .525 r
F P
0 g
s
.83024 .215 m .79064 .23385 L .78953 .24161 L p .447 .131 .38 r
F P
0 g
s
.66482 .32445 m .70095 .29425 L .70583 .29258 L p .738 1 .858 r
F P
0 g
s
.70095 .29425 m .66482 .32445 L .6579 .32342 L p .862 1 1 r
F P
0 g
s
.61701 .33729 m .6579 .32342 L .66482 .32445 L p .58 .991 1 r
F P
0 g
s
.5193 .35375 m .5596 .33998 L .56842 .34702 L p .746 .934 1 r
F P
0 g
s
.5596 .33998 m .5193 .35375 L .50993 .34391 L p .757 .875 1 r
F P
0 g
s
.6579 .32342 m .61701 .33729 L .60902 .33321 L p .703 .988 1 r
F P
0 g
s
.56842 .34702 m .60902 .33321 L .61701 .33729 L p .703 .988 1 r
F P
0 g
s
.60902 .33321 m .56842 .34702 L .5596 .33998 L p .746 .934 1 r
F P
0 g
s
.24236 .81193 m .20391 .82291 L .17245 .81036 L p .685 .962 1 r
F P
0 g
s
.13403 .82146 m .17245 .81036 L .20391 .82291 L p .685 .962 1 r
F P
0 g
s
.38778 .78223 m .34845 .79339 L .32064 .78957 L p .577 .978 1 r
F P
0 g
s
.28126 .80082 m .32064 .78957 L .34845 .79339 L p .577 .978 1 r
F P
0 g
s
.74815 .26756 m .78744 .24736 L .78953 .24161 L p .279 .1 .14 r
F P
0 g
s
.07004 .7361 m .10092 .79979 L .13403 .82146 L p .965 .517 .193 r
F P
0 g
s
.78953 .24161 m .82934 .22224 L .83024 .215 L p .447 .131 .38 r
F P
0 g
s
.82934 .22224 m .78953 .24161 L .78744 .24736 L p .294 .1 .145 r
F P
0 g
s
.50993 .34391 m .55038 .33002 L .5596 .33998 L p .757 .875 1 r
F P
0 g
s
.70095 .29425 m .74035 .27291 L .74472 .27147 L p .53 .958 .83 r
F P
0 g
s
.74035 .27291 m .70095 .29425 L .69517 .2932 L p .731 1 1 r
F P
0 g
s
.6579 .32342 m .69517 .2932 L .70095 .29425 L p .862 1 1 r
F P
0 g
s
.5596 .33998 m .60038 .32605 L .60902 .33321 L p .746 .934 1 r
F P
0 g
s
.60038 .32605 m .5596 .33998 L .55038 .33002 L p .757 .875 1 r
F P
0 g
s
.69517 .2932 m .6579 .32342 L .65011 .31924 L p .89 1 1 r
F P
0 g
s
.60902 .33321 m .65011 .31924 L .6579 .32342 L p .703 .988 1 r
F P
0 g
s
.65011 .31924 m .60902 .33321 L .60038 .32605 L p .746 .934 1 r
F P
0 g
s
.28126 .80082 m .24236 .81193 L .21134 .79913 L p .685 .962 1 r
F P
0 g
s
.17245 .81036 m .21134 .79913 L .24236 .81193 L p .685 .962 1 r
F P
0 g
s
.32064 .78957 m .36049 .77819 L .38778 .78223 L p .577 .978 1 r
F P
0 g
s
.78744 .24736 m .82753 .22757 L .82934 .22224 L p .294 .1 .145 r
F P
0 g
s
.17245 .81036 m .13403 .82146 L .10092 .79979 L p .725 .904 1 r
F P
0 g
s
.69517 .2932 m .73516 .27183 L .74035 .27291 L p .731 1 1 r
F P
0 g
s
.73516 .27183 m .69517 .2932 L .68864 .28943 L p .801 1 1 r
F P
0 g
s
.65011 .31924 m .68864 .28943 L .69517 .2932 L p .89 1 1 r
F P
0 g
s
.55038 .33002 m .59132 .31595 L .60038 .32605 L p .757 .875 1 r
F P
0 g
s
.74035 .27291 m .7805 .25219 L .7844 .25093 L p .508 .946 .825 r
F P
0 g
s
.7805 .25219 m .74035 .27291 L .73516 .27183 L p .716 1 1 r
F P
0 g
s
.60038 .32605 m .64165 .31195 L .65011 .31924 L p .746 .934 1 r
F P
0 g
s
.64165 .31195 m .60038 .32605 L .59132 .31595 L p .757 .875 1 r
F P
0 g
s
.68864 .28943 m .65011 .31924 L .64165 .31195 L p .885 .968 1 r
F P
0 g
s
.32064 .78957 m .28126 .80082 L .2507 .78775 L p .685 .962 1 r
F P
0 g
s
.21134 .79913 m .2507 .78775 L .28126 .80082 L p .685 .962 1 r
F P
0 g
s
.68864 .28943 m .72926 .2682 L .73516 .27183 L p .801 1 1 r
F P
0 g
s
.72926 .2682 m .68864 .28943 L .68153 .283 L p .817 .954 1 r
F P
0 g
s
.64165 .31195 m .68153 .283 L .68864 .28943 L p .885 .968 1 r
F P
0 g
s
.73516 .27183 m .77581 .25107 L .7805 .25219 L p .716 1 1 r
F P
0 g
s
.77581 .25107 m .73516 .27183 L .72926 .2682 L p .791 1 1 r
F P
0 g
s
.21134 .79913 m .17245 .81036 L .13972 .78839 L p .725 .904 1 r
F P
0 g
s
.10092 .79979 m .13972 .78839 L .17245 .81036 L p .725 .904 1 r
F P
0 g
s
.59132 .31595 m .63275 .30172 L .64165 .31195 L p .757 .875 1 r
F P
0 g
s
.7805 .25219 m .82135 .23195 L .82485 .23085 L p .49 .935 .821 r
F P
0 g
s
.82135 .23195 m .7805 .25219 L .77581 .25107 L p .703 1 1 r
F P
0 g
s
.68153 .283 m .64165 .31195 L .63275 .30172 L p .872 .909 .983 r
F P
0 g
s
.36049 .77819 m .32064 .78957 L .29054 .77624 L p .685 .962 1 r
F P
0 g
s
.2507 .78775 m .29054 .77624 L .32064 .78957 L p .685 .962 1 r
F P
0 g
s
.68153 .283 m .72281 .2621 L .72926 .2682 L p .817 .954 1 r
F P
0 g
s
.72281 .2621 m .68153 .283 L .67405 .27405 L p .815 .893 1 r
F P
0 g
s
.63275 .30172 m .67405 .27405 L .68153 .283 L p .872 .909 .983 r
F P
0 g
s
.72926 .2682 m .77047 .24758 L .77581 .25107 L p .791 1 1 r
F P
0 g
s
.77047 .24758 m .72926 .2682 L .72281 .2621 L p .81 .952 1 r
F P
0 g
s
.77581 .25107 m .81712 .23081 L .82135 .23195 L p .703 1 1 r
F P
0 g
s
.81712 .23081 m .77581 .25107 L .77047 .24758 L p .783 1 1 r
F P
0 g
s
.2507 .78775 m .21134 .79913 L .17899 .77685 L p .725 .904 1 r
F P
0 g
s
.13972 .78839 m .17899 .77685 L .21134 .79913 L p .725 .904 1 r
F P
0 g
s
.29054 .77624 m .33088 .76459 L .36049 .77819 L p .685 .962 1 r
F P
0 g
s
.67405 .27405 m .716 .25367 L .72281 .2621 L p .815 .893 1 r
F P
0 g
s
.72281 .2621 m .76461 .24177 L .77047 .24758 L p .81 .952 1 r
F P
0 g
s
.76461 .24177 m .72281 .2621 L .716 .25367 L p .808 .891 1 r
F P
0 g
s
.77047 .24758 m .81227 .22743 L .81712 .23081 L p .783 1 1 r
F P
0 g
s
.81227 .22743 m .77047 .24758 L .76461 .24177 L p .803 .95 1 r
F P
0 g
s
.29054 .77624 m .2507 .78775 L .21875 .76517 L p .725 .904 1 r
F P
0 g
s
.17899 .77685 m .21875 .76517 L .2507 .78775 L p .725 .904 1 r
F P
0 g
s
.716 .25367 m .75839 .23379 L .76461 .24177 L p .808 .891 1 r
F P
0 g
s
.76461 .24177 m .80693 .22189 L .81227 .22743 L p .803 .95 1 r
F P
0 g
s
.80693 .22189 m .76461 .24177 L .75839 .23379 L p .803 .89 1 r
F P
0 g
s
.75839 .23379 m .80123 .2143 L .80693 .22189 L p .803 .89 1 r
F P
0 g
s
.33088 .76459 m .29054 .77624 L .25901 .75334 L p .725 .904 1 r
F P
0 g
s
.21875 .76517 m .25901 .75334 L .29054 .77624 L p .725 .904 1 r
F P
0 g
s
.25901 .75334 m .29977 .74136 L .33088 .76459 L p .725 .904 1 r
F P
0 g
s
.25 Mabswid
.80061 0 m
.9327 .21103 L
s
.9327 .21103 m
1 .77902 L
s
1 .77902 m
.85247 .5896 L
s
.85247 .5896 m
.80061 0 L
s
.06337 .27882 m
0 .83882 L
s
0 .83882 m
.85247 .5896 L
s
.85247 .5896 m
.80061 0 L
s
.80061 0 m
.06337 .27882 L
s
0 .83882 m
.20049 .98709 L
s
.00518 .84266 m
.00589 .83634 L
s
[(0)] .00377 .85528 .11186 -1 Mshowa
.05603 .88026 m
.05666 .87394 L
s
[(0.5)] .05477 .8929 .09932 -1 Mshowa
.10481 .91634 m
.10537 .91001 L
s
[(1)] .10371 .92899 .08748 -1 Mshowa
.15167 .95098 m
.15215 .94465 L
s
[(1.5)] .1507 .96365 .0763 -1 Mshowa
.1967 .98429 m
.19711 .97795 L
s
[(2)] .19586 .99696 .06571 -1 Mshowa
.125 Mabswid
.01552 .8503 m
.01594 .84651 L
s
.02578 .85788 m
.02618 .8541 L
s
.03594 .8654 m
.03634 .86161 L
s
.04603 .87286 m
.04641 .86907 L
s
.06595 .88759 m
.06631 .8838 L
s
.07578 .89487 m
.07614 .89107 L
s
.08554 .90208 m
.08589 .89829 L
s
.09521 .90924 m
.09556 .90544 L
s
.11434 .92338 m
.11466 .91958 L
s
.12378 .93036 m
.12409 .92656 L
s
.13315 .93729 m
.13346 .93349 L
s
.14244 .94417 m
.14274 .94037 L
s
.16081 .95775 m
.1611 .95395 L
s
.16989 .96446 m
.17016 .96066 L
s
.1789 .97112 m
.17916 .96732 L
s
.18783 .97773 m
.18809 .97393 L
s
gsave
.09784 .99201 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(@) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 11.250000 moveto
(D) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
.25 Mabswid
.9327 .21103 m
1 .77902 L
s
.93409 .2228 m
.92808 .22484 L
s
[(-1)] .94612 .21871 -1 .34012 Mshowa
.94854 .34472 m
.9425 .34667 L
s
[(0)] .96063 .34082 -1 .32249 Mshowa
.96396 .47488 m
.95789 .47673 L
s
[(1)] .97612 .47119 -1 .30374 Mshowa
.98046 .61414 m
.97435 .61587 L
s
[(2)] .99268 .61067 -1 .28377 Mshowa
.99816 .76349 m
.99202 .7651 L
s
[(3)] 1.01044 .76026 -1 .26245 Mshowa
.125 Mabswid
.93691 .24656 m
.9333 .24778 L
s
.93976 .27063 m
.93615 .27183 L
s
.94265 .29501 m
.93903 .2962 L
s
.94558 .3197 m
.94195 .32088 L
s
.95154 .37007 m
.94791 .37123 L
s
.95459 .39575 m
.95095 .3969 L
s
.95767 .42178 m
.95403 .42291 L
s
.96079 .44815 m
.95715 .44927 L
s
.96717 .50197 m
.96352 .50307 L
s
.97043 .52944 m
.96677 .53052 L
s
.97373 .55728 m
.97007 .55835 L
s
.97707 .58551 m
.97341 .58657 L
s
.9839 .64317 m
.98023 .64419 L
s
.98739 .67261 m
.98372 .67362 L
s
.99093 .70247 m
.98725 .70347 L
s
.99452 .73276 m
.99084 .73374 L
s
gsave
1.03688 .45273 -61 -10.6002 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(@) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 11.250000 moveto
(D) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{227.812, 224.812},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003S0000h1000`40n007h00Ohooo003Sool00>?o
o`00hooo003Sool00>?oo`00hooo003Sool00>?oo`00hooo003Sool00>?oo`00hooo003Sool00>?o
o`00hooo002Dool2001=ool009;oo`800003ool0000004coo`00Sooo0`000ooo0P00C?oo002<ool3
0006ool00`00ool0001;ool008[oo`8000Woo`04003ooooo001:ool008Ooo`<000_oo`04003ooooo
001:ool008Coo`<000koo`05003oooooool00019ool008;oo`8001;oo`04003ooooo0019ool007oo
o`<001Coo`05003oooooool00018ool007coo`<001Ooo`03003ooooo00;oo`03003ooooo04Goo`00
N_oo0P006_oo00<00?ooool00_oo00<00?ooool0AOoo001gool3000Lool00`00ooooo`03ool00`00
ooooo`14ool007Coo`<001ooo`03003ooooo00Coo`03003ooooo04?oo`00L_oo0P008_oo00<00?oo
ool01?oo00<00?ooool0@ooo001_ool3000Tool00`00ooooo`05ool00`00ooooo`12ool006coo`<0
02Ooo`03003ooooo00Goo`03003ooooo04;oo`00J_oo0P009Ooo0`000_oo00<00?ooool01_oo00<0
0?ooool0@Ooo001Wool3000Sool400001[IO_Wl00000ool000Woo`03003ooooo043oo`00IOoo0P00
8ooo0`0000BnGkIO_UnfG`800004_Qh00?oo0009ool00`00ooooo`10ool006;oo`<0027oo`@00005
_UnfG`00]UnnO`0200000kGn_QjeoP0200000ooo003oo`08ool00`00ooooo`0oool005ooo`<001ko
o`H000FfG`@00003_QjeoKhN00800004]K`00?oo000:ool00`00ooooo`0nool005goo`8001[oo`L0
0007_UnfGkiO]Ul00;IO_Ul0100000Feo[hN002n7[Gn00800004_Kfe_Kfm]Kd2000:ool00`00oooo
o`0nool005[oo`<001Ooo`D00009YQnMojHOWOl00;Ho]Unf?kIO00D00005]OjeoKGn]OfeoP040000
1KFl]Kfe_;Fm]K`00`002ooo00<00?ooool0?Ooo001Gool3000Hool200001lJOaWh00:GoYQnUojHO
00L00006_Qjeo[hN002n7[Gn100000Fm_KFm002e_Kfm00@00003]G`0000000coo`03003ooooo03co
o`00EOoo0P006Ooo0P0000BnO/Io_Wk6O`D000>U_P03002eo[Gn00BeoP<00005]Kbe_KFl]Kfe_004
00001KEl]E^eO;EK]G`00P003?oo00<00?ooool0??oo001Bool3000Hool300000lJOaWk6W`030000
2lHn002U_jFnYKnU_ZFoYKjU_`00_Qh0100000Be_@00]Kfm_@@00003]EbeO00000:eO0@00003]C/0
0?oo00coo`03003ooooo03_oo`00Cooo0`005ooo100000C6O`00aWnnOP800006aSjn7LHn_Qg6?P00
0jFn1@0000Fe_KFL]KfeW;Fm00<00006]GbeFkEl]E^eO;EK0`0000F]6[Dk[AZe>jdJ008000coo`03
003ooooo03_oo`00COoo0P005ooo0`0000G6O/JOaWk6WlIn00<00006_Qg6?[hMaSjn7LHn1@0000BU
OZEmYGh000:e_@@00007]GbeG000]EbeO;EL]G`00P0000J]6`00]A^e>kDK]C/4000=ool00`00oooo
o`0jool004[oo`<001Ooo`800006_Wh00;inaWnnO/Io0P0000Nn7LHn0036?[hMaSjn7@0200002;gM
_M`00:EMYGfUGJEmYEd300002JeK]Gb]FkEl[E^eO000]GbeF`0200001[DK[AZe6jdJ002]6P800005
[=Z/f@00[=T0000=ool00`00ooooo`0iool004Ooo`<001Soo`800006aYo6O/JOaWh00<In0P0000O6
?[hMaSjn7@00_Qg6?P0200001kgLaMfmg<GM002UOZEm00<00008YA`00;EL]GbeG;El]EbeO0@00005
[A^e>jdK]C^]6`0300001:cJ]=Z/f[CJ0`003Ooo00<00?ooool0>Ooo0015ool2000Hool300001Kin
aWnnO/Io_Wh00`0000Jn7LHn_Qg6?[hMaSh300001[gM_MbmgKgL_Mfmg0<00007WAbU79dLYAbM7000
[E/0100000Ne6jdJ002]6[DK[AZe6`0200001:bi002/fJcJ100000>TN000ool03Ooo00<00?ooool0
>?oo0012ool3000Fool400001<In0036O/JO100000Fn7LHn0036?[hM00@00006aMd00<GM_Mc5gKgL
0`0000NmV`00YAbU?:DLYCbU700300002JdK]A^]6kDK[A^e6jdK002]6`0200000kCJ[=ZdfP030000
1:bI[7R/VJah0P003ooo00<00?ooool0=ooo000oool3000Fool300001LIo_Wk6OkinaWl00`0000K6
?[hMaSjn7LHn_Qd300002;gL_Mfmg;gM_Mbmg@00_Md200001kek_IbmNkfL_G/009dL00<00009W;^T
f`00]A^]6[DK[AZe6jdJ00<00003[;V/fP0000800005[7T00:aiY7R/N@04000?ool00`00ooooo`0g
ool003goo`8001Ooo`800006aYl00<JOaWk6WlIn0P0000O6?[hM002n7LHn_Qg6?P0200002;gL002m
g<GM_Mc5gKgLaMd300001[fL_I^mW;fK_IbmV`<00008W=^TficKY=^Lf`00[A^e6`@00003[;X00:cJ
00800006[7V/N:ai[7P00:ah0P0000>/=j@G00000P003ooo00<00?ooool0=_oo000jool3000Gool2
00001[inaWnnO/Io0036O`800007_Qg6?[hMaSh00<Hn_Qd00P0000JmgKgL_Mfmg000_M`400001kfL
_G/00;ek_IbmNkfL00800008]CX009bkY=^L^jCKW;^Tf`<00004[=Z/^JcJ[;T300001JAh[7VTN:ai
Y7P00`000j@G0P0000>S]P00ool03_oo00<00?ooool0=_oo000hool2000Gool300001LJOaWk6WlIn
aYl00`0000K6?[hMaSjn7LHn_Qd300001KgLaMfmg<GM_M`00`0000VmVkfL_I^mW;fK_I`00;fL_I/0
0P0000JmFkdj_E^m>P00W=/400001IbJ002/fZbj[=X00P0000JTN000[7R/NJah[7T200001:@G002T
5j`g0P0000BKU@00ViD0013oo`03003ooooo03Goo`00=Ooo0`001Ooo0P003_oo100000C6O`00aWnn
OP@00005aSjn7@00_Qg6?P0400001[gL002mg;gM_Mbmg@800009_G^mW000_IbmNkfL_G^mW;ek00<0
0005]CZmFkDj_E^e>P0300001iaiW7ZLNIajW7T00:bi00800006[7VTN:aiY7P00:Ah0P001:@G0P00
00JS]Y^EXkH009<C000Aool00`00ooooo`0dool003;oo`<000Coo`@00004aWk6O`000009ool30000
1LInaYo6O/JOaWh00`0000Jn7LHn_Qg6?[hMaSh300002<GM_Mc5gKgLaMfmg000_M`200001kfL_I^m
W;fK002mVkfL00@00006_CX00;dj_E^m>[eK0P0000Ndn@00W7ZLVYajW9ZLNP0300001JAh[7VTN:ai
Y7P00`001:@G00D00:>f002S]Y^E00<00003Ta<00?oo013oo`03003ooooo00Soo`@002Soo`00<?oo
0P001?oo0`0000O6OkinaWnnO/Io_Wk6O`020006ool200001[in002nO/Io_Wk6O`800007_Qg6?P00
aSjn7LHn_Qd00P0000Rmg@00_Mfmg;gM_MbmgKgL0`0000JmNkfL_G^mW;ek_I`300002;eK]CZmFkDj
_E^e>P00]CX200001kCj]?VdnP00W7ZLNIaj00800006U1P00:Ah[7VTN:ai0P0000>Sm`00Y1L00Z@G
0P0000>KUY^E00000P000Y<C00<008Y@00004Ooo0P002ooo00<00?ooool09ooo000]ool30004ool2
00003<IoaWh00<InaWo6O/IoaWk6OlIn000KG08000;oo`800006aYo6O/JOaWh00<In0P0000O6?[hM
aSjn7@00_Qg6?P0200001[gLaMfmg<GM0035g@@00007_I^mW000_IbmVkfL_I/00P0000RmF`00_E^m
>[eK_CZmFkdj0`0000JdnKdJ]?Vm6[Ci_AX300001Y`iU1RL>I@H002TN08000BT5`<00003ViFS]Y^E
00T000koo`<00003ool00?oo00[oo`03003ooooo02Ooo`00:_oo0`001Ooo0P00012nO/Io_Wk6O`00
aWnnO/Io_Wk6OkinaWnnOP006e`KO0800005_Wk6OkinaWnnOP0300001[hMaSjn7LHn_Qg6?P<00005
_Mfmg;gM_Mbmg@0300002KfL_G^mW;ek_IbmN`00_G^mW00200002;Dj_E^e>P00]CZmFkDj_E/20000
2;Ci002dnKCj]?Vdn[Ci]?X200001I@HW1VD69`IU1P00`0000FSmj@GXoNT5j?g00800003ViJKUI^F
00@000;?7`<00003L@`00?oo00[oo`<000Coo`04003oooooool40004ool00`00ooooo`0Wool002So
o`8000Koo`80000FaWo6O/IoaWk6OlInaWl00<IoaWk6OlInaWo6O/IoaWh001]L6gd00<JOaWh30000
1KhMaSh00<Hn_Qd0100000K5g@00aMfmg<GM_M`200002KfL_I/00;fK_IbmVkfL_I^mW00300002;eK
_CZmFkdj_E^m>P00_CX200001[dJ]?Vm6[Ci002dn@@00005];T009`iU1RL>@0200000i>g002T5`03
Y1L200000i^E002KU@0300001<lOccl000000/nO0P0000=8a000ool01ooo0`001_oo0P002ooo00<0
0?ooool09ooo000Uool30004ool400004[inaWl00<Io_Wk6OkinaWnnO/Io0036OkinaWnnO/Io_Wk6
O`H00005_Qg6?[hMaSjn7@0300002;gL_Mfmg;gM_Mbmg@00_Md200001kek_IbmNkfL002mW;ek00@0
0008_E/00;eK]CZmFkDj_E^e>P<00005]?Vdn[Ci]?Zdn@0300001[BH];RdV;Bh]9Rd^0<00006TkN;
Ui>gRiL00:?g0P0000>KUY^E00000`000/lO0P0000??W`0000000/Oo0P001_oo0`002_oo00<00?oo
ool01ooo0`00:Ooo000Rool30004ool300005<IoaWk6OlInaWo6O/Io0036OlInaWo6O/IoaWk6O`00
aWo6O/IoaWh200002LHnaQd001[k002n7LHn_Qg6?P0200002;gLaMd00<GM_Mc5gKgLaMd300001[fL
_I^mW;fK_IbmV`<00008_CZmFkdj_E^m>[eK002mF`@00006]?T00;Ci_AZdnKdJ0P0000Rd^000];Rd
^KBh];Vd^;Bi0P001I>g0`0000>KUI^F00000P0000???`00ccl00P0000C?W`00003?o`D000?oo`<0
00koo`03003ooooo00Soo`03003ooooo02Ooo`008?oo0P001Ooo0P0001G6Okin002nO/Io_Wk6Okin
aWnnOP00_Wk6OkinaWnnO/Io_Wh00;inaWl00P0000Fn7LHn_Qg6?P00008Jn`0;002n7@00002mgKgL
_Mfmg000_Mbmg@0300001kfL_G/00;ek_IbmNkfL00800008]CX00;Dj_E^e>[eK]CZmF`<00008]?Zd
nKCj]?Vdn[Ci002dn@800008];RdV;Bh002d^;BH];RdV0800005]5L009>gRiNC]`0200000h<E002K
U@030004cal01@00cil00000aol010000_oo00@00?oo0000017oo`03003ooooo03;oo`007Ooo0`00
1Ooo0P0001C6O/IoaWk6OlIn0036O/IoaWk6OlInaWo6OP00aWk6OlInaWo6O/Io0`0000W6?/HMaSk6
7LHnaQg6?P006a`00P0000Jmg<GM_Mc5gKgLaMd300002;fL_I^mW;fK_I`00;fL_I/200002;eK_CZm
F`00_E^m>[eK_CX200002;dJ002m6[Ci_AZdnKdJ]?T300002;Bh];Vd^;Bi];Rd^@00];T200001:aG
]5N/EkAG0`0000>;=X<ERcH00`0000C??llOcco?7`800003eil00<oo00@000?oo`<001;oo`8003Co
o`006_oo0`001_oo0P0001?6OkinaWnnO/Io_Wk6O`00aWnnO/Io_Wk6OkinaWnnOP00_Wk6O`030000
3<Hn_Qd00;hMaSjn7LHn_Qg6?[hM000Jn`800004_Mfmg;gM_M`200002;fL002mW;ek_IbmNkfL_G/3
00002;Dj_E^e>[eK]CZmF`00_E/200001[Ci]?ZdnKCj002dnP@00008]9P00;BH];RdV;Bh]9Rd^0<0
0005]5N/=kAG[3NdE`0200000h<ERcF35@030004cal200001<nO00000<oo0`000_oo0`0000?oo`00
ool04ooo00<00?ooool0<_oo000Hool20006ool300004/InaWl00<IoaWk6OlInaWo6O/Io0036OlIn
aWo6O/IoaWk6O`<0000>aSk67LHnaQg6?P00aSk67LHnaQg6?/HMaSk67@<0000:_Mc5g@00002mVkfL
_I^mW000_I`400002;dj002m>[eK_CZmFkdj_E/300001KdJ]?Vm6[Ci_AX00`0000Rd^KBh];Vd^;Bi
];P00;Bh100000FdE`00]5N/EkAG00<00004ZoH008/fPaD300000llO003?7`0200001=NO00000<Oo
0`0000Coo`00000000Coo`03003ooooo01;oo`03003ooooo03;oo`005Ooo0`001_oo0P00016nO/Io
_Wk6OkinaWl00<Io_Wk6OkinaWnnO/Io0036Okin00@0000AaSjn7@00_Qg6?[hMaSjn7LHn0036?[hM
aSjn7LHn_Qg6?P02000021Wh00000;fL_G^mW;ek_I`300002;eK]CZmFkDj_E^e>P00]CX400001[Cj
002dn[Ci]?Zdn@800008];P00;Bh]9Rd^;BH];RdV0<00007[3NdEj`g]5N/=`00[3L00P0000B[mZ?E
ZoJSe@@000C?7`03003?W`0000P000Ooo`03003ooooo01;oo`03003ooooo03;oo`004_oo0`002?oo
01400?ji0036OlInaWo6O/IoaWh00<InaWo6O/IoaWk6OlIn00<0000EaQg6?/HMaSk67LHn0036?/HM
aSk67LHnaQg6?P00aSk67LHnaQg6?/HM008000066QT00;fL_I^mW;fK0P0000RmF`00_E^m>[eK_CZm
Fkdj0`0000RdnKdJ]?Vm6[Ci_AX00;dJ0P0000Rd^;Bi];P00;Bh];Vd^;Bi0P0000N/E`00[5NdEjaG
]5N/E`0300001J_EZoJ[eJ_fZmD00`0000??7mLo00000P0000?G_lnO00001P002_oo00<00?ooool0
4_oo0P00<ooo000@ool2000;ool03`00mYWf^@00aWnnO/Io_Wk6Okin002nO/Io_Wk6O`0200005[hM
aSh00<Hn_Qg6?[hMaSjn7LHn0036?[hMaSjn7LHn_Qg6?P00aSjn7LHn0P0000>mg0006OP00P0000Vm
W000002e>[eK]CZmF`00_E/0100000Rdn@00]?Vdn[Ci]?ZdnKCj0`0000Rd^;BH];RdV;Bh]9P00;BH
0P0000FdEj`g]5L00;AG00@00006XmD00:?EZoJSeJ_f0`000/lO0P0000?GWlnO00000`0000>oo`00
00003Ooo00<00?ooool04_oo00<00?ooool0<_oo000>ool2000<ool03P00m[Wn^OJi0036O/IoaWk6
OlInaWo6OP00aWh200005/HnaQg6?/HM00367LHnaQg6?/HMaSk67@00aQg6?/HMaSk67LHnaQd00<HM
0P0000K5gKgLaMfmg0006OP200001KeK_CZmFkdj_E/00`0000Rm6[Ci_AZdnKdJ]?T00;Ci100000Rd
^@00];Vd^;Bi];Rd^KBh0`0000B/EkAG[5NdE`<00008ZoJ[eJ_fZmF[mZ_E002[e@@00005gel00000
cioG_`0400000koo0000000?ool00`00ooooo`0Cool00`00ooooo`0aool000koo`8000coo`0<003f
^OJIm[WfV@00_Wk6OkinaWnnO/Io0P0001Jn7LHn_Qg6?[hMaSjn7@00_Qg6?[hMaSjn7LHn_Qd00;hM
aSjn7LHn_Qg6?P800008_Mc5gKgLaMfmg<GM000J6@800003]CZmFkDj00800008]?X00;Cj]?Vdn[Ci
]?Zdn@<00008]9Rd^;BH];RdV;Bh002d^0@00005[3L00:`g]5N/=`0200002:_f002[mZ?EZoJSeJ_f
XmD500001Mlngcl00000cil0100000>oo`0000004Ooo00<00?ooool04ooo00<00?ooool0<Ooo000>
ool00`00ool0000:ool02P00o[Wf^Ojim[Wn^OJi0036O/Io0`0001C6?/HM00367LHnaQg6?/HMaSk6
7@00aQg6?/HMaSk67LHn0036?/HM100000W5gKgL002mg<GM_Mc5gKgLaMd00`0000^mF`00002dnKdJ
]?T00;Ci_AZdnKdJ00800008];P00;Bh];Vd^;Bi];Rd^@<00007]5N/EkAG[5NdE`00]5L00P0000J[
eJ_fZmF[mP00ZoH400002J=C00000=mOgckOG`00ikoO_`0300000koo0000000Cool00`00ooooo`0C
ool00`00ooooo`0aool000koo`04003ooooo0008ool02@00m[WfVOJimYWf^OJIm[WfV@0200005<Hn
_Qg6?[hMaSjn7@00_Qg6?[hMaSjn7LHn_Qd00;hMaSjn7LHn_Qd300003KgLaMfmg<GM_Mc5g@00aMfm
g<GM_Mc5gKgL0080000:6=@00;Cj]?Vdn[Ci]?Zdn@00]?T200001[Bh]9Rd^;BH002dV0@00007]5L0
0;AG[3NdEj`g]5L00`0000F[mZ?EZoJSeJ_f00<00004XeBKDj=DVe<200001Mlogch00000ikl00P00
00COo`00eol001Goo`03003ooooo01;oo`8003?oo`003_oo00D00?ooooooo`0000Ooo`0L003f^Oji
m[Wn^OJio[Wf^@00mWT00<HMaSk67LHnaQg6?P00aSk67LHnaQg6?/HMaSh00<HnaQd300004KgLaMd0
0<GM_Mc5gKgLaMfmg<GM0035gKgLaMfmg<GM_M`00`0000FdnKdJ]?Vm6[Ci00<00005];Rd^KBh];Vd
^00300001jaG]5N/EkAG[5L00:aG00@00006ZoH00:_fZmF[mZ_E0P0000FSM000XgBSDj=d00800007
gckOGmln003O_nNo00000^Oo00<00=oo00005ooo00<00?ooool04_oo00<00?ooool0<Ooo000=ool0
0`00ooooo`03ool00`00ooooo`03ool06P00mYWf^OJIm[WfVOJimYWf^@00mWT00<Hn_Qg6?[hMaSjn
7LHn0036?[hMaSjn7LHn_Qd300005;gLaMfmg<GM_M`00;gLaMfmg<GM_Mc5gKgL002mg<GM_Mc5gKgL
aMd300000kCj]?T0000200001[BH002dV;Bh]9Rd^0800007[3L00:`g]5N/=kAG[3L00`0000RSeJ_f
XmF[mZ?EZoH00:_f0P0000FKDj=DVe<009]C00800009a_nno`00gch00=noikl00>Oo00@001Soo`03
003ooooo01;oo`03003ooooo037oo`003Ooo00<00?ooool01?oo01`00?ooooooo`00m[Wn^OJio[Wf
^Ojim[Wn^@00mYWfNOJI0036?/HMaSk67LHnaQg6?P00aSk67@<0000G_M`00;gLaMfmg<GM_Mc5gKgL
002mg<GM_Mc5gKgLaMfmg000_Mc5gKgLaMfmg00500001kBh];Vd^;Bi];P00;Bh00800007]5N/EkAG
002dEjaG]5L00P0000N[e@00ZmF[mZ_EZoJ[e@0400001:=dXe>SM:=C0`0000NnolLO__o77kko003O
_`0200001>Oo003Oo`006_oo00<00?ooool04_oo00<00?ooool0<Ooo000=ool00`00ooooo`05ool0
6000ooooo`00m[WfVOJimYWf^OJIm[T00?IikWSfNNih002n7LHn_Qg6?[hMaSjn7@<0000G_Mc5gKgL
aMfmg000_Mc5gKgLaMfmg<GM0035gKgLaMfmg<GM_Mc5g@00aMfmg00300002Kek00000;Bh]9Rd^;BH
];RdV00300001j`g]5N/=kAG[3L00:`g00800005ZoJSeJ_f002[mP04000019]CXe@00:=D1@0000?6
o`0000000P000kmo00<00=ooiol00`006ooo00<00?ooool04_oo0P00<_oo000=ool00`00ooooo`06
ool200004ojim[Wn^OJio[Wf^Oji003fVOIimYWfNOJImWT00<HMaSk67LHn0080000G_M`00;gLaMfm
g<GM_Mc5g@00aMfmg<GM_Mc5gKgLaMd00<GM_Mc5gKgLaMfmg00300001KfK_IbmVkfL_I/00P0000>d
^KBh];T00P0000N/E`00[5NdEjaG]5N/E`0300001:_EZoJ[eJ_f0`0000FSM:=CXgBSDj=d00@00003
__o77kko00800004agnoOlMo_gl200000kOo]ml00003000Lool00`00ooooo`0Cool00`00ooooo`0`
ool000goo`03003ooooo00Ooo`0B003fVOJimYWf^OJIm[WfV@00kWSfNNihmWW^N?Ii0036?[hM0P00
01O5gKgLaMfmg000_Mc5gKgLaMfmg<GM0035gKgLaMfmg<GM_Mc5g@00aMfmg<GM00800008_G/00;ek
_IbmNkfL_G^mW0D00005]5N/=kAG002dE`0400001J?E002SeJ_fXmD00P0000FSE000XeBKDj=D00<0
0004a_nnolKo__l20002_gl00`00_gnoO`0200000joO00000003000Nool00`00ooooo`0Cool01000
oooooooo0P00;Ooo000=ool00`00ooooo`05ool200004?Jio[Wf^Ojim[Wn^OJi003fNOJImWWfVOIi
mYWfNOJI0P0001Nmg<GM_Mc5gKgLaMd00<GM_Mc5gKgLaMfmg000_Mc5gKgLaMfmg<GM_M`00;gL0080
000A_IbmVkfL_I/00;fK_IbmVkfL_I^mW000]5N/EkAG[5NdE`0300001Z_fZmF[mZ_EZoJ[e@<00004
Xe>SM:=CXg@300001LLO__l00;koaal00P0000FoOlMo_go7Okmo00<00004UmnGo`00000Qool00`00
ooooo`0Cool01000ooooo`000_oo00<00?ooool0:_oo000=ool00`00ooooo`02ool30000:LIo_Wh0
0?JIm[WfVOJimYT00>ihmWW^N?IikWSfNNihmWT00?II0035gKgLaMfmg<GM_Mc5g@00aMfmg<GM_Mc5
gKgL002mg<GM_Mc5gKgLaMd00P00012mNkfL_G^mW;ek_I`00;fL_G^mW;ek_IbmN`00[3NdE`<00008
ZoH00:_fXmF[mZ?EZoJSe@800005XeBKDj=DVe>SE00200001Kkoa_nnolKo__l00P0000>oO`000000
100000>GgiOo00000P008ooo00<00?ooool04ooo00@00?ooool000;oo`03003ooooo02[oo`003Ooo
00<00?ooool00P0002S6WlInaYo6O/JOaWh00?Jio[Wf^@00mWWfVOIimYWfNOJImWWfV@00mUW^F000
_Mc5gKgLaMfmg<GM_M`00;gLaMfmg<GM_Mc5g@00aMfmg0@0000@_I`00;fL_I^mW;fK_IbmVkfL002m
W;fK_IbmVkfL_I/200001j_fZmF[mZ_EZoH00:_f00@00003Xe<00:=C00@00003aal0000000800003
_go7Okmo00800005]ongg`00002Go`02000Wool00`00ooooo`0Bool20002ool01000ooooo`00;?oo
000<ool00`00ooooo`02ool09P00_Wk6OkinaWnnO/Io003f^OJIm[T00?IikWSfNNihmWW^N?Ii003f
FNiHmUW^F000_Mc5gKgLaMfmg<GM_M`00;gLaMfmg<GM_M`300006[ek_IbmNkfL_G^mW000_IbmNkfL
_G^mW;ek002mNkfL_G^mW;ek_IbSeJ_fXmF[mZ?EZoH300000j=DVe>SE00400000kkoa_nno`020004
_gl200000kOO[ml00004000Zool00`00ooooo`0Cool01000ooooo`000_oo00<00?ooool0:_oo000<
ool00`00ooooo`03ool08P00aWk6WlInaYo6O/JO003n^@00mYWfNOJImWWfVOIimYT00?IIkUSfFNiH
mUW^F000_Mc5gKgLaMfmg<GM0035gKgL0`0001VmV`00_I^mW;fK_IbmVkfL_I/00;fK_IbmVkfL_I^m
W;fK002mVkfL_I^mW;fKZoJ[e@0400000j=d002SM00300001<LO__o77kko0P0000G7Okmo002oOlMo
00800003]ml0000000<002goo`03003ooooo01?oo`04003ooooo0002ool00`00ooooo`0Zool000co
o`03003ooooo00?oo`070036OkinaWnnO/Io_Wh00P0001KfNNihmWW^N?IikWSfNNih003^F?IIkUSf
FNiHmUW^F000_Mc5gKgLaMfmg0<0000H_G^mW;ek_IbmN`00_G^mW;ek_IbmNkfL002mW;ek_IbmNkfL
_G^mW000_IbmNkfL100000FKDj=DVe>SE9]C00800005a_nno`00__o6o`020005_gl400000iOo0000
000`ool00`00ooooo`0Cool01000oooooooo0P00;Ooo000<ool00`00ooooo`04ool00`00aYo6OP03
00005KhMaSh00?JImWWfVOIimYWfN@00kUSfFNiHmUW^F?IIkUSfF@00aMfmg00300006;fK002mVkfL
_I^mW;fK_I`00;fL_I^mW;fK_IbmVkfL002mW;fK_IbmVkfL_I^mW0<00007_E^m>Z=dXe>SM:=CXg@0
0P0000FnolLO__o77kko00800003agnoO`0000D00004UonGg`00000cool00`00ooooo`0Cool00`00
ooooo`0_ool000coo`03003ooooo00Goo`80000F_Qg6?[hMaSjn7LHn003fNNihmWW^N000kUSfFNiH
mUW^F?IIkUSfFNiHmUT200006[ek_IbmNkfL_G/00;ek_IbmNkfL_G^mW000_IbmNkfL_G^mW;ek002m
NkfL_G^mW;ek_I`200001[Dj_E^e>[eKVe>SE0<00003a_nno`0000<000:oO`<00004[ml00000Uol3
000fool00`00ooooo`0Bool2000aool000coo`03003ooooo00Koo`0E0036?[hMaSjn7LHn_Qd00?Ii
mYWfN@00kUSfFNiHmUW^F?IIkUSfFNiH00<0000G_I^mW;fK_IbmVkfL002mW;fK_IbmVkfL_I/00;fK
_IbmVkfL_I^mW;fK002mV`0400001;eK002mFkdj100000?77kko00000P0000BoOlMo_go7O`800003
]mngo`0000@003Woo`03003ooooo01?oo`03003ooooo02ooo`003?oo00<00?ooool01ooo01@00<Hn
_Qg6?[hMaSjn7@00kWP00>iHmUW^F?IIkUSfFNiHmUW^F0000nhh01@00;ek_IbmNkfL_G^mW000_Ibm
NkfL_G^mW;ek002mNkfL_G^mW;ek0`0000Fe>[eK]CZmFkDj00<00004__o6okkoa_l20002_gl00`00
_gnoO`08000lool00`00ooooo`0Cool00`00ooooo`0_ool000coo`03003ooooo00Ooo`07002n7LHn
_Qg6?[hMaSh00P0001o^F?IIkUSfFNiHmUW^F?IIkUP00>hhmSS^>?Hh002mW;fK_IbmVkfL_I/00;fK
_IbmVkfL_I^mW000_IbmV`0300001[dj002m>[eK_CZmF`800005__o77`00aanno`0200001LMo_go7
Okmoagl01P00@?oo00<00?ooool04ooo00<00?ooool0;ooo000<ool00`00ooooo`08ool00`00_Qg6
?P0300002KgM_M`00>iHmUW^F?IIkUSfF@020006kSP03@00_IbmNkfL_G^mW;ek002mNkfL_G^mW;ek
00<00007]CZmFkDj_E^e>P00]CX00P0000G6okkoa_nnolKo00800003_gl0000000L004?oo`03003o
oooo01?oo`03003ooooo02ooo`002ooo00<00?ooool02_oo0P0001k5gKgLaMfmg<GM003fFNiHmUW^
F?II003f>>hhmSS^>?HhkSSf>>hh002mVkfL_I^mW;fK_I`00;fL_I/300002;dj002m>[eK_CZmFkdj
_E/200000kkoaal0000300001[moagl00000]mngo`@004Koo`03003ooooo01?oo`03003ooooo02oo
o`002ooo00<00?ooool02ooo00`00;gL_Mfmg;gM_Mbmg@00mUW^F?II000:kSP01P00_G^mW;ek_Ibm
N`<0000;]CZmFkDj_E^e>P00]CZmFkDj_E^e>P0400000kmo0000000200000joO]ml000040018ool0
0`00ooooo`0Cool00`00ooooo`0_ool000_oo`03003ooooo00_oo`090035gKgLaMfmg<GM_M`00>iH
0080000>mSS^>?HhkSSf>>hhmSS^>?HhkSSf>000_IbmV`<0000@_CX00;dj_E^m>[eK_CZmF`00_E^m
>[eK_CZmF`00oY@300000kOo]ml000080018ool00`00ooooo`0Dool00`00ooooo`0^ool000_oo`03
003ooooo00coo`06002mgKgL_Mfmg;gM0P003Nhh0`0001>e>[eK]CZmFkDj002e>[eK]CZmFkDj002e
>[eK]CZmFkDj003nU00?0017ool00`00ooooo`0Dool00`00ooooo`0^ool000_oo`03003ooooo00co
o`03002mg<GM00<0000=_I`00?HhkSSf>>hhmSS^>?HhkSSf>>hhmSP00P0001C^6000_CZmFkdj_E^m
>[eK002mFkdj_E^m>[eK_CX00;dj_E^m>P000_jD0P0000?nU000o[<00P0000?ndojb00001@00Aooo
00<00?ooool04ooo0P00<?oo000;ool00`00ooooo`0=ool200001[fL_G^mW;ek_I`000S^>080000F
kQS]mnhHkOL00;Dj_E^e>[eK]CX00;Dj_E^e>[eK]CZmF`00_E^e>[eK0002oY@200000ojd003n]002
00000ojbo]8000050017ool00`00ooooo`0Dool00`00ooooo`0^ool000_oo`03003ooooo00goo`0>
002mW;fK_IbmVkfL_I/00>hhmSS^>?HhkSSf>08000K^600B002mFkdj_E^m>[eK_CX00;dj_E^m>[eK
_CX00;dj_E^m>P000_jD0P0000?nU000o[<00P0000Cnd`00003n/@800003oXd0000004Ooo`03003o
oooo01?oo`03003ooooo02koo`002ooo00<00?ooool03_oo00P00;fL_G^mW;ek_IbmN`000nhh0P00
01S^6>ggkQS]mnhHkOO^6>gg002e>[eK]CZmFkDj_E/00;eK]CZmFkDj_E^e>P00]CX200001ojD003n
U000o[@00?jd00800004o[800000o]850017ool00`00ooooo`0Cool00`00ooooo`0Lool20008ool2
0006ool000[oo`03003ooooo00ooo`09002mVkfL_I^mW;fK_I`00?Hh008000_^600<002m>[eK_CZm
Fkdj002m>[eK_CZmFkdj0`0000^dn@00oY@00?jD003nU000o[<00?jc00<00003o[40000000<004Oo
o`03003ooooo01?oo`03003ooooo01coo`03003ooooo00Soo`03003ooooo00Coo`002_oo00<00?oo
ool04?oo00H00;ek_IbmNkfL_G/200005^hHkOO^6>ggkQS]mnhHkOO^6>ggkQS]mnhH002mFkDj_E^e
>[eK]CX00;Dj0`0000jdn@00]?VlnP00oY@00?jD003n]000o[@00?jc2P00A_oo00<00?ooool04ooo
00<00?ooool03Ooo0`000_oo00<00?ooool00`001?oo00@00?ooooooo`@000?oo`03003ooooo00Co
o`002_oo00<00?ooool04?oo00@00;fL_I^mW0800003_CX00>hH00c^6006002m>[eK_CZmFkdj0`00
00FdnKdJ]?Vm6[Ci008000;nU008003nU000oY@00?jc003n/`800003o]<0000000800003oVd00000
04Koo`03003ooooo01Coo`03003ooooo00@000Woo`04003ooooo0002ool00`00ooooo`05ool01@00
oooooooo00000_oo00D00?ooooooo`0000Koo`002_oo00<00?ooool04Ooo0`0001BmFkDj_E^e>P00
kOO^6>ggkQS]mnhHkOO^6>ggkQS]mnhHkOL00;Dj0`0000Rdn@00]?Vln[Ci_?ZdnKcj0P000_jD00P0
0?jD003n]000o[@00?jc0P0000Gn/P00o[400?j?00<004Koo`03003ooooo01Coo`03003ooooo00;o
o`03003ooooo00Soo`<000?oo`<000Goo`03003ooooo00;oo`03003ooooo00?oo`03003ooooo00Co
o`002_oo00<00?ooool04Ooo00P00;dj_E^m>[eK_CZmF`003>hH0`0000^dnKdJ]?Vm6[Ci002dnKdJ
]?Vm6[Ci008000;nU007003nU000oY@00?jco[@00`0000?nd`00o]801@00A_oo00<00?ooool05?oo
00<00?ooool00_oo00<00?ooool02?oo00@00?ooool000;oo`04003ooooo0004ool00`00ooooo`03
ool00`00ooooo`02ool00`00ooooo`04ool000[oo`03003ooooo017oo`0A002mFkDj_E^e>[eK]CZm
F`00kQS]mnhHkOO^6>ggkQS]m`0300004>gG002dnKcj]?Vln[Ci_?X00;cj]?Vln[Ci_?Zdn@000_jD
00L00?jD003n]000o[Cn/`0200001?jc00000?ja0`0000?nK0000000AOoo00<00?ooool05?oo00<0
0?ooool00_oo00<00?ooool02?oo00@00?ooool000?oo`8000Goo`04003oooooool40003ool00`00
ooooo`04ool000[oo`03003ooooo01;oo`08002mFkdj_E^m>[eK_CX000G^60<0000CkMO]n>gGkOS]
e`00]?Vm6[Ci_AZdn@00]?Vm6[Ci_AZdnKdJ00000_jD00L00?jD003nU000o[?n]00200001?jc0000
0?kB0`0000?nS@000000AOoo00<00?ooool04ooo0P001?oo00<00?ooool01ooo10002ooo0P002?oo
0P001_oo000:ool00`00ooooo`0Bool02P00]CZmFkDj_E^e>[eK]CX00>gg0`0001K]enggkMO]mngG
kOO]engg002ln[Ci_?ZdnKcj]?T00;Ci_?ZdnKcj]?T000;nU007003nU000o[@00?jdo[<00P0000On
/`0000000?kB003n[`030015ool00`00ooooo`0Dool00`00ooooo`03000Zool000[oo`03003ooooo
01?oo`06002m>[eK_CZmFkdj0`0001S]enghkMO]n>gGkOS]enghkMO]n>gG002dnKdJ]?Vm6[Ci_AX0
0;dJ]?Vm6[Ci0003oY@01`00oY@00?jD003n/ojd00800005o[<00?jb003n/@0300000oi]003oo`14
ool00`00ooooo`0Dool00`00ooooo`02ool00`00ooooo`0Xool000Woo`03003ooooo01Coo`03002m
FkDj00<0000H]?T00>gGkOO]enggkMO]mngGkOO]enggkMO]mngG002dnKcj]?Vln[Ci_?X00;cj0`00
0ojD00T00?jD003n]000o[Cn/`00o[<00P0000>d^@00o]800`0000?nK@000000A?oo00<00?ooool0
5?oo00<00?ooool0;Ooo0009ool00`00ooooo`0Eool200006;Ci_AZdnKdJ]?T00>gGkOS]enghkMO]
n>gGkOS]enghkMO]n000_AZdnKdJ]?Vm6P<00003];Vd^00000;nU00:003n]?jD003nU000o[?n]000
o]@200000ojb002d^00300000oi]00000015ool00`00ooooo`0Dool00`00ooooo`0/ool000Woo`03
003ooooo01Goo`0G002dnKcj]?Vln[Ci_?X00>ggkMO]mngGkOO]enggkMO]mngGkOO]engg002lnP03
00001[Bi002d^KBh];T000;nU00:003nT`00oY?n]000o[Cn/`00o[<400001?kB00000?j?0`00AOoo
00<00?ooool04ooo0P00;_oo0009ool00`00ooooo`0Fool04P00]?Vm6[Ci_AZdn@00kMO]n>gGkOS]
enghkMO]n>gGkOS]e`@00009];Vd^;Bi];P00;Bh];Vd^00000;nU00<003n]000o[CnU000o[?n]000
o]@00?kC0`0000KndP00oZl00?j=0015ool00`00ooooo`0Dool00`00ooooo`0/ool000Woo`03003o
oooo01Koo`0>002ln[Ci_?ZdnKcj003]mngGkOO]enggkMO]m`@0000=iKK]e`00];Vd^;Bi];Rd^KBh
002d^;Bi00000_jD01000?jC003nTojd003n]?jc003n/`00o[800?jb003n/@<00003oV`00?oo04Co
o`03003ooooo01Coo`03003ooooo02coo`002Ooo00<00?ooool05_oo00X00;Ci_AZdnKdJ]?Vm6P00
kOS]e`@0000AkMO]]^gGkKK]enff002d^;Bi];Rd^KBh];T00;Bi];P00002oY@04000o[@00?jdoY@0
0?jco[@00?kD003nd`00o]<00?kB0`0000?nS@000000A?oo00<00?ooool05?oo00<00?ooool0;?oo
0009ool00`00ooooo`0Fool01P00_?ZdnKcj]?VlnP@0000DiKK]enFfkMOU]^gGiKK]enFfkMOU]P00
];Rd^KBh];Vd^000];P000?nU00@003nT`00oY?n]000o[Cn/`00o[<00?jb003n/P00o[4200001?j_
003nK@00A?oo00<00?ooool05?oo00<00?ooool0;?oo0009ool00`00ooooo`0Gool00`00_AX00002
00006;Bi003]enffkMO]]^gGkKK]enffkMO]]^gGkKK]e`00];Vd^;Bi];Rd^@00];T000?nU00@003n
]000o[CnU000o[?n]000o]@00?kC003nd`00o]8200001?j?003nK@00A?oo00<00?ooool04ooo0P00
;_oo0008ool00`00ooooo`0Hool200006KBi]9Rd^KBH003U]^gGiKK]enFfkMOU]^gGiKK]enFfkMOU
]P00];Rd^KBh];Vd^;Bi008000?nU00A003nT`00oY?n]000o[Cn/`00o[<00?jb003n/P00o[7ndP03
00000oi]003oo`13ool00`00ooooo`0Dool00`00ooooo`0/ool000Soo`03003ooooo01Soo`0K002d
^KBh];Vd^;Bi003]enffkMO]]^gGkKK]enffkMO]]^gGkKK]e`00];Vd^;Bi];Rd^KBh008000?nU00;
003n]000o[CnU000o[?n]000o]Cn/`0500001Oja00000000oVd00P00@ooo00<00?ooool05Ooo00<0
0?ooool0:ooo0008ool00`00ooooo`0Hool06P00]9Rd^KBH];VdV;Bi003]enFfkMOU]^gGiKK]enFf
kMOU]^gGiKK]e`00];Vd^;Bi];P30003oY@02@00oY<00?jCo[@00?jdo[<00002o[<01P00o[<00?kB
003ndP800004oXl00?i/0013ool00`00ooooo`0Eool00`00ooooo`0[ool000Soo`03003ooooo01So
o`0G002d^KBh];Vd^;Bi];P00>ffkMO]]^gGkKK]enffkMO]]^gGkKK]enff002d^00300000kAG0000
0003oY@04@00o[@00?jdoY@00?jco[Cn/`00o[<00?jc003n/P00o[400P0000Cn[`00oXd004?oo`03
003ooooo01Goo`03003ooooo02_oo`002?oo00<00?ooool06Ooo00D00;Bi]9Rd^KBH01400005[5Nd
MjaG]7L00003oY@05000oYCnT`00oY?n]000o[Cn/ojd003n/`00o[<00?kB003ndP00o[4200000oi/
003oo`12ool00`00ooooo`0Dool2000]ool000Soo`03003ooooo01Woo`03002d^0000080000F[5L0
0>fFkKK]U^ffkIK]]^fFkKK]U^ffkIK]]^fF002dEkAh]5NdN;AG0003oY@04`00oYCn]000o[CnU000
o[?n]?jc003n/`00o[<00?jb003n/OkB00<00003oXd00?oo04?oo`03003ooooo01Coo`03003ooooo
02_oo`002?oo00<00?ooool06Ooo0P0001V/EkAG[5NdE`00kIKUU^fFiIK]U^FFkIKUU^fFiIK]U^FF
kIH00;Ag[5NdMjaG]7L00003oY@04`00oYCnT`00oY?n]000o[Cn/ojd003n/`00o[<00?kB002/5_ja
00800003oZl0000004Coo`03003ooooo01Coo`03003ooooo02_oo`002?oo00<00?ooool06Ooo01/0
0:aG]5N/EkAG[5L00>fFkKK]U^ffkIK]]^fFkKK]U^ffkIK]]^fF002dEkAh]5NdN;AG00000ojD01<0
0?jDo[@00?jdoY@00?jco[Cn/`00o[<00?jc003n/_kC003ndP0200000oj?00000014ool00`00oooo
o`0Dool00`00ooooo`0[ool000Soo`03003ooooo01Woo`0K002dEjaG]5N/EkAG[5L00>FFkIKUU^fF
iIK]U^FFkIKUU^fFiIK]UP00]7N/EkAg[5NdM`0000?nU00>003nU?jC003nTojd003n]?jco[@00?jc
003n/`800003o[800?ja00800003oZl0000004Coo`03003ooooo01Coo`03003ooooo02_oo`001ooo
00<00?ooool06_oo01/00:aG]5N/EkAG[5NdE`00kKK]U^ffkIK]]^fFkKK]U^ffkIK]]^fFkKH00;Ah
]5NdN;AG00000ojD01D00?jDo[@00?jdoY@00?jco[Cn/`00o[<00?jco]<00?kC003ndP00o]400`00
A?oo00<00?ooool05?oo0P00;?oo0007ool00`00ooooo`0Kool06@00[5NdEjaG]5N/E`00iIK]U^FF
kIKUU^fFiIK]U^FFkIKUU^fFiIH00:aG]7N/E`0000CnU00D003nU?jC003nTojd003n]?jco[@00?jc
003n/ojb003n/P00o[7ndP800003oXl00?oo04;oo`8001Ooo`03003ooooo02[oo`001ooo00<00?oo
ool06ooo00D00;AG[5NdEjaG00D0000?kIK]]^fFkKK]U^ffkIK]]^fFkKH00;Ah]5NdN00000CnU00D
003nU?jd003n]?jD003n/ojdo[<00?jco]@00?kC003nd`00o];n/@800003oZl00?oo03koo`@000;o
o`03003ooooo01Coo`03003ooooo00@002Koo`001ooo00<00?ooool06ooo00<00:aG00000P0000F[
m@00iGG]MNEe00L00005iIK]U^FF002/E`030004oY@02`00oYCnT`00oY?n]?jC003n/ojd00000_jc
00L00?jb003n/P00o[7ndP0200000oj?003oo`0kool30007ool00`00ooooo`0Cool01000ooooo`00
:Ooo0007ool00`00ooooo`0Kool20004ZoH02`00kIK]MNfFkGG]U^eekIK]MNfFkGD01@0000>[mZ`F
00001?jD01H00?jDo[@00?jdoYCn]000o[Cn/`00o[?ne000o]<00?kCo[800?ja003n/08003Soo`@0
00[oo`03003ooooo01?oo`05003oooooool0000Xool000Ooo`03003ooooo01_oo`0I002[mJ_fZoF[
mZ_e003UMNeeiGG]MNEekGGUMNeeiGG]MNEekGGUM@00ZoF/5Z_e00001?jD00/00?jDoY<00?jCo[Cn
T`00o[?n]00000;n/`08003n/P00o[;ndP00o];n/@<003Goo`<000ooo`03003ooooo01;oo`03003o
oooo00;oo`03003ooooo02Goo`001ooo00<00?ooool06ooo00<00:_fZoH00j_f01<00>fFkGG]U^ee
kIK]MNfFkGG]U^eekIK]MNfF002/5Z_f[1H00003oY@05P00o[CnU?jd003n]?jDo[@00?jdo[<00?jc
o]@00?kCo[<00?jb003n/OkB0`00<Ooo10004ooo00<00?ooool04?oo0P001Ooo00<00?ooool09?oo
0007ool00`00ooooo`0Kool06000ZoF[mZ_eZoJ[m@00iGG]MNEekGGUMNeeiGG]MNEekGGUMNeeiGD0
0:_e[1H000CnU00<003nTojDoY<00?jCo[CnT`00o[?n]0000_jc00P00?jbo[<00?kB003nd_ja0`00
;_oo0`006?oo00<00?ooool04?oo00@00?ooool000;oo`03003ooooo02Coo`001_oo00<00?ooool0
7?oo00<00:_fZoH00Z_f0`00013]MNfFkGG]U^eekIK]MNfFkGG]U^eekIH00:`FZoH000CnU00I003n
]?jDo[@00?jdoYCn]000o[Cn/`00o[?ne000o]?n/`00o[800?jao]800?kA0000:_oo10007?oo00<0
0?ooool04?oo00<00?ooool00P009ooo0006ool00`00ooooo`0Lool00`00ZoD0000200000j=c003U
E00400003>eeiGG]MNEekGGUMNeeiGD00:_e[1H000CnU00<003nTojDoY<00?jCo[CnT`00o[?n]000
0_jc00/00?jbo[<00?kB003nd_ja003n/00002Ooo`<0027oo`03003ooooo00ooo`03003ooooo02Wo
o`001_oo00<00?ooool07?oo0P0000ZSM:^DXgB[U000kGG]E>eekEC]M@<00009kIK]MNfFkGG]UP00
[1J[mP0000CnU00G003n]?jDo[@00?jdoYCn]000o[Cn/`00o[?ne000o]?n/`00o[;nd`00o];n/@02
000Tool3000Tool00`00ooooo`0?ool00`00ooooo`0Yool000Koo`03003ooooo01coo`0?002SLj=d
Xg>SM:=c003UE>eEiEC]ENEDkEGUE>eE00@00003iGD00:_e008000CnU00=003nTojDoY<00?jCo[Cn
T`00o[?n]?jc00000_jc00P00?jc003nd_jb003n/OkB0P008?oo1000:?oo00<00?ooool03Ooo0P00
:ooo0006ool00`00ooooo`0Lool04`00ZiBSM:^DXgB[U000kGG]E>eekEC]MNeDkGG]E>eekEC]MNeD
00<00003Xi@00?jD00?nU00G003n]?jDo[@00?jdoYCn]000o[Cn/ojd003ne?jc003n/`00o[;nd`00
o];n/@02000Mool3000]ool00`00ooooo`0=ool00`00ooooo`0Yool000Koo`03003ooooo01coo`0H
002SLj=dXg>SM:=c003UE>eEiEC]ENEDkEGUE>eEiEC]ENEDkEGUE000XiB[U0001?jD00d00?jCoYCn
T`00oY?n]?jC003n/ojdo[<00002o[<02000o[?n/P00o[800?jao]82000Iool4000aool00`00oooo
o`0<ool00`00ooooo`0Yool000Koo`03003ooooo01coo`05002[U:=dZiBSM00300003neDkGG]E>ee
kEC]MNeDkGG]E>eekEC]M@00Zi@00005oY@05`00o[CnU?jd003n]?jDo[@00?jdo[?n]000o]Cn/`00
o[?nd`00o]<00?kBo[400P005_oo0`00=_oo00<00?ooool02ooo00<00?ooool0:Ooo0006ool00`00
ooooo`0Lool00`00Xg<0000200000i[A003U<`0200003NeEiEC]ENEDkEGUE>eEiEC]ENED002SU000
00GnU00=003nTojDoY<00?jCo[CnT`00o[?n]?jc00000_jc00X00?jco[800?jbo]800?kBo[4001;o
o`@003[oo`03003ooooo00[oo`03003ooooo02Woo`001_oo00<00?ooool07?oo0P0000RJdJ;bV]6R
lP00kEC]<neD0P0000_]E>eekEC]MNeDkGG]E>ee002[U00000GnU00I003n]?jDo[@00?jdoYCn]000
o[Cn/ojd003ne?jc003n/okC003ndojb003n/OkB00003ooo0`00?_oo00<00?ooool02_oo0P00:_oo
0005ool00`00ooooo`0Mool03000V]6JlY[AV_800>ddiC?]=>DckCCU<`<00008iEC]ENEDkEGUE000
Xi@000CnU00>003nU?jCoYCnT`00oY?n]?jC003n/ojdo[<000;n/`0:003n/ojb003n/_kB003nd_ja
000;ool40012ool00`00ooooo`0:ool00`00ooooo`0Xool000Goo`03003ooooo01goo`0?002RlY[A
X_:Jd@00kC?]E>dckEC]<neDkC?]E>dc00800006kGG]E>ee002[U0001?jD01X00?jDo[CnU?jd003n
]?jDo[@00?jdo[?n]000o]Cn/`00o[?nd`00o]?n/P00o[7ndP002?oo0`00Aooo00<00?ooool02Ooo
00<00?ooool0:?oo0005ool00`00ooooo`0Lool01@00V_:JdI[bV]400P0000_]=>DckCCU<nddiC?]
=>DckCCU<ndd00800004iE@00:>D0004oY@03P00oYCnTojD003n]?jCo[CnT`00o[?n]?jc0002o[<0
2@00o[?n/P00o[;ndP00o]800005ool4001;ool00`00ooooo`08ool00`00ooooo`0Xool000Goo`03
003ooooo01coo`03002Jd@000080000@RPh00000003]<neDkC?]E>dckEC]<neDkC?]E>dckE@20005
oY@06@00oYCn]?jD003nU?jdoYCn]000o[Cn/ojd003ne?jco]@00?kCo[<00?jb003n/@0000;oo`<0
053oo`03003ooooo00Ooo`03003ooooo02Soo`001Ooo00<00?ooool07?oo0P0000N:3Y8>RPh00>dB
kA?]4P0200002^ddiC?]=>DckCCU<nddiC?]=>Dc0P001OjD00h00?jDoY?nU000o[CnTojdoY<00?jc
o[Cn/`000ojc00H00?jbo[<00?kBo[84001Cool00`00ooooo`07ool00`00ooooo`0Xool000Goo`03
003ooooo01coo`0;002:3Y8>RPjB3P00kA?]4^dCkA;]4`0200002>dckEC]<neDkC?]E>dckE@20005
oY@05000oYCn]?jD003nU?jdoYCn]000o[Cn/ojd003ne?jco]@00?kCo[<4001Gool00`00ooooo`05
ool2000Zool000Goo`03003ooooo01_oo`0F002:3Y8>RPjB3P00kA?]4^dCkA;]4ndBkA?]4P00iC?]
=>DckCCU<nddiC<20005oY@03P00oYCnTojD003n]?jCo[CnT`00o[?n]?jc0002o[<500000okBo[80
001Hool00`00ooooo`05ool00`00ooooo`0Xool000Goo`03003ooooo01_oo`04002B3XX>TPh20000
2NdBkA?]4^dCkA;]4ndBkA?]4P0200001neDkC?]E>dc002JlP0000GnU00=003nU?jdoY@00?jDo[Cn
U?jd003n]?jco[@00`0000Sne000o]?n/`00o[;nd`00FOoo00<00?ooool01Ooo00<00?ooool09ooo
0004ool00`00ooooo`0Kool00`00TPh0000200003GSY00000000kA?]4^dCkA;]4ndBkA?]4^dC0080
0005iC?]=000Xa800005oY@02@00oYCnTojD003n]?jCo[CnT`050003o[<01P00o[;n/`00o]8005_o
o`03003ooooo00Coo`03003ooooo02Ooo`001?oo00<00?ooool06ooo0P00019hjGS9N>T00>cbk?7/
lP00kA?]4^dCkA;]4ndBkA?]4^dCkA840006oY@01P00oYCn]?jD003nU0@0000;o[Cn/ojd003ne?jc
o]@00?kCo[?nd`02001Lool00`00ooooo`03ool00`00ooooo`0Wool000Coo`03003ooooo01_oo`06
001hjGS9N>Uhb@001>ca0P0000[]4^dCkA;]4ndBkA?]4^dCkA;]4`8000KnU003003nU?jC00<00009
oY?n]?jC003n/ojdo[?n]00000;n/`04003n/_jco[82001Lool00`00ooooo`02ool2000Yool000Co
o`03003ooooo00Ooo`80017oo`04001hjGS9N>T200004Ncak?;/lNcbk?7/l^ca003]4^dCkA;]4ndB
kA?]4^dCkA800P001OjD100001?nU000oYCn]?jDo[@00?jdo[?n]?jc003n/okD003ndojco]<0001N
ool00`00ooooo`02ool00`00ooooo`0Wool000Coo`03003ooooo00?oo`@00004]ano?`00000?ool3
00001533003/d0001nca0P0000K]4ndBkA?]4^dCkA830002oY@400003ojCoYCnTojD003n]?jCo[Cn
T`00o[?n]?jco[@00003o[<01000o[?n/P00Gooo00D00?ooooooo`0002Woo`001?oo00@00?oooooo
o`8000Jo?`04003d1P00000<ool200001USSD<<00>c@m=3/d0800007k?;/lNcbk?7/l^cak?800P00
00?]4P0000001P000ojD01H00?jdoYCn]?jD003nU?jdoYCn]000o[Cn/ojdo[<00?jco]Cn/`00o[<0
067oo`04003ooooo0003ool2000Tool000Coo`03003ooooo00;oo`800009_cng7klo]ano?kLO003d
1_@V008000Soo`H000K/d003003/lNca00?/l@@00009]cl00000002g?klo]cno?`0000;nU00@003n
TojDoY?nU000o[CnTojdoY<00?jco[Cn/ojd0003o[<00`00o[<0001Rool00`00ool00002ool01000
ooooo`008ooo0004ool00`00ooooo`04ool20002_cl300000o@Vm0Kd9P050002ool600002OB_k:h0
0>c@m=3/d?C@k=3dd00500000klo0000000200002;Lo_cng?klo]cno?kLo_cl300004_jdoYCn]?jD
003nU?jdoYCn]000o[Cn/ojdo[<00?jco]Cn/`8006Coo`03003oo`0000Coo`03003ooooo027oo`00
1?oo00<00?ooool01_oo0P0000>g_jnO]kl00P0000Gd1P00m2SdB?@X00D00008k6cdSNa/003/[_B^
k:h000?/d0<00003]cno?`0000800003_cng?klo00800007]cno?kLo_cng?klo]cl0100000knU?jC
oY@00?jdoY?n]?jC003n/ojdo[?n]0000_jc0P00IOoo0`000_oo0P009?oo0003ool00`00ooooo`09
ool400004_@6m2Kd1P00m2SdB@00m6_dJ_A[003dSOB<m8gdS000k:kd[`@00003]cl000000080000;
_cng?klo]cno?kLo_cng?`00]cno?`0400001:no]kn__kNo0P0000_nU000oYCn]?jDo[@00?jdo[?n
]?jc00<006Soo`8000Coo`03003ooooo027oo`000ooo00<00?ooool03Ooo00l00?@6m2H00?A8m2Sd
B000m6_dB_A[003dSNa/m8d00`0000>o?kLo00000P0000bg?klo]cl00;Lo_cng?klo]cno?kLo_cl3
00002Zno[in__jnO[kn_Wjno[in__jnO100000Gn]?jC003n/ojd00<006[oo`800005ool00?ooool0
000Sool000?oo`03003ooooo00goo`0;003d9_@6m2H00?A9m2SdB@00m6_dJP0400000klo00000002
00001kLo_cng?klo]cno?kLo00800003_cng?`0000<00003]kn__kNo00<00003]kn__kNo00@000BG
g`H006Soo`@000Coo`8002Coo`000ooo00<00?ooool03_oo00L00?@Vm0H00?@Xm4Sd:00300000kLo
_cl0000200003;lo]cno?`00_cng?klo]cno?kLo_cng?`<0000:[in__jnO[kn_Wjno[in__jnO[kl4
000019NoUmnG_iOO1@00J?oo1000;_oo0003ool00`00ooooo`0>ool00`00m0Kd9P0400000kLo0000
000200001klo]cno?kLo_cng?klo00800003]cno?`0000<00003[kng_jno00<00003[kng_jno00@0
00BGg`H006Ooo`@003;oo`000ooo00<00?ooool03ooo0P0000>o?kLo00000P0000bg?klo]cl00;Lo
_cng?klo]cno?kLo_cl300002Zno[in__jnO[kn_Wjno[in__jnO100000BGgiNoUmnG_`D006Soo`<0
03Koo`000ooo00<00?ooool04?oo0`0000Ng?klo]cno?kLo_cng?`0200000klo]cl0000300000kNo
[kng_`0300000kNo[kng_`040004Uml6001Wool4000iool000?oo`03003ooooo00coo`@000;oo`09
002o?kLo_cng?klo]cno?kLo00<0000:[in__jnO[kn_Wjno[in__jnO[kl4000019NoUmnG_iOO1@00
Iooo1000?Ooo0002ool00`00ooooo`0:ool30007ool200000kLo_cl0000300000jno]kn__`030000
0jno]kn__`040004Uml6001Vool40011ool000;oo`03003ooooo00Koo`@000coo`80000:[kn_Wjno
[in__jnO[kn_Wjno[il4000019OOUknGgiNo1@00Hooo00<00?ooool01000AOoo0002ool00`00oooo
o`03ool3000Bool300000kNo[kng_`040004Uml6001Vool30019ool000;oo`03003oo`0000<001So
o`@00004UknGgiNoUml5001Vool4001<ool000;oo`80023oo`H006Goo`@0053oo`000ooo00<00?oo
ool08ooo00<00?ooool0Gooo1000E?oo0004ool2000Sool00`00ooooo`1Lool3001Hool000Goo`<0
027oo`03003ooooo05Soo`@0013oo`03003ooooo04Soo`001ooo00<00?ooool07_oo00<00?ooool0
EOoo10005?oo00<00?ooool0B?oo0008ool2000Nool00`00ooooo`1?ool00`00ool00003000Eool5
0019ool000;oo`8000Koo`8001coo`03003ooooo04koo`<001Woo`04003ooooo001:ool00005ool0
0?ooool00006ool00`00ool0000Jool00`00ooooo`1:ool4000Mool00`00ool0001:ool00005ool0
0?ooool00007ool2000Jool00`00ooooo`16ool4000Rool2001:ool00005ool00?ooool00009ool0
0`00ooooo`0Gool00`00ooooo`11ool5000Wool00`00ooooo`18ool00005ool00?ooool0000:ool2
000Gool00`00ooooo`0nool4001fool00005ool00?ooool0000;ool3000Eool00`00ooooo`0kool3
001jool000;oo`8000koo`03003ooooo01?oo`03003ooooo03?oo`04003oooooool4001mool000Co
o`8000Coo`03003ooooo00;oo`<00004ool00000000Bool00`00ooooo`0cool40021ool000?oo`04
003ooooo000;ool01000ool000004_oo00<00?ooool0;ooo1000QOoo0003ool01000ooooo`002ooo
00<00?ooool00`004?oo00<00?ooool0;?oo0`00ROoo0003ool01000ooooo`002?oo0`001Ooo00<0
0?oo00003_oo00<00?ooool09_oo00<00?oo00000`00S?oo0003ool01000ooooo`002?oo00<00?oo
ool01_oo0P003_oo00<00?ooool09?oo1000T?oo0003ool01000ooooo`002?oo00<00?ooool02?oo
0P003?oo00<00?ooool08?oo10003Ooo1000Pooo0004ool20009ool40008ool2000;ool00`00oooo
o`0Iool01000oooooooo0`004Ooo00<00?ooool0Q?oo000Cool40005ool30008ool00`00ooooo`0J
ool4000Eool00`00ooooo`23ool001Goo`03003ooooo00Koo`03003ooooo00Koo`03003ooooo01Ko
o`@001[oo`03003ooooo08;oo`005Ooo00<00?ooool01ooo0P001_oo00<00?ooool04_oo10007ooo
00<00?ooool0POoo000Eool00`00ooooo`08ool20005ool00`00ooooo`0=ool5000Pool01000oooo
o`00Pooo000Eool00`00ooooo`0:ool20003ool00`00ooooo`0;ool3000Uool20024ool001?oo`<0
00goo`05003oo`00ool00009ool4002^ool001Goo`@000?oo`03003ooooo00;oo`L000;oo`03003o
oooo00@00;;oo`005ooo00<00?ooool02_oo00D00?oo00000?oo00@00;Koo`005ooo00<00?ooool0
2_oo00<00?ooool00P00^_oo000Gool00`00ooooo`07ool30016ool00`00ool0001fool001Ooo`03
003ooooo00Ooo`03003ooooo04Koo`03003oo`0007Koo`005Ooo0`002Ooo00<00?ooool0AOoo00D0
0?oo003oo`0007Goo`004ooo0P000_oo00<00?ooool00ooo0P000_oo1`0000?oo`000000?_oo00D0
0?oo003oo`0007Goo`004ooo00<00?ooool02?oo00<00?ooool00ooo00D00?ooooooo`0000;oo`03
003ooooo03_oo`04003oo`00ool2001dool000Coo`D00005ool00?oo003oo`020003ool01000oooo
oooo10000ooo00<00?ooool01?oo00@00?ooool000;oo`03003ooooo0;Goo`001_oo00T00?oooooo
o`00ool00?oo00001?oo00D00?ooooooo`0000;oo`05003oooooool00007ool00`00ool00002ool0
0`00ooooo`2eool000Koo`09003oooooool00?oo003oo`0000Coo`03003ooooo00;oo`03003ooooo
00?oo`03003ooooo00Koo`8000;oo`03003ooooo0;Goo`001_oo00T00?ooooooo`00ool00?oo0000
1?oo00<00?ooool00ooo00<00?ooool00_oo00<00?ooool00ooo00<00?ooool00P000_oo00<00?oo
ool0]Ooo0006ool01000oooooooo0P0000?oo`00ool01?oo00@00?ooooooo`@000?oo`03003ooooo
00Coo`8000;oo`800;Soo`001?oo1@002_oo0P002?oo0P00a?oo003Sool00>?oo`00hooo003Sool0
0>?oo`00hooo003Sool00>?oo`00hooo003Sool00>?oo`00hooo003Sool00>?oo`00hooo0000\
\>"],
  ImageRangeCache->{{{0, 226.812}, {223.812, 0}} -> {0.0016671, -0.0848322, 
  0.00536599, 0.00536599}}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5.3 Optimization with OPTDesign ", "Section"],

Cell[CellGroupData[{

Cell["\<\
OPTDesign[p,Radius->r0,Center->k0 (* GridPoints -> wpts *)];
\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \("\nParameterization for internal stability:"\)], "Print"],

Cell[BoxData[
    \("  T = A + B * T1, where T1 is RHP stable and"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("A[s]="\[InvisibleSpace]\(\(-5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2\),
      SequenceForm[ "A[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("B[s]="\[InvisibleSpace]\(\(-5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2\),
      SequenceForm[ "B[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("  "\)], "Print"],

Cell[BoxData[
    \("Processing performance function..."\)], "Print"],

Cell[BoxData[
    \("Sampling Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Smoothing Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Optimization routine Anopt now number-crunching ..."\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 2.2485924730083E+00 :   N/A   : 8.9E-01 : 1. E+00 :   N/A   : NON \
: 128"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.5972070924256E-01 : 1.8E+00 : 3.6E-03 : 0  E+00 : 1.3E-02 : NON \
: 128"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.597207092425609E-01"\)], "Print"],

Cell[BoxData[
    \("flat      = 3.5616725651E-03"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 1.32E-02"\)], "Print"],

Cell[BoxData[
    \("Calculating output functions ..."\)], "Print"],

Cell[BoxData[
    \("Resetting options ..."\)], "Print"],

Cell[BoxData[
    \("Done! "\)], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5.4 Producing a rational compensator", "Section"],

Cell[CellGroupData[{

Cell["\<\
Clear[Trat] ;
Trat[s_] = RationalModel[s,DegreeOfDenominator -> 3]\
\>", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]0.171724347370698593`\),
      SequenceForm[ "Error = ", 0.17172434737069858],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(-5.`\) + s\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2 + 
      \(\((\(1.9619799915290466`\[InvisibleSpace]\) + 
              0.805659853329512998`\ s)\)\ 
          \((\(-5.`\) + s)\)\)\/\(\((
            \(-1.58062453779134459`\) - 0.419375462208655491`\ s)\)\ 
          \((\(2.`\[InvisibleSpace]\) + s)\)\^2\)\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simplifying the compensator", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
comp1[s_] =   1/p[s]  Trat[s]/(1-Trat[s]) ; 
PlotZP[comp1[s],s,Frame -> True];\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.569044 0.0783556 0.5 0.454545 [
[.09891 -0.0125 -6 -9 ]
[.09891 -0.0125 6 0 ]
[.25562 -0.0125 -6 -9 ]
[.25562 -0.0125 6 0 ]
[.41233 -0.0125 -6 -9 ]
[.41233 -0.0125 6 0 ]
[.56904 -0.0125 -3 -9 ]
[.56904 -0.0125 3 0 ]
[.72576 -0.0125 -3 -9 ]
[.72576 -0.0125 3 0 ]
[.88247 -0.0125 -3 -9 ]
[.88247 -0.0125 3 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .04545 -12 -4.5 ]
[-0.0125 .04545 0 4.5 ]
[-0.0125 .27273 -24 -4.5 ]
[-0.0125 .27273 0 4.5 ]
[-0.0125 .5 -6 -4.5 ]
[-0.0125 .5 0 4.5 ]
[-0.0125 .72727 -18 -4.5 ]
[-0.0125 .72727 0 4.5 ]
[-0.0125 .95455 -6 -4.5 ]
[-0.0125 .95455 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.09891 0 m
.09891 .00625 L
s
[(-6)] .09891 -0.0125 0 1 Mshowa
.25562 0 m
.25562 .00625 L
s
[(-4)] .25562 -0.0125 0 1 Mshowa
.41233 0 m
.41233 .00625 L
s
[(-2)] .41233 -0.0125 0 1 Mshowa
.56904 0 m
.56904 .00625 L
s
[(0)] .56904 -0.0125 0 1 Mshowa
.72576 0 m
.72576 .00625 L
s
[(2)] .72576 -0.0125 0 1 Mshowa
.88247 0 m
.88247 .00625 L
s
[(4)] .88247 -0.0125 0 1 Mshowa
.125 Mabswid
.13809 0 m
.13809 .00375 L
s
.17727 0 m
.17727 .00375 L
s
.21644 0 m
.21644 .00375 L
s
.2948 0 m
.2948 .00375 L
s
.33398 0 m
.33398 .00375 L
s
.37316 0 m
.37316 .00375 L
s
.45151 0 m
.45151 .00375 L
s
.49069 0 m
.49069 .00375 L
s
.52987 0 m
.52987 .00375 L
s
.60822 0 m
.60822 .00375 L
s
.6474 0 m
.6474 .00375 L
s
.68658 0 m
.68658 .00375 L
s
.76493 0 m
.76493 .00375 L
s
.80411 0 m
.80411 .00375 L
s
.84329 0 m
.84329 .00375 L
s
.05973 0 m
.05973 .00375 L
s
.02056 0 m
.02056 .00375 L
s
.92164 0 m
.92164 .00375 L
s
.96082 0 m
.96082 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .04545 m
.00625 .04545 L
s
[(-1)] -0.0125 .04545 1 0 Mshowa
0 .27273 m
.00625 .27273 L
s
[(-0.5)] -0.0125 .27273 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(0)] -0.0125 .5 1 0 Mshowa
0 .72727 m
.00625 .72727 L
s
[(0.5)] -0.0125 .72727 1 0 Mshowa
0 .95455 m
.00625 .95455 L
s
[(1)] -0.0125 .95455 1 0 Mshowa
.125 Mabswid
0 .09091 m
.00375 .09091 L
s
0 .13636 m
.00375 .13636 L
s
0 .18182 m
.00375 .18182 L
s
0 .22727 m
.00375 .22727 L
s
0 .31818 m
.00375 .31818 L
s
0 .36364 m
.00375 .36364 L
s
0 .40909 m
.00375 .40909 L
s
0 .45455 m
.00375 .45455 L
s
0 .54545 m
.00375 .54545 L
s
0 .59091 m
.00375 .59091 L
s
0 .63636 m
.00375 .63636 L
s
0 .68182 m
.00375 .68182 L
s
0 .77273 m
.00375 .77273 L
s
0 .81818 m
.00375 .81818 L
s
0 .86364 m
.00375 .86364 L
s
0 .90909 m
.00375 .90909 L
s
0 1 m
.00375 1 L
s
.25 Mabswid
0 0 m
0 1 L
s
.09891 .99375 m
.09891 1 L
s
.25562 .99375 m
.25562 1 L
s
.41233 .99375 m
.41233 1 L
s
.56904 .99375 m
.56904 1 L
s
.72576 .99375 m
.72576 1 L
s
.88247 .99375 m
.88247 1 L
s
.125 Mabswid
.13809 .99625 m
.13809 1 L
s
.17727 .99625 m
.17727 1 L
s
.21644 .99625 m
.21644 1 L
s
.2948 .99625 m
.2948 1 L
s
.33398 .99625 m
.33398 1 L
s
.37316 .99625 m
.37316 1 L
s
.45151 .99625 m
.45151 1 L
s
.49069 .99625 m
.49069 1 L
s
.52987 .99625 m
.52987 1 L
s
.60822 .99625 m
.60822 1 L
s
.6474 .99625 m
.6474 1 L
s
.68658 .99625 m
.68658 1 L
s
.76493 .99625 m
.76493 1 L
s
.80411 .99625 m
.80411 1 L
s
.84329 .99625 m
.84329 1 L
s
.05973 .99625 m
.05973 1 L
s
.02056 .99625 m
.02056 1 L
s
.92164 .99625 m
.92164 1 L
s
.96082 .99625 m
.96082 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .04545 m
1 .04545 L
s
.99375 .27273 m
1 .27273 L
s
.99375 .5 m
1 .5 L
s
.99375 .72727 m
1 .72727 L
s
.99375 .95455 m
1 .95455 L
s
.125 Mabswid
.99625 .09091 m
1 .09091 L
s
.99625 .13636 m
1 .13636 L
s
.99625 .18182 m
1 .18182 L
s
.99625 .22727 m
1 .22727 L
s
.99625 .31818 m
1 .31818 L
s
.99625 .36364 m
1 .36364 L
s
.99625 .40909 m
1 .40909 L
s
.99625 .45455 m
1 .45455 L
s
.99625 .54545 m
1 .54545 L
s
.99625 .59091 m
1 .59091 L
s
.99625 .63636 m
1 .63636 L
s
.99625 .68182 m
1 .68182 L
s
.99625 .77273 m
1 .77273 L
s
.99625 .81818 m
1 .81818 L
s
.99625 .86364 m
1 .86364 L
s
.99625 .90909 m
1 .90909 L
s
.99625 0 m
1 0 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 .5 m
1 .5 L
s
.56904 0 m
.56904 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
.41233 .5 .02 0 365.73 arc
s
newpath
.41233 .5 .02 0 365.73 arc
s
newpath
.49169 .5 .02 0 365.73 arc
s
newpath
.96082 .5 .02 0 365.73 arc
s
.03355 .48182 m
.06991 .51818 L
s
.03355 .51818 m
.06991 .48182 L
s
.46906 .15883 m
.50542 .19519 L
s
.46906 .19519 m
.50542 .15883 L
s
.46906 .80481 m
.50542 .84117 L
s
.46906 .84117 m
.50542 .80481 L
s
.94264 .48182 m
.979 .51818 L
s
.94264 .51818 m
.979 .48182 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000181000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00
oooo8Ooo003ooolQool00?ooob7oo`00=_oo0P00:?oo00<00?ooool09?oo10008_oo0P009_oo1000
9ooo00<00?ooool07Ooo000eool01000ooooo`009ooo00<00?ooool09?oo00<00?ooool08_oo00@0
0?ooool002Goo`03003ooooo02Soo`03003ooooo01goo`00=Ooo00@00?ooool002Coo`D002Koo`03
003ooooo027oo`04003ooooo000Vool00`00ooooo`0Tool5000Nool002ooo`@000;oo`<0023oo`@0
0005ool00?ooool0000Pool40004ool00`00ooooo`0Pool01000ooooo`009ooo00<00?ooool08ooo
00@00?ooool001ooo`00=Ooo00<00?ooool09_oo00<00?oo0000:Ooo00<00?ooool07ooo00@00?oo
ool002Soo`03003ooooo02?oo`03003oo`0001ooo`00=Ooo00<00?ooool09ooo0P009_oo00@00?oo
ool0027oo`04003ooooo000Uool01000ooooo`009_oo0P007ooo000fool3000Wool00`00ooooo`0U
ool2000Sool2000Wool2000Xool00`00ooooo`0Mool00?ooob7oo`006oooo`001@000Ooo000Kool0
0`00ooooo`02ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`07ool0
0`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`08ool0
0`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool0
0`00ooooo`08ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool0
0`00ooooo`08ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool0
0`00ooooo`07ool10001ool001_oo`03003ooooo01Koo`03003ooooo02Koo`03003ooooo02Koo`03
003ooooo02Goo`03003ooooo02Koo`03003ooooo02Koo`03003ooooo01_oo`40007oo`006ooo00<0
0?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?oo
ool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo
093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000K
ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001?oo`@000Coo`03003ooooo093oo`03003o
oooo06goo`40007oo`005Ooo00<00?ooool00ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000Eool0
0`00ooooo`03ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool000goo`@000Coo`03003ooooo
00?oo`03003ooooo093oo`03003ooooo06goo`40007oo`005Ooo00<00?ooool00ooo00<00?ooool0
T?oo00<00?ooool0KOoo0@000Ooo000Cool30005ool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001Goo`03003ooooo00?oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?oo
ool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool0
01_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0
KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093o
o`03003ooooo06goo`40007oo`006ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000Kool00`00oooo
o`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`00
6ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]
ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo
00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03
003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@00
0Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003o
oooo06goo`40007oo`006ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000Kool00`00ooooo`2@ool0
0`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<0
0?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0Mooo00<00?oo
ool01_oo00<00?ooool03Ooo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`1hool00`00oooo
o`04ool00`00ooooo`0>ool00`00ooooo`1]ool10001ool001_oo`03003ooooo07Woo`03003ooooo
00;oo`03003ooooo00ooo`03003ooooo06goo`40007oo`006ooo00<00?ooool0N_oo00@00?ooool0
01;oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0Nooo0P004ooo00<00?ooool0KOoo0@00
0Ooo000Kool00`00ooooo`1kool2000Cool00`00ooooo`1]ool10001ool001_oo`8007_oo`04003o
oooo000Bool00`00ooooo`1/ool20001ool001_oo`03003ooooo07Woo`03003ooooo00;oo`03003o
oooo00ooo`03003ooooo06goo`40007oo`006ooo00<00?ooool0N?oo00<00?ooool01?oo00<00?oo
ool03_oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`1gool00`00ooooo`06ool00`00oooo
o`0=ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`00
6ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]
ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo
00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03
003ooooo093oo`03003ooooo06goo`40007oo`006ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000K
ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06go
o`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool0
0`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<0
0?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?oo
ool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool000Soo`8000Coo`03
003ooooo00;oo`<000Goo`03003ooooo093oo`03003ooooo06goo`40007oo`001ooo00@00?ooool0
00_oo`03003ooooo00;oo`80097oo`03003ooooo06coo`80007oo`001ooo00@00?ooool000_oo`03
003ooooo00;oo`03003ooooo093oo`03003ooooo06goo`40007oo`0000?oo`0000000P000_oo00@0
0?ooool000Soo`<000Goo`03003ooooo093oo`03003ooooo06goo`40007oo`001ooo00@00?ooool0
00Soo`03003ooooo00Goo`03003ooooo093oo`03003ooooo06goo`40007oo`001ooo00@00?ooool0
00Soo`03003ooooo00Goo`03003ooooo093oo`03003ooooo06goo`40007oo`002?oo0P002Ooo1000
1?oo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]
ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo
00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03
003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@00
0Ooo000Kool2002Aool00`00ooooo`1/ool20001ool001_oo`03003ooooo093oo`03003ooooo06go
o`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool0
0`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<0
0?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?oo
ool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo
093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000K
ool2002Aool00`00ooooo`1/ool20001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007o
o`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00oooo
o`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0
T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_o
o`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo
0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03
003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool2002A
ool00`00ooooo`1/ool20001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo
00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool1
0001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<0
0?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003o
oooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo
000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo
06goo`40007oo`006ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000Kool00`00ooooo`2@ool00`00
ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?oo
ool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool0
01_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0
KOoo0@000Ooo000Kool00`00ooooo`05ool00`00ooooo`07ool00`00ooooo`1Cool4000Aool4000B
ool00`00ooooo`1Nool00`00ooooo`040003ool00`00ooooo`02ool10001ool001_oo`03003ooooo
00Koo`03003ooooo00Goo`03003ooooo05?oo`03003ooooo00;oo`8000koo`03003ooooo00;oo`80
013oo`03003ooooo05ooo`8000Coo`<000Goo`40007oo`006ooo00<00?ooool01ooo00<00?ooool0
0ooo00<00?ooool0E?oo00<00?ooool01?oo0P003?oo00<00?ooool01?oo00<00?ooool03Ooo00<0
0?ooool0H?oo00<00?ooool00ooo0P001Ooo0@000Ooo000Kool00`00ooooo`08ool01@00oooooooo
0000E_oo00<00?ooool01_oo00<00?ooool02Ooo00<00?ooool01Ooo00<00?ooool03Ooo00<00?oo
ool0Gooo00<00?oo00000ooo00<00?oo00001Ooo0@000Ooo000Dool20005ool00`00ooooo`09ool0
0`00ool0001Gool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`06ool00`00ooooo`0<ool0
0`00ooooo`1Oool01P00ooooo`00ool000?oo`03003ooooo00;oo`40007oo`004ooo00@00?ooool0
00Coool000D0007oo`004ooo00@00?ooool000Coo`03003ooooo00Woo`03003oo`0005Ooo`03003o
oooo00Ooo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00_oo`03003ooooo05koo`07003o
ooooool00?oo00000ooo00<00?ooool00_oo0@000Ooo000Cool01000ooooo`001?oo00<00?ooool0
2?oo00D00?ooooooo`0005Koo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Koo`03003o
oooo00coo`03003ooooo05koo`04003ooooo0003ool01000ooooo`001?oo0@000Ooo000Cool01000
ooooo`001?oo00<00?ooool01ooo00<00?ooool00ooo00<00?ooool0Dooo00<00?ooool01_oo00<0
0?ooool02Ooo00<00?ooool01_oo00<00?ooool03?oo00<00?ooool0G_oo00<00?oo00001Ooo00<0
0?oo00001?oo0@000Ooo000Cool01000ooooo`001?oo00<00?ooool01_oo00<00?ooool01Ooo00<0
0?ooool0Dooo00<00?ooool01?oo00<00?ooool02ooo00<00?ooool00ooo0P003ooo00<00?ooool0
Gooo00<00?ooool01?oo0P001Ooo0@000Ooo000Dool20005ool00`00ooooo`05ool00`00ooooo`07
ool00`00ooooo`1Cool6000?ool5000Aool00`00ooooo`1Nool00`00ool000050002ool00`00oooo
o`02ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0
T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_o
o`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo
0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`80097oo`03003ooooo
06coo`80007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@
ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo
00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool1
0001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<0
0?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003o
oooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo
000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`80097oo`03003ooooo06coo`80
007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00
ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?oo
ool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool0
01_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0
KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093o
o`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool0
0`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`80097oo`03003ooooo06coo`80007oo`00
6ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]
ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo
00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03
003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@00
0Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003o
oooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool2002Aool0
0`00ooooo`1/ool20001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<0
0?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001
ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?oo
ool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo
093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000K
ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06go
o`40007oo`002?oo0P001?oo00<00?ooool00_oo0`001Ooo00<00?ooool0T?oo00<00?ooool0KOoo
0@000Ooo0007ool01000ooooo`002ooo00<00?ooool00_oo0P00TOoo00<00?ooool0K?oo0P000Ooo
0007ool01000ooooo`002ooo00<00?ooool00_oo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo
0007ool01000ooooo`002?oo0`001Ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo0007ool0
1000ooooo`002?oo00<00?ooool01Ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo0007ool0
1000ooooo`002?oo00<00?ooool01Ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo0008ool2
0009ool40004ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03
003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00
ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007o
o`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00oooo
o`1]ool10001ool001_oo`80097oo`03003ooooo06coo`80007oo`006ooo00<00?ooool0T?oo00<0
0?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003o
oooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo
000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo
06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`1g
ool00`00ooooo`06ool00`00ooooo`0=ool00`00ooooo`1]ool10001ool001_oo`03003ooooo07So
o`03003ooooo00Coo`03003ooooo00koo`03003ooooo06goo`40007oo`006ooo00<00?ooool0NOoo
00<00?ooool00_oo00<00?ooool03ooo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`1jool0
1000ooooo`004_oo00<00?ooool0KOoo0@000Ooo000Kool2001lool2000Cool00`00ooooo`1/ool2
0001ool001_oo`03003ooooo07_oo`8001?oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0
Nooo0P004ooo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`1jool01000ooooo`004_oo00<0
0?ooool0KOoo0@000Ooo000Kool00`00ooooo`1iool00`00ooooo`02ool00`00ooooo`0?ool00`00
ooooo`1]ool10001ool001_oo`03003ooooo07Soo`03003ooooo00Coo`03003ooooo00koo`03003o
oooo06goo`40007oo`006ooo00<00?ooool0Mooo00<00?ooool01_oo00<00?ooool03Ooo00<00?oo
ool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo
093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000K
ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06go
o`40007oo`006ooo0P00TOoo00<00?ooool0K?oo0P000Ooo000Kool00`00ooooo`2@ool00`00oooo
o`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0
T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_o
o`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo
0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03
003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00
ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`80097oo`03003ooooo06coo`80007oo`006ooo
00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool1
0001ool001_oo`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<0
0?ooool0KOoo0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003o
oooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo
000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03003ooooo
06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Cool40004ool00`00
ooooo`2@ool00`00ooooo`1]ool10001ool001Goo`03003ooooo00?oo`80097oo`03003ooooo06co
o`80007oo`005Ooo00<00?ooool00ooo00<00?ooool0T?oo00<00?ooool0KOoo0@000Ooo000Eool0
0`00ooooo`03ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001Goo`03003ooooo00?oo`03
003ooooo093oo`03003ooooo06goo`40007oo`004ooo0`001Ooo00<00?ooool0T?oo00<00?ooool0
KOoo0@000Ooo000Eool00`00ooooo`03ool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_o
o`03003ooooo093oo`03003ooooo06goo`40007oo`006ooo00<00?ooool0T?oo00<00?ooool0KOoo
0@000Ooo000Kool00`00ooooo`2@ool00`00ooooo`1]ool10001ool001_oo`03003ooooo093oo`03
003ooooo06goo`40007oo`006ooo00<00?ooool05_oo00<00?ooool09_oo00<00?ooool09_oo00<0
0?ooool09Ooo00<00?ooool09_oo00<00?ooool09_oo00<00?ooool06ooo0@000Ooo000Kool00`00
ooooo`02ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`07ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`08ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`08ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`08ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`07ool10001ool001_oool000D0007oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo
003ooolQool00?ooob7oo`00oooo8Ooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-8.60167, -1.26786, 0.0491564, 
  0.0084737}}]
}, Open  ]],

Cell["\<\
The pole-zero pair at s = 5 must be cancelled.  We do this \
with\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
comp2[s_] = CancelZP[comp1[s],s-5,s,Tolerance -> 0.001]

PlotZP[comp2[s],s,Frame -> True];\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \(\(\(-7.62710907475404553`\) - 15.3527968971711885`\ s - 
        9.63246509110565796`\ s\^2 - 
        1.93142195560428717`\ s\^3\)\/\(\(-17.662883529907476`\) - 
        25.7997339925336621`\ s - 14.577643110987295`\ s\^2 - 
        1.67750184883462205`\ s\^3\)\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.868458 0.119584 0.5 0.454545 [
[.15095 -0.0125 -6 -9 ]
[.15095 -0.0125 6 0 ]
[.39012 -0.0125 -6 -9 ]
[.39012 -0.0125 6 0 ]
[.62929 -0.0125 -6 -9 ]
[.62929 -0.0125 6 0 ]
[.86846 -0.0125 -3 -9 ]
[.86846 -0.0125 3 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .04545 -12 -4.5 ]
[-0.0125 .04545 0 4.5 ]
[-0.0125 .27273 -24 -4.5 ]
[-0.0125 .27273 0 4.5 ]
[-0.0125 .5 -6 -4.5 ]
[-0.0125 .5 0 4.5 ]
[-0.0125 .72727 -18 -4.5 ]
[-0.0125 .72727 0 4.5 ]
[-0.0125 .95455 -6 -4.5 ]
[-0.0125 .95455 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15095 0 m
.15095 .00625 L
s
[(-6)] .15095 -0.0125 0 1 Mshowa
.39012 0 m
.39012 .00625 L
s
[(-4)] .39012 -0.0125 0 1 Mshowa
.62929 0 m
.62929 .00625 L
s
[(-2)] .62929 -0.0125 0 1 Mshowa
.86846 0 m
.86846 .00625 L
s
[(0)] .86846 -0.0125 0 1 Mshowa
.125 Mabswid
.21075 0 m
.21075 .00375 L
s
.27054 0 m
.27054 .00375 L
s
.33033 0 m
.33033 .00375 L
s
.44991 0 m
.44991 .00375 L
s
.50971 0 m
.50971 .00375 L
s
.5695 0 m
.5695 .00375 L
s
.68908 0 m
.68908 .00375 L
s
.74887 0 m
.74887 .00375 L
s
.80867 0 m
.80867 .00375 L
s
.09116 0 m
.09116 .00375 L
s
.03137 0 m
.03137 .00375 L
s
.92825 0 m
.92825 .00375 L
s
.98804 0 m
.98804 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .04545 m
.00625 .04545 L
s
[(-1)] -0.0125 .04545 1 0 Mshowa
0 .27273 m
.00625 .27273 L
s
[(-0.5)] -0.0125 .27273 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(0)] -0.0125 .5 1 0 Mshowa
0 .72727 m
.00625 .72727 L
s
[(0.5)] -0.0125 .72727 1 0 Mshowa
0 .95455 m
.00625 .95455 L
s
[(1)] -0.0125 .95455 1 0 Mshowa
.125 Mabswid
0 .09091 m
.00375 .09091 L
s
0 .13636 m
.00375 .13636 L
s
0 .18182 m
.00375 .18182 L
s
0 .22727 m
.00375 .22727 L
s
0 .31818 m
.00375 .31818 L
s
0 .36364 m
.00375 .36364 L
s
0 .40909 m
.00375 .40909 L
s
0 .45455 m
.00375 .45455 L
s
0 .54545 m
.00375 .54545 L
s
0 .59091 m
.00375 .59091 L
s
0 .63636 m
.00375 .63636 L
s
0 .68182 m
.00375 .68182 L
s
0 .77273 m
.00375 .77273 L
s
0 .81818 m
.00375 .81818 L
s
0 .86364 m
.00375 .86364 L
s
0 .90909 m
.00375 .90909 L
s
0 1 m
.00375 1 L
s
.25 Mabswid
0 0 m
0 1 L
s
.15095 .99375 m
.15095 1 L
s
.39012 .99375 m
.39012 1 L
s
.62929 .99375 m
.62929 1 L
s
.86846 .99375 m
.86846 1 L
s
.125 Mabswid
.21075 .99625 m
.21075 1 L
s
.27054 .99625 m
.27054 1 L
s
.33033 .99625 m
.33033 1 L
s
.44991 .99625 m
.44991 1 L
s
.50971 .99625 m
.50971 1 L
s
.5695 .99625 m
.5695 1 L
s
.68908 .99625 m
.68908 1 L
s
.74887 .99625 m
.74887 1 L
s
.80867 .99625 m
.80867 1 L
s
.09116 .99625 m
.09116 1 L
s
.03137 .99625 m
.03137 1 L
s
.92825 .99625 m
.92825 1 L
s
.98804 .99625 m
.98804 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .04545 m
1 .04545 L
s
.99375 .27273 m
1 .27273 L
s
.99375 .5 m
1 .5 L
s
.99375 .72727 m
1 .72727 L
s
.99375 .95455 m
1 .95455 L
s
.125 Mabswid
.99625 .09091 m
1 .09091 L
s
.99625 .13636 m
1 .13636 L
s
.99625 .18182 m
1 .18182 L
s
.99625 .22727 m
1 .22727 L
s
.99625 .31818 m
1 .31818 L
s
.99625 .36364 m
1 .36364 L
s
.99625 .40909 m
1 .40909 L
s
.99625 .45455 m
1 .45455 L
s
.99625 .54545 m
1 .54545 L
s
.99625 .59091 m
1 .59091 L
s
.99625 .63636 m
1 .63636 L
s
.99625 .68182 m
1 .68182 L
s
.99625 .77273 m
1 .77273 L
s
.99625 .81818 m
1 .81818 L
s
.99625 .86364 m
1 .86364 L
s
.99625 .90909 m
1 .90909 L
s
.99625 0 m
1 0 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 .5 m
1 .5 L
s
.86846 0 m
.86846 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
.62929 .5 .02 0 365.73 arc
s
newpath
.62929 .5 .02 0 365.73 arc
s
newpath
.7504 .5 .02 0 365.73 arc
s
.06077 .48182 m
.09713 .51818 L
s
.06077 .51818 m
.09713 .48182 L
s
.72543 .15883 m
.7618 .19519 L
s
.72543 .19519 m
.7618 .15883 L
s
.72543 .80481 m
.7618 .84117 L
s
.72543 .84117 m
.7618 .80481 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000181000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00
oooo8Ooo003ooolQool00?ooob7oo`00A?oo0P00?Ooo00<00?ooool0>Ooo1000>?oo0P008ooo0013
ool01000ooooo`00??oo00<00?ooool0>Ooo00<00?ooool0>?oo00@00?ooool002;oo`00@ooo00@0
0?ooool003Woo`D003_oo`03003ooooo03Ooo`04003ooooo000Rool003goo`@000;oo`<003Goo`@0
0005ool00?ooool0000eool40004ool00`00ooooo`0fool01000ooooo`008_oo0013ool00`00oooo
o`0kool00`00ool0000nool00`00ooooo`0eool01000ooooo`008_oo0013ool00`00ooooo`0lool2
000kool01000ooooo`00=ooo00@00?ooool002;oo`00A?oo0`00??oo00<00?ooool0>_oo0P00>Ooo
0P008ooo003ooolQool001_oool000D0007oo`006ooo00<00?ooool01Ooo00<00?ooool03?oo00<0
0?ooool03Ooo00<00?ooool03?oo00<00?ooool03Ooo00<00?ooool03Ooo00<00?ooool03?oo00<0
0?ooool03Ooo00<00?ooool03?oo00<00?ooool03Ooo00<00?ooool03?oo00<00?ooool03Ooo00<0
0?ooool03?oo00<00?ooool03Ooo00<00?ooool03?oo00<00?ooool03Ooo00<00?ooool03?oo00@0
0?ooool0007oo`006ooo00<00?ooool09?oo00<00?ooool0>ooo00<00?ooool0>ooo00<00?ooool0
>ooo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_o
o`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo
0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03
003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00
ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007o
o`004ooo10001?oo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Eool00`00ooooo`03ool2
003Oool00`00ooooo`0Nool20001ool001Goo`03003ooooo00?oo`03003ooooo0=koo`03003ooooo
01ooo`40007oo`003Ooo10001?oo00<00?ooool00ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Eool00`00ooooo`03ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001?oo`<000Go
o`03003ooooo0=koo`03003ooooo01ooo`40007oo`005Ooo00<00?ooool00ooo00<00?ooool0g_oo
00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003o
oooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool2003Oool0
0`00ooooo`0Nool20001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<0
0?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001
ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?oo
ool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo
0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000K
ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01oo
o`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool2003Oool00`00oooo
o`0Nool20001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_o
o`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo
0@000Ooo000Kool00`00ooooo`2iool00`00ooooo`07ool00`00ooooo`0Hool00`00ooooo`0Oool1
0001ool001_oo`03003ooooo0;[oo`03003ooooo00Goo`03003ooooo01Woo`03003ooooo01ooo`40
007oo`006ooo00<00?ooool0^ooo00<00?ooool00ooo00<00?ooool06_oo00<00?ooool07ooo0@00
0Ooo000Kool00`00ooooo`2lool01@00oooooooo00007Ooo00<00?ooool07ooo0@000Ooo000Kool0
0`00ooooo`2mool3000Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0;goo`<001koo`03
003ooooo01ooo`40007oo`006ooo0P00_Ooo00D00?ooooooo`0001goo`03003ooooo01koo`80007o
o`006ooo00<00?ooool0^ooo00<00?ooool00ooo00<00?ooool06_oo00<00?ooool07ooo0@000Ooo
000Kool00`00ooooo`2jool00`00ooooo`05ool00`00ooooo`0Iool00`00ooooo`0Oool10001ool0
01_oo`03003ooooo0;Woo`03003ooooo00Ooo`03003ooooo01Soo`03003ooooo01ooo`40007oo`00
6ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo
00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Kool2003Oool00`00ooooo`0Nool20001ool001_oo`03003ooooo0=koo`03003ooooo01oo
o`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool0
0`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<0
0?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001
ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?oo
ool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo
0=koo`03003ooooo01ooo`40007oo`002?oo0P001?oo00<00?ooool00_oo0`001Ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo0007ool01000ooooo`002ooo00<00?ooool00_oo0P00gooo00<0
0?ooool07_oo0P000Ooo0007ool01000ooooo`002ooo00<00?ooool00_oo00<00?ooool0g_oo00<0
0?ooool07ooo0@000Ooo00000ooo000000020002ool01000ooooo`002?oo0`001Ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo0007ool01000ooooo`002?oo00<00?ooool01Ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo0007ool01000ooooo`002?oo00<00?ooool01Ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo0008ool20009ool40004ool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo
00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`800=ooo`03003ooooo01ko
o`80007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool0
0`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<0
0?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001
ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?oo
ool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo
0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000K
ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`800=ooo`03003ooooo01koo`80007o
o`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00oooo
o`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_o
o`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo
0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03
003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00
ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`800=ooo`03003ooooo01koo`80007oo`006ooo
00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool1
0001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<0
0?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003o
oooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo
000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo
01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool2003Oool00`00
ooooo`0Nool20001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?oo
ool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool0
01_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool0
7ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo00go
o`03003ooooo00Koo`03003ooooo08Goo`@001_oo`@001goo`03003ooooo01ooo`40007oo`006ooo
00<00?ooool03_oo00<00?ooool01?oo00<00?ooool0QOoo00<00?ooool00_oo00<00?ooool05ooo
00<00?ooool00_oo0P006ooo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`0?ool00`00oooo
o`02ool00`00ooooo`25ool00`00ooooo`04ool2000Gool00`00ooooo`04ool2000Iool00`00oooo
o`0Oool10001ool001_oo`03003ooooo013oo`04003ooooo0027ool00`00ooooo`06ool00`00oooo
o`0Dool00`00ooooo`06ool00`00ooooo`0Gool00`00ooooo`0Oool10001ool001Coo`8000Goo`03
003ooooo017oo`8008Soo`03003ooooo00Ooo`03003ooooo01?oo`03003ooooo00Ooo`03003ooooo
01Koo`03003ooooo01ooo`40007oo`004ooo00@00?ooool000Coool000D0007oo`004ooo00@00?oo
ool000Coo`03003ooooo017oo`8008Soo`03003ooooo00Ooo`03003ooooo01;oo`03003ooooo00So
o`03003ooooo01Koo`03003ooooo01ooo`40007oo`004ooo00@00?ooool000Coo`03003ooooo013o
o`04003ooooo0027ool00`00ooooo`07ool00`00ooooo`0Cool00`00ooooo`06ool00`00ooooo`0G
ool00`00ooooo`0Oool10001ool001?oo`04003ooooo0004ool00`00ooooo`0?ool00`00ooooo`02
ool00`00ooooo`24ool00`00ooooo`07ool00`00ooooo`0Cool00`00ooooo`06ool00`00ooooo`0G
ool00`00ooooo`0Oool10001ool001?oo`04003ooooo0004ool00`00ooooo`0>ool00`00ooooo`04
ool00`00ooooo`24ool00`00ooooo`04ool2000Gool00`00ooooo`03ool2000Jool00`00ooooo`0O
ool10001ool001Coo`8000Goo`03003ooooo00goo`03003ooooo00Koo`03003ooooo08Coo`H001[o
o`D001coo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo
000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo
01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3N
ool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo
0P00gooo00<00?ooool07_oo0P000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool0
01_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool0
7ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=ko
o`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool0
0`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40
007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00
ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo0P00gooo
00<00?ooool07_oo0P000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003o
oooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00oooo
o`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`00
6ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo0P00gooo00<00?oo
ool07_oo0P000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo
0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000K
ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01oo
o`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool0
0`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<0
0?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001
ool001_oo`800=ooo`03003ooooo01koo`80007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo
0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03
003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00
ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007o
o`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00oooo
o`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0
g_oo00<00?ooool07ooo0@000Ooo0008ool20004ool00`00ooooo`02ool30005ool00`00ooooo`3N
ool00`00ooooo`0Oool10001ool000Ooo`04003ooooo000;ool00`00ooooo`02ool2003Oool00`00
ooooo`0Nool20001ool000Ooo`04003ooooo000;ool00`00ooooo`02ool00`00ooooo`3Nool00`00
ooooo`0Oool10001ool000Ooo`04003ooooo0008ool30005ool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool000Ooo`04003ooooo0008ool00`00ooooo`05ool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool000Ooo`04003ooooo0008ool00`00ooooo`05ool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool000Soo`8000Woo`@000Coo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo
00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool1
0001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<0
0?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003o
oooo0=koo`03003ooooo01ooo`40007oo`006ooo0P00gooo00<00?ooool07_oo0P000Ooo000Kool0
0`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40
007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00
ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?oo
ool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool0
01_oo`03003ooooo0;Woo`03003ooooo00Ooo`03003ooooo01Soo`03003ooooo01ooo`40007oo`00
6ooo00<00?ooool0^_oo00<00?ooool01Ooo00<00?ooool06Ooo00<00?ooool07ooo0@000Ooo000K
ool00`00ooooo`2kool00`00ooooo`03ool00`00ooooo`0Jool00`00ooooo`0Oool10001ool001_o
o`03003ooooo0;coo`05003oooooool0000Mool00`00ooooo`0Oool10001ool001_oo`800;koo`03
003oo`0001koo`03003ooooo01koo`80007oo`006ooo00<00?ooool0__oo00<00?ooool07Ooo00<0
0?ooool07ooo0@000Ooo000Kool00`00ooooo`2mool00`00ool0000Nool00`00ooooo`0Oool10001
ool001_oo`03003ooooo0;coo`05003oooooool0000Mool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0;_oo`03003ooooo00?oo`03003ooooo01[oo`03003ooooo01ooo`40007oo`006ooo00<0
0?ooool0^_oo00<00?ooool01Ooo00<00?ooool06Ooo00<00?ooool07ooo0@000Ooo000Kool00`00
ooooo`2iool00`00ooooo`07ool00`00ooooo`0Hool00`00ooooo`0Oool10001ool001_oo`03003o
oooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo
000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo
01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool2003Oool00`00
ooooo`0Nool20001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?oo
ool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool0
01_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool0
7ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=ko
o`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool0
0`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40
007oo`006ooo0P00gooo00<00?ooool07_oo0P000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0O
ool10001ool001_oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo
00<00?ooool07ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03
003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@00
0Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=koo`03003o
oooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool00`00oooo
o`3Nool00`00ooooo`0Oool10001ool001?oo`@000Coo`03003ooooo0=koo`03003ooooo01ooo`40
007oo`005Ooo00<00?ooool00ooo0P00gooo00<00?ooool07_oo0P000Ooo000Eool00`00ooooo`03
ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001Goo`03003ooooo00?oo`03003ooooo0=ko
o`03003ooooo01ooo`40007oo`005Ooo00<00?ooool00ooo00<00?ooool0g_oo00<00?ooool07ooo
0@000Ooo000Cool30005ool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001Goo`03003ooooo
00?oo`03003ooooo0=koo`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool0
7ooo0@000Ooo000Kool00`00ooooo`3Nool00`00ooooo`0Oool10001ool001_oo`03003ooooo0=ko
o`03003ooooo01ooo`40007oo`006ooo00<00?ooool0g_oo00<00?ooool07ooo0@000Ooo000Kool0
0`00ooooo`0Tool00`00ooooo`0kool00`00ooooo`0kool00`00ooooo`0kool00`00ooooo`0Oool1
0001ool001_oo`03003ooooo00Goo`03003ooooo00coo`03003ooooo00goo`03003ooooo00coo`03
003ooooo00goo`03003ooooo00goo`03003ooooo00coo`03003ooooo00goo`03003ooooo00coo`03
003ooooo00goo`03003ooooo00coo`03003ooooo00goo`03003ooooo00coo`03003ooooo00goo`03
003ooooo00coo`03003ooooo00goo`03003ooooo00coo`04003ooooo0001ool001_oool000D0007o
o`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo0000
\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-8.13991, -1.26786, 0.032209, 
  0.0084737}}]
}, Open  ]],

Cell["Now we calculate the resulting closed loop function", "Text"],

Cell[CellGroupData[{

Cell["\<\
Trat2[s_] = Together[p[s] comp2[s]/(1 + p[s] comp2[s])]

PlotZP[Trat2[s],s,Frame -> True];\
\>", "Input"],

Cell[BoxData[
    \(\(4\ \((\(-5\) + s)\)\ 
        \((\(7.62710907475404553`\[InvisibleSpace]\) + 
            15.3527968971711885`\ s + 9.63246509110565796`\ s\^2 + 
            1.93142195560428717`\ s\^3)\)\)\/\(\(-505.799852093230395`\) - 
        1145.7998520932302`\ s - 1027.10007395338487`\ s\^2 - 
        454.200147906769835`\ s\^3 - 98.7125647092117652`\ s\^4 - 
        8.38750924417310983`\ s\^5\)\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.429809 0.103671 0.5 0.454545 [
[.01513 -0.0125 -6 -9 ]
[.01513 -0.0125 6 0 ]
[.22247 -0.0125 -6 -9 ]
[.22247 -0.0125 6 0 ]
[.42981 -0.0125 -3 -9 ]
[.42981 -0.0125 3 0 ]
[.63715 -0.0125 -3 -9 ]
[.63715 -0.0125 3 0 ]
[.84449 -0.0125 -3 -9 ]
[.84449 -0.0125 3 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .04545 -12 -4.5 ]
[-0.0125 .04545 0 4.5 ]
[-0.0125 .27273 -24 -4.5 ]
[-0.0125 .27273 0 4.5 ]
[-0.0125 .5 -6 -4.5 ]
[-0.0125 .5 0 4.5 ]
[-0.0125 .72727 -18 -4.5 ]
[-0.0125 .72727 0 4.5 ]
[-0.0125 .95455 -6 -4.5 ]
[-0.0125 .95455 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.01513 0 m
.01513 .00625 L
s
[(-4)] .01513 -0.0125 0 1 Mshowa
.22247 0 m
.22247 .00625 L
s
[(-2)] .22247 -0.0125 0 1 Mshowa
.42981 0 m
.42981 .00625 L
s
[(0)] .42981 -0.0125 0 1 Mshowa
.63715 0 m
.63715 .00625 L
s
[(2)] .63715 -0.0125 0 1 Mshowa
.84449 0 m
.84449 .00625 L
s
[(4)] .84449 -0.0125 0 1 Mshowa
.125 Mabswid
.06696 0 m
.06696 .00375 L
s
.1188 0 m
.1188 .00375 L
s
.17063 0 m
.17063 .00375 L
s
.2743 0 m
.2743 .00375 L
s
.32614 0 m
.32614 .00375 L
s
.37797 0 m
.37797 .00375 L
s
.48164 0 m
.48164 .00375 L
s
.53348 0 m
.53348 .00375 L
s
.58532 0 m
.58532 .00375 L
s
.68899 0 m
.68899 .00375 L
s
.74082 0 m
.74082 .00375 L
s
.79266 0 m
.79266 .00375 L
s
.89633 0 m
.89633 .00375 L
s
.94816 0 m
.94816 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .04545 m
.00625 .04545 L
s
[(-1)] -0.0125 .04545 1 0 Mshowa
0 .27273 m
.00625 .27273 L
s
[(-0.5)] -0.0125 .27273 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(0)] -0.0125 .5 1 0 Mshowa
0 .72727 m
.00625 .72727 L
s
[(0.5)] -0.0125 .72727 1 0 Mshowa
0 .95455 m
.00625 .95455 L
s
[(1)] -0.0125 .95455 1 0 Mshowa
.125 Mabswid
0 .09091 m
.00375 .09091 L
s
0 .13636 m
.00375 .13636 L
s
0 .18182 m
.00375 .18182 L
s
0 .22727 m
.00375 .22727 L
s
0 .31818 m
.00375 .31818 L
s
0 .36364 m
.00375 .36364 L
s
0 .40909 m
.00375 .40909 L
s
0 .45455 m
.00375 .45455 L
s
0 .54545 m
.00375 .54545 L
s
0 .59091 m
.00375 .59091 L
s
0 .63636 m
.00375 .63636 L
s
0 .68182 m
.00375 .68182 L
s
0 .77273 m
.00375 .77273 L
s
0 .81818 m
.00375 .81818 L
s
0 .86364 m
.00375 .86364 L
s
0 .90909 m
.00375 .90909 L
s
0 1 m
.00375 1 L
s
.25 Mabswid
0 0 m
0 1 L
s
.01513 .99375 m
.01513 1 L
s
.22247 .99375 m
.22247 1 L
s
.42981 .99375 m
.42981 1 L
s
.63715 .99375 m
.63715 1 L
s
.84449 .99375 m
.84449 1 L
s
.125 Mabswid
.06696 .99625 m
.06696 1 L
s
.1188 .99625 m
.1188 1 L
s
.17063 .99625 m
.17063 1 L
s
.2743 .99625 m
.2743 1 L
s
.32614 .99625 m
.32614 1 L
s
.37797 .99625 m
.37797 1 L
s
.48164 .99625 m
.48164 1 L
s
.53348 .99625 m
.53348 1 L
s
.58532 .99625 m
.58532 1 L
s
.68899 .99625 m
.68899 1 L
s
.74082 .99625 m
.74082 1 L
s
.79266 .99625 m
.79266 1 L
s
.89633 .99625 m
.89633 1 L
s
.94816 .99625 m
.94816 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 .04545 m
1 .04545 L
s
.99375 .27273 m
1 .27273 L
s
.99375 .5 m
1 .5 L
s
.99375 .72727 m
1 .72727 L
s
.99375 .95455 m
1 .95455 L
s
.125 Mabswid
.99625 .09091 m
1 .09091 L
s
.99625 .13636 m
1 .13636 L
s
.99625 .18182 m
1 .18182 L
s
.99625 .22727 m
1 .22727 L
s
.99625 .31818 m
1 .31818 L
s
.99625 .36364 m
1 .36364 L
s
.99625 .40909 m
1 .40909 L
s
.99625 .45455 m
1 .45455 L
s
.99625 .54545 m
1 .54545 L
s
.99625 .59091 m
1 .59091 L
s
.99625 .63636 m
1 .63636 L
s
.99625 .68182 m
1 .68182 L
s
.99625 .77273 m
1 .77273 L
s
.99625 .81818 m
1 .81818 L
s
.99625 .86364 m
1 .86364 L
s
.99625 .90909 m
1 .90909 L
s
.99625 0 m
1 0 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 .5 m
1 .5 L
s
.42981 0 m
.42981 1 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
newpath
.22247 .5 .02 0 365.73 arc
s
newpath
.22247 .5 .02 0 365.73 arc
s
newpath
.32746 .5 .02 0 365.73 arc
s
newpath
.94816 .5 .02 0 365.73 arc
s
.02089 .48182 m
.05726 .51818 L
s
.02089 .51818 m
.05726 .48182 L
s
.20429 .48182 m
.24065 .51818 L
s
.20429 .51818 m
.24065 .48182 L
s
.20429 .48182 m
.24065 .51818 L
s
.20429 .51818 m
.24065 .48182 L
s
.20429 .48182 m
.24065 .51818 L
s
.20429 .51818 m
.24065 .48182 L
s
.20429 .48182 m
.24065 .51818 L
s
.20429 .51818 m
.24065 .48182 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P000181000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00
oooo8Ooo003ooolQool00?ooob7oo`008_oo00<00?ooool0<Ooo1000;ooo0P00<ooo1000=?oo00<0
0?ooool09ooo000Rool00`00ooooo`0aool00`00ooooo`0_ool01000ooooo`00<_oo00<00?ooool0
=Ooo00<00?ooool09ooo000Oool5000cool00`00ooooo`0^ool01000ooooo`00<ooo00<00?ooool0
<Ooo1@00:?oo000Jool400001Ooo003ooooo0000;Ooo10001?oo00<00?ooool0;Ooo00@00?ooool0
03Coo`03003ooooo033oo`04003ooooo000Yool0023oo`03003oo`0003Koo`03003ooooo02coo`04
003ooooo000eool00`00ooooo`0`ool00`00ool0000Yool0027oo`8003?oo`04003ooooo000^ool0
1000ooooo`00<_oo00@00?ooool003?oo`8002Woo`008_oo00<00?ooool0<_oo0P00<?oo0P00=?oo
0P00=Ooo00<00?ooool09ooo003ooolQool001_oool000D0007oo`006ooo00D00?ooooooo`0000co
o`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00[o
o`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00_o
o`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[o
o`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`40007oo`006ooo00D00?ooooooo`00
03Goo`03003ooooo03;oo`03003ooooo03?oo`03003ooooo03?oo`03003ooooo02Goo`40007oo`00
6ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2A
ool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo
00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03
003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@00
0Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001?oo`@000Coo`03003ooooo06co
o`03003ooooo097oo`40007oo`005Ooo00<00?ooool00ooo0P00KOoo00<00?ooool0T?oo0P000Ooo
000Eool00`00ooooo`03ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000goo`@000Coo`03
003ooooo00?oo`03003ooooo06coo`03003ooooo097oo`40007oo`005Ooo00<00?ooool00ooo00<0
0?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Cool30005ool00`00ooooo`1/ool00`00ooooo`2A
ool10001ool001Goo`03003ooooo00?oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo
00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool1
0001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<0
0?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003o
oooo06coo`03003ooooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40
007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00
ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?oo
ool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool0
01_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0
TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06co
o`03003ooooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00oooo
o`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`00
6ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2A
ool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo
00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03
003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@00
0Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003o
oooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool0
0`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<0
0?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001
ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?oo
ool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo
06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000K
ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`8006goo`03003ooooo093oo`80007o
o`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00oooo
o`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0
K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_o
o`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo
0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03
003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo0008ool20004
ool00`00ooooo`02ool30005ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Ooo`04003o
oooo000;ool00`00ooooo`02ool2001]ool00`00ooooo`2@ool20001ool000Ooo`04003ooooo000;
ool00`00ooooo`02ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool00003ool00000008000;o
o`04003ooooo0008ool30005ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Ooo`04003o
oooo0008ool00`00ooooo`05ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Ooo`04003o
oooo0008ool00`00ooooo`05ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Soo`8000Wo
o`@000Coo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?oo
ool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo
06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000K
ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097o
o`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00oooo
o`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0
K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_o
o`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo
0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03
003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00
ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007o
o`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool1
0001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<0
0?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003o
oooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo
000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo
097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/
ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo
0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool0
01_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0
TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06co
o`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40
007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00
ooooo`2Aool10001ool001_oo`8006goo`03003ooooo093oo`80007oo`006ooo00<00?ooool0K?oo
00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03
003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@00
0Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003o
oooo097oo`40007oo`006ooo00<00?ooool00_oo00<00?ooool01ooo00<00?ooool08ooo00<00?oo
00000`000ooo00<00?ooool04Ooo10006Ooo00<00?ooool0POoo10003?oo0@000Ooo000Kool00`00
ooooo`03ool00`00ooooo`05ool00`00ooooo`0Uool00`00ooooo`02ool3000Cool00`00ooooo`02
ool2000Gool00`00ooooo`20ool00`00ooooo`02ool2000:ool10001ool001_oo`03003ooooo00Co
o`03003ooooo00?oo`03003ooooo02Koo`8000Coo`8001?oo`03003ooooo00Coo`8001Goo`03003o
oooo083oo`03003ooooo00Coo`8000Soo`40007oo`006ooo00<00?ooool01Ooo00D00?ooooooo`00
02Soo`04003ooooo0002ool00`00ool0000Bool00`00ooooo`06ool00`00ooooo`0Cool00`00oooo
o`1oool00`00ooooo`06ool00`00ooooo`06ool10001ool001Coo`8000Goo`03003ooooo00Koo`03
003oo`0002Woo`04003oooooool20003ool00`00ooooo`0?ool00`00ooooo`07ool00`00ooooo`0B
ool00`00ooooo`1oool00`00ooooo`07ool00`00ooooo`05ool10001ool001?oo`04003ooooo0004
oooo00050001ool001?oo`04003ooooo0004ool00`00ooooo`06ool00`00ool0000Xool00`00oooo
o`02ool20004ool00`00ooooo`0>ool00`00ooooo`07ool00`00ooooo`0Bool00`00ooooo`1oool0
0`00ooooo`07ool00`00ooooo`05ool10001ool001?oo`04003ooooo0004ool00`00ooooo`05ool0
1@00oooooooo0000:?oo00@00?ooool000;oo`04003ooooo000Aool00`00ooooo`07ool00`00oooo
o`0Bool00`00ooooo`1oool00`00ooooo`07ool00`00ooooo`05ool10001ool001?oo`04003ooooo
0004ool00`00ooooo`04ool00`00ooooo`03ool00`00ooooo`0Uool00`00ool00004ool00`00ool0
000Aool00`00ooooo`06ool00`00ooooo`0Cool00`00ooooo`1oool00`00ooooo`07ool00`00oooo
o`05ool10001ool001?oo`04003ooooo0004ool00`00ooooo`03ool00`00ooooo`05ool00`00oooo
o`0Uool00`00ooooo`03ool2000Cool00`00ooooo`04ool00`00ooooo`0Dool00`00ooooo`20ool0
0`00ooooo`04ool20008ool10001ool001Coo`8000Goo`03003ooooo00;oo`03003ooooo00Ooo`03
003ooooo02?oo`03003oo`0000@000;oo`03003ooooo017oo`H001Ooo`03003ooooo087oo`H000[o
o`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool0
0`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<0
0?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001
ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo
0P000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03
003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00
ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007o
o`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00oooo
o`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0
K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_o
o`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo
000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo
097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/
ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo
00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool1
0001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<0
0?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003o
oooo06coo`03003ooooo097oo`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40
007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00
ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?oo
ool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool0
01_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0
TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`8006goo`03003o
oooo093oo`80007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00oooo
o`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`00
6ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2A
ool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo
00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03
003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@00
0Ooo0008ool20004ool00`00ooooo`02ool30005ool00`00ooooo`1/ool00`00ooooo`2Aool10001
ool000Ooo`04003ooooo000;ool00`00ooooo`02ool2001]ool00`00ooooo`2@ool20001ool000Oo
o`04003ooooo000;ool00`00ooooo`02ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Oo
o`04003ooooo0008ool30005ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Ooo`04003o
oooo0008ool00`00ooooo`05ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Ooo`04003o
oooo0008ool00`00ooooo`05ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool000Soo`8000Wo
o`@000Coo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?oo
ool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo
06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000K
ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097o
o`40007oo`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00oooo
o`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0
K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_o
o`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo
0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03
003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00
ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007o
o`006ooo0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool1
0001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<0
0?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003o
oooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo
000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo
097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/
ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo
0P00KOoo00<00?ooool0T?oo0P000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool0
01_oo`03003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0
TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06co
o`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40
007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00
ooooo`2Aool10001ool001_oo`8006goo`03003ooooo093oo`80007oo`006ooo00<00?ooool0K?oo
00<00?ooool0TOoo0@000Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03
003ooooo06coo`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@00
0Ooo000Kool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003o
oooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool00`00oooo
o`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40007oo`00
6ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Cool40004ool00`00ooooo`1/ool00`00
ooooo`2Aool10001ool001Goo`03003ooooo00?oo`8006goo`03003ooooo093oo`80007oo`005Ooo
00<00?ooool00ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Eool00`00ooooo`03ool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001Goo`03003ooooo00?oo`03003ooooo06coo`03
003ooooo097oo`40007oo`004ooo0`001Ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000E
ool00`00ooooo`03ool00`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06co
o`03003ooooo097oo`40007oo`006ooo00<00?ooool0K?oo00<00?ooool0TOoo0@000Ooo000Kool0
0`00ooooo`1/ool00`00ooooo`2Aool10001ool001_oo`03003ooooo06coo`03003ooooo097oo`40
007oo`006ooo00D00?ooooooo`0003Goo`03003ooooo03;oo`03003ooooo03?oo`03003ooooo03?o
o`03003ooooo02Goo`40007oo`006ooo00D00?ooooooo`0000coo`03003ooooo00_oo`03003ooooo
00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo
00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo
00_oo`03003ooooo00[oo`03003ooooo00_oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo
00_oo`03003ooooo00[oo`40007oo`006oooo`001@000Ooo003ooolQool00?ooob7oo`00oooo8Ooo
003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00001\
\>"],
  ImageRangeCache->{{{0, 287}, {287, 0}} -> {-5.15819, -1.26786, 0.0371529, 
  0.0084737}}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5.5 How good is the answer?", "Section"],

Cell[CellGroupData[{

Cell["Check that Trat satisfies the original requirements", "Subsection"],

Cell[CellGroupData[{

Cell["\<\

perf = Plot[ 1/r0[w] Abs[k0[w]-Trat2[I w]],{w,0,10},
                          PlotRange->{0.,2.},
                          Frame -> True,
                          DisplayFunction->Identity];
\t   Show[perf,
                Graphics[{Dashing[{0.05,0.05}],Line[{{0,1},{10,1}}]}],
                          DisplayFunction->$DisplayFunction]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0952381 0 0.309017 [
[.02381 -0.0125 -3 -9 ]
[.02381 -0.0125 3 0 ]
[.21429 -0.0125 -3 -9 ]
[.21429 -0.0125 3 0 ]
[.40476 -0.0125 -3 -9 ]
[.40476 -0.0125 3 0 ]
[.59524 -0.0125 -3 -9 ]
[.59524 -0.0125 3 0 ]
[.78571 -0.0125 -3 -9 ]
[.78571 -0.0125 3 0 ]
[.97619 -0.0125 -6 -9 ]
[.97619 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .07725 -24 -4.5 ]
[-0.0125 .07725 0 4.5 ]
[-0.0125 .15451 -18 -4.5 ]
[-0.0125 .15451 0 4.5 ]
[-0.0125 .23176 -24 -4.5 ]
[-0.0125 .23176 0 4.5 ]
[-0.0125 .30902 -6 -4.5 ]
[-0.0125 .30902 0 4.5 ]
[-0.0125 .38627 -24 -4.5 ]
[-0.0125 .38627 0 4.5 ]
[-0.0125 .46353 -18 -4.5 ]
[-0.0125 .46353 0 4.5 ]
[-0.0125 .54078 -24 -4.5 ]
[-0.0125 .54078 0 4.5 ]
[-0.0125 .61803 -6 -4.5 ]
[-0.0125 .61803 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
[(0)] .02381 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(2)] .21429 -0.0125 0 1 Mshowa
.40476 0 m
.40476 .00625 L
s
[(4)] .40476 -0.0125 0 1 Mshowa
.59524 0 m
.59524 .00625 L
s
[(6)] .59524 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(8)] .78571 -0.0125 0 1 Mshowa
.97619 0 m
.97619 .00625 L
s
[(10)] .97619 -0.0125 0 1 Mshowa
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.5 0 m
.5 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .07725 m
.00625 .07725 L
s
[(0.25)] -0.0125 .07725 1 0 Mshowa
0 .15451 m
.00625 .15451 L
s
[(0.5)] -0.0125 .15451 1 0 Mshowa
0 .23176 m
.00625 .23176 L
s
[(0.75)] -0.0125 .23176 1 0 Mshowa
0 .30902 m
.00625 .30902 L
s
[(1)] -0.0125 .30902 1 0 Mshowa
0 .38627 m
.00625 .38627 L
s
[(1.25)] -0.0125 .38627 1 0 Mshowa
0 .46353 m
.00625 .46353 L
s
[(1.5)] -0.0125 .46353 1 0 Mshowa
0 .54078 m
.00625 .54078 L
s
[(1.75)] -0.0125 .54078 1 0 Mshowa
0 .61803 m
.00625 .61803 L
s
[(2)] -0.0125 .61803 1 0 Mshowa
.125 Mabswid
0 .01545 m
.00375 .01545 L
s
0 .0309 m
.00375 .0309 L
s
0 .04635 m
.00375 .04635 L
s
0 .0618 m
.00375 .0618 L
s
0 .09271 m
.00375 .09271 L
s
0 .10816 m
.00375 .10816 L
s
0 .12361 m
.00375 .12361 L
s
0 .13906 m
.00375 .13906 L
s
0 .16996 m
.00375 .16996 L
s
0 .18541 m
.00375 .18541 L
s
0 .20086 m
.00375 .20086 L
s
0 .21631 m
.00375 .21631 L
s
0 .24721 m
.00375 .24721 L
s
0 .26266 m
.00375 .26266 L
s
0 .27812 m
.00375 .27812 L
s
0 .29357 m
.00375 .29357 L
s
0 .32447 m
.00375 .32447 L
s
0 .33992 m
.00375 .33992 L
s
0 .35537 m
.00375 .35537 L
s
0 .37082 m
.00375 .37082 L
s
0 .40172 m
.00375 .40172 L
s
0 .41717 m
.00375 .41717 L
s
0 .43262 m
.00375 .43262 L
s
0 .44807 m
.00375 .44807 L
s
0 .47898 m
.00375 .47898 L
s
0 .49443 m
.00375 .49443 L
s
0 .50988 m
.00375 .50988 L
s
0 .52533 m
.00375 .52533 L
s
0 .55623 m
.00375 .55623 L
s
0 .57168 m
.00375 .57168 L
s
0 .58713 m
.00375 .58713 L
s
0 .60258 m
.00375 .60258 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
.02381 .61178 m
.02381 .61803 L
s
.21429 .61178 m
.21429 .61803 L
s
.40476 .61178 m
.40476 .61803 L
s
.59524 .61178 m
.59524 .61803 L
s
.78571 .61178 m
.78571 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .07725 m
1 .07725 L
s
.99375 .15451 m
1 .15451 L
s
.99375 .23176 m
1 .23176 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .38627 m
1 .38627 L
s
.99375 .46353 m
1 .46353 L
s
.99375 .54078 m
1 .54078 L
s
.125 Mabswid
.99625 .01545 m
1 .01545 L
s
.99625 .0309 m
1 .0309 L
s
.99625 .04635 m
1 .04635 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .09271 m
1 .09271 L
s
.99625 .10816 m
1 .10816 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .13906 m
1 .13906 L
s
.99625 .16996 m
1 .16996 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .20086 m
1 .20086 L
s
.99625 .21631 m
1 .21631 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26266 m
1 .26266 L
s
.99625 .27812 m
1 .27812 L
s
.99625 .29357 m
1 .29357 L
s
.99625 .32447 m
1 .32447 L
s
.99625 .33992 m
1 .33992 L
s
.99625 .35537 m
1 .35537 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .40172 m
1 .40172 L
s
.99625 .41717 m
1 .41717 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .44807 m
1 .44807 L
s
.99625 .47898 m
1 .47898 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .50988 m
1 .50988 L
s
.99625 .52533 m
1 .52533 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57168 m
1 .57168 L
s
.99625 .58713 m
1 .58713 L
s
.99625 .60258 m
1 .60258 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .14388 m
.02499 .14388 L
.02605 .14388 L
.02729 .14387 L
.02846 .14387 L
.03053 .14385 L
.03279 .14383 L
.0379 .14376 L
.04262 .14367 L
.04759 .14358 L
.05002 .14355 L
.05108 .14353 L
.05224 .14351 L
.05344 .10762 L
.05458 .10762 L
.05589 .10761 L
.0571 .1076 L
.0578 .1076 L
.05845 .1076 L
.05969 .1076 L
.06038 .1076 L
.06101 .1076 L
.06171 .10761 L
.06244 .10761 L
.06379 .10762 L
.06507 .10764 L
.06746 .10768 L
.0688 .10771 L
.07003 .10774 L
.07283 .10784 L
.07507 .10793 L
.07754 .10806 L
.08253 .1084 L
.088 .10891 L
.0939 .10964 L
.10458 .11144 L
.11415 .11359 L
.12462 .1165 L
.14572 .12375 L
.16697 .13215 L
.18628 .13994 L
.19592 .14371 L
.20099 .14564 L
.20637 .14764 L
.21106 .14934 L
.21235 .1498 L
.21306 .15005 L
.21373 .15028 L
.21504 .14292 L
.21623 .14267 L
Mistroke
.22533 .14068 L
.26357 .13137 L
.30427 .12091 L
.34345 .1112 L
.36346 .10652 L
.37398 .10415 L
.38508 .10172 L
.39506 .0996 L
.39779 .09903 L
.40072 .09842 L
.40214 .09813 L
.40348 .09785 L
.40468 .09761 L
.40599 .10683 L
.40718 .10697 L
.4083 .10709 L
.41084 .10738 L
.41598 .10796 L
.4252 .10896 L
.46394 .11278 L
.50514 .11624 L
.54481 .11909 L
.58297 .12144 L
.62359 .12361 L
.66268 .12541 L
.70423 .12707 L
.74427 .12846 L
.78279 .12964 L
.82376 .13074 L
.86321 .13167 L
.90115 .13247 L
.94154 .13323 L
.97619 .13382 L
Mfstroke
[ .05 .05 ] 0 setdash
.02381 .30902 m
.97619 .30902 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`008?oo0P00;ooo1000;ooo00<00?ooool0;Ooo0P00<?oo0P00:ooo10000ooo0P00
1?oo000Oool01000ooooo`00;_oo00<00?ooool0<?oo00<00?ooool0;?oo00@00?ooool002koo`04
003ooooo000/ool01@00oooooooo00000_oo00<00?ooool00Ooo000Oool01000ooooo`00;ooo00<0
0?ooool0;?oo1@00;Ooo00@00?ooool002koo`04003ooooo000/ool01@00oooooooo00000_oo00<0
0?ooool00Ooo000Oool01000ooooo`00<?oo00<00?ooool0:ooo00@00?ooool002koo`<0033oo`80
02goo`05003oooooool00002ool00`00ooooo`01ool001ooo`04003ooooo000aool00`00ooooo`0[
ool00`00ool0000^ool00`00ooooo`0_ool01000ooooo`00;?oo00D00?ooooooo`0000;oo`03003o
oooo007oo`007ooo00@00?ooool002koo`04003ooooo000^ool2000^ool00`00ooooo`0_ool01000
ooooo`00:_oo0`000ooo00@00?ooool000?oo`008?oo0P00<?oo0P00<?oo00<00?ooool0;Ooo0`00
;ooo0P00;Ooo00<00?ooool00_oo0P001?oo003ooolQool00?ooob7oo`006oooo`001@000Ooo000K
ool00`00ooooo`03ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`0:
ool00`00ooooo`09ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`09
ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09
ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`09ool00`00ooooo`0:
ool00`00ooooo`09ool00`00ooooo`03ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`00
6ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03003ooooo0?oo
o`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool10001
ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00oooo
o`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo0P00oooo0Ooo0P000Ooo
000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo00<0
0?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03003ooooo0?ooo`7oo`40
007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo0002ool20004ool00`00ooooo`02ool40002ool3
0005ool00`00ooooo`3oool1ool10001ool00005ool00?ooool00008ool00`00ooooo`06ool00`00
ooooo`02ool3003oool20001ool00005ool00?ooool00009ool00`00ooooo`05ool00`00ooooo`02
ool00`00ooooo`3oool1ool10001ool00005ool00?ooool0000:ool01000oooooooo0`001Ooo00<0
0?ooool0oooo0Ooo0@000Ooo00001Ooo003ooooo00002ooo00@00?ooool000Ooo`03003ooooo0?oo
o`7oo`40007oo`0000Goo`00ooooo`0000Soo`04003ooooo0002ool00`00ooooo`05ool2003oool1
ool20001ool000;oo`8000[oo`8000?oo`@000Coo`03003ooooo06Coo`<009Woo`40007oo`006ooo
00<00?ooool0H?oo10000_oo00<00?ooool0Uooo0@000Ooo000Kool00`00ooooo`1Jool60006ool0
0`00ooooo`2Gool10001ool001_oo`8000coo``0047oo`8000goo`T008ooo`80007oo`006ooo00<0
0?ooool02ooo00<00?ooool02Ooo1@00>?oo10006?oo2`00QOoo0@000Ooo000Kool00`00ooooo`0;
ool00`00ooooo`0>ool3000`ool5000Wool:001kool10001ool001_oo`03003ooooo00_oo`03003o
oooo017oo`<002Woo`@003KooaD006Koo`40007oo`006ooo0P003?oo00<00?ooool05?oo0`008_oo
1000Cooo5000DOoo0P000Ooo000Kool00`00ooooo`0:ool00`00ooooo`0Hool2000Lool4001WoolO
000cool10001ool001_oo`03003ooooo00[oo`03003ooooo01[oo`<001Goo`@008[oobh000Goo`40
007oo`006ooo00<00?ooool02_oo00<00?ooool07Ooo0P003_oo1@00`Ooo0@000Ooo000Kool2000;
ool00`00ooooo`0Oool30007ool40035ool20001ool001_oo`03003ooooo00?oo`P002Coo`8000Co
o`03003ooooo0<Soo`40007oo`006ooo00<00?ooool0<Ooo0P000_oo00<00?ooool0b?oo0@000Ooo
0008ool20004ool00`00ooooo`02ool30005ool00`00ooooo`0cool2003;ool10001ool000Ooo`04
003ooooo000;ool00`00ooooo`02ool3003oool20001ool000Ooo`04003ooooo000;ool00`00oooo
o`02ool00`00ooooo`3oool1ool10001ool000Ooo`04003ooooo0008ool30005ool00`00ooooo`3o
ool1ool10001ool000Ooo`04003ooooo0008ool00`00ooooo`05ool00`00ooooo`3oool1ool10001
ool000Ooo`04003ooooo0008ool00`00ooooo`05ool2003oool1ool20001ool000Soo`8000Woo`@0
00Coo`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00
ooooo`3oool1ool10001ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0oooo0Ooo0@00
0Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo
0P00oooo0Ooo0P000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7o
o`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03
003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo0002ool20004ool00`00
ooooo`03ool00`00ooooo`02ool30005ool00`00ooooo`3oool1ool10001ool00005ool00?ooool0
0009ool00`00ooooo`05ool00`00ooooo`02ool3003oool20001ool00005ool00?ooool0000:ool0
0`00ooooo`04ool00`00ooooo`02ool00`00ooooo`3oool1ool10001ool00005ool00?ooool0000:
ool01000oooooooo0`001Ooo00<00?ooool0oooo0Ooo0@000Ooo00001Ooo003ooooo00002ooo00@0
0?ooool000Ooo`03003ooooo0?ooo`7oo`40007oo`0000Goo`00ooooo`0000Soo`04003ooooo0002
ool00`00ooooo`05ool2003oool1ool20001ool000;oo`8000Woo`@000;oo`@000Coo`03003ooooo
0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool1
0001ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00
ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo0P00oooo0Ooo0P00
0Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo
00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03003ooooo0?ooo`7o
o`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Cool40004ool00`00ooooo`3oool1ool1
0001ool001Goo`03003ooooo00?oo`<000?oo`h000coo`h000coo`h000coo`h000coo`h000coo`h0
00coo`h000coo`h000_oo`h000coo`h000Goo`80007oo`005Ooo00<00?ooool00ooo00<00?ooool0
oooo0Ooo0@000Ooo000Eool00`00ooooo`03ool00`00ooooo`3oool1ool10001ool001Goo`03003o
oooo00?oo`03003ooooo0?ooo`7oo`40007oo`004ooo0`001Ooo0P00oooo0Ooo0P000Ooo000Eool0
0`00ooooo`03ool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`00
6ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03003ooooo0?oo
o`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool10001
ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00oooo
o`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo0P00oooo0Ooo0P000Ooo
000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`0000?oo`00
00000P000ooo00<00?ooool00_oo10000_oo0`001Ooo00<00?ooool0oooo0Ooo0@000Ooo0003ool0
0`00ooooo`07ool00`00ooooo`06ool00`00ooooo`02ool3003oool20001ool000?oo`03003ooooo
00Soo`03003ooooo00Goo`03003ooooo00;oo`03003ooooo0?ooo`7oo`40007oo`000ooo00<00?oo
ool02Ooo00@00?ooooooo`<000Goo`03003ooooo0?ooo`7oo`40007oo`000ooo00<00?ooool02_oo
00@00?ooool000Ooo`03003ooooo0?ooo`7oo`40007oo`0000Coo`00000000Woo`04003ooooo0002
ool00`00ooooo`05ool2003oool1ool20001ool000?oo`03003ooooo00Soo`8000?oo`@000Coo`03
003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3o
ool1ool10001ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000K
ool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo0P00oooo
0Ooo0P000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40007o
o`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1ool20001ool001_oo`03003ooooo
0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo0007ool40003ool00`00ooooo`02
ool30005ool00`00ooooo`3oool1ool10001ool000Woo`03003ooooo00[oo`03003ooooo00;oo`<0
0?ooo`80007oo`002Ooo00<00?ooool02_oo00<00?ooool00_oo00<00?ooool0oooo0Ooo0@000Ooo
0009ool00`00ooooo`07ool30005ool00`00ooooo`3oool1ool10001ool000Woo`03003ooooo00Oo
o`03003ooooo00Goo`03003ooooo0?ooo`7oo`40007oo`001ooo0`002Ooo00<00?ooool01Ooo0P00
oooo0Ooo0P000Ooo0009ool00`00ooooo`07ool40004ool00`00ooooo`3oool1ool10001ool001_o
o`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool2003oool1
ool20001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo
000Kool00`00ooooo`3oool1ool10001ool001_oo`800?ooo`7oo`80007oo`006ooo00<00?ooool0
oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003ooooo0?ooo`7oo`40
007oo`006ooo0P00oooo0Ooo0P000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03003o
oooo0?ooo`7oo`40007oo`0000?oo`0000000P000ooo00<00?ooool00ooo00<00?ooool00_oo0`00
1Ooo00<00?ooool0oooo0Ooo0@000Ooo0003ool00`00ooooo`08ool00`00ooooo`05ool00`00oooo
o`02ool3003oool20001ool000?oo`03003ooooo00Woo`03003ooooo00Coo`03003ooooo00;oo`03
003ooooo0?ooo`7oo`40007oo`000ooo00<00?ooool02Ooo00@00?ooooooo`<000Goo`03003ooooo
0?ooo`7oo`40007oo`000ooo00<00?ooool02_oo00@00?ooool000Ooo`03003ooooo0?ooo`7oo`40
007oo`0000Coo`00000000Woo`04003ooooo0002ool00`00ooooo`05ool2003oool1ool20001ool0
00?oo`03003ooooo00Ooo`@000;oo`@000Coo`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?oo
ool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`800?ooo`7oo`80007o
o`006ooo00<00?ooool0oooo0Ooo0@000Ooo000Kool00`00ooooo`3oool1ool10001ool001_oo`03
003ooooo0?ooo`7oo`40007oo`006ooo0P00oooo0Ooo0P000Ooo000Kool00`00ooooo`3oool1ool1
0001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool0oooo0Ooo0@000Ooo000K
ool2003oool1ool20001ool001_oo`03003ooooo0?ooo`7oo`40007oo`006ooo00<00?ooool00ooo
00<00?ooool0;ooo00<00?ooool0;_oo00<00?ooool0;_oo00<00?ooool0;ooo00<00?ooool0;_oo
00<00?ooool00ooo0@000Ooo000Cool40004ool00`00ooooo`03ool00`00ooooo`09ool00`00oooo
o`0:ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`09ool00`00oooo
o`0:ool00`00ooooo`09ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00oooo
o`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`09ool00`00ooooo`0:ool00`00oooo
o`09ool00`00ooooo`09ool00`00ooooo`0:ool00`00ooooo`09ool00`00ooooo`03ool10001ool0
01?oo`03003ooooo00Goool000D0007oo`005?oo00<00?ooool0oooo2_oo000Eool00`00ooooo`3o
ool9ool001Koo`03003ooooo0?ooo`Soo`004ooo00@00?ooool00?ooo`[oo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-1.36219, -0.152823, 
  0.0405086, 0.0124846}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Bode Plots of closed loop transfer function", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
T2disc = Discretize[Trat2[s]];
bm = BodeMagnitude[T2disc,PlotLabel ->\"BodeMagnitudePlot\"];
bp = BodePhase[T2disc,PlotLabel -> \"Phase Bode Plot\"];
\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.25 1.49012 0.0952624 [
[.25 -0.0125 -11 -9.8125 ]
[.25 -0.0125 11 0 ]
[.5 -0.0125 -5 -9.8125 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -8 -9.8125 ]
[.75 -0.0125 8 0 ]
[1 -0.0125 -11 -9.8125 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -19.875 ]
[.5 -0.0125 29 -9.8125 ]
[-0.0125 .06118 -18 -4.5 ]
[-0.0125 .06118 0 4.5 ]
[-0.0125 .15645 -18 -4.5 ]
[-0.0125 .15645 0 4.5 ]
[-0.0125 .25171 -18 -4.5 ]
[-0.0125 .25171 0 4.5 ]
[-0.0125 .34697 -18 -4.5 ]
[-0.0125 .34697 0 4.5 ]
[-0.0125 .44223 -18 -4.5 ]
[-0.0125 .44223 0 4.5 ]
[-0.0125 .5375 -18 -4.5 ]
[-0.0125 .5375 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34 -4.90625 ]
[-0.01875 .30902 -18 4.90625 ]
[0 .63053 -14 0 ]
[0 .63053 14 9.8125 ]
[.25 .63053 -11 0 ]
[.25 .63053 11 9.8125 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 9.8125 ]
[.75 .63053 -8 0 ]
[.75 .63053 8 9.8125 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -53 9.8125 ]
[.5 .63053 53 19.9375 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.25 0 m
.25 .61803 L
s
.5 0 m
.5 .61803 L
s
.75 0 m
.75 .61803 L
s
1 0 m
1 .61803 L
s
0 .06118 m
1 .06118 L
s
0 .15645 m
1 .15645 L
s
0 .25171 m
1 .25171 L
s
0 .34697 m
1 .34697 L
s
0 .44223 m
1 .44223 L
s
0 .5375 m
1 .5375 L
s
0 g
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -72 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(100) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.001 w
.07526 0 m
.07526 .00375 L
s
.11928 0 m
.11928 .00375 L
s
.15051 0 m
.15051 .00375 L
s
.17474 0 m
.17474 .00375 L
s
.19454 0 m
.19454 .00375 L
s
.21127 0 m
.21127 .00375 L
s
.22577 0 m
.22577 .00375 L
s
.23856 0 m
.23856 .00375 L
s
.32526 0 m
.32526 .00375 L
s
.36928 0 m
.36928 .00375 L
s
.40051 0 m
.40051 .00375 L
s
.42474 0 m
.42474 .00375 L
s
.44454 0 m
.44454 .00375 L
s
.46127 0 m
.46127 .00375 L
s
.47577 0 m
.47577 .00375 L
s
.48856 0 m
.48856 .00375 L
s
.57526 0 m
.57526 .00375 L
s
.61928 0 m
.61928 .00375 L
s
.65051 0 m
.65051 .00375 L
s
.67474 0 m
.67474 .00375 L
s
.69454 0 m
.69454 .00375 L
s
.71127 0 m
.71127 .00375 L
s
.72577 0 m
.72577 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.82526 0 m
.82526 .00375 L
s
.86928 0 m
.86928 .00375 L
s
.90051 0 m
.90051 .00375 L
s
.92474 0 m
.92474 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.96127 0 m
.96127 .00375 L
s
.97577 0 m
.97577 .00375 L
s
.98856 0 m
.98856 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(H) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 11.250000 moveto
(\\220) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 11.250000 moveto
(L) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 .06118 m
.00625 .06118 L
s
[(-15)] -0.0125 .06118 1 0 Mshowa
0 .15645 m
.00625 .15645 L
s
[(-14)] -0.0125 .15645 1 0 Mshowa
0 .25171 m
.00625 .25171 L
s
[(-13)] -0.0125 .25171 1 0 Mshowa
0 .34697 m
.00625 .34697 L
s
[(-12)] -0.0125 .34697 1 0 Mshowa
0 .44223 m
.00625 .44223 L
s
[(-11)] -0.0125 .44223 1 0 Mshowa
0 .5375 m
.00625 .5375 L
s
[(-10)] -0.0125 .5375 1 0 Mshowa
.125 Mabswid
0 .08024 m
.00375 .08024 L
s
0 .09929 m
.00375 .09929 L
s
0 .11834 m
.00375 .11834 L
s
0 .13739 m
.00375 .13739 L
s
0 .1755 m
.00375 .1755 L
s
0 .19455 m
.00375 .19455 L
s
0 .2136 m
.00375 .2136 L
s
0 .23266 m
.00375 .23266 L
s
0 .27076 m
.00375 .27076 L
s
0 .28981 m
.00375 .28981 L
s
0 .30887 m
.00375 .30887 L
s
0 .32792 m
.00375 .32792 L
s
0 .36602 m
.00375 .36602 L
s
0 .38508 m
.00375 .38508 L
s
0 .40413 m
.00375 .40413 L
s
0 .42318 m
.00375 .42318 L
s
0 .46129 m
.00375 .46129 L
s
0 .48034 m
.00375 .48034 L
s
0 .49939 m
.00375 .49939 L
s
0 .51844 m
.00375 .51844 L
s
0 .04213 m
.00375 .04213 L
s
0 .02308 m
.00375 .02308 L
s
0 .00403 m
.00375 .00403 L
s
0 .55655 m
.00375 .55655 L
s
0 .5756 m
.00375 .5756 L
s
0 .59465 m
.00375 .59465 L
s
0 .61371 m
.00375 .61371 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(dB) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 .61178 m
.25 .61803 L
s
gsave
.25 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.75 .61178 m
.75 .61803 L
s
gsave
.75 .63053 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.001 w
.07526 .61428 m
.07526 .61803 L
s
.11928 .61428 m
.11928 .61803 L
s
.15051 .61428 m
.15051 .61803 L
s
.17474 .61428 m
.17474 .61803 L
s
.19454 .61428 m
.19454 .61803 L
s
.21127 .61428 m
.21127 .61803 L
s
.22577 .61428 m
.22577 .61803 L
s
.23856 .61428 m
.23856 .61803 L
s
.32526 .61428 m
.32526 .61803 L
s
.36928 .61428 m
.36928 .61803 L
s
.40051 .61428 m
.40051 .61803 L
s
.42474 .61428 m
.42474 .61803 L
s
.44454 .61428 m
.44454 .61803 L
s
.46127 .61428 m
.46127 .61803 L
s
.47577 .61428 m
.47577 .61803 L
s
.48856 .61428 m
.48856 .61803 L
s
.57526 .61428 m
.57526 .61803 L
s
.61928 .61428 m
.61928 .61803 L
s
.65051 .61428 m
.65051 .61803 L
s
.67474 .61428 m
.67474 .61803 L
s
.69454 .61428 m
.69454 .61803 L
s
.71127 .61428 m
.71127 .61803 L
s
.72577 .61428 m
.72577 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.82526 .61428 m
.82526 .61803 L
s
.86928 .61428 m
.86928 .61803 L
s
.90051 .61428 m
.90051 .61803 L
s
.92474 .61428 m
.92474 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.96127 .61428 m
.96127 .61803 L
s
.97577 .61428 m
.97577 .61803 L
s
.98856 .61428 m
.98856 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .06118 m
1 .06118 L
s
.99375 .15645 m
1 .15645 L
s
.99375 .25171 m
1 .25171 L
s
.99375 .34697 m
1 .34697 L
s
.99375 .44223 m
1 .44223 L
s
.99375 .5375 m
1 .5375 L
s
.125 Mabswid
.99625 .08024 m
1 .08024 L
s
.99625 .09929 m
1 .09929 L
s
.99625 .11834 m
1 .11834 L
s
.99625 .13739 m
1 .13739 L
s
.99625 .1755 m
1 .1755 L
s
.99625 .19455 m
1 .19455 L
s
.99625 .2136 m
1 .2136 L
s
.99625 .23266 m
1 .23266 L
s
.99625 .27076 m
1 .27076 L
s
.99625 .28981 m
1 .28981 L
s
.99625 .30887 m
1 .30887 L
s
.99625 .32792 m
1 .32792 L
s
.99625 .36602 m
1 .36602 L
s
.99625 .38508 m
1 .38508 L
s
.99625 .40413 m
1 .40413 L
s
.99625 .42318 m
1 .42318 L
s
.99625 .46129 m
1 .46129 L
s
.99625 .48034 m
1 .48034 L
s
.99625 .49939 m
1 .49939 L
s
.99625 .51844 m
1 .51844 L
s
.99625 .04213 m
1 .04213 L
s
.99625 .02308 m
1 .02308 L
s
.99625 .00403 m
1 .00403 L
s
.99625 .55655 m
1 .55655 L
s
.99625 .5756 m
1 .5756 L
s
.99625 .59465 m
1 .59465 L
s
.99625 .61371 m
1 .61371 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -114 5.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(BodeMagnitudePlot) show
165.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.09751 .49839 Mdot
.17283 .49876 Mdot
.21696 .49938 Mdot
.24835 .50024 Mdot
.27277 .50134 Mdot
.29281 .50269 Mdot
.30984 .50427 Mdot
.32467 .50609 Mdot
.33783 .50814 Mdot
.3497 .51042 Mdot
.36052 .51291 Mdot
.37048 .51562 Mdot
.37975 .51854 Mdot
.38841 .52166 Mdot
.39657 .52497 Mdot
.40431 .52846 Mdot
.41167 .53213 Mdot
.41871 .53595 Mdot
.42547 .53991 Mdot
.43199 .544 Mdot
.4383 .5482 Mdot
.44443 .55249 Mdot
.45039 .55685 Mdot
.45622 .56126 Mdot
.46194 .56567 Mdot
.46755 .57007 Mdot
.47308 .57442 Mdot
.47854 .57868 Mdot
.48395 .58281 Mdot
.48932 .58674 Mdot
.49467 .59044 Mdot
.5 .59383 Mdot
.50533 .59684 Mdot
.51068 .59939 Mdot
.51605 .6014 Mdot
.52146 .60274 Mdot
.52692 .60332 Mdot
.53245 .60299 Mdot
.53806 .6016 Mdot
.54378 .59898 Mdot
.54961 .59493 Mdot
.55557 .58922 Mdot
.5617 .5816 Mdot
.56801 .57177 Mdot
.57453 .55939 Mdot
.58129 .54408 Mdot
.58833 .52537 Mdot
.59569 .50276 Mdot
.60343 .47562 Mdot
.61159 .44327 Mdot
.62025 .40486 Mdot
.62952 .35942 Mdot
.63948 .30579 Mdot
.6503 .24256 Mdot
.66217 .16801 Mdot
.67533 .07997 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{336, 207.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005@0000ca000`40n007h00OooooDOoo003ooomA
ool00?oooe7oo`00ooooDOoo002Zool00`00ooooo`0Nool00`00ooooo`22ool00:Woo`03003ooooo
01?oo`03003ooooo00[oo`03003ooooo087oo`00Vooo00<00?oo00002ooo00@00?ooooooo`@000?o
o`800004ool00?ooool200001Ooo003ooooo00001?oo0`001?oo00<00?ooool0POoo002Kool00`00
ool0000;ool00`00ooooo`02ool00`00ooooo`02ool01000ooooo`000_oo00@00?ooool000?oo`03
003ooooo00Goo`05003oooooool00023ool009[oo`05003oo`00ool0000:ool00`00ooooo`02ool0
0`00ooooo`03ool30002ool01000ooooo`000ooo0P001?oo0P001?oo00<00?ooool0POoo002Jool0
1@00ool00?oo00002_oo00<00?ooool00_oo00<00?ooool01Ooo00@00?ooool000;oo`03003ooooo
00;oo`05003oooooool00006ool00`00ooooo`21ool009[oo`04003oo`00ool20009ool01P00oooo
oooo003oo`<000;oo`8000Coo`<000Coo`03003ooooo00;oo`<000?oo`03003ooooo087oo`00Z_oo
00<00?ooool03ooo00<00?ooool00ooo00<00?ooool01_oo00<00?ooool0P_oo002kool2002Cool0
0?oooe7oo`00ooooDOoo003ooomAool006goo`8000Coo`03003ooooo00;oo`@003Ooo`@003[oo`@0
00?oo`8003Goo`@000?oo`8000Coo`8001;oo`00K?oo00@00?ooool000[oo`03003ooooo03Soo`03
003ooooo03_oo`05003oooooool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?oo
ool000;oo`03003ooooo00ooo`00K?oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_o
o`05003oooooool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03
003ooooo00ooo`00K?oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_oo`05003ooooo
ool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo00oo
o`00K?oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_oo`05003oooooool00002ool0
0`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo00ooo`00K?oo00@0
0?ooool000Soo`<003Soo`<003_oo`<000?oo`04003ooooo000dool30003ool01000ooooo`000_oo
00@00?ooool0017oo`00KOoo0P002ooo00<00?ooool0>?oo00<00?ooool0>ooo00<00?ooool00_oo
0P00=ooo00<00?ooool00_oo0P001?oo0P004_oo000coooo0006000Hool003?oo`8001;oo`03003o
oooo00Soo`03003ooooo00Goo`03003ooooo00Coo`03003ooooo00;oo`05003oooooool00003ool0
1000ooooo`000_oo00<00?ooool04Ooo00<00?ooool02?oo00<00?ooool01Ooo00<00?ooool00ooo
00<00?ooool00ooo00D00?ooooooo`0000?oo`04003ooooo0002ool00`00ooooo`0@ool00`00oooo
o`09ool00`00ooooo`05ool00`00ooooo`03ool00`00ooooo`02ool00`00ooooo`02ool01000oooo
o`000ooo00@00?ooool001;oo`03003ooooo00Woo`03003ooooo00Goo`03003ooooo00?oo`03003o
oooo00;oo`03003ooooo00;oo`04003ooooo0003ool01000ool000006?oo000cool00`00ooooo`0n
ool00`00ooooo`0nool00`00ooooo`0nool00`00ooooo`0nool00`00ooooo`0Fool003?oo`03003o
oooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo
00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo
000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00oooo
o`0Fool003?oo`8003ooo`0300oooooo03koo`03013ooooo03koo`0300oooooo03goo`8001Soo`00
<ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool0
5_oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00
ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03ko
o`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03ooo
ool0?_oo00<00?ooool05_oo000cool2000oool00`0@ooooo`0nool00`0?ooooo`0nool00`0@oooo
o`0mool2000Hool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo
03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<0
4?ooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0nool00`00ooooo`0Fool002Goo`@000;oo`<000Goo`03003ooooo03koo`03013o
oooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`009ooo00<00?ooool01?oo
00<00?ooool00_oo0`000?l03`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l00`006?oo000Wool0
0`00ooooo`04ool00`00ooooo`02ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool0
0`0@ooooo`0nool00`00ooooo`0Fool001ooo`@000Coo`04003oooooool30005ool00`00ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool002Ooo`05003o
ooooool00007ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool0
0`00ooooo`0Fool002Goo`<000?oo`03003ooooo00Goo`03003ooooo03koo`0300oooooo03koo`03
013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`009ooo00@00?ooooooo`@000Coo`8003oo
o`03013ooooo03koo`0300oooooo02[oo`8001;oo`03013ooooo03goo`8001Soo`00<ooo00<00?oo
ool0?_oo00<03oooool0?_oo00<04?ooool0:ooo00<00?ooool04?oo00<03oooool0?_oo00<00?oo
ool05_oo000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool0
0`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo
03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<0
4?ooool0?_oo00<00?ooool05_oo000cool2000oool00`0?ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0mool2000Hool003?oo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013o
oooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo
00<03oooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?oooo
o`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo03koo`03
013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo00<04?ooool0?_oo00<0
3oooool0?_oo00<04?ooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@
ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`8003ooo`03
00oooooo03koo`03013ooooo03koo`0300oooooo03goo`8001Soo`00<ooo00<00?ooool0?_oo00<0
4?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool002Goo`@000Co
o`03003ooooo00?oo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03ko
o`03003ooooo01Koo`009ooo00<00?ooool00ooo00<00?ooool00ooo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000Wool00`00ooooo`050004
ool30000o`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l04003000Hool001ooo`@000Coo`04003o
oooo0002ool00`00ooooo`03ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0nool00`00ooooo`0Fool002Ooo`07003oooooool00?oo00001Ooo00<00?ooool0?_oo00<0
4?ooool0?_oo00<03oooool09_oo0P005_oo00<04?ooool0?_oo00<00?ooool05_oo000Uool30004
ool20005ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0Wool00`00ooooo`0Dool00`0?
ooooo`0nool00`00ooooo`0Fool002Ooo`03003ooooo00?oo`03003ooooo00?oo`03003ooooo03ko
o`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo
00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?Ooo0P006?oo000cool00`00ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo
03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`00<ooo00<0
0?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000c
ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0F
ool003?oo`8003ooo`03013ooooo03koo`0300oooooo03koo`03013ooooo03goo`8001Soo`00<ooo
00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo
000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00oooo
o`0Fool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03
003ooooo01Koo`00<ooo0P00?ooo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?Ooo0P00
6?oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00
ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03ko
o`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03ooo
ool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool0
0`0@ooooo`0nool00`00ooooo`0Fool003?oo`8003ooo`0300oooooo03koo`03013ooooo03koo`03
00oooooo03goo`8001Soo`00<ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<0
4?ooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03koo`0300oo
oooo02?oo`8001Woo`03013ooooo03koo`03003ooooo01Koo`009Ooo10000ooo0P001Ooo00<00?oo
ool0?_oo00<03oooool0?_oo00<04?ooool09?oo00<00?ooool05ooo00<03oooool0?_oo00<00?oo
ool05_oo000Wool01@00oooooooo00000_oo00<00?ooool00_oo0`000?l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01000`006?oo000Wool00`00ooooo`04ool00`00ooooo`02ool00`00ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool001ooo`@000Co
o`03003ooooo00;oo`8000Goo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013o
oooo03koo`03003ooooo01Koo`009ooo00<00?ooool01?oo00<00?ooool00_oo00<00?ooool0?_oo
00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000Uool30003ool0
1000ooooo`001?oo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo
00<00?ooool05_oo000Wool00`00ooooo`02ool20005ool2000oool00`0?ooooo`0nool00`0@oooo
o`0nool00`0?ooooo`0mool2000Hool003?oo`03003ooooo03koo`03013ooooo03koo`0300oooooo
03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<03oooool0?_oo00<0
4?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oo
oooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo00<04?oo
ool0?_oo00<03oooool0?_oo00<04?ooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0?oooo
o`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03
013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0
?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000cool00`00
ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0Pool2000Lool00`0@ooooo`0nool00`00ooooo`0F
ool003?oo`8003ooo`0300oooooo03koo`03013ooooo027oo`03003ooooo01[oo`0300oooooo03go
o`8001Soo`004Ooo0P0000?oo`0000000`006_oo00<00?ooool0?_oo00<04?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000@ool01000ooooo`000_oo00@00?ooool001Wo
o`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo01Ko
o`004?oo00@00?ooool000;oo`04003ooooo000Iool00`00ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool0013oo`04003ooooo0002ool3000Jool00`00
ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool0017o
o`<000;oo`04003ooooo000Iool2000oool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0m
ool2000Hool001?oo`03003oo`0000<001[oo`03003ooooo03koo`0300oooooo03koo`03013ooooo
03koo`0300oooooo03koo`03003ooooo01Koo`004_oo0P007ooo00<00?ooool0?_oo00<04?ooool0
?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0?
ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool002Goo`@000;oo`@000Co
o`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Ko
o`009ooo00D00?ooooooo`0000Ooo`<0003o00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?00<0
01Soo`009ooo00<00?ooool00_oo00<00?ooool01?oo00<00?ooool0?_oo00<04?ooool0?_oo00<0
3oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000Oool40004ool00`00ooooo`03ool00`00
ooooo`03ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00
ooooo`0Fool002Ooo`03003ooooo00Coo`03003ooooo00;oo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo01koo`8001koo`03013ooooo03koo`03003ooooo01Koo`009Ooo0`000ooo00@00?oo
ool000Coo`03003ooooo03koo`0300oooooo03koo`03013ooooo01ooo`03003ooooo01coo`0300oo
oooo03koo`03003ooooo01Koo`009ooo00<00?ooool00_oo0P001Ooo0P00?ooo00<04?ooool0?_oo
00<03oooool0?_oo00<04?ooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0?ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo
03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<0
3oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0n
ool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`8003oo
o`0300oooooo03koo`03013ooooo03koo`0300oooooo03goo`8001Soo`00<ooo00<00?ooool0?_oo
00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000cool00`00oooo
o`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03
003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00
<ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool0
5_oo000cool2000oool00`0@ooooo`0nool00`0?ooooo`0Kool2000Qool00`0@ooooo`0mool2000H
ool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo01coo`03003ooooo01ooo`0300oo
oooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo
00<04?ooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0?ooooo`0nool00`0@oooo
o`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03koo`03
00oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo00<03oooool0?_oo00<0
4?ooool0?_oo00<03oooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo03ko
o`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`009Ooo10000_oo10001?oo00<00?oo
ool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000Wool0
0`00ooooo`03ool00`00ooooo`03ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool0
0`0?ooooo`0nool00`00ooooo`0Fool002Ooo`03003ooooo00?oo`03003ooooo00?oo`<0002K0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l040020000HP0?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0
400?01003`0@00l0400?01003`0@00l0400?0103000Hool001ooo`@000Coo`03003ooooo00?oo`03
003ooooo00?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo01[oo`03003ooooo027oo`03
00oooooo03koo`03003ooooo01Koo`009ooo00<00?ooool00ooo00<00?ooool00ooo00<00?ooool0
?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000Uool30003
ool30005ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00
ooooo`0Fool002Ooo`03003ooooo00?oo`03003ooooo00?oo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo00<03oooool0?_oo
00<04?ooool0?_oo00<03oooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0@ooooo`0nool0
0`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo
03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<0
4?ooool0?_oo00<03oooool05ooo0P009Ooo00<04?ooool0?_oo00<00?ooool05_oo000cool00`00
ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0Hool00`00ooooo`0Sool00`0?ooooo`0nool00`00
ooooo`0Fool003?oo`8003ooo`03013ooooo03koo`0300oooooo03koo`03013ooooo03goo`8001So
o`00<ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?oo
ool05_oo000cool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool0
0`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo
03koo`03003ooooo01Koo`00<ooo00<00?ooool05_oo0P004Ooo0P002_oo0P001ooo00<04?ooool0
?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool05_oo000cool2000Hool00`00ooooo`0@
ool00`00ooooo`09ool00`00ooooo`04ool20004ool20003ool2000eool00`0@ooooo`0Eool2000W
ool00`0?ooooo`0mool2000Hool003?oo`03003ooooo03koo`03003ooooo00?oo`03003ooooo00;o
o`04003oooooool20002ool2000/ool00`0?ooooo`0Fool00`00ooooo`0Uool00`0@ooooo`0nool0
0`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo00goo`06003oooooool00?oo0P0000?o
o`0000009_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool05_oo000cool00`00ooooo`0n
ool00`0@ooooo`0Dool01@00ooooo`00ool00P008ooo00<03oooool0?_oo00<04?ooool0?_oo00<0
0?ooool05_oo000cool00`00ooooo`0nool00`0?ooooo`0Jool00`00ool00003000Nool00`0@oooo
o`0nool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`8003ooo`03013ooooo01goo`03003oo`00
008001coo`0300oooooo03koo`03013ooooo03goo`8001Soo`00<ooo00<00?ooool0?_oo00<03ooo
ool08Ooo0`006_oo00<04?ooool04ooo0P00:Ooo00<03oooool0?_oo00<00?ooool05_oo000cool0
0`00ooooo`0nool00`0@ooooo`0Sool3000Hool00`0?ooooo`0Dool00`00ooooo`0Wool00`0@oooo
o`0nool00`00ooooo`0Fool002Goo`@000?oo`8000Goo`03003ooooo03koo`0300oooooo02Goo`<0
01Koo`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`009ooo00D00?ooooooo`0000;o
o`03003ooooo00;oo`<0001X01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`<0002D00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0103000Hool002Ooo`05
003oooooool00002ool00`00ooooo`02ool00`00ooooo`0nool00`0?ooooo`0Yool3000Bool00`0@
ooooo`0nool00`0?ooooo`0nool00`00ooooo`0Fool001ooo`@000Coo`05003oooooool00002ool0
0`00ooooo`02ool00`00ooooo`0nool00`0@ooooo`0[ool3000@ool00`0?ooooo`0Aool2000[ool0
0`0@ooooo`0nool00`00ooooo`0Fool002Ooo`05003oooooool00002ool00`00ooooo`02ool00`00
ooooo`0nool00`0?ooooo`0]ool2000?ool00`0@ooooo`0Bool00`00ooooo`0Yool00`0?ooooo`0n
ool00`00ooooo`0Fool002Goo`<000?oo`04003ooooo0004ool00`00ooooo`0nool00`0@ooooo`0^
ool3000=ool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0Fool002Ooo`03003ooooo00;o
o`8000Goo`8003ooo`0300oooooo033oo`8000coo`03013ooooo03koo`0300oooooo03goo`8001So
o`00<ooo00<00?ooool0?_oo00<04?ooool0<Ooo0`002_oo00<03oooool04?oo0P00;?oo00<04?oo
ool0?_oo00<00?ooool05_oo000cool00`00ooooo`0nool00`0?ooooo`0cool20009ool00`0@oooo
o`0Aool00`00ooooo`0Zool00`0?ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03
013ooooo03Coo`03003ooooo00Ooo`0300oooooo03koo`03013ooooo03koo`03003ooooo01Koo`00
<ooo00<00?ooool0?_oo00<03oooool0=Ooo0P001ooo00<04?ooool03_oo0P00;_oo00<03oooool0
?_oo00<00?ooool05_oo000cool2000oool00`0@ooooo`0fool20006ool00`0?ooooo`0?ool00`00
ooooo`0/ool00`0@ooooo`0mool2000Hool003?oo`03003ooooo03koo`0300oooooo03Ooo`<000Co
o`03013ooooo03koo`0300oooooo03koo`03003ooooo01Koo`00<ooo00<00?ooool0?_oo00<04?oo
ool0>Ooo0P000ooo00<03oooool03?oo0P00<?oo00<04?ooool0?_oo00<00?ooool05_oo000cool0
0`00ooooo`0nool00`0?ooooo`0jool20002ool00`0@ooooo`0=ool00`00ooooo`0^ool00`0?oooo
o`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`03013ooooo03_oo`<0000300oooooo00_o
o`80037oo`03013ooooo03koo`03003ooooo01Koo`00<ooo0P00?ooo00<03oooool0?Ooo0P002ooo
0P0000?oo`00ool0<?oo00<03oooool0?Ooo0P006?oo000cool00`00ooooo`0nool00`0@ooooo`0n
ool40007ool3000cool00`0@ooooo`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oo
oooo03koo`03013oo`0000P00003ool00?oo03?oo`0300oooooo03koo`03003ooooo01Koo`00<ooo
00<00?ooool0?_oo00<04?ooool0?_oo00H03oooooooo`00ool200000ooo0000000fool00`0@oooo
o`0nool00`00ooooo`0Fool003?oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`03
00oooooo03koo`03003ooooo01Koo`00<ooo0P004_oo00<00?ooool02?oo00<00?ooool01Ooo00<0
0?ooool01?oo00<00?ooool00_oo00D00?ooooooo`0000?oo`04003ooooo0002ool00`00ooooo`0A
ool00`00ooooo`08ool00`00ooooo`05ool00`00ooooo`03ool00`00ooooo`03ool01@00oooooooo
00000ooo00@00?ooool000;oo`03003ooooo013oo`03003ooooo00Woo`03003ooooo00Goo`03003o
oooo00?oo`03003ooooo00;oo`03003ooooo00;oo`04003ooooo0003ool01000ooooo`004_oo00<0
0?ooool02Ooo00<00?ooool01Ooo00<00?ooool00ooo00<00?ooool00_oo00<00?ooool00_oo00@0
0?ooool000?oo`04003oo`00000Hool003?oool000H001Soo`00ooooDOoo003ooomAool00?oooe7o
o`00ooooDOoo003ooomAool00?oooe7oo`00ooooDOoo003ooomAool002Woo`8000Coo`03003ooooo
00?oo`8000?oo`@002ooo`8000Coo`03003ooooo00;oo`@003Ooo`@003[oo`@000?oo`8005Koo`00
:?oo00@00?ooool000Soo`04003ooooo0004ool00`00ooooo`0]ool01000ooooo`002_oo00<00?oo
ool0>?oo00<00?ooool0>ooo00D00?ooooooo`0000;oo`03003ooooo05?oo`00:?oo00@00?ooool0
00Soo`04003ooooo0004ool00`00ooooo`0]ool01000ooooo`002_oo00<00?ooool0>?oo00<00?oo
ool0>ooo00D00?ooooooo`0000;oo`03003ooooo05?oo`00:?oo00@00?ooool000Soo`04003ooooo
0004ool00`00ooooo`0]ool01000ooooo`002_oo00<00?ooool0>?oo00<00?ooool0>ooo00D00?oo
ooooo`0000;oo`03003ooooo05?oo`00:?oo00@00?ooool000Soo`04003ooooo0004ool00`00oooo
o`0]ool01000ooooo`002_oo00<00?ooool0>?oo00<00?ooool0>ooo00D00?ooooooo`0000;oo`03
003ooooo05?oo`00:?oo00@00?ooool000Soo`04003ooooo0002ool3000_ool01000ooooo`002?oo
0`00>?oo0`00>ooo0`000ooo00@00?ooool005Goo`00:Ooo0P002_oo0P001Ooo00<00?ooool0;_oo
0P002ooo00<00?ooool0>?oo00<00?ooool0>ooo00<00?ooool00_oo0P00E_oo003ooomAool00:Oo
o`<00:Koo`00Z_oo00<00?ooool0Xooo0022ool40004ool20004ool200001?oo003ooooo0`0000Co
o`00003oo`8000?oo`800004ool00?ooool300001Ooo00000000ool00P0000?oo`0000000`000ooo
0P000ooo0P0000Coo`00ooooo`800004ool00?ooool300001?oo000000001?oo1@000_oo0P001Ooo
0P00JOoo0023ool01000ooooo`000_oo00@00?ooool000;oo`04003ooooo0002ool00`00ooooo`02
ool01@00oooooooo00000_oo00@00?ooool000;oo`04003ooooo0002ool01000ooooo`001?oo00<0
0?ooool00_oo00H00?ooool00?oo0002ool01000ooooo`000_oo00@00?ooool000Goo`03003ooooo
00Goo`05003oooooool00002ool01@00oooooooo00000_oo00<00?ooool0I_oo0023ool01000oooo
o`000_oo00@00?ooool000;oo`04003ooooo0002ool30002ool01@00ool00?oo00000ooo0`000_oo
00@00?ooool000;oo`04003ooooo0004ool00`00ooooo`02ool00`00ooooo`02ool01000ooooo`00
0_oo00@00?ooool000;oo`<000?oo`<000Goo`05003oooooool00002ool01@00oooooooo0000Jooo
0023ool30003ool01000ooooo`000_oo00@00?ooool000;oo`0:003ooooo003oo`00ool00?oo0005
ool01000ooooo`000_oo00@00?ooool000;oo`03003ooooo00;oo`03003ooooo00;oo`03003ooooo
00;oo`04003ooooo0002ool01000ooooo`000_oo00@00?ooool000;oo`04003ooooo0004ool01@00
oooooooo00000_oo00D00?ooooooo`0006_oo`00Pooo00@00?ooool000?oo`8000Coo`<000?oo`80
00;oo`800003ool0000000?oo`8000Coo`800004ool00000ool20004ool20003ool400001?oo0000
0?oo0P000ooo0`000ooo0P000ooo00@00?ooool000Coo`03003ooooo00;oo`8000?oo`@006Woo`00
\
\>"],
  ImageRangeCache->{{{0, 335}, {206.5, 0}} -> {-2.79964, -16.5784, 0.0154309, 
  0.0404959}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.25 0.603319 0.00232207 [
[.25 -0.0125 -11 -9.8125 ]
[.25 -0.0125 11 0 ]
[.5 -0.0125 -5 -9.8125 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -8 -9.8125 ]
[.75 -0.0125 8 0 ]
[1 -0.0125 -11 -9.8125 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -19.875 ]
[.5 -0.0125 29 -9.8125 ]
[-0.0125 .0228 -24 -4.5 ]
[-0.0125 .0228 0 4.5 ]
[-0.0125 .13891 -24 -4.5 ]
[-0.0125 .13891 0 4.5 ]
[-0.0125 .25501 -24 -4.5 ]
[-0.0125 .25501 0 4.5 ]
[-0.0125 .37111 -24 -4.5 ]
[-0.0125 .37111 0 4.5 ]
[-0.0125 .48722 -18 -4.5 ]
[-0.0125 .48722 0 4.5 ]
[-0.0125 .60332 -6 -4.5 ]
[-0.0125 .60332 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -46 -4.90625 ]
[-0.01875 .30902 -24 4.90625 ]
[0 .63053 -14 0 ]
[0 .63053 14 9.8125 ]
[.25 .63053 -11 0 ]
[.25 .63053 11 9.8125 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 9.8125 ]
[.75 .63053 -8 0 ]
[.75 .63053 8 9.8125 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -47 9.8125 ]
[.5 .63053 47 19.625 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.25 0 m
.25 .61803 L
s
.5 0 m
.5 .61803 L
s
.75 0 m
.75 .61803 L
s
1 0 m
1 .61803 L
s
0 .13891 m
1 .13891 L
s
0 .25501 m
1 .25501 L
s
0 .37111 m
1 .37111 L
s
0 .48722 m
1 .48722 L
s
0 g
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -72 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -13.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(100) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.001 w
.07526 0 m
.07526 .00375 L
s
.11928 0 m
.11928 .00375 L
s
.15051 0 m
.15051 .00375 L
s
.17474 0 m
.17474 .00375 L
s
.19454 0 m
.19454 .00375 L
s
.21127 0 m
.21127 .00375 L
s
.22577 0 m
.22577 .00375 L
s
.23856 0 m
.23856 .00375 L
s
.32526 0 m
.32526 .00375 L
s
.36928 0 m
.36928 .00375 L
s
.40051 0 m
.40051 .00375 L
s
.42474 0 m
.42474 .00375 L
s
.44454 0 m
.44454 .00375 L
s
.46127 0 m
.46127 .00375 L
s
.47577 0 m
.47577 .00375 L
s
.48856 0 m
.48856 .00375 L
s
.57526 0 m
.57526 .00375 L
s
.61928 0 m
.61928 .00375 L
s
.65051 0 m
.65051 .00375 L
s
.67474 0 m
.67474 .00375 L
s
.69454 0 m
.69454 .00375 L
s
.71127 0 m
.71127 .00375 L
s
.72577 0 m
.72577 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.82526 0 m
.82526 .00375 L
s
.86928 0 m
.86928 .00375 L
s
.90051 0 m
.90051 .00375 L
s
.92474 0 m
.92474 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.96127 0 m
.96127 .00375 L
s
.97577 0 m
.97577 .00375 L
s
.98856 0 m
.98856 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -23.875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.062500 moveto
 558.000000 14.062500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.062500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 11.250000 moveto
(H) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 11.250000 moveto
(\\220) show
105.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 11.250000 moveto
(L) show
117.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Math2Mono
%%+ Courier
grestore
0 .0228 m
.00625 .0228 L
s
[(-250)] -0.0125 .0228 1 0 Mshowa
0 .13891 m
.00625 .13891 L
s
[(-200)] -0.0125 .13891 1 0 Mshowa
0 .25501 m
.00625 .25501 L
s
[(-150)] -0.0125 .25501 1 0 Mshowa
0 .37111 m
.00625 .37111 L
s
[(-100)] -0.0125 .37111 1 0 Mshowa
0 .48722 m
.00625 .48722 L
s
[(-50)] -0.0125 .48722 1 0 Mshowa
0 .60332 m
.00625 .60332 L
s
[(0)] -0.0125 .60332 1 0 Mshowa
.125 Mabswid
0 .04602 m
.00375 .04602 L
s
0 .06924 m
.00375 .06924 L
s
0 .09246 m
.00375 .09246 L
s
0 .11569 m
.00375 .11569 L
s
0 .16213 m
.00375 .16213 L
s
0 .18535 m
.00375 .18535 L
s
0 .20857 m
.00375 .20857 L
s
0 .23179 m
.00375 .23179 L
s
0 .27823 m
.00375 .27823 L
s
0 .30145 m
.00375 .30145 L
s
0 .32467 m
.00375 .32467 L
s
0 .34789 m
.00375 .34789 L
s
0 .39433 m
.00375 .39433 L
s
0 .41755 m
.00375 .41755 L
s
0 .44077 m
.00375 .44077 L
s
0 .46399 m
.00375 .46399 L
s
0 .51044 m
.00375 .51044 L
s
0 .53366 m
.00375 .53366 L
s
0 .55688 m
.00375 .55688 L
s
0 .5801 m
.00375 .5801 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -107 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(deg) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
87.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.25 .61178 m
.25 .61803 L
s
gsave
.25 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.75 .61178 m
.75 .61803 L
s
gsave
.75 .63053 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
.001 w
.07526 .61428 m
.07526 .61803 L
s
.11928 .61428 m
.11928 .61803 L
s
.15051 .61428 m
.15051 .61803 L
s
.17474 .61428 m
.17474 .61803 L
s
.19454 .61428 m
.19454 .61803 L
s
.21127 .61428 m
.21127 .61803 L
s
.22577 .61428 m
.22577 .61803 L
s
.23856 .61428 m
.23856 .61803 L
s
.32526 .61428 m
.32526 .61803 L
s
.36928 .61428 m
.36928 .61803 L
s
.40051 .61428 m
.40051 .61803 L
s
.42474 .61428 m
.42474 .61803 L
s
.44454 .61428 m
.44454 .61803 L
s
.46127 .61428 m
.46127 .61803 L
s
.47577 .61428 m
.47577 .61803 L
s
.48856 .61428 m
.48856 .61803 L
s
.57526 .61428 m
.57526 .61803 L
s
.61928 .61428 m
.61928 .61803 L
s
.65051 .61428 m
.65051 .61803 L
s
.67474 .61428 m
.67474 .61803 L
s
.69454 .61428 m
.69454 .61803 L
s
.71127 .61428 m
.71127 .61803 L
s
.72577 .61428 m
.72577 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.82526 .61428 m
.82526 .61803 L
s
.86928 .61428 m
.86928 .61803 L
s
.90051 .61428 m
.90051 .61803 L
s
.92474 .61428 m
.92474 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.96127 .61428 m
.96127 .61803 L
s
.97577 .61428 m
.97577 .61803 L
s
.98856 .61428 m
.98856 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .0228 m
1 .0228 L
s
.99375 .13891 m
1 .13891 L
s
.99375 .25501 m
1 .25501 L
s
.99375 .37111 m
1 .37111 L
s
.99375 .48722 m
1 .48722 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .04602 m
1 .04602 L
s
.99625 .06924 m
1 .06924 L
s
.99625 .09246 m
1 .09246 L
s
.99625 .11569 m
1 .11569 L
s
.99625 .16213 m
1 .16213 L
s
.99625 .18535 m
1 .18535 L
s
.99625 .20857 m
1 .20857 L
s
.99625 .23179 m
1 .23179 L
s
.99625 .27823 m
1 .27823 L
s
.99625 .30145 m
1 .30145 L
s
.99625 .32467 m
1 .32467 L
s
.99625 .34789 m
1 .34789 L
s
.99625 .39433 m
1 .39433 L
s
.99625 .41755 m
1 .41755 L
s
.99625 .44077 m
1 .44077 L
s
.99625 .46399 m
1 .46399 L
s
.99625 .51044 m
1 .51044 L
s
.99625 .53366 m
1 .53366 L
s
.99625 .55688 m
1 .55688 L
s
.99625 .5801 m
1 .5801 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -108 5.8125 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Phase) show
99.000000 11.250000 moveto
(Bode) show
129.000000 11.250000 moveto
(Plot) show
153.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.09751 .60184 Mdot
.17283 .60036 Mdot
.21696 .59887 Mdot
.24835 .59736 Mdot
.27277 .59583 Mdot
.29281 .59428 Mdot
.30984 .5927 Mdot
.32467 .59109 Mdot
.33783 .58943 Mdot
.3497 .58774 Mdot
.36052 .58599 Mdot
.37048 .58418 Mdot
.37975 .58231 Mdot
.38841 .58038 Mdot
.39657 .57836 Mdot
.40431 .57627 Mdot
.41167 .57409 Mdot
.41871 .57181 Mdot
.42547 .56943 Mdot
.43199 .56694 Mdot
.4383 .56432 Mdot
.44443 .56158 Mdot
.45039 .5587 Mdot
.45622 .55567 Mdot
.46194 .55247 Mdot
.46755 .54911 Mdot
.47308 .54556 Mdot
.47854 .54181 Mdot
.48395 .53785 Mdot
.48932 .53366 Mdot
.49467 .52922 Mdot
.5 .52453 Mdot
.50533 .51955 Mdot
.51068 .51427 Mdot
.51605 .50866 Mdot
.52146 .5027 Mdot
.52692 .49637 Mdot
.53245 .48963 Mdot
.53806 .48245 Mdot
.54378 .4748 Mdot
.54961 .46664 Mdot
.55557 .45793 Mdot
.5617 .44861 Mdot
.56801 .43866 Mdot
.57453 .42799 Mdot
.58129 .41657 Mdot
.58833 .4043 Mdot
.59569 .39114 Mdot
.60343 .37697 Mdot
.61159 .36173 Mdot
.62025 .3453 Mdot
.62952 .32756 Mdot
.63948 .30841 Mdot
.6503 .2877 Mdot
.66217 .26528 Mdot
.67533 .24103 Mdot
.69016 .2148 Mdot
.70719 .18648 Mdot
.72723 .15601 Mdot
.75165 .12339 Mdot
.78304 .08875 Mdot
.82717 .05238 Mdot
.90249 .01472 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{336, 207.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005@0000ca000`40n007h00OooooDOoo003ooomA
ool00?oooe7oo`00ooooDOoo002`ool00`00ooooo`0Nool00`00ooooo`1lool00:ooo`03003ooooo
01?oo`03003ooooo00[oo`03003ooooo07_oo`00XOoo00<00?oo00002ooo00@00?ooooooo`@000?o
o`800004ool00?ooool200001Ooo003ooooo00001?oo0`001?oo00<00?ooool0Nooo002Qool00`00
ool0000;ool00`00ooooo`02ool00`00ooooo`02ool01000ooooo`000_oo00@00?ooool000?oo`03
003ooooo00Goo`05003oooooool0001mool00:3oo`05003oo`00ool0000:ool00`00ooooo`02ool0
0`00ooooo`03ool30002ool01000ooooo`000ooo0P001?oo0P001?oo00<00?ooool0Nooo002Pool0
1@00ool00?oo00002_oo00<00?ooool00_oo00<00?ooool01Ooo00@00?ooool000;oo`03003ooooo
00;oo`05003oooooool00006ool00`00ooooo`1kool00:3oo`04003oo`00ool20009ool01P00oooo
oooo003oo`<000;oo`8000Coo`<000Coo`03003ooooo00;oo`<000?oo`03003ooooo07_oo`00/?oo
00<00?ooool03ooo00<00?ooool00ooo00<00?ooool01_oo00<00?ooool0O?oo0031ool2002=ool0
0?oooe7oo`00ooooDOoo003ooomAool007?oo`8000Coo`03003ooooo00;oo`@003Ooo`@003[oo`@0
00?oo`8003Goo`@000?oo`8000Coo`8000coo`00L_oo00@00?ooool000[oo`03003ooooo03Soo`03
003ooooo03_oo`05003oooooool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?oo
ool000;oo`03003ooooo00Woo`00L_oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_o
o`05003oooooool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03
003ooooo00Woo`00L_oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_oo`05003ooooo
ool00002ool00`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo00Wo
o`00L_oo00@00?ooool000[oo`03003ooooo03Soo`03003ooooo03_oo`05003oooooool00002ool0
0`00ooooo`0dool01@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo00Woo`00L_oo00@0
0?ooool000Soo`<003Soo`<003_oo`<000?oo`04003ooooo000dool30003ool01000ooooo`000_oo
00@00?ooool000_oo`00Looo0P002ooo00<00?ooool0>?oo00<00?ooool0>ooo00<00?ooool00_oo
0P00=ooo00<00?ooool00_oo0P001?oo0P003?oo000ioooo0006000Bool003Woo`03003ooooo017o
o`03003ooooo00Soo`03003ooooo00Goo`03003ooooo00?oo`03003ooooo00?oo`05003oooooool0
0003ool01000ooooo`000_oo00<00?ooool04Ooo00<00?ooool02?oo00<00?ooool01Ooo00<00?oo
ool00ooo00<00?ooool00ooo00D00?ooooooo`0000?oo`04003ooooo0002ool00`00ooooo`0@ool0
0`00ooooo`09ool00`00ooooo`05ool00`00ooooo`03ool00`00ooooo`02ool00`00ooooo`02ool0
1@00oooooooo00000_oo00@00?ooool001;oo`03003ooooo00Woo`03003ooooo00Goo`03003ooooo
00?oo`03003ooooo00;oo`03003ooooo00;oo`04003ooooo0003ool01000ooooo`004_oo000iool0
0`00ooooo`0nool00`00ooooo`0nool00`00ooooo`0nool00`00ooooo`0nool00`00ooooo`0@ool0
03Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo
013oo`00>Ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool09?oo0P00
6?oo00<00?ooool04?oo000Uool40002ool30004ool20005ool00`00ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0Uool00`00ooooo`0Fool00`00ooooo`0@ool002Goo`03003o
oooo00Koo`04003ooooo0002ool00`00ooooo`02ool3000nool00`0?ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0lool3000Bool002Koo`03003ooooo00Goo`04003ooooo0002ool00`00ooooo`02
ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@
ool001ooo`@000Coo`04003oooooool30003ool01000ooooo`001?oo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool04?oo000Xool01000ooooo`001Ooo
00@00?ooool000Coo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03ko
o`03003ooooo013oo`009Ooo00@00?ooool000;oo`03003ooooo00?oo`04003ooooo0004ool00`00
ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool002Ko
o`8000?oo`@000?oo`8000Goo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013o
oooo03koo`03003ooooo013oo`00>Ooo0P00?ooo00<03oooool0?_oo00<04?ooool0?_oo00<03ooo
ool0?Ooo0P004_oo000iool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@oooo
o`0@ool2000/ool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo03koo`03013ooooo
03koo`0300oooooo017oo`03003ooooo02[oo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<0
4?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0n
ool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003o
oooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo013oo`00>Ooo
0P00?ooo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?Ooo0P004_oo000iool00`00oooo
o`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03
003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00
>Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool0
4?oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00
ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo00Go
o`8003Ooo`03003ooooo013oo`00>Ooo0P00?ooo00<03oooool0?_oo00<04?ooool0?_oo00<03ooo
ool01_oo00<00?ooool0=?oo0P004_oo000iool00`00ooooo`0nool00`0@ooooo`0nool00`0?oooo
o`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo03koo`03
013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<04?ooool0
?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0?
ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03ko
o`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo013oo`00>Ooo0P00?ooo
00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?Ooo0P004_oo000iool00`00ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo
03koo`0300oooooo03koo`03013ooooo03koo`8003ooo`03003ooooo013oo`00>Ooo00<00?ooool0
?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04000ool0?_oo00<00?ooool04?oo000iool00`00
ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool002Go
o`@000?oo`8000Coo`8000Goo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013o
oooo03koo`03003ooooo013oo`009Ooo00<00?ooool00ooo00@00?ooool000;oo`04003ooooo0004
ool30000o`0?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?0100
3`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`03000Bool002Koo`03003ooooo00;o
o`04003ooooo0002ool01000ooooo`001?oo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0
?_oo00<04?ooool0?_oo00<00?ooool04?oo000Oool40004ool01@00oooooooo00000_oo00@00?oo
ool000;oo`03003ooooo00;oo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oo
oooo03koo`03003ooooo013oo`00:?oo00@00?ooool000;oo`04003ooooo0002ool00`00ooooo`02
ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@
ool002Goo`04003ooooo0002ool01000ooooo`000_oo00@00?ooool000Coo`03003ooooo03koo`03
00oooooo03koo`03013ooooo03Ooo`8000Goo`0300oooooo03koo`03003ooooo013oo`009_oo0P00
1?oo0P001?oo0P001Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0>?oo00<00?ooool0
0ooo00<04?ooool0?_oo00<00?ooool04?oo000iool2000oool00`0?ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0mool2000Bool003Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo03ko
o`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?oo
ool0?_oo00<03oooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0@ooooo`0nool0
0`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo
03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<0
4?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool2000oool00`0?
ooooo`0nool00`0@ooooo`0bool2000:ool00`0?ooooo`0mool2000Bool003Woo`03003ooooo03ko
o`03013ooooo03koo`0300oooooo03?oo`03003ooooo00Soo`03013ooooo03koo`03003ooooo013o
o`00>Ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?oo
ool04?oo000iool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool0
0`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo
03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<0
4?ooool0?_oo00<00?ooool04?oo000iool2000oool00`0?ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0mool2000Bool003Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo03koo`03013o
oooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0;_oo
0P003_oo00<03oooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0@ooooo`0nool0
0`0?ooooo`0_ool00`00ooooo`0<ool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo
03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00>Ooo00<0
0?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000i
ool2000oool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0mool2000Bool003Woo`03003o
oooo03koo`03013ooooo03koo`0300oooooo03koo`03013ooooo03koo`03003ooooo013oo`00>Ooo
00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0:_oo0P004_oo00<03oooool0?_oo00<00?oo
ool04?oo000iool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0[ool00`00ooooo`0@ool0
0`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo03koo`03013ooooo
03koo`0300oooooo03koo`03003ooooo013oo`009Ooo10000_oo0`001?oo0P001Ooo00<00?ooool0
?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000Wool00`00
ooooo`04ool01000ooooo`000_oo00<00?ooool00_oo0`000?l03`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l00`004_oo000Wool00`00ooooo`04ool01000ooooo`000_oo00<00?ooool00_oo00<0
0?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000O
ool40004ool01000oooooooo0`000ooo00@00?ooool000Coo`03003ooooo03koo`0300oooooo03ko
o`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`009ooo00D00?ooooooo`0000Goo`04
003ooooo0004ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0Vool2000Fool00`0@oooo
o`0nool00`00ooooo`0@ool002Goo`<000?oo`03003ooooo00?oo`04003ooooo0004ool00`00oooo
o`0nool00`0?ooooo`0nool00`0@ooooo`0Wool00`00ooooo`0Dool00`0?ooooo`0nool00`00oooo
o`0@ool002Ooo`04003oooooool40003ool20005ool00`00ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`8003ooo`0300oooooo03koo`03013o
oooo03koo`0300oooooo03goo`8001;oo`00>Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0?ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo
03koo`0300oooooo02?oo`8001Woo`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0
?_oo00<03oooool0?_oo00<04?ooool09?oo00<00?ooool05ooo00<03oooool0?_oo00<00?ooool0
4?oo000iool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00
ooooo`0@ool003Woo`8003ooo`0300oooooo03koo`03013ooooo03koo`0300oooooo03goo`8001;o
o`005Ooo0`008Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo
00<00?ooool04?oo000Hool00`00ooooo`0Nool00`00ooooo`0nool00`0?ooooo`0nool00`0@oooo
o`0Pool2000Lool00`0?ooooo`0nool00`00ooooo`0@ool000[oo`800004ool00?ooool30003ool3
000Pool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0Qool00`00ooooo`0Jool00`0@oooo
o`0nool00`00ooooo`0@ool000Woo`04003ooooo0002ool00`00ooooo`03ool01000ooooo`008?oo
00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool04?oo
0009ool01000ooooo`000_oo0`000ooo00@00?ooool0023oo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo03koo`03013ooooo03koo`03003ooooo013oo`002Ooo00@00?ooool000;oo`04003o
oooo0002ool01000ooooo`008?oo0P00?ooo00<03oooool0?_oo00<04?ooool0?_oo00<03oooool0
?Ooo0P004_oo000:ool30003ool20004ool200000ooo003oo`0Nool00`00ooooo`0nool00`0@oooo
o`0nool00`0?ooooo`0Nool2000Nool00`0@ooooo`0nool00`00ooooo`0@ool000coo`03003ooooo
02[oo`03003ooooo03koo`0300oooooo03koo`03013ooooo01ooo`03003ooooo01coo`0300oooooo
03koo`03003ooooo013oo`002ooo0P00;?oo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0
?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@
ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo03koo`03013ooooo03koo`03003ooooo013oo`00>Ooo0P00?ooo00<03oooool0?_oo
00<04?ooool06ooo0P008Ooo00<03oooool0?Ooo0P004_oo000iool00`00ooooo`0nool00`0@oooo
o`0nool00`0?ooooo`0Lool00`00ooooo`0Oool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03
003ooooo03koo`0300oooooo03koo`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00
>Ooo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool0
4?oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0Iool2000Sool00`0?ooooo`0n
ool00`00ooooo`0@ool002Goo`@000?oo`8000Coo`8000Goo`03003ooooo03koo`03013ooooo03ko
o`0300oooooo01[oo`03003ooooo027oo`03013ooooo03koo`03003ooooo013oo`009ooo00D00?oo
ooooo`0000;oo`04003ooooo0002ool00`00ooooo`02ool30000o`0?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`03000Bool002Ooo`05003oooooool00002ool01000ooooo`000_oo00<00?ooool0
0_oo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool0
4?oo000Oool40004ool01@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo00;oo`03003o
oooo03koo`0300oooooo03koo`03013ooooo01Ooo`8002Goo`0300oooooo03koo`03003ooooo013o
o`009ooo00D00?ooooooo`0000;oo`04003ooooo0002ool00`00ooooo`02ool00`00ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0Hool00`00ooooo`0Sool00`0@ooooo`0nool00`00ooooo`0@ool0
02Goo`<000?oo`04003ooooo0002ool01000ooooo`001?oo00<00?ooool0?_oo00<03oooool0?_oo
00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool04?oo000Wool00`00ooooo`02ool20004ool2
0005ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0Eool2000Wool00`0@ooooo`0nool0
0`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo03koo`03013ooooo01Koo`03003ooooo
02Goo`0300oooooo03koo`03003ooooo013oo`00>Ooo0P00?ooo00<04?ooool0?_oo00<03oooool0
?_oo00<04?ooool0?Ooo0P004_oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0n
ool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03koo`0300oo
oooo01?oo`8002Woo`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool05?oo00<00?ooool09ooo00<03oooool0?_oo00<00?ooool04?oo000iool0
0`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool0
03Woo`03003ooooo03koo`0300oooooo03koo`03013ooooo017oo`8002_oo`0300oooooo03koo`03
003ooooo013oo`00>Ooo0P00?ooo00<04?ooool0?_oo00<03oooool04_oo00<00?ooool0:Ooo00<0
4?ooool0?Ooo0P004_oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?
ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo013o
o`8002coo`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03oooool0?_oo
00<04?ooool04Ooo00<00?ooool0:_oo00<03oooool0?_oo00<00?ooool04?oo000iool00`00oooo
o`0nool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03
003ooooo03koo`0300oooooo03koo`03013ooooo00koo`8002koo`0300oooooo03koo`03003ooooo
013oo`00>Ooo0P00?ooo00<04?ooool0?_oo00<03oooool03ooo00<00?ooool0;?oo00<04?ooool0
?Ooo0P004_oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0<ool2000`ool00`0?
ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03koo`0300oooooo00go
o`03003ooooo02koo`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool0
0`0@ooooo`0nool00`0?ooooo`0;ool2000aool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03
003ooooo03koo`0300oooooo03koo`03013ooooo00coo`03003ooooo02ooo`0300oooooo03koo`03
003ooooo013oo`00>Ooo0P00?ooo00<04?ooool0?_oo00<03oooool02Ooo0P00<ooo00<04?ooool0
?Ooo0P004_oo000iool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0:ool00`00ooooo`0a
ool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo03koo`0300oo
oooo00Soo`8003Coo`03013ooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool02Ooo00<00?ooool0<_oo00<03oooool0?_oo00<00?ooool04?oo000[ool3
0004ool20005ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`06ool2000fool00`0@oooo
o`0nool00`00ooooo`0@ool002koo`04003ooooo0002ool00`00ooooo`02ool00`00ooooo`0nool0
0`0?ooooo`0nool00`0@ooooo`07ool00`00ooooo`0dool00`0?ooooo`0nool00`00ooooo`0@ool0
02koo`04003ooooo0002ool00`00ooooo`02ool30000Q`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@
00l040020000MP0?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?01003`0@00l0400?
01003`0@0`004_oo000Uool40002ool30003ool01000ooooo`001?oo00<00?ooool0?_oo00<03ooo
ool0?_oo00<04?ooool01_oo00<00?ooool0=Ooo00<03oooool0?_oo00<00?ooool04?oo000[ool0
0`00ooooo`03ool01000ooooo`001?oo00<00?ooool0?_oo00<04?ooool0?_oo00<03oooool00ooo
0P00>Ooo00<04?ooool0?_oo00<00?ooool04?oo000[ool00`00ooooo`03ool01000ooooo`001?oo
00<00?ooool0?_oo00<03oooool0?_oo00<04?ooool00_oo0`00>Ooo00<03oooool0?_oo00<00?oo
ool04?oo000[ool40003ool20005ool00`00ooooo`0nool00`0@ooooo`0nool00`0?ooooo`03ool0
0`00ooooo`0hool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo
03koo`03013ooooo008003coo`0300oooooo03koo`03003ooooo013oo`00>Ooo0P00?ooo00<04?oo
ool0?_oo00<03ooo00000P00??oo00<04?ooool0?Ooo0P004_oo000iool00`00ooooo`0nool00`0?
ooooo`0nool0100@ooooo`00?Ooo00<03oooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0n
ool00`0@ooooo`0nool00`0?0000000nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003o
oooo03koo`0300oooooo03goo`800003ool00?oo03goo`0300oooooo03koo`03003ooooo013oo`00
>Ooo00<00?ooool0?_oo00<04?ooool0??oo0`00@?oo00<04?ooool0?_oo00<00?ooool04?oo000i
ool00`00ooooo`0nool00`0?ooooo`0jool200000ooo00004010ool00`0?ooooo`0nool00`00oooo
o`0@ool003Woo`8003ooo`03013ooooo03_oo`04003ooooo00m0ool00`0@ooooo`0mool2000Bool0
03Woo`03003ooooo03koo`0300oooooo03Woo`8000?oo`03013ooooo03koo`0300oooooo03koo`03
003ooooo013oo`00>Ooo00<00?ooool0?_oo00<04?ooool0>?oo0`000ooo00<03oooool0?_oo00<0
4?ooool0?_oo00<00?ooool04?oo000iool00`00ooooo`0nool00`0?ooooo`0eool300000ooo003o
o`03ool00`0@ooooo`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03
013ooooo03?oo`800003ool0000000Koo`0300oooooo03koo`03013ooooo03koo`03003ooooo013o
o`00>Ooo00<00?ooool0?_oo00<03oooool0<_oo0`002Ooo00<04?ooool0?_oo00<03oooool0?_oo
00<00?ooool04?oo000iool2000oool00`0@ooooo`0`ool200000ooo003oo`09ool00`0?ooooo`0n
ool00`0@ooooo`0mool2000Bool003Woo`03003ooooo03koo`0300oooooo02goo`D000coo`03013o
oooo03koo`0300oooooo03koo`03003ooooo013oo`00>Ooo00<00?ooool0?_oo00<04?ooool0;?oo
0`0000?oo`00ool03?oo00<03oooool0?_oo00<04?ooool0?_oo00<00?ooool04?oo000iool00`00
ooooo`0nool00`0?ooooo`0Xool400000ooo003oo`0?ool00`0@ooooo`0nool00`0?ooooo`0nool0
0`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo02Koo`800004ool00?oo000Bool00`0?
ooooo`0nool00`0@ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`0300oooooo02;o
o`@00003ool00?oo01Goo`03013ooooo03koo`0300oooooo03koo`03003ooooo013oo`00>Ooo0P00
?ooo00<04?ooool07_oo100000Coo`00ool001Soo`0300oooooo03koo`03013ooooo03goo`8001;o
o`00>Ooo00<00?ooool0?_oo00<03oooool06Ooo0P0000Ooo`00003oo`00ool0000Lool00`0@oooo
o`0nool00`0?ooooo`0nool00`00ooooo`0@ool003Woo`03003ooooo03koo`03013ooooo013oo`80
0004ool00000ool20002ool01000ooooo`008?oo00<03oooool0?_oo00<04?ooool0?_oo00<00?oo
ool04?oo000iool00`00ooooo`0nool00`0?ooooo`07ool20003ool20003ool01000ooooo`000_oo
00<00?ooool09?oo00<04?ooool0?_oo00<03oooool0?_oo00<00?ooool04?oo000bool20005ool0
0`00ooooo`0eool20006ool20004ool20004ool00`00ooooo`02ool00`00ooooo`0^ool00`0?oooo
o`0nool00`0@ooooo`0nool00`00ooooo`0@ool0037oo`04003ooooo0004ool00`00ooooo`0Fool2
000Aool2000;ool00`00ooooo`05ool00`00ooooo`03ool00`00ooooo`0hool00`0@ooooo`0nool0
0`0?ooooo`0nool00`00ooooo`0@ool0037oo`04003ooooo0004ool3000Gool00`00ooooo`0@ool0
0`00ooooo`0Aool00`0@ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0lool3000Bool0037oo`04
003ooooo0004ool00`00ooooo`0nool00`0?ooooo`0nool00`0@ooooo`0nool00`0?ooooo`0nool0
0`00ooooo`0@ool0037oo`04003ooooo0004ool00`00ooooo`0nool00`00ooooo`0nool00`00oooo
o`0nool00`00ooooo`0nool00`00ooooo`0@ool0037oo`04003ooooo0004ool00`00ooooo`0Aool0
0`00ooooo`08ool00`00ooooo`05ool00`00ooooo`03ool00`00ooooo`03ool01@00oooooooo0000
0ooo00@00?ooool000;oo`03003ooooo017oo`03003ooooo00Soo`03003ooooo00Goo`03003ooooo
00?oo`03003ooooo00?oo`05003oooooool00003ool01000ooooo`000_oo00<00?ooool04?oo00<0
0?ooool02Ooo00<00?ooool01Ooo00<00?ooool00ooo00<00?ooool00_oo00<00?ooool00_oo00D0
0?ooooooo`0000;oo`04003ooooo000Bool00`00ooooo`09ool00`00ooooo`05ool00`00ooooo`03
ool00`00ooooo`02ool00`00ooooo`02ool01000ooooo`000ooo00@00?ooool001;oo`00<_oo0P00
1Oooo`001P004_oo003ooomAool00?oooe7oo`00ooooDOoo003ooomAool00?oooe7oo`00ooooDOoo
003ooomAool00?oooe7oo`00;ooo0P001?oo00<00?ooool00ooo0P000ooo1000;ooo0P001?oo00<0
0?ooool00_oo1000=ooo1000>_oo10000ooo0P00D?oo000^ool01000ooooo`002?oo00@00?ooool0
00Coo`03003ooooo02goo`04003ooooo000:ool00`00ooooo`0hool00`00ooooo`0kool01@00oooo
oooo00000_oo00<00?ooool0COoo000^ool01000ooooo`002?oo00@00?ooool000Coo`03003ooooo
02goo`04003ooooo000:ool00`00ooooo`0hool00`00ooooo`0kool01@00oooooooo00000_oo00<0
0?ooool0COoo000^ool01000ooooo`002?oo00@00?ooool000Coo`03003ooooo02goo`04003ooooo
000:ool00`00ooooo`0hool00`00ooooo`0kool01@00oooooooo00000_oo00<00?ooool0COoo000^
ool01000ooooo`002?oo00@00?ooool000Coo`03003ooooo02goo`04003ooooo000:ool00`00oooo
o`0hool00`00ooooo`0kool01@00oooooooo00000_oo00<00?ooool0COoo000^ool01000ooooo`00
2?oo00@00?ooool000;oo`<002ooo`04003ooooo0008ool3000hool3000kool30003ool01000oooo
o`00Cooo000_ool2000:ool20005ool00`00ooooo`0^ool2000;ool00`00ooooo`0hool00`00oooo
o`0kool00`00ooooo`02ool2001@ool00?oooe7oo`00ooooDOoo002>ool30003ool300000ooo0000
0002ool200000ooo003oo`030004ool30007ool40004ool20004ool200001?oo003ooooo0`001ooo
0`001?oo1@000_oo0P001Ooo0P00JOoo002?ool00`00ooooo`03ool01000ooooo`000_oo00@00?oo
ool000Goo`04003ooooo000;ool01000ooooo`000_oo00@00?ooool000;oo`04003ooooo0002ool0
0`00ooooo`09ool00`00ooooo`05ool01@00oooooooo00000_oo00D00?ooooooo`0000;oo`03003o
oooo06Koo`00Sooo0`000ooo00@00?ooool000?oo`<000?oo`8000?oo`<000Woo`04003ooooo0002
ool01000ooooo`000_oo00@00?ooool000;oo`<000Woo`<000Goo`05003oooooool00002ool01@00
oooooooo0000Jooo002?ool01000ooooo`000_oo00@00?ooool000Goo`04003ooooo0005ool01000
ooooo`002?oo0`000ooo00@00?ooool000;oo`04003ooooo0002ool01000ooooo`002?oo00@00?oo
ool000Coo`05003oooooool00002ool01@00oooooooo0000Jooo002?ool01000ooooo`000_oo0`00
1?oo0P001?oo0`000ooo0P002Ooo00@00?ooool000?oo`8000Coo`<000?oo`8000Woo`04003ooooo
0004ool00`00ooooo`02ool20003ool4001Yool00001\
\>"],
  ImageRangeCache->{{{0, 335}, {206.5, 0}} -> {-2.88688, -298.161, 0.0154021, 
  1.65823}}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["5.6 Optimality diagnostics", "Section"],

Cell[CellGroupData[{

Cell["flat", "Input"],

Cell[BoxData[
    \(0.00356167256510193785`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["grAlign", "Input"],

Cell[BoxData[
    \(4.44089209850062616`*^-16\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["ned", "Input"],

Cell[BoxData[
    \(0.0131836236287413743`\)], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell["gammastar", "Input"],

Cell[BoxData[
    \(0.159720709242560987`\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example 2 : changing tracking and roll off", "Section"],

Cell[CellGroupData[{

Cell["\<\

     p[s_]= (1 - s/5)/(1+s/2)^2;

      wt  = 0.3;
  alphat  = 0.5;

 alphagpm = 2.0;

       wb = 2.0;
   alphab = 0.7;
              
       wr = 4.0;
   alphar = 30.;
 
    k0[w_]= Which[ Abs[w] < wb, 1.0,
                     True, 0.0];
 
   r0[w_] = Which[Abs[w] < wt,  alphat,
               Abs[w] < wb,  alphagpm,
               Abs[w] < wr,  alphab,
                     True ,  alphar*Abs[p[I w]/w^2]
                ];
                
     wpts = Grid[Ngrid ->128,GridSpread -> 1.0];


EnvelopePlot[Radius ->r0,Center -> k0,
\t\t\tFrequencyBand -> {0,wr+1} 
\t\t \t(* ,GridPoints -> wpts *) ];\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.77556e-17 0.2 0.161866 0.147151 [
[.2 .14937 -3 -9 ]
[.2 .14937 3 0 ]
[.4 .14937 -3 -9 ]
[.4 .14937 3 0 ]
[.6 .14937 -3 -9 ]
[.6 .14937 3 0 ]
[.8 .14937 -3 -9 ]
[.8 .14937 3 0 ]
[1 .14937 -3 -9 ]
[1 .14937 3 0 ]
[-0.0125 .01472 -12 -4.5 ]
[-0.0125 .01472 0 4.5 ]
[-0.0125 .30902 -6 -4.5 ]
[-0.0125 .30902 0 4.5 ]
[-0.0125 .45617 -6 -4.5 ]
[-0.0125 .45617 0 4.5 ]
[-0.0125 .60332 -6 -4.5 ]
[-0.0125 .60332 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 .16187 m
.2 .16812 L
s
[(1)] .2 .14937 0 1 Mshowa
.4 .16187 m
.4 .16812 L
s
[(2)] .4 .14937 0 1 Mshowa
.6 .16187 m
.6 .16812 L
s
[(3)] .6 .14937 0 1 Mshowa
.8 .16187 m
.8 .16812 L
s
[(4)] .8 .14937 0 1 Mshowa
1 .16187 m
1 .16812 L
s
[(5)] 1 .14937 0 1 Mshowa
.125 Mabswid
.04 .16187 m
.04 .16562 L
s
.08 .16187 m
.08 .16562 L
s
.12 .16187 m
.12 .16562 L
s
.16 .16187 m
.16 .16562 L
s
.24 .16187 m
.24 .16562 L
s
.28 .16187 m
.28 .16562 L
s
.32 .16187 m
.32 .16562 L
s
.36 .16187 m
.36 .16562 L
s
.44 .16187 m
.44 .16562 L
s
.48 .16187 m
.48 .16562 L
s
.52 .16187 m
.52 .16562 L
s
.56 .16187 m
.56 .16562 L
s
.64 .16187 m
.64 .16562 L
s
.68 .16187 m
.68 .16562 L
s
.72 .16187 m
.72 .16562 L
s
.76 .16187 m
.76 .16562 L
s
.84 .16187 m
.84 .16562 L
s
.88 .16187 m
.88 .16562 L
s
.92 .16187 m
.92 .16562 L
s
.96 .16187 m
.96 .16562 L
s
.25 Mabswid
0 .16187 m
1 .16187 L
s
0 .01472 m
.00625 .01472 L
s
[(-1)] -0.0125 .01472 1 0 Mshowa
0 .30902 m
.00625 .30902 L
s
[(1)] -0.0125 .30902 1 0 Mshowa
0 .45617 m
.00625 .45617 L
s
[(2)] -0.0125 .45617 1 0 Mshowa
0 .60332 m
.00625 .60332 L
s
[(3)] -0.0125 .60332 1 0 Mshowa
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .13244 m
.00375 .13244 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .25016 m
.00375 .25016 L
s
0 .27959 m
.00375 .27959 L
s
0 .33845 m
.00375 .33845 L
s
0 .36788 m
.00375 .36788 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .4856 m
.00375 .4856 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
1 .12227 m
.89064 .10101 L
.79844 .07715 L
.72271 .06301 L
.65931 .05917 L
.60541 .05886 L
.55896 .05886 L
.51848 .05886 L
.48284 .0586 L
.45119 .05548 L
.42286 .04471 L
.39733 .02887 L
.37417 .01809 L
.35305 .01497 L
.33368 .01472 L
.31583 .01472 L
.29932 .01472 L
.28398 .01472 L
.26967 .01472 L
.25628 .01472 L
.2437 .01472 L
.23186 .01472 L
.22067 .01472 L
.21007 .01472 L
.2 .01472 L
.19042 .01472 L
.18127 .01472 L
.17252 .01472 L
.16414 .01472 L
.15608 .01472 L
.14833 .01472 L
.14086 .01472 L
.13364 .01472 L
.12665 .01472 L
.11988 .01472 L
.1133 .01472 L
.1069 .01472 L
.10067 .01472 L
.09459 .01472 L
.08865 .01472 L
.08284 .01472 L
.07715 .01472 L
.07156 .016 L
.06607 .0316 L
.06067 .08549 L
.05535 .16467 L
.0501 .21856 L
.04491 .23415 L
.03978 .23544 L
.0347 .23544 L
Mistroke
.02967 .23544 L
.02467 .23544 L
.0197 .23544 L
.01475 .23544 L
.00983 .23544 L
.00491 .23544 L
0 .23544 L
Mfstroke
1 .20146 m
.89064 .22272 L
.79844 .24658 L
.72271 .26073 L
.65931 .26456 L
.60541 .26487 L
.55896 .26487 L
.51848 .26487 L
.48284 .26685 L
.45119 .29076 L
.42286 .37339 L
.39733 .4948 L
.37417 .57743 L
.35305 .60134 L
.33368 .60332 L
.31583 .60332 L
.29932 .60332 L
.28398 .60332 L
.26967 .60332 L
.25628 .60332 L
.2437 .60332 L
.23186 .60332 L
.22067 .60332 L
.21007 .60332 L
.2 .60332 L
.19042 .60332 L
.18127 .60332 L
.17252 .60332 L
.16414 .60332 L
.15608 .60332 L
.14833 .60332 L
.14086 .60332 L
.13364 .60332 L
.12665 .60332 L
.11988 .60332 L
.1133 .60332 L
.1069 .60332 L
.10067 .60332 L
.09459 .60332 L
.08865 .60332 L
.08284 .60332 L
.07715 .60332 L
.07156 .60203 L
.06607 .58644 L
.06067 .53255 L
.05535 .45336 L
.0501 .39947 L
.04491 .38388 L
.03978 .38259 L
.0347 .38259 L
Mistroke
.02967 .38259 L
.02467 .38259 L
.0197 .38259 L
.01475 .38259 L
.00983 .38259 L
.00491 .38259 L
0 .38259 L
Mfstroke
.5 g
.01 w
0 .38259 m
.01975 .38259 L
.04057 .38259 L
.04585 .38259 L
.0515 .38259 L
.05424 .38259 L
.05683 .38259 L
.05796 .38259 L
.05916 .38259 L
.06047 .60332 L
.06168 .60332 L
.08481 .60332 L
.12454 .60332 L
.16684 .60332 L
.20755 .60332 L
.25083 .60332 L
.29253 .60332 L
.33263 .60332 L
.35501 .60332 L
.37531 .60332 L
.38547 .60332 L
.39082 .60332 L
.39374 .60332 L
.39646 .60332 L
.39769 .60332 L
.39885 .60332 L
.40019 .26487 L
.40141 .26487 L
.40684 .26487 L
.41639 .26487 L
.45664 .26487 L
.49946 .26487 L
.54069 .26487 L
.58033 .26487 L
.62254 .26487 L
.66316 .26487 L
.70636 .26487 L
.74797 .26487 L
.76893 .26487 L
.78798 .26487 L
.79306 .26487 L
.79574 .26487 L
.7972 .26487 L
.79857 .26487 L
.79977 .26487 L
.80105 .23224 L
.80178 .23203 L
.80247 .23184 L
.80377 .23147 L
.80857 .23016 L
Mistroke
.83057 .22451 L
.87073 .21568 L
.91346 .20798 L
.9546 .20188 L
.99832 .1965 L
1 .19631 L
Mfstroke
0 .23544 m
.01975 .23544 L
.04057 .23544 L
.04585 .23544 L
.0515 .23544 L
.05424 .23544 L
.05683 .23544 L
.05796 .23544 L
.05916 .23544 L
.06047 .01472 L
.06168 .01472 L
.08481 .01472 L
.12454 .01472 L
.16684 .01472 L
.20755 .01472 L
.25083 .01472 L
.29253 .01472 L
.33263 .01472 L
.35501 .01472 L
.37531 .01472 L
.38547 .01472 L
.39082 .01472 L
.39374 .01472 L
.39646 .01472 L
.39769 .01472 L
.39885 .01472 L
.40019 .05886 L
.40141 .05886 L
.40684 .05886 L
.41639 .05886 L
.45664 .05886 L
.49946 .05886 L
.54069 .05886 L
.58033 .05886 L
.62254 .05886 L
.66316 .05886 L
.70636 .05886 L
.74797 .05886 L
.76893 .05886 L
.78798 .05886 L
.79306 .05886 L
.79574 .05886 L
.7972 .05886 L
.79857 .05886 L
.79977 .05886 L
.80105 .0915 L
.80178 .0917 L
.80247 .0919 L
.80377 .09226 L
.80857 .09358 L
Mistroke
.83057 .09922 L
.87073 .10805 L
.91346 .11575 L
.9546 .12185 L
.99832 .12723 L
1 .12742 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool000ooo`03003ooooo0?ooo`ooo`003ooo00<00?ooool0
oooo3ooo000?ool00`00ooooo`3oool?ool000Ooo`@000Coo`03003ooooo0?ooo`ooo`002Ooo00<0
0?ooool00ooo0`003?oo05ekkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnl0YOoo0009ool00`00ooooo`03ool00`00ooooo`0<ool0GH@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn4402Uool00003ool00000008000Coo`03003ooooo00?oo`03003ooooo00coo`05Nnn44?ooool0
001Bool20002ool00h@@Nnooo`2Tool000Woo`03003ooooo00?oo`03003ooooo00coo`04Q11kkooo
001Eool200000g__Q13oo`2Tool000Ooo`<000Goo`03003ooooo00coo`04Nnn44?oo001Gool00h@@
Nnooo`2Tool000Woo`03003ooooo00?oo`03003ooooo00coo`04Q11kkooo001Gool017__Q1000000
Xooo000?ool00`00ooooo`0<ool017__Q13oo`00Eooo00F447__ooooo`000:;oo`003ooo00<00?oo
ool03?oo00B447__ool005Ooo`03Nnn44?oo00;oo`800:3oo`003ooo0P003Ooo00Akkh@@ool005Oo
o`03Q11kkooo00Coo`8009koo`003ooo00<00?ooool03?oo00>447__0000F?oo00=kkh@@ool01_oo
0P00W?oo000?ool00`00ooooo`0<ool00g__Q100001Hool00h@@Nnooo`08ool3002Iool000ooo`03
003ooooo00coo`03Q11kk`0005Soo`03Nnn44?oo00_oo`H009?oo`003ooo00<00?ooool03?oo00=k
kh@@0000F?oo06j447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__>Ooo000?ool0
0`00ooooo`0<ool00h@@Nnl0001Hool0KW__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q10iool000ooo`03003ooooo00coo`03Nnn440000;?oo`D000coo`03Q11kkooo03Soo`003ooo
00<00?ooool03?oo00>447__0000^?oo1@001ooo00=kkh@@ool0>?oo000?ool2000=ool00g__Q13o
o`2mool50002ool00h@@Nnooo`0hool000ooo`03003ooooo00coo`03Q11kkooo0<;oo`800003Nnn4
400003Soo`003ooo00<00?ooool03?oo00=kkh@@ool0a?oo00>447__ool01000=?oo000?ool00`00
ooooo`0<ool00h@@Nnooo`34ool00g__Q13oo`04ool4000`ool000ooo`03003ooooo00coo`03Nnn4
4?oo0<Coo`04Q11kkoooNnl7ool4000/ool000ooo`03003ooooo00coo`03Q11kkooo0<Coo`07Nnn4
47__Q11kkh@@Nnl02?oo1000:?oo000?ool00`00ooooo`0<ool00g__Q13oo`34ool02X@@Nnn447__
Q11kkh@@Nnn447__2Ooo10009?oo000?ool00`00ooooo`0<ool00h@@Nnooo`3:ool02G__Q11kkh@@
Nnn447__Q11kk`08ool5000Oool000ooo`8000goo`03Nnn44?oo0<goo`0<Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@1ooo1@006_oo000?ool00`00ooooo`0<ool00h@@Nnooo`3Cool02h@@Nnn447__Q11k
kh@@Nnn447__Q1001ooo1@005Ooo000?ool00`00ooooo`0<ool00g__Q13oo`3Hool038@@Nnn447__
Q11kkh@@Nnn447__Q11kk`Koo`D0013oo`003ooo00<00?ooool03?oo00>447__ool0g_oo00]kkh@@
Nnn447__Q11kkh@@Nnn447__00Koo`D000_oo`003ooo00<00?ooool03?oo00=kkh@@ool0hooo00ak
kh@@Nnn447__Q11kkh@@Nnn447__Q105ool50006ool000ooo`03003ooooo00coo`03Q11kkooo0>Wo
o`0CQ11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__00000ooo000?ool00`00ooooo`0<
ool00g__Q13oo`3^ool03X@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnl3ool000ooo`03003ooooo
00coo`03Q11kkooo02;oo`@0037oo`@003?oo`8003Goo`03003ooooo037oo`03Q11kkh@@00?oo`00
3ooo0P003Ooo00=kkh@@ool09?oo00<00?ooool0<?oo00<00?ooool0<ooo00@00?ooool003Coo`03
003ooooo03Coo`40007oo`7oo`003ooo00<00?ooool03?oo00>447__ool09?oo00<00?ooool0<Ooo
00<00?ooool0=Ooo00<00?ooool0;ooo1@00=Ooo0@000Ooo0Ooo000?ool00`00ooooo`0<ool00g__
Q13oo`0Tool00`00ooooo`0bool00`00ooooo`0bool2000bool01000ooooo`00<ooo0`000ooo000?
ool00`00ooooo`0<ool00h@@Nnooo`0Tool00`00ooooo`0cool00`00ooooo`0cool00`00ooooo`0`
ool00`00ool0000cool00`00ooooo`03ool000ooo`03003ooooo00coo`03Nnn44?oo02;oo`<003;o
o`04003ooooo000bool01000ooooo`00<ooo0P00<ooo00<00?ooool00ooo000?ool00`00ooooo`0<
ool00h@@Nnooo`0Tool00`00ooooo`0aool2000dool2000eool00`00ooooo`0aool40002ool000oo
o`03003ooooo00coo`03Nnn44?oo0?ooo`003ooo3`0000>447__0000o0000ooo000?ool00`00oooo
o`08ool01P00ooooooooNnn440Coo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03
003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03
003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03
003ooooo00Soo`03003ooooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03
003ooooo00Soo`03003ooooo00Ooo`03003ooooo00Soo`03003ooooo00Soo`03003ooooo00Soo`03
003ooooo007oo`003ooo00<00?ooool02?oo00H00?oooooooh@@Nnl4ool00`00ooooo`08ool00`00
ooooo`08ool00`00ooooo`08ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`08ool00`00
ooooo`08ool00`00ooooo`07ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00
ooooo`07ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`07ool00`00
ooooo`08ool00`00ooooo`08ool00`00ooooo`08ool00`00ooooo`07ool00`00ooooo`08ool00`00
ooooo`08ool00`00ooooo`08ool00`00ooooo`01ool000ooo`03003ooooo00coo`03Nnn44?oo0?oo
o`003ooo00<00?ooool03?oo00>447__ool0oooo000?ool00`00ooooo`0;ool00`00Nnn4403oool1
ool000ooo`03003ooooo00_oo`03002447__0?ooo`7oo`003ooo00<00?ooool02ooo00<007__Q100
oooo0Ooo000?ool2000<ool00`00Q11kk`3oool1ool000ooo`03003ooooo00_oo`03001kkh@@0?_o
o`03Q13ooooo00;oo`003ooo00<00?ooool02ooo00<008@@Nnl0kooo00ikkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@0ooo000?ool00`00ooooo`0;ool00`00Nnn4403Zool04g__Q11kkh@@Nnn447__
Q11kkh@@Nnn447__Q11kkh@@Nnn447__00?oo`003ooo00<00?ooool02ooo00<008@@Nnl0i?oo00b4
47__Q11kkh@@Nnn447__Q11kkh@@Nnl5ool50006ool000ooo`03003ooooo00[oo`04003oog__Q13O
ool02h@@Nnn447__Q11kkh@@Nnn447__Q1001_oo1@002ooo000?ool00`00ooooo`0:ool01000oon4
47__fOoo00akkh@@Nnn447__Q11kkh@@Nnn447__Q106ool5000@ool000ooo`03003ooooo00[oo`04
003oog__Q13Dool02g__Q11kkh@@Nnn447__Q11kkh@@Nnl01ooo1@005Ooo000?ool2000;ool01000
oon447__c_oo00b447__Q11kkh@@Nnn447__Q11kkh@@Nnl7ool5000Jool000ooo`03003ooooo00[o
o`04003oog__Q13;ool02H@@Nnn447__Q11kkh@@Nnn44008ool5000Oool000ooo`03003ooooo00[o
o`04003ooh@@Nno8ool01h@@Nnn447__Q11kkh@@00Woo`@002Coo`003ooo00<00?ooool02Ooo00D0
0?oooomkkh@@0<Goo`07Q11kkh@@Nnn447__Q1002?oo1000:?oo000>ool04X@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kklGoo`04Nnn447__Q107ool4000/ool000koo`0BNnn447__Q11k
kh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@aOoo00>447__ool01?oo1000<?oo000?ool00`00oooo
o`3Cool00g__Q13oo`04000dool000ooo`03003ooooo0=7oo`800003Q11kk`0003Soo`003ooo0P00
cOoo1@000_oo00=kkh@@ool0>?oo000?ool00`00ooooo`37ool50007ool00h@@Nnooo`0hool000oo
o`03003ooooo0<;oo`D000coo`03Nnn44?oo03Soo`003ooo00<00?ooool0]ooo2`004Ooo00>447__
ool0>?oo000?ool00`00ooooo`1Wool0KW__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q10iool000ooo`03003ooooo06Ooo`1^Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkcWoo`003ooo00<00?ooool0Iooo00=kkh@@ool04ooo0P00Sooo000?ool00`00ooooo`1W
ool00h@@Nnooo`0Bool00`00ooooo`2?ool000ooo`8006Soo`03Nnn44?oo013oo`8009;oo`003ooo
00<00?ooool0Iooo00>447__ool03ooo00<00?ooool0T_oo000?ool00`00ooooo`1Wool00g__Q13o
o`0=ool2002Eool000ooo`03003ooooo06Ooo`03Q11kkooo00coo`03003ooooo09Goo`003ooo00<0
0?ooool0Iooo00=kkh@@ool03?oo00<00?ooool0UOoo000?ool00`00ooooo`1Wool00h@@Nnooo`0;
ool00`00ooooo`2Fool000ooo`03003ooooo06Ooo`03Nnn44?oo00_oo`03003ooooo09Koo`001ooo
10001?oo00<00?ooool0Iooo00>447__ool02ooo00<00?ooool0U_oo0009ool00`00ooooo`03ool3
001Wool00g__Q13oo`0:ool00`00ooooo`2Gool000Woo`03003ooooo00?oo`03003ooooo06Ooo`03
Q11kkooo00[oo`03003ooooo09Ooo`002Ooo00<00?ooool00ooo00<00?ooool0Iooo00=kkh@@ool0
2Ooo00<00?ooool0V?oo0009ool00`00ooooo`03ool00`00ooooo`1Wool00h@@Nnooo`09ool00`00
ooooo`2Hool000Ooo`<000Goo`03003ooooo06Ooo`03Nnn44?oo00Woo`03003ooooo09Soo`002Ooo
00<00?ooool00ooo00<00?ooool0Iooo00>447__ool02?oo00<00?ooool0VOoo000?ool00`00oooo
o`1Wool00g__Q13oo`08ool00`00ooooo`2Iool000ooo`03003ooooo06Ooo`03Q11kkooo00Soo`03
003ooooo09Woo`003ooo0P00J?oo00=kkh@@ool01ooo00<00?ooool0V_oo000?ool00`00ooooo`1W
ool00h@@Nnooo`07ool00`00ooooo`2Jool000ooo`03003ooooo06Ooo`03Nnn44?oo00Ooo`03003o
oooo09[oo`003ooo00<00?ooool0Iooo00>447__ool01_oo00<00?ooool0Vooo000?ool00`00oooo
o`1Wool00g__Q13oo`06ool00`00ooooo`2Kool000ooo`03003ooooo06Ooo`03Q11kkooo00Goo`03
003ooooo09coo`003ooo00<00?ooool0Iooo00=kkh@@ool01Ooo00<00?ooool0W?oo000?ool00`00
ooooo`1Wool00h@@Nnooo`05ool00`00ooooo`2Lool000ooo`8006Soo`03Nnn44?oo00Coo`03003o
oooo09goo`003ooo00<00?ooool0Iooo00>447__ool01?oo00<00?ooool0WOoo000?ool00`00oooo
o`1Wool00g__Q13oo`04ool00`00ooooo`2Mool000ooo`03003ooooo06Ooo`03Q11kkooo00Coo`03
003ooooo09goo`003_oo01:447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@NnmIool00g__
Q13oo`03ool00`00ooooo`2Nool000koo`0BNnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11k
kh@@FOoo00>447__ool00ooo00<00?ooool0W_oo000?ool00`00ooooo`09ool01@00oooooh@@Nnl0
FOoo00=kkh@@ool00ooo00<00?ooool0W_oo000?ool00`00ooooo`0:ool01000oomkkh@@FOoo00>4
47__ool00ooo00<00?ooool0W_oo000?ool2000;ool01000oon447__FOoo00=kkh@@ool00ooo00<0
0?ooool0W_oo000?ool00`00ooooo`0:ool01000oomkkh@@FOoo00>447__ool00ooo00<00?ooool0
W_oo000?ool00`00ooooo`0:ool01000oon447__FOoo00=kkh@@ool00_oo00<00?ooool0Wooo000?
ool00`00ooooo`0:ool01000oomkkh@@FOoo00>447__ool00_oo00<00?ooool0Wooo000?ool00`00
ooooo`0;ool00`00Q11kk`1Iool00g__Q13oo`02ool00`00ooooo`2Oool000ooo`03003ooooo00_o
o`03001kkh@@05Woo`03Q11kkooo00;oo`03003ooooo09ooo`003ooo00<00?ooool02ooo00<008@@
Nnl0FOoo00=kkh@@ool00_oo00<00?ooool0Wooo000?ool00`00ooooo`0;ool00`00Nnn4401Iool0
1H@@Nnoooooo0000X_oo000?ool2000<ool00`00Q11kk`1Iool01G__Q13ooooo0000X_oo000?ool0
0`00ooooo`0;ool00`00Nnn4401Iool01H@@Nnoooooo0000X_oo000?ool00`00ooooo`0;ool00`00
Q11kk`1Iool01G__Q13ooooo0000X_oo000?ool00`00ooooo`0;ool00`00Nnn4401Iool01H@@Nnoo
oooo0000X_oo000?ool00`00ooooo`0<ool00h@@Nnooo`1Hool01G__Q13ooooo0000X_oo000?ool0
0`00ooooo`0<ool00g__Q13oo`1Hool018@@Nnooo`00Xooo0007ool40004ool00`00ooooo`0<ool0
0h@@Nnooo`1Hool017__Q13oo`00Xooo0007ool00`00ooooo`05ool00`00ooooo`0<ool00g__Q13o
o`1Hool018@@Nnooo`00Xooo0008ool00`00ooooo`04ool3000<ool00h@@Nnooo`1Hool017__Q13o
o`00Xooo0009ool00`00ooooo`03ool00`00ooooo`0<ool00g__Q13oo`1Hool018@@Nnooo`00Xooo
000:ool00`00ooooo`02ool00`00ooooo`0<ool00h@@Nnooo`1Hool00g__Q100002Tool000Ooo`04
003ooooo0004ool00`00ooooo`0<ool00g__Q13oo`1Hool00h@@Nnl0002Tool000Soo`8000Goo`03
003ooooo00coo`03Q11kkooo05Soo`03Nnn440000:Coo`003ooo00<00?ooool03?oo00=kkh@@ool0
F?oo00>447__0000Y?oo000?ool00`00ooooo`0<ool00h@@Nnooo`1Hool00g__Q100002Tool000oo
o`8000goo`03Nnn44?oo05Soo`03Q11kk`000:Coo`003ooo00<00?ooool03?oo00>447__ool0F?oo
00=kkh@@ool0Y?oo000?ool00`00ooooo`0<ool00g__Q13oo`1Hool00h@@Nnooo`2Tool000ooo`03
003ooooo00coo`03Q11kkooo05Soo`03Nnn44?oo0:Coo`003ooo00<00?ooool03?oo00=kkh@@ool0
F?oo00>447__ool0Y?oo000?ool00`00ooooo`0<ool00h@@Nnooo`1Hool00g__Q13oo`2Tool000oo
o`03003ooooo00coo`03Nnn44?oo05Soo`03Q11kkooo0:Coo`003ooo00<00?ooool03?oo00>447__
ool0F?oo00=kkh@@ool0Y?oo000?ool2000=ool00g__Q13oo`1Gool00`00Q11kk`2Uool000ooo`03
003ooooo00coo`03Q11kkooo05Ooo`03001kkh@@0:Goo`003ooo00<00?ooool03?oo00=kkh@@ool0
Eooo00<008@@Nnl0YOoo000?ool00`00ooooo`0<ool00h@@Nnooo`1Fool01000oomkkh@@YOoo000?
ool00`00ooooo`0<ool00g__Q13oo`1Fool01000oon447__YOoo000?ool00`00ooooo`0<ool00h@@
Nnooo`1Fool01000oomkkh@@YOoo000?ool00`00ooooo`0<ool00g__Q13oo`1Fool01000oon447__
YOoo000?ool00`00ooooo`0<ool00h@@Nnooo`1Eool01@00ooooog__Q100YOoo000?ool2000=ool0
0g__Q13oo`1Eool01@00oooooh@@Nnl0YOoo000?ool00`00ooooo`0<ool00h@@Nnl0001Eool01@00
ooooog__Q100YOoo000?ool00`00ooooo`0<ool00g__Q100001Dool01P00ooooooooQ11kkjGoo`00
3ooo00<00?ooool03?oo00>447__0000E?oo00H00?ooooooog__Q12Uool000ooo`03003ooooo00co
o`03Nnn4400005Coo`06003ooooooon447__YOoo000?ool00`00ooooo`0<ool00h@@Nnl0001Cool0
0`00ooooo`02ool00g__Q13oo`2Tool000ooo`03003ooooo00coo`03Nnn4400005?oo`03003ooooo
00;oo`03Q11kkooo0:Coo`003ooo00<00?ooool03?oo00>447__0000Dooo00<00?ooool00_oo00=k
kh@@ool0Y?oo000?ool2000=ool00g__Q100001Bool00`00ooooo`03ool00h@@Nnooo`2Tool000oo
o`03003ooooo00coo`04Q11kkooo001Aool00`00ooooo`03ool00g__Q13oo`2Tool000ooo`03003o
oooo00coo`04Nnn44?oo001@ool00`00ooooo`04ool00h@@Nnooo`2Tool000ooo`03003ooooo00co
o`04Q11kkooo001@ool00`00ooooo`04ool00g__Q13oo`2Tool000ooo`03003ooooo00coo`04Nnn4
4?oo001?ool00`00ooooo`05ool00h@@Nnooo`2Tool000ooo`03003ooooo00coo`04Q11kkooo001>
ool00`00ooooo`06ool00g__Q13oo`2Tool000ooo`03003ooooo00coo`04Nnn44?oo001=ool00`00
ooooo`07ool00h@@Nnooo`2Tool000Soo`8000Goo`03003ooooo00coo`05Q11kkoooool0001<ool0
0`00ooooo`07ool00g__Q13oo`2Tool000Ooo`04003ooooo0004ool3000<ool0GG__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn4
47__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kk`2Uool0
00[oo`03003ooooo00;oo`03003ooooo00coo`1MQ11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@
Nnn447__Q11kkh@@Nnn447__Q11kkh@@Nnn447__Q11kkh@@0:Goo`002?oo0P001Ooo00<00?ooool0
oooo3ooo000:ool00`00ooooo`02ool00`00ooooo`3oool?ool000Ooo`04003ooooo0004ool00`00
ooooo`3oool?ool000Soo`800?oooaOoo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo0000
\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.285873, -1.23796, 
  0.0186123, 0.0252969}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\


OPTDesign[p,Radius->r0,Center->k0,Cdecay -> 2];\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[BoxData[
    \("\nParameterization for internal stability:"\)], "Print"],

Cell[BoxData[
    \("  T = A + B * T1, where T1 is RHP stable and"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("A[s]="\[InvisibleSpace]\(\(-5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4\),
      SequenceForm[ "A[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("B[s]="\[InvisibleSpace]\(\(-5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4\),
      SequenceForm[ "B[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -4]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("  "\)], "Print"],

Cell[BoxData[
    \("Processing performance function..."\)], "Print"],

Cell[BoxData[
    \("Sampling Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Smoothing Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Optimization routine Anopt now number-crunching ..."\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 6.8717898926677E+00 :   N/A   : 10.E-01 : 1. E+00 :   N/A   : NON \
: 128"\)], "Print"],

Cell[BoxData[
    \(" 1 : 2.6213878066206E-01 : 1.8E+01 : 1.5E-02 : 0  E+00 : 1.4E-02 : NON \
: 128"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 2.621387806620552E-01"\)], "Print"],

Cell[BoxData[
    \("flat      = 1.5147014939E-02"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 1.43E-02"\)], "Print"],

Cell[BoxData[
    \("Calculating output functions ..."\)], "Print"],

Cell[BoxData[
    \("Resetting options ..."\)], "Print"],

Cell[BoxData[
    \("Done! "\)], "Print"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Trat, the rational approximation to T", "Subsection"],

Cell[CellGroupData[{

Cell["Trat9 = RationalModel[s,DegreeOfDenominator -> 9]", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]0.805310352957018693`\),
      SequenceForm[ "Error = ", 0.80531035295701869],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(-5.`\) + s\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4 + 
      \((\((\(-5.`\) + s)\)\ 
            \((\(4.50319577736618725`\[InvisibleSpace]\) + 
                34.8332052647553513`\ s + 45.5341346094399401`\ s\^2 + 
                39.9900950725826831`\ s\^3 + 14.4099613488268918`\ s\^4 + 
                1.87402179416459766`\ s\^5)\))\)/
        \((\((\(2.`\[InvisibleSpace]\) + s)\)\^4\ 
            \((\(-1.54335440756903441`\) - 7.4885253309155999`\ s - 
                16.5908993000363942`\ s\^2 - 4.83380609779030567`\ s\^3 - 
                1.40600243770370658`\ s\^4 - 0.13741242598496095`\ s\^5)\))
          \)\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["How close are T and its rational approximation Trat?", "Subsection"],

Cell["\<\
The Mma. command ListPlot is used for plotting data.  In the plot \
below the horizontal axis represents the gridpoint number (and not \
frequency).\
\>", "Text"],

Cell[CellGroupData[{

Cell["ListPlot[  Abs[T-Sample[Trat9]],PlotRange -> All]", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0163105 0.00749906 0.0147151 3.44445 [
[.16629 .00222 -6 -9 ]
[.16629 .00222 6 0 ]
[.31627 .00222 -6 -9 ]
[.31627 .00222 6 0 ]
[.46625 .00222 -6 -9 ]
[.46625 .00222 6 0 ]
[.61624 .00222 -6 -9 ]
[.61624 .00222 6 0 ]
[.76622 .00222 -9 -9 ]
[.76622 .00222 9 0 ]
[.9162 .00222 -9 -9 ]
[.9162 .00222 9 0 ]
[.00381 .10083 -30 -4.5 ]
[.00381 .10083 0 4.5 ]
[.00381 .18694 -24 -4.5 ]
[.00381 .18694 0 4.5 ]
[.00381 .27305 -30 -4.5 ]
[.00381 .27305 0 4.5 ]
[.00381 .35916 -18 -4.5 ]
[.00381 .35916 0 4.5 ]
[.00381 .44527 -30 -4.5 ]
[.00381 .44527 0 4.5 ]
[.00381 .53138 -24 -4.5 ]
[.00381 .53138 0 4.5 ]
[.00381 .61749 -30 -4.5 ]
[.00381 .61749 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.16629 .01472 m
.16629 .02097 L
s
[(20)] .16629 .00222 0 1 Mshowa
.31627 .01472 m
.31627 .02097 L
s
[(40)] .31627 .00222 0 1 Mshowa
.46625 .01472 m
.46625 .02097 L
s
[(60)] .46625 .00222 0 1 Mshowa
.61624 .01472 m
.61624 .02097 L
s
[(80)] .61624 .00222 0 1 Mshowa
.76622 .01472 m
.76622 .02097 L
s
[(100)] .76622 .00222 0 1 Mshowa
.9162 .01472 m
.9162 .02097 L
s
[(120)] .9162 .00222 0 1 Mshowa
.125 Mabswid
.05381 .01472 m
.05381 .01847 L
s
.0913 .01472 m
.0913 .01847 L
s
.1288 .01472 m
.1288 .01847 L
s
.20379 .01472 m
.20379 .01847 L
s
.24128 .01472 m
.24128 .01847 L
s
.27878 .01472 m
.27878 .01847 L
s
.35377 .01472 m
.35377 .01847 L
s
.39126 .01472 m
.39126 .01847 L
s
.42876 .01472 m
.42876 .01847 L
s
.50375 .01472 m
.50375 .01847 L
s
.54124 .01472 m
.54124 .01847 L
s
.57874 .01472 m
.57874 .01847 L
s
.65373 .01472 m
.65373 .01847 L
s
.69123 .01472 m
.69123 .01847 L
s
.72872 .01472 m
.72872 .01847 L
s
.80371 .01472 m
.80371 .01847 L
s
.84121 .01472 m
.84121 .01847 L
s
.8787 .01472 m
.8787 .01847 L
s
.95369 .01472 m
.95369 .01847 L
s
.99119 .01472 m
.99119 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.01631 .10083 m
.02256 .10083 L
s
[(0.025)] .00381 .10083 1 0 Mshowa
.01631 .18694 m
.02256 .18694 L
s
[(0.05)] .00381 .18694 1 0 Mshowa
.01631 .27305 m
.02256 .27305 L
s
[(0.075)] .00381 .27305 1 0 Mshowa
.01631 .35916 m
.02256 .35916 L
s
[(0.1)] .00381 .35916 1 0 Mshowa
.01631 .44527 m
.02256 .44527 L
s
[(0.125)] .00381 .44527 1 0 Mshowa
.01631 .53138 m
.02256 .53138 L
s
[(0.15)] .00381 .53138 1 0 Mshowa
.01631 .61749 m
.02256 .61749 L
s
[(0.175)] .00381 .61749 1 0 Mshowa
.125 Mabswid
.01631 .03194 m
.02006 .03194 L
s
.01631 .04916 m
.02006 .04916 L
s
.01631 .06638 m
.02006 .06638 L
s
.01631 .0836 m
.02006 .0836 L
s
.01631 .11805 m
.02006 .11805 L
s
.01631 .13527 m
.02006 .13527 L
s
.01631 .15249 m
.02006 .15249 L
s
.01631 .16972 m
.02006 .16972 L
s
.01631 .20416 m
.02006 .20416 L
s
.01631 .22138 m
.02006 .22138 L
s
.01631 .2386 m
.02006 .2386 L
s
.01631 .25583 m
.02006 .25583 L
s
.01631 .29027 m
.02006 .29027 L
s
.01631 .30749 m
.02006 .30749 L
s
.01631 .32472 m
.02006 .32472 L
s
.01631 .34194 m
.02006 .34194 L
s
.01631 .37638 m
.02006 .37638 L
s
.01631 .3936 m
.02006 .3936 L
s
.01631 .41083 m
.02006 .41083 L
s
.01631 .42805 m
.02006 .42805 L
s
.01631 .46249 m
.02006 .46249 L
s
.01631 .47972 m
.02006 .47972 L
s
.01631 .49694 m
.02006 .49694 L
s
.01631 .51416 m
.02006 .51416 L
s
.01631 .5486 m
.02006 .5486 L
s
.01631 .56583 m
.02006 .56583 L
s
.01631 .58305 m
.02006 .58305 L
s
.01631 .60027 m
.02006 .60027 L
s
.25 Mabswid
.01631 0 m
.01631 .61803 L
s
.008 w
.02381 .01472 Mdot
.03131 .01473 Mdot
.03881 .01489 Mdot
.04631 .01536 Mdot
.05381 .01639 Mdot
.0613 .01815 Mdot
.0688 .02102 Mdot
.0763 .02582 Mdot
.0838 .03439 Mdot
.0913 .04742 Mdot
.0988 .05851 Mdot
.1063 .06001 Mdot
.1138 .05745 Mdot
.1213 .06054 Mdot
.1288 .06724 Mdot
.1363 .0752 Mdot
.14379 .09694 Mdot
.15129 .12831 Mdot
.15879 .16153 Mdot
.16629 .18874 Mdot
.17379 .20061 Mdot
.18129 .20745 Mdot
.18879 .21775 Mdot
.19629 .21522 Mdot
.20379 .23717 Mdot
.21129 .24117 Mdot
.21879 .25924 Mdot
.22628 .27154 Mdot
.23378 .28106 Mdot
.24128 .30048 Mdot
.24878 .3029 Mdot
.25628 .32608 Mdot
.26378 .3248 Mdot
.27128 .34795 Mdot
.27878 .34642 Mdot
.28628 .36634 Mdot
.29378 .3672 Mdot
.30127 .38185 Mdot
.30877 .38648 Mdot
.31627 .39522 Mdot
.32377 .40369 Mdot
.33127 .40738 Mdot
.33877 .41843 Mdot
.34627 .41954 Mdot
.35377 .43092 Mdot
.36127 .43354 Mdot
.36877 .44343 Mdot
.37627 .45354 Mdot
.38376 .46642 Mdot
.39126 .49588 Mdot
.39876 .54193 Mdot
.40626 .58234 Mdot
.41376 .60332 Mdot
.42126 .59187 Mdot
.42876 .53978 Mdot
.43626 .50606 Mdot
.44376 .51118 Mdot
.45126 .48998 Mdot
.45876 .49381 Mdot
.46625 .50218 Mdot
.47375 .49817 Mdot
.48125 .50928 Mdot
.48875 .50773 Mdot
.49625 .51117 Mdot
.50375 .51202 Mdot
.51125 .51117 Mdot
.51875 .50773 Mdot
.52625 .50928 Mdot
.53375 .49817 Mdot
.54124 .50218 Mdot
.54874 .49381 Mdot
.55624 .48998 Mdot
.56374 .51118 Mdot
.57124 .50606 Mdot
.57874 .53978 Mdot
.58624 .59187 Mdot
.59374 .60332 Mdot
.60124 .58234 Mdot
.60874 .54193 Mdot
.61624 .49588 Mdot
.62373 .46642 Mdot
.63123 .45354 Mdot
.63873 .44343 Mdot
.64623 .43354 Mdot
.65373 .43092 Mdot
.66123 .41954 Mdot
.66873 .41843 Mdot
.67623 .40738 Mdot
.68373 .40369 Mdot
.69123 .39522 Mdot
.69873 .38648 Mdot
.70622 .38185 Mdot
.71372 .3672 Mdot
.72122 .36634 Mdot
.72872 .34642 Mdot
.73622 .34795 Mdot
.74372 .3248 Mdot
.75122 .32608 Mdot
.75872 .3029 Mdot
.76622 .30048 Mdot
.77372 .28106 Mdot
.78121 .27154 Mdot
.78871 .25924 Mdot
.79621 .24117 Mdot
.80371 .23717 Mdot
.81121 .21522 Mdot
.81871 .21775 Mdot
.82621 .20745 Mdot
.83371 .20061 Mdot
.84121 .18874 Mdot
.84871 .16153 Mdot
.85621 .12831 Mdot
.8637 .09694 Mdot
.8712 .0752 Mdot
.8787 .06724 Mdot
.8862 .06054 Mdot
.8937 .05745 Mdot
.9012 .06001 Mdot
.9087 .05851 Mdot
.9162 .04742 Mdot
.9237 .03439 Mdot
.9312 .02582 Mdot
.9387 .02102 Mdot
.94619 .01815 Mdot
.95369 .01639 Mdot
.96119 .01536 Mdot
.96869 .01489 Mdot
.97619 .01473 Mdot
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00@_oo10000ooo0P008?oo00<00?ooool0
0_oo0P007ooo0P001?oo0P007_oo0P001?oo0P006ooo10000ooo0P001?oo0P006?oo10000_oo1000
0ooo0P004?oo0012ool00`00ooooo`03ool01000ooooo`007ooo00D00?ooooooo`0000;oo`03003o
oooo01_oo`04003ooooo0002ool01000ooooo`007?oo00@00?ooool000;oo`04003ooooo000Lool0
1@00oooooooo00000_oo00@00?ooool000;oo`03003ooooo01Ooo`05003oooooool00005ool01000
ooooo`003ooo0013ool00`00ooooo`02ool01000ooooo`007?oo1@000_oo00@00?ooool001goo`04
003ooooo0002ool01000ooooo`007?oo00@00?ooool000;oo`04003ooooo000Lool01@00oooooooo
00000_oo00@00?ooool000;oo`03003ooooo01Ooo`03003ooooo00;oo`03003ooooo00;oo`04003o
oooo000?ool004Coo`05003oooooool00002ool00`00ooooo`0Jool01000ooooo`000ooo00@00?oo
ool001goo`<000?oo`04003ooooo000Mool20003ool01000ooooo`007?oo00D00?ooooooo`0000;o
o`04003ooooo0002ool00`00ooooo`0Gool00`00ooooo`03ool01@00oooooooo00000_oo00<00?oo
ool03Ooo000Qool00`00ooooo`0Qool01000ooooo`000_oo00<00?ooool06ooo00<00?oo00000ooo
00@00?ooool001goo`03003ooooo00?oo`04003ooooo000Lool01000ooooo`000_oo00@00?ooool0
01coo`05003oooooool00002ool01000ooooo`000_oo00<00?ooool05ooo00<00?ooool01?oo00@0
0?ooool000;oo`03003ooooo00goo`008Ooo00<00?ooool07_oo00@00?ooool000;oo`04003ooooo
000Nool20003ool01000ooooo`007Ooo00<00?ooool00ooo00@00?ooool001coo`04003ooooo0002
ool01000ooooo`006_oo0`000ooo00@00?ooool000;oo`04003ooooo000Gool30003ool01000oooo
o`000_oo00@00?ooool000ooo`008Ooo00<00?ooool07ooo0P001?oo0P008?oo00<00?ooool00_oo
0P007ooo0`000ooo0P007_oo0P001?oo0P007Ooo00<00?ooool00_oo0P001?oo0P006_oo00<00?oo
ool00_oo0P001?oo0P004?oo000Qool00`00ooooo`3lool001goool000@000008Ooo00P00?oo003o
o`00ool00?oo1`001?oo00<00?ooool01ooo00<00?ooool01_oo00<00?ooool01ooo00<00?ooool0
1ooo00<00?ooool01_oo00<00?ooool01ooo00<00?ooool01ooo00<00?ooool01_oo00<00?ooool0
1ooo00<00?ooool01ooo00<00?ooool01_oo00<00?ooool01ooo00<00?ooool01ooo00<00?ooool0
1_oo00<00?ooool01ooo00<00?ooool01ooo00<00?ooool01ooo00<00?ooool01_oo00<00?ooool0
1ooo00<00?ooool01ooo00<00?ooool01_oo00<00?ooool01ooo00<00?ooool00_oo100000Soo`00
ool00?oo003oo`000_oo00<00?ooool00Ooo000Qool00`00ooooo`09ool00`00ool0000Gool00`00
ooooo`0Tool00`00ooooo`0Tool00`00ooooo`0Sool00`00ooooo`0Tool00`00ooooo`0Tool00`00
ooooo`03ool00`00ool0000>ool0027oo`03003ooooo00coo`800=[oo`8001;oo`008Ooo00<00?oo
ool03Ooo00<00?ooool0fOoo00<00?ooool04?oo000Qool2000?ool2003Fool2000Dool0027oo`03
003ooooo00ooo`03003ooooo0=Goo`03003ooooo01;oo`008Ooo00<00?ooool0o?oo000Qool00`00
ooooo`0@ool2003Bool2000Fool0027oo`8001;oo`03003ooooo0=7oo`03003ooooo01Coo`008Ooo
00<00?ooool0o?oo000Qool00`00ooooo`0Bool70034ool7000Hool0027oo`03003ooooo01?oo`80
0004ool00?oo0035ool01000ool00?oo0P006?oo000Qool2000Jool20030ool2000Oool0027oo`03
003ooooo01[oo`03003ooooo0;ooo`03003ooooo01goo`008Ooo00<00?ooool06ooo0P00_?oo0P00
8Ooo000Qool00`00ooooo`0Lool00`00ooooo`2kool00`00ooooo`0Oool0027oo`03003ooooo0?co
o`008Ooo0P00oOoo000Qool00`00ooooo`3lool0027oo`03003ooooo0?coo`000_oo0P001?oo00<0
0?ooool00ooo0P000ooo10000_oo0`001Ooo00<00?ooool07Ooo0P00^?oo0P008ooo00001Ooo003o
oooo00002?oo00@00?ooool000;oo`03003ooooo00Koo`03003ooooo00;oo`8001ooo`03003ooooo
0;Ooo`03003ooooo027oo`0000Goo`00ooooo`0000Soo`04003ooooo0003ool00`00ooooo`05ool0
0`00ooooo`02ool00`00ooooo`3lool00005ool00?ooool00008ool01000ooooo`001?oo00@00?oo
ooooo`<000Goo`03003ooooo0?coo`0000Goo`00ooooo`0000Soo`04003ooooo0005ool01000oooo
o`001ooo00<00?ooool0o?oo00001Ooo003ooooo00002?oo00@00?ooool000;oo`04003ooooo0002
ool00`00ooooo`05ool00`00ooooo`3lool000;oo`8000[oo`8000Coo`8000?oo`@000Coo`800?go
o`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`0Oool2002dool2000Uool0027oo`03003ooooo
023oo`03003ooooo0;?oo`03003ooooo02?oo`008Ooo0P00oOoo000Qool00`00ooooo`3lool0027o
o`03003ooooo0?coo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0027oo`800?goo`00
8Ooo00<00?ooool0o?oo000Qool00`00ooooo`0Qool2002`ool2000Wool0027oo`03003ooooo02;o
o`03003ooooo0:ooo`03003ooooo02Goo`008Ooo0P00oOoo000Qool00`00ooooo`3lool0027oo`03
003ooooo0?coo`002?oo0P001?oo00<00?ooool00ooo0P000ooo0`001Ooo00<00?ooool0o?oo0007
ool01000ooooo`002?oo00@00?ooool000Goo`03003ooooo00;oo`03003ooooo0?coo`001ooo00@0
0?ooool000Soo`04003ooooo0005ool00`00ooooo`02ool2000Tool2002/ool2000Yool000Ooo`04
003ooooo0008ool01000ooooo`000_oo0`001Ooo00<00?ooool09?oo00<00?ooool0Zooo00<00?oo
ool09ooo0007ool01000ooooo`002?oo00@00?ooool000;oo`03003ooooo00Goo`03003ooooo0?co
o`001ooo00@00?ooool000Soo`04003ooooo0002ool00`00ooooo`05ool00`00ooooo`0Uool2002X
ool2000[ool000Soo`8000[oo`8000?oo`@000Coo`8002Ooo`<00:Coo`800003ool00?oo02[oo`00
8Ooo00<00?ooool0:?oo00<00?ooool0Xooo00<00?ooool0:ooo000Qool00`00ooooo`0[ool2002L
ool2000aool0027oo`03003ooooo02Woo`800003ool00?oo09coo`<002ooo`008Ooo0P00:ooo00<0
0?ooool0Wooo00<00?ooool0;Ooo000Qool00`00ooooo`3lool0027oo`03003ooooo0?coo`008Ooo
00<00?ooool0o?oo000Qool00`00ooooo`0]ool2002Hool2000cool0027oo`8002ooo`<009Coo`80
0003ool00?oo03;oo`008Ooo00<00?ooool0<?oo00<00?ooool0Tooo00<00?ooool0<ooo000Qool0
0`00ooooo`3lool0027oo`03003ooooo0?coo`008Ooo0P00oOoo000Qool00`00ooooo`0`ool2002A
ool2000gool0027oo`03003ooooo037oo`03003ooooo093oo`03003ooooo03Goo`000_oo0P001?oo
00<00?ooool00ooo0P001?oo00<00?ooool00_oo0`001Ooo00<00?ooool0o?oo00001Ooo003ooooo
00002?oo00@00?ooool000?oo`03003ooooo00Goo`03003ooooo00;oo`03003ooooo03;oo`8008ko
o`8003Soo`0000Goo`00ooooo`0000Soo`04003ooooo0004ool00`00ooooo`04ool00`00ooooo`02
ool2000dool00`00ooooo`2=ool00`00ooooo`0fool00005ool00?ooool00008ool01000ooooo`00
1?oo00@00?ooooooo`<000Goo`03003ooooo03Coo`8008[oo`8003[oo`0000Goo`00ooooo`0000So
o`04003ooooo0005ool01000ooooo`001ooo00<00?ooool0=Ooo00<00?ooool0ROoo00<00?ooool0
>?oo00001Ooo003ooooo00002?oo00@00?ooool000;oo`04003ooooo0002ool00`00ooooo`05ool0
0`00ooooo`3lool000;oo`8000[oo`8000?oo`@000;oo`@000Coo`800?goo`008Ooo00<00?ooool0
o?oo000Qool00`00ooooo`0fool20026ool2000lool0027oo`03003ooooo03Ooo`<008;oo`800003
ool00?oo03_oo`008Ooo00<00?ooool0>Ooo00<00?ooool0POoo00<00?ooool0??oo000Qool2003m
ool0027oo`03003ooooo0?coo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0027oo`80
03_oo`@007[oo`@0043oo`008Ooo00<00?ooool0>ooo00<00?oo0000Nooo00<00?oo0000@?oo000Q
ool00`00ooooo`3lool0027oo`03003ooooo0?coo`008Ooo00<00?ooool0o?oo000Qool20011ool2
001bool20016ool0027oo`03003ooooo03koo`800003ool00?oo07;oo`<004Coo`008Ooo00<00?oo
ool0?ooo00<00?ooool0MOoo00<00?ooool0@_oo000>ool20004ool00`00ooooo`02ool40004ool0
0`00ooooo`3lool000goo`04003ooooo000:ool00`00ooooo`03ool2003mool000goo`04003ooooo
000:ool00`00ooooo`03ool00`00ooooo`12ool2001^ool20018ool000goo`04003ooooo000:ool0
0`00ooooo`03ool00`00ooooo`13ool3001Zool200000ooo003oo`17ool000goo`04003ooooo000:
ool00`00ooooo`03ool00`00ooooo`15ool00`00ooooo`1Yool00`00ooooo`18ool000goo`04003o
oooo0008ool30005ool2003mool000koo`8000_oo`03003ooooo00?oo`03003ooooo04Koo`8006Ko
o`8004coo`008Ooo00<00?ooool0Aooo00<00?ooool0IOoo00<00?ooool0B_oo000Qool00`00oooo
o`18ool2001Rool2001>ool0027oo`03003ooooo04Woo`03003ooooo067oo`03003ooooo04coo`00
8Ooo0P00Booo0P00G_oo0P00D?oo000Qool00`00ooooo`1;ool00`00ooooo`1Mool00`00ooooo`1>
ool0027oo`03003ooooo04coo`8005[oo`8005;oo`008Ooo00<00?ooool0COoo0P00Eooo0P0000?o
o`00ool0DOoo000Qool2001?ool00`00ooooo`1Fool00`00ooooo`1Bool0027oo`03003ooooo0?co
o`008Ooo00<00?ooool0Cooo1000D?oo1000EOoo000Qool00`00ooooo`1@ool00`00ool0001Aool0
0`00ool0001Eool0027oo`03003ooooo0?coo`008Ooo0P00E?oo0P00C?oo0P00FOoo000Qool00`00
ooooo`1Dool30018ool200000ooo003oo`1Hool0027oo`03003ooooo05Koo`03003ooooo04Ooo`03
003ooooo05Woo`000_oo0P001?oo00<00?ooool00_oo10000_oo10000_oo0`001Ooo00<00?ooool0
Eooo0P00A?oo0P00GOoo00001Ooo003ooooo00002_oo00D00?ooooooo`0000Soo`03003ooooo00;o
o`8005Woo`03003ooooo04?oo`03003ooooo05_oo`0000Goo`00ooooo`0000[oo`03003ooooo00;o
o`03003ooooo00Goo`03003ooooo00;oo`03003ooooo0?coo`0000Goo`00ooooo`0000[oo`03003o
oooo00?oo`04003oooooool30005ool00`00ooooo`1Iool20010ool2001Oool00005ool00?ooool0
000:ool00`00ooooo`04ool01000ooooo`001ooo00<00?ooool0F_oo00<00?ooool0?ooo00<00?oo
ool0GOoo00001Ooo003ooooo00002?oo0`000ooo00@00?ooool000;oo`03003ooooo00Goo`03003o
oooo0?coo`000_oo0P002ooo00<00?ooool00_oo0P000ooo10001?oo0P00G?oo0P00??oo0P00HOoo
000Qool00`00ooooo`1Lool00`00ooooo`0kool00`00ooooo`1Oool0027oo`03003ooooo0?coo`00
8Ooo00<00?ooool0o?oo000Qool2003mool0027oo`03003ooooo0?coo`008Ooo00<00?ooool0K?oo
0P006_oo0P00L_oo000Qool00`00ooooo`1]ool3000Fool200000ooo003oo`1aool0027oo`03003o
oooo05goo`80013oo`03003ooooo008000koo`8000?oo`03003ooooo00goo`8006?oo`008Ooo0P00
Gooo00<00?ooool03ooo0P0000?oo`00ool03_oo0`004_oo00<00?ooool0HOoo000Qool00`00oooo
o`1Yool20006ool00`00ooooo`0Aool00`00ooooo`03ool2001_ool0027oo`03003ooooo06[oo`03
003ooooo00Ooo`@000Koo`@000[oo`03003ooooo06goo`008Ooo00<00?ooool0Jooo0P002?oo00<0
0?oo00001P0000Coo`00ool000Ooo`80077oo`008Ooo0P00KOoo00<00?ooool02_oo00D00?oo003o
o`0000coo`03003ooooo06ooo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool000Soo`80
00Coo`03003ooooo00;oo`@000;oo`<000Goo`03003ooooo0?coo`001ooo00@00?ooool000[oo`03
003ooooo00Coo`03003ooooo00;oo`800?goo`001ooo00@00?ooool000[oo`03003ooooo00Coo`03
003ooooo00;oo`03003ooooo0?coo`001ooo00@00?ooool000[oo`04003oooooool30005ool00`00
ooooo`1Wool2000Tool2001]ool000Ooo`04003ooooo000:ool01@00oooooooo00001ooo00<00?oo
ool0Gooo0P001ooo00<00?ooool08ooo00<00?ooool01?oo0P00IOoo0007ool01000ooooo`002?oo
0`000ooo00<00?ooool01Ooo00<00?ooool0H?oo00<00?ooool0<ooo00<00?ooool0Hooo0008ool2
000;ool01000oooooooo10001?oo0P00oOoo000Qool00`00ooooo`3lool0027oo`03003ooooo0?co
o`008Ooo00<00?ooool0o?oo000Qool2003mool0027oo`03003ooooo0?coo`008Ooo00<00?ooool0
o?oo000Qool00`00ooooo`3lool0027oo`03003ooooo067oo`80033oo`8006Ooo`008Ooo0P00Hooo
00<00?ooool0;ooo00<00?ooool0IOoo000Qool00`00ooooo`3lool0027oo`03003ooooo06Goo`80
02Soo`8006_oo`008Ooo00<00?ooool0I_oo00<00?ooool09ooo00<00?ooool0JOoo000Qool2001T
ool2000/ool2001Yool0027oo`03003ooooo06Coo`03003ooooo02_oo`03003ooooo06Ooo`008Ooo
00<00?ooool0o?oo0002ool20004ool00`00ooooo`02ool40003ool00`00ooooo`02ool30005ool0
0`00ooooo`3lool00005ool00?ooool0000:ool00`00ooooo`02ool00`00ooooo`05ool00`00oooo
o`02ool00`00ooooo`3lool00005ool00?ooool0000:ool00`00ooooo`03ool00`00ooooo`04ool0
0`00ooooo`02ool2003mool00005ool00?ooool0000:ool00`00ooooo`03ool01000oooooooo0`00
oooo1Ooo00001Ooo003ooooo00002_oo00<00?ooool01?oo00@00?ooool00?ooo`Ooo`0000Goo`00
ooooo`0000Soo`<000?oo`04003ooooo0002ool00`00ooooo`3oool5ool000;oo`8000_oo`04003o
ooooool40002ool4003oool4ool00?ooob7oo`00oooo8Ooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-17.1746, -0.0164061, 
  0.5169, 0.00112537}}],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]],

Cell["\<\
Here is a similar plot, only now the horizontal axis represents \
frequency w\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
ListPlot[ Transpose[{wpts,Abs[T-Sample[Trat9]]}],
          PlotRange -> {{0.,5.},All}]\
\>", "Input",
  PageWidth->Infinity,
  InitializationCell->True,
  ShowSpecialCharacters->False,
  FormatType->InputForm],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.77556e-17 0.2 0.0147151 3.44445 [
[.2 .00222 -3 -9 ]
[.2 .00222 3 0 ]
[.4 .00222 -3 -9 ]
[.4 .00222 3 0 ]
[.6 .00222 -3 -9 ]
[.6 .00222 3 0 ]
[.8 .00222 -3 -9 ]
[.8 .00222 3 0 ]
[1 .00222 -3 -9 ]
[1 .00222 3 0 ]
[-0.0125 .10083 -30 -4.5 ]
[-0.0125 .10083 0 4.5 ]
[-0.0125 .18694 -24 -4.5 ]
[-0.0125 .18694 0 4.5 ]
[-0.0125 .27305 -30 -4.5 ]
[-0.0125 .27305 0 4.5 ]
[-0.0125 .35916 -18 -4.5 ]
[-0.0125 .35916 0 4.5 ]
[-0.0125 .44527 -30 -4.5 ]
[-0.0125 .44527 0 4.5 ]
[-0.0125 .53138 -24 -4.5 ]
[-0.0125 .53138 0 4.5 ]
[-0.0125 .61749 -30 -4.5 ]
[-0.0125 .61749 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 .01472 m
.2 .02097 L
s
[(1)] .2 .00222 0 1 Mshowa
.4 .01472 m
.4 .02097 L
s
[(2)] .4 .00222 0 1 Mshowa
.6 .01472 m
.6 .02097 L
s
[(3)] .6 .00222 0 1 Mshowa
.8 .01472 m
.8 .02097 L
s
[(4)] .8 .00222 0 1 Mshowa
1 .01472 m
1 .02097 L
s
[(5)] 1 .00222 0 1 Mshowa
.125 Mabswid
.04 .01472 m
.04 .01847 L
s
.08 .01472 m
.08 .01847 L
s
.12 .01472 m
.12 .01847 L
s
.16 .01472 m
.16 .01847 L
s
.24 .01472 m
.24 .01847 L
s
.28 .01472 m
.28 .01847 L
s
.32 .01472 m
.32 .01847 L
s
.36 .01472 m
.36 .01847 L
s
.44 .01472 m
.44 .01847 L
s
.48 .01472 m
.48 .01847 L
s
.52 .01472 m
.52 .01847 L
s
.56 .01472 m
.56 .01847 L
s
.64 .01472 m
.64 .01847 L
s
.68 .01472 m
.68 .01847 L
s
.72 .01472 m
.72 .01847 L
s
.76 .01472 m
.76 .01847 L
s
.84 .01472 m
.84 .01847 L
s
.88 .01472 m
.88 .01847 L
s
.92 .01472 m
.92 .01847 L
s
.96 .01472 m
.96 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
0 .10083 m
.00625 .10083 L
s
[(0.025)] -0.0125 .10083 1 0 Mshowa
0 .18694 m
.00625 .18694 L
s
[(0.05)] -0.0125 .18694 1 0 Mshowa
0 .27305 m
.00625 .27305 L
s
[(0.075)] -0.0125 .27305 1 0 Mshowa
0 .35916 m
.00625 .35916 L
s
[(0.1)] -0.0125 .35916 1 0 Mshowa
0 .44527 m
.00625 .44527 L
s
[(0.125)] -0.0125 .44527 1 0 Mshowa
0 .53138 m
.00625 .53138 L
s
[(0.15)] -0.0125 .53138 1 0 Mshowa
0 .61749 m
.00625 .61749 L
s
[(0.175)] -0.0125 .61749 1 0 Mshowa
.125 Mabswid
0 .03194 m
.00375 .03194 L
s
0 .04916 m
.00375 .04916 L
s
0 .06638 m
.00375 .06638 L
s
0 .0836 m
.00375 .0836 L
s
0 .11805 m
.00375 .11805 L
s
0 .13527 m
.00375 .13527 L
s
0 .15249 m
.00375 .15249 L
s
0 .16972 m
.00375 .16972 L
s
0 .20416 m
.00375 .20416 L
s
0 .22138 m
.00375 .22138 L
s
0 .2386 m
.00375 .2386 L
s
0 .25583 m
.00375 .25583 L
s
0 .29027 m
.00375 .29027 L
s
0 .30749 m
.00375 .30749 L
s
0 .32472 m
.00375 .32472 L
s
0 .34194 m
.00375 .34194 L
s
0 .37638 m
.00375 .37638 L
s
0 .3936 m
.00375 .3936 L
s
0 .41083 m
.00375 .41083 L
s
0 .42805 m
.00375 .42805 L
s
0 .46249 m
.00375 .46249 L
s
0 .47972 m
.00375 .47972 L
s
0 .49694 m
.00375 .49694 L
s
0 .51416 m
.00375 .51416 L
s
0 .5486 m
.00375 .5486 L
s
0 .56583 m
.00375 .56583 L
s
0 .58305 m
.00375 .58305 L
s
0 .60027 m
.00375 .60027 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.89064 .04742 Mdot
.79844 .05851 Mdot
.72271 .06001 Mdot
.65931 .05745 Mdot
.60541 .06054 Mdot
.55896 .06724 Mdot
.51848 .0752 Mdot
.48284 .09694 Mdot
.45119 .12831 Mdot
.42286 .16153 Mdot
.39733 .18874 Mdot
.37417 .20061 Mdot
.35305 .20745 Mdot
.33368 .21775 Mdot
.31583 .21522 Mdot
.29932 .23717 Mdot
.28398 .24117 Mdot
.26967 .25924 Mdot
.25628 .27154 Mdot
.2437 .28106 Mdot
.23186 .30048 Mdot
.22067 .3029 Mdot
.21007 .32608 Mdot
.2 .3248 Mdot
.19042 .34795 Mdot
.18127 .34642 Mdot
.17252 .36634 Mdot
.16414 .3672 Mdot
.15608 .38185 Mdot
.14833 .38648 Mdot
.14086 .39522 Mdot
.13364 .40369 Mdot
.12665 .40738 Mdot
.11988 .41843 Mdot
.1133 .41954 Mdot
.1069 .43092 Mdot
.10067 .43354 Mdot
.09459 .44343 Mdot
.08865 .45354 Mdot
.08284 .46642 Mdot
.07715 .49588 Mdot
.07156 .54193 Mdot
.06607 .58234 Mdot
.06067 .60332 Mdot
.05535 .59187 Mdot
.0501 .53978 Mdot
.04491 .50606 Mdot
.03978 .51118 Mdot
.0347 .48998 Mdot
.02967 .49381 Mdot
.02467 .50218 Mdot
.0197 .49817 Mdot
.01475 .50928 Mdot
.00983 .50773 Mdot
.00491 .51117 Mdot
0 .51202 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`00oooo8Ooo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool0057oo`@002ko
o`@002ooo`80037oo`03003ooooo02koo`<000?oo`00Dooo00<00?ooool0;Ooo00<00?ooool0;ooo
00@00?ooool0033oo`03003ooooo037oo`40007oo`7oo`00Dooo00<00?ooool0;_oo00<00?ooool0
<Ooo00<00?ooool0:ooo1@00<_oo0@000Ooo0Ooo001Cool00`00ooooo`0_ool00`00ooooo`0^ool2
000^ool01000ooooo`00<?oo0`000ooo001Cool00`00ooooo`0`ool00`00ooooo`0_ool00`00oooo
o`0/ool00`00ool0000`ool00`00ooooo`03ool0027oo`03003ooooo02goo`<002ooo`04003ooooo
000^ool01000ooooo`00;ooo0P00<?oo00<00?ooool00ooo000Qool00`00ooooo`0_ool00`00oooo
o`0^ool2000`ool2000aool00`00ooooo`0^ool40002ool0027oo`03003ooooo0?coo`008Oooo000
0ooo000Qool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00ooooo`07ool00`00
ooooo`08ool00`00ooooo`01ool0027oo`03003ooooo02ooo`03003ooooo02ooo`03003ooooo02oo
o`03003ooooo02ooo`03003ooooo033oo`03003ooooo007oo`008Ooo00<00?ooool0o?oo000Qool0
0`00ooooo`3lool0027oo`800?goo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0027o
o`03003ooooo0=coo`8001koo`008Ooo0P00g_oo00<00?ooool07?oo000Qool00`00ooooo`2Rool2
001Hool0027oo`03003ooooo09Coo`8000goo`03003ooooo00_oo`80017oo`8003Koo`008Ooo00<0
0?ooool0UOoo00<00?ooool06_oo00<00?ooool04?oo00<00?ooool0=?oo000Qool20029ool2001b
ool0027oo`03003ooooo08Woo`03003ooooo073oo`008Ooo00<00?ooool0O_oo0P00O?oo000Qool0
0`00ooooo`1oool00`00ooooo`1jool0027oo`03003ooooo0?coo`008Ooo0P00oOoo000Qool00`00
ooooo`3lool0027oo`03003ooooo07Goo`8008Goo`000_oo0P001?oo00<00?ooool00ooo0P000ooo
10000_oo0`001Ooo00<00?ooool0M_oo00<00?ooool0Pooo00001Ooo003ooooo00002?oo00@00?oo
ool000;oo`03003ooooo00Koo`03003ooooo00;oo`800?goo`0000Goo`00ooooo`0000Soo`04003o
oooo0003ool00`00ooooo`05ool00`00ooooo`02ool00`00ooooo`3lool00005ool00?ooool00008
ool01000ooooo`001?oo00@00?ooooooo`<000Goo`03003ooooo0?coo`0000Goo`00ooooo`0000So
o`04003ooooo0005ool01000ooooo`001ooo00<00?ooool0o?oo00001Ooo003ooooo00002?oo00@0
0?ooool000;oo`04003ooooo0002ool00`00ooooo`05ool2003mool000;oo`8000[oo`8000Coo`80
00?oo`@000Coo`03003ooooo0?coo`008Ooo00<00?ooool0KOoo0P00SOoo000Qool00`00ooooo`1^
ool00`00ooooo`2;ool0027oo`800?goo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0
027oo`03003ooooo0?coo`008Ooo00<00?ooool0o?oo000Qool2003mool0027oo`03003ooooo0?co
o`008Ooo00<00?ooool0I_oo0P00U?oo000Qool00`00ooooo`1Wool00`00ooooo`2Bool0027oo`80
0?goo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool000Soo`8000Coo`03003ooooo00?o
o`8000?oo`<000Goo`03003ooooo0?coo`001ooo00@00?ooool000Soo`04003ooooo0005ool00`00
ooooo`02ool2001Qool2002Jool000Ooo`04003ooooo0008ool01000ooooo`001Ooo00<00?ooool0
0_oo00<00?ooool0HOoo00<00?ooool0V?oo0007ool01000ooooo`002?oo00@00?ooool000;oo`<0
00Goo`03003ooooo0?coo`001ooo00@00?ooool000Soo`04003ooooo0002ool00`00ooooo`05ool0
0`00ooooo`1Jool2002Pool000Ooo`04003ooooo0008ool01000ooooo`000_oo00<00?ooool01Ooo
00<00?ooool0Fooo00<00?ooool0W_oo0008ool2000:ool20003ool40004ool2001Fool2002Uool0
027oo`03003ooooo05Koo`03003ooooo0:?oo`008Ooo00<00?ooool0Booo0P00[ooo000Qool00`00
ooooo`1<ool01000oooooooo0P00Z_oo000Qool2001Bool00`00ooooo`2Xool0027oo`03003ooooo
0?coo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0027oo`8004Soo`800;?oo`008Ooo
00<00?ooool0@ooo0P000ooo00<00?ooool0/Ooo000Qool00`00ooooo`14ool00`00ooooo`2eool0
027oo`03003ooooo0?coo`008Ooo00<00?ooool0o?oo000Qool20011ool2002jool0027oo`03003o
oooo047oo`03003ooooo0;Soo`008Ooo00<00?ooool0o?oo0002ool20004ool00`00ooooo`03ool2
0004ool00`00ooooo`02ool30005ool00`00ooooo`0lool2002nool00005ool00?ooool00008ool0
1000ooooo`000ooo00<00?ooool01Ooo00<00?ooool00_oo0P00?_oo00<00?ooool0_?oo00001Ooo
003ooooo00002?oo00@00?ooool000Coo`03003ooooo00Coo`03003ooooo00;oo`03003ooooo0?co
o`0000Goo`00ooooo`0000Soo`04003ooooo0004ool01000oooooooo0`001Ooo00<00?ooool0>Ooo
0P00`Ooo00001Ooo003ooooo00002?oo00@00?ooool000Goo`04003ooooo0007ool00`00ooooo`0j
ool00`00ooooo`2oool00005ool00?ooool00008ool01000ooooo`000_oo00@00?ooool000;oo`03
003ooooo00Goo`800?goo`000_oo0P002_oo0P000ooo10000_oo10001?oo00<00?ooool0o?oo000Q
ool00`00ooooo`0fool20034ool0027oo`03003ooooo03?oo`8000;oo`03003ooooo0<;oo`008Ooo
00<00?ooool0=?oo00<00?ooool0aOoo000Qool2003mool0027oo`03003ooooo0?coo`008Ooo00<0
0?ooool0o?oo000Qool00`00ooooo`3lool0027oo`8002ooo`800003ool000000<Woo`008Ooo00<0
0?ooool0;ooo00@00?ooool00<Woo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3lool0027o
o`800?goo`008Ooo00<00?ooool0:_oo1000c_oo000Qool00`00ooooo`0[ool00`00ool0003>ool0
00koo`8000Coo`03003ooooo00;oo`@000Coo`03003ooooo0?coo`003Ooo00@00?ooool000[oo`03
003ooooo00?oo`03003ooooo0?coo`003Ooo00@00?ooool000[oo`03003ooooo00?oo`800?goo`00
3Ooo00@00?ooool000[oo`03003ooooo00?oo`03003ooooo02Goo`@00=?oo`003Ooo00@00?ooool0
00[oo`03003ooooo00?oo`03003ooooo02Koo`03003oo`000=?oo`003Ooo00@00?ooool000Soo`<0
00Goo`03003ooooo0?coo`003_oo0P002ooo00<00?ooool00ooo0P00oOoo000Qool00`00ooooo`0S
ool2003Gool0027oo`03003ooooo027oo`800003ool00?oo0=Koo`008Ooo00<00?ooool08_oo00<0
0?ooool0eooo000Qool2000Pool2003Kool0027oo`03003ooooo023oo`03003ooooo0=Woo`008Ooo
00<00?ooool07_oo0P00g?oo000Qool00`00ooooo`0Lool200000ooo003oo`3Kool0027oo`03003o
oooo01goo`03003ooooo0=coo`008Ooo0P00oOoo000Qool00`00ooooo`0Iool3003Pool0027oo`03
003ooooo01[oo`800>3oo`008Ooo00<00?ooool0o?oo000Qool2000Hool2003Sool0027oo`03003o
oooo01Goo`800003ool00?oo0>;oo`008Ooo00<00?ooool05_oo00<00?ooool0hooo0002ool20004
ool00`00ooooo`02ool40002ool40002ool30005ool00`00ooooo`0Dool2003Vool00005ool00?oo
ool0000:ool01@00oooooooo00002?oo00<00?ooool00_oo0P005_oo00<00?ooool0i?oo00001Ooo
003ooooo00002_oo00<00?ooool00_oo00<00?ooool01Ooo00<00?ooool00_oo00<00?ooool0o?oo
00001Ooo003ooooo00002_oo00<00?ooool00ooo00@00?ooooooo`<000Goo`03003ooooo01;oo`80
0>Soo`0000Goo`00ooooo`0000[oo`03003ooooo00Coo`04003ooooo0007ool00`00ooooo`0Cool0
0`00ooooo`3Vool00005ool00?ooool00008ool30003ool01000ooooo`000_oo00<00?ooool01Ooo
00<00?ooool0o?oo0002ool2000;ool00`00ooooo`02ool20003ool40004ool2000Bool2003Yool0
027oo`03003ooooo01;oo`03003ooooo0>Ooo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`3l
ool0027oo`800?goo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`05ool2003eool0027oo`03
003ooooo00Coo`<000Soo`800>_oo`008Ooo0P000_oo0P000_oo00<00?ooool02?oo00<00?ooool0
jOoo000Qool00`00ooooo`02ool2003hool0027oo`04003oo`000002ool01000oooooooo0P00looo
000Pool300000ooo00000004ool3003cool0027oo`8000;oo`03003ooooo00?oo`03003ooooo0?;o
o`008Ooo0P00oOoo000Qool00`00ooooo`3lool0027oo`03003ooooo0?coo`002?oo0P001?oo00<0
0?ooool00_oo10000_oo0`001Ooo00<00?ooool0o?oo0007ool01000ooooo`002_oo00<00?ooool0
1?oo00<00?ooool00_oo0P00oOoo0007ool01000ooooo`002_oo00<00?ooool01?oo00<00?ooool0
0_oo00<00?ooool02Ooo0P00lOoo0007ool01000ooooo`002_oo00@00?ooooooo`<000Goo`03003o
oooo00[oo`04003oooooool2003/ool000Ooo`04003ooooo000:ool01@00oooooooo00001ooo00<0
0?ooool03ooo00<00?ooool0j_oo0007ool01000ooooo`002?oo0`000ooo00<00?ooool01Ooo0P00
oOoo0008ool2000;ool01000oooooooo10001?oo00<00?ooool0o?oo000Qool00`00ooooo`3lool0
027oo`03003ooooo0?coo`008Ooo0P00oOoo000Qool00`00ooooo`3lool0027oo`03003ooooo0?co
o`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`0=ool2003]ool0027oo`8000ooo`03003ooooo
0>_oo`008Ooo00<00?ooool0o?oo000Qool00`00ooooo`0:ool2003`ool0027oo`03003ooooo00_o
o`03003ooooo0>koo`008Ooo0P003?oo0P00kooo000Qool00`00ooooo`0<ool00`00ooooo`3]ool0
027oo`03003ooooo0?coo`000_oo0P001?oo00<00?ooool00_oo10000ooo00<00?ooool00_oo0`00
1Ooo00<00?ooool0o?oo00001Ooo003ooooo00002_oo00<00?ooool00_oo00<00?ooool01Ooo00<0
0?ooool00_oo0P00oOoo00001Ooo003ooooo00002_oo00<00?ooool00ooo00<00?ooool01?oo00<0
0?ooool00_oo00<00?ooool0o?oo00001Ooo003ooooo00002_oo00<00?ooool00ooo00@00?oooooo
o`<00?ooo`Goo`0000Goo`00ooooo`0000[oo`03003ooooo00Coo`04003ooooo003oool7ool00005
ool00?ooool00008ool30003ool01000ooooo`000_oo00<00?ooool0oooo1Ooo0002ool2000;ool0
1000oooooooo10000_oo1000oooo1?oo003ooolQool00?ooob7oo`00oooo8Ooo003ooolQool00001
\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.660464, -0.0193031, 
  0.0199313, 0.0011573}}],

Cell[BoxData[
    RowBox[{
    \(Graphics::"gptn"\), \( : \ \), 
      "\<\"Coordinate \\!\\(\\*InterpretationBox[\\\"\\\\[Infinity]\\\", \
DirectedInfinity[1]]\\) in \
\\!\\({\\*InterpretationBox[\\\"\\\\[Infinity]\\\", DirectedInfinity[1]], \
0}\\) is not a floating-point number.\"\>"}]], "Message"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["More on Rational Approximation", "Section"],

Cell["\<\
RationalModel[ ] does rational approximation by calling the \
Caratheodory-Fejer approximation algorithm by N.Threfethen, implemented here \
under the name CFApproximant[].  (this is transparent to the user).  As most \
input is typically the default one, results may be somewhat unusual at times. \
 If RationalModel[ ] does not do a good job, you may try the routine \
NewtonFit[ ].  Rational approximation is a subject unto itself, and there are \
many nontrivial aspects to the problem of rational approximation.\
\>", "Text"],

Cell["\<\
Here we attempt to do a rational approximation with low degree \
rationals\
\>", "Text"],

Cell[CellGroupData[{

Cell["Trat3 = RationalModel[s,DegreeOfDenominator -> 3]", "Input"],

Cell[BoxData[
    InterpretationBox[
      \("Error in input: \n  Internal stability requires that the option \n  \
OrderOfDenominator \n  must specify an integer greater than or equal to "\
\[InvisibleSpace]4\),
      SequenceForm[ 
      "Error in input: \n  Internal stability requires that the option \n  \
OrderOfDenominator \n  must specify an integer greater than or equal to ", 
        4],
      Editable->False]], "Print"],

Cell[BoxData[
    \($Failed\)], "Output"]
}, Open  ]],

Cell["Too low!  Now try a rational with denominator with degree 4", "Text"],

Cell[CellGroupData[{

Cell["Trat4 = RationalModel[s,DegreeOfDenominator -> 4]", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]15.070049233509164`\),
      SequenceForm[ "Error = ", 15.070049233509163],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(-\(\(9.37912141739275106`\ 
            \((\(-5.`\) + s)\)\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^4
            \)\)\)], "Output"]
}, Open  ]],

Cell["\<\
The approximation is not good as the error figure suggests.  A plot \
similar to the one produced before is more informative.\
\>", "Text"]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
AutoGeneratedPackage->None,
ScreenStyleEnvironment->"Working",
PrintingStyleEnvironment->"Working",
WindowSize->{550, 652},
WindowMargins->{{4, Automatic}, {-14, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of \
all cells in a given style.  Make modifications to any definition using \
commands in the Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->WindowWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->8,
    FontSize->11],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->5,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the \
Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    CellMargins->{{12, Inherited}, {20, 40}},
    CellGroupingRules->{"TitleGrouping", 0},
    PageBreakBelow->False,
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->36,
    FontWeight->"Bold"],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{24, 10}, {20, 40}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    FontSize->20],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{2, 10}, {15, 30}},
    FontSize->24]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{12, Inherited}, {10, 15}},
    CellGroupingRules->{"TitleGrouping", 10},
    PageBreakBelow->False,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->24],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{24, 10}, {15, 20}},
    LineSpacing->{1, 0},
    FontSize->36],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{8, 10}, {4, 4}},
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{2, 10}, {10, 15}},
    FontSize->18]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    CellMargins->{{12, Inherited}, {10, 20}},
    CellGroupingRules->{"TitleGrouping", 20},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    CellMargins->{{24, 10}, {10, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    CellMargins->{{8, 10}, {8, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    CellMargins->{{2, 10}, {8, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold"],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{40, 10}, {11, 32}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{18, Inherited}, {6, 12}},
    FontSize->12],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{13, 0}, {7, 22}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellDingbat->"\[FilledSmallSquare]",
    CellMargins->{{22, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontSize->14,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{36, 10}, {11, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{16, Inherited}, {6, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {7, 22}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSmallSquare]",
    CellMargins->{{22, Inherited}, {8, 18}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{34, 10}, {11, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{17, Inherited}, {6, 12}},
    FontSize->10],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{9, 0}, {7, 14}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {7, 7}},
    LineSpacing->{1, 3},
    CounterIncrements->"Text"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{24, 10}, {10, 10}},
    LineSpacing->{1, 5}],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{8, 10}, {6, 6}},
    LineSpacing->{1, 1}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{2, 2}, {6, 6}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {6, 6}},
    LineSpacing->{1, 3},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{24, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {5, 5}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{2, 2}, {5, 5}},
    FontSize->7]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellMargins->{{45, 10}, {5, 7}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{11, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{72, Inherited}, {8, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{40, 10}, {2, 3}}],
  
  Cell[StyleData["Input", "Printout"],
    CellMargins->{{39, 0}, {4, 6}},
    FontSize->9]
  }, Closed]],
  
  Cell[StyleData["InputOnly"],
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    StyleMenuListing->None,
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellMargins->{{47, 10}, {7, 5}},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{11, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Output"],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{72, Inherited}, {10, 8}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{41, Inherited}, {3, 2}}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{39, 0}, {6, 4}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{45, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{11, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{72, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{41, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{45, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    CellLabelMargins->{{11, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None,
    FontSize->9],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{72, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0},
    FontSize->9],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{41, Inherited}, {Inherited, Inherited}},
    FontSize->9],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{39, Inherited}, {Inherited, Inherited}},
    FontSize->9]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{4, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    ImageMargins->{{62, Inherited}, {Inherited, 0}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    ImageSize->{175, 175},
    ImageMargins->{{38, Inherited}, {Inherited, 0}}],
  
  Cell[StyleData["Graphics", "Printout"],
    ImageSize->{250, 250},
    ImageMargins->{{30, Inherited}, {Inherited, 0}},
    FontSize->9]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontColor->RGBColor[0, 0, 1]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->9],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontFamily->"Courier",
    FontSize->8,
    FontSlant->"Italic",
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["InlineFormula"],
    CellMargins->{{10, 4}, {0, 8}},
    CellHorizontalScrolling->True,
    ScriptLevel->1,
    SingleLetterItalics->True],
  
  Cell[StyleData["InlineFormula", "Presentation"],
    CellMargins->{{24, 10}, {10, 10}},
    LineSpacing->{1, 5}],
  
  Cell[StyleData["InlineFormula", "Condensed"],
    CellMargins->{{8, 10}, {6, 6}},
    LineSpacing->{1, 1}],
  
  Cell[StyleData["InlineFormula", "Printout"],
    CellMargins->{{2, 0}, {6, 6}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{42, Inherited}, {Inherited, Inherited}},
    CellHorizontalScrolling->True,
    ScriptLevel->0,
    SingleLetterItalics->True,
    StyleMenuListing->None,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    LineSpacing->{1, 5}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    LineSpacing->{1, 1}],
  
  Cell[StyleData["DisplayFormula", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontSize->10,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Times",
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Palette Styles", "Section"],
  
  Cell["\<\
The cells below define styles that define standard \
ButtonFunctions, for use in palette buttons.\
\>", "Text"],
  
  Cell[StyleData["Paste"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, After]}]&)}],
  
  Cell[StyleData["Evaluate"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        SelectionEvaluate[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["EvaluateCell"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionMove[ 
          FrontEnd`InputNotebook[ ], All, Cell, 1], 
        FrontEnd`SelectionEvaluateCreateCell[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["CopyEvaluate"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`SelectionCreateCell[ 
          FrontEnd`InputNotebook[ ], All], 
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionEvaluate[ 
          FrontEnd`InputNotebook[ ], All]}]&)}],
  
  Cell[StyleData["CopyEvaluateCell"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`SelectionCreateCell[ 
          FrontEnd`InputNotebook[ ], All], 
        FrontEnd`NotebookApply[ 
          FrontEnd`InputNotebook[ ], #, All], 
        FrontEnd`SelectionEvaluateCreateCell[ 
          FrontEnd`InputNotebook[ ], All]}]&)}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line \
help system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontColor->RGBColor[0, 0, 1],
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    FontVariations->{"Underline"->False}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Placeholder Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making placeholder \
objects in palette templates.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Placeholder"],
    Editable->False,
    Selectable->False,
    StyleBoxAutoDelete->True,
    Placeholder->True,
    StyleMenuListing->None],
  
  Cell[StyleData["Placeholder", "Presentation"]],
  
  Cell[StyleData["Placeholder", "Condensed"]],
  
  Cell[StyleData["Placeholder", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SelectionPlaceholder"],
    Editable->False,
    Selectable->False,
    StyleBoxAutoDelete->True,
    Placeholder->Primary,
    StyleMenuListing->None,
    DrawHighlighted->True],
  
  Cell[StyleData["SelectionPlaceholder", "Presentation"]],
  
  Cell[StyleData["SelectionPlaceholder", "Condensed"]],
  
  Cell[StyleData["SelectionPlaceholder", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["FormatType Styles", "Section"],
  
  Cell["\<\
The cells below define styles that are mixed in with the styles \
of most cells.  If a cell's FormatType matches the name of one of the styles \
defined below, then that style is applied between the cell's style and its \
own options.\
\>", "Text"],
  
  Cell[StyleData["CellExpression"],
    PageWidth->Infinity,
    CellMargins->{{6, Inherited}, {Inherited, Inherited}},
    ShowCellLabel->False,
    ShowSpecialCharacters->False,
    AllowInlineCells->False,
    AutoItalicWords->{},
    StyleMenuListing->None,
    FontFamily->"Courier",
    Background->GrayLevel[1]],
  
  Cell[StyleData["InputForm"],
    AllowInlineCells->False,
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["OutputForm"],
    PageWidth->Infinity,
    TextAlignment->Left,
    LineSpacing->{1, -5},
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["StandardForm"],
    LineSpacing->{1.25, 0},
    StyleMenuListing->None,
    FontFamily->"Courier"],
  
  Cell[StyleData["TraditionalForm"],
    LineSpacing->{1.25, 0},
    SingleLetterItalics->True,
    TraditionalFunctionNotation->True,
    DelimiterMatching->None,
    StyleMenuListing->None],
  
  Cell["\<\
The style defined below is mixed in to any cell that is in an \
inline cell within another.\
\>", "Text"],
  
  Cell[StyleData["InlineCell"],
    TextAlignment->Left,
    ScriptLevel->1,
    StyleMenuListing->None],
  
  Cell[StyleData["InlineCellEditing"],
    StyleMenuListing->None,
    Background->RGBColor[1, 0.749996, 0.8]]
  }, Closed]]
  }, Open  ]]
  }]
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 44, 0, 148, "Title"],
Cell[1778, 53, 263, 5, 86, "Text"],

Cell[CellGroupData[{
Cell[2066, 62, 43, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[2134, 66, 29, 0, 27, "Input"],
Cell[2166, 68, 126, 2, 49, "Print"],
Cell[2295, 72, 118, 2, 35, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[2462, 80, 45, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[2532, 84, 836, 38, 507, "Input",
  InitializationCell->True],
Cell[3371, 124, 179, 3, 39, "Message"],
Cell[3553, 129, 185, 3, 39, "Message"],

Cell[CellGroupData[{
Cell[3763, 136, 18557, 604, 186, 6261, 448, "GraphicsData", 
"PostScript", "Graphics"],
Cell[22323, 742, 108546, 4371, 233, 73708, 3937, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[130930, 5120, 51, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[131006, 5124, 194, 7, 42, "Input",
  InitializationCell->True],
Cell[131203, 5133, 77, 1, 35, "Print"],
Cell[131283, 5136, 79, 1, 21, "Print"],
Cell[131365, 5139, 299, 9, 37, "Print"],
Cell[131667, 5150, 299, 9, 37, "Print"],
Cell[131969, 5161, 37, 1, 21, "Print"],
Cell[132009, 5164, 69, 1, 21, "Print"],
Cell[132081, 5167, 65, 1, 21, "Print"],
Cell[132149, 5170, 66, 1, 21, "Print"],
Cell[132218, 5173, 86, 1, 21, "Print"],
Cell[132307, 5176, 36, 1, 21, "Print"],
Cell[132346, 5179, 114, 2, 21, "Print"],
Cell[132463, 5183, 109, 2, 21, "Print"],
Cell[132575, 5187, 114, 2, 21, "Print"],
Cell[132692, 5191, 113, 2, 21, "Print"],
Cell[132808, 5195, 113, 2, 21, "Print"],
Cell[132924, 5199, 36, 1, 21, "Print"],
Cell[132963, 5202, 36, 1, 21, "Print"],
Cell[133002, 5205, 42, 1, 21, "Print"],
Cell[133047, 5208, 42, 1, 21, "Print"],
Cell[133092, 5211, 68, 1, 21, "Print"],
Cell[133163, 5214, 63, 1, 21, "Print"],
Cell[133229, 5217, 63, 1, 21, "Print"],
Cell[133295, 5220, 54, 1, 21, "Print"],
Cell[133352, 5223, 67, 1, 21, "Print"],
Cell[133422, 5226, 56, 1, 21, "Print"],
Cell[133481, 5229, 41, 1, 21, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[133571, 5236, 55, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[133651, 5240, 91, 3, 42, "Input"],
Cell[133745, 5245, 177, 3, 21, "Print"],
Cell[133925, 5250, 349, 6, 46, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[134311, 5261, 49, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[134385, 5265, 211, 7, 42, "Input",
  InitializationCell->True],
Cell[134599, 5274, 21401, 630, 296, 4849, 421, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[156015, 5907, 89, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[156129, 5914, 223, 8, 57, "Input",
  InitializationCell->True],
Cell[156355, 5924, 296, 5, 45, "Output"],
Cell[156654, 5931, 19831, 557, 296, 4126, 358, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[176500, 6491, 67, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[176592, 6495, 115, 4, 57, "Input"],
Cell[176710, 6501, 424, 7, 45, "Output"],
Cell[177137, 6510, 20166, 586, 296, 4493, 388, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[197364, 7103, 46, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[197435, 7107, 73, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[197533, 7111, 475, 13, 132, "Input",
  InitializationCell->True],
Cell[198011, 7126, 16468, 682, 186, 7007, 561, "GraphicsData", 
"PostScript", "Graphics"],
Cell[214482, 7810, 130, 3, 27, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[214661, 7819, 65, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[214751, 7823, 283, 9, 72, "Input",
  InitializationCell->True],

Cell[CellGroupData[{
Cell[215059, 7836, 46015, 1472, 216, 21395, 1163, "GraphicsData", 
"PostScript", "Graphics"],
Cell[261077, 9310, 45005, 1425, 216, 21153, 1126, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[306155, 10743, 45, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[306225, 10747, 21, 0, 27, "Input"],
Cell[306249, 10749, 57, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[306343, 10755, 24, 0, 27, "Input"],
Cell[306370, 10757, 59, 1, 29, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[306466, 10763, 20, 0, 27, "Input"],
Cell[306489, 10765, 56, 1, 27, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[306582, 10771, 26, 0, 27, "Input"],
Cell[306611, 10773, 55, 1, 27, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[306715, 10780, 61, 0, 52, "Section"],

Cell[CellGroupData[{
Cell[306801, 10784, 740, 34, 447, "Input",
  InitializationCell->True],
Cell[307544, 10820, 20192, 625, 186, 6259, 448, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[327773, 11450, 182, 8, 57, "Input",
  InitializationCell->True],
Cell[327958, 11460, 77, 1, 35, "Print"],
Cell[328038, 11463, 79, 1, 21, "Print"],
Cell[328120, 11466, 299, 9, 37, "Print"],
Cell[328422, 11477, 299, 9, 37, "Print"],
Cell[328724, 11488, 37, 1, 21, "Print"],
Cell[328764, 11491, 69, 1, 21, "Print"],
Cell[328836, 11494, 65, 1, 21, "Print"],
Cell[328904, 11497, 66, 1, 21, "Print"],
Cell[328973, 11500, 86, 1, 21, "Print"],
Cell[329062, 11503, 36, 1, 21, "Print"],
Cell[329101, 11506, 114, 2, 21, "Print"],
Cell[329218, 11510, 109, 2, 21, "Print"],
Cell[329330, 11514, 114, 2, 21, "Print"],
Cell[329447, 11518, 113, 2, 21, "Print"],
Cell[329563, 11522, 113, 2, 21, "Print"],
Cell[329679, 11526, 36, 1, 21, "Print"],
Cell[329718, 11529, 36, 1, 21, "Print"],
Cell[329757, 11532, 42, 1, 21, "Print"],
Cell[329802, 11535, 42, 1, 21, "Print"],
Cell[329847, 11538, 68, 1, 21, "Print"],
Cell[329918, 11541, 63, 1, 21, "Print"],
Cell[329984, 11544, 63, 1, 21, "Print"],
Cell[330050, 11547, 54, 1, 21, "Print"],
Cell[330107, 11550, 67, 1, 21, "Print"],
Cell[330177, 11553, 56, 1, 21, "Print"],
Cell[330236, 11556, 41, 1, 21, "Print"]
}, Open  ]],

Cell[CellGroupData[{
Cell[330314, 11562, 59, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[330398, 11566, 66, 0, 27, "Input"],
Cell[330467, 11568, 177, 3, 21, "Print"],
Cell[330647, 11573, 663, 11, 103, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[331359, 11590, 74, 0, 45, "Subsection"],
Cell[331436, 11592, 172, 4, 50, "Text"],

Cell[CellGroupData[{
Cell[331633, 11600, 174, 4, 27, "Input",
  InitializationCell->True],
Cell[331810, 11606, 16739, 534, 186, 6238, 400, "GraphicsData", 
"PostScript", "Graphics"],
Cell[348552, 12142, 130, 3, 27, "Output"]
}, Open  ]],
Cell[348697, 12148, 101, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[348823, 12155, 220, 7, 42, "Input",
  InitializationCell->True],
Cell[349046, 12164, 12647, 429, 186, 4382, 322, "GraphicsData", 
"PostScript", "Graphics"],
Cell[361696, 12595, 302, 6, 39, "Message"],
Cell[362001, 12603, 130, 3, 27, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[362192, 12613, 49, 0, 52, "Section"],
Cell[362244, 12615, 540, 8, 122, "Text"],
Cell[362787, 12625, 98, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[362910, 12632, 66, 0, 27, "Input"],
Cell[362979, 12634, 432, 9, 63, "Print"],
Cell[363414, 12645, 41, 1, 27, "Output"]
}, Open  ]],
Cell[363470, 12649, 75, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[363570, 12653, 66, 0, 27, "Input"],
Cell[363639, 12655, 174, 3, 21, "Print"],
Cell[363816, 12660, 152, 3, 46, "Output"]
}, Open  ]],
Cell[363983, 12666, 149, 3, 50, "Text"]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

