(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     79368,       2341]*)
(*NotebookOutlinePosition[     80259,       2372]*)
(*  CellTagsIndexPosition[     80215,       2368]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData["NewtonFit"], "Title",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["Introduction"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "This ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " notebook is the documentation for the package NewtonFit, for treating \
nonlinear L2 data fitting  problems.  The main function is ",
  StyleBox["NewtonFit[ ] ",
    FontWeight->"Bold"],
  ",   which is an  implementation of the Newton algorithm .  "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["An L2 Approximation problem"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell["\<\
The L2 optimization or approximation problem  considered here is

\t\t\tGiven a function  F ( x  , a1 , a2 , . . . , am ) and two sets of n \
numbers each,
\t
\t\t\t(data points) \t \tP   =  { x1 , x2 , x3 , . . . , xn }
 (Prob)\t\t\t(data values)  \t\tV  =  { y1 , y2 , y3 , . . . , yn }
\t
\t\t\tFind ( if it exists ) an m-tuple ( a1 , a2 , a3 , . . . , am ) that \
minimizes
\t
\t\t\t\tchi2( a1 , a2 ,  .  .  .  , am ) =    Sum      (  yk - F ( xk  , a1 , \
a2 , . . . , am ) )^2 
\t\t\t\t\t\t\t\tk=1,n\
\>", "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function Fit[ ]  can be used to treat Prob when the function F( x , a1 , \
. . . ) is linear in the a's.  For the general problem Prob a more powerful \
algorithm is necessary.  One example of this is the  classical Newton method, \
implemented here as the function ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  ".  Another is the well known Gauss-Newton algorithm.  These are iterative \
procedures that when converge produce a local solution to Prob.  This is the \
best one can hope for such a general optimization problem.  In practice, one \
way to go about finding global solutions is to run the algorithm repeatedly \
with different initial guesses, with the hope that the global solution will \
be found by one of these runs.\n\nThe emphasis in these notes is on models F( \
x , a1 , . . . , am ) which are rational functions in x, with coefficients \
given by a1 , a2, a3, . . . am.  The algorithm implemented  here is very \
general in that  in principle it solves Prob for F rational and for many \
other functions not necessarily rational.  The wide range of problems that \
can be treated with this implementation is due to the fact that ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can do both simbolic and numerical calculations.  "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  StyleBox["What does NewtonFit[ ]  do for you?",
    FontWeight->"Bold"],
  "\nThe package NewtonFit.m contains the function  ",
  StyleBox["NewtonFit[]",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  ".  You can\n     ",
  StyleBox["a.",
    FontWeight->"Bold"],
  " Find local solutions to  standard nonlinear L2 data fitting  problems for \
a wide range of models, \n          in particular models  which are rational \
functions of given order.\n     ",
  StyleBox["b.",
    FontWeight->"Bold"],
  " Solve weigthed nonlinear L2 data fitting  problems.\n     ",
  StyleBox["c.",
    FontWeight->"Bold"],
  " Find rational approximations to the data, where  the degree and the  ",
  StyleBox["minimum number of stable zeros \n         and poles",
    FontSlant->"Italic"],
  " is specified a priori."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  StyleBox[
  "Problems you may encounter when using NewtonFit  when treating rational \
models",
    FontWeight->"Bold"],
  ".\n\n",
  StyleBox["1",
    FontWeight->"Bold"],
  ".  ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "  will fail frequently in arriving at  a local solution.  The nature of \
the problem of approximation with rational models is such that for many cases \
there are many narrow, skinny and curved  valleys in parameter space.  This \
produces a high degree of instability in local algorithms, such as Newton and \
Gauss-Newton.  A good initial guess is very important.\n\n",
  StyleBox["2.",
    FontWeight->"Bold"],
  " Even if a local solution is produced  you can never be sure if it is a \
global solution (unless the model is linear in the parameters, i.e., the \
classical linear fit).\n\nIn practice a partial cure for both problems 1. and \
2. is to produce lots of runs, initializing each time at a different location \
in parameter space.  More satisfactory would be to complement the local \
algorithms  with different algorithms having better global behavior.  This is \
not done here though. for lack of time."
}], "SmallText",
  Evaluatable->False,
  PageBreakWithin->False,
  AspectRatioFixed->True],

Cell[TextData[{
  StyleBox["References.",
    FontWeight->"Bold"],
  "\n\nZ. Shafiei and A. T. Shenton, Theory and application of H-intinity \
disk method.  Report No.     \n       MES/ATS/BAE/002/90, Department of \
Mechanical Engineering, University of Liverpool, U. K.\n\t\nLuus, R. (1980) \
Optimization in model reduction\", International Journal of Control, \n       \
Vol. 32, N0. 5, pp. 741-747.\n\t\nP. Gill,  Murray and M. Wright, Practical \
Optimization, Academic Press, 1986.\n\t\n\t"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["1.First Example"], "Section",
  Evaluatable->False,
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[TextData[
"In this example we sample a high order rationalfunction of \"s\" on the \
imaginary axis \" I w \" , and fit a low order rational to the data that we \
obtained.  \nConsider the function \n\n( 1 )\t\tF[ s ] = 1 / ( s^3 + 6 s^2 + \
11 s + 6 ).\n\t\t\nA grid of points  j*w i  on the imaginary axis is given by \
\n\n( 2 ) \t\tw [[ i+1 ]]  = 0.01  ( 1.1 )^i ,      where i = 0, . . . , 99.\n\
\t\t\nand corresponding values v[[ i ]] are generated as \n\n( 3 ) \t\tv[[ i \
]] = F[  I * w[[ i ]]  ].\n\t\t\nFollowing Luus and Shenton-Shafiei, a \
\"reduced\" second order order model \n\n\t\t\t\t\t     a [ 1 ] s  +  1.\n( 4 \
)\t   g[ s , a ]       = \t  -----------------------------------\n\t    \t\t\t\
  a[ 2 ] + a [ 3 ] s + a [ 4 ]  s^2\n\nis sought.  More precisely,"], 
  "SmallText",
  Evaluatable->False,
  PageBreakBelow->True,
  AspectRatioFixed->True],

Cell[TextData[{
  StyleBox["Problem",
    FontWeight->"Bold"],
  ".\n\nGiven  datapoints ( 2 ),  datavalues (  3  ) , and the model (  4  ) \
find parameters a[1], a[2], a[3] , a[4]\nthat minimize\n( 5 )\t\tSum[      \
Abs[  v[[ i ]]  -  g[  I * w[[ i ]]  , a ]  ]^2  ,      { i , 1 , n } ] \n"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData["Load the package"], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(<< "\<NewtonFit.m\>"; \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"First we produce  the grid points and grid values"], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(w = Table[0.01\ \ 1.1\^i, {i, 0, 99}]; \n
    \(F[s_] := 1\/\(s\^3 + 6. \ s\^2 + 11. \ s + 6. \); \)\), 
    \(\(values = F[I\ w]; \)\), 
    \(\(points = I\ w; \)\)}], "Input",
  AspectRatioFixed->True],

Cell[TextData[{
  "The gridpoints s and gridvalues v are combined as a list of pairs pairs ( \
s , v ) called data below.   This is one of the input for ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "[]."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(data = Transpose[{points, values}]; \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"The model is set below in terms of parameters a[1], a[2],   ... and the \
variable s.  "], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(model = \(a[1]\ s + 1. \)\/\(a[2]\ s\^2 + a[3]\ s + a[4]\); \)\)], 
  "Input",
  AspectRatioFixed->True],

Cell[TextData[{
  "We now choose an initial set of parameters to start the iteration.  If \
these are not specified the function ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "[] defaults all of them as 0.  "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(initial1 = {\(-1. \), 5. , 9. , 4. }; \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[{
  "The output of Newton is assigned to the variable output1 below.  This will \
be helpful when manipulating the results of the call to ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "[]."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(output1 = NewtonFit[data, model, s, a, Parameters \[Rule] initial1]\)], 
  "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]0.754419163461352582`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.581506845340728517`\),
      SequenceForm[ 
      0, "                   ", 0.75441916346135263, "         ", 
        0.5815068453407285],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]4.0583840354378653`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.482129412645315547`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.313686041121274605`\),
      SequenceForm[ 
      1, "   ", 4.0583840354378653, "         ", 0.48212941264531556, 
        "         ", 0.31368604112127463],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]3.07859708775156404`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0714609019078547813`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0123888661205283856`
        \),
      SequenceForm[ 
      2, "   ", 3.078597087751564, "         ", 0.071460901907854785, 
        "         ", 0.012388866120528386],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(3\[InvisibleSpace]"   "\[InvisibleSpace]2.09594405152195095`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00824722435888532956`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000601043198262158639`
        \),
      SequenceForm[ 
      3, "   ", 2.095944051521951, "         ", 0.0082472243588853304, 
        "         ", 0.00060104319826215864],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(4\[InvisibleSpace]"   "\[InvisibleSpace]0.389809259163741028`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000208885044555711862`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000327769154408939344`
        \),
      SequenceForm[ 
      4, "   ", 0.38980925916374104, "         ", 0.00020888504455571187, 
        "         ", 0.00032776915440893936],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(5\[InvisibleSpace]"   "\[InvisibleSpace]0.0285536182108090308`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.6621903088433565`*^-6
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000327117421150067366`
        \),
      SequenceForm[ 
      5, "   ", 0.028553618210809029, "         ", 1.6621903088433565*^-06, 
        "         ", 0.00032711742115006738],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(6\[InvisibleSpace]"   "\[InvisibleSpace]0.0000899071574970145981`
        \[InvisibleSpace]"         "\[InvisibleSpace]3.82928691093854922`*^-11
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000327117408386177066`
        \),
      SequenceForm[ 
      6, "   ", 8.9907157497014594*^-05, "         ", 
        3.8292869109385491*^-11, "         ", 0.00032711740838617707],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Parameters 
        \[Rule] {\(-0.134521167400646035`\), 4.86549099652770244`, 
          10.0684113995232893`, 6.01588312172409978`}, 
      SingValHessian 
        \[Rule] {0.401004689233975852`, 0.0583707210569155243`, 
          0.00385516699559253073`, 0.00126541511206969792`}, 
      Cost \[Rule] 0.000327117408386177066`, 
      NormGradient \[Rule] 3.82928691093854922`*^-11, NIterations \[Rule] 6}
      \)], "Output"]
}, Open  ]],

Cell[TextData[{
  "For an explanation of what the diagnostics mean, for example \
SingValHessian, type ?SingValHessian at the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " prompt to get a description.  \n\nThe results of the run above say that a \
local minimum has been found, since the gradient of the objective function  \
is close enough to 0, and the hessian has the correct signature.  This is how \
you form a function with the optimal parameters out of the model:"
}], "SmallText",
  Evaluatable->False,
  PageBreakWithin->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(par = Parameters /. output1; \)\), 
    \(r[s_] = 
      model /. {
          a[j_] \[RuleDelayed] par\[LeftDoubleBracket]j\[RightDoubleBracket]}
          \)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    FractionBox[
      RowBox[{
        RowBox[{
          StyleBox["1.`",
            StyleBoxAutoDelete->True,
            PrintPrecision->1], "\[InvisibleSpace]"}], "-", 
        \(0.134521167400646035`\ s\)}], 
      \(\(6.01588312172409978`\[InvisibleSpace]\) + 10.0684113995232893`\ s + 
        4.86549099652770244`\ s\^2\)]], "Output"]
}, Open  ]],

Cell[TextData[
"A plot is now produced which contains the (discrete) data being \
approximated, and the optimal function r[s]."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
      ListPlot[Transpose[{Re[values], Im[values]}], 
        DisplayFunction \[Rule] Identity], 
      ParametricPlot[{Re[r[I\ t]], Im[r[I\ t]]}, {t, 0, 10. }, 
        DisplayFunction \[Rule] Identity], 
      DisplayFunction \[Rule] $DisplayFunction]; \)\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.191734 4.70785 0.589865 4.88171 [
[.42713 .57737 -12 -9 ]
[.42713 .57737 12 0 ]
[.66252 .57737 -9 -9 ]
[.66252 .57737 9 0 ]
[.89791 .57737 -12 -9 ]
[.89791 .57737 12 0 ]
[.17923 .00406 -30 -4.5 ]
[.17923 .00406 0 4.5 ]
[.17923 .10169 -24 -4.5 ]
[.17923 .10169 0 4.5 ]
[.17923 .19933 -30 -4.5 ]
[.17923 .19933 0 4.5 ]
[.17923 .29696 -30 -4.5 ]
[.17923 .29696 0 4.5 ]
[.17923 .3946 -30 -4.5 ]
[.17923 .3946 0 4.5 ]
[.17923 .49223 -30 -4.5 ]
[.17923 .49223 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.42713 .58987 m
.42713 .59612 L
s
[(0.05)] .42713 .57737 0 1 Mshowa
.66252 .58987 m
.66252 .59612 L
s
[(0.1)] .66252 .57737 0 1 Mshowa
.89791 .58987 m
.89791 .59612 L
s
[(0.15)] .89791 .57737 0 1 Mshowa
.125 Mabswid
.23881 .58987 m
.23881 .59362 L
s
.28589 .58987 m
.28589 .59362 L
s
.33297 .58987 m
.33297 .59362 L
s
.38005 .58987 m
.38005 .59362 L
s
.4742 .58987 m
.4742 .59362 L
s
.52128 .58987 m
.52128 .59362 L
s
.56836 .58987 m
.56836 .59362 L
s
.61544 .58987 m
.61544 .59362 L
s
.7096 .58987 m
.7096 .59362 L
s
.75668 .58987 m
.75668 .59362 L
s
.80375 .58987 m
.80375 .59362 L
s
.85083 .58987 m
.85083 .59362 L
s
.14466 .58987 m
.14466 .59362 L
s
.09758 .58987 m
.09758 .59362 L
s
.0505 .58987 m
.0505 .59362 L
s
.00342 .58987 m
.00342 .59362 L
s
.94499 .58987 m
.94499 .59362 L
s
.99207 .58987 m
.99207 .59362 L
s
.25 Mabswid
0 .58987 m
1 .58987 L
s
.19173 .00406 m
.19798 .00406 L
s
[(-0.12)] .17923 .00406 1 0 Mshowa
.19173 .10169 m
.19798 .10169 L
s
[(-0.1)] .17923 .10169 1 0 Mshowa
.19173 .19933 m
.19798 .19933 L
s
[(-0.08)] .17923 .19933 1 0 Mshowa
.19173 .29696 m
.19798 .29696 L
s
[(-0.06)] .17923 .29696 1 0 Mshowa
.19173 .3946 m
.19798 .3946 L
s
[(-0.04)] .17923 .3946 1 0 Mshowa
.19173 .49223 m
.19798 .49223 L
s
[(-0.02)] .17923 .49223 1 0 Mshowa
.125 Mabswid
.19173 .02847 m
.19548 .02847 L
s
.19173 .05288 m
.19548 .05288 L
s
.19173 .07729 m
.19548 .07729 L
s
.19173 .1261 m
.19548 .1261 L
s
.19173 .15051 m
.19548 .15051 L
s
.19173 .17492 m
.19548 .17492 L
s
.19173 .22374 m
.19548 .22374 L
s
.19173 .24815 m
.19548 .24815 L
s
.19173 .27255 m
.19548 .27255 L
s
.19173 .32137 m
.19548 .32137 L
s
.19173 .34578 m
.19548 .34578 L
s
.19173 .37019 m
.19548 .37019 L
s
.19173 .41901 m
.19548 .41901 L
s
.19173 .44341 m
.19548 .44341 L
s
.19173 .46782 m
.19548 .46782 L
s
.19173 .51664 m
.19548 .51664 L
s
.19173 .54105 m
.19548 .54105 L
s
.19173 .56546 m
.19548 .56546 L
s
.19173 .61427 m
.19548 .61427 L
s
.25 Mabswid
.19173 0 m
.19173 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.97619 .57495 Mdot
.97615 .57346 Mdot
.9761 .57182 Mdot
.97605 .57002 Mdot
.97598 .56803 Mdot
.9759 .56585 Mdot
.97579 .56345 Mdot
.97567 .56081 Mdot
.97552 .55791 Mdot
.97535 .55472 Mdot
.97513 .55121 Mdot
.97487 .54736 Mdot
.97455 .54312 Mdot
.97417 .53846 Mdot
.97371 .53334 Mdot
.97315 .52771 Mdot
.97247 .52153 Mdot
.97166 .51475 Mdot
.97067 .5073 Mdot
.96948 .49913 Mdot
.96804 .49017 Mdot
.9663 .48035 Mdot
.9642 .4696 Mdot
.96166 .45784 Mdot
.95861 .44499 Mdot
.95493 .43097 Mdot
.9505 .4157 Mdot
.94518 .3991 Mdot
.93879 .38111 Mdot
.93112 .36166 Mdot
.92196 .34072 Mdot
.91102 .31828 Mdot
.898 .29436 Mdot
.88255 .26905 Mdot
.8643 .24249 Mdot
.84283 .21491 Mdot
.81774 .18665 Mdot
.78859 .15818 Mdot
.755 .13011 Mdot
.71667 .10321 Mdot
.67339 .0784 Mdot
.62514 .05679 Mdot
.57217 .03956 Mdot
.51499 .02799 Mdot
.45449 .02332 Mdot
.39192 .02662 Mdot
.32888 .03871 Mdot
.26727 .05997 Mdot
.20914 .09028 Mdot
.15653 .12888 Mdot
.11128 .17443 Mdot
.07482 .22507 Mdot
.04806 .27855 Mdot
.0312 .33246 Mdot
.02381 .38448 Mdot
.02482 .43256 Mdot
.03273 .47514 Mdot
.04576 .51125 Mdot
.06206 .5405 Mdot
.07993 .56307 Mdot
.09792 .57955 Mdot
.11495 .59081 Mdot
.1303 .59787 Mdot
.14358 .60173 Mdot
.15468 .60329 Mdot
.16369 .60332 Mdot
.17082 .60242 Mdot
.17634 .60104 Mdot
.18053 .59947 Mdot
.18365 .59791 Mdot
.18596 .59646 Mdot
.18764 .59519 Mdot
.18884 .59411 Mdot
.18971 .59321 Mdot
.19032 .59248 Mdot
.19075 .5919 Mdot
.19105 .59143 Mdot
.19126 .59107 Mdot
.19141 .59079 Mdot
.19151 .59057 Mdot
.19158 .5904 Mdot
.19163 .59027 Mdot
.19166 .59017 Mdot
.19168 .5901 Mdot
.1917 .59004 Mdot
.19171 .59 Mdot
.19172 .58997 Mdot
.19172 .58994 Mdot
.19173 .58992 Mdot
.19173 .58991 Mdot
.19173 .5899 Mdot
.19173 .58989 Mdot
.19173 .58988 Mdot
.19173 .58988 Mdot
.19173 .58988 Mdot
.19173 .58987 Mdot
.19173 .58987 Mdot
.19173 .58987 Mdot
.19173 .58987 Mdot
.19173 .58987 Mdot
.5 Mabswid
.9743 .58986 m
.97404 .57176 L
.97334 .5553 L
.97199 .53634 L
.97018 .51848 L
.9657 .48693 L
.95901 .4531 L
.94951 .41649 L
.93706 .37864 L
.90893 .3138 L
.87276 .25198 L
.83315 .19964 L
.78443 .14922 L
.72734 .10381 L
.66697 .06789 L
.63202 .05182 L
.59939 .03961 L
.56867 .03043 L
.53993 .02379 L
.52415 .02092 L
.50722 .01845 L
.49199 .01676 L
.47811 .01566 L
.46542 .01502 L
.45197 .01472 L
.43942 .01479 L
.42813 .01515 L
.41506 .01591 L
.40777 .01651 L
.40107 .01716 L
.37629 .02045 L
.35341 .02476 L
.32915 .03072 L
.28335 .04616 L
.2452 .06368 L
.20933 .08468 L
.15582 .12639 L
.11215 .17381 L
.07837 .22502 L
.06636 .24883 L
.05725 .27045 L
.04506 .30833 L
.04054 .328 L
.03759 .34511 L
.03645 .35367 L
.03549 .36279 L
.03511 .36726 L
.03478 .37198 L
.03453 .37636 L
.03436 .38027 L
Mistroke
.03424 .38418 L
.03418 .3878 L
.03415 .39156 L
.03415 .39364 L
.03417 .39554 L
.03422 .39869 L
.03432 .40209 L
.03445 .40558 L
.03461 .40878 L
.03509 .41598 L
.03565 .4223 L
.03731 .43571 L
.03911 .4465 L
.04137 .45741 L
.04608 .47523 L
.05696 .50459 L
.06898 .52797 L
.0891 .55626 L
.10695 .57428 L
.12071 .58482 L
.1323 .59176 L
.14115 .59592 L
.14802 .59846 L
.15396 .60016 L
.15864 .60116 L
.16085 .60152 L
.16275 .60177 L
.16447 .60195 L
.16592 .60206 L
.16664 .6021 L
.16729 .60213 L
.16765 .60214 L
.16797 .60215 L
.16815 .60215 L
.16835 .60216 L
.16851 .60216 L
.16868 .60216 L
.16884 .60216 L
.16891 .60216 L
.16898 .60216 L
.16905 .60216 L
.16912 .60216 L
.1692 .60217 L
.16927 .60217 L
.16931 .60217 L
.16935 .60217 L
.16939 .60217 L
.16944 .60217 L
.16952 .60216 L
.16956 .60216 L
Mistroke
.1696 .60216 L
.16974 .60216 L
.16982 .60216 L
.1699 .60216 L
.1702 .60216 L
.17037 .60215 L
.17052 .60215 L
.17109 .60213 L
.17166 .60211 L
.17217 .60208 L
.17307 .60202 L
.17399 .60194 L
.17477 .60185 L
.17627 .60164 L
.17754 .60142 L
.17873 .60116 L
.1797 .60092 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  Evaluatable->False,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True,
  ImageSize->{282, 174},
  ImageMargins->{{34, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004J0000[Q000`40n007h00Ooooo6ooo003ooolK
ool00?oooa_oo`00oooo6ooo000Nool20004ool00`00ooooo`02ool40002ool40004ool00`00oooo
o`3Pool001goo`04003ooooo000:ool01@00oooooooo00001ooo0`00h?oo000Mool01000ooooo`00
2_oo00<00?ooool00_oo00<00?ooool01?oo00<00?ooool0h?oo000Gool40002ool01000ooooo`00
2_oo00<00?ooool00ooo00<00?ooool00ooo00<00?ooool0h?oo000Mool01000ooooo`002_oo00<0
0?ooool01?oo00<00?ooool00_oo00<00?ooool0>?oo6P00S_oo000Mool01000ooooo`002?oo0`00
0ooo00@00?ooool000Coo`03003ooooo02ooo`T001[oo`D008Woo`007_oo0P002ooo00<00?ooool0
0_oo0P001Ooo00<00?ooool0;?oo0`005_oo0P004?oo2000POoo000gool00`00ooooo`0Wool50008
ool2000@ool00`00ooooo`0=ool20007ool4001mool003Ooo`8002?oo`D000koo`03003ooooo01oo
o`03003ooooo00Woo`@007Woo`00=ooo00<00?ooool08?oo0P00AOoo0P00Mooo000gool00`00oooo
o`0Mool30002ool20011ool20002ool3001dool003Ooo`03003ooooo01_oo`8000Koo`03003ooooo
043oo`03003ooooo00?oo`<0077oo`00=ooo00<00?ooool06Ooo0P00E?oo0`00K_oo000gool00`00
ooooo`0Fool3001Iool3001[ool003Ooo`8001Coo`<005ooo`8006Woo`00=ooo00<00?ooool04Ooo
0P00HOoo0P0000?oo`000000Iooo000gool00`00ooooo`0>ool300000ooo0000001Qool01000oooo
oooo0P00IOoo000gool00`00ooooo`0<ool20005ool00`00ooooo`1Uool2001Sool003Ooo`03003o
oooo00[oo`80077oo`80067oo`00=ooo00<00?ooool02?oo0P00MOoo0P00Gooo000gool00`00oooo
o`07ool00`00ooooo`1gool2001Mool003Ooo`8000Koo`8007[oo`@005_oo`00=ooo00<00?ooool0
0ooo0P00OOoo00<00?ooool00P00FOoo000gool00`00ooooo`02ool00`00ooooo`22ool00`00oooo
o`1Fool003Ooo`04003oooooool30024ool2001Fool003Ooo`04003ooooo0002ool00`00ooooo`24
ool2001Dool002Coo`8000Coo`03003ooooo00;oo`@000Coo`<008coo`8005;oo`008ooo00@00?oo
ool000[oo`03003ooooo00?oo`03003ooooo08koo`80053oo`008ooo00@00?ooool000[oo`03003o
oooo00;oo`@008goo`800003ool00?oo04koo`007Ooo10000_oo00@00?ooool000[oo`07003ooooo
ool00?oo0000T?oo00@00?ooool004koo`008ooo00@00?ooool000[oo`04003ooooo0002ool00`00
ooooo`2Bool2001<ool002?oo`04003ooooo0008ool300000ooo003oo`02ool00`00ooooo`2Dool0
0`00ooooo`19ool002Coo`8000_oo`8000Coo`03003ooooo09Goo`03003ooooo04Soo`00<?oo00<0
0?ooool01?oo00<00?ooool0U_oo00<00?ooool0Aooo000_ool00`00ooooo`05ool2002Hool00`00
ooooo`16ool002goo`8000Soo`03003ooooo09Ooo`<004Koo`00;Ooo0P002?oo00<00?ooool0V?oo
00<00?oo0000AOoo000/ool00`00ooooo`08ool00`00ooooo`2Kool00`00ooooo`12ool002_oo`03
003ooooo00Woo`03003ooooo09coo`03003ooooo047oo`00:_oo00<00?ooool02_oo00<00?ooool0
WOoo0P00@Ooo000Yool00`00ooooo`0;ool00`00ooooo`2Oool00`00ooooo`0nool002Soo`03003o
oooo00coo`800:7oo`03003ooooo03goo`00:?oo00<00?ooool03?oo00<00?ooool0XOoo00<00?oo
ool0??oo000Wool00`00ooooo`0=ool00`00ooooo`2Qool2000mool002Koo`03003ooooo00koo`03
003ooooo0:;oo`8003coo`009Ooo00<00?ooool03ooo00<00?ooool0Y?oo00<00?ooool0>Ooo000T
ool00`00ooooo`0@ool00`00ooooo`2Uool00`00ooooo`0hool0027oo`<001?oo`03003ooooo0:Ko
o`03003ooooo03Ooo`008_oo00<00?ooool04_oo0P00Yooo00<00?ooool0=ooo000Qool00`00oooo
o`0Cool00`00ooooo`2Wool00`00ooooo`0fool0027oo`03003ooooo01?oo`03003ooooo0:Soo`03
003ooooo03Goo`008?oo00<00?ooool05?oo00<00?ooool0ZOoo0P00=Ooo000Oool00`00ooooo`0E
ool00`00ooooo`2Zool00`00ooooo`0cool001koo`8000Coo`03003ooooo00?oo`8000Coo`8000Go
o`03003ooooo0:_oo`03003ooooo03;oo`007Ooo0P0000?oo`00ool01ooo00@00?ooool000;oo`04
003ooooo0004ool00`00ooooo`2/ool00`00ooooo`0aool001goo`04003ooooo0008ool01000oooo
o`000_oo00@00?ooool000Coo`<00:goo`03003ooooo033oo`005ooo100000?oo`0000000_oo00<0
0?ooool01_oo00@00?ooool000?oo`8000Goo`03003ooooo0:koo`03003ooooo02ooo`007?oo0P00
0_oo00<00?ooool01_oo00@00?ooool000;oo`04003ooooo0004ool00`00ooooo`2_ool00`00oooo
o`0^ool001_oo`03003oo`0000;oo`03003ooooo00Koo`04003ooooo0002ool01000ooooo`001?oo
00<00?ooool0[ooo0`00;_oo000Jool01000oooooooo0P002_oo0P001?oo0P001Ooo00<00?ooool0
/?oo0P00;_oo000Iool00`00ooooo`0Kool00`00ooooo`2aool00`00ooooo`0/ool001Ooo`<001go
o`800;?oo`03003ooooo02_oo`006?oo00<00?ooool07?oo00<00?ooool0/_oo00<00?ooool0:ooo
000Hool00`00ooooo`0Lool00`00ooooo`2cool00`00ooooo`0Zool001Ooo`03003ooooo01goo`03
003ooooo0;Coo`03003ooooo02Woo`005ooo00<00?ooool07Ooo00<00?ooool0]Ooo00<00?ooool0
:?oo000Fool00`00ooooo`0Nool00`00ooooo`2eool2000Yool001Koo`03003ooooo01koo`03003o
oooo0;Koo`03003ooooo02Ooo`005Ooo00<00?ooool07ooo0P00^?oo00<00?ooool09_oo000Eool0
0`00ooooo`0Oool00`00ooooo`2hool00`00ooooo`0Uool001Coo`03003ooooo023oo`03003ooooo
0;Woo`03003ooooo02Coo`005?oo00<00?ooool08?oo00<00?ooool0^Ooo00<00?ooool09?oo000C
ool00`00ooooo`0Qool00`00ooooo`2jool00`00ooooo`0Sool001?oo`03003ooooo027oo`03003o
oooo0;[oo`8002Coo`004_oo00<00?ooool08_oo00<00?ooool0^ooo00<00?ooool08_oo000Bool0
0`00ooooo`0Rool2002mool00`00ooooo`0Qool000ooo`<002Goo`03003ooooo0;coo`03003ooooo
027oo`004?oo0P009Ooo00<00?ooool0_Ooo00<00?ooool08?oo000Aool00`00ooooo`0Sool00`00
ooooo`2mool00`00ooooo`0Pool0017oo`03003ooooo02?oo`03003ooooo0;koo`03003ooooo01oo
o`004?oo00<00?ooool02ooo0P001?oo00<00?ooool00ooo0P001?oo0P001Ooo00<00?ooool0__oo
0`007ooo000@ool00`00ooooo`0:ool01000ooooo`002?oo00@00?ooool000;oo`04003ooooo0004
ool3002oool2000Oool0013oo`03003ooooo00[oo`04003ooooo0008ool01000ooooo`000_oo00@0
0?ooool000Coo`03003ooooo0<3oo`03003ooooo01goo`004?oo00<00?ooool01?oo10000_oo00@0
0?ooool000Soo`04003ooooo0002ool30005ool00`00ooooo`30ool00`00ooooo`0Mool000ooo`03
003ooooo00_oo`04003ooooo0008ool01000ooooo`000_oo00<00?ooool01Ooo00<00?ooool0`Ooo
00<00?ooool07?oo000?ool00`00ooooo`0;ool01000ooooo`002?oo00@00?ooool000;oo`03003o
oooo00Goo`03003ooooo0<7oo`03003ooooo01coo`003ooo00<00?ooool03?oo0P002_oo0P001?oo
0`001?oo00<00?ooool0`_oo00<00?ooool06ooo000?ool00`00ooooo`0Uool00`00ooooo`32ool2
000Lool000koo`03003ooooo02Koo`800<Coo`03003ooooo01[oo`003_oo00<00?ooool09_oo00<0
0?ooool0`ooo00<00?ooool06_oo000>ool00`00ooooo`0Vool00`00ooooo`34ool00`00ooooo`0I
ool000_oo`800003ool00?oo02Ooo`03003ooooo0<Coo`03003ooooo01Woo`003?oo00<00?oo0000
:?oo00<00?ooool0aOoo00<00?ooool06?oo000=ool00`00ooooo`0Wool00`00ooooo`35ool2000I
ool000goo`03003ooooo02Ooo`03003ooooo0<Koo`03003ooooo01Ooo`003Ooo00<00?ooool09ooo
0P00aooo00<00?ooool05ooo000=ool00`00ooooo`0Wool00`00ooooo`36ool00`00ooooo`0Gool0
00goo`03003ooooo02Ooo`03003ooooo0<Ooo`03003ooooo01Koo`003Ooo00<00?ooool09ooo00<0
0?ooool0aooo00<00?ooool05_oo000=ool00`00ooooo`0Wool00`00ooooo`38ool2000Fool000go
o`03003ooooo02Ooo`03003ooooo0<Soo`8001Koo`003?oo00<00?ooool0:?oo0P00b_oo00<00?oo
ool05?oo000<ool00`00ooooo`0Xool00`00ooooo`39ool00`00ooooo`0Dool000coo`03003ooooo
02Soo`03003ooooo0<[oo`03003ooooo01?oo`003?oo00<00?ooool0:?oo00<00?ooool0b_oo0P00
5?oo0009ool200000ooo003oo`0Yool00`00ooooo`3:ool2000Dool000[oo`03003oo`0002[oo`03
003ooooo0<_oo`03003ooooo01;oo`003?oo00<00?ooool03ooo0P001?oo00<00?ooool00ooo0P00
1Ooo00<00?ooool00ooo00<00?ooool0booo00<00?ooool04_oo000<ool00`00ooooo`0>ool01000
ooooo`002?oo00@00?ooool000Coo`03003ooooo00?oo`<00<_oo`03003ooooo01;oo`003?oo00<0
0?ooool03_oo00@00?ooool000Soo`05003ooooo003oo`050004ool00`00ooooo`3;ool3000Bool0
00coo`03003ooooo00Soo`@000;oo`04003ooooo0008ool01P00ooooo`00ool000;oo`03003ooooo
00?oo`03003ooooo0<coo`8001;oo`003?oo00<00?ooool03_oo00@00?ooool000Soo`04003ooooo
0002ool00`00ool00005ool00`00ooooo`3<ool00`00ooooo`0Aool000coo`03003ooooo00koo`04
003ooooo0008ool01000ooooo`000ooo0P001Ooo00<00?ooool0c?oo00<00?ooool04Ooo000<ool0
0`00ooooo`0?ool2000:ool20005ool00`00ooooo`03ool00`00ooooo`3=ool00`00ooooo`0@ool0
00coo`03003ooooo02Soo`03003ooooo0<goo`80017oo`003?oo00<00?ooool0:?oo0P00c_oo0P00
4Ooo000=ool00`00ooooo`0Wool00`00ooooo`3>ool00`00ooooo`0?ool000goo`03003ooooo02Oo
o`03003ooooo0<koo`03003ooooo00ooo`002Ooo0P000_oo00<00?ooool09ooo00<00?ooool0c_oo
0P004?oo000:ool01000ooooo`00:Ooo00<00?ooool0c_oo0P004?oo000=ool00`00ooooo`0Wool0
0`00ooooo`3?ool00`00ooooo`0>ool000koo`03003ooooo02Koo`03003ooooo0<ooo`03003ooooo
00koo`003_oo00<00?ooool09_oo0P00d?oo0P003ooo000>ool00`00ooooo`0Vool00`00ooooo`3@
ool00`00ooooo`0=ool000koo`03003ooooo02Koo`03003ooooo0=3oo`03003ooooo00goo`003_oo
00<00?ooool09_oo00<00?ooool0d?oo0P003_oo000>ool00`00ooooo`0Vool00`00ooooo`3@ool2
000>ool000koo`03003ooooo02Koo`03003ooooo0=7oo`03003ooooo00coo`003ooo00<00?ooool0
9Ooo0P00d_oo0P003Ooo000?ool00`00ooooo`0Uool00`00ooooo`3Aool2000=ool000_oo`8000;o
o`03003ooooo02Goo`03003ooooo0=7oo`03003ooooo00coo`003?oo00@00?ooool002Ooo`03003o
oooo0=7oo`8000goo`004?oo00<00?ooool09?oo00<00?ooool0d_oo00<00?ooool02ooo000@ool0
0`00ooooo`0Tool00`00ooooo`3Bool00`00ooooo`0;ool0013oo`03003ooooo00_oo`8000Coo`03
003ooooo00?oo`8000?oo`@000Coo`03003ooooo0=;oo`8000coo`004Ooo00<00?ooool02Ooo00@0
0?ooool000Soo`04003ooooo0002ool00`00ooooo`05ool3003Bool2000<ool0017oo`03003ooooo
00Woo`04003ooooo0008ool01000ooooo`000ooo00<00?ooool01?oo00<00?ooool0d_oo0P003?oo
000Bool00`00ooooo`02ool40002ool01000ooooo`002?oo00@00?ooool000Coo`03003ooooo00?o
o`03003ooooo0=;oo`8000coo`004_oo00<00?ooool02?oo00@00?ooool000Soo`04003ooooo0005
ool00`00ooooo`02ool00`00ooooo`3Bool00`00ooooo`0;ool001?oo`03003ooooo00Ooo`04003o
oooo0008ool01000ooooo`000_oo00@00?ooool000Coo`03003ooooo0=?oo`8000_oo`003ooo0P00
0_oo00<00?ooool02?oo0P002_oo0P001?oo0P001Ooo00<00?ooool0dooo0P002ooo000@ool01@00
oooooooo00008_oo00<00?ooool0dooo0P002ooo000Dool00`00ooooo`0Pool2003Dool2000;ool0
01Goo`03003ooooo01ooo`03003ooooo0=?oo`8000_oo`005Ooo00<00?ooool07ooo00<00?ooool0
dooo0P002ooo000Fool00`00ooooo`0Nool00`00ooooo`3Dool00`00ooooo`09ool001Ooo`03003o
oooo01goo`03003ooooo0=?oo`8000_oo`005ooo00<00?ooool07Ooo00<00?ooool0e?oo0P002_oo
000Cool20003ool00`00ooooo`0Lool2003Eool2000:ool001Coo`05003oooooool0000Nool00`00
ooooo`3Dool2000:ool001Woo`03003ooooo01_oo`03003ooooo0=Coo`8000[oo`006_oo00<00?oo
ool06_oo00<00?ooool0e?oo0P002_oo000Jool00`00ooooo`0Jool00`00ooooo`3Dool2000:ool0
01_oo`03003ooooo01Woo`03003ooooo03Coo`8000Coo`03003ooooo00?oo`8000?oo`<0033oo`80
00Coo`03003ooooo00;oo`@002koo`8000Coo`03003ooooo00;oo`@000;oo`<000_oo`8000[oo`00
6?oo0P000_oo00<00?ooool06?oo00<00?ooool0<ooo00@00?ooool000Soo`04003ooooo0005ool0
0`00ooooo`0/ool01000ooooo`002_oo00<00?ooool0;?oo00@00?ooool000[oo`03003ooooo00Co
o`03003ooooo00Soo`8000[oo`006Ooo00@00?ooooooo`8001Soo`8003Coo`04003ooooo0008ool0
1000ooooo`001Ooo00<00?ooool0;?oo00@00?ooool000[oo`03003ooooo02coo`04003ooooo000:
ool00`00ooooo`04ool00`00ooooo`08ool2000:ool001ooo`03003ooooo01Goo`03003ooooo03?o
o`04003ooooo0008ool01000ooooo`000_oo0`00;ooo00@00?ooool000[oo`03003ooooo02coo`04
003ooooo000:ool01000oooooooo0`002ooo0P002_oo000Pool00`00ooooo`0Dool00`00ooooo`0c
ool01000ooooo`002?oo00@00?ooool000;oo`03003ooooo02ooo`04003ooooo000:ool00`00oooo
o`0/ool01000ooooo`002_oo00D00?ooooooo`0000goo`8000[oo`007Ooo0P000_oo00<00?ooool0
4ooo00<00?ooool0<ooo00@00?ooool000Soo`04003ooooo0002ool00`00ooooo`0_ool01000oooo
o`002?oo0`00;_oo00@00?ooool000Soo`<000?oo`03003ooooo00_oo`8000[oo`007_oo00@00?oo
ooooo`8001?oo`03003ooooo03Coo`8000[oo`8000?oo`@002ooo`8000_oo`03003ooooo02goo`80
00_oo`04003oooooool4000:ool00`00ooooo`09ool002Coo`03003ooooo013oo`03003ooooo0=Co
o`03003ooooo00Woo`009Ooo0P004?oo0P00eOoo00<00?ooool02Ooo0003oooo000D0004ool000Co
o`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00;oo`03003ooooo00?oo`8000[oo`@000_o
o`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[o
o`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[o
o`03003ooooo00Woo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[oo`03003ooooo00[o
o`03003ooooo00[oo`03003ooooo00Coo`009_oo0P000ooo00<00?ooool01Ooo1@00h_oo000Wool0
0`00ooooo`0>003Rool002_oo`08003ooooo003oo`00ool000Coo`03003ooooo0>3oo`00=ooo00<0
0?ooool0h?oo000gool2003Qool003Ooo`03003ooooo0>3oo`00\
\>"],
  ImageRangeCache->{{{0, 281}, {173, 0}} -> {-0.0432649, -0.123359, 
  0.000773979, 0.000746414}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["2. Template for many runs"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"The example above is not typical in that the run produced satisfactory \
results in the first trial.  Approximation by rational functions is tricky \
business:  usually there are lots of local minima, and problems due to \
indefinite hessians are not uncommon.  Below you will find an example of a \
set of commands for running the program many times starting the iteration at \
different random locations in the  space of parameters.  \n\nThe example \
treated is the same as above.  Initial sets of parameters are generated \
randomly with values a[i] in the interval ( - 5 , 5 ) .  The output of \
individual runs is stored for later inspection in a file named \"output1\"."],
   "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(<< "\<NewtonFit.m\>"; \)\), 
    \(\(w = Table[0.01\ 1.1\^i, {i, 0, 99}]; \)\), 
    \(\(F[s_] = 1\/\(s\^3 + 6. \ s\^2 + 11. \ s + 6. \); \)\), 
    \(\(values = F[I\ w]; \)\), 
    \(\(points = I\ w; \)\), 
    \(\(data = Transpose[{points, values}]; \)\), 
    \(\(model = \(a[1]\ s + 1. \)\/\(a[2]\ s\^2 + a[3]\ s + a[4]\); \)\), 
    \(Do[initial = 10. \ {Random[], Random[], Random[], Random[]} - 5. ; 
      output = NewtonFit[data, model, s, a, Parameters \[Rule] initial]; 
      \(Save["\<output1\>", initial, output]; \), {i, 1, 7}]\)}], "Input",
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]26.3504758402348526`
        \[InvisibleSpace]"         "\[InvisibleSpace]35.5601585145964671`\),
      SequenceForm[ 
      0, "                   ", 26.350475840234854, "         ", 
        35.560158514596466],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]89.1129320777314504`
        \[InvisibleSpace]"         "\[InvisibleSpace]2.21989899535168033`
        \[InvisibleSpace]"         "\[InvisibleSpace]29.7575883943585761`\),
      SequenceForm[ 
      1, "   ", 89.112932077731443, "         ", 2.2198989953516803, 
        "         ", 29.757588394358578],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]2.21989899535168033`
        \[InvisibleSpace]"         "\[InvisibleSpace]29.7575883943585761`\),
      SequenceForm[ 
      2, "   ", 0.0, "         ", 2.2198989953516803, "         ", 
        29.757588394358578],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]11.3794639833306221`
        \[InvisibleSpace]"         "\[InvisibleSpace]21.3938267097425072`\),
      SequenceForm[ 
      0, "                   ", 11.379463983330622, "         ", 
        21.393826709742505],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]152064.526738398919`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000195562957441189215`
        \[InvisibleSpace]"         "\[InvisibleSpace]4.042502723227825`\),
      SequenceForm[ 
      1, "   ", 152064.52673839891, "         ", 0.00019556295744118922, 
        "         ", 4.042502723227825],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]4.98346069149952519`*^9
        \[InvisibleSpace]"         "\[InvisibleSpace]5.96896257774647942`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374416463`\),
      SequenceForm[ 
      2, "   ", 4983460691.4995251, "         ", 5.9689625777464791*^-09, 
        "         ", 4.0425018837441646],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(3\[InvisibleSpace]"   "\[InvisibleSpace]5.66326028693214045`*^6
        \[InvisibleSpace]"         "\[InvisibleSpace]5.96218728844483791`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374416374`\),
      SequenceForm[ 
      3, "   ", 5663260.2869321406, "         ", 5.9621872884448378*^-09, 
        "         ", 4.0425018837441637],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(4\[InvisibleSpace]"   "\[InvisibleSpace]6.20577457910896868`*^8
        \[InvisibleSpace]"         "\[InvisibleSpace]5.30263369788978433`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374416374`\),
      SequenceForm[ 
      4, "   ", 620577457.9108969, "         ", 5.3026336978897839*^-09, 
        "         ", 4.0425018837441637],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(5\[InvisibleSpace]"   "\[InvisibleSpace]3.58129789368013451`*^8
        \[InvisibleSpace]"         "\[InvisibleSpace]4.98443054765535897`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374416374`\),
      SequenceForm[ 
      5, "   ", 358129789.36801344, "         ", 4.9844305476553591*^-09, 
        "         ", 4.0425018837441637],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(6\[InvisibleSpace]"   "\[InvisibleSpace]3.5830135562008607`*^9
        \[InvisibleSpace]"         "\[InvisibleSpace]3.11454382534488605`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374416641`\),
      SequenceForm[ 
      6, "   ", 3583013556.2008605, "         ", 3.1145438253448862*^-09, 
        "         ", 4.0425018837441664],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(7\[InvisibleSpace]"   "\[InvisibleSpace]9.51160137232376357`*^9
        \[InvisibleSpace]"         "\[InvisibleSpace]1.5604902301887602`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374415486`\),
      SequenceForm[ 
      7, "   ", 9511601372.3237629, "         ", 1.5604902301887602*^-09, 
        "         ", 4.0425018837441549],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(8\[InvisibleSpace]"   "\[InvisibleSpace]2.63764663355800488`*^8
        \[InvisibleSpace]"         "\[InvisibleSpace]1.53919278367195992`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374415397`\),
      SequenceForm[ 
      8, "   ", 263764663.35580051, "         ", 1.5391927836719599*^-09, 
        "         ", 4.042501883744154],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(9\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.53919278367195992`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]4.04250188374415397`\),
      SequenceForm[ 
      9, "   ", 0.0, "         ", 1.5391927836719599*^-09, "         ", 
        4.042501883744154],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]5.8687115397279097`
        \[InvisibleSpace]"         "\[InvisibleSpace]12.1954676454047805`\),
      SequenceForm[ 
      0, "                   ", 5.8687115397279097, "         ", 
        12.195467645404781],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]1.63852188777914831`
        \[InvisibleSpace]"         "\[InvisibleSpace]3.78127835942940748`
        \[InvisibleSpace]"         "\[InvisibleSpace]9.52219877551861593`\),
      SequenceForm[ 
      1, "   ", 1.6385218877791483, "         ", 3.7812783594294075, 
        "         ", 9.5221987755186159],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]46612.6942462005278`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0000601189003352862716`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.25764519517919204`\),
      SequenceForm[ 
      2, "   ", 46612.694246200525, "         ", 6.0118900335286274*^-05, 
        "         ", 1.257645195179192],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(3\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0000601189003352862716`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.25764519517919204`\),
      SequenceForm[ 
      3, "   ", 0.0, "         ", 6.0118900335286274*^-05, "         ", 
        1.257645195179192],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]65.2344750161316433`
        \[InvisibleSpace]"         "\[InvisibleSpace]44.160996924820921`\),
      SequenceForm[ 
      0, "                   ", 65.23447501613164, "         ", 
        44.160996924820921],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]13.3858215738860297`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.98195002769786317`
        \[InvisibleSpace]"         "\[InvisibleSpace]9.03230651852539168`\),
      SequenceForm[ 
      1, "   ", 13.38582157388603, "         ", 1.9819500276978632, 
        "         ", 9.0323065185253917],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.98195002769786317`
        \[InvisibleSpace]"         "\[InvisibleSpace]9.03230651852539168`\),
      SequenceForm[ 
      2, "   ", 0.0, "         ", 1.9819500276978632, "         ", 
        9.0323065185253917],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]1.87734686339552414`
        \[InvisibleSpace]"         "\[InvisibleSpace]2.7350834702349922`\),
      SequenceForm[ 
      0, "                   ", 1.8773468633955241, "         ", 
        2.7350834702349922],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.87734686339552414`
        \[InvisibleSpace]"         "\[InvisibleSpace]2.7350834702349922`\),
      SequenceForm[ 
      1, "   ", 0.0, "         ", 1.8773468633955241, "         ", 
        2.7350834702349922],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]109.551766893277768`
        \[InvisibleSpace]"         "\[InvisibleSpace]51.2558793871123086`\),
      SequenceForm[ 
      0, "                   ", 109.55176689327776, "         ", 
        51.255879387112309],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]78.5153088102024376`
        \[InvisibleSpace]"         "\[InvisibleSpace]4.0980313526573493`
        \[InvisibleSpace]"         "\[InvisibleSpace]34.7307618640400628`\),
      SequenceForm[ 
      1, "   ", 78.515308810202441, "         ", 4.0980313526573493, 
        "         ", 34.730761864040062],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]4.0980313526573493`
        \[InvisibleSpace]"         "\[InvisibleSpace]34.7307618640400628`\),
      SequenceForm[ 
      2, "   ", 0.0, "         ", 4.0980313526573493, "         ", 
        34.730761864040062],
      Editable->False]], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]34.0949218872674908`
        \[InvisibleSpace]"         "\[InvisibleSpace]53.0158967617231713`\),
      SequenceForm[ 
      0, "                   ", 34.094921887267489, "         ", 
        53.01589676172317],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]34.0949218872674908`
        \[InvisibleSpace]"         "\[InvisibleSpace]53.0158967617231713`\),
      SequenceForm[ 
      1, "   ", 0.0, "         ", 34.094921887267489, "         ", 
        53.01589676172317],
      Editable->False]], "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["3. Using a weight."], "Section",
  Evaluatable->False,
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[TextData[{
  "It is possible to emphasize some data points over other by means of a \
weight function as in\n\n\t\tSum[  weight[[i]]*( Abs[ v[[i]]  -  g[  I * w[[ \
i ]]  , a ]  ]^2  )  ,  { i , 1 , n } ] \n\nA weight can be specified as \
input for the function ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "[ ] .\nSuppose that the example considered in section 1. gives an answer \
that we reject as producing too large deviation from the model to the data \
for values of w[[i]]  i = 25 , . . . , 45.  To reduce the error at these 21 \
points a weight function is defined as "
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(wt = Join[Table[1. , {24}], Table[10. , {21}], Table[1. , {55}]]; 
    \)\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"We can use the parameter values found in the previous run to initialize the \
current one."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(initial2 = Parameters /. output1; \)\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(output2 = 
      NewtonFit[data, model, s, a, Parameters \[Rule] initial2, 
        Weight \[Rule] wt]\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]0.00372872677728563006`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00052630932121879832`
        \),
      SequenceForm[ 
      0, "                   ", 0.0037287267772856299, "         ", 
        0.00052630932121879828],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]0.165521753017844446`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0000226568023759128323`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000401831434579619717`
        \),
      SequenceForm[ 
      1, "   ", 0.16552175301784444, "         ", 2.2656802375912831*^-05, 
        "         ", 0.0004018314345796197],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]0.00417434053524024673`
        \[InvisibleSpace]"         "\[InvisibleSpace]2.57386147646970542`*^-8
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000401816523290127758`
        \),
      SequenceForm[ 
      2, "   ", 0.0041743405352402466, "         ", 2.5738614764697052*^-08, 
        "         ", 0.00040181652329012774],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(3\[InvisibleSpace]"   "\[InvisibleSpace]3.43865925458171917`*^-7
        \[InvisibleSpace]"         "\[InvisibleSpace]3.06372891818851789`*^-10
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000401816523288686067`
        \),
      SequenceForm[ 
      3, "   ", 3.4386592545817194*^-07, "         ", 
        3.0637289181885178*^-10, "         ", 0.00040181652328868603],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(4\[InvisibleSpace]"   "\[InvisibleSpace]0.`
        \[InvisibleSpace]"         "\[InvisibleSpace]3.06372891818851789`*^-10
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000401816523288686067`
        \),
      SequenceForm[ 
      4, "   ", 0.0, "         ", 3.0637289181885178*^-10, "         ", 
        0.00040181652328868603],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Parameters 
        \[Rule] {\(-0.135427820545090726`\), 4.71744820426683819`, 
          10.1359981048913083`, 6.01399215928739927`}, 
      SingValHessian 
        \[Rule] {1.39787380055665533`, 0.231320130811744029`, 
          0.0153349387091865274`, 0.00155874725556147648`}, 
      Cost \[Rule] 0.000401816523288686067`, 
      NormGradient \[Rule] 3.06372891818851789`*^-10, NIterations \[Rule] 4}
      \)], "Output"]
}, Open  ]],

Cell[TextData[
"To plot the resulting function set just proceed as in the first example."], 
  "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["4. Stable Zeros and Poles."], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "A zero z of a rational function R[x] is stable if Re[z] < 0.  A pole z of \
R[x] is stable if Re[z]<0.   We now consider the problem of finding the set \
of optimal parameters in (Prob) with rational model, with the additional \
constraint that the resulting rational function has either (a) stable poles, \
(b) stable zeros, and (c) stable poles and zeros.  Key in solving this \
problem with the tools we have is the following result.\n\nProposition.\nLet \
P[ x ] be a monic polynomial with real coefficients.  Then P[ x ] has all its \
zeros in the left open (closed) left half plane if and only if  when P is \
factored over the reals as a product of degree 1 and degree 2 monic \
polynomials the coefficients of each  factor are positive (nonnegative).\n\n\
We can take advantage of this result because the function ",
  StyleBox["NewtonFit",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "[ ] calculates derivatives by symbolic differentiation.  As an example, \
consider the problem from section 1. with the additional constraint that all \
zeros and poles of the resulting rational function are located in the left \
half plane.  Note that the answer obtained in section 1. is not satisfactory, \
because it has a zero in the RHP.   Define as model"
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(model3 = 
      \(a[4]\ \((a[1]\^2\ s + 1. )\)\)\/\(a[2]\^2\ s\^2 + a[3]\^2\ s + 1\); 
    \)\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[TextData[{
  "Note that any choice of parameters in model3 produces a rational functions \
with zeros and poles off the Right Hand Plane.  \n     The set of initial \
parameters given below has been determined with a multiple run of ",
  StyleBox["NewtonFit[] ",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "with several initial sets of parameters.  This set in particular produces \
a local solution as shown below."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(initial3 = {0. ,  .9, 1.3,  .17}; \)\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(output3 = 
      NewtonFit[data, model3, s, a, Parameters \[Rule] initial3]\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    \("Iter    step          grad           cost"\)], "Print"],

Cell[BoxData[
    \("--------------------------------------------------"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(0\[InvisibleSpace]"                   \
"\[InvisibleSpace]0.32370331744577192`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00430163209508041877`
        \),
      SequenceForm[ 
      0, "                   ", 0.32370331744577191, "         ", 
        0.0043016320950804189],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(1\[InvisibleSpace]"   "\[InvisibleSpace]0.112942896597330388`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.022004967063217542`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00156777404612294768`
        \),
      SequenceForm[ 
      1, "   ", 0.11294289659733039, "         ", 0.022004967063217543, 
        "         ", 0.0015677740461229478],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(2\[InvisibleSpace]"   "\[InvisibleSpace]0.00762117850806631036`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.000017777053082513472`
        \[InvisibleSpace]"         "\[InvisibleSpace]0.0015446888275479413`
        \),
      SequenceForm[ 
      2, "   ", 0.0076211785080663105, "         ", 1.7777053082513472*^-05, 
        "         ", 0.0015446888275479414],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(3\[InvisibleSpace]"   "\[InvisibleSpace]0.0000305235121581233315`
        \[InvisibleSpace]"         "\[InvisibleSpace]1.18668481912743351`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00154468865411384648`
        \),
      SequenceForm[ 
      3, "   ", 3.0523512158123331*^-05, "         ", 
        1.1866848191274335*^-09, "         ", 0.0015446886541138464],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \(4\[InvisibleSpace]"   "\[InvisibleSpace]4.53644917637885836`*^-11
        \[InvisibleSpace]"         "\[InvisibleSpace]1.08998216903212119`*^-9
        \[InvisibleSpace]"         "\[InvisibleSpace]0.00154468865411384541`
        \),
      SequenceForm[ 
      4, "   ", 4.5364491763788579*^-11, "         ", 
        1.0899821690321213*^-09, "         ", 0.0015446886541138453],
      Editable->False]], "Print"],

Cell[BoxData[
    \({Parameters 
        \[Rule] {0.`, 1.00935969011286696`, 1.31713509982247423`, 
          0.166462800628930196`}, 
      SingValHessian 
        \[Rule] {91.4902997084746871`, 0.775119586193551945`, 
          0.350922023253075732`, 0.0313337397597327482`}, 
      Cost \[Rule] 0.00154468865411384541`, 
      NormGradient \[Rule] 1.08998216903212119`*^-9, NIterations \[Rule] 4}
      \)], "Output"]
}, Open  ]],

Cell[TextData[
"Now the rational function obtained above is produced, and a plot ais \
generated.."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(par3 = Parameters /. output3; \)\), 
    \(r3[s_] = 
      model3 /. {
          a[j_] \[RuleDelayed] par3\[LeftDoubleBracket]j\[RightDoubleBracket]}
          \)}], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    FractionBox[
      RowBox[{"0.166462800628930196`", " ", 
        RowBox[{"(", 
          RowBox[{
            RowBox[{
              StyleBox["1.`",
                StyleBoxAutoDelete->True,
                PrintPrecision->1], "\[InvisibleSpace]"}], "+", \(0.`\ s\)}], 
          ")"}]}], 
      \(1 + 1.73484487118435914`\ s + 1.0188069840247429`\ s\^2\)]], "Output"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
      ListPlot[Transpose[{Re[values], Im[values]}], 
        DisplayFunction \[Rule] Identity], 
      ParametricPlot[{Re[r3[I\ t]], Im[r3[I\ t]]}, {t, 0, 10. }, 
        DisplayFunction \[Rule] Identity], 
      DisplayFunction \[Rule] $DisplayFunction]; \)\)], "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.191734 4.70785 0.589843 4.88963 [
[.42713 .57734 -12 -9 ]
[.42713 .57734 12 0 ]
[.66252 .57734 -9 -9 ]
[.66252 .57734 9 0 ]
[.89791 .57734 -12 -9 ]
[.89791 .57734 12 0 ]
[.17923 .00309 -30 -4.5 ]
[.17923 .00309 0 4.5 ]
[.17923 .10088 -24 -4.5 ]
[.17923 .10088 0 4.5 ]
[.17923 .19867 -30 -4.5 ]
[.17923 .19867 0 4.5 ]
[.17923 .29647 -30 -4.5 ]
[.17923 .29647 0 4.5 ]
[.17923 .39426 -30 -4.5 ]
[.17923 .39426 0 4.5 ]
[.17923 .49205 -30 -4.5 ]
[.17923 .49205 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.42713 .58984 m
.42713 .59609 L
s
[(0.05)] .42713 .57734 0 1 Mshowa
.66252 .58984 m
.66252 .59609 L
s
[(0.1)] .66252 .57734 0 1 Mshowa
.89791 .58984 m
.89791 .59609 L
s
[(0.15)] .89791 .57734 0 1 Mshowa
.125 Mabswid
.23881 .58984 m
.23881 .59359 L
s
.28589 .58984 m
.28589 .59359 L
s
.33297 .58984 m
.33297 .59359 L
s
.38005 .58984 m
.38005 .59359 L
s
.4742 .58984 m
.4742 .59359 L
s
.52128 .58984 m
.52128 .59359 L
s
.56836 .58984 m
.56836 .59359 L
s
.61544 .58984 m
.61544 .59359 L
s
.7096 .58984 m
.7096 .59359 L
s
.75668 .58984 m
.75668 .59359 L
s
.80375 .58984 m
.80375 .59359 L
s
.85083 .58984 m
.85083 .59359 L
s
.14466 .58984 m
.14466 .59359 L
s
.09758 .58984 m
.09758 .59359 L
s
.0505 .58984 m
.0505 .59359 L
s
.00342 .58984 m
.00342 .59359 L
s
.94499 .58984 m
.94499 .59359 L
s
.99207 .58984 m
.99207 .59359 L
s
.25 Mabswid
0 .58984 m
1 .58984 L
s
.19173 .00309 m
.19798 .00309 L
s
[(-0.12)] .17923 .00309 1 0 Mshowa
.19173 .10088 m
.19798 .10088 L
s
[(-0.1)] .17923 .10088 1 0 Mshowa
.19173 .19867 m
.19798 .19867 L
s
[(-0.08)] .17923 .19867 1 0 Mshowa
.19173 .29647 m
.19798 .29647 L
s
[(-0.06)] .17923 .29647 1 0 Mshowa
.19173 .39426 m
.19798 .39426 L
s
[(-0.04)] .17923 .39426 1 0 Mshowa
.19173 .49205 m
.19798 .49205 L
s
[(-0.02)] .17923 .49205 1 0 Mshowa
.125 Mabswid
.19173 .02754 m
.19548 .02754 L
s
.19173 .05198 m
.19548 .05198 L
s
.19173 .07643 m
.19548 .07643 L
s
.19173 .12533 m
.19548 .12533 L
s
.19173 .14978 m
.19548 .14978 L
s
.19173 .17423 m
.19548 .17423 L
s
.19173 .22312 m
.19548 .22312 L
s
.19173 .24757 m
.19548 .24757 L
s
.19173 .27202 m
.19548 .27202 L
s
.19173 .32091 m
.19548 .32091 L
s
.19173 .34536 m
.19548 .34536 L
s
.19173 .36981 m
.19548 .36981 L
s
.19173 .41871 m
.19548 .41871 L
s
.19173 .44315 m
.19548 .44315 L
s
.19173 .4676 m
.19548 .4676 L
s
.19173 .5165 m
.19548 .5165 L
s
.19173 .54095 m
.19548 .54095 L
s
.19173 .5654 m
.19548 .5654 L
s
.19173 .61429 m
.19548 .61429 L
s
.25 Mabswid
.19173 0 m
.19173 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.97619 .57491 Mdot
.97615 .57341 Mdot
.9761 .57177 Mdot
.97605 .56996 Mdot
.97598 .56798 Mdot
.9759 .56579 Mdot
.97579 .56339 Mdot
.97567 .56074 Mdot
.97552 .55784 Mdot
.97535 .55464 Mdot
.97513 .55113 Mdot
.97487 .54727 Mdot
.97455 .54302 Mdot
.97417 .53835 Mdot
.97371 .53323 Mdot
.97315 .52759 Mdot
.97247 .5214 Mdot
.97166 .51461 Mdot
.97067 .50715 Mdot
.96948 .49896 Mdot
.96804 .48999 Mdot
.9663 .48015 Mdot
.9642 .46939 Mdot
.96166 .45761 Mdot
.95861 .44473 Mdot
.95493 .43069 Mdot
.9505 .4154 Mdot
.94518 .39877 Mdot
.93879 .38075 Mdot
.93112 .36127 Mdot
.92196 .34029 Mdot
.91102 .31782 Mdot
.898 .29386 Mdot
.88255 .26851 Mdot
.8643 .2419 Mdot
.84283 .21428 Mdot
.81774 .18598 Mdot
.78859 .15746 Mdot
.755 .12934 Mdot
.71667 .1024 Mdot
.67339 .07755 Mdot
.62514 .0559 Mdot
.57217 .03865 Mdot
.51499 .02706 Mdot
.45449 .02237 Mdot
.39192 .02568 Mdot
.32888 .03779 Mdot
.26727 .05909 Mdot
.20914 .08945 Mdot
.15653 .12811 Mdot
.11128 .17373 Mdot
.07482 .22445 Mdot
.04806 .27802 Mdot
.0312 .33202 Mdot
.02381 .38412 Mdot
.02482 .43228 Mdot
.03273 .47493 Mdot
.04576 .5111 Mdot
.06206 .5404 Mdot
.07993 .563 Mdot
.09792 .57951 Mdot
.11495 .59079 Mdot
.1303 .59786 Mdot
.14358 .60172 Mdot
.15468 .60329 Mdot
.16369 .60332 Mdot
.17082 .60242 Mdot
.17634 .60103 Mdot
.18053 .59946 Mdot
.18365 .5979 Mdot
.18596 .59645 Mdot
.18764 .59518 Mdot
.18884 .59409 Mdot
.18971 .5932 Mdot
.19032 .59246 Mdot
.19075 .59188 Mdot
.19105 .59141 Mdot
.19126 .59105 Mdot
.19141 .59077 Mdot
.19151 .59055 Mdot
.19158 .59038 Mdot
.19163 .59025 Mdot
.19166 .59015 Mdot
.19168 .59008 Mdot
.1917 .59002 Mdot
.19171 .58998 Mdot
.19172 .58994 Mdot
.19172 .58992 Mdot
.19173 .5899 Mdot
.19173 .58989 Mdot
.19173 .58988 Mdot
.19173 .58987 Mdot
.19173 .58986 Mdot
.19173 .58986 Mdot
.19173 .58985 Mdot
.19173 .58985 Mdot
.19173 .58985 Mdot
.19173 .58985 Mdot
.19173 .58985 Mdot
.19173 .58985 Mdot
.5 Mabswid
.97542 .58984 m
.97518 .57242 L
.97455 .55657 L
.97333 .53832 L
.97171 .5211 L
.96766 .49065 L
.96161 .45791 L
.95298 .42236 L
.94162 .38541 L
.91569 .32158 L
.88185 .2599 L
.84419 .20691 L
.79705 .15508 L
.74076 .1077 L
.68015 .06976 L
.64463 .05269 L
.61122 .03974 L
.57901 .02991 L
.54989 .02316 L
.53383 .02029 L
.51659 .01787 L
.50872 .01699 L
.50044 .01622 L
.48573 .01523 L
.47155 .01475 L
.45877 .01472 L
.4442 .01513 L
.4306 .01597 L
.41823 .01711 L
.40696 .01847 L
.38209 .02256 L
.35804 .02802 L
.33267 .03544 L
.28538 .0543 L
.24742 .07487 L
.21247 .09903 L
.16222 .14565 L
.14038 .17222 L
.12357 .19667 L
.10867 .22259 L
.09601 .24953 L
.0867 .27417 L
.08028 .29557 L
.07577 .3148 L
.074 .32416 L
.07246 .33386 L
.07135 .34231 L
.07055 .34979 L
.06995 .35707 L
.06954 .36376 L
Mistroke
.06937 .36744 L
.06924 .37127 L
.06916 .37483 L
.06913 .378 L
.06913 .38167 L
.06915 .38371 L
.06918 .38556 L
.06926 .38892 L
.06938 .3924 L
.06953 .39571 L
.06973 .39924 L
.07019 .40546 L
.07132 .41649 L
.07301 .42842 L
.07487 .4386 L
.07874 .45524 L
.0833 .47054 L
.09322 .4959 L
.10232 .51382 L
.11929 .53946 L
.13344 .5557 L
.14386 .56544 L
.15162 .57163 L
.15807 .57614 L
.16297 .57918 L
.16712 .58151 L
.17034 .58316 L
.17289 .58436 L
.17514 .58534 L
.17695 .58607 L
.17843 .58664 L
.17978 .58712 L
.18089 .5875 L
.18192 .58783 L
.18277 .58809 L
.18349 .5883 L
.18417 .58849 L
.18419 .5885 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic,
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40n007h00Ooooo8Ooo003ooolQ
ool00?ooob7oo`007ooo0P001?oo00<00?ooool00_oo10000_oo10001Ooo00<00?ooool0i?oo000N
ool01000ooooo`002_oo00D00?ooooooo`0000Soo`800>Goo`007_oo00@00?ooool000[oo`03003o
oooo00;oo`03003ooooo00Goo`03003ooooo0>Coo`006?oo10000_oo00@00?ooool000[oo`03003o
oooo00?oo`03003ooooo00Coo`03003ooooo0>Coo`007_oo00@00?ooool000[oo`03003ooooo00Co
o`03003ooooo00?oo`03003ooooo04;oo`l009?oo`007_oo00@00?ooool000Soo`<000?oo`04003o
oooo0005ool00`00ooooo`0hool:000?ool9002:ool001ooo`8000_oo`03003ooooo00;oo`8000Ko
o`03003ooooo03Coo`@000koo`8001;oo`D008Goo`00>Ooo00<00?ooool0;ooo1`004Ooo00<00?oo
ool03Ooo0P001_oo1000POoo000iool2000[ool50006ool00`00ooooo`0Pool00`00ooooo`08ool4
001mool003Woo`03003ooooo02Koo`@003goo`D007Soo`00>Ooo00<00?ooool08_oo1000A_oo1000
M?oo000iool00`00ooooo`0Qool30013ool20007ool4001`ool003Woo`03003ooooo01koo`<004Oo
o`03003ooooo00Woo`8006koo`00>Ooo00<00?ooool06ooo0`00F?oo0`00Jooo000iool00`00oooo
o`0Iool2001Nool2001Yool003Woo`8001Soo`8005goo`8000?oo`8006Ooo`00>Ooo00<00?ooool0
4Ooo0P000_oo0P00H?oo00<00?ooool00ooo0P00IOoo000iool00`00ooooo`0Bool3001Zool2001S
ool003Woo`03003ooooo017oo`8006koo`80067oo`00>Ooo00<00?ooool03ooo0P00L_oo0P00Gooo
000iool00`00ooooo`0=ool2001fool2001Mool003Woo`03003ooooo00coo`03003ooooo07Soo`80
05_oo`00>Ooo0P002ooo0P00Mooo0P001?oo0P00FOoo000iool00`00ooooo`09ool00`00ooooo`1h
ool00`00ooooo`04ool00`00ooooo`1Fool003Woo`03003ooooo00Ooo`8008?oo`8005Koo`00>Ooo
00@00?ooooooo`8000?oo`03003ooooo08Goo`8005Coo`00>Ooo00<00?ooool00_oo00@00?ooool0
08[oo`8005;oo`009Ooo0P001?oo00<00?ooool00_oo10001Ooo00<00?ooool00ooo0P00SOoo00<0
0?ooool0Cooo000Tool01000ooooo`002_oo00<00?ooool01?oo0P000ooo00<00?ooool0R_oo0P00
0_oo0P00Cooo000Tool01000ooooo`002_oo00<00?ooool01?oo00D00?ooooooo`0008koo`03003o
oooo00;oo`8004goo`007_oo10000_oo00@00?ooool000[oo`03003ooooo00Coo`04003ooooo002F
ool00`00ooooo`1:ool002Coo`04003ooooo000:ool00`00ooooo`04ool00`00ool0002Hool00`00
ooooo`19ool002Coo`04003ooooo0008ool30006ool2002Jool00`00ooooo`18ool002Goo`8000_o
o`03003ooooo00Coo`03003ooooo09[oo`03003ooooo04Ooo`00=ooo0`00WOoo00<00?ooool0A_oo
000fool00`00ooooo`02002Mool00`00ooooo`15ool002koo`8000Goo`05003oooooool0002Lool2
00000ooo003oo`15ool002ooo`03003ooooo00;oo`03003ooooo00;oo`03003ooooo09_oo`03003o
oooo008004Coo`00<ooo00<00?ooool00ooo00<00?ooool0X?oo00<00?ooool0@Ooo000bool00`00
ooooo`04ool00`00ooooo`2Qool00`00ooooo`10ool0037oo`03003ooooo00Goo`03003ooooo0:;o
o`03003ooooo03ooo`00<?oo00<00?ooool01_oo00<00?ooool0Xooo00<00?ooool0?_oo000_ool0
0`00ooooo`07ool2002Uool00`00ooooo`0mool002koo`03003ooooo00Soo`03003ooooo0:Goo`03
003ooooo03coo`00;Ooo00<00?ooool02Ooo00<00?ooool0Xooo0P0000?oo`00ool0??oo000]ool0
0`00ooooo`09ool00`00ooooo`2Tool01000ooooo`00??oo000/ool00`00ooooo`0:ool00`00oooo
o`2Xool00`00ooooo`0iool002_oo`03003ooooo00_oo`03003ooooo0:Woo`03003ooooo03Soo`00
8Ooo0P001ooo00<00?ooool03?oo00<00?ooool0Z_oo00<00?ooool0=ooo000Rool00`00ooooo`04
ool00`00ooooo`0=ool2002/ool00`00ooooo`0fool002Woo`03003ooooo00goo`03003ooooo0:co
o`03003ooooo03Goo`00:?oo00<00?ooool03_oo00<00?ooool0[Ooo00<00?ooool0=?oo000Wool0
0`00ooooo`0?ool00`00ooooo`2/ool3000eool002Koo`03003ooooo013oo`03003ooooo0:goo`03
003oo`0003Coo`009_oo00<00?ooool04?oo00<00?ooool0/?oo00<00?ooool0<Ooo000Oool20004
ool00`00ooooo`03ool20004ool20006ool00`00ooooo`2aool00`00ooooo`0`ool001koo`04003o
oooo0003ool00`00ooooo`02ool01000ooooo`000_oo00@00?ooool000Goo`800;?oo`03003ooooo
02ooo`007_oo00@00?ooool000;oo`03003ooooo00?oo`04003ooooo0002ool01000ooooo`001Ooo
00<00?ooool0/ooo00<00?ooool0;_oo000Hool40002ool01000ooooo`000_oo00<00?ooool00ooo
00@00?ooool000?oo`8000Koo`03003ooooo0;Coo`03003ooooo02goo`007_oo00H00?ooool00?oo
0006ool01000ooooo`000_oo00@00?ooool000Goo`03003ooooo0;Goo`03003ooooo02coo`007_oo
00H00?ooool00?oo0006ool01000ooooo`000_oo00@00?ooool000Goo`03003ooooo0;?oo`<002ko
o`007ooo0P0000?oo`00ool01ooo0P001?oo0P001_oo00<00?ooool0]?oo00<00?oo0000;Ooo000R
ool00`00ooooo`0Dool00`00ooooo`2gool00`00ooooo`0Zool001Ooo`8000Soo`03003ooooo01Go
o`800;Soo`03003ooooo02[oo`006?oo00<00?ooool01_oo00<00?ooool05Ooo00<00?ooool0^?oo
00<00?ooool0:Ooo000Pool00`00ooooo`0Fool00`00ooooo`2iool00`00ooooo`0Xool0023oo`03
003ooooo01Koo`03003ooooo0;Woo`03003ooooo02Soo`007ooo00<00?ooool05ooo00<00?ooool0
^_oo00<00?ooool09ooo000Oool00`00ooooo`0Gool00`00ooooo`2iool3000Xool001koo`03003o
oooo01Soo`03003ooooo0;[oo`8002Soo`007_oo00<00?ooool06?oo0P00_Ooo00<00?ooool09Ooo
000Nool00`00ooooo`0Hool00`00ooooo`2mool00`00ooooo`0Tool001goo`03003ooooo01Woo`03
003ooooo0;goo`03003ooooo02Coo`007Ooo00<00?ooool06Ooo00<00?ooool0__oo00<00?ooool0
8ooo000Lool00`00ooooo`0Jool00`00ooooo`2oool00`00ooooo`0Rool001coo`03003ooooo01[o
o`03003ooooo0;koo`8002Coo`006ooo00<00?ooool06ooo0P00`?oo0P008ooo000Kool00`00oooo
o`0Kool00`00ooooo`30ool00`00ooooo`0Qool0013oo`8000Woo`03003ooooo01_oo`03003ooooo
0<7oo`03003ooooo023oo`004Ooo00<00?ooool01_oo00<00?ooool07?oo00<00?ooool0`_oo00<0
0?ooool07ooo000Jool00`00ooooo`0Lool00`00ooooo`32ool00`00ooooo`0Oool001[oo`03003o
oooo01coo`03003ooooo0<?oo`03003ooooo01koo`006Ooo00<00?ooool00ooo0P001?oo00<00?oo
ool00ooo0P001?oo0P001_oo00<00?ooool0`_oo0P008?oo000Iool00`00ooooo`02ool01000oooo
o`002?oo00@00?ooool000;oo`04003ooooo0005ool20034ool2000Oool001Woo`03003ooooo00;o
o`04003ooooo0008ool01000ooooo`000_oo00@00?ooool000Goo`03003ooooo0<Coo`03003ooooo
01goo`006?oo10000_oo00@00?ooool000Soo`04003ooooo0002ool30006ool00`00ooooo`35ool0
0`00ooooo`0Lool001Soo`03003ooooo00?oo`04003ooooo0008ool01000ooooo`000_oo00<00?oo
ool01_oo00<00?ooool0a_oo00<00?ooool06ooo000Hool00`00ooooo`03ool01000ooooo`002?oo
00@00?ooool000;oo`03003ooooo00Koo`03003ooooo0<Koo`03003ooooo01_oo`006?oo00<00?oo
ool01?oo0P002_oo0P001?oo0`001Ooo00<00?ooool0aooo00<00?ooool06_oo000Hool00`00oooo
o`0Nool00`00ooooo`36ool2000Lool001Soo`03003ooooo01koo`800<Soo`8001_oo`006?oo00<0
0?ooool07_oo00<00?ooool0b?oo00<00?ooool06Ooo000Hool00`00ooooo`0Nool00`00ooooo`39
ool00`00ooooo`0Hool000_oo`8000[oo`03003ooooo01ooo`03003ooooo0<Woo`03003ooooo01So
o`003?oo00<00?ooool02?oo00<00?ooool07ooo00<00?ooool0b_oo00<00?ooool05ooo000Gool0
0`00ooooo`0Oool00`00ooooo`39ool2000Iool001Ooo`03003ooooo01ooo`03003ooooo0<[oo`03
003ooooo01Ooo`005ooo00<00?ooool07ooo0P00c?oo00<00?ooool05_oo000Gool00`00ooooo`0O
ool00`00ooooo`3;ool00`00ooooo`0Fool001Ooo`03003ooooo01ooo`03003ooooo0<_oo`03003o
oooo01Koo`005_oo00<00?ooool08?oo00<00?ooool0c?oo00<00?ooool05Ooo000Fool00`00oooo
o`0Pool00`00ooooo`3;ool2000Gool001Koo`03003ooooo023oo`03003ooooo0<coo`8001Koo`00
5_oo00<00?ooool08?oo00<00?ooool0cOoo00<00?ooool05?oo000Fool00`00ooooo`0Pool2003>
ool00`00ooooo`0Dool001Koo`03003ooooo023oo`03003ooooo0<koo`03003ooooo01?oo`005_oo
00<00?ooool08?oo00<00?ooool0c_oo00<00?ooool04ooo0009ool2000;ool00`00ooooo`0Pool0
0`00ooooo`3=ool3000Dool000[oo`03003ooooo00Woo`03003ooooo023oo`03003ooooo0<koo`80
01Coo`005_oo00<00?ooool01_oo0P001?oo00<00?ooool00ooo0P001Ooo00<00?ooool01?oo00<0
0?ooool0cooo00<00?ooool04_oo000Fool00`00ooooo`05ool01000ooooo`002?oo00@00?ooool0
00Coo`03003ooooo00Coo`03003ooooo0=3oo`03003ooooo017oo`005_oo00<00?ooool01Ooo00@0
0?ooool000Soo`05003ooooo003oo`050005ool2003Aool00`00ooooo`0Aool001Koo`03003oo`00
00<000;oo`04003ooooo0008ool01P00ooooo`00ool000;oo`03003ooooo00Coo`03003ooooo0<oo
o`8001?oo`005_oo00<00?ooool01Ooo00@00?ooool000Soo`04003ooooo0002ool00`00ool00006
ool00`00ooooo`3@ool00`00ooooo`0Aool001Koo`03003ooooo00Goo`04003ooooo0008ool01000
ooooo`000ooo0P001_oo00<00?ooool0dOoo00<00?ooool04?oo000Fool00`00ooooo`06ool2000:
ool20005ool00`00ooooo`04ool00`00ooooo`3Aool00`00ooooo`0@ool001Ooo`03003ooooo01oo
o`03003ooooo0=7oo`80017oo`005ooo00<00?ooool07ooo00<00?ooool0dOoo0P004Ooo000Gool0
0`00ooooo`0Oool2003Cool00`00ooooo`0?ool001Ooo`03003ooooo01ooo`03003ooooo0=;oo`03
003ooooo00ooo`005ooo00<00?ooool07ooo00<00?ooool0d_oo00<00?ooool03ooo0009ool2000<
ool00`00ooooo`0Oool00`00ooooo`3Bool2000@ool000[oo`03003ooooo00_oo`03003ooooo01ko
o`03003ooooo0=?oo`03003ooooo00koo`006?oo00<00?ooool07_oo00<00?ooool0dooo00<00?oo
ool03_oo000Hool00`00ooooo`0Nool2003Dool2000?ool001Soo`03003ooooo01koo`03003ooooo
0=?oo`8000ooo`006Ooo00<00?ooool07Ooo00<00?ooool0dooo00<00?ooool03_oo000Iool00`00
ooooo`0Mool00`00ooooo`3Dool00`00ooooo`0=ool001Woo`03003ooooo01goo`03003ooooo0=Co
o`8000koo`006Ooo00<00?ooool07Ooo00<00?ooool0e?oo0P003_oo000Iool00`00ooooo`0Mool0
0`00ooooo`3Dool00`00ooooo`0=ool001[oo`03003ooooo01coo`800=Goo`<000goo`006_oo00<0
0?ooool07?oo00<00?ooool0eOoo0P003Ooo000;ool2000=ool00`00ooooo`0Lool00`00ooooo`3E
ool00`00ooooo`0<ool000coo`03003ooooo00coo`03003ooooo01_oo`03003ooooo0=Goo`8000go
o`006ooo00<00?ooool06ooo00<00?ooool0eOoo0P003Ooo000Lool00`00ooooo`0Jool00`00oooo
o`3Fool00`00ooooo`0;ool001coo`03003ooooo008000Coo`03003ooooo00?oo`8000?oo`@000Go
o`03003ooooo0=Koo`8000coo`007Ooo0P000_oo00<00?ooool01_oo00@00?ooool000;oo`03003o
oooo00Koo`800=Ooo`8000coo`007Ooo0P000_oo00<00?ooool01_oo00@00?ooool000?oo`03003o
oooo00Goo`03003ooooo0=Koo`03003ooooo00_oo`006?oo100000?oo`0000000_oo00<00?ooool0
1_oo00@00?ooool000Coo`03003ooooo00Coo`03003ooooo0=Koo`8000coo`007_oo00@00?ooool0
00Soo`04003ooooo0005ool00`00ooooo`03ool00`00ooooo`3Fool2000<ool001koo`04003ooooo
0008ool01000ooooo`000_oo00@00?ooool000Goo`03003ooooo0=Koo`8000coo`003ooo0P003_oo
0P002_oo0P001?oo0P001_oo00<00?ooool0eooo00<00?ooool02_oo000@ool00`00ooooo`0<ool0
0`00ooooo`0Gool00`00ooooo`3Gool2000;ool0023oo`03003ooooo01Koo`800=Soo`8000_oo`00
8?oo00<00?ooool05_oo00<00?ooool0eooo0P002ooo000Qool00`00ooooo`0Eool00`00ooooo`3G
ool2000;ool002;oo`03003ooooo01Coo`03003ooooo0=Ooo`8000_oo`008ooo00<00?ooool04ooo
00<00?ooool0eooo0P002ooo000Sool00`00ooooo`0Cool00`00ooooo`3Hool00`00ooooo`09ool0
01Coo`8000koo`03003ooooo01;oo`03003ooooo0=Ooo`8000_oo`005Ooo00<00?ooool03Ooo00<0
0?ooool04Ooo0P00fOoo0P002_oo000Vool00`00ooooo`0@ool00`00ooooo`3Hool2000:ool002Ko
o`03003ooooo013oo`03003ooooo0=Soo`8000[oo`009ooo00<00?ooool03ooo00<00?ooool0f?oo
0P002_oo000Xool00`00ooooo`0>ool00`00ooooo`3Hool2000:ool002Woo`03003ooooo00goo`03
003ooooo03Goo`8000Coo`03003ooooo00?oo`8000?oo`<0037oo`8000Coo`03003ooooo00;oo`@0
033oo`8000Coo`03003ooooo00;oo`@000;oo`<000_oo`8000[oo`006Ooo0P003ooo00<00?ooool0
3?oo00<00?ooool0=?oo00@00?ooool000Soo`04003ooooo0005ool00`00ooooo`0]ool01000oooo
o`002_oo00<00?ooool0;_oo00@00?ooool000[oo`03003ooooo00Coo`03003ooooo00Soo`8000[o
o`006_oo00<00?ooool03_oo00<00?ooool02ooo0P00=Ooo00@00?ooool000Soo`04003ooooo0005
ool00`00ooooo`0]ool01000ooooo`002_oo00<00?ooool0;_oo00@00?ooool000[oo`03003ooooo
00Coo`03003ooooo00Soo`8000[oo`00;?oo0P002ooo00<00?ooool0=?oo00@00?ooool000Soo`04
003ooooo0002ool3000`ool01000ooooo`002_oo00<00?ooool0;_oo00@00?ooool000[oo`04003o
ooooool3000;ool2000:ool002koo`03003ooooo00Soo`03003ooooo03Coo`04003ooooo0008ool0
1000ooooo`000_oo00<00?ooool0<?oo00@00?ooool000[oo`03003ooooo02koo`04003ooooo000:
ool01@00oooooooo00003Ooo0P002_oo000Nool2000?ool00`00ooooo`07ool00`00ooooo`0dool0
1000ooooo`002?oo00@00?ooool000;oo`03003ooooo033oo`04003ooooo0008ool3000`ool01000
ooooo`002?oo0`000ooo00<00?ooool02ooo0P002_oo000Oool00`00ooooo`0>ool30006ool00`00
ooooo`0eool2000:ool20003ool4000`ool2000;ool00`00ooooo`0_ool2000;ool01000oooooooo
10002_oo00<00?ooool02Ooo000cool20004ool00`00ooooo`3Hool00`00ooooo`09ool002;oo`80
013oo`H00=[oo`03003ooooo00Woo`000oooo`006P001?oo0004ool00`00ooooo`0:ool00`00oooo
o`0:ool00`00ooooo`06ool20002ool00`00ooooo`08ool4000<ool00`00ooooo`0:ool00`00oooo
o`0:ool00`00ooooo`0:ool00`00ooooo`0;ool00`00ooooo`0:ool00`00ooooo`0:ool00`00oooo
o`0:ool00`00ooooo`0:ool00`00ooooo`0;ool00`00ooooo`0:ool00`00ooooo`0:ool00`00oooo
o`0:ool00`00ooooo`0:ool00`00ooooo`0;ool00`00ooooo`0:ool00`00ooooo`0:ool00`00oooo
o`04ool002Soo`04003oo`000008ool6003Vool002_oo`03003ooooo00H00003ool0000000;oo`03
003ooooo0>Coo`00;ooo00D00?oo003oo`0000Goo`03003ooooo0>Coo`00>Ooo00<00?ooool0i?oo
000iool2003Uool003Woo`03003ooooo0>Coo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-0.0432221, -0.123282, 
  0.0007575, 0.000729339}}]
}, Open  ]],

Cell["", "Input",
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["6.  Non Rational Models."], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The functions that were  fit to the data in the examples above were \
rational, but the function ",
  StyleBox["NewtonFit[]",
    FontFamily->"Courier",
    FontSize->12,
    FontWeight->"Bold"],
  "   can solve problems with more general models."
}], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
WindowToolbars->{},
CellGrouping->Automatic,
WindowSize->{628, 658},
WindowMargins->{{4, Automatic}, {4, Automatic}},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->True,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
CharacterEncoding->"XAutomaticEncoding"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 84, 2, 104, "Title",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[1840, 57, 89, 2, 52, "Section",
  Evaluatable->False],
Cell[1932, 61, 398, 11, 42, "SmallText",
  Evaluatable->False]
}, Open  ]],

Cell[CellGroupData[{
Cell[2367, 77, 104, 2, 52, "Section",
  Evaluatable->False],
Cell[2474, 81, 582, 17, 177, "SmallText",
  Evaluatable->False],
Cell[3059, 100, 1461, 28, 181, "SmallText",
  Evaluatable->False],
Cell[4523, 130, 910, 25, 118, "SmallText",
  Evaluatable->False],
Cell[5436, 157, 1309, 30, 194, "SmallText",
  Evaluatable->False,
  PageBreakWithin->False],
Cell[6748, 189, 562, 11, 176, "SmallText",
  Evaluatable->False]
}, Open  ]],

Cell[CellGroupData[{
Cell[7347, 205, 151, 4, 52, "Section",
  Evaluatable->False,
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[7501, 211, 876, 14, 327, "SmallText",
  Evaluatable->False,
  PageBreakBelow->True],
Cell[8380, 227, 359, 8, 102, "SmallText",
  Evaluatable->False],
Cell[8742, 237, 95, 2, 27, "SmallText",
  Evaluatable->False],
Cell[8840, 241, 85, 2, 27, "Input"],
Cell[8928, 245, 129, 3, 27, "SmallText",
  Evaluatable->False],
Cell[9060, 250, 224, 5, 100, "Input"],
Cell[9287, 257, 322, 10, 46, "SmallText",
  Evaluatable->False],
Cell[9612, 269, 99, 2, 27, "Input"],
Cell[9714, 273, 167, 4, 27, "SmallText",
  Evaluatable->False],
Cell[9884, 279, 126, 3, 45, "Input"],
Cell[10013, 284, 326, 10, 44, "SmallText",
  Evaluatable->False],
Cell[10342, 296, 101, 2, 27, "Input"],
Cell[10446, 300, 321, 10, 46, "SmallText",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[10792, 314, 188, 5, 27, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[10983, 321, 85, 1, 23, "Print"],
Cell[11071, 324, 76, 1, 23, "Print"],
Cell[11150, 327, 85, 1, 23, "Print"],
Cell[11238, 330, 353, 8, 23, "Print"],
Cell[11594, 340, 426, 8, 23, "Print"],
Cell[12023, 350, 439, 9, 23, "Print"],
Cell[12465, 361, 445, 9, 23, "Print"],
Cell[12913, 372, 450, 9, 23, "Print"],
Cell[13366, 383, 452, 9, 25, "Print"],
Cell[13821, 394, 460, 9, 25, "Print"],
Cell[14284, 405, 449, 9, 61, "Output"]
}, Open  ]],
Cell[14748, 417, 576, 12, 87, "SmallText",
  Evaluatable->False,
  PageBreakWithin->False],

Cell[CellGroupData[{
Cell[15349, 433, 219, 6, 43, "Input"],
Cell[15571, 441, 357, 9, 43, "Output"]
}, Open  ]],
Cell[15943, 453, 190, 4, 27, "SmallText",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[16158, 461, 382, 9, 75, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[16543, 472, 18339, 585, 182, 6790, 436, "GraphicsData", 
"PostScript", "Graphics",
  Evaluatable->False,
  PageBreakBelow->Automatic]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[34931, 1063, 102, 2, 52, "Section",
  Evaluatable->False],
Cell[35036, 1067, 759, 12, 117, "SmallText",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[35820, 1083, 634, 12, 219, "Input",
  PageBreakBelow->Automatic],
Cell[36457, 1097, 85, 1, 23, "Print"],
Cell[36545, 1100, 76, 1, 23, "Print"],
Cell[36624, 1103, 85, 1, 23, "Print"],
Cell[36712, 1106, 350, 8, 23, "Print"],
Cell[37065, 1116, 423, 8, 23, "Print"],
Cell[37491, 1126, 391, 8, 23, "Print"],
Cell[37885, 1136, 85, 1, 23, "Print"],
Cell[37973, 1139, 76, 1, 23, "Print"],
Cell[38052, 1142, 85, 1, 23, "Print"],
Cell[38140, 1145, 350, 8, 23, "Print"],
Cell[38493, 1155, 428, 8, 23, "Print"],
Cell[38924, 1165, 435, 8, 25, "Print"],
Cell[39362, 1175, 435, 8, 25, "Print"],
Cell[39800, 1185, 434, 8, 25, "Print"],
Cell[40237, 1195, 435, 8, 25, "Print"],
Cell[40675, 1205, 434, 8, 25, "Print"],
Cell[41112, 1215, 434, 8, 25, "Print"],
Cell[41549, 1225, 434, 8, 25, "Print"],
Cell[41986, 1235, 399, 8, 25, "Print"],
Cell[42388, 1245, 85, 1, 23, "Print"],
Cell[42476, 1248, 76, 1, 23, "Print"],
Cell[42555, 1251, 85, 1, 23, "Print"],
Cell[42643, 1254, 349, 8, 23, "Print"],
Cell[42995, 1264, 423, 8, 23, "Print"],
Cell[43421, 1274, 432, 8, 23, "Print"],
Cell[43856, 1284, 400, 8, 23, "Print"],
Cell[44259, 1294, 85, 1, 23, "Print"],
Cell[44347, 1297, 76, 1, 23, "Print"],
Cell[44426, 1300, 85, 1, 23, "Print"],
Cell[44514, 1303, 348, 8, 23, "Print"],
Cell[44865, 1313, 422, 8, 23, "Print"],
Cell[45290, 1323, 391, 8, 23, "Print"],
Cell[45684, 1333, 85, 1, 23, "Print"],
Cell[45772, 1336, 76, 1, 23, "Print"],
Cell[45851, 1339, 85, 1, 23, "Print"],
Cell[45939, 1342, 349, 8, 23, "Print"],
Cell[46291, 1352, 390, 8, 23, "Print"],
Cell[46684, 1362, 85, 1, 23, "Print"],
Cell[46772, 1365, 76, 1, 23, "Print"],
Cell[46851, 1368, 85, 1, 23, "Print"],
Cell[46939, 1371, 350, 8, 23, "Print"],
Cell[47292, 1381, 422, 8, 23, "Print"],
Cell[47717, 1391, 390, 8, 23, "Print"],
Cell[48110, 1401, 85, 1, 23, "Print"],
Cell[48198, 1404, 76, 1, 23, "Print"],
Cell[48277, 1407, 85, 1, 23, "Print"],
Cell[48365, 1410, 349, 8, 23, "Print"],
Cell[48717, 1420, 390, 8, 23, "Print"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[49156, 1434, 154, 4, 52, "Section",
  Evaluatable->False,
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[49313, 1440, 703, 15, 119, "SmallText",
  Evaluatable->False],
Cell[50019, 1457, 134, 3, 27, "Input"],
Cell[50156, 1462, 170, 4, 27, "SmallText",
  Evaluatable->False],
Cell[50329, 1468, 156, 4, 27, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],

Cell[CellGroupData[{
Cell[50510, 1476, 220, 6, 27, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[50733, 1484, 85, 1, 23, "Print"],
Cell[50821, 1487, 76, 1, 23, "Print"],
Cell[50900, 1490, 85, 1, 23, "Print"],
Cell[50988, 1493, 372, 9, 23, "Print"],
Cell[51363, 1504, 451, 9, 23, "Print"],
Cell[51817, 1515, 455, 9, 25, "Print"],
Cell[52275, 1526, 459, 9, 25, "Print"],
Cell[52737, 1537, 418, 9, 25, "Print"],
Cell[53158, 1548, 446, 9, 61, "Output"]
}, Open  ]],
Cell[53619, 1560, 155, 4, 27, "SmallText",
  Evaluatable->False]
}, Open  ]],

Cell[CellGroupData[{
Cell[53811, 1569, 103, 2, 52, "Section",
  Evaluatable->False],
Cell[53917, 1573, 1368, 23, 209, "SmallText",
  Evaluatable->False],
Cell[55288, 1598, 213, 6, 48, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[55504, 1606, 515, 12, 59, "SmallText",
  Evaluatable->False],
Cell[56022, 1620, 156, 4, 27, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],

Cell[CellGroupData[{
Cell[56203, 1628, 193, 5, 27, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[56399, 1635, 85, 1, 23, "Print"],
Cell[56487, 1638, 76, 1, 23, "Print"],
Cell[56566, 1641, 85, 1, 23, "Print"],
Cell[56654, 1644, 366, 9, 23, "Print"],
Cell[57023, 1655, 443, 9, 23, "Print"],
Cell[57469, 1666, 452, 9, 23, "Print"],
Cell[57924, 1677, 457, 9, 25, "Print"],
Cell[58384, 1688, 457, 9, 25, "Print"],
Cell[58844, 1699, 420, 9, 61, "Output"]
}, Open  ]],
Cell[59279, 1711, 162, 4, 27, "SmallText",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[59466, 1719, 282, 8, 43, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[59751, 1729, 387, 10, 43, "Output"]
}, Open  ]],

Cell[CellGroupData[{
Cell[60175, 1744, 384, 9, 75, "Input",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic],
Cell[60562, 1755, 18248, 559, 186, 6318, 406, "GraphicsData", 
"PostScript", "Graphics",
  PageBreakWithin->Automatic,
  PageBreakBelow->Automatic]
}, Open  ]],
Cell[78825, 2317, 43, 1, 27, "Input"]
}, Open  ]],

Cell[CellGroupData[{
Cell[78905, 2323, 101, 2, 52, "Section",
  Evaluatable->False],
Cell[79009, 2327, 331, 10, 44, "SmallText",
  Evaluatable->False]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

