(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     10404,        307]*)
(*NotebookOutlinePosition[     11316,        338]*)
(*  CellTagsIndexPosition[     11272,        334]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData[
"Interpolation with rational functions"], "Title",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[TextData["using NewtonInterpolant "], "Subtitle",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"  Copyright ( c )  (1993-97) J. W. Helton and O. Merino.\nLab. for \
Mathematics and Statistics, University of California San Diego, La Jolla, CA \
92093."], "SmallText",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["0.  Introduction"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"Let R[ s ] be a function of the complex variable s, and let\n\n           { \
s1 ,  s2 , . . . ,  sn } , \n           { z1 , z2 , . . . ,  zn } \n          \
 \nbe given sets of  n complex numbers.  The relation R[ s1 ] = z1 is called \
interpolation condition, and several of these form a \"set of interpolation \
conditions\":\n\n                  R[ s1 ] = z1 ,  R[ s2 ] = z2 ,  . . .  , \
R[ sn ] = zn                  ( INT ) \n\nA function R[ ] that satisfies ( \
INT ) is called an interpolant for ( INT ).\n\nThe function \
NewtonInterpolant[ ] produces a rational function, say R[ s ], that satisfies \
a set of interpolation conditions specified by the user as input, and has all \
its poles at a single location, typically a negative real number.  \n\nThere \
are many rational functions that satisfy ( INT ).  NewtonInterpolant[ ] \
produces one that is the most economical in the sense that it is proper and \
the denominator has the smallest degree possible.\n\nNewtonInterpolant[ ] can \
also deal with more general problems, with interpolation conditions on the \
derivatives of the function."], "Text",
  Evaluatable->False,
  AspectRatioFixed->True]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["1.  First calculation of an interpolant"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData["Begin by loading the package:"], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[
    \(<< "\<NewtonInterpolant`\>"\)], "Input",
  AspectRatioFixed->True],

Cell[TextData[
"The problem is to find an interpolant  R[ s ]  for the interpolation \
conditions\n\n                 ( INT )                 R[ 3 ] = 1 ,   R[ 1 ] \
= -2\n\nThe answer is obtained with                   "], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(data = {{3, 1}, {1, \(-2\)}}; \)\), 
    \(rat1 = NewtonInterpolant[data, s]\)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(4\ \((\(-2\) + s)\)\)\/\(1 + s\)\)], "Output"]
}, Open  ]],

Cell[TextData[
"To verify that the rational function is in fact an interpolant is easy:"], 
  "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(rat1 /. s \[Rule] 3\)], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(1\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["2.  Specifying the pole location"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"In Example 1 the pole location of the interpolant is s0 = -1.  You can \
change this value by giving a  pole location as input.  For example, suppose \
you want pole location -4 for the interpolant in example 1.  This is what you \
type:"], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(rat2 = NewtonInterpolant[data, s, PoleLocation \[Rule] \(-4\)]\)], 
  "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(\(-37\) + 17\ s\)\/\(2\ \((4 + s)\)\)\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["3.  Specifying of the relative degree"], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[
"If R = N/D is a rational function with N as numerator and D as denominator,  \
the relative degree of R is the integer\n\n             d( R  )  =  degree( N \
)  -  degree( D )\n\nfunctions with nonnegative relative degree are called \
proper, and if the relative degree is positive the function is called \
strictly proper.\n\nThe user can specify a relative degree as input for \
NewtonInterpolant[ ].  Suppose that you want to determine an interpolant for  \
   \n\n                 ( INT )                 R[ 3 ] = 1 ,   R[ 1 ] = -2\n \
\nwith pole location -4 and relative degree 2.  This is how you can do it:"], 
  "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(rat2 = 
      NewtonInterpolant[data, s, PoleLocation \[Rule] \(-4\), 
        RelativeDegree \[Rule] 2]\)], "Input",
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(\(-1093\) + 593\ s\)\/\(2\ \((4 + s)\)\^3\)\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["4.  Complex numbers as data."], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "Rational functions that appear in engineering have real coefficients.  \
This corresponds in terms of zeros and poles to the statement that complex \
zeros and complex poles come in conjugate pairs.  These functions are called \
Real-rational, and they have the property :\n\n( RR )                         \
   R[ Conjugate[ s] ] = Conjugate[ R[ s ] ]\n\nIf some of the data you use to \
set up an interpolation problem is complex and you want an answer that is \
real-rational, then you must include pairs of data points that reflect \
property (RR) .   For example, if one interpolation condition is \n\n         \
                              R[ 3 + 2 I ] = 1 - 5 I\n\nin order to obtain a \
real-rational interpolant your set of interpolation conditions must include  \
both\n\n                                       R[ 3 + 2 I ] = 1 - 5 I    and  \
   R[ 3 - 2 I ] = 1 + 5 I\n\nNote: if one of the conditions is not given as \
input, NewtonInterpolant[ ] will give as interpolant a non real-rational \
function, since it will assume that you are willing to give up real rational \
in order to produce the most \"economical\" answer.\n\n",
  StyleBox["Problem:",
    FontSlant->"Italic"],
  " \n\nFind a real-rational function R such that R[ I ] = 0,  R[ 2 + 3 I ] = \
4 - 2 I,  R[ 1 ] = 3.\n\n",
  StyleBox["Solution:",
    FontSlant->"Italic"],
  "    First expand the set of interpolation conditions to\n\n( INT )         \
 R[ I ] = 0 ,  R[ - I ] = 0 ,  R[ 2 + 3 I ] = 4 - 2 I,  R[ 2 - 3 I ] = 4 + 2 \
I ,  R[ 1 ] = 3.\n\nand follow with "
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(data
       = {{I, 0}, {\(-I\), 0}, {2 + 3\ I, 4 - 2\ I}, {2 - 3\ I, 4 + 2\ I}, {
          1, 3}}; \)\), 
    \(rat4 = NewtonInterpolant[data, s]\)}], "Input",
  PageBreakBelow->Automatic,
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(-\(\(3\ \((4 - 13\ s + 5\ s\^2 - 13\ s\^3 + s\^4)\)\)\/\((1 + s)\)\^4
            \)\)\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData["5.  Higher order interpolation."], "Section",
  Evaluatable->False,
  PageBreakAbove->True,
  AspectRatioFixed->True],

Cell[TextData[
"More general interpolation conditions are produced when, besides the value \
of the function at a point, values of the derivatives of this function at the \
point are specified.  NewtonInterpolant can handle this case too.  \n\n\
Problem:   Find an interpolant R  such that    R[ 1 ] = -3  ,  R[ 2 ] = 0 , \
and   R'[ 2 ] = 1                "], "Text",
  Evaluatable->False,
  PageBreakWithin->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(data = {{1, \(-3\)}, {2, 0, 1}}; \)\), 
    \(rat = NewtonInterpolant[data, s]\)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(-\(\(3\ \((10 - 7\ s + s\^2)\)\)\/\((1 + s)\)\^2\)\)\)], "Output"]
}, Open  ]],

Cell[TextData[
"Problem:   Find a real rational interpolant R  with relative degree 1 such \
that    \n                                   R[ 2 I ] = 0 , and   R'[ 2 I ] = \
1 - 3 I \n                                   \nSolution:   In this case it is \
necessary to consider pairs of complex data as follows.                       \
            "], "Text",
  Evaluatable->False,
  PageBreakWithin->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(data = {{2\ I, 0, 1 - 3\ I}, {\(-2\)\ I, 0, 1 + 3\ I}}; \)\), 
    \(NewtonInterpolant[data, s, RelativeDegree \[Rule] 1]\)}], "Input",
  AspectRatioFixed->True],

Cell[BoxData[
    \(\(\(-24\) + 316\ s - 6\ s\^2 + 79\ s\^3\)\/\(8\ \((1 + s)\)\^4\)\)], 
  "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
WindowToolbars->{"RulerBar", "EditBar"},
CellGrouping->Manual,
WindowSize->{720, 589},
WindowMargins->{{Automatic, 107}, {16, Automatic}},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, 128}},
ShowCellLabel->True,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
CharacterEncoding->"XAutomaticEncoding"
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 113, 3, 107, "Title"],

Cell[CellGroupData[{
Cell[1869, 58, 102, 2, 54, "Subtitle"],
Cell[1974, 62, 234, 5, 42, "SmallText"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2245, 72, 93, 2, 54, "Section"],
Cell[2341, 76, 1183, 17, 356, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3561, 98, 116, 2, 54, "Section"],
Cell[3680, 102, 103, 2, 32, "Text"],
Cell[3786, 106, 86, 2, 27, "Input"],
Cell[3875, 110, 279, 5, 104, "Text"],

Cell[CellGroupData[{
Cell[4179, 119, 139, 3, 43, "Input"],
Cell[4321, 124, 68, 1, 44, "Output"]
}, Open  ]],
Cell[4404, 128, 149, 4, 32, "Text"],

Cell[CellGroupData[{
Cell[4578, 136, 78, 2, 27, "Input"],
Cell[4659, 140, 35, 1, 27, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[4743, 147, 109, 2, 54, "Section"],
Cell[4855, 151, 312, 6, 50, "Text"],

Cell[CellGroupData[{
Cell[5192, 161, 124, 3, 27, "Input"],
Cell[5319, 166, 73, 1, 46, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[5441, 173, 114, 2, 54, "Section"],
Cell[5558, 177, 693, 11, 212, "Text"],

Cell[CellGroupData[{
Cell[6276, 192, 192, 5, 27, "Input",
  PageBreakBelow->Automatic],
Cell[6471, 199, 79, 1, 46, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[6599, 206, 105, 2, 54, "Section"],
Cell[6707, 210, 1628, 27, 482, "Text"],

Cell[CellGroupData[{
Cell[8360, 241, 238, 6, 43, "Input",
  PageBreakBelow->Automatic],
Cell[8601, 249, 122, 2, 48, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[8772, 257, 132, 3, 54, "Section",
  PageBreakAbove->True],
Cell[8907, 262, 442, 8, 86, "Text",
  PageBreakWithin->False],

Cell[CellGroupData[{
Cell[9374, 274, 141, 3, 43, "Input"],
Cell[9518, 279, 88, 1, 48, "Output"]
}, Open  ]],
Cell[9621, 283, 430, 8, 86, "Text",
  PageBreakWithin->False],

Cell[CellGroupData[{
Cell[10076, 295, 184, 3, 43, "Input"],
Cell[10263, 300, 101, 2, 48, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

