(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    357321,      13710]*)
(*NotebookOutlinePosition[    358113,      13738]*)
(*  CellTagsIndexPosition[    358069,      13734]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["OPTDesign plots, data and functions", "Title"],

Cell[TextData[StyleBox["(C) 1998 J.W.Helton and O. Merino"]], "SmallText"],

Cell[CellGroupData[{

Cell["\<\
Some users may want to manipulate the output of an OPTDesign run \
before dealing with rational fits.  We present examples of OPTDesign commands \
to plot  and manipulate  T ,  L,  Co or other lists of  data.\
\>", 
  "Subsubtitle"],

Cell[CellGroupData[{

Cell[TextData[StyleBox[
"Example from Chapter 5:  run this before proceeding with the rest of the \
notebook"]], "Subsection"],

Cell[TextData[{
  StyleBox["The gridsize "],
  StyleBox["n",
    FontWeight->"Bold"],
  StyleBox[" has been set to 32.  You may change it if you wish."]
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
<<OPTDesign`;

        n = 32;  (* this sets the gridsize *)

     p[s_]= (1 - s/5)/(1+s/2)^2;

      wt  = 0.3;
  alphat  = 1.5;

 alphagpm = 2.0;

       wb = 2.0;
   alphab = 0.7;
              
       wr = 4.0;
   alphar = 1.5;
 
k0[w_]= Which[ Abs[w] < wb, 1.0,
                     True, 0.0];
 
r0[w_] = Which[Abs[w] < wt,  alphat,
               Abs[w] < wb,  alphagpm,
               Abs[w] < wr,  alphab,
                     True ,  alphar*Abs[p[I w]]
                ];     

OPTDesign[p,Radius->r0,Center->k0,Ngrid -> n];\
\>", "Input"],

Cell[BoxData[
    \(" \n Anopt 3.0 (c) Copyright 1991-94 \n J.W.Helton and O.Merino.  All \
rights reserved."\)], "Print"],

Cell[BoxData[
    \("OPTDesign 2.0 (c) 1991-98 \n   J.W. Helton and O.Merino.  All rights \
reserved."\)], "Print"],

Cell[BoxData[
    \(General::"spell1" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(alphab\)\" is similar \
to existing symbol \"\!\(alphat\)\"."\)], "Message"],

Cell[BoxData[
    \(General::"spell" \( : \ \) 
      "Possible spelling error: new symbol name \"\!\(alphar\)\" is similar \
to existing symbols \!\({alphab, alphat}\)."\)], "Message"],

Cell[BoxData[
    \("\nParameterization for internal stability:"\)], "Print"],

Cell[BoxData[
    \("  T = A + B * T1, where T1 is RHP stable and"\)], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("A[s]="\[InvisibleSpace]\(\( - 5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2\),
      SequenceForm[ "A[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    InterpretationBox[
      \("B[s]="\[InvisibleSpace]\(\( - 5.`\) + s
            \)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2\),
      SequenceForm[ "B[s]=", 
        Times[ 
          Plus[ -5.0, s], 
          Power[ 
            Plus[ 2.0, s], -2]]],
      Editable->False]], "Print"],

Cell[BoxData[
    \("  "\)], "Print"],

Cell[BoxData[
    \("Processing performance function..."\)], "Print"],

Cell[BoxData[
    \("Sampling Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Smoothing Radius and Center ..."\)], "Print"],

Cell[BoxData[
    \("Optimization routine Anopt now number-crunching ..."\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("It :    Current Value    :  Step   : Optimality Tests  :  Error  : Sm. \
: Grid"\)], "Print"],

Cell[BoxData[
    \("   :     gammastar       :         :   flat   grAlign  :   ned   :     \
:"\)], "Print"],

Cell[BoxData[
    \("-----------------------------------------------------------------------\
------"\)], "Print"],

Cell[BoxData[
    \(" 0 : 2.2096082076563E+00 :   N/A   : 8.7E-01 : 1. E+00 :   N/A   : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" 1 : 1.4876204762846E-01 : 1.6E+00 : 1.3E-03 : 0  E+00 : 3.7E-03 : NON \
: 32"\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \(" "\)], "Print"],

Cell[BoxData[
    \("Summary"\)], "Print"],

Cell[BoxData[
    \("-------"\)], "Print"],

Cell[BoxData[
    \("gammastar = 1.487620476284562E-01"\)], "Print"],

Cell[BoxData[
    \("flat      = 1.2642945402E-03"\)], "Print"],

Cell[BoxData[
    \("grAlign   = 0           E+00"\)], "Print"],

Cell[BoxData[
    \("ned      = 3.73E-03"\)], "Print"],

Cell[BoxData[
    \("Calculating output functions ..."\)], "Print"],

Cell[BoxData[
    \("Resetting options ..."\)], "Print"],

Cell[BoxData[
    \("Done! "\)], "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Functions and grids"]], "Section"],

Cell[CellGroupData[{

Cell["Output functions of OPTDesign runs", "Subsection"],

Cell[TextData[{
  "A run of OPTDesign produces the calculated closed loop T, the open loop L \
and the compensator Co as lists.  If you type T (say) in a session after you \
run OPTDesign, then Mathematica returns a list of complex numbers that are \
the values of the calculated closed loop function values on a w-axis grid.  \n\
",
  StyleBox[
  "Note that the grid points on the w axis do not appear explicitly when you \
ask Mathematica to show T",
    FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell["T ", "Input"],

Cell[BoxData[
    \({0, \(-0.0371551406288340135`\) + 0.0341142236154496458`\ I, 
      \(-0.124852139294746411`\) + 0.00349232094116091484`\ I, 
      \(-0.184927783065699795`\) - 0.129108921856047164`\ I, 
      \(-0.114989309350481461`\) - 0.306714653006541837`\ I, 
      \(0.0780775913006389643`\[InvisibleSpace]\) - 0.371381458962195099`\ I, 
      \(0.212941763857875709`\[InvisibleSpace]\) - 0.285463537533383604`\ I, 
      \(0.236438544137187012`\[InvisibleSpace]\) - 0.199328365308105048`\ I, 
      \(0.231823252482546848`\[InvisibleSpace]\) - 0.160042465352175033`\ I, 
      \(0.230234597326233858`\[InvisibleSpace]\) - 0.141843233028048469`\ I, 
      \(0.230141527259049105`\[InvisibleSpace]\) - 0.133257078822847652`\ I, 
      \(0.233163554454544197`\[InvisibleSpace]\) - 0.13625322983822552`\ I, 
      \(0.250685132540256372`\[InvisibleSpace]\) - 0.152636898312056734`\ I, 
      \(0.301351692212556088`\[InvisibleSpace]\) - 0.162629006848771018`\ I, 
      \(0.365522868563008085`\[InvisibleSpace]\) - 0.13106073022916882`\ I, 
      \(0.403420109773417756`\[InvisibleSpace]\) - 0.066914265044986454`\ I, 
      \(0.412220079811065787`\[InvisibleSpace]\) + 0.`\ I, 
      \(0.403420109773417756`\[InvisibleSpace]\) + 0.0669142650449864095`\ I, 
      \(0.365522868563008085`\[InvisibleSpace]\) + 0.131060730229168753`\ I, 
      \(0.301351692212556088`\[InvisibleSpace]\) + 0.162629006848770796`\ I, 
      \(0.250685132540256372`\[InvisibleSpace]\) + 0.152636898312056734`\ I, 
      \(0.23316355445454393`\[InvisibleSpace]\) + 0.136253229838225298`\ I, 
      \(0.230141527259049105`\[InvisibleSpace]\) + 0.133257078822847652`\ I, 
      \(0.23023459732623408`\[InvisibleSpace]\) + 0.141843233028048469`\ I, 
      \(0.231823252482546848`\[InvisibleSpace]\) + 0.160042465352175033`\ I, 
      \(0.236438544137187101`\[InvisibleSpace]\) + 0.199328365308105261`\ I, 
      \(0.212941763857875621`\[InvisibleSpace]\) + 0.285463537533383604`\ I, 
      \(0.0780775913006388222`\[InvisibleSpace]\) + 0.371381458962195321`\ I, 
      \(-0.114989309350481461`\) + 0.306714653006541837`\ I, 
      \(-0.184927783065699902`\) + 0.129108921856047098`\ I, 
      \(-0.124852139294746411`\) - 0.00349232094116090951`\ I, 
      \(-0.0371551406288340135`\) - 0.0341142236154496458`\ I}\)], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
What is the grid you are currently using in an OPTDesign session?\
\
\>", "Subsection"],

Cell["\<\
The assumption is that there is a grid that is used for sampling \
all functions of the OPTDesign session.  To see it type\
\>", "Text"],

Cell[CellGroupData[{

Cell["Grid[]", "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        InterpretationBox["\[Infinity]",
          DirectedInfinity[ 1]], ",", "10.1531703876088497`", ",", 
        "5.02733949212584363`", ",", "3.29655820893832007`", ",", 
        "2.41421356237309447`", ",", "1.870868411789389`", ",", 
        "1.49660576266548916`", ",", "1.21850352558797636`", ",", "1.`", ",", 
        "0.820678790828660176`", ",", "0.668178637919298701`", ",", 
        "0.534511135950791693`", ",", "0.414213562373095012`", ",", 
        "0.303346683607342271`", ",", "0.198912367379658014`", ",", 
        "0.0984914033571643443`", ",", "0.`", ",", 
        \(-0.0984914033571643443`\), ",", \(-0.198912367379658014`\), ",", 
        \(-0.303346683607342271`\), ",", \(-0.414213562373095012`\), ",", 
        \(-0.534511135950791693`\), ",", \(-0.668178637919298701`\), ",", 
        \(-0.820678790828660176`\), ",", \(-1.`\), ",", 
        \(-1.21850352558797636`\), ",", \(-1.49660576266548916`\), ",", 
        \(-1.870868411789389`\), ",", \(-2.41421356237309447`\), ",", 
        \(-3.29655820893832007`\), ",", \(-5.02733949212584363`\), ",", 
        \(-10.1531703876088497`\)}], "}"}]], "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Putting together the grid and the values of a function.", "Subsection"],

Cell["\<\
If you wish you may produce a list of pairs of the form { w , T[ I \
w ] }.  To do this, type\
\>", "Text"],

Cell[CellGroupData[{

Cell["Tpairs = OPTDParametrize[T]", "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{"{", 
          RowBox[{
            InterpretationBox["\[Infinity]",
              DirectedInfinity[ 1]], ",", "0"}], "}"}], ",", 
        \({10.1531703876088497`, 
          \(-0.0371551406288340135`\) + 0.0341142236154496458`\ I}\), ",", 
        \({5.02733949212584363`, 
          \(-0.124852139294746411`\) + 0.00349232094116091484`\ I}\), ",", 
        \({3.29655820893832007`, 
          \(-0.184927783065699795`\) - 0.129108921856047164`\ I}\), ",", 
        \({2.41421356237309447`, 
          \(-0.114989309350481461`\) - 0.306714653006541837`\ I}\), ",", 
        \({1.870868411789389`, 
          \(0.0780775913006389643`\[InvisibleSpace]\) - 
            0.371381458962195099`\ I}\), ",", 
        \({1.49660576266548916`, 
          \(0.212941763857875709`\[InvisibleSpace]\) - 
            0.285463537533383604`\ I}\), ",", 
        \({1.21850352558797636`, 
          \(0.236438544137187012`\[InvisibleSpace]\) - 
            0.199328365308105048`\ I}\), ",", 
        \({1.`, \(0.231823252482546848`\[InvisibleSpace]\) - 
            0.160042465352175033`\ I}\), ",", 
        \({0.820678790828660176`, 
          \(0.230234597326233858`\[InvisibleSpace]\) - 
            0.141843233028048469`\ I}\), ",", 
        \({0.668178637919298701`, 
          \(0.230141527259049105`\[InvisibleSpace]\) - 
            0.133257078822847652`\ I}\), ",", 
        \({0.534511135950791693`, 
          \(0.233163554454544197`\[InvisibleSpace]\) - 
            0.13625322983822552`\ I}\), ",", 
        \({0.414213562373095012`, 
          \(0.250685132540256372`\[InvisibleSpace]\) - 
            0.152636898312056734`\ I}\), ",", 
        \({0.303346683607342271`, 
          \(0.301351692212556088`\[InvisibleSpace]\) - 
            0.162629006848771018`\ I}\), ",", 
        \({0.198912367379658014`, 
          \(0.365522868563008085`\[InvisibleSpace]\) - 
            0.13106073022916882`\ I}\), ",", 
        \({0.0984914033571643443`, 
          \(0.403420109773417756`\[InvisibleSpace]\) - 
            0.066914265044986454`\ I}\), ",", 
        \({0.`, \(0.412220079811065787`\[InvisibleSpace]\) + 0.`\ I}\), ",", 
        \({\(-0.0984914033571643443`\), 
          \(0.403420109773417756`\[InvisibleSpace]\) + 
            0.0669142650449864095`\ I}\), ",", 
        \({\(-0.198912367379658014`\), 
          \(0.365522868563008085`\[InvisibleSpace]\) + 
            0.131060730229168753`\ I}\), ",", 
        \({\(-0.303346683607342271`\), 
          \(0.301351692212556088`\[InvisibleSpace]\) + 
            0.162629006848770796`\ I}\), ",", 
        \({\(-0.414213562373095012`\), 
          \(0.250685132540256372`\[InvisibleSpace]\) + 
            0.152636898312056734`\ I}\), ",", 
        \({\(-0.534511135950791693`\), 
          \(0.23316355445454393`\[InvisibleSpace]\) + 
            0.136253229838225298`\ I}\), ",", 
        \({\(-0.668178637919298701`\), 
          \(0.230141527259049105`\[InvisibleSpace]\) + 
            0.133257078822847652`\ I}\), ",", 
        \({\(-0.820678790828660176`\), 
          \(0.23023459732623408`\[InvisibleSpace]\) + 
            0.141843233028048469`\ I}\), ",", 
        \({\(-1.`\), 
          \(0.231823252482546848`\[InvisibleSpace]\) + 
            0.160042465352175033`\ I}\), ",", 
        \({\(-1.21850352558797636`\), 
          \(0.236438544137187101`\[InvisibleSpace]\) + 
            0.199328365308105261`\ I}\), ",", 
        \({\(-1.49660576266548916`\), 
          \(0.212941763857875621`\[InvisibleSpace]\) + 
            0.285463537533383604`\ I}\), ",", 
        \({\(-1.870868411789389`\), 
          \(0.0780775913006388222`\[InvisibleSpace]\) + 
            0.371381458962195321`\ I}\), ",", 
        \({\(-2.41421356237309447`\), 
          \(-0.114989309350481461`\) + 0.306714653006541837`\ I}\), ",", 
        \({\(-3.29655820893832007`\), 
          \(-0.184927783065699902`\) + 0.129108921856047098`\ I}\), ",", 
        \({\(-5.02733949212584363`\), 
          \(-0.124852139294746411`\) - 0.00349232094116090951`\ I}\), ",", 
        \({\(-10.1531703876088497`\), 
          \(-0.0371551406288340135`\) - 0.0341142236154496458`\ I}\)}], 
      "}"}]], "Output"]
}, Open  ]],

Cell["Let's check the size of the list Tpairs:", "Text"],

Cell[CellGroupData[{

Cell["Dimensions[Tpairs]", "Input"],

Cell[BoxData[
    \({32, 2}\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Plots"]], "Section"],

Cell[CellGroupData[{

Cell["Plotting the envelope and T simultaneously", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
EnvelopePlot3D[Radius -> r0 , Center -> k0 , ClosedLoop -> T \
];\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.01205 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0265144 1.04793 0 1.04793 [
[.2389 1.02513 -3.17888 0 ]
[.2389 1.02513 2.82112 9 ]
[.37774 .98992 -3.0835 0 ]
[.37774 .98992 2.9165 9 ]
[.52296 .95309 -2.98205 0 ]
[.52296 .95309 3.01795 9 ]
[.67501 .91451 -2.87394 0 ]
[.67501 .91451 3.12606 9 ]
[.83437 .87405 -2.75849 0 ]
[.83437 .87405 3.24151 9 ]
[1.00158 .83158 -2.63492 0 ]
[1.00158 .83158 3.36508 9 ]
[.59965 .9995 -4.88145 0 ]
[.59965 .9995 5.11855 15 ]
[.04392 .88272 -3.32443 0 ]
[.04392 .88272 2.67557 9 ]
[.09461 .91974 -9.85694 0 ]
[.09461 .91974 8.14306 9 ]
[.14316 .9552 -3.24914 0 ]
[.14316 .9552 2.75086 9 ]
[.18969 .98919 -9.64411 0 ]
[.18969 .98919 8.35589 9 ]
[.23434 1.0218 -3.18217 0 ]
[.23434 1.0218 2.81783 9 ]
[.11933 1.00721 -18.4886 0 ]
[.11933 1.00721 15.5114 15 ]
[.94483 .25907 0 -6.02028 ]
[.94483 .25907 12 2.97972 ]
[.95965 .38048 0 -5.93776 ]
[.95965 .38048 6 3.06224 ]
[.9755 .51042 0 -5.84984 ]
[.9755 .51042 6 3.15016 ]
[.99251 .64981 0 -5.75596 ]
[.99251 .64981 6 3.24404 ]
[1.01079 .79973 0 -5.65551 ]
[1.01079 .79973 6 3.34449 ]
[1.03824 .4916 0 -9.74973 ]
[1.03824 .4916 34 5.25027 ]
[ 0 0 0 0 ]
[ 1 1.01205 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.23968 1.01205 m
1 .81858 L
s
.23968 1.01205 m
.24007 1.00551 L
s
[(0)] .2389 1.02513 .05963 -1 Mshowa
.37811 .97683 m
.37829 .97028 L
s
[(1)] .37774 .98992 .02783 -1 Mshowa
.52288 .93999 m
.52284 .93344 L
s
[(2)] .52296 .95309 -0.00598 -1 Mshowa
.67446 .90142 m
.67418 .89488 L
s
[(3)] .67501 .91451 -0.04202 -1 Mshowa
.83331 .861 m
.83279 .85447 L
s
[(4)] .83437 .87405 -0.0805 -1 Mshowa
1 .81858 m
.99921 .81208 L
s
[(5)] 1.00158 .83158 -0.12169 -1 Mshowa
.125 Mabswid
.26688 1.00513 m
.26709 1.00121 L
s
.29432 .99815 m
.2945 .99422 L
s
.322 .99111 m
.32216 .98718 L
s
.34993 .984 m
.35006 .98007 L
s
.40654 .96959 m
.40662 .96566 L
s
.43523 .96229 m
.43529 .95836 L
s
.46418 .95493 m
.46421 .951 L
s
.4934 .94749 m
.4934 .94356 L
s
.55264 .93242 m
.55259 .92849 L
s
.58267 .92478 m
.58259 .92085 L
s
.61298 .91706 m
.61287 .91313 L
s
.64357 .90928 m
.64344 .90535 L
s
.70563 .89349 m
.70543 .88956 L
s
.7371 .88548 m
.73687 .88156 L
s
.76886 .8774 m
.76861 .87347 L
s
.80093 .86924 m
.80065 .86532 L
s
.86601 .85268 m
.86566 .84876 L
s
.89902 .84428 m
.89864 .84037 L
s
.93235 .83579 m
.93194 .83189 L
s
.96601 .82723 m
.96557 .82333 L
s
gsave
.59965 .9995 -65.8815 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
0 .83649 m
.23968 1.01205 L
s
.04533 .86969 m
.04603 .86318 L
s
[(0)] .04392 .88272 .10814 -1 Mshowa
.09585 .9067 m
.09647 .90018 L
s
[(0.5)] .09461 .91974 .09522 -1 Mshowa
.14424 .94215 m
.14478 .93562 L
s
[(1)] .14316 .9552 .08305 -1 Mshowa
.19063 .97612 m
.1911 .96959 L
s
[(1.5)] .18969 .98919 .07157 -1 Mshowa
.23513 1.00872 m
.23553 1.00218 L
s
[(2)] .23434 1.0218 .06072 -1 Mshowa
.125 Mabswid
.05561 .87722 m
.05602 .87332 L
s
.0658 .88469 m
.0662 .88078 L
s
.07591 .89209 m
.0763 .88818 L
s
.08592 .89943 m
.0863 .89552 L
s
.1057 .91391 m
.10606 .91 L
s
.11546 .92106 m
.11581 .91715 L
s
.12513 .92815 m
.12548 .92424 L
s
.13473 .93518 m
.13506 .93126 L
s
.15368 .94906 m
.15399 .94514 L
s
.16303 .95591 m
.16334 .95199 L
s
.17231 .9627 m
.1726 .95878 L
s
.18151 .96944 m
.18179 .96552 L
s
.19968 .98275 m
.19995 .97883 L
s
.20865 .98932 m
.20891 .9854 L
s
.21755 .99584 m
.2178 .99192 L
s
.22638 1.00231 m
.22662 .99838 L
s
.03495 .86209 m
.03538 .85819 L
s
.02448 .85443 m
.02493 .85052 L
s
.01392 .84669 m
.01438 .84279 L
s
.00327 .83889 m
.00373 .83499 L
s
gsave
.11933 1.00721 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.931 .25158 m
1 .81858 L
s
.93242 .26326 m
.92622 .26536 L
s
[(-1)] .94483 .25907 -1 .33784 Mshowa
.94717 .38447 m
.94093 .38646 L
s
[(0)] .95965 .38048 -1 .3195 Mshowa
.96296 .51418 m
.95668 .51606 L
s
[(1)] .9755 .51042 -1 .29996 Mshowa
.97989 .65333 m
.97358 .65509 L
s
[(2)] .99251 .64981 -1 .2791 Mshowa
.9981 .80299 m
.99176 .80462 L
s
[(3)] 1.01079 .79973 -1 .25678 Mshowa
.125 Mabswid
.93529 .28686 m
.93157 .28811 L
s
.9382 .31078 m
.93447 .31201 L
s
.94115 .33501 m
.93742 .33623 L
s
.94414 .35957 m
.9404 .36078 L
s
.95024 .4097 m
.94649 .41089 L
s
.95335 .43528 m
.9496 .43645 L
s
.95651 .46122 m
.95276 .46238 L
s
.95971 .48752 m
.95595 .48866 L
s
.96625 .54123 m
.96248 .54234 L
s
.96959 .56865 m
.96581 .56975 L
s
.97297 .59648 m
.96919 .59756 L
s
.97641 .6247 m
.97262 .62577 L
s
.98343 .68239 m
.97964 .68343 L
s
.98701 .71187 m
.98322 .7129 L
s
.99066 .74179 m
.98686 .7428 L
s
.99435 .77216 m
.99055 .77316 L
s
gsave
1.03824 .4916 -61 -13.7497 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.23968 1.01205 L
s
.23968 1.01205 m
.27167 .47549 L
s
.27167 .47549 m
.0677 .27203 L
s
.77416 0 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.77416 0 m
.0677 .27203 L
s
.27167 .47549 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.23968 1.01205 L
s
.23968 1.01205 m
.27167 .47549 L
s
0 0 m
1 0 L
1 1.01205 L
0 1.01205 L
closepath
clip
newpath
.5 Mabswid
.30241 .4381 m .26906 .44957 L .28671 .49523 L p .735 .66 .852 r
F P
0 g
s
.28671 .49523 m .32009 .48403 L .30241 .4381 L p .735 .66 .852 r
F P
0 g
s
.22116 .51827 m .28671 .49523 L .26906 .44957 L p .933 .755 .762 r
F P
0 g
s
.32009 .48403 m .28671 .49523 L .30094 .54206 L p .725 .625 .82 r
F P
0 g
s
.28671 .49523 m .22116 .51827 L .23201 .55408 L p .928 .737 .746 r
F P
0 g
s
.23201 .55408 m .30094 .54206 L .28671 .49523 L p .928 .737 .746 r
F P
0 g
s
.30094 .54206 m .33441 .53111 L .32009 .48403 L p .725 .625 .82 r
F P
0 g
s
.33441 .53111 m .30094 .54206 L .31146 .58917 L p .713 .588 .786 r
F P
0 g
s
.30094 .54206 m .23201 .55408 L .23988 .59009 L p .924 .72 .727 r
F P
0 g
s
.23988 .59009 m .31146 .58917 L .30094 .54206 L p .924 .72 .727 r
F P
0 g
s
.28171 .39424 m .24831 .406 L .26906 .44957 L p .745 .694 .883 r
F P
0 g
s
.26906 .44957 m .30241 .4381 L .28171 .39424 L p .745 .694 .883 r
F P
0 g
s
.20756 .48342 m .26906 .44957 L .24831 .406 L p .94 .774 .776 r
F P
0 g
s
.26906 .44957 m .20756 .48342 L .22116 .51827 L p .933 .755 .762 r
F P
0 g
s
.31146 .58917 m .34509 .57843 L .33441 .53111 L p .713 .588 .786 r
F P
0 g
s
.34509 .57843 m .31146 .58917 L .31803 .63563 L p .699 .548 .748 r
F P
0 g
s
.31146 .58917 m .23988 .59009 L .24457 .62553 L p .921 .703 .707 r
F P
0 g
s
.24457 .62553 m .31803 .63563 L .31146 .58917 L p .921 .703 .707 r
F P
0 g
s
.33612 .42651 m .30241 .4381 L .32009 .48403 L p .735 .66 .852 r
F P
0 g
s
.32009 .48403 m .35382 .47271 L .33612 .42651 L p .735 .66 .852 r
F P
0 g
s
.35382 .47271 m .32009 .48403 L .33441 .53111 L p .725 .625 .82 r
F P
0 g
s
.33441 .53111 m .36824 .52004 L .35382 .47271 L p .725 .625 .82 r
F P
0 g
s
.36824 .52004 m .33441 .53111 L .34509 .57843 L p .713 .588 .786 r
F P
0 g
s
.25834 .35332 m .22482 .36539 L .24831 .406 L p .753 .73 .914 r
F P
0 g
s
.24831 .406 m .28171 .39424 L .25834 .35332 L p .753 .73 .914 r
F P
0 g
s
.1915 .45028 m .24831 .406 L .22482 .36539 L p .949 .793 .788 r
F P
0 g
s
.24831 .406 m .1915 .45028 L .20756 .48342 L p .94 .774 .776 r
F P
0 g
s
.31547 .38235 m .28171 .39424 L .30241 .4381 L p .745 .694 .883 r
F P
0 g
s
.30241 .4381 m .33612 .42651 L .31547 .38235 L p .745 .694 .883 r
F P
0 g
s
.34509 .57843 m .37909 .56758 L .36824 .52004 L p .713 .588 .786 r
F P
0 g
s
.37909 .56758 m .34509 .57843 L .3519 .62508 L p .699 .548 .748 r
F P
0 g
s
.31803 .63563 m .3519 .62508 L .34509 .57843 L p .699 .548 .748 r
F P
0 g
s
.3519 .62508 m .31803 .63563 L .32046 .68049 L p .682 .502 .703 r
F P
0 g
s
.31803 .63563 m .24457 .62553 L .24596 .65963 L p .92 .687 .685 r
F P
0 g
s
.24596 .65963 m .32046 .68049 L .31803 .63563 L p .92 .687 .685 r
F P
0 g
s
.3702 .41479 m .33612 .42651 L .35382 .47271 L p .735 .66 .852 r
F P
0 g
s
.35382 .47271 m .38792 .46126 L .3702 .41479 L p .735 .66 .852 r
F P
0 g
s
.38792 .46126 m .35382 .47271 L .36824 .52004 L p .725 .625 .82 r
F P
0 g
s
.36824 .52004 m .40243 .50884 L .38792 .46126 L p .725 .625 .82 r
F P
0 g
s
.40243 .50884 m .36824 .52004 L .37909 .56758 L p .713 .588 .786 r
F P
0 g
s
.29222 .34112 m .25834 .35332 L .28171 .39424 L p .753 .73 .914 r
F P
0 g
s
.28171 .39424 m .31547 .38235 L .29222 .34112 L p .753 .73 .914 r
F P
0 g
s
.34959 .37033 m .31547 .38235 L .33612 .42651 L p .745 .694 .883 r
F P
0 g
s
.33612 .42651 m .3702 .41479 L .34959 .37033 L p .745 .694 .883 r
F P
0 g
s
.20756 .48342 m .17428 .4948 L .18786 .52944 L p .733 .659 .853 r
F P
0 g
s
.18786 .52944 m .22116 .51827 L .20756 .48342 L p .733 .659 .853 r
F P
0 g
s
.22116 .51827 m .18786 .52944 L .19863 .56506 L p .722 .623 .821 r
F P
0 g
s
.23272 .31625 m .19902 .32864 L .22482 .36539 L p .761 .767 .945 r
F P
0 g
s
.22482 .36539 m .25834 .35332 L .23272 .31625 L p .761 .767 .945 r
F P
0 g
s
.17329 .41958 m .22482 .36539 L .19902 .32864 L p .959 .813 .799 r
F P
0 g
s
.22482 .36539 m .17329 .41958 L .1915 .45028 L p .949 .793 .788 r
F P
0 g
s
.19863 .56506 m .23201 .55408 L .22116 .51827 L p .722 .623 .821 r
F P
0 g
s
.23201 .55408 m .19863 .56506 L .20638 .6009 L p .71 .585 .785 r
F P
0 g
s
.37909 .56758 m .41345 .55661 L .40243 .50884 L p .713 .588 .786 r
F P
0 g
s
.41345 .55661 m .37909 .56758 L .38613 .61442 L p .699 .548 .748 r
F P
0 g
s
.3519 .62508 m .38613 .61442 L .37909 .56758 L p .699 .548 .748 r
F P
0 g
s
.38613 .61442 m .3519 .62508 L .35462 .6701 L p .682 .502 .703 r
F P
0 g
s
.32046 .68049 m .35462 .6701 L .3519 .62508 L p .682 .502 .703 r
F P
0 g
s
.1915 .45028 m .15819 .46188 L .17428 .4948 L p .743 .695 .885 r
F P
0 g
s
.17428 .4948 m .20756 .48342 L .1915 .45028 L p .743 .695 .885 r
F P
0 g
s
.40465 .40294 m .3702 .41479 L .38792 .46126 L p .735 .66 .852 r
F P
0 g
s
.38792 .46126 m .42239 .44969 L .40465 .40294 L p .735 .66 .852 r
F P
0 g
s
.42239 .44969 m .38792 .46126 L .40243 .50884 L p .725 .625 .82 r
F P
0 g
s
.20638 .6009 m .23988 .59009 L .23201 .55408 L p .71 .585 .785 r
F P
0 g
s
.23988 .59009 m .20638 .6009 L .2109 .6362 L p .695 .543 .745 r
F P
0 g
s
.40243 .50884 m .437 .49753 L .42239 .44969 L p .725 .625 .82 r
F P
0 g
s
.437 .49753 m .40243 .50884 L .41345 .55661 L p .713 .588 .786 r
F P
0 g
s
.32647 .32879 m .29222 .34112 L .31547 .38235 L p .753 .73 .914 r
F P
0 g
s
.31547 .38235 m .34959 .37033 L .32647 .32879 L p .753 .73 .914 r
F P
0 g
s
.35462 .6701 m .32046 .68049 L .31858 .72281 L p .66 .447 .648 r
F P
0 g
s
.32046 .68049 m .24596 .65963 L .24393 .69159 L p .921 .67 .659 r
F P
0 g
s
.24393 .69159 m .31858 .72281 L .32046 .68049 L p .921 .67 .659 r
F P
0 g
s
.3841 .35818 m .34959 .37033 L .3702 .41479 L p .745 .694 .883 r
F P
0 g
s
.3702 .41479 m .40465 .40294 L .3841 .35818 L p .745 .694 .883 r
F P
0 g
s
.17329 .41958 m .13989 .43141 L .15819 .46188 L p .751 .731 .917 r
F P
0 g
s
.15819 .46188 m .1915 .45028 L .17329 .41958 L p .751 .731 .917 r
F P
0 g
s
.26679 .30373 m .23272 .31625 L .25834 .35332 L p .761 .767 .945 r
F P
0 g
s
.25834 .35332 m .29222 .34112 L .26679 .30373 L p .761 .767 .945 r
F P
0 g
s
.41345 .55661 m .4482 .54552 L .437 .49753 L p .713 .588 .786 r
F P
0 g
s
.4482 .54552 m .41345 .55661 L .42073 .60364 L p .699 .548 .748 r
F P
0 g
s
.38613 .61442 m .42073 .60364 L .41345 .55661 L p .699 .548 .748 r
F P
0 g
s
.42073 .60364 m .38613 .61442 L .38915 .6596 L p .682 .502 .703 r
F P
0 g
s
.35462 .6701 m .38915 .6596 L .38613 .61442 L p .682 .502 .703 r
F P
0 g
s
.2109 .6362 m .24457 .62553 L .23988 .59009 L p .695 .543 .745 r
F P
0 g
s
.24457 .62553 m .2109 .6362 L .21207 .67017 L p .677 .495 .699 r
F P
0 g
s
.43948 .39096 m .40465 .40294 L .42239 .44969 L p .735 .66 .852 r
F P
0 g
s
.42239 .44969 m .45724 .438 L .43948 .39096 L p .735 .66 .852 r
F P
0 g
s
.45724 .438 m .42239 .44969 L .437 .49753 L p .725 .625 .82 r
F P
0 g
s
.437 .49753 m .47195 .48609 L .45724 .438 L p .725 .625 .82 r
F P
0 g
s
.47195 .48609 m .437 .49753 L .4482 .54552 L p .713 .588 .786 r
F P
0 g
s
.3611 .31632 m .32647 .32879 L .34959 .37033 L p .753 .73 .914 r
F P
0 g
s
.34959 .37033 m .3841 .35818 L .3611 .31632 L p .753 .73 .914 r
F P
0 g
s
.38915 .6596 m .35462 .6701 L .35309 .71254 L p .66 .447 .648 r
F P
0 g
s
.31858 .72281 m .35309 .71254 L .35462 .6701 L p .66 .447 .648 r
F P
0 g
s
.41898 .3459 m .3841 .35818 L .40465 .40294 L p .745 .694 .883 r
F P
0 g
s
.40465 .40294 m .43948 .39096 L .41898 .3459 L p .745 .694 .883 r
F P
0 g
s
.20532 .28393 m .17137 .29664 L .19902 .32864 L p .768 .809 .978 r
F P
0 g
s
.19902 .32864 m .23272 .31625 L .20532 .28393 L p .768 .809 .978 r
F P
0 g
s
.15329 .39205 m .19902 .32864 L .17137 .29664 L p .971 .833 .807 r
F P
0 g
s
.19902 .32864 m .15329 .39205 L .17329 .41958 L p .959 .813 .799 r
F P
0 g
s
.15329 .39205 m .11976 .40411 L .13989 .43141 L p .759 .77 .949 r
F P
0 g
s
.13989 .43141 m .17329 .41958 L .15329 .39205 L p .759 .77 .949 r
F P
0 g
s
.30123 .29107 m .26679 .30373 L .29222 .34112 L p .761 .767 .945 r
F P
0 g
s
.29222 .34112 m .32647 .32879 L .30123 .29107 L p .761 .767 .945 r
F P
0 g
s
.4482 .54552 m .48332 .5343 L .47195 .48609 L p .713 .588 .786 r
F P
0 g
s
.48332 .5343 m .4482 .54552 L .45572 .59274 L p .699 .548 .748 r
F P
0 g
s
.42073 .60364 m .45572 .59274 L .4482 .54552 L p .699 .548 .748 r
F P
0 g
s
.45572 .59274 m .42073 .60364 L .42406 .64898 L p .682 .502 .703 r
F P
0 g
s
.38915 .6596 m .42406 .64898 L .42073 .60364 L p .682 .502 .703 r
F P
0 g
s
.21207 .67017 m .24596 .65963 L .24457 .62553 L p .677 .495 .699 r
F P
0 g
s
.43948 .39096 m .45724 .438 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.42011 .32877 m .41362 .31159 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.45724 .438 m .42011 .32877 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.46193 .33341 m .45655 .31584 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42011 .32877 m .45724 .438 L .47195 .48609 L p .851 .545 .1 r
F P
0 g
s
.42546 .34624 m .42011 .32877 L p .42133 .33248 L .851 .545 .1 r
F P
0 g
s
.47195 .48609 m .42546 .34624 L p .42133 .33248 L .851 .545 .1 r
F P
0 g
s
.41441 .31401 m .43948 .39096 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.41898 .3459 m .43948 .39096 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.40615 .29511 m .40856 .30463 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.40856 .30463 m .41898 .3459 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.46607 .35084 m .46193 .33341 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.41362 .31159 m .40615 .29511 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.43948 .39096 m .41753 .32357 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.41753 .32357 m .41362 .31159 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.47195 .48609 m .48332 .5343 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.42546 .34624 m .42644 .3492 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.42644 .3492 m .47195 .48609 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.45572 .59274 m .4911 .58172 L .48332 .5343 L p .699 .548 .748 r
F P
0 g
s
.43223 .38037 m .42953 .36359 L p .43058 .36731 L 1 .766 .313 r
F P
0 g
s
.4911 .58172 m .43223 .38037 L p .43058 .36731 L 1 .766 .313 r
F P
0 g
s
.4911 .58172 m .45572 .59274 L .45937 .63824 L p .682 .502 .703 r
F P
0 g
s
.45937 .63824 m .49506 .62738 L .4911 .58172 L p .682 .502 .703 r
F P
0 g
s
.43347 .39618 m .43223 .38037 L p .43316 .38403 L 1 .81 .413 r
F P
0 g
s
.49506 .62738 m .43347 .39618 L p .43316 .38403 L 1 .81 .413 r
F P
0 g
s
.49506 .62738 m .45937 .63824 L .45895 .68104 L p .66 .447 .648 r
F P
0 g
s
.45895 .68104 m .49504 .67031 L .49506 .62738 L p .66 .447 .648 r
F P
0 g
s
.43322 .41062 m .43347 .39618 L p .43428 .39977 L 1 .833 .481 r
F P
0 g
s
.49504 .67031 m .43322 .41062 L p .43428 .39977 L 1 .833 .481 r
F P
0 g
s
.49504 .67031 m .45895 .68104 L .45436 .72015 L p .63 .378 .578 r
F P
0 g
s
.45436 .72015 m .49089 .7095 L .49504 .67031 L p .63 .378 .578 r
F P
0 g
s
.43147 .4233 m .43322 .41062 L p .4339 .41412 L 1 .843 .528 r
F P
0 g
s
.49089 .7095 m .43147 .4233 L p .4339 .41412 L 1 .843 .528 r
F P
0 g
s
.49089 .7095 m .45436 .72015 L .44553 .75455 L p .585 .285 .481 r
F P
0 g
s
.44553 .75455 m .48255 .74393 L .49089 .7095 L p .585 .285 .481 r
F P
0 g
s
.42823 .43388 m .43147 .4233 L p .43201 .4267 L 1 .847 .562 r
F P
0 g
s
.48255 .74393 m .42823 .43388 L p .43201 .4267 L 1 .847 .562 r
F P
0 g
s
.42406 .64898 m .45937 .63824 L .45572 .59274 L p .682 .502 .703 r
F P
0 g
s
.45937 .63824 m .42406 .64898 L .42327 .69166 L p .66 .447 .648 r
F P
0 g
s
.42327 .69166 m .45895 .68104 L .45937 .63824 L p .66 .447 .648 r
F P
0 g
s
.45895 .68104 m .42327 .69166 L .41825 .73068 L p .63 .378 .578 r
F P
0 g
s
.41825 .73068 m .45436 .72015 L .45895 .68104 L p .63 .378 .578 r
F P
0 g
s
.48255 .74393 m .44553 .75455 L .43247 .78321 L p .515 .153 .339 r
F P
0 g
s
.43247 .78321 m .47003 .7726 L .48255 .74393 L p .515 .153 .339 r
F P
0 g
s
.42357 .44206 m .42823 .43388 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42406 .64898 m .38915 .6596 L .38799 .70216 L p .66 .447 .648 r
F P
0 g
s
.38799 .70216 m .42327 .69166 L .42406 .64898 L p .66 .447 .648 r
F P
0 g
s
.45436 .72015 m .41825 .73068 L .40893 .76504 L p .585 .285 .481 r
F P
0 g
s
.40893 .76504 m .44553 .75455 L .45436 .72015 L p .585 .285 .481 r
F P
0 g
s
.44553 .75455 m .40893 .76504 L .39535 .79371 L p .515 .153 .339 r
F P
0 g
s
.39535 .79371 m .43247 .78321 L .44553 .75455 L p .515 .153 .339 r
F P
0 g
s
.43247 .78321 m .39535 .79371 L .37763 .81571 L p .392 .1 .116 r
F P
0 g
s
.47003 .7726 m .43247 .78321 L .4153 .80516 L p .392 .1 .116 r
F P
0 g
s
.37763 .81571 m .4153 .80516 L .43247 .78321 L p .392 .1 .116 r
F P
0 g
s
.4153 .80516 m .45342 .79449 L .47003 .7726 L p .392 .1 .116 r
F P
0 g
s
.41042 .45021 m .41759 .44756 L p .41772 .45064 L 1 .842 .617 r
F P
0 g
s
.45342 .79449 m .4153 .80516 L .39424 .81946 L p .178 .1 .1 r
F P
0 g
s
.39424 .81946 m .43293 .80867 L .45342 .79449 L p .178 .1 .1 r
F P
0 g
s
.43293 .80867 m .41042 .45021 L p .41772 .45064 L 1 .842 .617 r
F P
0 g
s
.41042 .45021 m .40223 .44986 L p .40266 .44975 L .573 .989 1 r
F P
0 g
s
.41759 .44756 m .41042 .45021 L p .41182 .44973 L .1 .1 .1 r
F P
0 g
s
.47003 .7726 m .42401 .44518 L p .42938 .44318 L 1 .847 .586 r
F P
0 g
s
.48332 .5343 m .42953 .36359 L .42546 .34624 L p .98 .684 .169 r
F P
0 g
s
.42953 .36359 m .48332 .5343 L .4911 .58172 L p 1 .766 .313 r
F P
0 g
s
.43223 .38037 m .4911 .58172 L .49506 .62738 L p 1 .81 .413 r
F P
0 g
s
.43347 .39618 m .49506 .62738 L .49504 .67031 L p 1 .833 .481 r
F P
0 g
s
.43322 .41062 m .49504 .67031 L .49089 .7095 L p 1 .843 .528 r
F P
0 g
s
.43147 .4233 m .49089 .7095 L .48255 .74393 L p 1 .847 .562 r
F P
0 g
s
.48255 .74393 m .47003 .7726 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42823 .43388 m .42878 .43704 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42878 .43704 m .48255 .74393 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42938 .44318 m .42357 .44206 L .42864 .43718 L 1 .847 .586 r
F
.42401 .44518 m .42357 .44206 L p .42938 .44318 L F P
0 g
s
.47003 .7726 m .45342 .79449 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.42357 .44206 m .42486 .45121 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.42486 .45121 m .47003 .7726 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.40266 .44975 m .40357 .4495 L .41042 .45021 L .573 .989 1 r
F
.41759 .44756 m .42357 .44206 L p .42384 .44525 L 1 .845 .604 r
F P
0 g
s
.45342 .79449 m .41759 .44756 L p .42384 .44525 L 1 .845 .604 r
F P
0 g
s
.41759 .44756 m .45342 .79449 L .43293 .80867 L p 1 .842 .617 r
F P
0 g
s
.44823 .4375 m .4552 .43492 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.41042 .45021 m .41079 .45008 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.41079 .45008 m .44823 .4375 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.4552 .43492 m .46099 .42945 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.41759 .44756 m .41799 .44743 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.41799 .44743 m .4552 .43492 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.4552 .43492 m .44474 .43843 L p .44549 .43823 L .1 .1 .1 r
F P
0 g
s
.44474 .43843 m .41799 .44743 L p .44549 .43823 L .1 .1 .1 r
F P
0 g
s
.41182 .44973 m .44549 .43823 L .41799 .44743 L .1 .1 .1 r
F
.41799 .44743 m .41759 .44756 L p .41182 .44973 L F P
0 g
s
.42357 .44206 m .41759 .44756 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.46099 .42945 m .42357 .44206 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.44823 .4375 m .41079 .45008 L p .40357 .4495 L .573 .989 1 r
F P
0 g
s
.41079 .45008 m .41042 .45021 L p .40357 .4495 L .573 .989 1 r
F P
0 g
s
.42823 .43388 m .42357 .44206 L p .42406 .44182 L .42 .1 .21 r
F P
0 g
s
.46547 .42131 m .42823 .43388 L p .42406 .44182 L .42 .1 .21 r
F P
0 g
s
.43147 .4233 m .42823 .43388 L p .42875 .43359 L .535 .21 .414 r
F P
0 g
s
.46853 .41073 m .43147 .4233 L p .42875 .43359 L .535 .21 .414 r
F P
0 g
s
.43322 .41062 m .43147 .4233 L p .43203 .42294 L .601 .335 .543 r
F P
0 g
s
.47012 .39803 m .43322 .41062 L p .43203 .42294 L .601 .335 .543 r
F P
0 g
s
.43347 .39618 m .43322 .41062 L p .43384 .41017 L .642 .422 .63 r
F P
0 g
s
.47023 .38356 m .43347 .39618 L p .43384 .41017 L .642 .422 .63 r
F P
0 g
s
.43223 .38037 m .43347 .39618 L p .43415 .39564 L .67 .487 .693 r
F P
0 g
s
.46886 .3677 m .43223 .38037 L p .43415 .39564 L .67 .487 .693 r
F P
0 g
s
.42953 .36359 m .43223 .38037 L p .43299 .3797 L .691 .539 .743 r
F P
0 g
s
.43299 .3797 m .44387 .37021 L .42953 .36359 L .691 .539 .743 r
F
.46607 .35084 m .43839 .3605 L p .44387 .37021 L F P
0 g
s
.43839 .3605 m .42953 .36359 L p .44387 .37021 L .691 .539 .743 r
F P
0 g
s
.42953 .36359 m .43839 .3605 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.43839 .3605 m .46607 .35084 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.42546 .34624 m .42953 .36359 L p .43041 .36277 L .708 .583 .785 r
F P
0 g
s
.46193 .33341 m .42546 .34624 L p .43041 .36277 L .708 .583 .785 r
F P
0 g
s
.42546 .34624 m .42669 .34581 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42669 .34581 m .46193 .33341 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42011 .32877 m .42546 .34624 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.45655 .31584 m .42011 .32877 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.41362 .31159 m .41441 .31401 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.24596 .65963 m .21207 .67017 L .20978 .70204 L p .653 .438 .642 r
F P
0 g
s
.35309 .71254 m .31858 .72281 L .31231 .76157 L p .63 .378 .578 r
F P
0 g
s
.31858 .72281 m .24393 .69159 L .23846 .72064 L p .923 .654 .631 r
F P
0 g
s
.23846 .72064 m .31231 .76157 L .31858 .72281 L p .923 .654 .631 r
F P
0 g
s
.39612 .30371 m .3611 .31632 L .3841 .35818 L p .753 .73 .914 r
F P
0 g
s
.3841 .35818 m .41898 .3459 L .39612 .30371 L p .753 .73 .914 r
F P
0 g
s
.35309 .71254 m .38799 .70216 L .38915 .6596 L p .66 .447 .648 r
F P
0 g
s
.23965 .27108 m .20532 .28393 L .23272 .31625 L p .768 .809 .978 r
F P
0 g
s
.23272 .31625 m .26679 .30373 L .23965 .27108 L p .768 .809 .978 r
F P
0 g
s
.33607 .27826 m .30123 .29107 L .32647 .32879 L p .761 .767 .945 r
F P
0 g
s
.32647 .32879 m .3611 .31632 L .33607 .27826 L p .761 .767 .945 r
F P
0 g
s
.13192 .36836 m .09821 .38067 L .11976 .40411 L p .765 .813 .983 r
F P
0 g
s
.11976 .40411 m .15329 .39205 L .13192 .36836 L p .765 .813 .983 r
F P
0 g
s
.20978 .70204 m .24393 .69159 L .24596 .65963 L p .653 .438 .642 r
F P
0 g
s
.38799 .70216 m .35309 .71254 L .34723 .75139 L p .63 .378 .578 r
F P
0 g
s
.31231 .76157 m .34723 .75139 L .35309 .71254 L p .63 .378 .578 r
F P
0 g
s
.39787 .27976 m .39612 .30371 L .41898 .3459 L p .1 .253 .795 r
F P
0 g
s
.41898 .3459 m .40615 .29511 L .39787 .27976 L p .1 .253 .795 r
F P
0 g
s
.27436 .25808 m .23965 .27108 L .26679 .30373 L p .768 .809 .978 r
F P
0 g
s
.26679 .30373 m .30123 .29107 L .27436 .25808 L p .768 .809 .978 r
F P
0 g
s
.24393 .69159 m .20978 .70204 L .20402 .73103 L p .621 .364 .567 r
F P
0 g
s
.17137 .29664 m .13192 .36836 L .15329 .39205 L p .971 .833 .807 r
F P
0 g
s
.37129 .26532 m .33607 .27826 L .3611 .31632 L p .761 .767 .945 r
F P
0 g
s
.3611 .31632 m .39612 .30371 L .37129 .26532 L p .761 .767 .945 r
F P
0 g
s
.42327 .69166 m .38799 .70216 L .38254 .74109 L p .63 .378 .578 r
F P
0 g
s
.34723 .75139 m .38254 .74109 L .38799 .70216 L p .63 .378 .578 r
F P
0 g
s
.30947 .24493 m .27436 .25808 L .30123 .29107 L p .768 .809 .978 r
F P
0 g
s
.30123 .29107 m .33607 .27826 L .30947 .24493 L p .768 .809 .978 r
F P
0 g
s
.20402 .73103 m .23846 .72064 L .24393 .69159 L p .621 .364 .567 r
F P
0 g
s
.34723 .75139 m .31231 .76157 L .30161 .7958 L p .585 .285 .481 r
F P
0 g
s
.31231 .76157 m .23846 .72064 L .22958 .74601 L p .927 .637 .598 r
F P
0 g
s
.22958 .74601 m .30161 .7958 L .31231 .76157 L p .927 .637 .598 r
F P
0 g
s
.38898 .26592 m .37129 .26532 L .39612 .30371 L p .1 .373 .878 r
F P
0 g
s
.39612 .30371 m .39787 .27976 L .38898 .26592 L p .1 .373 .878 r
F P
0 g
s
.23846 .72064 m .20402 .73103 L .19481 .75637 L p .572 .265 .463 r
F P
0 g
s
.38254 .74109 m .41825 .73068 L .42327 .69166 L p .63 .378 .578 r
F P
0 g
s
.34497 .23164 m .30947 .24493 L .33607 .27826 L p .768 .809 .978 r
F P
0 g
s
.33607 .27826 m .37129 .26532 L .34497 .23164 L p .768 .809 .978 r
F P
0 g
s
.38254 .74109 m .34723 .75139 L .33698 .78566 L p .585 .285 .481 r
F P
0 g
s
.30161 .7958 m .33698 .78566 L .34723 .75139 L p .585 .285 .481 r
F P
0 g
s
.19481 .75637 m .22958 .74601 L .23846 .72064 L p .572 .265 .463 r
F P
0 g
s
.3797 .25395 m .34497 .23164 L .37129 .26532 L p .1 .423 .909 r
F P
0 g
s
.37129 .26532 m .38898 .26592 L .3797 .25395 L p .1 .423 .909 r
F P
0 g
s
.41825 .73068 m .38254 .74109 L .37275 .77541 L p .585 .285 .481 r
F P
0 g
s
.33698 .78566 m .37275 .77541 L .38254 .74109 L p .585 .285 .481 r
F P
0 g
s
.33698 .78566 m .30161 .7958 L .28654 .82447 L p .515 .153 .339 r
F P
0 g
s
.30161 .7958 m .22958 .74601 L .21739 .76695 L p .933 .621 .561 r
F P
0 g
s
.21739 .76695 m .28654 .82447 L .30161 .7958 L p .933 .621 .561 r
F P
0 g
s
.22958 .74601 m .19481 .75637 L .18227 .77732 L p .492 .12 .307 r
F P
0 g
s
.37275 .77541 m .40893 .76504 L .41825 .73068 L p .585 .285 .481 r
F P
0 g
s
.37275 .77541 m .33698 .78566 L .32239 .81433 L p .515 .153 .339 r
F P
0 g
s
.28654 .82447 m .32239 .81433 L .33698 .78566 L p .515 .153 .339 r
F P
0 g
s
.18227 .77732 m .21739 .76695 L .22958 .74601 L p .492 .12 .307 r
F P
0 g
s
.45005 .29855 m .41362 .31159 L .42011 .32877 L p .733 .659 .854 r
F P
0 g
s
.42011 .32877 m .45655 .31584 L .45005 .29855 L p .733 .659 .854 r
F P
0 g
s
.40893 .76504 m .37275 .77541 L .35866 .80408 L p .515 .153 .339 r
F P
0 g
s
.32239 .81433 m .35866 .80408 L .37275 .77541 L p .515 .153 .339 r
F P
0 g
s
.4426 .28196 m .40615 .29511 L .41362 .31159 L p .743 .695 .885 r
F P
0 g
s
.41362 .31159 m .45005 .29855 L .4426 .28196 L p .743 .695 .885 r
F P
0 g
s
.21739 .76695 m .18227 .77732 L .1666 .79319 L p .353 .1 .1 r
F P
0 g
s
.43437 .26649 m .39787 .27976 L .40615 .29511 L p .752 .731 .916 r
F P
0 g
s
.40615 .29511 m .4426 .28196 L .43437 .26649 L p .752 .731 .916 r
F P
0 g
s
.43223 .38037 m .46886 .3677 L .46607 .35084 L p .691 .539 .743 r
F P
0 g
s
.02 w
.08602 .49343 Mdot
.5 Mabswid
.32239 .81433 m .28654 .82447 L .26726 .84661 L p .392 .1 .116 r
F P
0 g
s
.28654 .82447 m .21739 .76695 L .20207 .78278 L p .94 .604 .518 r
F P
0 g
s
.20207 .78278 m .26726 .84661 L .28654 .82447 L p .94 .604 .518 r
F P
0 g
s
.42556 .25251 m .38898 .26592 L .39787 .27976 L p .76 .769 .948 r
F P
0 g
s
.39787 .27976 m .43437 .26649 L .42556 .25251 L p .76 .769 .948 r
F P
0 g
s
.4869 .28536 m .45005 .29855 L .45655 .31584 L p .733 .659 .854 r
F P
0 g
s
.45655 .31584 m .4934 .30277 L .4869 .28536 L p .733 .659 .854 r
F P
0 g
s
.4934 .30277 m .45655 .31584 L .46193 .33341 L p .721 .622 .821 r
F P
0 g
s
.46193 .33341 m .49883 .32044 L .4934 .30277 L p .721 .622 .821 r
F P
0 g
s
.49883 .32044 m .46193 .33341 L .46607 .35084 L p .708 .583 .785 r
F P
0 g
s
.35866 .80408 m .39535 .79371 L .40893 .76504 L p .515 .153 .339 r
F P
0 g
s
.47947 .26866 m .4426 .28196 L .45005 .29855 L p .743 .695 .885 r
F P
0 g
s
.45005 .29855 m .4869 .28536 L .47947 .26866 L p .743 .695 .885 r
F P
0 g
s
.43347 .39618 m .47023 .38356 L .46886 .3677 L p .67 .487 .693 r
F P
0 g
s
.1666 .79319 m .20207 .78278 L .21739 .76695 L p .353 .1 .1 r
F P
0 g
s
.46607 .35084 m .50304 .33795 L .49883 .32044 L p .708 .583 .785 r
F P
0 g
s
.50304 .33795 m .46607 .35084 L .46886 .3677 L p .691 .539 .743 r
F P
0 g
s
.4713 .25305 m .43437 .26649 L .4426 .28196 L p .752 .731 .916 r
F P
0 g
s
.4426 .28196 m .47947 .26866 L .4713 .25305 L p .752 .731 .916 r
F P
0 g
s
.41639 .24041 m .3797 .25395 L .38898 .26592 L p .766 .811 .981 r
F P
0 g
s
.38898 .26592 m .42556 .25251 L .41639 .24041 L p .766 .811 .981 r
F P
0 g
s
.02 w
.09289 .48216 Mdot
.5 Mabswid
.46886 .3677 m .50592 .35487 L .50304 .33795 L p .691 .539 .743 r
F P
0 g
s
.50592 .35487 m .46886 .3677 L .47023 .38356 L p .67 .487 .693 r
F P
0 g
s
.43322 .41062 m .47012 .39803 L .47023 .38356 L p .642 .422 .63 r
F P
0 g
s
.35866 .80408 m .32239 .81433 L .30362 .83643 L p .392 .1 .116 r
F P
0 g
s
.26726 .84661 m .30362 .83643 L .32239 .81433 L p .392 .1 .116 r
F P
0 g
s
.46257 .23895 m .42556 .25251 L .43437 .26649 L p .76 .769 .948 r
F P
0 g
s
.43437 .26649 m .4713 .25305 L .46257 .23895 L p .76 .769 .948 r
F P
0 g
s
.52418 .27202 m .4869 .28536 L .4934 .30277 L p .733 .659 .854 r
F P
0 g
s
.4934 .30277 m .53069 .28954 L .52418 .27202 L p .733 .659 .854 r
F P
0 g
s
.53069 .28954 m .4934 .30277 L .49883 .32044 L p .721 .622 .821 r
F P
0 g
s
.49883 .32044 m .53615 .30731 L .53069 .28954 L p .721 .622 .821 r
F P
0 g
s
.53615 .30731 m .49883 .32044 L .50304 .33795 L p .708 .583 .785 r
F P
0 g
s
.51678 .2552 m .47947 .26866 L .4869 .28536 L p .743 .695 .885 r
F P
0 g
s
.4869 .28536 m .52418 .27202 L .51678 .2552 L p .743 .695 .885 r
F P
0 g
s
.47023 .38356 m .50741 .37079 L .50592 .35487 L p .67 .487 .693 r
F P
0 g
s
.50741 .37079 m .47023 .38356 L .47012 .39803 L p .642 .422 .63 r
F P
0 g
s
.02 w
.10074 .46722 Mdot
.5 Mabswid
.50304 .33795 m .54043 .32491 L .53615 .30731 L p .708 .583 .785 r
F P
0 g
s
.54043 .32491 m .50304 .33795 L .50592 .35487 L p .691 .539 .743 r
F P
0 g
s
.20207 .78278 m .1666 .79319 L .14808 .80337 L p .114 .1 .1 r
F P
0 g
s
.02 w
.11252 .45146 Mdot
.5 Mabswid
.50865 .23947 m .4713 .25305 L .47947 .26866 L p .752 .731 .916 r
F P
0 g
s
.47947 .26866 m .51678 .2552 L .50865 .23947 L p .752 .731 .916 r
F P
0 g
s
.02 w
.12883 .44069 Mdot
.5 Mabswid
.4535 .22672 m .41639 .24041 L .42556 .25251 L p .766 .811 .981 r
F P
0 g
s
.42556 .25251 m .46257 .23895 L .4535 .22672 L p .766 .811 .981 r
F P
0 g
s
.50592 .35487 m .54342 .3419 L .54043 .32491 L p .691 .539 .743 r
F P
0 g
s
.54342 .3419 m .50592 .35487 L .50741 .37079 L p .67 .487 .693 r
F P
0 g
s
.43147 .4233 m .46853 .41073 L .47012 .39803 L p .601 .335 .543 r
F P
0 g
s
.02 w
.14661 .43422 Mdot
.5 Mabswid
.47012 .39803 m .50745 .38529 L .50741 .37079 L p .642 .422 .63 r
F P
0 g
s
.50745 .38529 m .47012 .39803 L .46853 .41073 L p .601 .335 .543 r
F P
0 g
s
.39535 .79371 m .35866 .80408 L .3404 .82613 L p .392 .1 .116 r
F P
0 g
s
.30362 .83643 m .3404 .82613 L .35866 .80408 L p .392 .1 .116 r
F P
0 g
s
.50001 .22523 m .46257 .23895 L .4713 .25305 L p .76 .769 .948 r
F P
0 g
s
.4713 .25305 m .50865 .23947 L .50001 .22523 L p .76 .769 .948 r
F P
0 g
s
.56189 .25853 m .52418 .27202 L .53069 .28954 L p .733 .659 .854 r
F P
0 g
s
.53069 .28954 m .56841 .27616 L .56189 .25853 L p .733 .659 .854 r
F P
0 g
s
.56841 .27616 m .53069 .28954 L .53615 .30731 L p .721 .622 .821 r
F P
0 g
s
.53615 .30731 m .57391 .29403 L .56841 .27616 L p .721 .622 .821 r
F P
0 g
s
.57391 .29403 m .53615 .30731 L .54043 .32491 L p .708 .583 .785 r
F P
0 g
s
.55451 .24158 m .51678 .2552 L .52418 .27202 L p .743 .695 .885 r
F P
0 g
s
.52418 .27202 m .56189 .25853 L .55451 .24158 L p .743 .695 .885 r
F P
0 g
s
.50741 .37079 m .54503 .35787 L .54342 .3419 L p .67 .487 .693 r
F P
0 g
s
.54503 .35787 m .50741 .37079 L .50745 .38529 L p .642 .422 .63 r
F P
0 g
s
.02 w
.16485 .42775 Mdot
.5 Mabswid
.54043 .32491 m .57826 .31171 L .57391 .29403 L p .708 .583 .785 r
F P
0 g
s
.57826 .31171 m .54043 .32491 L .54342 .3419 L p .691 .539 .743 r
F P
0 g
s
.26726 .84661 m .20207 .78278 L .18392 .79288 L p .948 .588 .469 r
F P
0 g
s
.14808 .80337 m .18392 .79288 L p .19802 .78433 L .114 .1 .1 r
F P
0 g
s
.18392 .79288 m .14808 .80337 L .12712 .80735 L p .39 .887 .881 r
F P
0 g
s
.16332 .79674 m .18392 .79288 L p .17966 .79397 L .39 .887 .881 r
F P
0 g
s
.12712 .80735 m .16332 .79674 L p .17966 .79397 L .39 .887 .881 r
F P
0 g
s
.16332 .79674 m .12712 .80735 L .10422 .80476 L p .605 .985 1 r
F P
0 g
s
.14075 .794 m .16332 .79674 L p .15888 .79735 L .605 .985 1 r
F P
0 g
s
.10422 .80476 m .14075 .794 L p .15888 .79735 L .605 .985 1 r
F P
0 g
s
.11682 .78448 m .14075 .794 L p .13618 .79411 L .699 .958 1 r
F P
0 g
s
.07998 .79543 m .11682 .78448 L p .13618 .79411 L .699 .958 1 r
F P
0 g
s
.05509 .77939 m .0922 .76821 L .11682 .78448 L p .733 .899 1 r
F P
0 g
s
.11682 .78448 m .0922 .76821 L p .09699 .77461 L .968 .523 .206 r
F P
0 g
s
.14075 .794 m .10422 .80476 L .07998 .79543 L p .699 .958 1 r
F P
0 g
s
.11682 .78448 m .07998 .79543 L .05509 .77939 L p .733 .899 1 r
F P
0 g
s
.15598 .85356 m .11682 .78448 L p .09699 .77461 L .968 .523 .206 r
F P
0 g
s
.14075 .794 m .11682 .78448 L p .12217 .79057 L .969 .538 .279 r
F P
0 g
s
.18775 .8653 m .14075 .794 L p .12217 .79057 L .969 .538 .279 r
F P
0 g
s
.16332 .79674 m .14075 .794 L p .14656 .79959 L .964 .554 .349 r
F P
0 g
s
.21736 .86775 m .16332 .79674 L p .14656 .79959 L .964 .554 .349 r
F P
0 g
s
.18392 .79288 m .16332 .79674 L p .16946 .80166 L .957 .571 .413 r
F P
0 g
s
.24406 .86131 m .18392 .79288 L p .16946 .80166 L .957 .571 .413 r
F P
0 g
s
.18392 .79288 m .20207 .78278 L p .19802 .78433 L .114 .1 .1 r
F P
0 g
s
.54645 .22572 m .50865 .23947 L .51678 .2552 L p .752 .731 .916 r
F P
0 g
s
.51678 .2552 m .55451 .24158 L .54645 .22572 L p .752 .731 .916 r
F P
0 g
s
.42823 .43388 m .46547 .42131 L .46853 .41073 L p .535 .21 .414 r
F P
0 g
s
.49106 .21287 m .4535 .22672 L .46257 .23895 L p .766 .811 .981 r
F P
0 g
s
.46257 .23895 m .50001 .22523 L .49106 .21287 L p .766 .811 .981 r
F P
0 g
s
.54342 .3419 m .58135 .32878 L .57826 .31171 L p .691 .539 .743 r
F P
0 g
s
.58135 .32878 m .54342 .3419 L .54503 .35787 L p .67 .487 .693 r
F P
0 g
s
.46853 .41073 m .50602 .39801 L .50745 .38529 L p .601 .335 .543 r
F P
0 g
s
.50602 .39801 m .46853 .41073 L .46547 .42131 L p .535 .21 .414 r
F P
0 g
s
.30362 .83643 m .26726 .84661 L .24406 .86131 L p .178 .1 .1 r
F P
0 g
s
.18392 .79288 m .24406 .86131 L .26726 .84661 L p .948 .588 .469 r
F P
0 g
s
.50745 .38529 m .54522 .37241 L .54503 .35787 L p .642 .422 .63 r
F P
0 g
s
.54522 .37241 m .50745 .38529 L .50602 .39801 L p .601 .335 .543 r
F P
0 g
s
.02 w
.18468 .41973 Mdot
.5 Mabswid
.3404 .82613 m .37763 .81571 L .39535 .79371 L p .392 .1 .116 r
F P
0 g
s
.53789 .21135 m .50001 .22523 L .50865 .23947 L p .76 .769 .948 r
F P
0 g
s
.50865 .23947 m .54645 .22572 L .53789 .21135 L p .76 .769 .948 r
F P
0 g
s
.60005 .24488 m .56189 .25853 L .56841 .27616 L p .733 .659 .854 r
F P
0 g
s
.56841 .27616 m .60657 .26262 L .60005 .24488 L p .733 .659 .854 r
F P
0 g
s
.60657 .26262 m .56841 .27616 L .57391 .29403 L p .721 .622 .821 r
F P
0 g
s
.57391 .29403 m .61211 .28059 L .60657 .26262 L p .721 .622 .821 r
F P
0 g
s
.61211 .28059 m .57391 .29403 L .57826 .31171 L p .708 .583 .785 r
F P
0 g
s
.59269 .2278 m .55451 .24158 L .56189 .25853 L p .743 .695 .885 r
F P
0 g
s
.56189 .25853 m .60005 .24488 L .59269 .2278 L p .743 .695 .885 r
F P
0 g
s
.54503 .35787 m .58309 .3448 L .58135 .32878 L p .67 .487 .693 r
F P
0 g
s
.58309 .3448 m .54503 .35787 L .54522 .37241 L p .642 .422 .63 r
F P
0 g
s
.57826 .31171 m .61653 .29836 L .61211 .28059 L p .708 .583 .785 r
F P
0 g
s
.61653 .29836 m .57826 .31171 L .58135 .32878 L p .691 .539 .743 r
F P
0 g
s
.58469 .21181 m .54645 .22572 L .55451 .24158 L p .752 .731 .916 r
F P
0 g
s
.55451 .24158 m .59269 .2278 L .58469 .21181 L p .752 .731 .916 r
F P
0 g
s
.42357 .44206 m .46099 .42945 L .46547 .42131 L p .42 .1 .21 r
F P
0 g
s
.46547 .42131 m .50314 .40858 L .50602 .39801 L p .535 .21 .414 r
F P
0 g
s
.50314 .40858 m .46547 .42131 L .46099 .42945 L p .42 .1 .21 r
F P
0 g
s
.52905 .19885 m .49106 .21287 L .50001 .22523 L p .766 .811 .981 r
F P
0 g
s
.50001 .22523 m .53789 .21135 L .52905 .19885 L p .766 .811 .981 r
F P
0 g
s
.58135 .32878 m .61972 .3155 L .61653 .29836 L p .691 .539 .743 r
F P
0 g
s
.61972 .3155 m .58135 .32878 L .58309 .3448 L p .67 .487 .693 r
F P
0 g
s
.50602 .39801 m .54396 .38514 L .54522 .37241 L p .601 .335 .543 r
F P
0 g
s
.54396 .38514 m .50602 .39801 L .50314 .40858 L p .535 .21 .414 r
F P
0 g
s
.02 w
.20744 .40968 Mdot
.5 Mabswid
.3404 .82613 m .30362 .83643 L .28093 .85103 L p .178 .1 .1 r
F P
0 g
s
.24406 .86131 m .28093 .85103 L .30362 .83643 L p .178 .1 .1 r
F P
0 g
s
.54522 .37241 m .58343 .35937 L .58309 .3448 L p .642 .422 .63 r
F P
0 g
s
.58343 .35937 m .54522 .37241 L .54396 .38514 L p .601 .335 .543 r
F P
0 g
s
.57622 .1973 m .53789 .21135 L .54645 .22572 L p .76 .769 .948 r
F P
0 g
s
.54645 .22572 m .58469 .21181 L .57622 .1973 L p .76 .769 .948 r
F P
0 g
s
.63865 .23106 m .60005 .24488 L .60657 .26262 L p .733 .659 .854 r
F P
0 g
s
.60657 .26262 m .64518 .24892 L .63865 .23106 L p .733 .659 .854 r
F P
0 g
s
.64518 .24892 m .60657 .26262 L .61211 .28059 L p .721 .622 .821 r
F P
0 g
s
.61211 .28059 m .65077 .267 L .64518 .24892 L p .721 .622 .821 r
F P
0 g
s
.65077 .267 m .61211 .28059 L .61653 .29836 L p .708 .583 .785 r
F P
0 g
s
.63133 .21386 m .59269 .2278 L .60005 .24488 L p .743 .695 .885 r
F P
0 g
s
.60005 .24488 m .63865 .23106 L .63133 .21386 L p .743 .695 .885 r
F P
0 g
s
.58309 .3448 m .6216 .33157 L .61972 .3155 L p .67 .487 .693 r
F P
0 g
s
.6216 .33157 m .58309 .3448 L .58343 .35937 L p .642 .422 .63 r
F P
0 g
s
.61653 .29836 m .65526 .28486 L .65077 .267 L p .708 .583 .785 r
F P
0 g
s
.65526 .28486 m .61653 .29836 L .61972 .3155 L p .691 .539 .743 r
F P
0 g
s
.62338 .19773 m .58469 .21181 L .59269 .2278 L p .752 .731 .916 r
F P
0 g
s
.59269 .2278 m .63133 .21386 L .62338 .19773 L p .752 .731 .916 r
F P
0 g
s
.46099 .42945 m .49886 .4167 L .50314 .40858 L p .42 .1 .21 r
F P
0 g
s
.49886 .4167 m .46099 .42945 L .4552 .43492 L p .212 .1 .1 r
F P
0 g
s
.50314 .40858 m .54127 .3957 L .54396 .38514 L p .535 .21 .414 r
F P
0 g
s
.54127 .3957 m .50314 .40858 L .49886 .4167 L p .42 .1 .21 r
F P
0 g
s
.5675 .18467 m .52905 .19885 L .53789 .21135 L p .766 .811 .981 r
F P
0 g
s
.53789 .21135 m .57622 .1973 L .5675 .18467 L p .766 .811 .981 r
F P
0 g
s
.61972 .3155 m .65856 .30206 L .65526 .28486 L p .691 .539 .743 r
F P
0 g
s
.65856 .30206 m .61972 .3155 L .6216 .33157 L p .67 .487 .693 r
F P
0 g
s
.54396 .38514 m .58235 .37211 L .58343 .35937 L p .601 .335 .543 r
F P
0 g
s
.58235 .37211 m .54396 .38514 L .54127 .3957 L p .535 .21 .414 r
F P
0 g
s
.02 w
.23392 .3961 Mdot
.5 Mabswid
.37763 .81571 m .3404 .82613 L .31825 .84064 L p .178 .1 .1 r
F P
0 g
s
.28093 .85103 m .31825 .84064 L .3404 .82613 L p .178 .1 .1 r
F P
0 g
s
.58343 .35937 m .6221 .34618 L .6216 .33157 L p .642 .422 .63 r
F P
0 g
s
.6221 .34618 m .58343 .35937 L .58235 .37211 L p .601 .335 .543 r
F P
0 g
s
.61501 .18309 m .57622 .1973 L .58469 .21181 L p .76 .769 .948 r
F P
0 g
s
.58469 .21181 m .62338 .19773 L .61501 .18309 L p .76 .769 .948 r
F P
0 g
s
.67772 .21708 m .63865 .23106 L .64518 .24892 L p .733 .659 .854 r
F P
0 g
s
.64518 .24892 m .68426 .23506 L .67772 .21708 L p .733 .659 .854 r
F P
0 g
s
.68426 .23506 m .64518 .24892 L .65077 .267 L p .721 .622 .821 r
F P
0 g
s
.65077 .267 m .68988 .25324 L .68426 .23506 L p .721 .622 .821 r
F P
0 g
s
.68988 .25324 m .65077 .267 L .65526 .28486 L p .708 .583 .785 r
F P
0 g
s
.67042 .19975 m .63133 .21386 L .63865 .23106 L p .743 .695 .885 r
F P
0 g
s
.63865 .23106 m .67772 .21708 L .67042 .19975 L p .743 .695 .885 r
F P
0 g
s
.6216 .33157 m .66057 .31819 L .65856 .30206 L p .67 .487 .693 r
F P
0 g
s
.66057 .31819 m .6216 .33157 L .6221 .34618 L p .642 .422 .63 r
F P
0 g
s
.65526 .28486 m .69445 .27119 L .68988 .25324 L p .708 .583 .785 r
F P
0 g
s
.69445 .27119 m .65526 .28486 L .65856 .30206 L p .691 .539 .743 r
F P
0 g
s
.66254 .18349 m .62338 .19773 L .63133 .21386 L p .752 .731 .916 r
F P
0 g
s
.63133 .21386 m .67042 .19975 L .66254 .18349 L p .752 .731 .916 r
F P
0 g
s
.49886 .4167 m .53718 .4038 L .54127 .3957 L p .42 .1 .21 r
F P
0 g
s
.53718 .4038 m .49886 .4167 L .49327 .42212 L p .212 .1 .1 r
F P
0 g
s
.4552 .43492 m .49327 .42212 L .49886 .4167 L p .212 .1 .1 r
F P
0 g
s
.54127 .3957 m .57984 .38267 L .58235 .37211 L p .535 .21 .414 r
F P
0 g
s
.57984 .38267 m .54127 .3957 L .53718 .4038 L p .42 .1 .21 r
F P
0 g
s
.60641 .17031 m .5675 .18467 L .57622 .1973 L p .766 .811 .981 r
F P
0 g
s
.57622 .1973 m .61501 .18309 L .60641 .17031 L p .766 .811 .981 r
F P
0 g
s
.49327 .42212 m .4552 .43492 L .44823 .4375 L p .1 .1 .1 r
F P
0 g
s
.4153 .80516 m .37763 .81571 L .35601 .83011 L p .178 .1 .1 r
F P
0 g
s
.35601 .83011 m .39424 .81946 L .4153 .80516 L p .178 .1 .1 r
F P
0 g
s
.41042 .45021 m .43293 .80867 L .40893 .81434 L p 1 .838 .627 r
F P
0 g
s
.40893 .81434 m .40223 .44986 L .41042 .45021 L p 1 .838 .627 r
F P
0 g
s
.65856 .30206 m .69785 .28847 L .69445 .27119 L p .691 .539 .743 r
F P
0 g
s
.69785 .28847 m .65856 .30206 L .66057 .31819 L p .67 .487 .693 r
F P
0 g
s
.58235 .37211 m .62119 .35893 L .6221 .34618 L p .601 .335 .543 r
F P
0 g
s
.62119 .35893 m .58235 .37211 L .57984 .38267 L p .535 .21 .414 r
F P
0 g
s
.31825 .84064 m .35601 .83011 L .37763 .81571 L p .178 .1 .1 r
F P
0 g
s
.6221 .34618 m .66122 .33283 L .66057 .31819 L p .642 .422 .63 r
F P
0 g
s
.66122 .33283 m .6221 .34618 L .62119 .35893 L p .601 .335 .543 r
F P
0 g
s
.16332 .79674 m .21736 .86775 L .24406 .86131 L p .957 .571 .413 r
F P
0 g
s
.28093 .85103 m .24406 .86131 L .21736 .86775 L p .322 .837 .818 r
F P
0 g
s
.65426 .16871 m .61501 .18309 L .62338 .19773 L p .76 .769 .948 r
F P
0 g
s
.62338 .19773 m .66254 .18349 L .65426 .16871 L p .76 .769 .948 r
F P
0 g
s
.70451 .21523 m .67772 .21708 L .68426 .23506 L p .247 .358 .822 r
F P
0 g
s
.68426 .23506 m .70796 .22473 L .70451 .21523 L p .247 .358 .822 r
F P
0 g
s
.70796 .22473 m .68426 .23506 L .68988 .25324 L p .228 .307 .778 r
F P
0 g
s
.68988 .25324 m .71093 .23434 L .70796 .22473 L p .228 .307 .778 r
F P
0 g
s
.71093 .23434 m .68988 .25324 L .69445 .27119 L p .19 .242 .723 r
F P
0 g
s
.70069 .20608 m .67042 .19975 L .67772 .21708 L p .248 .396 .859 r
F P
0 g
s
.67772 .21708 m .70451 .21523 L .70069 .20608 L p .248 .396 .859 r
F P
0 g
s
.66057 .31819 m .7 .30465 L .69785 .28847 L p .67 .487 .693 r
F P
0 g
s
.7 .30465 m .66057 .31819 L .66122 .33283 L p .642 .422 .63 r
F P
0 g
s
.69445 .27119 m .71336 .24381 L .71093 .23434 L p .19 .242 .723 r
F P
0 g
s
.71336 .24381 m .69445 .27119 L .69785 .28847 L p .129 .157 .654 r
F P
0 g
s
.02 w
.26519 .37146 Mdot
.5 Mabswid
.69658 .19752 m .66254 .18349 L .67042 .19975 L p .23 .423 .89 r
F P
0 g
s
.67042 .19975 m .70069 .20608 L .69658 .19752 L p .23 .423 .89 r
F P
0 g
s
.53718 .4038 m .57596 .39074 L .57984 .38267 L p .42 .1 .21 r
F P
0 g
s
.57596 .39074 m .53718 .4038 L .53179 .40917 L p .212 .1 .1 r
F P
0 g
s
.49327 .42212 m .53179 .40917 L .53718 .4038 L p .212 .1 .1 r
F P
0 g
s
.57984 .38267 m .61888 .36948 L .62119 .35893 L p .535 .21 .414 r
F P
0 g
s
.61888 .36948 m .57984 .38267 L .57596 .39074 L p .42 .1 .21 r
F P
0 g
s
.64579 .15579 m .60641 .17031 L .61501 .18309 L p .766 .811 .981 r
F P
0 g
s
.61501 .18309 m .65426 .16871 L .64579 .15579 L p .766 .811 .981 r
F P
0 g
s
.53179 .40917 m .49327 .42212 L .48649 .42463 L p .1 .1 .1 r
F P
0 g
s
.44823 .4375 m .48649 .42463 L .49327 .42212 L p .1 .1 .1 r
F P
0 g
s
.48649 .42463 m .44823 .4375 L .44023 .43707 L p .573 .989 1 r
F P
0 g
s
.40223 .44986 m .44023 .43707 L .44823 .4375 L p .573 .989 1 r
F P
0 g
s
.69785 .28847 m .71518 .25289 L .71336 .24381 L p .129 .157 .654 r
F P
0 g
s
.71518 .25289 m .69785 .28847 L .7 .30465 L p .1 .1 .562 r
F P
0 g
s
.62119 .35893 m .6605 .3456 L .66122 .33283 L p .601 .335 .543 r
F P
0 g
s
.6605 .3456 m .62119 .35893 L .61888 .36948 L p .535 .21 .414 r
F P
0 g
s
.44023 .43707 m .40223 .44986 L .39325 .44648 L p .701 .989 1 r
F P
0 g
s
.40223 .44986 m .40893 .81434 L .3819 .81083 L p 1 .833 .635 r
F P
0 g
s
.3819 .81083 m .39325 .44648 L .40223 .44986 L p 1 .833 .635 r
F P
0 g
s
.66122 .33283 m .70082 .31932 L .7 .30465 L p .642 .422 .63 r
F P
0 g
s
.70082 .31932 m .66122 .33283 L .6605 .3456 L p .601 .335 .543 r
F P
0 g
s
.31825 .84064 m .28093 .85103 L .25474 .85733 L p .322 .837 .818 r
F P
0 g
s
.21736 .86775 m .25474 .85733 L .28093 .85103 L p .322 .837 .818 r
F P
0 g
s
.69229 .18976 m .65426 .16871 L .66254 .18349 L p .189 .437 .911 r
F P
0 g
s
.66254 .18349 m .69658 .19752 L .69229 .18976 L p .189 .437 .911 r
F P
0 g
s
.7 .30465 m .71634 .26136 L .71518 .25289 L p .1 .1 .562 r
F P
0 g
s
.71634 .26136 m .7 .30465 L .70082 .31932 L p .1 .1 .436 r
F P
0 g
s
.57596 .39074 m .61521 .37752 L .61888 .36948 L p .42 .1 .21 r
F P
0 g
s
.61521 .37752 m .57596 .39074 L .57078 .39606 L p .212 .1 .1 r
F P
0 g
s
.53179 .40917 m .57078 .39606 L .57596 .39074 L p .212 .1 .1 r
F P
0 g
s
.61888 .36948 m .65839 .35614 L .6605 .3456 L p .535 .21 .414 r
F P
0 g
s
.65839 .35614 m .61888 .36948 L .61521 .37752 L p .42 .1 .21 r
F P
0 g
s
.68793 .18303 m .64579 .15579 L .65426 .16871 L p .118 .434 .917 r
F P
0 g
s
.65426 .16871 m .69229 .18976 L .68793 .18303 L p .118 .434 .917 r
F P
0 g
s
.57078 .39606 m .53179 .40917 L .52522 .41161 L p .1 .1 .1 r
F P
0 g
s
.48649 .42463 m .52522 .41161 L .53179 .40917 L p .1 .1 .1 r
F P
0 g
s
.02 w
.31147 .32721 Mdot
.5 Mabswid
.52522 .41161 m .48649 .42463 L .47869 .42412 L p .573 .989 1 r
F P
0 g
s
.44023 .43707 m .47869 .42412 L .48649 .42463 L p .573 .989 1 r
F P
0 g
s
.6605 .3456 m .70028 .3321 L .70082 .31932 L p .601 .335 .543 r
F P
0 g
s
.70028 .3321 m .6605 .3456 L .65839 .35614 L p .535 .21 .414 r
F P
0 g
s
.47869 .42412 m .44023 .43707 L .43143 .43359 L p .701 .989 1 r
F P
0 g
s
.39325 .44648 m .43143 .43359 L .44023 .43707 L p .701 .989 1 r
F P
0 g
s
.43143 .43359 m .39325 .44648 L .3837 .44011 L p .745 .934 1 r
F P
0 g
s
.39325 .44648 m .3819 .81083 L .35251 .79774 L p 1 .828 .64 r
F P
0 g
s
.35251 .79774 m .3837 .44011 L .39325 .44648 L p 1 .828 .64 r
F P
0 g
s
.70082 .31932 m .7168 .26898 L .71634 .26136 L p .1 .1 .436 r
F P
0 g
s
.7168 .26898 m .70082 .31932 L .70028 .3321 L p .1 .1 .262 r
F P
0 g
s
.35601 .83011 m .31825 .84064 L .29258 .84678 L p .322 .837 .818 r
F P
0 g
s
.25474 .85733 m .29258 .84678 L .31825 .84064 L p .322 .837 .818 r
F P
0 g
s
.02 w
.39591 .28068 Mdot
.5 Mabswid
.61521 .37752 m .65493 .36414 L .65839 .35614 L p .42 .1 .21 r
F P
0 g
s
.65493 .36414 m .61521 .37752 L .61024 .38279 L p .212 .1 .1 r
F P
0 g
s
.57078 .39606 m .61024 .38279 L .61521 .37752 L p .212 .1 .1 r
F P
0 g
s
.65839 .35614 m .69838 .34263 L .70028 .3321 L p .535 .21 .414 r
F P
0 g
s
.69838 .34263 m .65839 .35614 L .65493 .36414 L p .42 .1 .21 r
F P
0 g
s
.61024 .38279 m .57078 .39606 L .56442 .39844 L p .1 .1 .1 r
F P
0 g
s
.52522 .41161 m .56442 .39844 L .57078 .39606 L p .1 .1 .1 r
F P
0 g
s
.56442 .39844 m .52522 .41161 L .51762 .41102 L p .573 .989 1 r
F P
0 g
s
.47869 .42412 m .51762 .41102 L .52522 .41161 L p .573 .989 1 r
F P
0 g
s
.3837 .44011 m .42204 .42711 L .43143 .43359 L p .745 .934 1 r
F P
0 g
s
.42204 .42711 m .3837 .44011 L .37386 .43088 L p .757 .875 1 r
F P
0 g
s
.3837 .44011 m .35251 .79774 L .32158 .775 L p 1 .823 .644 r
F P
0 g
s
.32158 .775 m .37386 .43088 L .3837 .44011 L p 1 .823 .644 r
F P
0 g
s
.70028 .3321 m .71656 .27556 L .7168 .26898 L p .1 .1 .262 r
F P
0 g
s
.71656 .27556 m .70028 .3321 L .69838 .34263 L p .1 .1 .1 r
F P
0 g
s
.51762 .41102 m .47869 .42412 L .47007 .42054 L p .701 .989 1 r
F P
0 g
s
.43143 .43359 m .47007 .42054 L .47869 .42412 L p .701 .989 1 r
F P
0 g
s
.47007 .42054 m .43143 .43359 L .42204 .42711 L p .745 .934 1 r
F P
0 g
s
.14075 .794 m .18775 .8653 L .21736 .86775 L p .964 .554 .349 r
F P
0 g
s
.25474 .85733 m .21736 .86775 L .18775 .8653 L p .564 .975 1 r
F P
0 g
s
.39424 .81946 m .35601 .83011 L .33089 .8361 L p .322 .837 .818 r
F P
0 g
s
.29258 .84678 m .33089 .8361 L .35601 .83011 L p .322 .837 .818 r
F P
0 g
s
.37386 .43088 m .41234 .41775 L .42204 .42711 L p .757 .875 1 r
F P
0 g
s
.65493 .36414 m .69514 .3506 L .69838 .34263 L p .42 .1 .21 r
F P
0 g
s
.69514 .3506 m .65493 .36414 L .65018 .36936 L p .212 .1 .1 r
F P
0 g
s
.61024 .38279 m .65018 .36936 L .65493 .36414 L p .212 .1 .1 r
F P
0 g
s
.69838 .34263 m .71561 .28091 L p .716 .27763 L .1 .1 .1 r
F P
0 g
s
.71561 .28091 m .69838 .34263 L .69514 .3506 L p .871 .898 .409 r
F P
0 g
s
.71398 .28488 m .71561 .28091 L p .71498 .28305 L .871 .898 .409 r
F P
0 g
s
.69514 .3506 m .71398 .28488 L p .71498 .28305 L .871 .898 .409 r
F P
0 g
s
.65018 .36936 m .69061 .35577 L .69514 .3506 L p .212 .1 .1 r
F P
0 g
s
.71169 .28734 m .71398 .28488 L p .71326 .28705 L 1 1 .616 r
F P
0 g
s
.69061 .35577 m .71169 .28734 L p .71326 .28705 L 1 1 .616 r
F P
0 g
s
.71169 .28734 m .7088 .28823 L p .7101 .28764 L .472 .927 .828 r
F P
0 g
s
.71398 .28488 m .71169 .28734 L p .71296 .28671 L .1 .1 .1 r
F P
0 g
s
.71398 .28488 m .69514 .3506 L .69061 .35577 L p 1 1 .616 r
F P
0 g
s
.75361 .26659 m .71398 .28488 L p .71296 .28671 L .1 .1 .1 r
F P
0 g
s
.7516 .26885 m .75361 .26659 L p .72224 .28212 L .1 .1 .1 r
F P
0 g
s
.71169 .28734 m .7516 .26885 L p .72224 .28212 L .1 .1 .1 r
F P
0 g
s
.79398 .24855 m .75361 .26659 L .7516 .26885 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .79398 .24855 L p .76153 .26409 L .1 .1 .1 r
F P
0 g
s
.7516 .26885 m .7922 .25062 L p .76153 .26409 L .1 .1 .1 r
F P
0 g
s
.83505 .23069 m .79398 .24855 L .7922 .25062 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .83349 .23258 L .83505 .23069 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .78991 .25129 L p .82322 .23699 L .447 .912 .822 r
F P
0 g
s
.83349 .23258 m .7922 .25062 L p .82322 .23699 L .447 .912 .822 r
F P
0 g
s
.7516 .26885 m .74903 .26962 L p .78128 .25542 L .458 .919 .825 r
F P
0 g
s
.7922 .25062 m .7516 .26885 L p .78128 .25542 L .458 .919 .825 r
F P
0 g
s
.7516 .26885 m .71169 .28734 L p .7101 .28764 L .472 .927 .828 r
F P
0 g
s
.71561 .28091 m .71398 .28488 L p .71522 .28421 L .313 .1 .146 r
F P
0 g
s
.75503 .26294 m .71561 .28091 L p .71522 .28421 L .313 .1 .146 r
F P
0 g
s
.71561 .28091 m .71656 .27556 L p .716 .27763 L .1 .1 .1 r
F P
0 g
s
.65018 .36936 m .61024 .38279 L .60409 .3851 L p .1 .1 .1 r
F P
0 g
s
.56442 .39844 m .60409 .3851 L .61024 .38279 L p .1 .1 .1 r
F P
0 g
s
.60409 .3851 m .56442 .39844 L .55702 .39775 L p .573 .989 1 r
F P
0 g
s
.51762 .41102 m .55702 .39775 L .56442 .39844 L p .573 .989 1 r
F P
0 g
s
.42204 .42711 m .46085 .41394 L .47007 .42054 L p .745 .934 1 r
F P
0 g
s
.46085 .41394 m .42204 .42711 L .41234 .41775 L p .757 .875 1 r
F P
0 g
s
.55702 .39775 m .51762 .41102 L .50918 .40734 L p .701 .989 1 r
F P
0 g
s
.47007 .42054 m .50918 .40734 L .51762 .41102 L p .701 .989 1 r
F P
0 g
s
.50918 .40734 m .47007 .42054 L .46085 .41394 L p .745 .934 1 r
F P
0 g
s
.02 w
.54011 .2423 Mdot
.5 Mabswid
.29258 .84678 m .25474 .85733 L .22562 .85468 L p .564 .975 1 r
F P
0 g
s
.18775 .8653 m .22562 .85468 L .25474 .85733 L p .564 .975 1 r
F P
0 g
s
.43293 .80867 m .39424 .81946 L .36967 .82528 L p .322 .837 .818 r
F P
0 g
s
.33089 .8361 m .36967 .82528 L .39424 .81946 L p .322 .837 .818 r
F P
0 g
s
.74425 .20166 m .70451 .21523 L .70796 .22473 L p .707 .645 .859 r
F P
0 g
s
.70796 .22473 m .74746 .21056 L .74425 .20166 L p .707 .645 .859 r
F P
0 g
s
.74746 .21056 m .70796 .22473 L .71093 .23434 L p .695 .606 .824 r
F P
0 g
s
.71093 .23434 m .75024 .21955 L .74746 .21056 L p .695 .606 .824 r
F P
0 g
s
.75024 .21955 m .71093 .23434 L .71336 .24381 L p .68 .565 .787 r
F P
0 g
s
.74071 .19309 m .70069 .20608 L .70451 .21523 L p .717 .682 .892 r
F P
0 g
s
.70451 .21523 m .74425 .20166 L .74071 .19309 L p .717 .682 .892 r
F P
0 g
s
.71336 .24381 m .75254 .2284 L .75024 .21955 L p .68 .565 .787 r
F P
0 g
s
.75254 .2284 m .71336 .24381 L .71518 .25289 L p .661 .517 .743 r
F P
0 g
s
.73692 .18506 m .69658 .19752 L .70069 .20608 L p .725 .719 .924 r
F P
0 g
s
.70069 .20608 m .74071 .19309 L .73692 .18506 L p .725 .719 .924 r
F P
0 g
s
.71518 .25289 m .75428 .23688 L .75254 .2284 L p .661 .517 .743 r
F P
0 g
s
.75428 .23688 m .71518 .25289 L .71634 .26136 L p .635 .46 .69 r
F P
0 g
s
.41234 .41775 m .4513 .40446 L .46085 .41394 L p .757 .875 1 r
F P
0 g
s
.69061 .35577 m .65018 .36936 L .64426 .3716 L p .1 .1 .1 r
F P
0 g
s
.60409 .3851 m .64426 .3716 L .65018 .36936 L p .1 .1 .1 r
F P
0 g
s
.73298 .17778 m .69229 .18976 L .69658 .19752 L p .731 .758 .958 r
F P
0 g
s
.69658 .19752 m .73692 .18506 L .73298 .17778 L p .731 .758 .958 r
F P
0 g
s
.64426 .3716 m .60409 .3851 L .59691 .38432 L p .573 .989 1 r
F P
0 g
s
.55702 .39775 m .59691 .38432 L .60409 .3851 L p .573 .989 1 r
F P
0 g
s
.46085 .41394 m .50014 .40062 L .50918 .40734 L p .745 .934 1 r
F P
0 g
s
.50014 .40062 m .46085 .41394 L .4513 .40446 L p .757 .875 1 r
F P
0 g
s
.59691 .38432 m .55702 .39775 L .54878 .39397 L p .701 .989 1 r
F P
0 g
s
.50918 .40734 m .54878 .39397 L .55702 .39775 L p .701 .989 1 r
F P
0 g
s
.54878 .39397 m .50918 .40734 L .50014 .40062 L p .745 .934 1 r
F P
0 g
s
.71634 .26136 m .75543 .24478 L .75428 .23688 L p .635 .46 .69 r
F P
0 g
s
.75543 .24478 m .71634 .26136 L .7168 .26898 L p .6 .387 .621 r
F P
0 g
s
.33089 .8361 m .29258 .84678 L .26396 .84392 L p .564 .975 1 r
F P
0 g
s
.22562 .85468 m .26396 .84392 L .29258 .84678 L p .564 .975 1 r
F P
0 g
s
.36967 .82528 m .40893 .81434 L .43293 .80867 L p .322 .837 .818 r
F P
0 g
s
.72899 .17146 m .68793 .18303 L .69229 .18976 L p .734 .801 .994 r
F P
0 g
s
.69229 .18976 m .73298 .17778 L .72899 .17146 L p .734 .801 .994 r
F P
0 g
s
.7168 .26898 m .75595 .25187 L .75543 .24478 L p .6 .387 .621 r
F P
0 g
s
.75595 .25187 m .7168 .26898 L .71656 .27556 L p .547 .287 .525 r
F P
0 g
s
.11682 .78448 m .15598 .85356 L .18775 .8653 L p .969 .538 .279 r
F P
0 g
s
.7847 .18777 m .74425 .20166 L .74746 .21056 L p .71 .647 .858 r
F P
0 g
s
.74746 .21056 m .78769 .19614 L .7847 .18777 L p .71 .647 .858 r
F P
0 g
s
.78769 .19614 m .74746 .21056 L .75024 .21955 L p .698 .608 .824 r
F P
0 g
s
.75024 .21955 m .79032 .20458 L .78769 .19614 L p .698 .608 .824 r
F P
0 g
s
.79032 .20458 m .75024 .21955 L .75254 .2284 L p .683 .567 .786 r
F P
0 g
s
.7814 .17971 m .74071 .19309 L .74425 .20166 L p .72 .683 .891 r
F P
0 g
s
.74425 .20166 m .7847 .18777 L .7814 .17971 L p .72 .683 .891 r
F P
0 g
s
.75254 .2284 m .7925 .21289 L .79032 .20458 L p .683 .567 .786 r
F P
0 g
s
.7925 .21289 m .75254 .2284 L .75428 .23688 L p .664 .519 .743 r
F P
0 g
s
.71656 .27556 m .75581 .25798 L .75595 .25187 L p .547 .287 .525 r
F P
0 g
s
.75581 .25798 m .71656 .27556 L .71561 .28091 L p .462 .142 .381 r
F P
0 g
s
.4513 .40446 m .49073 .391 L .50014 .40062 L p .757 .875 1 r
F P
0 g
s
.7779 .17215 m .73692 .18506 L .74071 .19309 L p .728 .721 .924 r
F P
0 g
s
.74071 .19309 m .7814 .17971 L .7779 .17215 L p .728 .721 .924 r
F P
0 g
s
.64426 .3716 m .68491 .35793 L .69061 .35577 L p .1 .1 .1 r
F P
0 g
s
.71169 .28734 m .69061 .35577 L .68491 .35793 L p 1 1 .749 r
F P
0 g
s
.68491 .35793 m .64426 .3716 L .63729 .37073 L p .573 .989 1 r
F P
0 g
s
.59691 .38432 m .63729 .37073 L .64426 .3716 L p .573 .989 1 r
F P
0 g
s
.50014 .40062 m .53991 .38713 L .54878 .39397 L p .745 .934 1 r
F P
0 g
s
.53991 .38713 m .50014 .40062 L .49073 .391 L p .757 .875 1 r
F P
0 g
s
.75428 .23688 m .79418 .22084 L .7925 .21289 L p .664 .519 .743 r
F P
0 g
s
.79418 .22084 m .75428 .23688 L .75543 .24478 L p .639 .463 .69 r
F P
0 g
s
.63729 .37073 m .59691 .38432 L .58886 .38043 L p .701 .989 1 r
F P
0 g
s
.54878 .39397 m .58886 .38043 L .59691 .38432 L p .701 .989 1 r
F P
0 g
s
.58886 .38043 m .54878 .39397 L .53991 .38713 L p .745 .934 1 r
F P
0 g
s
.22562 .85468 m .18775 .8653 L .15598 .85356 L p .68 .963 1 r
F P
0 g
s
.77427 .16529 m .73298 .17778 L .73692 .18506 L p .734 .76 .957 r
F P
0 g
s
.73692 .18506 m .7779 .17215 L .77427 .16529 L p .734 .76 .957 r
F P
0 g
s
.36967 .82528 m .33089 .8361 L .30278 .83303 L p .564 .975 1 r
F P
0 g
s
.26396 .84392 m .30278 .83303 L .33089 .8361 L p .564 .975 1 r
F P
0 g
s
.75543 .24478 m .79531 .22823 L .79418 .22084 L p .639 .463 .69 r
F P
0 g
s
.79531 .22823 m .75543 .24478 L .75595 .25187 L p .605 .39 .622 r
F P
0 g
s
.71561 .28091 m .75503 .26294 L .75581 .25798 L p .462 .142 .381 r
F P
0 g
s
.77061 .15933 m .72899 .17146 L .73298 .17778 L p .738 .802 .993 r
F P
0 g
s
.73298 .17778 m .77427 .16529 L .77061 .15933 L p .738 .802 .993 r
F P
0 g
s
.75595 .25187 m .79587 .23486 L .79531 .22823 L p .605 .39 .622 r
F P
0 g
s
.79587 .23486 m .75595 .25187 L .75581 .25798 L p .553 .292 .526 r
F P
0 g
s
.71398 .28488 m .75361 .26659 L .75503 .26294 L p .313 .1 .146 r
F P
0 g
s
.49073 .391 m .53066 .37738 L .53991 .38713 L p .757 .875 1 r
F P
0 g
s
.82583 .17359 m .7847 .18777 L .78769 .19614 L p .712 .648 .858 r
F P
0 g
s
.78769 .19614 m .82864 .18148 L .82583 .17359 L p .712 .648 .858 r
F P
0 g
s
.82864 .18148 m .78769 .19614 L .79032 .20458 L p .7 .61 .824 r
F P
0 g
s
.79032 .20458 m .83112 .18944 L .82864 .18148 L p .7 .61 .824 r
F P
0 g
s
.83112 .18944 m .79032 .20458 L .7925 .21289 L p .686 .568 .786 r
F P
0 g
s
.75581 .25798 m .79583 .24056 L .79587 .23486 L p .553 .292 .526 r
F P
0 g
s
.79583 .24056 m .75581 .25798 L .75503 .26294 L p .469 .147 .383 r
F P
0 g
s
.82276 .16597 m .7814 .17971 L .7847 .18777 L p .722 .685 .891 r
F P
0 g
s
.7847 .18777 m .82583 .17359 L .82276 .16597 L p .722 .685 .891 r
F P
0 g
s
.68491 .35793 m .7088 .28823 L .71169 .28734 L p 1 1 .749 r
F P
0 g
s
.7088 .28823 m .68491 .35793 L .67817 .35697 L p 1 1 .82 r
F P
0 g
s
.63729 .37073 m .67817 .35697 L .68491 .35793 L p .573 .989 1 r
F P
0 g
s
.53991 .38713 m .58018 .37347 L .58886 .38043 L p .745 .934 1 r
F P
0 g
s
.58018 .37347 m .53991 .38713 L .53066 .37738 L p .757 .875 1 r
F P
0 g
s
.7925 .21289 m .8332 .19726 L .83112 .18944 L p .686 .568 .786 r
F P
0 g
s
.8332 .19726 m .7925 .21289 L .79418 .22084 L p .667 .521 .743 r
F P
0 g
s
.67817 .35697 m .63729 .37073 L .62945 .36673 L p .701 .989 1 r
F P
0 g
s
.58886 .38043 m .62945 .36673 L .63729 .37073 L p .701 .989 1 r
F P
0 g
s
.62945 .36673 m .58886 .38043 L .58018 .37347 L p .745 .934 1 r
F P
0 g
s
.81951 .15883 m .7779 .17215 L .7814 .17971 L p .731 .722 .923 r
F P
0 g
s
.7814 .17971 m .82276 .16597 L .81951 .15883 L p .731 .722 .923 r
F P
0 g
s
.26396 .84392 m .22562 .85468 L .1943 .84268 L p .68 .963 1 r
F P
0 g
s
.15598 .85356 m .1943 .84268 L .22562 .85468 L p .68 .963 1 r
F P
0 g
s
.79418 .22084 m .83482 .20474 L .8332 .19726 L p .667 .521 .743 r
F P
0 g
s
.83482 .20474 m .79418 .22084 L .79531 .22823 L p .642 .465 .69 r
F P
0 g
s
.40893 .81434 m .36967 .82528 L .34209 .822 L p .564 .975 1 r
F P
0 g
s
.30278 .83303 m .34209 .822 L .36967 .82528 L p .564 .975 1 r
F P
0 g
s
.81615 .15235 m .77427 .16529 L .7779 .17215 L p .737 .761 .957 r
F P
0 g
s
.7779 .17215 m .81951 .15883 L .81615 .15235 L p .737 .761 .957 r
F P
0 g
s
.0922 .76821 m .12295 .83241 L .15598 .85356 L p .968 .523 .206 r
F P
0 g
s
.75503 .26294 m .7952 .24517 L .79583 .24056 L p .469 .147 .383 r
F P
0 g
s
.7952 .24517 m .75503 .26294 L .75361 .26659 L p .323 .1 .15 r
F P
0 g
s
.79531 .22823 m .83595 .21168 L .83482 .20474 L p .642 .465 .69 r
F P
0 g
s
.83595 .21168 m .79531 .22823 L .79587 .23486 L p .608 .393 .622 r
F P
0 g
s
.81279 .1467 m .77061 .15933 L .77427 .16529 L p .741 .804 .992 r
F P
0 g
s
.77427 .16529 m .81615 .15235 L .81279 .1467 L p .741 .804 .992 r
F P
0 g
s
.79587 .23486 m .83655 .21791 L .83595 .21168 L p .608 .393 .622 r
F P
0 g
s
.83655 .21791 m .79587 .23486 L .79583 .24056 L p .558 .295 .527 r
F P
0 g
s
.53066 .37738 m .57109 .36359 L .58018 .37347 L p .757 .875 1 r
F P
0 g
s
.75361 .26659 m .79398 .24855 L .7952 .24517 L p .323 .1 .15 r
F P
0 g
s
.7088 .28823 m .74903 .26962 L .7516 .26885 L p .472 .927 .828 r
F P
0 g
s
.74903 .26962 m .7088 .28823 L .70539 .2875 L p .689 1 1 r
F P
0 g
s
.67817 .35697 m .70539 .2875 L .7088 .28823 L p 1 1 .82 r
F P
0 g
s
.58018 .37347 m .62095 .35964 L .62945 .36673 L p .745 .934 1 r
F P
0 g
s
.62095 .35964 m .58018 .37347 L .57109 .36359 L p .757 .875 1 r
F P
0 g
s
.70539 .2875 m .67817 .35697 L .67054 .35285 L p 1 .996 .851 r
F P
0 g
s
.62945 .36673 m .67054 .35285 L .67817 .35697 L p .701 .989 1 r
F P
0 g
s
.67054 .35285 m .62945 .36673 L .62095 .35964 L p .745 .934 1 r
F P
0 g
s
.79583 .24056 m .8366 .22324 L .83655 .21791 L p .558 .295 .527 r
F P
0 g
s
.8366 .22324 m .79583 .24056 L .7952 .24517 L p .475 .152 .384 r
F P
0 g
s
.30278 .83303 m .26396 .84392 L .2331 .83166 L p .68 .963 1 r
F P
0 g
s
.1943 .84268 m .2331 .83166 L .26396 .84392 L p .68 .963 1 r
F P
0 g
s
.34209 .822 m .3819 .81083 L .40893 .81434 L p .564 .975 1 r
F P
0 g
s
.70539 .2875 m .74597 .26888 L .74903 .26962 L p .689 1 1 r
F P
0 g
s
.74597 .26888 m .70539 .2875 L .70155 .28515 L p .772 1 1 r
F P
0 g
s
.67054 .35285 m .70155 .28515 L .70539 .2875 L p 1 .996 .851 r
F P
0 g
s
.7952 .24517 m .8361 .22754 L .8366 .22324 L p .475 .152 .384 r
F P
0 g
s
.8361 .22754 m .7952 .24517 L .79398 .24855 L p .331 .1 .153 r
F P
0 g
s
.1943 .84268 m .15598 .85356 L .12295 .83241 L p .723 .906 1 r
F P
0 g
s
.57109 .36359 m .61202 .34962 L .62095 .35964 L p .757 .875 1 r
F P
0 g
s
.74903 .26962 m .78991 .25129 L .7922 .25062 L p .458 .919 .825 r
F P
0 g
s
.78991 .25129 m .74903 .26962 L .74597 .26888 L p .68 1 1 r
F P
0 g
s
.79398 .24855 m .83505 .23069 L .8361 .22754 L p .331 .1 .153 r
F P
0 g
s
.70155 .28515 m .7425 .26663 L .74597 .26888 L p .772 1 1 r
F P
0 g
s
.7425 .26663 m .70155 .28515 L .69738 .28124 L p .794 .943 1 r
F P
0 g
s
.70155 .28515 m .67054 .35285 L .66223 .34564 L p 1 .956 .86 r
F P
0 g
s
.66223 .34564 m .69738 .28124 L .70155 .28515 L p 1 .956 .86 r
F P
0 g
s
.62095 .35964 m .66223 .34564 L .67054 .35285 L p .745 .934 1 r
F P
0 g
s
.66223 .34564 m .62095 .35964 L .61202 .34962 L p .757 .875 1 r
F P
0 g
s
.34209 .822 m .30278 .83303 L .2724 .8205 L p .68 .963 1 r
F P
0 g
s
.2331 .83166 m .2724 .8205 L .30278 .83303 L p .68 .963 1 r
F P
0 g
s
.74597 .26888 m .78716 .25054 L .78991 .25129 L p .68 1 1 r
F P
0 g
s
.78716 .25054 m .74597 .26888 L .7425 .26663 L p .765 1 1 r
F P
0 g
s
.69738 .28124 m .73872 .26291 L .7425 .26663 L p .794 .943 1 r
F P
0 g
s
.73872 .26291 m .69738 .28124 L .69299 .27585 L p .795 .883 1 r
F P
0 g
s
.69738 .28124 m .66223 .34564 L .65348 .33548 L p .999 .916 .857 r
F P
0 g
s
.65348 .33548 m .69299 .27585 L .69738 .28124 L p .999 .916 .857 r
F P
0 g
s
.2331 .83166 m .1943 .84268 L .16166 .82122 L p .723 .906 1 r
F P
0 g
s
.12295 .83241 m .16166 .82122 L .1943 .84268 L p .723 .906 1 r
F P
0 g
s
.7425 .26663 m .78403 .24837 L .78716 .25054 L p .765 1 1 r
F P
0 g
s
.78403 .24837 m .7425 .26663 L .73872 .26291 L p .789 .941 1 r
F P
0 g
s
.61202 .34962 m .65348 .33548 L .66223 .34564 L p .757 .875 1 r
F P
0 g
s
.78991 .25129 m .83145 .23316 L .83349 .23258 L p .447 .912 .822 r
F P
0 g
s
.83145 .23316 m .78991 .25129 L .78716 .25054 L p .672 1 1 r
F P
0 g
s
.69299 .27585 m .73473 .25784 L .73872 .26291 L p .795 .883 1 r
F P
0 g
s
.3819 .81083 m .34209 .822 L .3122 .80919 L p .68 .963 1 r
F P
0 g
s
.2724 .8205 m .3122 .80919 L .34209 .822 L p .68 .963 1 r
F P
0 g
s
.73872 .26291 m .7806 .24484 L .78403 .24837 L p .789 .941 1 r
F P
0 g
s
.7806 .24484 m .73872 .26291 L .73473 .25784 L p .791 .882 1 r
F P
0 g
s
.78716 .25054 m .82898 .2324 L .83145 .23316 L p .672 1 1 r
F P
0 g
s
.82898 .2324 m .78716 .25054 L .78403 .24837 L p .76 1 1 r
F P
0 g
s
.73473 .25784 m .77696 .24004 L .7806 .24484 L p .791 .882 1 r
F P
0 g
s
.78403 .24837 m .82614 .23032 L .82898 .2324 L p .76 1 1 r
F P
0 g
s
.82614 .23032 m .78403 .24837 L .7806 .24484 L p .785 .94 1 r
F P
0 g
s
.2724 .8205 m .2331 .83166 L .20087 .80989 L p .723 .906 1 r
F P
0 g
s
.16166 .82122 m .20087 .80989 L .2331 .83166 L p .723 .906 1 r
F P
0 g
s
.3122 .80919 m .35251 .79774 L .3819 .81083 L p .68 .963 1 r
F P
0 g
s
.7806 .24484 m .82302 .22695 L .82614 .23032 L p .785 .94 1 r
F P
0 g
s
.82302 .22695 m .7806 .24484 L .77696 .24004 L p .788 .881 1 r
F P
0 g
s
.77696 .24004 m .81969 .22239 L .82302 .22695 L p .788 .881 1 r
F P
0 g
s
.3122 .80919 m .2724 .8205 L .24059 .79841 L p .723 .906 1 r
F P
0 g
s
.20087 .80989 m .24059 .79841 L .2724 .8205 L p .723 .906 1 r
F P
0 g
s
.35251 .79774 m .3122 .80919 L .28082 .78678 L p .723 .906 1 r
F P
0 g
s
.24059 .79841 m .28082 .78678 L .3122 .80919 L p .723 .906 1 r
F P
0 g
s
.28082 .78678 m .32158 .775 L .35251 .79774 L p .723 .906 1 r
F P
0 g
s
.25 Mabswid
.77416 0 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.77416 0 m
.0677 .27203 L
s
0 .83649 m
.23968 1.01205 L
s
.04533 .86969 m
.04603 .86318 L
s
[(0)] .04392 .88272 .10814 -1 Mshowa
.09585 .9067 m
.09647 .90018 L
s
[(0.5)] .09461 .91974 .09522 -1 Mshowa
.14424 .94215 m
.14478 .93562 L
s
[(1)] .14316 .9552 .08305 -1 Mshowa
.19063 .97612 m
.1911 .96959 L
s
[(1.5)] .18969 .98919 .07157 -1 Mshowa
.23513 1.00872 m
.23553 1.00218 L
s
[(2)] .23434 1.0218 .06072 -1 Mshowa
.125 Mabswid
.05561 .87722 m
.05602 .87332 L
s
.0658 .88469 m
.0662 .88078 L
s
.07591 .89209 m
.0763 .88818 L
s
.08592 .89943 m
.0863 .89552 L
s
.1057 .91391 m
.10606 .91 L
s
.11546 .92106 m
.11581 .91715 L
s
.12513 .92815 m
.12548 .92424 L
s
.13473 .93518 m
.13506 .93126 L
s
.15368 .94906 m
.15399 .94514 L
s
.16303 .95591 m
.16334 .95199 L
s
.17231 .9627 m
.1726 .95878 L
s
.18151 .96944 m
.18179 .96552 L
s
.19968 .98275 m
.19995 .97883 L
s
.20865 .98932 m
.20891 .9854 L
s
.21755 .99584 m
.2178 .99192 L
s
.22638 1.00231 m
.22662 .99838 L
s
.03495 .86209 m
.03538 .85819 L
s
.02448 .85443 m
.02493 .85052 L
s
.01392 .84669 m
.01438 .84279 L
s
.00327 .83889 m
.00373 .83499 L
s
gsave
.11933 1.00721 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.931 .25158 m
1 .81858 L
s
.93242 .26326 m
.92622 .26536 L
s
[(-1)] .94483 .25907 -1 .33784 Mshowa
.94717 .38447 m
.94093 .38646 L
s
[(0)] .95965 .38048 -1 .3195 Mshowa
.96296 .51418 m
.95668 .51606 L
s
[(1)] .9755 .51042 -1 .29996 Mshowa
.97989 .65333 m
.97358 .65509 L
s
[(2)] .99251 .64981 -1 .2791 Mshowa
.9981 .80299 m
.99176 .80462 L
s
[(3)] 1.01079 .79973 -1 .25678 Mshowa
.125 Mabswid
.93529 .28686 m
.93157 .28811 L
s
.9382 .31078 m
.93447 .31201 L
s
.94115 .33501 m
.93742 .33623 L
s
.94414 .35957 m
.9404 .36078 L
s
.95024 .4097 m
.94649 .41089 L
s
.95335 .43528 m
.9496 .43645 L
s
.95651 .46122 m
.95276 .46238 L
s
.95971 .48752 m
.95595 .48866 L
s
.96625 .54123 m
.96248 .54234 L
s
.96959 .56865 m
.96581 .56975 L
s
.97297 .59648 m
.96919 .59756 L
s
.97641 .6247 m
.97262 .62577 L
s
.98343 .68239 m
.97964 .68343 L
s
.98701 .71187 m
.98322 .7129 L
s
.99066 .74179 m
.98686 .7428 L
s
.99435 .77216 m
.99055 .77316 L
s
gsave
1.03824 .4916 -61 -13.7497 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{284.562, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004L000180P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<ol07@000?l07@000?l07@000?l07@000?l0
7@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l0
7@000;T00P5Q0000]`020@0300410600002d00<10`020F00002a00<11P000`400@1O0000[`020@T0
00<10040G`000:`00`4;00040@000Eh0002Z00813`001040005M0000Y`030A4000@10001G@000:D0
0P4D00050@000040G0000:800`4F00050@000040G00009l00`4I00030@00008000<10000F@0009d0
0P4L00030@0000<000<10000F00009X00`4N00030@0000<000<10000F00009P00P4Q00030@0000@0
00<10000E`0009D00`4S00030@0000@000<10000E`0009<00P4V00030@0000D000<10000EP000900
0`4X00030@0000H000<10000E@0008d00`4[00030@0000H000<10000E@0008/00P4_00030@0000H0
00<10000E00008P00`4a00030@0000L000<10000D`0008H00P4d00030@0000L000<10000D`0008<0
0`4f00030@0000P000<10000DP0008400P4i00030@0000P000<10000DP0007h00`4k00030@0000T0
00<10000D@0007/00`4n00030@0000X000<10000D00007T00P5100030@0000X000<10000D00007H0
0`5300030@0000/000<10000C`0007@00P5600030@0000/000<10000C`0007400`5800030@0000`0
00<10000CP0006l00P5;00030@0000d000<10000C@0006`00`5>00030@0000`000<10000C@0006T0
0`5A00030@0000d000<10000C00006L00P5D00030@0000d000<10000C00006@00`5F00030@0000h0
00<10000B`0006800P5I00030@0000l000<10000BP0005l00`5K00030@0000l000<10000BP0005`0
0`5N00030@0000800P4<00030@0004T0001J0081H@040@:;0P4;00030@0004T0001G00<1@0020Al0
0`400i010@020@03R`4000/000<10000B00005D00P4o00@10X/20AX00`4018/10@44R`813@000`40
00170000DP030Ch00`40202;08/0R`4m0P4D00<100>;T0400P400h/1R`040@03R`4100`000<10000
A`0005000P4n00<12H/20@03=e@101000`400h/10@020@03R`6;00<100>;RX/00P400hF;0@0<0003
0@0004H0001=00<1?P020@0808/1R`2;08/40@05T8/10E@00P4:00@100B;0@4118/30@:;1043R`81
30000`4000160000BP030Cl00P44R`030H^;00<11X/01@5D0@4g00810`050@:;0`44R`<100B;RP6:
0P4018^5RhD20@03RhD100`000<10000A@0004P00P50008100H0R`2;08/30@0;Ri2;T8^@Ri2;T040
0U@20@07E0410I2;T0040@:;00<1R`400P44R`<100>;0H/00`400hZ;RP020@d000<10000A00004D0
0`4o00<11X/20@:;00<1Rh/01h/30@03E3MD00<100B;0@411H/30@04RhZ;RP<100B5R`6;0P4018F;
QH/20@03K04000`000<10000A00004<00P4o00<100N;0040R`2;008100Z@Ri2;0H^@Ri2;1042R`@1
0U@20@B;1042R`030H/1008118/40@03RX/1008100B;Q@413P000`4000130000@0030Ch00`45R`03
0H^;00811h/00`6;0@020@0:RX^:RhZ;0E@gE08100<g0H/00`401HZ;RX^:00<100>5RhD010400h^5
0@030@03Rfb;008100=/Q@403P000`4000120000?P020Ch00`4028/0R`2;08/00`40292;T8^@Ri2;
0`49R`<10e@30@B;10400h/1R`030@05QH^5RhD00`400hF;0@020@03RhF;00813P000`4000120000
>`030Ch00P42R`030H^;00>;0`42R`030H^;00>;0`402XZ;0H^:RhZ;RX/20@03Q@5>00@100=D0H/0
0`4018^5RhD30@04QH^5R`@100=/0@400P4018E/QF`20@03Q@4000h000<10000@@0003P00`4o0081
00L0R`2;0H/000<100R@Ri2;T06@R`<11H/00`6;R`02R`<10h/01P5DCUA>E0<118/30@04QH/1R`<1
00B;QH^50`400hF;Q@030@:50P4?00030@000440000f0081@0020@J;0`48R`<100^:RhZ;RX^:RhX1
RP020@0@RhF;QH^5R`5DCUA>E4h1Q@<100B5RhF;0`4018]/Rf`30@04K8D1Q@8100=/QF`0104@0003
0@000400000c00<1?`030@0608/0R`2;0P402I2;0H^@Ri2;T0020@:;00<1Rh/01X/30@V;00@1CUA>
1@4018F;QH/30@04RhD1Q@<100B5RhF;0`42Q@D100=VQ@4040000`40000o0000<@020Cl00`42R`03
0H^;00:;0P44R`030H^;00:;0P402X^:RhX1RX^:RhX30@0;QH/1RhF;QH^5RhD01@401TiCCU<1R`<1
00B;K8]/0`400fb5K0040@03Q@65008100IV0@65IXD20A0000<10000?`0002h00`4n00<100P0R`2;
0040R`8100V@Ri2;T8^@0I000P47R`030H/100811H/00`6;R`03R`<100R:R`5DCUA>E0<100F;QH^5
R`020@03Q@6500<118D30@04Q@41IP<100=V0@404@000`40000n0000;0020Ch00`48R`<128/30@08
RhZ;RX^:RhX30@0;QH^5RhF;QH^50HD00`402HF;QH^50DiCCP020@05CP6;K8/00`400hE/0@020@04
K8E/Q@<100>50@4010400h/10@020A8000<10000?P0002T00`4U00815P030@0808/1R`2;08/30@08
Ri01T8^@Ri030@:;00<1Rh/018/20@:;00<1Rh/01X/30@09RX^:RhZ;RX^:00<100E>E4h1Q@020@03
RhF;00810XD00`650@020@05QF`10HD00P401H/10@2@00<14`000`40000m00009P030BD00`42R`81
4@030@F;00<1Rh/00`45R`030H^;00<100Z;RX^:R`6;RX^:0P402X^5RhD1QH^5RhD30@0;QH/1RhF;
QH^5RhD00P401DhgCSM>008100=/QF`00P400hE/Q@030@03QFH100P100>;0@400P4F00030@0003`0
000T00819P020@0608/0R`2;0`4<00<100P0R`2;08/0R`<100R;T8^@Ri2;T0<118/30@:;00@1R`41
1h/00`6;0@020@0;RX^:RhX1RX^:RhX00`402XD1CUA>E05>0H/20@>50P400hD1Q@050@03004100H1
60000`40000l00008@030BD00`49R`040CL10@L00`42R`030H^;00>;0`42R`030H^;00>;0`401X^:
0HZ;RP@100b;0@41RhF;QH^5RhD30@0;QH^5RhF;QH^50HD00`401hF:QHZ50CL00P401Thg0@65K0<1
00=VQ@4010401I2;0@6@00D16`000`40000k00007`020BD00`4048/00@2;08/0R`2;08/1?E@20@80
0`40202;08/00@2;0`4028^@Ri2;0H^@0`45R`030H^;00:;1042R`030H^;00B;0P402hX1RX^:RhZ;
RX^:00<100R5RhF;QH^5R`810dh20@06Q@41K8E/1@401H/10@2@00D17@000`40000900@1;P0001`0
0`4T00<11X/00`6;R`07R`050Cdg?CL00P48R`<128/30@0;RhZ;RX^:RhZ;0H/010402HF;QH/1RhF;
Q@020@0;RhF;0H^5RhF;QH/00P402XD1QHZ5RXF:QHX20@04=dhgCP<100>5IXD010401h^@0@6;T040
0Y030Ah00P4<00030@0002d0000J008190030@0OR`2;08/0R`2;0H/0R`2;08/0R`01E3eD?@6;08/0
R`030@09Ri01T8^@Ri2;00810X/00`6;R`06R`<12H/03P6;0@6:RhZ;RX/1RhZ;0`402X^5R`6;QH^5
RhD20@03R`5>009>00D1CS81K0040@0508/10@000P4019000@4200030@0001T00`400`01000;0003
0@0002d0000G00<190020@:;00<1Rh/028/00`6;R`08R`050Cdg?@400X/30@F;00<1Rh/00X/20@0:
RX^:R`6;RX^:R`<100^;0H^5RhF;QH^5R`030@0:RhF;QH^5RhD1Q@8100^:QHZ5RXF:0HZ5RP020@>5
00L1=dh1CP4b00<100>@R`400P4018/10I030@@000<100005P030@@000<100000P040@@000<10000
;@0001@00`4U008101X0R`2;0040R`2;08/0R`2;0H/0R`2;08/0R`L100^;T8^@Ri2;T8/1R`020@N;
00<1R`400P45R`030H^;00:;10402X/1RhZ;RX^:RhX30@0;QH^5RhF;QH^50HD00P401X^5RhF;Q@81
00=><`400P401Y00T00100D11`000`40000C00<11P030@/000<10000;@0001800P4V008128/00`6;
R`08R`030H^;00>;0`42R`050Cdg?@4028/30@08RX^:RhZ;RX/30@0:RhF;QH^5RhD1Q@<100^5RhF;
QH/1RhF;Q@030@0;QHX1RXF:QHZ5RXD00`48Q@8100Hb0@6@Ri020@:@104900030@0001000`4;0003
0@0000L00`4_000040020BH00P40602;08/0R`2;08/00@2;08/0R`2;0040R`<100n;T8^@R`4gE3L1
Ri2;T8/00`42R`030H^;00>;0`42R`030H^;00F;0`402hZ;RX^:RhZ;RP6:00<100^5RhF;Q@65RhF;
Q@030@0:Q@65RhF;QH^5R`@100>@00400P400`010@030@`000<1000030030@h000<100002@000`40
000]000040020B<00`42R`030H^;00R;00<1Rh/01h/30@V;00<1=cd01P4028Z;RX^:0HZ;0`402X^5
RhF;0H^5RhD30@0;QH/1RhF;QH^5RhD00`402hF:QHZ5RXF:Q@6500<11HD00`65Q@02Q@<100B50I2;
0P400i010@040@h000<100002@030A4000<10000>@00010000<1004080020@0GR`2;08/00@2;08/0
R`2;08/1R`2;08/00`403H^@0I2;T8^@Ri2;T8/01@48R`<12H/03`6;0@6:RhZ;RP6:RhZ;RP030@0;
Q@65RhF;QH^5RhD00`402XF;QH^5RhF;0H/30@05RXF:0@000P400i010@040A0000<100001`020A@0
00<10000>@00010000@1000170030@V;00<1Rh/028/00`6;R`020@F;00<1Rh/028/00`4m0@020@05
RhZ;RX/00`402h^50HF;QH^5RhF;00<100Z;QH^5RhF;Q@650`402XF:QHZ50HF:QHX40@03Q@6500N5
0`4028E/QFb5K06Q1`4B00030@0000@00`4F00030@0003T0000@00050@00004060030@0G08/1R`2;
08/0R`2;0040R`2;08/0R`000`405I2;T8^@Ri2;0H^@Ri2;T8^@Ri01E0020@:;0`45R`030H^;00>;
0`402X/1RhZ;RX^:RhX30@0:QH^5RhF;QH/1R`<100^;QH^5RhD1QH^5R`030@0<RP6:QHZ5RXD1Y@6U
104D00040@0000<160020C/0000?00030@0000<000<100004`020@J;00<1Rh/028/00`6;R`03R`<1
0X/00`6;R`07R`030H^;00R;0P403SL10H^5RhF;QH^5R`6;0`402X^5RhF;0H^5RhD30@0:Q@65RXF:
QHZ5RP<12HD00`650@020@0=QFb5K8D1QFb5K06Q00020@04Y@410A@0104L00030@0003T0000?0003
0@0000@000<1000040020@0GR`2;08/0R`2;0H/0R`2;08/0R`0108/00P406Y2;T8^@0I2;T8^@Ri2;
T06@Ri2;T8^@Ri2;0P400dl10@08R`<100Z;RX^:RhZ;RP6:0`402XF;QH^50HF;QH/30@0;R`6;QH^5
RhF;QH/00`402XZ5RXF:QHZ50HD40@03004000810ZD00`40000@00<17`000`40000i00003`000`40
000500030@0000h000<1Y@402H/00`6;R`08R`<128/00`6;R`07R`030H^;00N;0P4023L10H^5RhF;
0`402X/1RhF;QH^5RhD30@0:QHZ5RXF:QHX1RP<11HD00`65Q@02Q@@100Z50HE/QFb5K8E/0`401F^5
0@2Q008100@00@4140030@8000<100007P000`40000h00003`000`40000600030@0000d001@1X@01
08/0R`2;08/00@2;08/0R`<101b;T06@Ri2;T8^@Ri01T8^@Ri2;T8^@0I2;T8^@0`4018/1CcL20@0>
R`410H^:RhZ;0H^:RhX30@0:Q@65RhF;QH^5R`<100Z;QH^5RhF;Q@650`402XF:QHZ50HF:QHX40@03
Q@6500:500H1Y@6U0JD30@d00`4500030@0001h000<10000>00000l000<100001`000`40000;0006
0J6UXJD12H/00`6;R`030@F;00<1Rh/01h/00`6;R`07R`030H^;00<100N;RX^:R`4g00<100Z;QH^5
RhF;Q@650`402XF:QHZ50HF:QHX30@03Q@6500N50`402Vb5K8E/QFb50HD30@0;QF^5JhD1QF/1X@00
0P400jD10@020@X00`4800030@0001d00P4j00003`000`40000800030@0000X000l10:40X@6;08/0
R`2;08/00`4078^@Ri2;T8^@R`6;T8^@Ri2;T8/1Ri2;T8^@Ri030@R;00`1=`41RhZ;RX^:RhX30@0:
QH^5RhF;QH/1R`<100^;QH^50HF;QH^5R`020@0:Q@65RXF:QHZ5RP<128D02@650JD00@0100030@P0
0`4;00030@0001h000<10000>00000l000<100002@000`40000800080JFQYJ6UX@45R`<10X/00`6;
R`08R`030H^;00N;00<1Rh/00X/30@0<RhX1RX^:RhZ;RX^:0P401S<1RhF;Q@<100Z50HF:QHZ5RXF:
0`48Q@040HF5Q@8100V5K8E/0Fb5K8D00`402XD1QF^5JhE[QF/40@040:41Y@<100>U0@0010030@h0
00<100007P000`40000h00003`000`40000:00030@0000H000`10:40X@2Q0:41R`020@0FRi2;T8^@
0I2;T8^@Ri2;T06@Ri2;T0<100B@0I2;0P45R`030H^;00N;00D1C`41RP030@0;QH^5RhD1QH^5RhD0
0P403H/1RhF;QH^5RhF;0H/00P4028F:QHZ5RP6:0`45Q@030HF500:50`402F/10@2U0JD1Y@020@04
Y@4000<14P000`40000M00030@0003P0000>00030@0000`000<100001@002P6QYJ6UXJFQYJ420@R;
00<1Rh/028/0106;Rh/40@0FR`410HZ;RX^:RhZ;0H^:RhZ;RX^:R`@100b5RXF:QHZ5RP6:QHX20@>5
00<1QHD01XD30@07QFb5K8E/Q@030@0:QF^5JhE[QF/1J`<100aVQFJ50@2Q00400@060AD000<10000
7@000`40000h00003P000`40000=00030@0000<002D1X@2Q0:40X@2Q0J41Ri2;T8^@Ri2;0H^@Ri2;
T8^@Ri01T8^@00@10X/00`6;R`06R`030H^;00N;0`402XF;QH^5RhF;QH/30@0;RhF;QH^50HF;QH/0
0`4028X1RXF:QHZ50P400hD1Q@07Q@<100I[QF^5JhD20@06Y@0100400P400jD1000G00030@0001d0
00<10000>00000h000<100003P000`400002000;0JFQYJ6UXJFQY@400Z400`6;R`07R`030H^;00N;
00<1Rh/00`408h^:RhX1RX^:RhZ;RX^:0HZ;RX^:RhZ;RP4bRXF:QHZ5RXF:008128D00`650@020@09
QFb5K8D1QFb5008100U[QF^50HE[QF/00`401fH1IXEVQFH00`401j40X@6U0JD00P400jD1000G0003
0@0001d00P4i00003P000`40000?000U0@000@2Q0:40X@2Q0040X@01T8^@Ri2;T8^@R`6;T8^@Ri2;
T0030@R;00<1Rh/01X/00`6;R`07R`<100F;QH^5R`030@09R`6;QH^5RhF;00<100Z:QHZ5RXF:Q@65
0P46Q@030HD1008100M[QF^50HE[00@100V@R`6U0JD1Y@00104H00030@0001h000<10000=`0000h0
00<10000400030400J6UXJFQYJ6U0@FQ00<1Rh/01h/00`6;R`03R`<101n;RP6:RhZ;RX^:R`6;RX^:
RhZ;RX/1RhZ;RX^:RhZ;00<100B:0@411HD01065QHD30@0:Q@65K8E/QFb5K0<100M[QF^5JhE[00<1
00MVQFJ5IXEV00<100b;T8^@Ri0100400@020@03Y@4101P000<100007P000`40000g00003P000`40
000A000N0J40X@2Q0:40X@6Q0:40X@01T8^@Ri2;T8^@0I2;0P45R`030H^;00N;00<1Rh/01X/00`6;
R`06R`8100dc0@6;QH^5RhF;Q@65008100Z:QHZ5RP6:QHZ50`400hD1Q@04Q@<100M[0F^5JhE[00<1
00Z;T8^@Ri2;T8^@0`401P2U0@6U0AT000<100007P000`40000g00003P000`40000@000;0J6UXJFQ
YJ6UX@402:400`6;R`06R`<1022:RhZ;RX^:R`6;RX^:RhZ;RX/1RhZ;RX^:RhX1RX^:R`<100B;Q@41
28D30@09K8E/QFb5K05/00<100M[QF^5J`5[00<100QVQFJ50HEVQ@8100R;T06@Ri2;T0@100>;T8/0
10400jD10@0I00030@0001h000<10000=`0000h000<1000040006@40X@2Q0:40X@6Q0:40X@2Q0:41
Ri2;T8/00`42R`030H^;00J;00<1Rh/01h/00`6;R`06R`030H^;00811H/01P5>QH^5R`@100V50HF:
QHZ5RXD00`47Q@<100M[QF^5JhE[00@100N@Ri2;0H^@00<119020@>@104J00030@0001h000<10000
=`0000h000<100003`002`6UXJFQYJ6UXJD100ZQ00<1Rh/00P4088Z;RX^:0HZ;RX^:RhZ;RP6:RhZ;
RX^:RhX1RX^:RhZ;10402HF;QH^5RhF;0@02Q@8100Z5K01/Q@65K8E/0`4026/1JhE[QF^50P402FH1
IXEVQFJ5IP020@08T8^@Ri2;T8/20@04T8^@R`<100=gT040104L00030@0001h000<10000=`0000d0
00<1000040005P6Q0:40X@2Q0040X@2Q0:40X@2Q0:420@R;00<1Rh/01X/00`6;R`07R`030H^;00@1
0h/00`6;R`06R`8100Z:QHZ508F:Q@650`44Q@040HF5Q@<100R5JhD1QF^5J`<100F;T8^@R`040@:@
00<1T9000P42T0@180000`40000N00030@0003L0000=00030@0000h00P402J6UXJFQYJ6U0@0<X@09
0J41RX^:RhZ;00<101:;RX^:RhZ;RX/1RhZ;RX^:RhX30@0IQH^5RhF;Q@65RhF;QH^5RhD1K8D0QFb5
K0030@09JhE[QF^5J`5[008100J5IXEVQFH50@05Ri01T8/00`401h^@Ri2;T8/00P400i1g0@020B<0
00<100007`000`4000020081<P0000d000<100002`030@0J08/1X@2Q0:400@2Q0:40X@2Q0:40X@6Q
0045R`D128/00`6;R`03R`<100>;0H/028/00`6;R`06R`050HZ5RXD00`400hD1Q@06Q@<100I[QF^5
JhD30@07T8^@Ri2;T0020@F@1043T0<19P000`40000N00050@0000400P000`40000_00003@000`40
000900811X/01`6QYJ6UX@402j400`6QX@02X@060H^:RhZ;10402hZ;RX^:RhZ;RP6:00<101V5RhF;
Q@65RhF;QH^5R`6;QH^5RhF;Q@5/00<100U[QF^5J`5[QF/00`401VH1IXEVQ@<100N@Ri2;T8^@0081
00F;T06@R`030@03Mi0100@1:@000`40000N00050@0000400P000`40000_00003@000`4000070081
022;08/0R`2;08/1X@2Q0J40X@2Q0:40X@2Q0J40X@2Q0@J;00@1Rh/11h/30@R;00<1Rh/01h/00`6;
R`05R`8128D020650@5[QF^51@401i2;0H^@Ri000`45T0810Y050B`000<100007P001@4000010080
00<10000;`0000d000<1000020000`6;R`07R`040JFQ0@^Q00<1XJ401J403P6:RhZ;RX^:R`6;RX^:
0`409XD1QH^5RhF;QH^50HF;QH^5RhF;Q@65RhF;QH^50F^5JhE[QF^50`4018EVQFH30@09Ri2;T8^@
0I2;008100>@Ri0010400gN@M`040Bl000<100007@020@<000@10001<@0000d000<100002@008@40
R`2;08/0R`010040X@2Q0:40X@2Q0J40X@2Q0:400@09R`040H/10@F;00<1Rh/01h/00`6;R`07R`03
0H^;00F;1HD30@04Q@65J`<100Z;T8^@Ri2;T8^@0P42T0<10Y060C8000<100007P001@4000010080
00<10000;`0000d000<100002P000`6;R`06R`812Z400`6QX@05X@@100N;RX^:RhZ;00<1022;QH^5
RhF;0H^5RhF;QH^5R`6;QH^5RhF;QH/1RhF;Q@<100=[Q@400P4018EVQFH30@09Ri01T8^@Ri2;00<1
00>;T8/00P400i1g0@050CD000<100007P000`4000020081<P0000`000<100002`001P6;08/0R`<1
01>;T040X@2Q0:40X@6Q0:40X@2Q00D118/30@03R`6;00R;00<1Rh/01h/00`6;R`07R`070H/10HZ;
RP020@06QF^5JhE[0P40292;T8^@0I2;0`42T0<10i060CL000<100007P000`40000f000030000`40
000<00030H^;00811X/00`6QX@04X@030J6Q00NQ0`400j41RP020@0QRhF;QH/1RhF;QH^5RhD1QH^5
RhF;QH^50HF;QH^5RhF;00<100Z;QH^5RfJ5IXEV0`4028^@Ri2;T8^@0P400h^@0@020@03T7L100H1
>P000`40000O00030@0003D0000<00030@0000d00P404Y2;T8^@Ri2;T040X@2Q0J40X@<100RQ0:40
X@6Q00811h/00`6;R`07R`030H^;00J;00<1Rh/00h/30@0:R`6;RX^:RhZ5J`@100F@Ri2;T0040@>@
0P42T0H1?P000`40000O00030@0003D0000<00030@0000h000<1Rh/01h/01@6QXJ4100>Q1047X@81
01j5RhF;QH^5RhD1QH^5RhF;QH/1RhF;QH^5RhF;0H/30@0;RhF;QH/1RhF;QH/00`401I2;0H^@00<1
00B@Ri2;0P400i1gT0050D4000<100007P020CL0000<00030@0000l000`1Ri2;T8^@Ri2;0J440@03
0:41008100ZQ0:40X@010:4128/00`6;R`07R`030H^;00F;0`402h^:RhZ;RX^:0HZ;00<100J@Ri2;
T8/20@B@0`400i010@040D@000<100007`000`40000e000030000`40000@00030H^;00J;1P43X@03
0J6Q00>Q0P406j6:X@6;QH^5RhF;QH/1RhF;QH^5RhD1QH^5R`030@0;R`6;QH^5RhF;QH/00`401I2;
T8^@00<100F@R`6;T0020@03M`4100@1A`000`40000O00030@0003D0000<00030@00010000H1T8^@
Ri040@03R`4100810100X@2Q0:40X@6Q0:40X@0128/00`6;R`07R`0>0H/10H^:RhZ;0H^:RhX40@05
T8/1Ri000`46T0D100=g0@40B`000`40000N00030@0003D0000<00030@00014000<1Rh/01P4018/1
X@48X@0J0HZQRZ6:XHX1QH^5RhF;QH^50HF;QH^5RhD30@09QH^5RhF;Q@6500<100N;T8^@Ri2;0081
00F@Ri2;T0030@03T041009g0P5=00030@0001h000<10000=@0000`000<100004P040@04R`410@:;
01@10:40X@2Q0:41X@2Q0:40X@2Q0@R;00<1Rh/00X/30@0:RP6:RhZ;RX^:R`8100J;T8^@Ri030@:@
00<1T9000`401I010@6@0081D0000`40000N00030@0003D0000;00030@0001001`4028^:0HZ;RX/1
1Z405`6:XHZQRZ6:XHZQ0H^5RhF;QH^5R`6;008100f5RhF;Q@65RhF;QH^500<100>;T0400P401Y2;
T8^@R`8100>@M`40105C00030@0001h00P4f00002`000`40000?00@100@00@411h/01@6Q0:400081
00fQ0:40X@2Q0:40X@0100N;0`404X^:RhZ;RX/1RhZ;RX^:R`6UX0811I030@>@1@5E00030@0001l0
00<10000=00000/000<1000040030@8000/1RX^:RhZ;RX^:0@02X@0D0J6:XHZQRZ6:XHZQRZ6:0HF;
QH/30@0ER`6;QH^5RhF;QH/1RhF;QH^50J2Q008100>;T0400P400i1g0@090E@000<100007`000`40
000d00002`000`40000<00<10P000`40000200030H^;00J;01D10040X@2Q0:40X@2Q0:40X@2Q0H/0
0P406X^:RhZ;0H^:RhZ;RX^:0HZ;RX^:0J2UX06P0P42T0L100>P0@400`5D00030@0001l000<10000
=00000/000<100002`040@P000L1RX^:RhZ;00<1016QRZ6:XHZQRZ6:XHZQRZ6:X@020@0NQH^5RhF;
Q@65RhF;QH^5RhD1QH^5R`6QX:41Y@6U1@401JD1Y06T00<100>:0@00D`000`40000O00030@0003@0
000;00030@0000`00`4800030H^;00@10X/0406Q0:40X@2Q0:40X@2Q0:430@0WRX^:RhZ;RX/1RhZ;
RX^:RhZ;0H^:RhX1X:FP0J01X06P0J01X06P00<100>P0@400P5C00030@0001l000<10000=00000/0
00<100003@000`400008008101B5RhF;QH^50HZQRZ6:XHZQRZ6:X@8102Z:XHX1QH^5RhF;QH/1RhF;
QH^5RhF;0H^5R`6QX:41Y@6U0JD1Y@6U0J@20@03X04100<1D`000`40000O00030@0003@0000;0003
0@0000T00`4<00030H^;00J;00/10:40X@2Q0:40X@020@FQ02L1RhZ;RX^:RhX1RX^:RhZ;RX^:0HZ;
RP6PY@6UX06P0J01X06P0J000P401Z@10@6T0E<000<100007`000`40000R008120020@H0000:0003
0@0000T0104=000A0HF;QH^5RhF;0J6:XHZQRZ400P407HZQRZ6:XHZQ0H^5RhF;QH^50HF;QH^5RhF;
Q@65008100fQX06PY@6U0JD1Y@6U00@100BP0@6P0`5B00030@00020000<100008@000`4000080003
0@0000@0000:00030@0000X00`4=00030H^;00J;00H1X@2Q0:420@ZQ0181RX^:RhZ;RX/1RhZ;RX^:
RhX30@0>Q@6PY@6UX06P0J01X:D50@03Y06T00@1D`000`40000O00030@0001800`4200030@0000<1
1000104000040@<000<10000100000X000<1000020040@l000`1RhF;QH^5R`6QRZ420@0KRZ6:XHZQ
RZ6:XHZQ0H^5RhF;QH^50HF;QH^500<101650HD1X:6P0J2U0JD1Y@6UX0060@06Y0410Il1D`000`40
000O00030@0001<000@100010P000`40000500050@0000400P001@40000100H0000:00030@0000L0
104@00030H^;00J;0`4?X@0;0H^:RhZ;RX^:0HX00`401HF;QH^5008100fUX:D1YJ01X06P0J2U00<1
00BU0@650P401:@10@5B00030@0001l000<100004`030@<00`4500030@00008000<100000`000`40
000400002P000`40000800<140001`6;QH^5RhD00`4068ZQRZ6:XHZQRZ6:XHZQRZ6:0HF;QH^5R`@1
00R5RXF:QHZ5RP8100fPXJ01X:D1Y@6UX06P00<100BP0@6T0P4019l10@5B00030@0001l000<10000
4`001040004200040@000@@000<100000`000`40000200030@0000@0000:00030@0000H0104B0003
0H/100<100B:RhX14J400`6:0@030@0<RhD1QH^5RhF;QH^50P404JFPY@6UX06P0J2U0JD1R`6U0081
00>P0J000P400j@1001A00030@0001h00P4E00040@000@<00P4500040@0000@10`000`4000040000
2P000`40000500@14`020@0HRhF;QH^5R`6QRZ6:XHZQRZ6:XHZQRZ6:0`4048F:QHZ5RXD1QHZ5RXF:
QHX20@0AX:6P0J2U0JD1YJ01X0650J000P400j@1Y0020@03W`41054000<100007`000`40000B00@1
2`020@P00P4600002P000`40000600<14`002`6;RX^:RhZ;RX/100fQ0P40:Z41QH^5RhF;QH/1RhF;
QH^5RhF;0JFPY@6UX06P0J2U0JD1Y@6U0JD1Y08100BP0@41D@000`40000O00030@0003<0000:0003
0@0000L000<100004`00506;QH^5RhF;Q@6:XHZQRZ6:XHZQ0`40:J6:XHX1QHZ5RXF:Q@65RXF:QHZ5
RXD1X:6P0J2U0JD1YJ01X06PY@6T008100BP0@6P0`5A00030@00020000@10000104]00002@000`40
000N000;0HZ;RX^:RhZ;RP401j420@NQ02D1RhF;QH^5RhD1QH^5RhF;Q@6PYJ2U0JFP0J01X:D1Y@6U
X06P008100>T0J@00P400il1001@00030@00020000<100000`000`40000/00002@000`40000O000>
0HF;QH^5RhF;0J6:XHX30@0[RZ6:XHZQRZ6:0HF:QHZ5RXF:0HZ5RXF:QHX1XJ2Q0JFP0J2U0JFP0J01
X0020@08Y041X06P0Il30E0000<1000080000`40000300030@0002`0000900030@0001l000`1RhZ;
RX^:RhZ;0J420@fQ0201QH^5RhF;QH/1RhF;QH^50J2UX06PY@6UX06PY@6UX0H100FTX041Y0030E00
00<100007P030@D000<10000;00000T000<100008000206;QH^5RhF;0`40=j6:XHZQRZ6:XHZQRZ6:
X@6:QHZ5RXF:Q@65RXF:QHX1XJ2Q0JFP0J2U0JFP0J2U0HD1Y06T0J@00P4019l1W`5@00030@000200
00<100000`000`40000/00002@000`40000P00050HZ;RX/00`400h/1X@0?X@0>0HF;QH^5RhF;0H^5
RhD30@0KX:FP0J2U0JFP0J2U0JFP0HD1X06P0J01X06P0081D@000`40000O00040@0000<1;P0000T0
00<100008@009@650@6:QHZ50HZQRZ6:XHZQRZ6:XHZQRZ6:0HF:QHZ5RXF:0HX00P42Q@0E0J6PX@6U
X06PY@6UX06PY@650JBP008100F50J@1X0030E0000<100007`000`40000300030@0002`000090003
0@0002400P4028F;QH^5RhD14:40206;QH^5RhF;0`406H^5RhD1X:FP0J2U0JFP0J2U0JFP0HD1X:D0
0P402:01X06P0Il1D0000`40000O00030@000380000900030@00024001l1QHZ5RXF:QHX1XHZQRZ6:
XHZQRZ6:XHZQRP65RXF:00<100>50HD00hD05P6PXJ2Q0JFP0J2U0JFP0J2U0HFP0J020@06Y:010J2O
0P5@00030@00020000<10000<@0000P000<100008`002@65RhF;QH^50@0AX@0[0HD10H^5RhD1QH^5
R`6UX:FP0J2U0JFP0J2U0JF50HFU0JD1Y@6T0J@1Y0020E0000<1000080000`40000a000020000`40
000S000H0HZ5RXF:QHX1XHZQRZ6:XHZQRZ6:XHZQ1047Q@0R0HF5Q@6PXJ2Q0JFP0J2UX06PY@6U0HFP
0J01X06P0J01W`81D0000`40000O0081<`0000P000<100008`002P65RhF;QH^5R`48X@H10Z40;065
RhF;QH^5R`6;QH/1YJ2UX06PY@6UX:D1YJ01X06PY@6U0JD1Y06T0JBP0P5?00030@00020000<10000
<@0000P000<100008`00306:QHZ5RXF:Q@6:X@H100V:XHZQRZ6:X@4028D06`65QHD1X:6P0J2UX06P
YJ01X:D1Y@6UX05/Y0020@06X:@10InP0P5?00030@00020000<10000<@0000P000<1000090001P6;
QH^5R`D13j408`6;QH^5RhF;0H^5R`6UX:D1YJ2U0JFPY@6UX06PY@6U0JFP008100JTX06P0J020Dl0
00<1000080000`40000a000020000`40000T00030HD100<10XD04`6:XHZQRZ6:XHZQRZ6:XHZQRP40
1hD08`65QHD1X:6P0J2UX06PYJ01X:D1YJ01X06PY06T0J@1Y06P0081C`000`40000P00030@000340
000800030@0002@00P4028F;QH^5RhD14:40;@6;QH^5RhF;0H^50J2UX:D1YJ2U0JFPY@6UX06PY@6U
X06P0J01X06P0J2T0@1?00030@00020000<10000<@0000P000<1000090000`65Q@06Q@0B0J6:XHZQ
RZ6:XHZQRZ6:XHX128D090650J6PXJ01X:FP0J2UX06PY@6UX05/Y@6T0J@1Y06T0J2O0Dl000<10000
80020C80000800030@0002@000X1RhF;QH^5RhD14:40;06;QH^5RhF;Q@650J2UX:D1YJ2U0JFPY@6U
X06PY@6:X06P0J01QJ@1Y06T0P5?00030@00020000<10000<00000L000<100009P000`65Q@05Q@0B
0J6:XHZQRZ6:XHZQRZ6:XHX11hD40@0RXJ2QX06PYJ01X:FP0J2U0JFP0FbU0JBP0J01X06P0InP0Dl0
00<1000080000`40000`00001`000`40000V000:0HF;QH^5RhF;0A2Q00@1QH^50`409XD1Q@6PYJ2U
0JFPY@6UX:D1YJ01X:D1RZ01X:D1Y@6T0J@1Y:01C`000`40000P00030@000300000700030@0002H0
00<1QHD01XD04`6:XHZQRZ6:XHZQRZ6:XHZQ0HD00P40:Fb5K8D1Q@6QX:6P0J2UX06PYJ01X:D1YJ01
K:D1Y:01X06P0J01Wj0104l000<1000080000`40000`00001`000`40000V00080HF;QH^5RhDD0@07
RXF:QHZ5RP020@0RX:FPY@6UX:D1YJ2U0JFP0J2U0HZP0HFU0JD1Y:01X06P0Dl000<1000080000`40
000`00001`000`40000W00040HF5Q@<103m/0HFQQJ65XHFQQJ65XHFQQJ41QFb5K8E/QF`1X:6PXJ01
X:FP0J2UX06PY@6UX05/Y@6TX06P0J2T0J@1X040CP020B8000<10000<00000L000<100009`002@6;
0@65RXF:0@0@X@0]0HF:QHZ5RXF:0JFPYJ01X:FPY@6UX:D1YJ01X:D1YJ2U0JD1YJ01X06P0J0104X0
104200030@0001h00P4b00001`000`40000W008104B5K8E/QF`1QJ65XHFQQJ65XHFQQJ65XHD1K8E/
QFb5K06PXJ2Q0JFPYJ01X:FP0J2UX06PY@6UX06P0J2T0J@1Y06P0DL00`4700030@0001l000<10000
;`0000L000<100009`002@6:QHZ5RXF:0@0AX@0Z0HZ5RXF:QHX1YJ2UX06PYJ2U0JFPY@6UX:D1YJ01
X:D1YJ01X06P0J2T0P5400<12P000`40000O00030@0002l0000600030@0002P004@1K8E/QFb5K8D1
XHFQQJ65XHFQQJ65XHFQQ@5/QFb5K8E/0J2QX:41YJ2UX06PYJ01X:FP0J2U0JFP0J2T0J@1Y:01X081
@0040@h000<100007P000`40000_00001P000`40000X000:0HZ5RXF:QHZ50A2Q02X1RXF:QHZ5RP6U
X:FP0J2UX:D1YJ2U0JFPY@6UX06PY@6UX06P0J2T0J@20Cd00`4C00030@0001d000<10000;`0000H0
00<10000:@00@`65K8E/QFb50J65XHFQQJ65XHFQQJ65XHD1K8E/QFb50J6PXJ2Q0JFPYJ01X:FP0J2U
X06PY@6UX06PY06TX06P0J000P4j00<15`000`40000L00030@0002l0000600030@0002T000T1QHZ5
RXF:Q@404:40;06:QHZ5RXD1X:FPYJ01X:FPY@6UX:D1YJ2U0JFP0J2U0JFP0J2U0J@1Y:01=P040AX0
00<1000070000`40000_00001P000`40000Y00060HE/QFb51`40>:65XHFQQJ65XHFQQJ41QFb5K8D1
XJ2QX:41YJ2UX06PYJ01X:FP0J2U0HFP0J2T0JBP0J2T0J@1<`030Al000<100006P020C4000060003
0@0002T000<1Q@400`42Q@030J6Q00:Q2@44X@030HD100@102BPYJ2UX06PYJ01X:FPY@6UX:D1QJ01
QJFP0J2U0JD1Y:01X04`00<18`000`40000J00030@0002l0000600030@0002T00P4056^5JhE[Q@6Q
Q:64XHBQQ:64XHBQ1P40:F^5JhD1XJ2QX:41YJ2U0JFPYJ01X:FP0J2U0HFPY@6TX06P0J2T0J@102`0
104W00030@0001X000<10000;P0000H000<10000:@000`65Q@05Q@030J6Q00nQ00<1QHD00hD09@6P
YJ2UX06PYJ01X:FPY@6UX:D1YJ01QJFP0J2U0JFP0J01X040:@030B`000<100006@00104000040BT0
000600030@0002T004D1QF^5JhE[QF/1Q:64XHBQQ:64XHBQQ:64X@65JhE[Q@6QX:6P0J2UX:D1YJ2U
X06PYJ01X:FP0J2U0JBP0J2T0J@1Y0409P030Bl000<100006@001@40000102`0000500030@0002X0
00<1QHD01XD00`6QX@0>X@030HF500:502@1YJ2UX:D1YJ2UX06PYJ2U0JFPY@6UX:D1YJ01X:D1YJ01
X:@20B80104c00030@0001P000<100000P000`40000Y00001@000`40000Z00130HE[QF^5JhE[0HBQ
Q:64XHBQQ:64XHBQQ:41QF^5J`6PXJ2QX06PYJ2U0JFPYJ01X:FP0J2UX06PY@6TX06PY06TX0020Al0
0`4h00030@0001L000<100000`000`40000X00001@000`40000Z00030HF500J500<1XJ403Z400`65
Q@02Q@0T0JFPYJ2U0JFPYJ01X:FPY@6UX:D1YJ2U0JFP0J2UX06P0J2T0P4K00@1?0000`40000E0081
1P000`40000W00001@000`40000Z00080HE[QF^5JhD30@0hXHBQQ:64XHBQQ:64XHBQQ05[QF/1X:6P
XJ01X:FPY@6UX:FP0J2UX06PYJ01X:D1Y:2T0JBP0J020AP00`5000030@0001H000D100000@020003
0@0002L0000500030@0002X000<1QHD00XD30@03Q@6Q00@13:40:P65QHD1YJ2UX:D1YJ2UX06PYJ2U
0JFPYJ01X:D1YJ2U0JFP0J2U0JBP0AD00`5400030@0001D000<100000P020BX0000500030@0002X0
00<1Q@400P402XEVQFH1Q:64XH@40@0bXHBQQ:64XH@1JhE[0J2QX:6P0J2UX:D1YJ2UX06PYJ2U0JFP
0J2UX06PY06TX06PY04A00@1B0000`40000E00030@0002d0000500030@0002X00P4028F4QHB5Q8D1
2J430@FQ02X1QHF50JFPYJ2U0JFPYJ01X:FPY@6UX:FP0J2U0JFPY@6UX06PY@6TX04>00<1C@000`40
000D00030@0002d0000500030@0002X001H1QFJ5IXEVQFH1Q:64XHBQQ:64XHBQ10400h@1J`020@0V
XJ2QX:6P0J2UX:D1YJ2UX06PYJ2U0JFP0J2UX06PY06TX06PY04;00<1D@000`40000C00030@0002d0
000400030@0002/000T1Q8F4QHB5Q0404J430@0XQ@6PYJ2UX:D1YJ2UX06PYJ2U0JFPYJ01X:D1QJ2U
0JFPY@6UX06P0@L0105D00030@0001<000<10000;@0000@000<10000:`00=`65IXEVQFJ50J64XHBQ
Q:64XHBQQ:64XHBQQ05VQ@6QX:6PXJ01X:FPY@6UX:FP0J2UX:D1YJ000P403:FP0J2TX06PY06T0@@0
0`5I00030@0001400P4_000010000`40000[00090HB5Q8F4QH@101:Q02/1QHD1X:FPYJ01X:FPYJ01
X:FPY@6UX:FP0J2UX06PY@6UX:D1YJ01X04000<1G@000`40000A00030@0002d0000400030@0002/0
00H1QFJ5IXD40@0hQ:64XHBQQ:64XHBQQ:64XH@1IXD1XJ2QX:41YJ2UX:D1YJ2UX06PYJ2U0JFPY@6U
X:D1Y:01X:@40F4000<1000040000`40000]000010000`40000[00030H@100<100B5IP6P0P4?X@0T
0HF50J2UX:FP0J2UX:FP0J2UX:D1YJ2UX06PYJ01X:FP0J2U1@5S00030@00010000<10000;@0000@0
00<10000:`020@0:IXAVQ6J40J64X@<102fQQ:64XHBQQ:64XH@1IXD1XJ2QX:41YJ2UX:D1YJ2UX06P
YJ2U0JFPY@6UX:D010401:2TX05T00030@00010000<10000;00000@000<10000:`003`5VQFJ5IXEV
0J2QX:6PX@030@RQ01h1QHD1YJ2UX:FP0J2UX:FP0J2UX:D1YJ2UX06PYJ050@07X:FP0J2U0@1U0003
0@0000l000<10000;00000@000<10000:`004P64IXAVQ6H1Q:64XHBQQ:64X@810224XHBQQ:41QFH1
X:6PXJ2Q0JFPYJ2U0JFPYJ01X:FPY@D100^UX:D1Y:2T0JBP0@1V00030@0000h000<10000;00000@0
00<10000:`00505VQFJ5IXD1XJ2QX:6PXJ2QX:6P0`43X@0G0HF50JFPYJ2UX06PYJ2UX06PYJ2U0JD0
10403j2UX06PYJ01X:FP0J2U0@1W00030@0000`00P4^00000`000`40000[000H0FJ4IXAVQ6H1Q:64
XHBQQ:64XHBQQ:640P400j41Q@020@0@X:6PXJ2Q0JFPYJ2U0JFPY@<101>UX:D1YJ2U0JFPY@6TX:@1
Y:0106L000<100003@000`40000/00000`000`40000[00060HEVQFJ50`404J6PXJ2QX:6PXJ2QX:6P
XJ2Q00<100j50JFPYJ2UX06PYJ2UX0@101FU0JFPYJ2U0JFP0J2UX06PYJ01X040J@000`40000<0003
0@0002`0000300030@0002/000<1IP400`400fIU0@020@0MQ:64XHBQQ:64XHBQQ:64X@64IP6PXJ2Q
X:41YJ0010406:FPYJ2U0JFPYJ01X:D1YJ2UX06PY06T0FX000<100002`000`40000/00000`000`40
000[008100UVOfIoIP6KXI/00P405j6PXJ2QX:6PXJ2QX:6P0FJ50JFPYJ2U00@101ZPYJ01X:FPYJ01
X:FPY@6UX:D1YJ2U0JFPY@81J`000`40000:00030@0002`0000300030@0002/000d1IFIUIVEV0J64
XHBQ008101>4XHBQQ:64XHBQQ:41Q06QX:6P00<101fPYJ2UX:D1YJ2UX:D1YJ2UX06PYJ01X:FP0J2T
X0020F/000<100002`000`40000[00000`000`40000Z000@0FIoIWmVOfH1Vj6KXI^QV`8100bQX:6P
XJ2QX:6P0FH50@0QX:D1YJ2UX:FP0J2UX:FP0J2UX:D1YJ2U0JFPY@6UX:D106d000<100002P000`40
000[00000`000`40000Z000L0FIUIVEVI@64XHBQQ:64XHBQ0J64XHBQQ:64X@<102FQX:6PXJ01X:FP
YJ2U0JFPYJ2U0JFPYJ01X:FP0J2UX06PY:0106h000<1000020020Bd0000300030@0002T000H1OfIo
IWl30@0;XI^QVj6KXI^QVj400P401:2QX:430@0XQFH1X:FPYJ2U0JFPYJ2UX06PYJ2UX06PYJ2U0JFP
Y@6UX:D1YJ2U0Fl000<1000020000`40000[00000P000`40000Z00030FD100<100=U0J000P402hBQ
Q:64XHBQQ:6400@102ZQQ05VQ06QX:6PXJ01X:FPYJ01X:FPYJ2U0JFPYJ01X:FP0J2UX06PY05a0003
0@0000L000<10000:`00008000<10000:@030@0AH6IPIV01Vj6K0I^QVj6KXI/00`40;Y^Q0@6PX@65
IP6PYJ2UX:D1YJ2UX:D1YJ2UX:FP0J2UX:D1YJ2U0JFPY@6UX05a00030@0000L000<10000:`000080
00<10000:0004P5PIF1UH6EP0HBPQ:240HBQQ0<100N4XHBQQ:6400<102AV0@6QX:6PXJ01X:FPYJ01
X:FPYJ2U0JFPYJ01X:FP0J2UX:D20G<000<100001P000`40000[00000P000`40000X000>0FIPIV1V
H6H1XI^QVj440@0<Vj6KXI^QVj6KXI^Q0`408ZFPYJ2UX:D1YJ2UX:D1YJ2UX:FP0J2UX:D1YJ2U0JFP
YJ020G@000<100001@000`40000[00000P000`40000@00815@001P5UH6EPI@<100>PQ:000`40=Z24
X06QQ:64XHBQQ:64XH@1IP6PXJ2QX:6P0J2UX:FP0J2UX:FPY@6UX:FP0J2UX06PYJ2U0GH000<10000
1@000`40000Z00000P000`40000=00<100B@0@414`000`5P0@030@03I@6P00<100NQVj6KXI^Q0081
032KXI^QVj6KXI/1IWl1YJ2UX:FPY@6UX:FPY@6UX:FPYJ01X:FPY@6UX:FP0J2UX05f00030@0000@0
0P4/00000P000`40000:00<100R;T8^@R`6I0A400`400dIOAP040@0GX06PQ:24X8BPQ:240HBQQ:64
XHBQ0FH00P408Z2QX:6PX@6UX:FPYJ01X:FPYJ2U0JFPYJ01X:FPY@6UX05h00030@0000@000<10000
:P00008000<1000020020@R@00<1VYX00P4>00040DIUAP@100FPVj2KX0020@09Vj6KXI^QVj6K0081
00BQVj6K1042T08101nUX:FP0J2UX:FPY@6UX:FPYJ01X:FPYJ01X:FP0J0107X000<100000`000`40
000Z00000P000`400009008100b;T8^@Ri2;T06JVIX20@`01`404j24X8BPQ:240HBPQ:24X8BPQ:00
10409h^@Ri2;T8^@0J6PX@6UX:FPYJ01X:FPYJ2U0JFPYJ2U0JFPY@6U0@1k00030@00008000<10000
:P000003004000d000<1T9001I000`6JVP02VP<11`080@0BVj01X9^PVj2KX9/1Vj6KXI^Q104<T00K
0JFP0J2UX:FPY@6UX:FPYJ01X:FPYJ01X:FP0081O@001@40000100@00P4V000000<00@003P020@03
T8/1008100Ig0IVJV@43VPH100<00@400`403Y^JViZK0J24X8BPQ:241P400i2;T0020@0<Ri2;T8^@
Ri2;T8^@0`406:FPYJ2UX06PYJ2UX:D1YJ2UX:D1YJ2U0Gh000D100000@0300040@000BD000000`01
000@008119000`6JVP02VP030IX100<100>P0@400`403Z2KX06PVj2KX06PVj2K104:T0812i030@0F
YJ2UX:D1YJ2UX:FP0J2UX:FP0J2U0H0000@100011P000`40000S000000<00@004P030@05Mi01VYT0
0`42VP030IZJ00>J00`1ViZKVY/1ViZKVY/40@0HRi010H^@Ri2;T8^@Ri2;T06@Ri2;T8^@0`400i1g
T0020@0CX:FP0J2UX:FPY@6UX:FPY@6U0@2200040@010@<00P4V000000<00@005@050@:J01<1WiZO
0J2JX9ZP0J2KX9^P0J2K00<11i000`6@T00:T0810Y030@R@0P403ZD1YJ2UX:FP0J2UX:FP0P640003
0@0100D000<100008`000003004001<01@401IZIVYT100BJ00<1VYX00YX01P6JViZKVP<101R;T8^@
Ri2;T8^@R`6;T8^@Ri2;T8^@Ri030@0=T7N@Mi1gT7N@Mi1gT0050@06YJ2U0JFP106500030@010080
00@100019@000003004001000`4500030IZJ00:J00X1VYnJW`6PVZ2J1@42T0812i020@J@1043T081
29030@05T7N@Mi001`6900810`020BH000000`01000=00<120001P6IVYVJ0@BJ00<1VYX00`40292;
T8^@Ri2;0P403Y2;T8^@Ri2;T8^@0I2;0`402GN@Mi1gT7N@M`020@03T7N@00<11WL40Hh00P4[0000
00<100002@040@`000<1VYX00YX0106OVYl30@f@00<1T9001i040@j@1@40191gT7L30H/000<10000
104]000000<100001P030A0000D1VYVJV@040@0AT8/10I2;T8^@Ri2;T8^@Ri000P400h^@R`040@04
T7N@M`8100R@Mi1gT7N@M`@11GL40Hd0104a000000<100000`030A@000<1VP400P47T08139040@Z@
0P43T0<100EgT7N@M`050HT000<10000104e000000<100400`4G008101B@Ri2;T8^@Ri2;T8/1Ri2;
T8^@R`@100>@Mi000P402WN@Mi1gT7N@Mi0>0H/0104i00000P4L008139020@:@0`49T08119040@05
T7N@Mi001P6700030@0000@1?@00000400410A`00P402h^@Ri2;T8^@Ri2;00<100igT7N@Mi1gT7N@
Mi1gT0D11GL60HT0104D00<1:P0000<000<1000070000`6@T005T0<10i030@R@0`401I1gT7N@00L1
QP000`4000040A/000<100009`0000@00P4M008100>@R`400P402GN@Mi1gT7N@M`020@03T7N@00<1
1gL40HT0104O00030@0002L0000500817P020@j@1@401I1gT7N@00<1QP00104000030B000`4Z0000
1`020Al0104017N@Mi040@Ig1@6900@18`000`40000Z000020001040004P00D100IgT7N@Mi050H@0
00@10000104W00030@0002X0000900819P060@@000<10000PP040B/0104Y00002`000`400@0]0003
0@0007X000@10000105L000030020Bd000<10000NP040F00000=00<1:`000`40001c00030@0000@1
I00000l000<10000:@000`40001b00@150000`40001A000040000`400@0X00030@0006/000<10000
104H00030@000540000900811P020BP000<10000JP040AT01@5B000020001040004700819P000`40
001T00030@0000<17@001040005C0000200010400048008100<00@008P000`40001S00@18@000`40
0@1C000020001040004:00818`000`40001L00030@0000@19P020E<0000800040@000@/000<10040
8@000`40001K00@1:`000`40001A000020001040004<008180000`40001F00030@0100<1P`0000T0
0P4?00817P000`40001D00@1Q`0001/000<1000070000`40001?00D1R`0001`00`4K00030@0004`0
104D0081N@0000h00P4400030@0000800`4200030@0101T000<10000A`050AL000@10001N00000d0
00@100012`000`4000020@03004001H000<10000@0000`40000200<17P000`40001f00003@001040
004;00040@0000<15`000`40001000D17`020GT0000=00040@000@P00`4600030@0001D000<10000
>@00104000040BD000<10000MP0000d000@1000120000`40000700030@0101@000<10000>@040BH0
00@10001N00000d000@1000120000`400008008100<00@004@000`40000b00030@0000@1:`020GT0
000>00812@040@T00P4B00030@000340106Z00009`030A0000<10000:`000`400@030Jh0000Y0003
0@0000h000<10000:@040A40106M000080040@H00P4>00030@0002@01@4E00030@0009h0000R0003
0@0000H00`4<00030@0001d000@10000104J00030@0009d0000R00030@0000P000<100002P000`40
000M00@17`000`40002L00008P000`40000900030@0100P000<1000060000`4000040B<000<10000
V`00028000<100002P020@03004000D000<100005`040B@000@10001W@0002000`4>00811P000`40
000A00030@0100<1:@020Ih0000R00030@0000d000<1004010000`40000:00030@000080107=0000
<`020@050040004030050M40000T00@10`000`40000200<10P020@03004100H000<10000104A00@1
`00002H000<100002P001@40004000811@040AL000<10000_`0002H000<100002P00104000030@03
004100816`000`40002o00009P000`40000700<11P020Al000<10000_`0002H000<100001`000`40
000W00030@000;l0000K00811`030@T000<100009@030C<000<10040R`0001/000<1000020000`40
000700@19P000`40000a00030@0108/0000<00D100D00@0100020@<000@10000104300030@0006@0
00D100400@2:00003P002@40000100400@0400050@0000400P001@40000100l0105C00050@010040
RP0000h000T100000@01004010000`40000200030@0000<000<100003@020@8000<10000D0001040
0@020HT0000>00090@0000400@0100@000<100000`000`40000200030@0000h000@10001h@0000h0
00@100000P400`01000400040@0000@10`000`40000>008100<00@00h00000`01@4:008120020A00
00@10041h@0003H00P400`010@3Q0000=`030N80003o01d0003o01d0003o01d0003o01d0003o01d0
003o01d0003o01d0003o01d0003o01d0003o01d0003o01d0003o01d0003o01d0003o01d0003o01d0
003o01d0003o01d00000\
\>"],
  ImageRangeCache->{{{101, 384.562}, {714.062, 427.062}} -> {-0.375673, 
  1.63609, 0.00397, 0.00397}}]
}, Open  ]],

Cell["The points of T may be joined by typing", "Text"],

Cell[CellGroupData[{

Cell["\<\
EnvelopePlot3D[Radius -> r0 , Center -> k0 , ClosedLoop -> T \
,PlotJoined -> True];\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.01205 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0265144 1.04793 0 1.04793 [
[.2389 1.02513 -3.17888 0 ]
[.2389 1.02513 2.82112 9 ]
[.37774 .98992 -3.0835 0 ]
[.37774 .98992 2.9165 9 ]
[.52296 .95309 -2.98205 0 ]
[.52296 .95309 3.01795 9 ]
[.67501 .91451 -2.87394 0 ]
[.67501 .91451 3.12606 9 ]
[.83437 .87405 -2.75849 0 ]
[.83437 .87405 3.24151 9 ]
[1.00158 .83158 -2.63492 0 ]
[1.00158 .83158 3.36508 9 ]
[.59965 .9995 -4.88145 0 ]
[.59965 .9995 5.11855 15 ]
[.04392 .88272 -3.32443 0 ]
[.04392 .88272 2.67557 9 ]
[.09461 .91974 -9.85694 0 ]
[.09461 .91974 8.14306 9 ]
[.14316 .9552 -3.24914 0 ]
[.14316 .9552 2.75086 9 ]
[.18969 .98919 -9.64411 0 ]
[.18969 .98919 8.35589 9 ]
[.23434 1.0218 -3.18217 0 ]
[.23434 1.0218 2.81783 9 ]
[.11933 1.00721 -18.4886 0 ]
[.11933 1.00721 15.5114 15 ]
[.94483 .25907 0 -6.02028 ]
[.94483 .25907 12 2.97972 ]
[.95965 .38048 0 -5.93776 ]
[.95965 .38048 6 3.06224 ]
[.9755 .51042 0 -5.84984 ]
[.9755 .51042 6 3.15016 ]
[.99251 .64981 0 -5.75596 ]
[.99251 .64981 6 3.24404 ]
[1.01079 .79973 0 -5.65551 ]
[1.01079 .79973 6 3.34449 ]
[1.03824 .4916 0 -9.74973 ]
[1.03824 .4916 34 5.25027 ]
[ 0 0 0 0 ]
[ 1 1.01205 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.23968 1.01205 m
1 .81858 L
s
.23968 1.01205 m
.24007 1.00551 L
s
[(0)] .2389 1.02513 .05963 -1 Mshowa
.37811 .97683 m
.37829 .97028 L
s
[(1)] .37774 .98992 .02783 -1 Mshowa
.52288 .93999 m
.52284 .93344 L
s
[(2)] .52296 .95309 -0.00598 -1 Mshowa
.67446 .90142 m
.67418 .89488 L
s
[(3)] .67501 .91451 -0.04202 -1 Mshowa
.83331 .861 m
.83279 .85447 L
s
[(4)] .83437 .87405 -0.0805 -1 Mshowa
1 .81858 m
.99921 .81208 L
s
[(5)] 1.00158 .83158 -0.12169 -1 Mshowa
.125 Mabswid
.26688 1.00513 m
.26709 1.00121 L
s
.29432 .99815 m
.2945 .99422 L
s
.322 .99111 m
.32216 .98718 L
s
.34993 .984 m
.35006 .98007 L
s
.40654 .96959 m
.40662 .96566 L
s
.43523 .96229 m
.43529 .95836 L
s
.46418 .95493 m
.46421 .951 L
s
.4934 .94749 m
.4934 .94356 L
s
.55264 .93242 m
.55259 .92849 L
s
.58267 .92478 m
.58259 .92085 L
s
.61298 .91706 m
.61287 .91313 L
s
.64357 .90928 m
.64344 .90535 L
s
.70563 .89349 m
.70543 .88956 L
s
.7371 .88548 m
.73687 .88156 L
s
.76886 .8774 m
.76861 .87347 L
s
.80093 .86924 m
.80065 .86532 L
s
.86601 .85268 m
.86566 .84876 L
s
.89902 .84428 m
.89864 .84037 L
s
.93235 .83579 m
.93194 .83189 L
s
.96601 .82723 m
.96557 .82333 L
s
gsave
.59965 .9995 -65.8815 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
0 .83649 m
.23968 1.01205 L
s
.04533 .86969 m
.04603 .86318 L
s
[(0)] .04392 .88272 .10814 -1 Mshowa
.09585 .9067 m
.09647 .90018 L
s
[(0.5)] .09461 .91974 .09522 -1 Mshowa
.14424 .94215 m
.14478 .93562 L
s
[(1)] .14316 .9552 .08305 -1 Mshowa
.19063 .97612 m
.1911 .96959 L
s
[(1.5)] .18969 .98919 .07157 -1 Mshowa
.23513 1.00872 m
.23553 1.00218 L
s
[(2)] .23434 1.0218 .06072 -1 Mshowa
.125 Mabswid
.05561 .87722 m
.05602 .87332 L
s
.0658 .88469 m
.0662 .88078 L
s
.07591 .89209 m
.0763 .88818 L
s
.08592 .89943 m
.0863 .89552 L
s
.1057 .91391 m
.10606 .91 L
s
.11546 .92106 m
.11581 .91715 L
s
.12513 .92815 m
.12548 .92424 L
s
.13473 .93518 m
.13506 .93126 L
s
.15368 .94906 m
.15399 .94514 L
s
.16303 .95591 m
.16334 .95199 L
s
.17231 .9627 m
.1726 .95878 L
s
.18151 .96944 m
.18179 .96552 L
s
.19968 .98275 m
.19995 .97883 L
s
.20865 .98932 m
.20891 .9854 L
s
.21755 .99584 m
.2178 .99192 L
s
.22638 1.00231 m
.22662 .99838 L
s
.03495 .86209 m
.03538 .85819 L
s
.02448 .85443 m
.02493 .85052 L
s
.01392 .84669 m
.01438 .84279 L
s
.00327 .83889 m
.00373 .83499 L
s
gsave
.11933 1.00721 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.931 .25158 m
1 .81858 L
s
.93242 .26326 m
.92622 .26536 L
s
[(-1)] .94483 .25907 -1 .33784 Mshowa
.94717 .38447 m
.94093 .38646 L
s
[(0)] .95965 .38048 -1 .3195 Mshowa
.96296 .51418 m
.95668 .51606 L
s
[(1)] .9755 .51042 -1 .29996 Mshowa
.97989 .65333 m
.97358 .65509 L
s
[(2)] .99251 .64981 -1 .2791 Mshowa
.9981 .80299 m
.99176 .80462 L
s
[(3)] 1.01079 .79973 -1 .25678 Mshowa
.125 Mabswid
.93529 .28686 m
.93157 .28811 L
s
.9382 .31078 m
.93447 .31201 L
s
.94115 .33501 m
.93742 .33623 L
s
.94414 .35957 m
.9404 .36078 L
s
.95024 .4097 m
.94649 .41089 L
s
.95335 .43528 m
.9496 .43645 L
s
.95651 .46122 m
.95276 .46238 L
s
.95971 .48752 m
.95595 .48866 L
s
.96625 .54123 m
.96248 .54234 L
s
.96959 .56865 m
.96581 .56975 L
s
.97297 .59648 m
.96919 .59756 L
s
.97641 .6247 m
.97262 .62577 L
s
.98343 .68239 m
.97964 .68343 L
s
.98701 .71187 m
.98322 .7129 L
s
.99066 .74179 m
.98686 .7428 L
s
.99435 .77216 m
.99055 .77316 L
s
gsave
1.03824 .4916 -61 -13.7497 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.23968 1.01205 L
s
.23968 1.01205 m
.27167 .47549 L
s
.27167 .47549 m
.0677 .27203 L
s
.77416 0 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.77416 0 m
.0677 .27203 L
s
.27167 .47549 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.23968 1.01205 L
s
.23968 1.01205 m
.27167 .47549 L
s
0 0 m
1 0 L
1 1.01205 L
0 1.01205 L
closepath
clip
newpath
.5 Mabswid
.30241 .4381 m .26906 .44957 L .28671 .49523 L p .735 .66 .852 r
F P
0 g
s
.28671 .49523 m .32009 .48403 L .30241 .4381 L p .735 .66 .852 r
F P
0 g
s
.22116 .51827 m .28671 .49523 L .26906 .44957 L p .933 .755 .762 r
F P
0 g
s
.32009 .48403 m .28671 .49523 L .30094 .54206 L p .725 .625 .82 r
F P
0 g
s
.28671 .49523 m .22116 .51827 L .23201 .55408 L p .928 .737 .746 r
F P
0 g
s
.23201 .55408 m .30094 .54206 L .28671 .49523 L p .928 .737 .746 r
F P
0 g
s
.30094 .54206 m .33441 .53111 L .32009 .48403 L p .725 .625 .82 r
F P
0 g
s
.33441 .53111 m .30094 .54206 L .31146 .58917 L p .713 .588 .786 r
F P
0 g
s
.30094 .54206 m .23201 .55408 L .23988 .59009 L p .924 .72 .727 r
F P
0 g
s
.23988 .59009 m .31146 .58917 L .30094 .54206 L p .924 .72 .727 r
F P
0 g
s
.28171 .39424 m .24831 .406 L .26906 .44957 L p .745 .694 .883 r
F P
0 g
s
.26906 .44957 m .30241 .4381 L .28171 .39424 L p .745 .694 .883 r
F P
0 g
s
.20756 .48342 m .26906 .44957 L .24831 .406 L p .94 .774 .776 r
F P
0 g
s
.26906 .44957 m .20756 .48342 L .22116 .51827 L p .933 .755 .762 r
F P
0 g
s
.31146 .58917 m .34509 .57843 L .33441 .53111 L p .713 .588 .786 r
F P
0 g
s
.34509 .57843 m .31146 .58917 L .31803 .63563 L p .699 .548 .748 r
F P
0 g
s
.31146 .58917 m .23988 .59009 L .24457 .62553 L p .921 .703 .707 r
F P
0 g
s
.24457 .62553 m .31803 .63563 L .31146 .58917 L p .921 .703 .707 r
F P
0 g
s
.33612 .42651 m .30241 .4381 L .32009 .48403 L p .735 .66 .852 r
F P
0 g
s
.32009 .48403 m .35382 .47271 L .33612 .42651 L p .735 .66 .852 r
F P
0 g
s
.35382 .47271 m .32009 .48403 L .33441 .53111 L p .725 .625 .82 r
F P
0 g
s
.33441 .53111 m .36824 .52004 L .35382 .47271 L p .725 .625 .82 r
F P
0 g
s
.36824 .52004 m .33441 .53111 L .34509 .57843 L p .713 .588 .786 r
F P
0 g
s
.25834 .35332 m .22482 .36539 L .24831 .406 L p .753 .73 .914 r
F P
0 g
s
.24831 .406 m .28171 .39424 L .25834 .35332 L p .753 .73 .914 r
F P
0 g
s
.1915 .45028 m .24831 .406 L .22482 .36539 L p .949 .793 .788 r
F P
0 g
s
.24831 .406 m .1915 .45028 L .20756 .48342 L p .94 .774 .776 r
F P
0 g
s
.31547 .38235 m .28171 .39424 L .30241 .4381 L p .745 .694 .883 r
F P
0 g
s
.30241 .4381 m .33612 .42651 L .31547 .38235 L p .745 .694 .883 r
F P
0 g
s
.34509 .57843 m .37909 .56758 L .36824 .52004 L p .713 .588 .786 r
F P
0 g
s
.37909 .56758 m .34509 .57843 L .3519 .62508 L p .699 .548 .748 r
F P
0 g
s
.31803 .63563 m .3519 .62508 L .34509 .57843 L p .699 .548 .748 r
F P
0 g
s
.3519 .62508 m .31803 .63563 L .32046 .68049 L p .682 .502 .703 r
F P
0 g
s
.31803 .63563 m .24457 .62553 L .24596 .65963 L p .92 .687 .685 r
F P
0 g
s
.24596 .65963 m .32046 .68049 L .31803 .63563 L p .92 .687 .685 r
F P
0 g
s
.3702 .41479 m .33612 .42651 L .35382 .47271 L p .735 .66 .852 r
F P
0 g
s
.35382 .47271 m .38792 .46126 L .3702 .41479 L p .735 .66 .852 r
F P
0 g
s
.38792 .46126 m .35382 .47271 L .36824 .52004 L p .725 .625 .82 r
F P
0 g
s
.36824 .52004 m .40243 .50884 L .38792 .46126 L p .725 .625 .82 r
F P
0 g
s
.40243 .50884 m .36824 .52004 L .37909 .56758 L p .713 .588 .786 r
F P
0 g
s
.29222 .34112 m .25834 .35332 L .28171 .39424 L p .753 .73 .914 r
F P
0 g
s
.28171 .39424 m .31547 .38235 L .29222 .34112 L p .753 .73 .914 r
F P
0 g
s
.34959 .37033 m .31547 .38235 L .33612 .42651 L p .745 .694 .883 r
F P
0 g
s
.33612 .42651 m .3702 .41479 L .34959 .37033 L p .745 .694 .883 r
F P
0 g
s
.20756 .48342 m .17428 .4948 L .18786 .52944 L p .733 .659 .853 r
F P
0 g
s
.18786 .52944 m .22116 .51827 L .20756 .48342 L p .733 .659 .853 r
F P
0 g
s
.22116 .51827 m .18786 .52944 L .19863 .56506 L p .722 .623 .821 r
F P
0 g
s
.23272 .31625 m .19902 .32864 L .22482 .36539 L p .761 .767 .945 r
F P
0 g
s
.22482 .36539 m .25834 .35332 L .23272 .31625 L p .761 .767 .945 r
F P
0 g
s
.17329 .41958 m .22482 .36539 L .19902 .32864 L p .959 .813 .799 r
F P
0 g
s
.22482 .36539 m .17329 .41958 L .1915 .45028 L p .949 .793 .788 r
F P
0 g
s
.19863 .56506 m .23201 .55408 L .22116 .51827 L p .722 .623 .821 r
F P
0 g
s
.23201 .55408 m .19863 .56506 L .20638 .6009 L p .71 .585 .785 r
F P
0 g
s
.37909 .56758 m .41345 .55661 L .40243 .50884 L p .713 .588 .786 r
F P
0 g
s
.41345 .55661 m .37909 .56758 L .38613 .61442 L p .699 .548 .748 r
F P
0 g
s
.3519 .62508 m .38613 .61442 L .37909 .56758 L p .699 .548 .748 r
F P
0 g
s
.38613 .61442 m .3519 .62508 L .35462 .6701 L p .682 .502 .703 r
F P
0 g
s
.32046 .68049 m .35462 .6701 L .3519 .62508 L p .682 .502 .703 r
F P
0 g
s
.1915 .45028 m .15819 .46188 L .17428 .4948 L p .743 .695 .885 r
F P
0 g
s
.17428 .4948 m .20756 .48342 L .1915 .45028 L p .743 .695 .885 r
F P
0 g
s
.40465 .40294 m .3702 .41479 L .38792 .46126 L p .735 .66 .852 r
F P
0 g
s
.38792 .46126 m .42239 .44969 L .40465 .40294 L p .735 .66 .852 r
F P
0 g
s
.42239 .44969 m .38792 .46126 L .40243 .50884 L p .725 .625 .82 r
F P
0 g
s
.20638 .6009 m .23988 .59009 L .23201 .55408 L p .71 .585 .785 r
F P
0 g
s
.23988 .59009 m .20638 .6009 L .2109 .6362 L p .695 .543 .745 r
F P
0 g
s
.40243 .50884 m .437 .49753 L .42239 .44969 L p .725 .625 .82 r
F P
0 g
s
.437 .49753 m .40243 .50884 L .41345 .55661 L p .713 .588 .786 r
F P
0 g
s
.32647 .32879 m .29222 .34112 L .31547 .38235 L p .753 .73 .914 r
F P
0 g
s
.31547 .38235 m .34959 .37033 L .32647 .32879 L p .753 .73 .914 r
F P
0 g
s
.35462 .6701 m .32046 .68049 L .31858 .72281 L p .66 .447 .648 r
F P
0 g
s
.32046 .68049 m .24596 .65963 L .24393 .69159 L p .921 .67 .659 r
F P
0 g
s
.24393 .69159 m .31858 .72281 L .32046 .68049 L p .921 .67 .659 r
F P
0 g
s
.3841 .35818 m .34959 .37033 L .3702 .41479 L p .745 .694 .883 r
F P
0 g
s
.3702 .41479 m .40465 .40294 L .3841 .35818 L p .745 .694 .883 r
F P
0 g
s
.17329 .41958 m .13989 .43141 L .15819 .46188 L p .751 .731 .917 r
F P
0 g
s
.15819 .46188 m .1915 .45028 L .17329 .41958 L p .751 .731 .917 r
F P
0 g
s
.26679 .30373 m .23272 .31625 L .25834 .35332 L p .761 .767 .945 r
F P
0 g
s
.25834 .35332 m .29222 .34112 L .26679 .30373 L p .761 .767 .945 r
F P
0 g
s
.41345 .55661 m .4482 .54552 L .437 .49753 L p .713 .588 .786 r
F P
0 g
s
.4482 .54552 m .41345 .55661 L .42073 .60364 L p .699 .548 .748 r
F P
0 g
s
.38613 .61442 m .42073 .60364 L .41345 .55661 L p .699 .548 .748 r
F P
0 g
s
.42073 .60364 m .38613 .61442 L .38915 .6596 L p .682 .502 .703 r
F P
0 g
s
.35462 .6701 m .38915 .6596 L .38613 .61442 L p .682 .502 .703 r
F P
0 g
s
.2109 .6362 m .24457 .62553 L .23988 .59009 L p .695 .543 .745 r
F P
0 g
s
.24457 .62553 m .2109 .6362 L .21207 .67017 L p .677 .495 .699 r
F P
0 g
s
.43948 .39096 m .40465 .40294 L .42239 .44969 L p .735 .66 .852 r
F P
0 g
s
.42239 .44969 m .45724 .438 L .43948 .39096 L p .735 .66 .852 r
F P
0 g
s
.45724 .438 m .42239 .44969 L .437 .49753 L p .725 .625 .82 r
F P
0 g
s
.437 .49753 m .47195 .48609 L .45724 .438 L p .725 .625 .82 r
F P
0 g
s
.47195 .48609 m .437 .49753 L .4482 .54552 L p .713 .588 .786 r
F P
0 g
s
.3611 .31632 m .32647 .32879 L .34959 .37033 L p .753 .73 .914 r
F P
0 g
s
.34959 .37033 m .3841 .35818 L .3611 .31632 L p .753 .73 .914 r
F P
0 g
s
.38915 .6596 m .35462 .6701 L .35309 .71254 L p .66 .447 .648 r
F P
0 g
s
.31858 .72281 m .35309 .71254 L .35462 .6701 L p .66 .447 .648 r
F P
0 g
s
.41898 .3459 m .3841 .35818 L .40465 .40294 L p .745 .694 .883 r
F P
0 g
s
.40465 .40294 m .43948 .39096 L .41898 .3459 L p .745 .694 .883 r
F P
0 g
s
.20532 .28393 m .17137 .29664 L .19902 .32864 L p .768 .809 .978 r
F P
0 g
s
.19902 .32864 m .23272 .31625 L .20532 .28393 L p .768 .809 .978 r
F P
0 g
s
.15329 .39205 m .19902 .32864 L .17137 .29664 L p .971 .833 .807 r
F P
0 g
s
.19902 .32864 m .15329 .39205 L .17329 .41958 L p .959 .813 .799 r
F P
0 g
s
.15329 .39205 m .11976 .40411 L .13989 .43141 L p .759 .77 .949 r
F P
0 g
s
.13989 .43141 m .17329 .41958 L .15329 .39205 L p .759 .77 .949 r
F P
0 g
s
.30123 .29107 m .26679 .30373 L .29222 .34112 L p .761 .767 .945 r
F P
0 g
s
.29222 .34112 m .32647 .32879 L .30123 .29107 L p .761 .767 .945 r
F P
0 g
s
.4482 .54552 m .48332 .5343 L .47195 .48609 L p .713 .588 .786 r
F P
0 g
s
.48332 .5343 m .4482 .54552 L .45572 .59274 L p .699 .548 .748 r
F P
0 g
s
.42073 .60364 m .45572 .59274 L .4482 .54552 L p .699 .548 .748 r
F P
0 g
s
.45572 .59274 m .42073 .60364 L .42406 .64898 L p .682 .502 .703 r
F P
0 g
s
.38915 .6596 m .42406 .64898 L .42073 .60364 L p .682 .502 .703 r
F P
0 g
s
.21207 .67017 m .24596 .65963 L .24457 .62553 L p .677 .495 .699 r
F P
0 g
s
.43948 .39096 m .45724 .438 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.42011 .32877 m .41362 .31159 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.45724 .438 m .42011 .32877 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.46193 .33341 m .45655 .31584 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42011 .32877 m .45724 .438 L .47195 .48609 L p .851 .545 .1 r
F P
0 g
s
.42546 .34624 m .42011 .32877 L p .42133 .33248 L .851 .545 .1 r
F P
0 g
s
.47195 .48609 m .42546 .34624 L p .42133 .33248 L .851 .545 .1 r
F P
0 g
s
.41441 .31401 m .43948 .39096 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.41898 .3459 m .43948 .39096 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.40615 .29511 m .40856 .30463 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.40856 .30463 m .41898 .3459 L p .41115 .30949 L .1 .1 .653 r
F P
0 g
s
.46607 .35084 m .46193 .33341 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.41362 .31159 m .40615 .29511 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.43948 .39096 m .41753 .32357 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.41753 .32357 m .41362 .31159 L p .41367 .31674 L .1 .1 .653 r
F P
0 g
s
.47195 .48609 m .48332 .5343 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.42546 .34624 m .42644 .3492 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.42644 .3492 m .47195 .48609 L p .42661 .34999 L .98 .684 .169 r
F P
0 g
s
.45572 .59274 m .4911 .58172 L .48332 .5343 L p .699 .548 .748 r
F P
0 g
s
.43223 .38037 m .42953 .36359 L p .43058 .36731 L 1 .766 .313 r
F P
0 g
s
.4911 .58172 m .43223 .38037 L p .43058 .36731 L 1 .766 .313 r
F P
0 g
s
.4911 .58172 m .45572 .59274 L .45937 .63824 L p .682 .502 .703 r
F P
0 g
s
.45937 .63824 m .49506 .62738 L .4911 .58172 L p .682 .502 .703 r
F P
0 g
s
.43347 .39618 m .43223 .38037 L p .43316 .38403 L 1 .81 .413 r
F P
0 g
s
.49506 .62738 m .43347 .39618 L p .43316 .38403 L 1 .81 .413 r
F P
0 g
s
.49506 .62738 m .45937 .63824 L .45895 .68104 L p .66 .447 .648 r
F P
0 g
s
.45895 .68104 m .49504 .67031 L .49506 .62738 L p .66 .447 .648 r
F P
0 g
s
.43322 .41062 m .43347 .39618 L p .43428 .39977 L 1 .833 .481 r
F P
0 g
s
.49504 .67031 m .43322 .41062 L p .43428 .39977 L 1 .833 .481 r
F P
0 g
s
.49504 .67031 m .45895 .68104 L .45436 .72015 L p .63 .378 .578 r
F P
0 g
s
.45436 .72015 m .49089 .7095 L .49504 .67031 L p .63 .378 .578 r
F P
0 g
s
.43147 .4233 m .43322 .41062 L p .4339 .41412 L 1 .843 .528 r
F P
0 g
s
.49089 .7095 m .43147 .4233 L p .4339 .41412 L 1 .843 .528 r
F P
0 g
s
.49089 .7095 m .45436 .72015 L .44553 .75455 L p .585 .285 .481 r
F P
0 g
s
.44553 .75455 m .48255 .74393 L .49089 .7095 L p .585 .285 .481 r
F P
0 g
s
.42823 .43388 m .43147 .4233 L p .43201 .4267 L 1 .847 .562 r
F P
0 g
s
.48255 .74393 m .42823 .43388 L p .43201 .4267 L 1 .847 .562 r
F P
0 g
s
.42406 .64898 m .45937 .63824 L .45572 .59274 L p .682 .502 .703 r
F P
0 g
s
.45937 .63824 m .42406 .64898 L .42327 .69166 L p .66 .447 .648 r
F P
0 g
s
.42327 .69166 m .45895 .68104 L .45937 .63824 L p .66 .447 .648 r
F P
0 g
s
.45895 .68104 m .42327 .69166 L .41825 .73068 L p .63 .378 .578 r
F P
0 g
s
.41825 .73068 m .45436 .72015 L .45895 .68104 L p .63 .378 .578 r
F P
0 g
s
.48255 .74393 m .44553 .75455 L .43247 .78321 L p .515 .153 .339 r
F P
0 g
s
.43247 .78321 m .47003 .7726 L .48255 .74393 L p .515 .153 .339 r
F P
0 g
s
.42357 .44206 m .42823 .43388 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42406 .64898 m .38915 .6596 L .38799 .70216 L p .66 .447 .648 r
F P
0 g
s
.38799 .70216 m .42327 .69166 L .42406 .64898 L p .66 .447 .648 r
F P
0 g
s
.45436 .72015 m .41825 .73068 L .40893 .76504 L p .585 .285 .481 r
F P
0 g
s
.40893 .76504 m .44553 .75455 L .45436 .72015 L p .585 .285 .481 r
F P
0 g
s
.44553 .75455 m .40893 .76504 L .39535 .79371 L p .515 .153 .339 r
F P
0 g
s
.39535 .79371 m .43247 .78321 L .44553 .75455 L p .515 .153 .339 r
F P
0 g
s
.43247 .78321 m .39535 .79371 L .37763 .81571 L p .392 .1 .116 r
F P
0 g
s
.47003 .7726 m .43247 .78321 L .4153 .80516 L p .392 .1 .116 r
F P
0 g
s
.37763 .81571 m .4153 .80516 L .43247 .78321 L p .392 .1 .116 r
F P
0 g
s
.4153 .80516 m .45342 .79449 L .47003 .7726 L p .392 .1 .116 r
F P
0 g
s
.41042 .45021 m .41759 .44756 L p .41772 .45064 L 1 .842 .617 r
F P
0 g
s
.45342 .79449 m .4153 .80516 L .39424 .81946 L p .178 .1 .1 r
F P
0 g
s
.39424 .81946 m .43293 .80867 L .45342 .79449 L p .178 .1 .1 r
F P
0 g
s
.43293 .80867 m .41042 .45021 L p .41772 .45064 L 1 .842 .617 r
F P
0 g
s
.41042 .45021 m .40223 .44986 L p .40266 .44975 L .573 .989 1 r
F P
0 g
s
.41759 .44756 m .41042 .45021 L p .41182 .44973 L .1 .1 .1 r
F P
0 g
s
.47003 .7726 m .42401 .44518 L p .42938 .44318 L 1 .847 .586 r
F P
0 g
s
.48332 .5343 m .42953 .36359 L .42546 .34624 L p .98 .684 .169 r
F P
0 g
s
.42953 .36359 m .48332 .5343 L .4911 .58172 L p 1 .766 .313 r
F P
0 g
s
.43223 .38037 m .4911 .58172 L .49506 .62738 L p 1 .81 .413 r
F P
0 g
s
.43347 .39618 m .49506 .62738 L .49504 .67031 L p 1 .833 .481 r
F P
0 g
s
.43322 .41062 m .49504 .67031 L .49089 .7095 L p 1 .843 .528 r
F P
0 g
s
.43147 .4233 m .49089 .7095 L .48255 .74393 L p 1 .847 .562 r
F P
0 g
s
.48255 .74393 m .47003 .7726 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42823 .43388 m .42878 .43704 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42878 .43704 m .48255 .74393 L p .42864 .43718 L 1 .847 .586 r
F P
0 g
s
.42938 .44318 m .42357 .44206 L .42864 .43718 L 1 .847 .586 r
F
.42401 .44518 m .42357 .44206 L p .42938 .44318 L F P
0 g
s
.47003 .7726 m .45342 .79449 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.42357 .44206 m .42486 .45121 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.42486 .45121 m .47003 .7726 L p .42434 .45116 L 1 .845 .604 r
F P
0 g
s
.40266 .44975 m .40357 .4495 L .41042 .45021 L .573 .989 1 r
F
.41759 .44756 m .42357 .44206 L p .42384 .44525 L 1 .845 .604 r
F P
0 g
s
.45342 .79449 m .41759 .44756 L p .42384 .44525 L 1 .845 .604 r
F P
0 g
s
.41759 .44756 m .45342 .79449 L .43293 .80867 L p 1 .842 .617 r
F P
0 g
s
.44823 .4375 m .4552 .43492 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.41042 .45021 m .41079 .45008 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.41079 .45008 m .44823 .4375 L p .41086 .45006 L .1 .1 .1 r
F P
0 g
s
.4552 .43492 m .46099 .42945 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.41759 .44756 m .41799 .44743 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.41799 .44743 m .4552 .43492 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.4552 .43492 m .44474 .43843 L p .44549 .43823 L .1 .1 .1 r
F P
0 g
s
.44474 .43843 m .41799 .44743 L p .44549 .43823 L .1 .1 .1 r
F P
0 g
s
.41182 .44973 m .44549 .43823 L .41799 .44743 L .1 .1 .1 r
F
.41799 .44743 m .41759 .44756 L p .41182 .44973 L F P
0 g
s
.42357 .44206 m .41759 .44756 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.46099 .42945 m .42357 .44206 L p .41805 .44737 L .212 .1 .1 r
F P
0 g
s
.44823 .4375 m .41079 .45008 L p .40357 .4495 L .573 .989 1 r
F P
0 g
s
.41079 .45008 m .41042 .45021 L p .40357 .4495 L .573 .989 1 r
F P
0 g
s
.42823 .43388 m .42357 .44206 L p .42406 .44182 L .42 .1 .21 r
F P
0 g
s
.46547 .42131 m .42823 .43388 L p .42406 .44182 L .42 .1 .21 r
F P
0 g
s
.43147 .4233 m .42823 .43388 L p .42875 .43359 L .535 .21 .414 r
F P
0 g
s
.46853 .41073 m .43147 .4233 L p .42875 .43359 L .535 .21 .414 r
F P
0 g
s
.43322 .41062 m .43147 .4233 L p .43203 .42294 L .601 .335 .543 r
F P
0 g
s
.47012 .39803 m .43322 .41062 L p .43203 .42294 L .601 .335 .543 r
F P
0 g
s
.43347 .39618 m .43322 .41062 L p .43384 .41017 L .642 .422 .63 r
F P
0 g
s
.47023 .38356 m .43347 .39618 L p .43384 .41017 L .642 .422 .63 r
F P
0 g
s
.43223 .38037 m .43347 .39618 L p .43415 .39564 L .67 .487 .693 r
F P
0 g
s
.46886 .3677 m .43223 .38037 L p .43415 .39564 L .67 .487 .693 r
F P
0 g
s
.42953 .36359 m .43223 .38037 L p .43299 .3797 L .691 .539 .743 r
F P
0 g
s
.43299 .3797 m .44387 .37021 L .42953 .36359 L .691 .539 .743 r
F
.46607 .35084 m .43839 .3605 L p .44387 .37021 L F P
0 g
s
.43839 .3605 m .42953 .36359 L p .44387 .37021 L .691 .539 .743 r
F P
0 g
s
.42953 .36359 m .43839 .3605 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.43839 .3605 m .46607 .35084 L p .43655 .35705 L .708 .583 .785 r
F P
0 g
s
.42546 .34624 m .42953 .36359 L p .43041 .36277 L .708 .583 .785 r
F P
0 g
s
.46193 .33341 m .42546 .34624 L p .43041 .36277 L .708 .583 .785 r
F P
0 g
s
.42546 .34624 m .42669 .34581 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42669 .34581 m .46193 .33341 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.42011 .32877 m .42546 .34624 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.45655 .31584 m .42011 .32877 L p .42647 .34525 L .721 .622 .821 r
F P
0 g
s
.41362 .31159 m .41441 .31401 L p .41486 .31516 L .1 .1 .452 r
F P
0 g
s
.24596 .65963 m .21207 .67017 L .20978 .70204 L p .653 .438 .642 r
F P
0 g
s
.35309 .71254 m .31858 .72281 L .31231 .76157 L p .63 .378 .578 r
F P
0 g
s
.31858 .72281 m .24393 .69159 L .23846 .72064 L p .923 .654 .631 r
F P
0 g
s
.23846 .72064 m .31231 .76157 L .31858 .72281 L p .923 .654 .631 r
F P
0 g
s
.39612 .30371 m .3611 .31632 L .3841 .35818 L p .753 .73 .914 r
F P
0 g
s
.3841 .35818 m .41898 .3459 L .39612 .30371 L p .753 .73 .914 r
F P
0 g
s
.35309 .71254 m .38799 .70216 L .38915 .6596 L p .66 .447 .648 r
F P
0 g
s
.23965 .27108 m .20532 .28393 L .23272 .31625 L p .768 .809 .978 r
F P
0 g
s
.23272 .31625 m .26679 .30373 L .23965 .27108 L p .768 .809 .978 r
F P
0 g
s
.33607 .27826 m .30123 .29107 L .32647 .32879 L p .761 .767 .945 r
F P
0 g
s
.32647 .32879 m .3611 .31632 L .33607 .27826 L p .761 .767 .945 r
F P
0 g
s
.13192 .36836 m .09821 .38067 L .11976 .40411 L p .765 .813 .983 r
F P
0 g
s
.11976 .40411 m .15329 .39205 L .13192 .36836 L p .765 .813 .983 r
F P
0 g
s
.20978 .70204 m .24393 .69159 L .24596 .65963 L p .653 .438 .642 r
F P
0 g
s
.38799 .70216 m .35309 .71254 L .34723 .75139 L p .63 .378 .578 r
F P
0 g
s
.31231 .76157 m .34723 .75139 L .35309 .71254 L p .63 .378 .578 r
F P
0 g
s
.39787 .27976 m .39612 .30371 L .41898 .3459 L p .1 .253 .795 r
F P
0 g
s
.41898 .3459 m .40615 .29511 L .39787 .27976 L p .1 .253 .795 r
F P
0 g
s
.27436 .25808 m .23965 .27108 L .26679 .30373 L p .768 .809 .978 r
F P
0 g
s
.26679 .30373 m .30123 .29107 L .27436 .25808 L p .768 .809 .978 r
F P
0 g
s
.24393 .69159 m .20978 .70204 L .20402 .73103 L p .621 .364 .567 r
F P
0 g
s
.17137 .29664 m .13192 .36836 L .15329 .39205 L p .971 .833 .807 r
F P
0 g
s
.37129 .26532 m .33607 .27826 L .3611 .31632 L p .761 .767 .945 r
F P
0 g
s
.3611 .31632 m .39612 .30371 L .37129 .26532 L p .761 .767 .945 r
F P
0 g
s
.42327 .69166 m .38799 .70216 L .38254 .74109 L p .63 .378 .578 r
F P
0 g
s
.34723 .75139 m .38254 .74109 L .38799 .70216 L p .63 .378 .578 r
F P
0 g
s
.30947 .24493 m .27436 .25808 L .30123 .29107 L p .768 .809 .978 r
F P
0 g
s
.30123 .29107 m .33607 .27826 L .30947 .24493 L p .768 .809 .978 r
F P
0 g
s
.20402 .73103 m .23846 .72064 L .24393 .69159 L p .621 .364 .567 r
F P
0 g
s
.34723 .75139 m .31231 .76157 L .30161 .7958 L p .585 .285 .481 r
F P
0 g
s
.31231 .76157 m .23846 .72064 L .22958 .74601 L p .927 .637 .598 r
F P
0 g
s
.22958 .74601 m .30161 .7958 L .31231 .76157 L p .927 .637 .598 r
F P
0 g
s
.38898 .26592 m .37129 .26532 L .39612 .30371 L p .1 .373 .878 r
F P
0 g
s
.39612 .30371 m .39787 .27976 L .38898 .26592 L p .1 .373 .878 r
F P
0 g
s
.23846 .72064 m .20402 .73103 L .19481 .75637 L p .572 .265 .463 r
F P
0 g
s
.38254 .74109 m .41825 .73068 L .42327 .69166 L p .63 .378 .578 r
F P
0 g
s
.34497 .23164 m .30947 .24493 L .33607 .27826 L p .768 .809 .978 r
F P
0 g
s
.33607 .27826 m .37129 .26532 L .34497 .23164 L p .768 .809 .978 r
F P
0 g
s
.38254 .74109 m .34723 .75139 L .33698 .78566 L p .585 .285 .481 r
F P
0 g
s
.30161 .7958 m .33698 .78566 L .34723 .75139 L p .585 .285 .481 r
F P
0 g
s
.19481 .75637 m .22958 .74601 L .23846 .72064 L p .572 .265 .463 r
F P
0 g
s
.3797 .25395 m .34497 .23164 L .37129 .26532 L p .1 .423 .909 r
F P
0 g
s
.37129 .26532 m .38898 .26592 L .3797 .25395 L p .1 .423 .909 r
F P
0 g
s
.41825 .73068 m .38254 .74109 L .37275 .77541 L p .585 .285 .481 r
F P
0 g
s
.33698 .78566 m .37275 .77541 L .38254 .74109 L p .585 .285 .481 r
F P
0 g
s
.33698 .78566 m .30161 .7958 L .28654 .82447 L p .515 .153 .339 r
F P
0 g
s
.30161 .7958 m .22958 .74601 L .21739 .76695 L p .933 .621 .561 r
F P
0 g
s
.21739 .76695 m .28654 .82447 L .30161 .7958 L p .933 .621 .561 r
F P
0 g
s
.22958 .74601 m .19481 .75637 L .18227 .77732 L p .492 .12 .307 r
F P
0 g
s
.37275 .77541 m .40893 .76504 L .41825 .73068 L p .585 .285 .481 r
F P
0 g
s
.37275 .77541 m .33698 .78566 L .32239 .81433 L p .515 .153 .339 r
F P
0 g
s
.28654 .82447 m .32239 .81433 L .33698 .78566 L p .515 .153 .339 r
F P
0 g
s
.18227 .77732 m .21739 .76695 L .22958 .74601 L p .492 .12 .307 r
F P
0 g
s
.45005 .29855 m .41362 .31159 L .42011 .32877 L p .733 .659 .854 r
F P
0 g
s
.42011 .32877 m .45655 .31584 L .45005 .29855 L p .733 .659 .854 r
F P
0 g
s
.40893 .76504 m .37275 .77541 L .35866 .80408 L p .515 .153 .339 r
F P
0 g
s
.32239 .81433 m .35866 .80408 L .37275 .77541 L p .515 .153 .339 r
F P
0 g
s
.4426 .28196 m .40615 .29511 L .41362 .31159 L p .743 .695 .885 r
F P
0 g
s
.41362 .31159 m .45005 .29855 L .4426 .28196 L p .743 .695 .885 r
F P
0 g
s
.21739 .76695 m .18227 .77732 L .1666 .79319 L p .353 .1 .1 r
F P
0 g
s
.43437 .26649 m .39787 .27976 L .40615 .29511 L p .752 .731 .916 r
F P
0 g
s
.40615 .29511 m .4426 .28196 L .43437 .26649 L p .752 .731 .916 r
F P
0 g
s
.43223 .38037 m .46886 .3677 L .46607 .35084 L p .691 .539 .743 r
F P
0 g
s
.01 w
.08602 .49343 m
.08602 .49343 L
s
.09289 .48216 m
.08602 .49343 L
s
.5 Mabswid
.32239 .81433 m .28654 .82447 L .26726 .84661 L p .392 .1 .116 r
F P
0 g
s
.28654 .82447 m .21739 .76695 L .20207 .78278 L p .94 .604 .518 r
F P
0 g
s
.20207 .78278 m .26726 .84661 L .28654 .82447 L p .94 .604 .518 r
F P
0 g
s
.42556 .25251 m .38898 .26592 L .39787 .27976 L p .76 .769 .948 r
F P
0 g
s
.39787 .27976 m .43437 .26649 L .42556 .25251 L p .76 .769 .948 r
F P
0 g
s
.4869 .28536 m .45005 .29855 L .45655 .31584 L p .733 .659 .854 r
F P
0 g
s
.45655 .31584 m .4934 .30277 L .4869 .28536 L p .733 .659 .854 r
F P
0 g
s
.4934 .30277 m .45655 .31584 L .46193 .33341 L p .721 .622 .821 r
F P
0 g
s
.46193 .33341 m .49883 .32044 L .4934 .30277 L p .721 .622 .821 r
F P
0 g
s
.49883 .32044 m .46193 .33341 L .46607 .35084 L p .708 .583 .785 r
F P
0 g
s
.35866 .80408 m .39535 .79371 L .40893 .76504 L p .515 .153 .339 r
F P
0 g
s
.47947 .26866 m .4426 .28196 L .45005 .29855 L p .743 .695 .885 r
F P
0 g
s
.45005 .29855 m .4869 .28536 L .47947 .26866 L p .743 .695 .885 r
F P
0 g
s
.43347 .39618 m .47023 .38356 L .46886 .3677 L p .67 .487 .693 r
F P
0 g
s
.1666 .79319 m .20207 .78278 L .21739 .76695 L p .353 .1 .1 r
F P
0 g
s
.46607 .35084 m .50304 .33795 L .49883 .32044 L p .708 .583 .785 r
F P
0 g
s
.50304 .33795 m .46607 .35084 L .46886 .3677 L p .691 .539 .743 r
F P
0 g
s
.4713 .25305 m .43437 .26649 L .4426 .28196 L p .752 .731 .916 r
F P
0 g
s
.4426 .28196 m .47947 .26866 L .4713 .25305 L p .752 .731 .916 r
F P
0 g
s
.41639 .24041 m .3797 .25395 L .38898 .26592 L p .766 .811 .981 r
F P
0 g
s
.38898 .26592 m .42556 .25251 L .41639 .24041 L p .766 .811 .981 r
F P
0 g
s
.01 w
.10074 .46722 m
.09289 .48216 L
s
.5 Mabswid
.46886 .3677 m .50592 .35487 L .50304 .33795 L p .691 .539 .743 r
F P
0 g
s
.50592 .35487 m .46886 .3677 L .47023 .38356 L p .67 .487 .693 r
F P
0 g
s
.43322 .41062 m .47012 .39803 L .47023 .38356 L p .642 .422 .63 r
F P
0 g
s
.35866 .80408 m .32239 .81433 L .30362 .83643 L p .392 .1 .116 r
F P
0 g
s
.26726 .84661 m .30362 .83643 L .32239 .81433 L p .392 .1 .116 r
F P
0 g
s
.46257 .23895 m .42556 .25251 L .43437 .26649 L p .76 .769 .948 r
F P
0 g
s
.43437 .26649 m .4713 .25305 L .46257 .23895 L p .76 .769 .948 r
F P
0 g
s
.52418 .27202 m .4869 .28536 L .4934 .30277 L p .733 .659 .854 r
F P
0 g
s
.4934 .30277 m .53069 .28954 L .52418 .27202 L p .733 .659 .854 r
F P
0 g
s
.53069 .28954 m .4934 .30277 L .49883 .32044 L p .721 .622 .821 r
F P
0 g
s
.49883 .32044 m .53615 .30731 L .53069 .28954 L p .721 .622 .821 r
F P
0 g
s
.53615 .30731 m .49883 .32044 L .50304 .33795 L p .708 .583 .785 r
F P
0 g
s
.51678 .2552 m .47947 .26866 L .4869 .28536 L p .743 .695 .885 r
F P
0 g
s
.4869 .28536 m .52418 .27202 L .51678 .2552 L p .743 .695 .885 r
F P
0 g
s
.47023 .38356 m .50741 .37079 L .50592 .35487 L p .67 .487 .693 r
F P
0 g
s
.50741 .37079 m .47023 .38356 L .47012 .39803 L p .642 .422 .63 r
F P
0 g
s
.01 w
.11252 .45146 m
.10074 .46722 L
s
.5 Mabswid
.50304 .33795 m .54043 .32491 L .53615 .30731 L p .708 .583 .785 r
F P
0 g
s
.54043 .32491 m .50304 .33795 L .50592 .35487 L p .691 .539 .743 r
F P
0 g
s
.20207 .78278 m .1666 .79319 L .14808 .80337 L p .114 .1 .1 r
F P
0 g
s
.01 w
.12883 .44069 m
.11252 .45146 L
s
.5 Mabswid
.50865 .23947 m .4713 .25305 L .47947 .26866 L p .752 .731 .916 r
F P
0 g
s
.47947 .26866 m .51678 .2552 L .50865 .23947 L p .752 .731 .916 r
F P
0 g
s
.01 w
.14661 .43422 m
.12883 .44069 L
s
.5 Mabswid
.4535 .22672 m .41639 .24041 L .42556 .25251 L p .766 .811 .981 r
F P
0 g
s
.42556 .25251 m .46257 .23895 L .4535 .22672 L p .766 .811 .981 r
F P
0 g
s
.50592 .35487 m .54342 .3419 L .54043 .32491 L p .691 .539 .743 r
F P
0 g
s
.54342 .3419 m .50592 .35487 L .50741 .37079 L p .67 .487 .693 r
F P
0 g
s
.43147 .4233 m .46853 .41073 L .47012 .39803 L p .601 .335 .543 r
F P
0 g
s
.01 w
.16485 .42775 m
.14661 .43422 L
s
.5 Mabswid
.47012 .39803 m .50745 .38529 L .50741 .37079 L p .642 .422 .63 r
F P
0 g
s
.50745 .38529 m .47012 .39803 L .46853 .41073 L p .601 .335 .543 r
F P
0 g
s
.39535 .79371 m .35866 .80408 L .3404 .82613 L p .392 .1 .116 r
F P
0 g
s
.30362 .83643 m .3404 .82613 L .35866 .80408 L p .392 .1 .116 r
F P
0 g
s
.50001 .22523 m .46257 .23895 L .4713 .25305 L p .76 .769 .948 r
F P
0 g
s
.4713 .25305 m .50865 .23947 L .50001 .22523 L p .76 .769 .948 r
F P
0 g
s
.56189 .25853 m .52418 .27202 L .53069 .28954 L p .733 .659 .854 r
F P
0 g
s
.53069 .28954 m .56841 .27616 L .56189 .25853 L p .733 .659 .854 r
F P
0 g
s
.56841 .27616 m .53069 .28954 L .53615 .30731 L p .721 .622 .821 r
F P
0 g
s
.53615 .30731 m .57391 .29403 L .56841 .27616 L p .721 .622 .821 r
F P
0 g
s
.57391 .29403 m .53615 .30731 L .54043 .32491 L p .708 .583 .785 r
F P
0 g
s
.55451 .24158 m .51678 .2552 L .52418 .27202 L p .743 .695 .885 r
F P
0 g
s
.52418 .27202 m .56189 .25853 L .55451 .24158 L p .743 .695 .885 r
F P
0 g
s
.50741 .37079 m .54503 .35787 L .54342 .3419 L p .67 .487 .693 r
F P
0 g
s
.54503 .35787 m .50741 .37079 L .50745 .38529 L p .642 .422 .63 r
F P
0 g
s
.01 w
.18468 .41973 m
.16485 .42775 L
s
.5 Mabswid
.54043 .32491 m .57826 .31171 L .57391 .29403 L p .708 .583 .785 r
F P
0 g
s
.57826 .31171 m .54043 .32491 L .54342 .3419 L p .691 .539 .743 r
F P
0 g
s
.26726 .84661 m .20207 .78278 L .18392 .79288 L p .948 .588 .469 r
F P
0 g
s
.14808 .80337 m .18392 .79288 L p .19802 .78433 L .114 .1 .1 r
F P
0 g
s
.18392 .79288 m .14808 .80337 L .12712 .80735 L p .39 .887 .881 r
F P
0 g
s
.16332 .79674 m .18392 .79288 L p .17966 .79397 L .39 .887 .881 r
F P
0 g
s
.12712 .80735 m .16332 .79674 L p .17966 .79397 L .39 .887 .881 r
F P
0 g
s
.16332 .79674 m .12712 .80735 L .10422 .80476 L p .605 .985 1 r
F P
0 g
s
.14075 .794 m .16332 .79674 L p .15888 .79735 L .605 .985 1 r
F P
0 g
s
.10422 .80476 m .14075 .794 L p .15888 .79735 L .605 .985 1 r
F P
0 g
s
.11682 .78448 m .14075 .794 L p .13618 .79411 L .699 .958 1 r
F P
0 g
s
.07998 .79543 m .11682 .78448 L p .13618 .79411 L .699 .958 1 r
F P
0 g
s
.05509 .77939 m .0922 .76821 L .11682 .78448 L p .733 .899 1 r
F P
0 g
s
.11682 .78448 m .0922 .76821 L p .09699 .77461 L .968 .523 .206 r
F P
0 g
s
.14075 .794 m .10422 .80476 L .07998 .79543 L p .699 .958 1 r
F P
0 g
s
.11682 .78448 m .07998 .79543 L .05509 .77939 L p .733 .899 1 r
F P
0 g
s
.15598 .85356 m .11682 .78448 L p .09699 .77461 L .968 .523 .206 r
F P
0 g
s
.14075 .794 m .11682 .78448 L p .12217 .79057 L .969 .538 .279 r
F P
0 g
s
.18775 .8653 m .14075 .794 L p .12217 .79057 L .969 .538 .279 r
F P
0 g
s
.16332 .79674 m .14075 .794 L p .14656 .79959 L .964 .554 .349 r
F P
0 g
s
.21736 .86775 m .16332 .79674 L p .14656 .79959 L .964 .554 .349 r
F P
0 g
s
.18392 .79288 m .16332 .79674 L p .16946 .80166 L .957 .571 .413 r
F P
0 g
s
.24406 .86131 m .18392 .79288 L p .16946 .80166 L .957 .571 .413 r
F P
0 g
s
.18392 .79288 m .20207 .78278 L p .19802 .78433 L .114 .1 .1 r
F P
0 g
s
.54645 .22572 m .50865 .23947 L .51678 .2552 L p .752 .731 .916 r
F P
0 g
s
.51678 .2552 m .55451 .24158 L .54645 .22572 L p .752 .731 .916 r
F P
0 g
s
.42823 .43388 m .46547 .42131 L .46853 .41073 L p .535 .21 .414 r
F P
0 g
s
.49106 .21287 m .4535 .22672 L .46257 .23895 L p .766 .811 .981 r
F P
0 g
s
.46257 .23895 m .50001 .22523 L .49106 .21287 L p .766 .811 .981 r
F P
0 g
s
.54342 .3419 m .58135 .32878 L .57826 .31171 L p .691 .539 .743 r
F P
0 g
s
.58135 .32878 m .54342 .3419 L .54503 .35787 L p .67 .487 .693 r
F P
0 g
s
.46853 .41073 m .50602 .39801 L .50745 .38529 L p .601 .335 .543 r
F P
0 g
s
.50602 .39801 m .46853 .41073 L .46547 .42131 L p .535 .21 .414 r
F P
0 g
s
.30362 .83643 m .26726 .84661 L .24406 .86131 L p .178 .1 .1 r
F P
0 g
s
.18392 .79288 m .24406 .86131 L .26726 .84661 L p .948 .588 .469 r
F P
0 g
s
.50745 .38529 m .54522 .37241 L .54503 .35787 L p .642 .422 .63 r
F P
0 g
s
.54522 .37241 m .50745 .38529 L .50602 .39801 L p .601 .335 .543 r
F P
0 g
s
.01 w
.20744 .40968 m
.18468 .41973 L
s
.5 Mabswid
.3404 .82613 m .37763 .81571 L .39535 .79371 L p .392 .1 .116 r
F P
0 g
s
.53789 .21135 m .50001 .22523 L .50865 .23947 L p .76 .769 .948 r
F P
0 g
s
.50865 .23947 m .54645 .22572 L .53789 .21135 L p .76 .769 .948 r
F P
0 g
s
.60005 .24488 m .56189 .25853 L .56841 .27616 L p .733 .659 .854 r
F P
0 g
s
.56841 .27616 m .60657 .26262 L .60005 .24488 L p .733 .659 .854 r
F P
0 g
s
.60657 .26262 m .56841 .27616 L .57391 .29403 L p .721 .622 .821 r
F P
0 g
s
.57391 .29403 m .61211 .28059 L .60657 .26262 L p .721 .622 .821 r
F P
0 g
s
.61211 .28059 m .57391 .29403 L .57826 .31171 L p .708 .583 .785 r
F P
0 g
s
.59269 .2278 m .55451 .24158 L .56189 .25853 L p .743 .695 .885 r
F P
0 g
s
.56189 .25853 m .60005 .24488 L .59269 .2278 L p .743 .695 .885 r
F P
0 g
s
.54503 .35787 m .58309 .3448 L .58135 .32878 L p .67 .487 .693 r
F P
0 g
s
.58309 .3448 m .54503 .35787 L .54522 .37241 L p .642 .422 .63 r
F P
0 g
s
.57826 .31171 m .61653 .29836 L .61211 .28059 L p .708 .583 .785 r
F P
0 g
s
.61653 .29836 m .57826 .31171 L .58135 .32878 L p .691 .539 .743 r
F P
0 g
s
.58469 .21181 m .54645 .22572 L .55451 .24158 L p .752 .731 .916 r
F P
0 g
s
.55451 .24158 m .59269 .2278 L .58469 .21181 L p .752 .731 .916 r
F P
0 g
s
.42357 .44206 m .46099 .42945 L .46547 .42131 L p .42 .1 .21 r
F P
0 g
s
.46547 .42131 m .50314 .40858 L .50602 .39801 L p .535 .21 .414 r
F P
0 g
s
.50314 .40858 m .46547 .42131 L .46099 .42945 L p .42 .1 .21 r
F P
0 g
s
.52905 .19885 m .49106 .21287 L .50001 .22523 L p .766 .811 .981 r
F P
0 g
s
.50001 .22523 m .53789 .21135 L .52905 .19885 L p .766 .811 .981 r
F P
0 g
s
.58135 .32878 m .61972 .3155 L .61653 .29836 L p .691 .539 .743 r
F P
0 g
s
.61972 .3155 m .58135 .32878 L .58309 .3448 L p .67 .487 .693 r
F P
0 g
s
.50602 .39801 m .54396 .38514 L .54522 .37241 L p .601 .335 .543 r
F P
0 g
s
.54396 .38514 m .50602 .39801 L .50314 .40858 L p .535 .21 .414 r
F P
0 g
s
.01 w
.23392 .3961 m
.20744 .40968 L
s
.5 Mabswid
.3404 .82613 m .30362 .83643 L .28093 .85103 L p .178 .1 .1 r
F P
0 g
s
.24406 .86131 m .28093 .85103 L .30362 .83643 L p .178 .1 .1 r
F P
0 g
s
.54522 .37241 m .58343 .35937 L .58309 .3448 L p .642 .422 .63 r
F P
0 g
s
.58343 .35937 m .54522 .37241 L .54396 .38514 L p .601 .335 .543 r
F P
0 g
s
.57622 .1973 m .53789 .21135 L .54645 .22572 L p .76 .769 .948 r
F P
0 g
s
.54645 .22572 m .58469 .21181 L .57622 .1973 L p .76 .769 .948 r
F P
0 g
s
.63865 .23106 m .60005 .24488 L .60657 .26262 L p .733 .659 .854 r
F P
0 g
s
.60657 .26262 m .64518 .24892 L .63865 .23106 L p .733 .659 .854 r
F P
0 g
s
.64518 .24892 m .60657 .26262 L .61211 .28059 L p .721 .622 .821 r
F P
0 g
s
.61211 .28059 m .65077 .267 L .64518 .24892 L p .721 .622 .821 r
F P
0 g
s
.65077 .267 m .61211 .28059 L .61653 .29836 L p .708 .583 .785 r
F P
0 g
s
.63133 .21386 m .59269 .2278 L .60005 .24488 L p .743 .695 .885 r
F P
0 g
s
.60005 .24488 m .63865 .23106 L .63133 .21386 L p .743 .695 .885 r
F P
0 g
s
.58309 .3448 m .6216 .33157 L .61972 .3155 L p .67 .487 .693 r
F P
0 g
s
.6216 .33157 m .58309 .3448 L .58343 .35937 L p .642 .422 .63 r
F P
0 g
s
.61653 .29836 m .65526 .28486 L .65077 .267 L p .708 .583 .785 r
F P
0 g
s
.65526 .28486 m .61653 .29836 L .61972 .3155 L p .691 .539 .743 r
F P
0 g
s
.62338 .19773 m .58469 .21181 L .59269 .2278 L p .752 .731 .916 r
F P
0 g
s
.59269 .2278 m .63133 .21386 L .62338 .19773 L p .752 .731 .916 r
F P
0 g
s
.46099 .42945 m .49886 .4167 L .50314 .40858 L p .42 .1 .21 r
F P
0 g
s
.49886 .4167 m .46099 .42945 L .4552 .43492 L p .212 .1 .1 r
F P
0 g
s
.50314 .40858 m .54127 .3957 L .54396 .38514 L p .535 .21 .414 r
F P
0 g
s
.54127 .3957 m .50314 .40858 L .49886 .4167 L p .42 .1 .21 r
F P
0 g
s
.5675 .18467 m .52905 .19885 L .53789 .21135 L p .766 .811 .981 r
F P
0 g
s
.53789 .21135 m .57622 .1973 L .5675 .18467 L p .766 .811 .981 r
F P
0 g
s
.61972 .3155 m .65856 .30206 L .65526 .28486 L p .691 .539 .743 r
F P
0 g
s
.65856 .30206 m .61972 .3155 L .6216 .33157 L p .67 .487 .693 r
F P
0 g
s
.54396 .38514 m .58235 .37211 L .58343 .35937 L p .601 .335 .543 r
F P
0 g
s
.58235 .37211 m .54396 .38514 L .54127 .3957 L p .535 .21 .414 r
F P
0 g
s
.01 w
.26519 .37146 m
.23392 .3961 L
s
.5 Mabswid
.37763 .81571 m .3404 .82613 L .31825 .84064 L p .178 .1 .1 r
F P
0 g
s
.28093 .85103 m .31825 .84064 L .3404 .82613 L p .178 .1 .1 r
F P
0 g
s
.58343 .35937 m .6221 .34618 L .6216 .33157 L p .642 .422 .63 r
F P
0 g
s
.6221 .34618 m .58343 .35937 L .58235 .37211 L p .601 .335 .543 r
F P
0 g
s
.61501 .18309 m .57622 .1973 L .58469 .21181 L p .76 .769 .948 r
F P
0 g
s
.58469 .21181 m .62338 .19773 L .61501 .18309 L p .76 .769 .948 r
F P
0 g
s
.67772 .21708 m .63865 .23106 L .64518 .24892 L p .733 .659 .854 r
F P
0 g
s
.64518 .24892 m .68426 .23506 L .67772 .21708 L p .733 .659 .854 r
F P
0 g
s
.68426 .23506 m .64518 .24892 L .65077 .267 L p .721 .622 .821 r
F P
0 g
s
.65077 .267 m .68988 .25324 L .68426 .23506 L p .721 .622 .821 r
F P
0 g
s
.68988 .25324 m .65077 .267 L .65526 .28486 L p .708 .583 .785 r
F P
0 g
s
.67042 .19975 m .63133 .21386 L .63865 .23106 L p .743 .695 .885 r
F P
0 g
s
.63865 .23106 m .67772 .21708 L .67042 .19975 L p .743 .695 .885 r
F P
0 g
s
.6216 .33157 m .66057 .31819 L .65856 .30206 L p .67 .487 .693 r
F P
0 g
s
.66057 .31819 m .6216 .33157 L .6221 .34618 L p .642 .422 .63 r
F P
0 g
s
.65526 .28486 m .69445 .27119 L .68988 .25324 L p .708 .583 .785 r
F P
0 g
s
.69445 .27119 m .65526 .28486 L .65856 .30206 L p .691 .539 .743 r
F P
0 g
s
.66254 .18349 m .62338 .19773 L .63133 .21386 L p .752 .731 .916 r
F P
0 g
s
.63133 .21386 m .67042 .19975 L .66254 .18349 L p .752 .731 .916 r
F P
0 g
s
.49886 .4167 m .53718 .4038 L .54127 .3957 L p .42 .1 .21 r
F P
0 g
s
.53718 .4038 m .49886 .4167 L .49327 .42212 L p .212 .1 .1 r
F P
0 g
s
.4552 .43492 m .49327 .42212 L .49886 .4167 L p .212 .1 .1 r
F P
0 g
s
.54127 .3957 m .57984 .38267 L .58235 .37211 L p .535 .21 .414 r
F P
0 g
s
.57984 .38267 m .54127 .3957 L .53718 .4038 L p .42 .1 .21 r
F P
0 g
s
.60641 .17031 m .5675 .18467 L .57622 .1973 L p .766 .811 .981 r
F P
0 g
s
.57622 .1973 m .61501 .18309 L .60641 .17031 L p .766 .811 .981 r
F P
0 g
s
.49327 .42212 m .4552 .43492 L .44823 .4375 L p .1 .1 .1 r
F P
0 g
s
.4153 .80516 m .37763 .81571 L .35601 .83011 L p .178 .1 .1 r
F P
0 g
s
.35601 .83011 m .39424 .81946 L .4153 .80516 L p .178 .1 .1 r
F P
0 g
s
.41042 .45021 m .43293 .80867 L .40893 .81434 L p 1 .838 .627 r
F P
0 g
s
.40893 .81434 m .40223 .44986 L .41042 .45021 L p 1 .838 .627 r
F P
0 g
s
.65856 .30206 m .69785 .28847 L .69445 .27119 L p .691 .539 .743 r
F P
0 g
s
.69785 .28847 m .65856 .30206 L .66057 .31819 L p .67 .487 .693 r
F P
0 g
s
.58235 .37211 m .62119 .35893 L .6221 .34618 L p .601 .335 .543 r
F P
0 g
s
.62119 .35893 m .58235 .37211 L .57984 .38267 L p .535 .21 .414 r
F P
0 g
s
.31825 .84064 m .35601 .83011 L .37763 .81571 L p .178 .1 .1 r
F P
0 g
s
.6221 .34618 m .66122 .33283 L .66057 .31819 L p .642 .422 .63 r
F P
0 g
s
.66122 .33283 m .6221 .34618 L .62119 .35893 L p .601 .335 .543 r
F P
0 g
s
.16332 .79674 m .21736 .86775 L .24406 .86131 L p .957 .571 .413 r
F P
0 g
s
.28093 .85103 m .24406 .86131 L .21736 .86775 L p .322 .837 .818 r
F P
0 g
s
.65426 .16871 m .61501 .18309 L .62338 .19773 L p .76 .769 .948 r
F P
0 g
s
.62338 .19773 m .66254 .18349 L .65426 .16871 L p .76 .769 .948 r
F P
0 g
s
.70451 .21523 m .67772 .21708 L .68426 .23506 L p .247 .358 .822 r
F P
0 g
s
.68426 .23506 m .70796 .22473 L .70451 .21523 L p .247 .358 .822 r
F P
0 g
s
.70796 .22473 m .68426 .23506 L .68988 .25324 L p .228 .307 .778 r
F P
0 g
s
.68988 .25324 m .71093 .23434 L .70796 .22473 L p .228 .307 .778 r
F P
0 g
s
.71093 .23434 m .68988 .25324 L .69445 .27119 L p .19 .242 .723 r
F P
0 g
s
.70069 .20608 m .67042 .19975 L .67772 .21708 L p .248 .396 .859 r
F P
0 g
s
.67772 .21708 m .70451 .21523 L .70069 .20608 L p .248 .396 .859 r
F P
0 g
s
.66057 .31819 m .7 .30465 L .69785 .28847 L p .67 .487 .693 r
F P
0 g
s
.7 .30465 m .66057 .31819 L .66122 .33283 L p .642 .422 .63 r
F P
0 g
s
.69445 .27119 m .71336 .24381 L .71093 .23434 L p .19 .242 .723 r
F P
0 g
s
.71336 .24381 m .69445 .27119 L .69785 .28847 L p .129 .157 .654 r
F P
0 g
s
.01 w
.31147 .32721 m
.26519 .37146 L
s
.5 Mabswid
.69658 .19752 m .66254 .18349 L .67042 .19975 L p .23 .423 .89 r
F P
0 g
s
.67042 .19975 m .70069 .20608 L .69658 .19752 L p .23 .423 .89 r
F P
0 g
s
.53718 .4038 m .57596 .39074 L .57984 .38267 L p .42 .1 .21 r
F P
0 g
s
.57596 .39074 m .53718 .4038 L .53179 .40917 L p .212 .1 .1 r
F P
0 g
s
.49327 .42212 m .53179 .40917 L .53718 .4038 L p .212 .1 .1 r
F P
0 g
s
.57984 .38267 m .61888 .36948 L .62119 .35893 L p .535 .21 .414 r
F P
0 g
s
.61888 .36948 m .57984 .38267 L .57596 .39074 L p .42 .1 .21 r
F P
0 g
s
.64579 .15579 m .60641 .17031 L .61501 .18309 L p .766 .811 .981 r
F P
0 g
s
.61501 .18309 m .65426 .16871 L .64579 .15579 L p .766 .811 .981 r
F P
0 g
s
.53179 .40917 m .49327 .42212 L .48649 .42463 L p .1 .1 .1 r
F P
0 g
s
.44823 .4375 m .48649 .42463 L .49327 .42212 L p .1 .1 .1 r
F P
0 g
s
.48649 .42463 m .44823 .4375 L .44023 .43707 L p .573 .989 1 r
F P
0 g
s
.40223 .44986 m .44023 .43707 L .44823 .4375 L p .573 .989 1 r
F P
0 g
s
.69785 .28847 m .71518 .25289 L .71336 .24381 L p .129 .157 .654 r
F P
0 g
s
.71518 .25289 m .69785 .28847 L .7 .30465 L p .1 .1 .562 r
F P
0 g
s
.62119 .35893 m .6605 .3456 L .66122 .33283 L p .601 .335 .543 r
F P
0 g
s
.6605 .3456 m .62119 .35893 L .61888 .36948 L p .535 .21 .414 r
F P
0 g
s
.44023 .43707 m .40223 .44986 L .39325 .44648 L p .701 .989 1 r
F P
0 g
s
.40223 .44986 m .40893 .81434 L .3819 .81083 L p 1 .833 .635 r
F P
0 g
s
.3819 .81083 m .39325 .44648 L .40223 .44986 L p 1 .833 .635 r
F P
0 g
s
.66122 .33283 m .70082 .31932 L .7 .30465 L p .642 .422 .63 r
F P
0 g
s
.70082 .31932 m .66122 .33283 L .6605 .3456 L p .601 .335 .543 r
F P
0 g
s
.31825 .84064 m .28093 .85103 L .25474 .85733 L p .322 .837 .818 r
F P
0 g
s
.21736 .86775 m .25474 .85733 L .28093 .85103 L p .322 .837 .818 r
F P
0 g
s
.69229 .18976 m .65426 .16871 L .66254 .18349 L p .189 .437 .911 r
F P
0 g
s
.66254 .18349 m .69658 .19752 L .69229 .18976 L p .189 .437 .911 r
F P
0 g
s
.7 .30465 m .71634 .26136 L .71518 .25289 L p .1 .1 .562 r
F P
0 g
s
.71634 .26136 m .7 .30465 L .70082 .31932 L p .1 .1 .436 r
F P
0 g
s
.57596 .39074 m .61521 .37752 L .61888 .36948 L p .42 .1 .21 r
F P
0 g
s
.61521 .37752 m .57596 .39074 L .57078 .39606 L p .212 .1 .1 r
F P
0 g
s
.53179 .40917 m .57078 .39606 L .57596 .39074 L p .212 .1 .1 r
F P
0 g
s
.61888 .36948 m .65839 .35614 L .6605 .3456 L p .535 .21 .414 r
F P
0 g
s
.65839 .35614 m .61888 .36948 L .61521 .37752 L p .42 .1 .21 r
F P
0 g
s
.68793 .18303 m .64579 .15579 L .65426 .16871 L p .118 .434 .917 r
F P
0 g
s
.65426 .16871 m .69229 .18976 L .68793 .18303 L p .118 .434 .917 r
F P
0 g
s
.57078 .39606 m .53179 .40917 L .52522 .41161 L p .1 .1 .1 r
F P
0 g
s
.48649 .42463 m .52522 .41161 L .53179 .40917 L p .1 .1 .1 r
F P
0 g
s
.01 w
.39591 .28068 m
.31147 .32721 L
s
.5 Mabswid
.52522 .41161 m .48649 .42463 L .47869 .42412 L p .573 .989 1 r
F P
0 g
s
.44023 .43707 m .47869 .42412 L .48649 .42463 L p .573 .989 1 r
F P
0 g
s
.6605 .3456 m .70028 .3321 L .70082 .31932 L p .601 .335 .543 r
F P
0 g
s
.70028 .3321 m .6605 .3456 L .65839 .35614 L p .535 .21 .414 r
F P
0 g
s
.47869 .42412 m .44023 .43707 L .43143 .43359 L p .701 .989 1 r
F P
0 g
s
.39325 .44648 m .43143 .43359 L .44023 .43707 L p .701 .989 1 r
F P
0 g
s
.43143 .43359 m .39325 .44648 L .3837 .44011 L p .745 .934 1 r
F P
0 g
s
.39325 .44648 m .3819 .81083 L .35251 .79774 L p 1 .828 .64 r
F P
0 g
s
.35251 .79774 m .3837 .44011 L .39325 .44648 L p 1 .828 .64 r
F P
0 g
s
.70082 .31932 m .7168 .26898 L .71634 .26136 L p .1 .1 .436 r
F P
0 g
s
.7168 .26898 m .70082 .31932 L .70028 .3321 L p .1 .1 .262 r
F P
0 g
s
.35601 .83011 m .31825 .84064 L .29258 .84678 L p .322 .837 .818 r
F P
0 g
s
.25474 .85733 m .29258 .84678 L .31825 .84064 L p .322 .837 .818 r
F P
0 g
s
.01 w
.54011 .2423 m
.39591 .28068 L
s
.5 Mabswid
.61521 .37752 m .65493 .36414 L .65839 .35614 L p .42 .1 .21 r
F P
0 g
s
.65493 .36414 m .61521 .37752 L .61024 .38279 L p .212 .1 .1 r
F P
0 g
s
.57078 .39606 m .61024 .38279 L .61521 .37752 L p .212 .1 .1 r
F P
0 g
s
.65839 .35614 m .69838 .34263 L .70028 .3321 L p .535 .21 .414 r
F P
0 g
s
.69838 .34263 m .65839 .35614 L .65493 .36414 L p .42 .1 .21 r
F P
0 g
s
.61024 .38279 m .57078 .39606 L .56442 .39844 L p .1 .1 .1 r
F P
0 g
s
.52522 .41161 m .56442 .39844 L .57078 .39606 L p .1 .1 .1 r
F P
0 g
s
.56442 .39844 m .52522 .41161 L .51762 .41102 L p .573 .989 1 r
F P
0 g
s
.47869 .42412 m .51762 .41102 L .52522 .41161 L p .573 .989 1 r
F P
0 g
s
.3837 .44011 m .42204 .42711 L .43143 .43359 L p .745 .934 1 r
F P
0 g
s
.42204 .42711 m .3837 .44011 L .37386 .43088 L p .757 .875 1 r
F P
0 g
s
.3837 .44011 m .35251 .79774 L .32158 .775 L p 1 .823 .644 r
F P
0 g
s
.32158 .775 m .37386 .43088 L .3837 .44011 L p 1 .823 .644 r
F P
0 g
s
.70028 .3321 m .71656 .27556 L .7168 .26898 L p .1 .1 .262 r
F P
0 g
s
.71656 .27556 m .70028 .3321 L .69838 .34263 L p .1 .1 .1 r
F P
0 g
s
.51762 .41102 m .47869 .42412 L .47007 .42054 L p .701 .989 1 r
F P
0 g
s
.43143 .43359 m .47007 .42054 L .47869 .42412 L p .701 .989 1 r
F P
0 g
s
.47007 .42054 m .43143 .43359 L .42204 .42711 L p .745 .934 1 r
F P
0 g
s
.14075 .794 m .18775 .8653 L .21736 .86775 L p .964 .554 .349 r
F P
0 g
s
.25474 .85733 m .21736 .86775 L .18775 .8653 L p .564 .975 1 r
F P
0 g
s
.39424 .81946 m .35601 .83011 L .33089 .8361 L p .322 .837 .818 r
F P
0 g
s
.29258 .84678 m .33089 .8361 L .35601 .83011 L p .322 .837 .818 r
F P
0 g
s
.37386 .43088 m .41234 .41775 L .42204 .42711 L p .757 .875 1 r
F P
0 g
s
.65493 .36414 m .69514 .3506 L .69838 .34263 L p .42 .1 .21 r
F P
0 g
s
.69514 .3506 m .65493 .36414 L .65018 .36936 L p .212 .1 .1 r
F P
0 g
s
.61024 .38279 m .65018 .36936 L .65493 .36414 L p .212 .1 .1 r
F P
0 g
s
.69838 .34263 m .71561 .28091 L p .716 .27763 L .1 .1 .1 r
F P
0 g
s
.71561 .28091 m .69838 .34263 L .69514 .3506 L p .871 .898 .409 r
F P
0 g
s
.71398 .28488 m .71561 .28091 L p .71498 .28305 L .871 .898 .409 r
F P
0 g
s
.69514 .3506 m .71398 .28488 L p .71498 .28305 L .871 .898 .409 r
F P
0 g
s
.65018 .36936 m .69061 .35577 L .69514 .3506 L p .212 .1 .1 r
F P
0 g
s
.71169 .28734 m .71398 .28488 L p .71326 .28705 L 1 1 .616 r
F P
0 g
s
.69061 .35577 m .71169 .28734 L p .71326 .28705 L 1 1 .616 r
F P
0 g
s
.71169 .28734 m .7088 .28823 L p .7101 .28764 L .472 .927 .828 r
F P
0 g
s
.71398 .28488 m .71169 .28734 L p .71296 .28671 L .1 .1 .1 r
F P
0 g
s
.71398 .28488 m .69514 .3506 L .69061 .35577 L p 1 1 .616 r
F P
0 g
s
.75361 .26659 m .71398 .28488 L p .71296 .28671 L .1 .1 .1 r
F P
0 g
s
.7516 .26885 m .75361 .26659 L p .72224 .28212 L .1 .1 .1 r
F P
0 g
s
.71169 .28734 m .7516 .26885 L p .72224 .28212 L .1 .1 .1 r
F P
0 g
s
.79398 .24855 m .75361 .26659 L .7516 .26885 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .79398 .24855 L p .76153 .26409 L .1 .1 .1 r
F P
0 g
s
.7516 .26885 m .7922 .25062 L p .76153 .26409 L .1 .1 .1 r
F P
0 g
s
.83505 .23069 m .79398 .24855 L .7922 .25062 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .83349 .23258 L .83505 .23069 L p .1 .1 .1 r
F P
0 g
s
.7922 .25062 m .78991 .25129 L p .82322 .23699 L .447 .912 .822 r
F P
0 g
s
.83349 .23258 m .7922 .25062 L p .82322 .23699 L .447 .912 .822 r
F P
0 g
s
.7516 .26885 m .74903 .26962 L p .78128 .25542 L .458 .919 .825 r
F P
0 g
s
.7922 .25062 m .7516 .26885 L p .78128 .25542 L .458 .919 .825 r
F P
0 g
s
.7516 .26885 m .71169 .28734 L p .7101 .28764 L .472 .927 .828 r
F P
0 g
s
.71561 .28091 m .71398 .28488 L p .71522 .28421 L .313 .1 .146 r
F P
0 g
s
.75503 .26294 m .71561 .28091 L p .71522 .28421 L .313 .1 .146 r
F P
0 g
s
.71561 .28091 m .71656 .27556 L p .716 .27763 L .1 .1 .1 r
F P
0 g
s
.65018 .36936 m .61024 .38279 L .60409 .3851 L p .1 .1 .1 r
F P
0 g
s
.56442 .39844 m .60409 .3851 L .61024 .38279 L p .1 .1 .1 r
F P
0 g
s
.60409 .3851 m .56442 .39844 L .55702 .39775 L p .573 .989 1 r
F P
0 g
s
.51762 .41102 m .55702 .39775 L .56442 .39844 L p .573 .989 1 r
F P
0 g
s
.42204 .42711 m .46085 .41394 L .47007 .42054 L p .745 .934 1 r
F P
0 g
s
.46085 .41394 m .42204 .42711 L .41234 .41775 L p .757 .875 1 r
F P
0 g
s
.55702 .39775 m .51762 .41102 L .50918 .40734 L p .701 .989 1 r
F P
0 g
s
.47007 .42054 m .50918 .40734 L .51762 .41102 L p .701 .989 1 r
F P
0 g
s
.50918 .40734 m .47007 .42054 L .46085 .41394 L p .745 .934 1 r
F P
0 g
s
.29258 .84678 m .25474 .85733 L .22562 .85468 L p .564 .975 1 r
F P
0 g
s
.18775 .8653 m .22562 .85468 L .25474 .85733 L p .564 .975 1 r
F P
0 g
s
.43293 .80867 m .39424 .81946 L .36967 .82528 L p .322 .837 .818 r
F P
0 g
s
.33089 .8361 m .36967 .82528 L .39424 .81946 L p .322 .837 .818 r
F P
0 g
s
.74425 .20166 m .70451 .21523 L .70796 .22473 L p .707 .645 .859 r
F P
0 g
s
.70796 .22473 m .74746 .21056 L .74425 .20166 L p .707 .645 .859 r
F P
0 g
s
.74746 .21056 m .70796 .22473 L .71093 .23434 L p .695 .606 .824 r
F P
0 g
s
.71093 .23434 m .75024 .21955 L .74746 .21056 L p .695 .606 .824 r
F P
0 g
s
.75024 .21955 m .71093 .23434 L .71336 .24381 L p .68 .565 .787 r
F P
0 g
s
.74071 .19309 m .70069 .20608 L .70451 .21523 L p .717 .682 .892 r
F P
0 g
s
.70451 .21523 m .74425 .20166 L .74071 .19309 L p .717 .682 .892 r
F P
0 g
s
.71336 .24381 m .75254 .2284 L .75024 .21955 L p .68 .565 .787 r
F P
0 g
s
.75254 .2284 m .71336 .24381 L .71518 .25289 L p .661 .517 .743 r
F P
0 g
s
.73692 .18506 m .69658 .19752 L .70069 .20608 L p .725 .719 .924 r
F P
0 g
s
.70069 .20608 m .74071 .19309 L .73692 .18506 L p .725 .719 .924 r
F P
0 g
s
.71518 .25289 m .75428 .23688 L .75254 .2284 L p .661 .517 .743 r
F P
0 g
s
.75428 .23688 m .71518 .25289 L .71634 .26136 L p .635 .46 .69 r
F P
0 g
s
.41234 .41775 m .4513 .40446 L .46085 .41394 L p .757 .875 1 r
F P
0 g
s
.69061 .35577 m .65018 .36936 L .64426 .3716 L p .1 .1 .1 r
F P
0 g
s
.60409 .3851 m .64426 .3716 L .65018 .36936 L p .1 .1 .1 r
F P
0 g
s
.73298 .17778 m .69229 .18976 L .69658 .19752 L p .731 .758 .958 r
F P
0 g
s
.69658 .19752 m .73692 .18506 L .73298 .17778 L p .731 .758 .958 r
F P
0 g
s
.01 w
.71388 .19536 m
.71578 .19484 L
s
.71578 .19484 m
.71804 .19423 L
s
.71804 .19423 m
.72059 .19354 L
s
.54011 .2423 m
.71388 .19536 L
s
.5 Mabswid
.64426 .3716 m .60409 .3851 L .59691 .38432 L p .573 .989 1 r
F P
0 g
s
.55702 .39775 m .59691 .38432 L .60409 .3851 L p .573 .989 1 r
F P
0 g
s
.46085 .41394 m .50014 .40062 L .50918 .40734 L p .745 .934 1 r
F P
0 g
s
.50014 .40062 m .46085 .41394 L .4513 .40446 L p .757 .875 1 r
F P
0 g
s
.59691 .38432 m .55702 .39775 L .54878 .39397 L p .701 .989 1 r
F P
0 g
s
.50918 .40734 m .54878 .39397 L .55702 .39775 L p .701 .989 1 r
F P
0 g
s
.54878 .39397 m .50918 .40734 L .50014 .40062 L p .745 .934 1 r
F P
0 g
s
.71634 .26136 m .75543 .24478 L .75428 .23688 L p .635 .46 .69 r
F P
0 g
s
.75543 .24478 m .71634 .26136 L .7168 .26898 L p .6 .387 .621 r
F P
0 g
s
.33089 .8361 m .29258 .84678 L .26396 .84392 L p .564 .975 1 r
F P
0 g
s
.22562 .85468 m .26396 .84392 L .29258 .84678 L p .564 .975 1 r
F P
0 g
s
.36967 .82528 m .40893 .81434 L .43293 .80867 L p .322 .837 .818 r
F P
0 g
s
.72899 .17146 m .68793 .18303 L .69229 .18976 L p .734 .801 .994 r
F P
0 g
s
.69229 .18976 m .73298 .17778 L .72899 .17146 L p .734 .801 .994 r
F P
0 g
s
.7168 .26898 m .75595 .25187 L .75543 .24478 L p .6 .387 .621 r
F P
0 g
s
.75595 .25187 m .7168 .26898 L .71656 .27556 L p .547 .287 .525 r
F P
0 g
s
.11682 .78448 m .15598 .85356 L .18775 .8653 L p .969 .538 .279 r
F P
0 g
s
.7847 .18777 m .74425 .20166 L .74746 .21056 L p .71 .647 .858 r
F P
0 g
s
.74746 .21056 m .78769 .19614 L .7847 .18777 L p .71 .647 .858 r
F P
0 g
s
.78769 .19614 m .74746 .21056 L .75024 .21955 L p .698 .608 .824 r
F P
0 g
s
.75024 .21955 m .79032 .20458 L .78769 .19614 L p .698 .608 .824 r
F P
0 g
s
.79032 .20458 m .75024 .21955 L .75254 .2284 L p .683 .567 .786 r
F P
0 g
s
.7814 .17971 m .74071 .19309 L .74425 .20166 L p .72 .683 .891 r
F P
0 g
s
.74425 .20166 m .7847 .18777 L .7814 .17971 L p .72 .683 .891 r
F P
0 g
s
.75254 .2284 m .7925 .21289 L .79032 .20458 L p .683 .567 .786 r
F P
0 g
s
.7925 .21289 m .75254 .2284 L .75428 .23688 L p .664 .519 .743 r
F P
0 g
s
.71656 .27556 m .75581 .25798 L .75595 .25187 L p .547 .287 .525 r
F P
0 g
s
.75581 .25798 m .71656 .27556 L .71561 .28091 L p .462 .142 .381 r
F P
0 g
s
.4513 .40446 m .49073 .391 L .50014 .40062 L p .757 .875 1 r
F P
0 g
s
.7779 .17215 m .73692 .18506 L .74071 .19309 L p .728 .721 .924 r
F P
0 g
s
.74071 .19309 m .7814 .17971 L .7779 .17215 L p .728 .721 .924 r
F P
0 g
s
.64426 .3716 m .68491 .35793 L .69061 .35577 L p .1 .1 .1 r
F P
0 g
s
.71169 .28734 m .69061 .35577 L .68491 .35793 L p 1 1 .749 r
F P
0 g
s
.68491 .35793 m .64426 .3716 L .63729 .37073 L p .573 .989 1 r
F P
0 g
s
.59691 .38432 m .63729 .37073 L .64426 .3716 L p .573 .989 1 r
F P
0 g
s
.50014 .40062 m .53991 .38713 L .54878 .39397 L p .745 .934 1 r
F P
0 g
s
.53991 .38713 m .50014 .40062 L .49073 .391 L p .757 .875 1 r
F P
0 g
s
.75428 .23688 m .79418 .22084 L .7925 .21289 L p .664 .519 .743 r
F P
0 g
s
.79418 .22084 m .75428 .23688 L .75543 .24478 L p .639 .463 .69 r
F P
0 g
s
.63729 .37073 m .59691 .38432 L .58886 .38043 L p .701 .989 1 r
F P
0 g
s
.54878 .39397 m .58886 .38043 L .59691 .38432 L p .701 .989 1 r
F P
0 g
s
.58886 .38043 m .54878 .39397 L .53991 .38713 L p .745 .934 1 r
F P
0 g
s
.22562 .85468 m .18775 .8653 L .15598 .85356 L p .68 .963 1 r
F P
0 g
s
.77427 .16529 m .73298 .17778 L .73692 .18506 L p .734 .76 .957 r
F P
0 g
s
.73692 .18506 m .7779 .17215 L .77427 .16529 L p .734 .76 .957 r
F P
0 g
s
.36967 .82528 m .33089 .8361 L .30278 .83303 L p .564 .975 1 r
F P
0 g
s
.26396 .84392 m .30278 .83303 L .33089 .8361 L p .564 .975 1 r
F P
0 g
s
.75543 .24478 m .79531 .22823 L .79418 .22084 L p .639 .463 .69 r
F P
0 g
s
.79531 .22823 m .75543 .24478 L .75595 .25187 L p .605 .39 .622 r
F P
0 g
s
.71561 .28091 m .75503 .26294 L .75581 .25798 L p .462 .142 .381 r
F P
0 g
s
.77061 .15933 m .72899 .17146 L .73298 .17778 L p .738 .802 .993 r
F P
0 g
s
.73298 .17778 m .77427 .16529 L .77061 .15933 L p .738 .802 .993 r
F P
0 g
s
.75595 .25187 m .79587 .23486 L .79531 .22823 L p .605 .39 .622 r
F P
0 g
s
.79587 .23486 m .75595 .25187 L .75581 .25798 L p .553 .292 .526 r
F P
0 g
s
.71398 .28488 m .75361 .26659 L .75503 .26294 L p .313 .1 .146 r
F P
0 g
s
.49073 .391 m .53066 .37738 L .53991 .38713 L p .757 .875 1 r
F P
0 g
s
.82583 .17359 m .7847 .18777 L .78769 .19614 L p .712 .648 .858 r
F P
0 g
s
.78769 .19614 m .82864 .18148 L .82583 .17359 L p .712 .648 .858 r
F P
0 g
s
.82864 .18148 m .78769 .19614 L .79032 .20458 L p .7 .61 .824 r
F P
0 g
s
.79032 .20458 m .83112 .18944 L .82864 .18148 L p .7 .61 .824 r
F P
0 g
s
.83112 .18944 m .79032 .20458 L .7925 .21289 L p .686 .568 .786 r
F P
0 g
s
.75581 .25798 m .79583 .24056 L .79587 .23486 L p .553 .292 .526 r
F P
0 g
s
.79583 .24056 m .75581 .25798 L .75503 .26294 L p .469 .147 .383 r
F P
0 g
s
.82276 .16597 m .7814 .17971 L .7847 .18777 L p .722 .685 .891 r
F P
0 g
s
.7847 .18777 m .82583 .17359 L .82276 .16597 L p .722 .685 .891 r
F P
0 g
s
.68491 .35793 m .7088 .28823 L .71169 .28734 L p 1 1 .749 r
F P
0 g
s
.7088 .28823 m .68491 .35793 L .67817 .35697 L p 1 1 .82 r
F P
0 g
s
.63729 .37073 m .67817 .35697 L .68491 .35793 L p .573 .989 1 r
F P
0 g
s
.53991 .38713 m .58018 .37347 L .58886 .38043 L p .745 .934 1 r
F P
0 g
s
.58018 .37347 m .53991 .38713 L .53066 .37738 L p .757 .875 1 r
F P
0 g
s
.7925 .21289 m .8332 .19726 L .83112 .18944 L p .686 .568 .786 r
F P
0 g
s
.8332 .19726 m .7925 .21289 L .79418 .22084 L p .667 .521 .743 r
F P
0 g
s
.67817 .35697 m .63729 .37073 L .62945 .36673 L p .701 .989 1 r
F P
0 g
s
.58886 .38043 m .62945 .36673 L .63729 .37073 L p .701 .989 1 r
F P
0 g
s
.62945 .36673 m .58886 .38043 L .58018 .37347 L p .745 .934 1 r
F P
0 g
s
.81951 .15883 m .7779 .17215 L .7814 .17971 L p .731 .722 .923 r
F P
0 g
s
.7814 .17971 m .82276 .16597 L .81951 .15883 L p .731 .722 .923 r
F P
0 g
s
.26396 .84392 m .22562 .85468 L .1943 .84268 L p .68 .963 1 r
F P
0 g
s
.15598 .85356 m .1943 .84268 L .22562 .85468 L p .68 .963 1 r
F P
0 g
s
.79418 .22084 m .83482 .20474 L .8332 .19726 L p .667 .521 .743 r
F P
0 g
s
.83482 .20474 m .79418 .22084 L .79531 .22823 L p .642 .465 .69 r
F P
0 g
s
.40893 .81434 m .36967 .82528 L .34209 .822 L p .564 .975 1 r
F P
0 g
s
.30278 .83303 m .34209 .822 L .36967 .82528 L p .564 .975 1 r
F P
0 g
s
.81615 .15235 m .77427 .16529 L .7779 .17215 L p .737 .761 .957 r
F P
0 g
s
.7779 .17215 m .81951 .15883 L .81615 .15235 L p .737 .761 .957 r
F P
0 g
s
.01 w
.72059 .19354 m
.81482 .16809 L
s
.5 Mabswid
.0922 .76821 m .12295 .83241 L .15598 .85356 L p .968 .523 .206 r
F P
0 g
s
.75503 .26294 m .7952 .24517 L .79583 .24056 L p .469 .147 .383 r
F P
0 g
s
.7952 .24517 m .75503 .26294 L .75361 .26659 L p .323 .1 .15 r
F P
0 g
s
.79531 .22823 m .83595 .21168 L .83482 .20474 L p .642 .465 .69 r
F P
0 g
s
.83595 .21168 m .79531 .22823 L .79587 .23486 L p .608 .393 .622 r
F P
0 g
s
.81279 .1467 m .77061 .15933 L .77427 .16529 L p .741 .804 .992 r
F P
0 g
s
.77427 .16529 m .81615 .15235 L .81279 .1467 L p .741 .804 .992 r
F P
0 g
s
.79587 .23486 m .83655 .21791 L .83595 .21168 L p .608 .393 .622 r
F P
0 g
s
.83655 .21791 m .79587 .23486 L .79583 .24056 L p .558 .295 .527 r
F P
0 g
s
.53066 .37738 m .57109 .36359 L .58018 .37347 L p .757 .875 1 r
F P
0 g
s
.75361 .26659 m .79398 .24855 L .7952 .24517 L p .323 .1 .15 r
F P
0 g
s
.7088 .28823 m .74903 .26962 L .7516 .26885 L p .472 .927 .828 r
F P
0 g
s
.74903 .26962 m .7088 .28823 L .70539 .2875 L p .689 1 1 r
F P
0 g
s
.67817 .35697 m .70539 .2875 L .7088 .28823 L p 1 1 .82 r
F P
0 g
s
.58018 .37347 m .62095 .35964 L .62945 .36673 L p .745 .934 1 r
F P
0 g
s
.62095 .35964 m .58018 .37347 L .57109 .36359 L p .757 .875 1 r
F P
0 g
s
.70539 .2875 m .67817 .35697 L .67054 .35285 L p 1 .996 .851 r
F P
0 g
s
.62945 .36673 m .67054 .35285 L .67817 .35697 L p .701 .989 1 r
F P
0 g
s
.67054 .35285 m .62945 .36673 L .62095 .35964 L p .745 .934 1 r
F P
0 g
s
.79583 .24056 m .8366 .22324 L .83655 .21791 L p .558 .295 .527 r
F P
0 g
s
.8366 .22324 m .79583 .24056 L .7952 .24517 L p .475 .152 .384 r
F P
0 g
s
.30278 .83303 m .26396 .84392 L .2331 .83166 L p .68 .963 1 r
F P
0 g
s
.1943 .84268 m .2331 .83166 L .26396 .84392 L p .68 .963 1 r
F P
0 g
s
.34209 .822 m .3819 .81083 L .40893 .81434 L p .564 .975 1 r
F P
0 g
s
.70539 .2875 m .74597 .26888 L .74903 .26962 L p .689 1 1 r
F P
0 g
s
.74597 .26888 m .70539 .2875 L .70155 .28515 L p .772 1 1 r
F P
0 g
s
.67054 .35285 m .70155 .28515 L .70539 .2875 L p 1 .996 .851 r
F P
0 g
s
.7952 .24517 m .8361 .22754 L .8366 .22324 L p .475 .152 .384 r
F P
0 g
s
.8361 .22754 m .7952 .24517 L .79398 .24855 L p .331 .1 .153 r
F P
0 g
s
.1943 .84268 m .15598 .85356 L .12295 .83241 L p .723 .906 1 r
F P
0 g
s
.57109 .36359 m .61202 .34962 L .62095 .35964 L p .757 .875 1 r
F P
0 g
s
.74903 .26962 m .78991 .25129 L .7922 .25062 L p .458 .919 .825 r
F P
0 g
s
.78991 .25129 m .74903 .26962 L .74597 .26888 L p .68 1 1 r
F P
0 g
s
.79398 .24855 m .83505 .23069 L .8361 .22754 L p .331 .1 .153 r
F P
0 g
s
.70155 .28515 m .7425 .26663 L .74597 .26888 L p .772 1 1 r
F P
0 g
s
.7425 .26663 m .70155 .28515 L .69738 .28124 L p .794 .943 1 r
F P
0 g
s
.70155 .28515 m .67054 .35285 L .66223 .34564 L p 1 .956 .86 r
F P
0 g
s
.66223 .34564 m .69738 .28124 L .70155 .28515 L p 1 .956 .86 r
F P
0 g
s
.62095 .35964 m .66223 .34564 L .67054 .35285 L p .745 .934 1 r
F P
0 g
s
.66223 .34564 m .62095 .35964 L .61202 .34962 L p .757 .875 1 r
F P
0 g
s
.34209 .822 m .30278 .83303 L .2724 .8205 L p .68 .963 1 r
F P
0 g
s
.2331 .83166 m .2724 .8205 L .30278 .83303 L p .68 .963 1 r
F P
0 g
s
.74597 .26888 m .78716 .25054 L .78991 .25129 L p .68 1 1 r
F P
0 g
s
.78716 .25054 m .74597 .26888 L .7425 .26663 L p .765 1 1 r
F P
0 g
s
.69738 .28124 m .73872 .26291 L .7425 .26663 L p .794 .943 1 r
F P
0 g
s
.73872 .26291 m .69738 .28124 L .69299 .27585 L p .795 .883 1 r
F P
0 g
s
.69738 .28124 m .66223 .34564 L .65348 .33548 L p .999 .916 .857 r
F P
0 g
s
.65348 .33548 m .69299 .27585 L .69738 .28124 L p .999 .916 .857 r
F P
0 g
s
.2331 .83166 m .1943 .84268 L .16166 .82122 L p .723 .906 1 r
F P
0 g
s
.12295 .83241 m .16166 .82122 L .1943 .84268 L p .723 .906 1 r
F P
0 g
s
.7425 .26663 m .78403 .24837 L .78716 .25054 L p .765 1 1 r
F P
0 g
s
.78403 .24837 m .7425 .26663 L .73872 .26291 L p .789 .941 1 r
F P
0 g
s
.61202 .34962 m .65348 .33548 L .66223 .34564 L p .757 .875 1 r
F P
0 g
s
.78991 .25129 m .83145 .23316 L .83349 .23258 L p .447 .912 .822 r
F P
0 g
s
.83145 .23316 m .78991 .25129 L .78716 .25054 L p .672 1 1 r
F P
0 g
s
.69299 .27585 m .73473 .25784 L .73872 .26291 L p .795 .883 1 r
F P
0 g
s
.3819 .81083 m .34209 .822 L .3122 .80919 L p .68 .963 1 r
F P
0 g
s
.2724 .8205 m .3122 .80919 L .34209 .822 L p .68 .963 1 r
F P
0 g
s
.73872 .26291 m .7806 .24484 L .78403 .24837 L p .789 .941 1 r
F P
0 g
s
.7806 .24484 m .73872 .26291 L .73473 .25784 L p .791 .882 1 r
F P
0 g
s
.78716 .25054 m .82898 .2324 L .83145 .23316 L p .672 1 1 r
F P
0 g
s
.82898 .2324 m .78716 .25054 L .78403 .24837 L p .76 1 1 r
F P
0 g
s
.73473 .25784 m .77696 .24004 L .7806 .24484 L p .791 .882 1 r
F P
0 g
s
.78403 .24837 m .82614 .23032 L .82898 .2324 L p .76 1 1 r
F P
0 g
s
.82614 .23032 m .78403 .24837 L .7806 .24484 L p .785 .94 1 r
F P
0 g
s
.2724 .8205 m .2331 .83166 L .20087 .80989 L p .723 .906 1 r
F P
0 g
s
.16166 .82122 m .20087 .80989 L .2331 .83166 L p .723 .906 1 r
F P
0 g
s
.3122 .80919 m .35251 .79774 L .3819 .81083 L p .68 .963 1 r
F P
0 g
s
.7806 .24484 m .82302 .22695 L .82614 .23032 L p .785 .94 1 r
F P
0 g
s
.82302 .22695 m .7806 .24484 L .77696 .24004 L p .788 .881 1 r
F P
0 g
s
.77696 .24004 m .81969 .22239 L .82302 .22695 L p .788 .881 1 r
F P
0 g
s
.3122 .80919 m .2724 .8205 L .24059 .79841 L p .723 .906 1 r
F P
0 g
s
.20087 .80989 m .24059 .79841 L .2724 .8205 L p .723 .906 1 r
F P
0 g
s
.35251 .79774 m .3122 .80919 L .28082 .78678 L p .723 .906 1 r
F P
0 g
s
.24059 .79841 m .28082 .78678 L .3122 .80919 L p .723 .906 1 r
F P
0 g
s
.28082 .78678 m .32158 .775 L .35251 .79774 L p .723 .906 1 r
F P
0 g
s
.25 Mabswid
.77416 0 m
.931 .25158 L
s
.931 .25158 m
1 .81858 L
s
1 .81858 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.0677 .27203 m
0 .83649 L
s
0 .83649 m
.82381 .59403 L
s
.82381 .59403 m
.77416 0 L
s
.77416 0 m
.0677 .27203 L
s
0 .83649 m
.23968 1.01205 L
s
.04533 .86969 m
.04603 .86318 L
s
[(0)] .04392 .88272 .10814 -1 Mshowa
.09585 .9067 m
.09647 .90018 L
s
[(0.5)] .09461 .91974 .09522 -1 Mshowa
.14424 .94215 m
.14478 .93562 L
s
[(1)] .14316 .9552 .08305 -1 Mshowa
.19063 .97612 m
.1911 .96959 L
s
[(1.5)] .18969 .98919 .07157 -1 Mshowa
.23513 1.00872 m
.23553 1.00218 L
s
[(2)] .23434 1.0218 .06072 -1 Mshowa
.125 Mabswid
.05561 .87722 m
.05602 .87332 L
s
.0658 .88469 m
.0662 .88078 L
s
.07591 .89209 m
.0763 .88818 L
s
.08592 .89943 m
.0863 .89552 L
s
.1057 .91391 m
.10606 .91 L
s
.11546 .92106 m
.11581 .91715 L
s
.12513 .92815 m
.12548 .92424 L
s
.13473 .93518 m
.13506 .93126 L
s
.15368 .94906 m
.15399 .94514 L
s
.16303 .95591 m
.16334 .95199 L
s
.17231 .9627 m
.1726 .95878 L
s
.18151 .96944 m
.18179 .96552 L
s
.19968 .98275 m
.19995 .97883 L
s
.20865 .98932 m
.20891 .9854 L
s
.21755 .99584 m
.2178 .99192 L
s
.22638 1.00231 m
.22662 .99838 L
s
.03495 .86209 m
.03538 .85819 L
s
.02448 .85443 m
.02493 .85052 L
s
.01392 .84669 m
.01438 .84279 L
s
.00327 .83889 m
.00373 .83499 L
s
gsave
.11933 1.00721 -79.4886 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.931 .25158 m
1 .81858 L
s
.93242 .26326 m
.92622 .26536 L
s
[(-1)] .94483 .25907 -1 .33784 Mshowa
.94717 .38447 m
.94093 .38646 L
s
[(0)] .95965 .38048 -1 .3195 Mshowa
.96296 .51418 m
.95668 .51606 L
s
[(1)] .9755 .51042 -1 .29996 Mshowa
.97989 .65333 m
.97358 .65509 L
s
[(2)] .99251 .64981 -1 .2791 Mshowa
.9981 .80299 m
.99176 .80462 L
s
[(3)] 1.01079 .79973 -1 .25678 Mshowa
.125 Mabswid
.93529 .28686 m
.93157 .28811 L
s
.9382 .31078 m
.93447 .31201 L
s
.94115 .33501 m
.93742 .33623 L
s
.94414 .35957 m
.9404 .36078 L
s
.95024 .4097 m
.94649 .41089 L
s
.95335 .43528 m
.9496 .43645 L
s
.95651 .46122 m
.95276 .46238 L
s
.95971 .48752 m
.95595 .48866 L
s
.96625 .54123 m
.96248 .54234 L
s
.96959 .56865 m
.96581 .56975 L
s
.97297 .59648 m
.96919 .59756 L
s
.97641 .6247 m
.97262 .62577 L
s
.98343 .68239 m
.97964 .68343 L
s
.98701 .71187 m
.98322 .7129 L
s
.99066 .74179 m
.98686 .7428 L
s
.99435 .77216 m
.99055 .77316 L
s
gsave
1.03824 .4916 -61 -13.7497 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(z) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{284.562, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004L000180P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<ol07@000?l07@000?l07@000?l07@000?l0
7@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l07@000?l0
7@000;T00P5Q0000]`020@0300410600002d00<10`020F00002a00<11P000`400@1O0000[`020@T0
00<10040G`000:`00`4;00040@000Eh0002Z00813`001040005M0000Y`030A4000@10001G@000:D0
0P4D00050@000040G0000:800`4F00050@000040G00009l00`4I00030@00008000<10000F@0009d0
0P4L00030@0000<000<10000F00009X00`4N00030@0000<000<10000F00009P00P4Q00030@0000@0
00<10000E`0009D00`4S00030@0000@000<10000E`0009<00P4V00030@0000D000<10000EP000900
0`4X00030@0000H000<10000E@0008d00`4[00030@0000H000<10000E@0008/00P4_00030@0000H0
00<10000E00008P00`4a00030@0000L000<10000D`0008H00P4d00030@0000L000<10000D`0008<0
0`4f00030@0000P000<10000DP0008400P4i00030@0000P000<10000DP0007h00`4k00030@0000T0
00<10000D@0007/00`4n00030@0000X000<10000D00007T00P5100030@0000X000<10000D00007H0
0`5300030@0000/000<10000C`0007@00P5600030@0000/000<10000C`0007400`5800030@0000`0
00<10000CP0006l00P5;00030@0000d000<10000C@0006`00`5>00030@0000`000<10000C@0006T0
0`5A00030@0000d000<10000C00006L00P5D00030@0000d000<10000C00006@00`5F00030@0000h0
00<10000B`0006800P5I00030@0000l000<10000BP0005l00`5K00030@0000l000<10000BP0005`0
0`5N00030@0000800P4<00030@0004T0001J0081H@040@:;0P4;00030@0004T0001G00<1@0020Al0
0`400i010@020@03R`4000/000<10000B00005D00P4o00@10X/20AX00`4018/10@44R`813@000`40
00170000DP030Ch00`40202;08/0R`4m0P4D00<100>;T0400P400h/1R`070@`000<10000A`000500
0P4n00<12H/20@03=e@101000`400h/10@020@03R`6;00<100>;0@400`400h/1000;00030@0004H0
001=00<1?P020@0808/1R`2;08/40@05T8/10E@00P4:00@100B;0@4118/;0@03R`4100`000<10000
AP0004X00`4o008118/00`6;R`030@J;00D1E041=`020@<01@42R`<118/:0@03QH^5008100>;Q@40
30000`4000150000B0020D000P401P2;08/0R`<100^;T8^@Ri2;T8^@0@02E08100MD0@41T8^@00@1
0X/00`6;0@090@03R`6;00<100>:RhX00P4=00030@0004@0001500<1?`030@J;0P42R`030H^;00N;
0`400e@gE0030@04R`410@F;2P4018F;0H/20@04QH^5R`8100=/0@0030000`4000140000@`020Cl0
0`401h/00@2;08/00P402Y2;T8/1Ri2;T8/40@:;1042E0810h/<0@B;10400hZ;0@020@04RhD10@h0
00<10000@`0004000`4n00<11H/00`6;R`020@N;00<1R`400P402XZ;RX^:R`5D=e@20@03=`4100/1
00>5RhD010400h^50@030@03Rfb;008100=/Q@403P000`4000120000?P020Ch00`4028/0R`2;08/0
0`40292;T8^@Ri2;0`49R`<100=D0@402`400h/1R`030@05QH^5RhD00`400hF;0@020@03RhF;0081
3P000`4000120000>`030Ch00P42R`030H^;00>;0`42R`030H^;00>;0`402XZ;0H^:RhZ;RX/20@03
Q@4100H100R;0@41RhF;Q@<100B5RhF;10400f`10@020@04QFb5K08100>50@003P000`4000110000
>0030Cl00P401`2;08/1R`000`40292;T8^@0I2;0`45R`030H^;00:;2P400diD0@020@B;0`4018F;
0H/30@04RhF;Q@<100>5RhD00`42Q@813`000`4000110000=P020D000P46R`<128/30@0;RX^:RhZ;
RX^:0HX020402X/1E4iDCUA>0HD30@04QH^5R`<100B;K8]/0`4016b50HD20@03K8E/00@140000`40
00100000<`030Cl00`401P2;08/0R`8100V@R`6;T8^@Ri000P42R`030H^;00J;2044R`040DiDCPD1
00B5RhF;0`4018^50HD30@04QH^5R`<10XD50@03IXD1010000<10000?`0003400P4o00<10X/00`6;
R`02R`8118/00`6;R`02R`8100Z;RX^:0HZ;RX^:1`401hF;QH^5RhD01@401TiCCU<1R`<100B;K8]/
0`400fb5K0040@03Q@65008100IV0@65IXD20A0000<10000?`0002h00`4n00<100P0R`2;0040R`81
00V@Ri2;T8^@0I000P47R`T100>;0H/018/30@08RX/1E4iDCU@30@05RhF;QH/00P400hD1Q@030@B5
0`4018D10FH30@03IP41014000<10000?P0002`00P4n00<128/30@R;0`401h^:RhZ;RX/01`4028^5
RhF;Q@650`402HF;QH^50DiCCP020@05CP6;K8/00`400hE/0@020@04K8E/Q@<100>50@4010400h/1
0@020A8000<10000?P0002T00`4U00815P030@0808/1R`2;08/30@08Ri01T8^@Ri030@:;00@1Rh^;
2048R`<100V:RhZ;RX^:RhX00`401DiDCP65008100>;QH/00P42Q@030HD1008100F5K041Q@020@05
R`4109000`4C00030@0003d0000V00<19@030@:;0P4A00<11H/00`6;R`030@F;00<1Rh/00`401H^:
RhZ;00P100V5RhD1QH^5RhD00`402hF;0H^5RhF;QH^5008100E>=dhgCP020@03K8E/008100>5K8D0
0`400hEV0@080@03R`4100815P000`40000l000090020BH00P401P2;08/0R`<130030@0808/0R`2;
08/30@08Ri2;T8^@Ri030@F;2047R`030H/1008100^:RhZ;RP6:RhZ;RP030@0:Q@5>E4iD0Dh1R`81
0hD20@03Q@6500D100<00@401P4H00030@0003`0000Q00<19@030@V;00@1=`411`030@:;00<1Rh/0
0h/30@:;00<1Rh/00h/30@05RhX1RX/02@4028^5RhF;QH^50`402hF;QH^5RhF;Q@6500<100N5RXF:
Q@4g008100I>=`41QF`30@03IXD100@100F@R`41T0050A/000<10000>`0001l00P4U00<1012;0040
R`2;08/0R`2;0CeD0P4200<100P0R`2;0040R`<100R;T8^@R`6;T0<118/:0@:;00<1Rh/018/20@0;
RP6:RhZ;RX^:RhX00`4028F;QH^5RhF;0P43CP8100J50@5/QF`50@05R`4109001@4M00030@0000T0
104^000070030B@00`46R`030H^;00N;00D1?CLm=`020@R;0`48R`<100>;RX/02`402X^5RhF;0H^5
RhD20@0;RhF;0H^5RhF;QH/00P402XD1QHZ5RXF:QHX20@04=dhgCP<100>5IXD010401h^@0@6;T040
0Y030Ah00P4<00030@0002d0000J008190030@0OR`2;08/0R`2;0H/0R`2;08/0R`01E3eD?@6;08/0
R`030@09Ri01T8^@Ri2;00810X/80@04R`410@V;00h1R`41RX^:RhZ;0H^:R`<100Z;QH/1RhF;QH^5
0P400h/1CP02CP050Dhb0F`010401@2;0@40008100B@00410P000`40000I00<100<00@002`000`40
000]00005`030B@00P42R`030H^;00R;00<1Rh/028/01@4m=cd100:;0`45R`030H^;00:;2P400hZ;
0@020@0;R`6;QH^5RhF;QH/00`402X^5RhF;QH^50HD20@0;RXF:QHZ5RP6:QHX00P43Q@070CM>0Dh1
<P030@03T8/1008100B;0@6@0`4400030@0001H00`4400030@000080104400030@0002d0000D00<1
9@020@0J08/0R`0108/0R`2;08/0R`6;08/0R`2;08/70@0;Ri2;T8^@Ri2;0H/02P4018/10@45R`03
0H^;00:;10402X/1RhZ;RX^:RhX30@0;QH^5RhF;QH^50HD00P401X^5RhF;Q@8100=><`400P401Y00
T00100D11`000`40000C00<11P030@/000<10000;@0001800P4V008128/00`6;R`08R`030H^;00>;
0`42R`050Cdg?@4028/80@03RhZ;00<100Z;QH^5RhF;Q@650`402hF;QH^5R`6;QH^500<100^5RP6:
QHZ5RXF:Q@030@R50P401S810I2;T0810Y040@T000<1000040030@/000<100001`030Bl0000@0081
9P020@0H08/0R`2;08/0R`0108/0R`2;08/00@2;0`403X^@Ri2;0CMD=`6;T8^@2044R`<10X/00`6;
R`05R`<100^:RhZ;RX^:RhX1RP030@0;QH^5RhD1QH^5RhD00`402XD1QH^5RhF;QH/40@03T0010081
00<00@400`4<00030@0000`00`4>00030@0000T000<10000;@0001000P4S00<10X/00`6;R`08R`03
0H^;00N;0`49R`040CLm=`P100F;RP6:R`030@0:RhF;QH/1RhF;Q@<100^5R`6;QH^5RhF;Q@030@0;
QHZ5RXF:QHZ50HD00`45Q@030HF500:50`4018D1T8/20@03T04100@13P000`40000900<14@000`40
000i000040000`400@0P008101N;08/0R`0108/0R`2;08/0R`6;08/0R`030@0=Ri01T8^@Ri2;T8^@
R`070@J;0`49R`0?0H/10HZ;RX^:0HZ;RX^:00<100^50HF;QH^5RhF;Q@030@0:QH^5RhF;QH/1R`<1
00F:QHX100020@03T04100@140000`400007008150000`40000i000040001040004L00<12H/00`6;
R`08R`030H^;00811H/00`6;R`07R`H100F;RX^:R`030@0;RhD1QH^5RhF;QH/00`402X^5RhF;QH^5
0HD30@0:QHZ5RXD1QHZ5RP@100>50HD01hD30@08QFb5K8E/0J470A8000<1000010030AH000<10000
>@00010000D100000@0H00<101L0R`6;08/0R`2;08/00@2;08/0R`2;00030@0@T8^@Ri2;T8/1Ri2;
T8^@R`@100=D0@400X/30@F;00<1Rh/00h/30@0:R`6;RX^:RhZ;RP<100Z5RhF;QH^5R`6;0`402h^5
RhF;Q@65RhF;00<100b:0HZ5RXF:Q@6U0JD40A@000@100000`4H0081>`0000l000<100000`000`40
000C00811X/00`6;R`08R`030H^;00>;0`42R`030H^;00N;00<1Rh/00X/40@:;0P403SL10H^5RhF;
QH^5R`6;0`402X^5RhF;0H^5RhD30@0:Q@65RXF:QHZ5RP<12HD00`650@020@0=QFb5K8D1QFb5K06Q
00020@04Y@410A@0104L00030@0003T0000?00030@0000@000<1000040020@0GR`2;08/0R`2;0H/0
R`2;08/0R`0108/00P404Y2;T8^@0I2;T8^@Ri2;T06@R`@100B@Ri2;0P400dl10@08R`<100Z;RX^:
RhZ;RP6:0`402XF;QH^50HF;QH/30@0;R`6;QH^5RhF;QH/00`402XZ5RXF:QHZ50HD40@0300400081
0ZD00`40000@00<17`000`40000i00003`000`40000500030@0000h000<1Y@402H/00`6;R`08R`<1
28/00`6;R`07R`@11X/20@08=`41RhF;QH/30@0:R`6;QH^5RhF;Q@<100Z5RXF:QHZ5RP6:0`45Q@03
0HF500:510402XD1QFb5K8E/QF`30@05JhD10:400P4010010@4@00<10P000`40000N00030@0003P0
000?00030@0000H000<100003@00506Q0040R`2;08/0R`0108/0R`2;0`404h^@0I2;T8^@Ri2;T06@
Ri2;T8/010401I2;T8^@00<100B;0Dlg0P403X/10@6;RX^:R`6;RX^:0`402XD1QH^5RhF;QH/30@0:
RhF;QH^5RhD1Q@<100Z5RXF:Q@65RXF:10400hD1Q@02Q@060JD1Y@6U0`4=00<11@000`40000N0003
0@0003P0000?00030@0000L000<100002`001P6QYJ6U0@V;00<1Rh/00`45R`030H^;00N;00<1Rh/0
0X/40@04R`6;R`<100N;RX^:R`4g00<100Z;QH^5RhF;Q@650`402XF:QHZ50HF:QHX30@03Q@6500N5
0`402Vb5K8E/QFb50HD30@0;QF^5JhD1QF/1X@000P400jD10@020@X00`4800030@0001d00P4j0000
3`000`40000800030@0000X000l10:40X@6;08/0R`2;08/00`405H^@Ri2;T8^@R`6;T8^@Ri2;T8/1
R`040@03T8^@00<128/0304g0@6;RX^:RhZ;RP<100Z5RhF;QH^5R`6;0`402h^5RhD1QH^5RhF;0081
00Z50HF:QHZ5RXF:0`48Q@090HD1Y@01004000<120030@/000<100007P000`40000h00003`000`40
000900030@0000P000P1YJ6UXJFQ0@F;0`42R`030H^;00R;00<1Rh/01X/40@:;0`4038^:0HZ;RX^:
RhZ;RP8100Hc0H^5RhD30@0:Q@65RXF:QHZ5RP<128D01065QHD20@09QFb5K05/QFb500<100Z50HE[
QF^5JhE[1040102Q0JD30@03Y@4000@00`4>00030@0001h000<10000>00000l000<100002P000`40
0006000<0@2Q0:40X@2Q0H/00P405h^@Ri2;T06@Ri2;T8^@Ri01T8^@Ri2;00D100>;0@401H/00`6;
R`07R`050Dl10HX00`402hF;QH^50HF;QH^5008100f;0H^5RhF;QH^5R`6;008100R5RXF:QHX1RP<1
1HD00`65Q@02Q@<100U[0@40Y@6U0JD00P401:D100030A8000<100007@000`40000h00003P000`40
000<00030@0000D000X1XJFQYJ6UXJFQ0P48R`030H^;00R;00<1Rh/00X/70@0BRX^:RhZ;RX/1RhZ;
RX^:RhZ;104038F:QHZ5RXF:0HZ5RP810hD00`65Q@06Q@<100N5K8E/QFb500<100Z5JhE[QF^5J`5[
0`4036J5IXD10:400@0100H15@000`40000M00030@0003P0000>00030@0000d000<100000`00906Q
0:40X@2Q0:41X@6;T8^@Ri2;T8/1Ri2;T8^@Ri2;T06@R`D10X/00`6;R`06R`030H^;00N;0`402XF;
QH^5RhF;QH/30@0;RhF;QH^50HF;QH/00`4028X1RXF:QHZ50P400hD1Q@07Q@<100I[QF^5JhD20@06
Y@0100400P400jD1000G00030@0001d000<10000>00000h000<100003P000`400002000;0JFQYJ6U
XJFQY@400Z400`6;R`07R`030H^;00N;00<1R`400P4098Z;RX^:0HZ;RX^:RhZ;RP6:RhZ;RX^:RhX1
<XZ5RXF:QHZ5RP8128D00`650@020@09QFb5K8D1QFb5008100U[QF^50HE[QF/00`401fH1IXEVQFH0
0`401j40X@6U0JD00P400jD1000G00030@0001d00P4i00003P000`40000?000U0@000@2Q0:40X@2Q
0040X@01T8^@Ri2;T8^@R`6;T8^@Ri2;T0040@N;00<1Rh/01X/00`6;R`07R`<100F;QH^5R`030@09
R`6;QH^5RhF;00<100Z:QHZ5RXF:Q@650P46Q@030HD1008100M[QF^50HE[00@100V@R`6U0JD1Y@00
104H00030@0001h000<10000=`0000h000<10000400030400J6UXJFQYJ6U0@FQ00<1Rh/01h/00`6;
R`03R`H101b:RhZ;RX^:R`6;RX^:RhZ;RX/1RhZ;RX^:RhZ;0`4018X10@45Q@040HF5Q@<100Z50HE/
QFb5K8E/0`401f^5JhE[QF/00`401fJ5IXEVQFH00`4038^@Ri2;T0400@01008100>U0@4060000`40
000N00030@0003L0000>00030@00014001h1X@2Q0:40X@2Q0J40X@2Q006@Ri2;T8^@Ri01T8/20@:;
1049R`030H^;00J;00<1Rh/01X/20@0=<`41RhF;QH^5RhD1Q@020@0:RXF:QHX1RXF:Q@<100>50HD0
18D30@07J`5[QF^5J`030@0:Ri2;T8^@Ri2;T0<100H0Y@41Y@4I00030@0001h000<10000=`0000h0
00<1000040002`6QYJ6UXJFQYJ4100RQ00<1Rh/01X/30@03RX^:00<101Z:R`6;RX^:RhZ;RX/1RhZ;
RX^:RhX1RX^:R`<100B;Q@4128D30@09K8E/QFb5K05/00<100M[QF^5J`5[00<100QVQFJ50HEVQ@81
00R;T06@Ri2;T0@100>;T8/010400jD10@0I00030@0001h000<10000=`0000h000<1000040006@40
X@2Q0:40X@6Q0:40X@2Q0:41Ri2;T8/00`42R`030H^;00<10h/00`6;R`07R`030H^;00J;00<1Rh/0
0P45R`060Dj5RhF;10402HD1QHZ5RXF:Q@030@N50`401f^5JhE[QF/010401i2;T8/1Ri000`44T081
0i040AX000<100007P000`40000g00003P000`40000?000;0JFQYJ6UXJFQY@402Z400`6;R`020@07
RX^:RhX1RP030@0FRX^:RhX1RX^:RhZ;RX^:0HZ;RX^:R`@100V5RhF;QH^5R`400XD20@0:QF`0K8D1
QFb5K0<100Q[0F^5JhE[Q@8100UV0FJ5IXEVQFH00P40292;T8^@Ri2;0P40192;T8/30@03Mi0100@1
70000`40000N00030@0003L0000=00030@00010001H1X@2Q0:40X@010:40X@2Q0:40X@2Q0P48R`<1
1X/00`6;R`07R`030H^;00@10h/00`6;R`06R`8100Z:QHZ508F:Q@650`44Q@040HF5Q@<100R5JhD1
QF^5J`<100F;T8^@R`040@:@00<1T9000P42T0@180000`40000N00030@0003L0000=00030@0000h0
0P402J6UXJFQYJ6U0@0<X@0:0J41RX^:RhZ;RP<1016:RhZ;RX^:R`6;RX^:RhZ;RP030@0IQH^5RhF;
Q@65RhF;QH^5RhD1K8D0QFb5K0030@09JhE[QF^5J`5[008100J5IXEVQFH50@05Ri01T8/00`401h^@
Ri2;T8/00P400i1g0@020B<000<100007`000`4000020081<P0000d000<100002`030@0J08/1X@2Q
0:400@2Q0:40X@2Q0:40X@6Q0046R`@128/00`6;R`03R`<100>;0H/028/00`6;R`06R`050HZ5RXD0
0`400hD1Q@06Q@<100I[QF^5JhD30@07T8^@Ri2;T0020@F@1043T0<19P000`40000N00050@000040
0P000`40000_00003@000`40000900811X/01`6QYJ6UX@402j400`6QX@02X@050H^:RhX00`403HX1
RX^:RhZ;RX^:0HX00`406HF;QH^50HF;QH^5RhF;0H^5RhF;QH^50F`00`402F^5JhE[0F^5J`030@06
IP5VQFJ50`401i2;T8^@Ri000P401H^@0I2;00<100=gT040104Y00030@0001h000D100000@020003
0@0002l0000=00030@0000L00P4088/0R`2;08/0R`6Q0:41X@2Q0:40X@2Q0:41X@2Q0:410h/30@>;
00<1Rh/01H/30@R;00<1Rh/01h/00`6;R`05R`8128D020650@5[QF^51@401i2;0H^@Ri000`45T081
0Y050B`000<100007P001@400001008000<10000;`0000d000<1000020000`6;R`07R`040JFQ0@^Q
00<1XJ401J450@09RX^:R`6;RX^:00<102J50HF;QH^5RhF;Q@65RhF;QH^5RhD1QH^5RhF;Q@5[QF^5
JhE[Q@<100B5IXEV0`402H^@Ri2;T06@R`020@03T8^@00@100=gT7L0104_00030@0001d00P430004
0@000C40000=00030@0000T0020108/0R`2;08/00@010:40X@2Q0:40X@6Q0:40X@2Q00<11h/0106;
0@45R`030H^;00N;00<1Rh/01h/00`6;R`05R`F50`4018D1QF/30@0:Ri2;T8^@Ri2;T0810Y030@:@
1P4b00030@0001h000D100000@0200030@0002l0000=00030@0000X000<1Rh/01X/20@ZQ00<1XJ40
1Z430@07RhZ;RX^:R`030@0PRhF;QH^5R`6;QH^5RhF;QH/1RhF;QH^5RhF;0H^5RhD30@03JhD10081
00B5IXEV0`402H^@0I2;T8^@R`030@03Ri2;008100>@M`401@4e00030@0001h000<100000P020C80
000<00030@0000/000H1R`2;08/30@0CRi010:40X@2Q0:41X@2Q0:40X@050@B;0`400h/1R`08R`03
0H^;00N;00<1Rh/01h/01`6;0@6:RhX00P401XE[QF^5J`8100R@Ri2;T06@R`<10Y030@>@1P4g0003
0@0001h000<10000=P0000`000<1000030000`6;R`020@J;00<1XJ401:400`6QX@04X@D10Z409@6:
0@6;QH^5R`6;QH^5RhF;Q@65RhF;QH^5RhD1QH^5RhF;QH/00`402X^5RhF;IXEVQFH30@08Ri2;T8^@
Ri020@03Ri01008100>@M`401P4j00030@0001l000<10000=@0000`000<100003@020@0CT8^@Ri2;
T8^@0@2Q0:41X@2Q00050@050:40X@000P47R`030H^;00N;00<1Rh/01X/00`6;R`03R`<100Z;0H^:
RhZ;RXE[10401I2;T8^@00@10i020@:@1P4n00030@0001l000<10000=@0000`000<100003P000`6;
R`07R`050J6QX@400j450@JQ0P407XF;QH^5RhF;Q@65RhF;QH^5R`6;QH^5RhF;QH/1R`<100^;QH^5
R`6;QH^5R`030@05T8/1Ri000`40192;T8/20@03T7N@00D1@@000`40000N0081=`0000`000<10000
3`003P6;T8^@Ri2;T8/1X@6Q1P402`2Q0:40X@010:4100R;00<1Rh/01h/00`6;R`05R`<100^;RX^:
RhZ;RP6:R`030@06T8^@Ri2;0P44T0<100>@0@40105400030@0001l000<10000=@0000`000<10000
40000`6;R`06R`L12:420@0KXHZQ0H^5RhF;QH^5R`6;QH^5RhF;Q@65RhF;00<100^;0H^5RhF;QH^5
R`030@05T8^@Ri000`401I2;0H^@008100=g0@40105700030@0001l000<10000=@0000`000<10000
40001P6@Ri2;T0P1016Q0:40X@2Q0:41X@2Q0:400@08R`030H^;00N;00h1R`41RhZ;RX/1RhZ;RP@1
00F@R`6;T0030@J@1@400gL10@1;00030@0001h000<10000=@0000`000<100004@000`6;R`080@ZQ
01X1RZ6:XHZQRP65RhF;QH^5RhD1QH^5RhF;Q@<100V5RhF;QH^50HD00`401h^@Ri2;T8/00P401I2;
T8^@00<100>@0@400WL20Dd000<100007P000`40000e000030000`40000B00L10h/05040X@2Q0:40
X@6Q0:40X@2Q0:4128/00`6;R`02R`<100Z:0HZ;RX^:RhZ;0P401X^@Ri2;T0<10Y000`6@T0030@05
T0410I000P5@00030@0001h000<10000=@0000/000<1000040080@07RX^:RhZ;0@06X@0G0HZQRZ6:
XHZQRZ41RhF;QH^5RhF;0H/00P403HF;QH^50HF;QH^5RhD00`400h^@0@020@06T8^@Ri2;0P400i1g
0@040E<000<100007P020CH0000;00030@0000l01`48R`050J40X@000P403J40X@2Q0:40X@2Q0040
1h/30@0BRhZ;RX^:R`6;RX^:RhZ;0JFP0P45T0<10i050ED000<100007`000`40000d00002`000`40
000>00<110002`6:RhZ;RX^:RhX100:Q01@1XHZQRZ6:XHZQRZ6:XHX1QH^5R`<101F;0H^5RhF;QH^5
R`6;QH^5RhD1X:400P400h^@0@020@03T7L100T1E0000`40000O00030@0003@0000;00030@0000d0
0`4600030H^;00J;01D10040X@2Q0:40X@2Q0:40X@2Q0H/00P406X^:RhZ;0H^:RhZ;RX^:0HZ;RX^:
0J2UX06P0P42T0L100>P0@400`5D00030@0001l000<10000=00000/000<1000030030@P000L1RX^:
RhZ;00<1016QRZ6:XHZQRZ6:XHZQRZ6:X@020@0NQH^5RhF;Q@65RhF;QH^5RhD1QH^5R`6QX:41Y@6U
1@401JD1Y06T00<100>:0@00D`000`40000O00030@0003@0000;00030@0000/00`4900030H^;00@1
0X/0406Q0:40X@2Q0:40X@2Q0:430@0WRX^:RhZ;RX/1RhZ;RX^:RhZ;0H^:RhX1X:FP0J01X06P0J01
X06P00<100>P0@400P5C00030@0001l000<10000=00000/000<100002`020@/00P4058F;QH^5RhD1
RZ6:XHZQRZ6:XHZQ0P40:XZQRP65RhF;QH^5R`6;QH^5RhF;QH/1RhF;0J6PX@6U0JD1Y@6U0JD1Y081
00>P0@400`5C00030@0001l000<10000=00000/000<100002P020@`000<1Rh/01X/02`40X@2Q0:40
X@2Q00811J409`6;RX^:RhZ;RP6:RhZ;RX^:RhX1RX^:0J2U0JFP0J01X06P0J01X0020@06Y0410J@1
D`000`40000O00030@0002800P4800811P0000X000<100002P030@d00141QH^5RhF;QH/1XHZQRZ6:
X@020@0MRZ6:XHZQRZ41RhF;QH^5RhD1QH^5RhF;QH^50HD00P403J6P0J2U0JD1Y@6U0JD010401:01
0J030E8000<1000080000`40000Q00030@0000P000<10000100000X000<100002P020@h000<1Rh/0
1X/01P6Q0:40X@812Z404P6:RhZ;RX^:R`6;RX^:RhZ;RP<100j50J2U0JFP0J01X06PY@D100>T0J@0
105C00030@0001l000<100004P030@8000<100000`4400040@0000@10`000`40000400002P000`40
000900814000306;QH^5RhF;0J6:X@8101^:XHZQRZ6:XHZQRZ41RhF;QH^5RhD1QH^5RhD00`404HD1
Q@6PXJ01X:D1Y@6U0JFP00H100JT0@41W`5C00030@0001l000<100004`001040004200030@0000D0
00D100000@0200050@0000401P0000X000<1000020030A0000<1Rh/01X/30@nQ00/1RhZ;RX^:RhX1
RP030@05QH^5RhD00P403JFPY@6UX06P0J01X:D00`401:D10HD20@04Y0410E8000<100007`000`40
000C00<10`030@D000<100000P000`40000300030@0000@0000:00030@0000P00P4A00070H^5RhF;
Q@030@0HRZ6:XHZQRZ6:XHZQRZ6:XHX1QH^5RhF;104028F:QHZ5RXF:0P403J2QX06PY@6U0JFP0J00
0`401:010J@20@04W`410E8000<100007`000`40000C00040@000@8000@1000110000`4000030003
0@00008000<10000100000X000<100001`020A<000<1R`400`4018Z;RP4AX@030HX100<100b;Q@65
RhF;QH^5RhD20@0AYJ2U0JFP0J01X:D1Y@6;0JD00P400j01X0020@03Y040054000<100007P020AD0
00@100010`020@D000@10000104300030@0000@0000:00030@0000H00`4C008101R;QH^5RhF;0J6:
XHZQRZ6:XHZQRZ6:XHX30@0@QHZ5RXF:Q@65RXF:QHZ5RP81016PXJ01X:D1Y@6UX06P0HD1X0020@03
Y06T008100>O0@40D@000`40000O00030@000180104;008120020@H0000:00030@0000L000<10000
4P002`6;RX^:RhZ;RX/100fQ0P40:Z41QH^5RhF;QH/1RhF;QH^5RhF;0JFPY@6UX06P0J2U0JD1Y@6U
0JD1Y08100BP0@41D@000`40000O00030@0003<0000:00030@0001d001@1RhF;QH^5RhD1RZ6:XHZQ
RZ6:X@<102VQRZ6:0HF:QHZ5RXD1QHZ5RXF:QHZ50J2QX06PY@6U0JFP0J01X:D1Y0020@04X041X0<1
D@000`40000P00040@0000@1;@0000T000<100007P002`6:RhZ;RX^:RhX100NQ0P47X@0U0H^5RhF;
QH^50HF;QH^5RhD1X:FPY@6UX06P0J2U0JD1YJ01X0020@03Y06T008100>O0@00D0000`40000P0003
0@0000<000<10000;00000T000<100007`003P65RhF;QH^5R`6QRZ6:0`40:hZQRZ6:XHZQRP65RXF:
QHZ5RP6:QHZ5RXF:0J6PX@6UX06PY@6UX06P0J000P402:@10J01X06O0`5@00030@00020000<10000
0`000`40000/00002@000`40000O000<0H^:RhZ;RX^:R`6Q0P4=X@0P0HF;QH^5RhF;0H^5RhF;Q@6P
YJ01X:D1YJ01X:D1YJ060@05Y:010J@00`5@00030@0001h00`4500030@0002`0000900030@000200
00P1RhF;QH^5R`<103NQRZ6:XHZQRZ6:XHZQRZ41RXF:QHZ5RXD1QHZ5RXF:0J6PX@6UX06PY@6UX06P
Y@650J@1Y06T008100BO0Il1D0000`40000P00030@0000<000<10000;00000T000<1000080001@6:
RhZ;00<100>;0J403j403P65RhF;QH^5R`6;QH^50`406j2UX06PY@6UX06PY@6UX0650J01X06P0J01
X0020E4000<100007`00104000030Bh0000900030@00024002D1Q@41RXF:Q@6:XHZQRZ6:XHZQRZ6:
XHZQRP65RXF:QHZ5RP6:00810XD05@6QX:41YJ01X:D1YJ01X:D1Q@6TX0020@05Q@6T0J000`5@0003
0@0001l000<100000`000`40000/00002@000`40000Q008100R5RhF;QH^50A2Q00P1RhF;QH^5R`<1
01V;QH^50J2UX06PY@6UX06PY@6UX0650J2U008100RP0J01X06O0E0000<100007`000`40000b0000
2@000`40000Q000O0HF:QHZ5RXF:0J6:XHZQRZ6:XHZQRZ6:XHX1QHZ5RP030@03Q@6500>501H1X:6P
X@6UX06PY@6UX06PY@65X06P0P401ZBP0@6PW`81D0000`40000P00030@000340000800030@0002<0
00T1QH^5RhF;Q@404J40:`650@6;QH^50HF;QH/1YJ2UX06PY@6UX06PY@6UQ@65Y@6U0JD1Y06T0J@0
0P5@00030@00020000<10000<@0000P000<100008`00606:QHZ5RXF:0J6:XHZQRZ6:XHZQRZ6:X@@1
1hD08P65QHD1X:6PX@6UX06PYJ01X:D1Y@65X06P0J01X06P0Il20E0000<100007`020C<000080003
0@0002<000X1QH^5RhF;QH/12:460@:Q02`1QH^5RhF;QH/1RhF;0JFPYJ01X:D1YJ2U0JFP0J01X:D1
Y@6U0J@1Y06TX081C`000`40000P00030@000340000800030@0002<000`1RXF:QHZ5RXD1RZ460@09
RZ6:XHZQRZ4100R501/1QHF50J2QX06PYJ01X:FP0J2U0JD1YJ01K:@00P401Z2T0@6OX081C`000`40
000P00030@000340000800030@0002@000H1RhF;QH/50@nQ02<1RhF;QH^5R`6;QH/1YJ2U0JFPY@6U
X:D1YJ01X:D1Y@6UX0020@06Y:01X06P0P5?00030@00020000<10000<@0000P000<1000090000`65
0@030@:501<1RZ6:XHZQRZ6:XHZQRZ6:XHX100N502<1QHF50J2QX06PYJ01X:FP0J2U0JFP0J01X:@1
Y06T0J@1X0020Dl000<1000080000`40000a000020000`40000T008100R5RhF;QH^50A2Q02d1RhF;
QH^5R`6;Q@6PYJ2U0JFPY@6UX:D1YJ01X:D1YJ01X06P0J01X06PY040C`000`40000P00030@000340
000800030@0002@000<1QHD01XD04P6QRZ6:XHZQRZ6:XHZQRZ6:0@R502@1Q@6QX:6P0J2UX06PYJ01
X:D1YJ01K:D1Y06T0J@1Y06PW`5?00030@0002000P4b000020000`40000T000:0H^5RhF;QH^50A2Q
02`1RhF;QH^5RhD1Q@6PYJ2U0JFPY@6UX:D1YJ01X:D1RZ01X06P0HFT0J@1Y081C`000`40000P0003
0@000300000700030@0002H000<1QHD01HD04P6QRZ6:XHZQRZ6:XHZQRZ6:0@N510408Z6PXJ01X:FP
0J2UX06PY@6UX05/Y@6TX06P0J01X06OX05?00030@00020000<10000<00000L000<100009P002P65
RhF;QH^5R`4@X@040HF;Q@<102J50HD1X:FPY@6UX:D1YJ2U0JFP0J2U0HZP0J2U0JD1Y06T0JBP0Dl0
00<1000080000`40000`00001`000`40000V00030HF500J501<1RZ6:XHZQRZ6:XHZQRZ6:X@650081
02U/QFb50HD1XJ2QX06PYJ01X:FP0J2U0JFP0FbU0JBP0J01X06P0InP0@1?00030@00020000<10000
<00000L000<100009P002065RhF;QH^550401hZ5RXF:QHX00P408Z2UX:D1YJ2U0JFPY@6UX06PY@6:
X065Y@6U0JBP0J01X05?00030@00020000<10000<00000L000<100009`001065QHD30@0oK065XHFQ
QJ65XHFQQJ65XHFQ0HE/QFb5K8E/0J2QX:6P0J2UX06PYJ01X:D1YJ01K:D1Y:01X06PY06T0J0104h0
0P4R00030@000300000700030@0002L000T1R`41QHZ5RP404:40;@65RXF:QHZ5RP6UX:FP0J2UX:D1
YJ2U0JFP0J2U0JFPY@6U0JFP0J01X06P0@1:00@10P000`40000N0081<P0000L000<100009`020@14
QFb5K8E/0HFQQJ65XHFQQJ65XHFQQJ650Fb5K8E/QF`1X:6PX@6UX:FP0J2UX06PYJ01X:D1YJ01X06P
Y06T0J@1X05700<11`000`40000O00030@0002l0000700030@0002L000T1RXF:QHZ5RP404J40:P6:
QHZ5RXF:0JFPYJ01X:FPY@6UX:D1YJ2U0JFP0J2U0JFP0J01X06PY081A0030@X000<100007`000`40
000_00001P000`40000X00140Fb5K8E/QFb50J65XHFQQJ65XHFQQJ65XHD1K8E/QFb5K06PXJ2Q0JFP
YJ01X:FP0J2UX06PY@6UX06PY06T0JBP0J020D00104>00030@0001h000<10000;`0000H000<10000
:0002P6:QHZ5RXF:Q@4@X@0Z0HZ5RXF:QHX1YJ2UX06PYJ2U0JFPY@6UX:D1YJ01X:D1YJ01X06PY06T
0P4m00<14`000`40000M00030@0002l0000600030@0002T004<1QFb5K8E/Q@6QQJ65XHFQQJ65XHFQ
QJ650Fb5K8E/Q@6QX:6PX@6UX:FP0J2UX06PYJ01X:D1YJ01X:@1Y:01X06P0081>P030AL000<10000
70000`40000_00001P000`40000Y00090HF:QHZ5RXD1012Q02`1RXF:QHZ50J2UX:FP0J2UX:D1YJ2U
0JFPY@6UX06PY@6UX06PY@6T0JBP0CH0104J00030@0001`000<10000;`0000H000<10000:@001P65
K8E/Q@L103RQQJ65XHFQQJ65XHFQ0HE/QFb50J6PXJ2Q0JFPYJ01X:FP0J2UX06PY@65X06PY06TX06P
Y06T0C<00`4O00030@0001X00P4a00001P000`40000Y00030HD100<10XD00`6QX@02X@T11:400`65
0@040@0TX:FPYJ01X:FP0J2UX:D1YJ2U0HFP0HFUX06PY@6U0JBP0J01<0030B<000<100006P000`40
000_00001P000`40000Y008101A[QF^5JhD1XHBQQ:64XHBQQ:64X@H102U[QF^50J6PXJ2Q0JFPY@6U
X:FP0J2UX06PY@65X:D1Y:01X06PY06T0@0/00@19`000`40000J00030@0002h0000600030@0002T0
00<1QHD01HD00`6QX@0?X@030HF500>502D1X:FPYJ01X:FP0J2UX:D1YJ2U0JFP0HFUX06PY@6UX06P
0J0102T00`4/00030@0001T000@10000104Y00001P000`40000Y00150HE[QF^5JhE[0HBQQ:64XHBQ
Q:64XHBQQ:41QF^5JhD1XJ2QX06PYJ2U0JFPYJ01X:FP0J2UX06PY@6TX06PY06T0J@102H00`4_0003
0@0001T000D100000@0/00001@000`40000Z00030HF500J500<1XJ403Z400`65Q@02Q@0T0JFPYJ2U
0JFPYJ01X:FPY@6UX:D1YJ2U0JFP0J2U0JFP0J2T0P4R00@1<`000`40000H00030@00008000<10000
:@0000D000<10000:P00@`65JhE[QF^5J`64XHBQQ:64XHBQQ:64XHBQ0HE[QF/1X:6PXJ01X:FPY@6U
X:FP0J2UX06PYJ01X:D1Y:01X:@1Y:000P4O00<1>0000`40000G00030@0000<000<10000:00000D0
00<10000:P000`65Q@06Q@030J6Q00jQ00<1QHD00XD0906UX:FPY@6UX:FP0J2UX:D1YJ2U0JFPY@6U
X06PYJ01X06PY0816`040C`000<100005@020@H000<100009`0000D000<10000:P002065JhE[QF^5
0`40>:64XHBQQ:64XHBQQ:64XH@1JhE[0J2QX:6P0J2UX:D1YJ2UX06PYJ01X:FP0J2U0JBPY06TX06P
0P4H00<1@0000`40000F00050@0000400P000`40000W00001@000`40000Z00030HF500:50`400hD1
X@040@bQ02X1QHF50JFPYJ2U0JFPYJ01X:FPY@6UX:FP0J2U0JFPY@6UX06PY@6TX04E00<1A0000`40
000E00030@0000800P4Z00001@000`40000Z00030HD1008100Z5IXEV0HBQQ:641040<Z64XHBQQ:64
0F^5J`6PXJ2QX06PYJ2U0JFPYJ01X:FPY@6UX06PYJ01X:@1Y:01X:@14@040DP000<100005@000`40
000]00001@000`40000Z008100R5Q8F4QHB50@VQ0`45X@0Z0HF5Q@6UX:FPY@6UX:FP0J2UX:D1YJ2U
X06PY@6UX:D1YJ01X:D1Y:013P030Dd000<1000050000`40000]00001@000`40000Z000F0HEVQFJ5
IXEV0HBQQ:64XHBQQ:64X@@100>40F/00P409Z6PXJ2QX06PYJ2U0JFPYJ01X:FPY@6UX06PYJ01X:@1
Y:01X:@12`030E4000<100004`000`40000]000010000`40000[00090HB5Q8F4QH@1016Q0`40:8D1
X:FPYJ2U0JFPYJ01X:FPY@6UX:FP0J2U0HFPY@6UX:D1YJ01X04700@1E0000`40000C00030@0002d0
000400030@0002/003L1QFJ5IXEVQ@6QQ:64XHBQQ:64XHBQQ:64XH@1IXD1XJ2QX:6P0J2UX:D1YJ2U
X06PYJ2U0JFP008100bUX06PY:01X:@1Y04400<1F@000`40000A0081;`0000@000<10000:`002@64
QHB5Q8F40@0BX@0[0HF50J2UX:FP0J2UX:FP0J2UX:D1YJ2UX06PYJ01X:D1YJ2U0JFP0J0100030Ed0
00<100004@000`40000]000010000`40000[00060HEVQFJ51040>8BQQ:64XHBQQ:64XHBQQ:640FJ5
0J6PXJ2Q0JFPYJ2U0JFPYJ01X:FPY@6UX:D1YJ2U0JBP0J2T105Q00030@00010000<10000;@0000@0
00<10000:`000`640@030@04QFH1X0813j409065Q@6PYJ2UX06PYJ2UX06PYJ2U0JFPYJ01X:FP0J2U
X06PY@D1H`000`40000@00030@0002d0000400030@0002/00P402VJ4IXAVQ06QQ:430@0]XHBQQ:64
XHBQQ:640FJ50J6PXJ2Q0JFPYJ2U0JFPYJ01X:FPY@6UX:D1YJ2U00@100BPY:01I0000`40000@0003
0@0002`0000400030@0002/000l1IXEVQFJ5IP6PXJ2QX:400`48X@0N0HF50JFPYJ2UX06PYJ2UX06P
YJ2U0JFPYJ01X:FP1@401j2UX06PY@40I@000`40000?00030@0002`0000400030@0002/00181Q6J4
IXAV0HBQQ:64XHBQQ:420@0PQ:64XHBQ0HEV0J2QX:6PX@6UX:FPY@6UX:FP0J2UX:D50@0;YJ2U0JBP
Y06TX040IP000`40000>00030@0002`0000400030@0002/001@1IXEVQFJ50J6PXJ2QX:6PXJ2QX0<1
0j405`65Q@6UX:FPYJ01X:FPYJ01X:FPY@6U00@100nPYJ01X:FP0J2UX06PY@40I`000`40000<0081
;P0000<000<10000:`00605VQ6J4IXAV0HBQQ:64XHBQQ:64XHBQQ08100>Q0HD00P404:2QX:6PX@6U
X:FPY@6UX:D30@0CYJ2U0JFPY@6UX:D1Y:2T0JBP0@1W00030@0000d000<10000;00000<000<10000
:`001P65IXEVQ@<1016QX:6PXJ2QX:6PXJ2QX:6PX@030@0>Q@6UX:FPYJ01X:FPYJ040@0EY@6UX:FP
Y@6UX06PYJ01X:FP0J0106T000<1000030000`40000/00000`000`40000[00030FH100<100=VI@40
0P407HBQQ:64XHBQQ:64XHBQQ:41Q6H1X:6PXJ2Q0JFP00@101RUX:FPY@6UX:FP0J2U0JFPYJ01X:@1
Y05Z00030@0000/000<10000;00000<000<10000:`020@09IWmVOfH1Vj6K008101NQX:6PXJ2QX:6P
XJ2QX05VQ@6UX:FPY@040@0JX:FP0J2UX:FP0J2UX:D1YJ2U0JFPY@6UX:D20F/000<100002P000`40
000/00000`000`40000[000=0FEVIFIUIP6QQ:64X@020@0CQ:64XHBQQ:64XHBQ0H@1XJ2QX0030@0M
X:FPYJ2U0JFPYJ2U0JFPYJ01X:FP0J2UX06PY:000P5[00030@0000/000<10000:`0000<000<10000
:P00405VOfIoIWmV0I^QVj6KXI/20@0<XJ2QX:6PXJ2QX05V1@408J2U0JFPYJ2UX06PYJ2UX06PYJ2U
0JFPY@6UX:D1YJ2U0@1]00030@0000X000<10000:`0000<000<10000:P00705VIFIUIVD1Q:64XHBQ
Q:64X@6QQ:64XHBQQ:430@0UXJ2QX:6P0J2UX:FPY@6UX:FPY@6UX:FP0J2UX06PYJ01X:BP0@1^0003
0@0000P00P4]00000`000`40000Y00060GmVOfIo0`402j6KXI^QVj6KXI^Q008100BPXJ2Q0`40:8EV
0J2UX:FPY@6UX:FPYJ01X:FPYJ01X:FPY@6UX:D1YJ2U0JFPY@5_00030@0000P000<10000:`000080
00<10000:P000`5U0@030@03I@6P008100^4XHBQQ:64XHBQQ0040@0ZXH@1IX@1XJ2QX:6P0J2UX:FP
0J2UX:FPY@6UX:FP0J2UX06PYJ01X:@1L@000`40000700030@0002/0000200030@0002T00`404F1V
H6IP0I^QV`6KXI^QVj6K00<102jKX@41X:41QFH1X:FPYJ2U0JFPYJ2U0JFPYJ2UX06PYJ2U0JFPY@6U
X:D1YJ01L@000`40000700030@0002/0000200030@0002P00181H6EPIF1UH064X8BPQ064XH@30@07
Q:64XHBQQ0030@0TIP41XJ2QX:6P0J2UX:FP0J2UX:FPY@6UX:FP0J2UX06PYJ2U0P5c00030@0000H0
00<10000:`00008000<10000:0003P5VH6IPIV1V0J6KXI^Q104039^QVj6KXI^QVj6KX@<102:UX:FP
YJ2U0JFPYJ2U0JFPYJ2UX06PYJ2U0JFPY@6UX:FP0P5d00030@0000D000<10000:`00008000<10000
40020AD000H1IF1UH6D30@03X8BP00<103JPQ:01XHBQQ:64XHBQQ:640FH1X:6PXJ2QX06PYJ2UX06P
YJ2UX:D1YJ2UX06PYJ01X:FPY@5f00030@0000D000<10000:P00008000<100003@030@04T0410A<0
00<1H0400`400fD1X0030@07XI^QVj6KX@020@0`Vj6KXI^QVj6K0FIo0JFPYJ2UX:D1YJ2UX:D1YJ2U
X:FP0J2UX:D1YJ2UX06PYJ01MP000`4000040081;000008000<100002P030@08Ri2;T8/1V@4A00<1
00=6GdH010405j01X8BPQ:24X8BPQ064XHBQQ:64X@5V008102:PXJ2QX:41YJ2UX:FP0J2UX:FPY@6U
X:FP0J2UX:D1YJ01N0000`40000400030@0002X0000200030@0000P00P48T0030IZJ00813P001056
IDH40@05X9^PVj000P402I^QVj6KXI^QV`020@04XI^QV`@10Y020@0OYJ2UX06PYJ2UX:D1YJ2UX:FP
0J2UX:FP0J2UX06P0@1j00030@0000<000<10000:P00008000<100002@020@0<Ri2;T8^@Ri01VYVJ
0P4<00L101>PQ:24X8BPQ064X8BPQ:24X8BP00@102N;T8^@Ri2;T06QX:41YJ2UX:FP0J2UX:FPY@6U
X:FPY@6UX:D1Y@40N`000`40000200030@0002X000000`01000=00030I2@00F@00<1VYX00YX30@L0
20404Y^P0J2KX9^PVj2K0I^QVj6KX@@139006`6UX06PYJ2UX:D1YJ2UX:FP0J2UX:FP0J2UX0020Gd0
00D100000@0400819P000003004000h00P400i2;0@020@06M`6IVYT10iX60@03004100<100jKVY^J
V`6PQ:24X8BPQ0H100>@Ri000P4038^@Ri2;T8^@Ri2;T0<101RUX:FPYJ01X:FPYJ2U0JFPYJ2U0JFP
Y@5n00050@0000400`001040004U000000<00@0040020@B@00<1VYX00YX00`6J0@030@03X04100<1
00jPVj01X9^PVj01X9^PV`@12Y020@^@0`405ZFPYJ2U0JFPYJ2UX06PYJ2UX06PY@6000040@000@H0
00<100008`000003004001800`401GN@0IZI00<10YX00`6JVP03VP0<0I^JViZK0I^JViZK104068^@
0@6;T8^@Ri2;T8^@Ri01T8^@Ri2;T0<100>@Mi000P404j2UX06PYJ2UX:D1YJ2UX:D1Y@40PP001040
0@4300819P000003004001D01@42VP0C0InJW`6PVZ2JX06PVj2KX06PV`030@N@00<1T9002Y020@:@
0`48T08100jU0JFPYJ2UX06PYJ2UX081Q0000`400@0500030@0002<000000`01000C00D100FJVIZI
0@04VP030IZJ00:J00H1VY^JViX30@0HRi2;T8^@Ri2;T8/1Ri2;T8^@Ri2;T8^@0`403I1gT7N@Mi1g
T7N@Mi001@401ZFPY@6UX0@1Q@000`400@0200040@000BD000000`01000@00<11@000`6JVP02VP0:
0IZOVYl1X9ZPVPD10Y020@^@0P46T0@10i020@R@0`401I1gT7N@00L1R@020@<00P4V000000<00@00
3@030@P000H1VIZIVP44VP030IZJ00<100R@Ri2;T8^@R`8100j@Ri2;T8^@Ri2;T06@R`<100UgT7N@
Mi1gT7L00P400i1gT0030@Ig106>0081:`0000030@0000T0104<00030IZJ00:J00@1WiZO0`4=T003
0I2@00N@104>T0D100B@Mi1g0`6;00030@0000@1;@0000030@0000H00`4@00050IZIVYT010404I2;
0@6@Ri2;T8^@Ri2;T8^@008100>;T8/01040191gT7L20@08T7N@Mi1gT7L40@Eg106=00@1<@000003
0@0000<00`4D00030IX100811i020@b@104:T0810i030@05Mi1gT7L01@6900030@0000@1=@000003
0@0100<15`020@0DT8^@Ri2;T8^@Ri2;0H^@Ri2;T8/40@03T7N@008100YgT7N@Mi1gT7N@3P6;00@1
>@00008170020@b@0P42T0<12I020@B@10401I1gT7N@00H1Q`000`4000040Cd0000010010@4L0081
00^;T8^@Ri2;T8^@R`030@0>Mi1gT7N@Mi1gT7N@Mi050@Eg1P6900@150030BX0000300030@0001`0
00<1T9001I030@>@0`48T0<100F@Mi1gT0070HH000<10000104K00030@0002L0000400817@020@03
T8/1008100UgT7N@Mi1gT7L00P400i1gT0030@Mg106900@17`000`40000W00001@020Ah00P4>T0D1
00F@Mi1gT0030HH000@100000`4P00<1:P0000L00P4O00@100AgT7N@1046M`D1R@040B<000<10000
:P0000P000@1000180050@06Mi1gT7N@1@6400040@0000@19`000`40000Z00002@020BH01P440003
0@000880104[00@1:@0000/000<10040;@000`40001j00040@0000@1G00000`00P4]00030@0007X0
105P00003@030B/000<10000L`000`4000040F@0000?00030@0002T000<10000LP040A@000<10000
D@00010000<10040:0000`40001[00030@0000@160000`40001A00002@020@H00P4X00030@0006X0
104I00D1DP0000P000@100011`020BH000<10000I0000`4000030Ad000@10001D`0000P000@10001
20020@030040028000<10000H`040B4000<10040D`0000P000@100012P020B<000<10000G0000`40
00040BH00P5C000020001040004;00030@01024000<10000F`040B/000<10000D@0000P000@10001
30020B0000<10000EP000`400@030H<0000900813`020Ah000<10000E0040HL0000K00030@0001`0
00<10000C`050H/0000L00<16`000`40001<00@150020GT0000>008110000`40000200<10P000`40
0@0I00030@0004L01@4G00040@000GP0000=00040@000@/000<100000P400`01000F00030@000400
00<100000P030Ah000<10000MP0000d000@100012`00104000030AL000<10000@0050Al00P5i0000
3@001040004800<11P000`40000E00030@0003T000@10000104U00030@0007H0000=00040@000@P0
00<100001`000`400@0D00030@0003T0104V00040@000GP0000=00040@000@P000<1000020020@03
0040014000<10000<P000`4000040B/00P5i00003P020@T0104900814P000`40000a00@1ZP0002L0
0`4@00030@0002/000<100400`6^0000:@000`40000>00030@0002T0104A00@1W@00020010460081
3P000`40000T00D15@000`40002N00008P000`40000600<130000`40000M00040@0000@16P000`40
002M00008P000`40000800030@0000X000<100007@040Al000<10000W000028000<100002@000`40
0@0800030@0001P000<10000104S00030@0009/0000R00030@0000X00P400`01000500030@0001L0
104T00040@000Id0000P00<13P020@H000<100004@000`400@030BT00P6N00008P000`40000=0003
0@0100@000<100002P000`40000200@1c@0003<00P401@01000100`01@7A000090040@<000<10000
0P030@800P400`010@0600030@0000@14@040L00000V00030@0000X000D1000100020@D0104G0003
0@000;l0000V00030@0000X000@100000`400`010@020A/000<10000_`0002H000<100001`030@H0
0P4O00030@000;l0000V00030@0000L000<100009`000`40002o00006`020@L00`4900030@0002D0
0`4c00030@0108/0000K00030@0000P000<100001`040BH000<10000<@000`400@2;000030050@05
00400@000P4300040@0000@10`000`40001T00050@010040RP0000h000T100000@01004010001@40
0001008000D100000@0?00@1D`001@400@0108X0000>00090@0000400@0100@000<100000P000`40
000300030@0000d00P4200030@00050000@100400P6900003P002@40000100400@0400030@0000<0
00<100000P000`40000>00040@000N40000>00040@00008100<00@001000104000040@<000<10000
3P020@0300400>00000<00D12P020@P00P4@00040@010N40000f008100<00@40h@0003L00`7R0000
o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000
o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000o`0M0000\
\>"],
  ImageRangeCache->{{{101, 384.562}, {409, 122}} -> {-0.375673, 0.424991, 
  0.00397, 0.00397}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Bode Plots of the discrete function T", "Subsection"],

Cell["\<\
The commands BodeMagnitude[f] and  BodePhase[f] take as input an f \
that is a list of data, or f that is a rational expresion of s.\
\>", "Text"],

Cell[CellGroupData[{

Cell["BodeMagnitude[T];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.25 0.985937 0.0492622 [
[.25 -0.0125 -11 -15 ]
[.25 -0.0125 11 0 ]
[.5 -0.0125 -5 -15 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -8 -15 ]
[.75 -0.0125 8 0 ]
[1 -0.0125 -11 -15 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -30 ]
[.5 -0.0125 29 -15 ]
[-0.0125 .00069 -18 -4.5 ]
[-0.0125 .00069 0 4.5 ]
[-0.0125 .09922 -18 -4.5 ]
[-0.0125 .09922 0 4.5 ]
[-0.0125 .19774 -18 -4.5 ]
[-0.0125 .19774 0 4.5 ]
[-0.0125 .29627 -18 -4.5 ]
[-0.0125 .29627 0 4.5 ]
[-0.0125 .39479 -18 -4.5 ]
[-0.0125 .39479 0 4.5 ]
[-0.0125 .49332 -18 -4.5 ]
[-0.0125 .49332 0 4.5 ]
[-0.0125 .59184 -12 -4.5 ]
[-0.0125 .59184 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34 -7.5 ]
[-0.01875 .30902 -18 7.5 ]
[0 .63053 -14 0 ]
[0 .63053 14 15 ]
[.25 .63053 -11 0 ]
[.25 .63053 11 15 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 15 ]
[.75 .63053 -8 0 ]
[.75 .63053 8 15 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.25 0 m
.25 .61803 L
s
.5 0 m
.5 .61803 L
s
.75 0 m
.75 .61803 L
s
1 0 m
1 .61803 L
s
0 .09922 m
1 .09922 L
s
0 .19774 m
1 .19774 L
s
0 .29627 m
1 .29627 L
s
0 .39479 m
1 .39479 L
s
0 .49332 m
1 .49332 L
s
0 .59184 m
1 .59184 L
s
0 g
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(100) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 0 m
.07526 .00375 L
s
.11928 0 m
.11928 .00375 L
s
.15051 0 m
.15051 .00375 L
s
.17474 0 m
.17474 .00375 L
s
.19454 0 m
.19454 .00375 L
s
.21127 0 m
.21127 .00375 L
s
.22577 0 m
.22577 .00375 L
s
.23856 0 m
.23856 .00375 L
s
.32526 0 m
.32526 .00375 L
s
.36928 0 m
.36928 .00375 L
s
.40051 0 m
.40051 .00375 L
s
.42474 0 m
.42474 .00375 L
s
.44454 0 m
.44454 .00375 L
s
.46127 0 m
.46127 .00375 L
s
.47577 0 m
.47577 .00375 L
s
.48856 0 m
.48856 .00375 L
s
.57526 0 m
.57526 .00375 L
s
.61928 0 m
.61928 .00375 L
s
.65051 0 m
.65051 .00375 L
s
.67474 0 m
.67474 .00375 L
s
.69454 0 m
.69454 .00375 L
s
.71127 0 m
.71127 .00375 L
s
.72577 0 m
.72577 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.82526 0 m
.82526 .00375 L
s
.86928 0 m
.86928 .00375 L
s
.90051 0 m
.90051 .00375 L
s
.92474 0 m
.92474 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.96127 0 m
.96127 .00375 L
s
.97577 0 m
.97577 .00375 L
s
.98856 0 m
.98856 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -34 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(H) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 15.000000 moveto
(\\220) show
105.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 15.000000 moveto
(L) show
117.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
0 .00069 m
.00625 .00069 L
s
[(-20)] -0.0125 .00069 1 0 Mshowa
0 .09922 m
.00625 .09922 L
s
[(-18)] -0.0125 .09922 1 0 Mshowa
0 .19774 m
.00625 .19774 L
s
[(-16)] -0.0125 .19774 1 0 Mshowa
0 .29627 m
.00625 .29627 L
s
[(-14)] -0.0125 .29627 1 0 Mshowa
0 .39479 m
.00625 .39479 L
s
[(-12)] -0.0125 .39479 1 0 Mshowa
0 .49332 m
.00625 .49332 L
s
[(-10)] -0.0125 .49332 1 0 Mshowa
0 .59184 m
.00625 .59184 L
s
[(-8)] -0.0125 .59184 1 0 Mshowa
.125 Mabswid
0 .02532 m
.00375 .02532 L
s
0 .04996 m
.00375 .04996 L
s
0 .07459 m
.00375 .07459 L
s
0 .12385 m
.00375 .12385 L
s
0 .14848 m
.00375 .14848 L
s
0 .17311 m
.00375 .17311 L
s
0 .22237 m
.00375 .22237 L
s
0 .247 m
.00375 .247 L
s
0 .27164 m
.00375 .27164 L
s
0 .3209 m
.00375 .3209 L
s
0 .34553 m
.00375 .34553 L
s
0 .37016 m
.00375 .37016 L
s
0 .41942 m
.00375 .41942 L
s
0 .44405 m
.00375 .44405 L
s
0 .46868 m
.00375 .46868 L
s
0 .51795 m
.00375 .51795 L
s
0 .54258 m
.00375 .54258 L
s
0 .56721 m
.00375 .56721 L
s
0 .61647 m
.00375 .61647 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(dB) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -75 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.01) show
87.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 .61178 m
.25 .61803 L
s
gsave
.25 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.75 .61178 m
.75 .61803 L
s
gsave
.75 .63053 -69 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.07526 .61428 m
.07526 .61803 L
s
.11928 .61428 m
.11928 .61803 L
s
.15051 .61428 m
.15051 .61803 L
s
.17474 .61428 m
.17474 .61803 L
s
.19454 .61428 m
.19454 .61803 L
s
.21127 .61428 m
.21127 .61803 L
s
.22577 .61428 m
.22577 .61803 L
s
.23856 .61428 m
.23856 .61803 L
s
.32526 .61428 m
.32526 .61803 L
s
.36928 .61428 m
.36928 .61803 L
s
.40051 .61428 m
.40051 .61803 L
s
.42474 .61428 m
.42474 .61803 L
s
.44454 .61428 m
.44454 .61803 L
s
.46127 .61428 m
.46127 .61803 L
s
.47577 .61428 m
.47577 .61803 L
s
.48856 .61428 m
.48856 .61803 L
s
.57526 .61428 m
.57526 .61803 L
s
.61928 .61428 m
.61928 .61803 L
s
.65051 .61428 m
.65051 .61803 L
s
.67474 .61428 m
.67474 .61803 L
s
.69454 .61428 m
.69454 .61803 L
s
.71127 .61428 m
.71127 .61803 L
s
.72577 .61428 m
.72577 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.82526 .61428 m
.82526 .61803 L
s
.86928 .61428 m
.86928 .61803 L
s
.90051 .61428 m
.90051 .61803 L
s
.92474 .61428 m
.92474 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.96127 .61428 m
.96127 .61803 L
s
.97577 .61428 m
.97577 .61803 L
s
.98856 .61428 m
.98856 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .00069 m
1 .00069 L
s
.99375 .09922 m
1 .09922 L
s
.99375 .19774 m
1 .19774 L
s
.99375 .29627 m
1 .29627 L
s
.99375 .39479 m
1 .39479 L
s
.99375 .49332 m
1 .49332 L
s
.99375 .59184 m
1 .59184 L
s
.125 Mabswid
.99625 .02532 m
1 .02532 L
s
.99625 .04996 m
1 .04996 L
s
.99625 .07459 m
1 .07459 L
s
.99625 .12385 m
1 .12385 L
s
.99625 .14848 m
1 .14848 L
s
.99625 .17311 m
1 .17311 L
s
.99625 .22237 m
1 .22237 L
s
.99625 .247 m
1 .247 L
s
.99625 .27164 m
1 .27164 L
s
.99625 .3209 m
1 .3209 L
s
.99625 .34553 m
1 .34553 L
s
.99625 .37016 m
1 .37016 L
s
.99625 .41942 m
1 .41942 L
s
.99625 .44405 m
1 .44405 L
s
.99625 .46868 m
1 .46868 L
s
.99625 .51795 m
1 .51795 L
s
.99625 .54258 m
1 .54258 L
s
.99625 .56721 m
1 .56721 L
s
.99625 .61647 m
1 .61647 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.24835 .60332 Mdot
.32467 .58118 Mdot
.37048 .52738 Mdot
.40431 .4614 Mdot
.43199 .42578 Mdot
.45622 .4192 Mdot
.47854 .42636 Mdot
.5 .44385 Mdot
.52146 .48377 Mdot
.54378 .54417 Mdot
.56801 .57136 Mdot
.59569 .50838 Mdot
.62952 .3487 Mdot
.67533 .09583 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<ol08@000?l08@000?l08@00094000<10000
7P000`40001[0000T0000`40000C00030@0000X000<10000JP00088000<100402`00104000040@<0
0P40100100020@050040004010030@@000<10000JP00088000<100402`000`40000200030@000080
00@100010P001040004300030@0000D000D100000@1/0000P@001@400@0100X000<100000P000`40
000300<10P0010400043008110020@@000<10000JP00084000D100400@0:00030@00008000<10000
1@001040004200030@00008000D100000@0600030@0006X0002100040@0100812@001P40000100<1
0P020@@00`4400030@0000800`4300030@0006X0002A00030@0000l000<100000`000`4000060003
0@0006/0002R0081O0000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@0006800P4400030@000080104Y00@1;@040@<00P4W00@10`020@@00P4F0000H@001040004:0003
0@0002X000<10000;P001@400001008000<100009P001@400001008000@100010P000`40000C0000
H@001040004:00030@0002X000<10000;P001@400001008000<100009P001@400001008000@10001
0P000`40000C0000H@001040004:00030@0002X000<10000;P001@400001008000<100009P001@40
0001008000@100010P000`40000C0000H@001040004:00030@0002X000<10000;P001@4000010080
00<100009P001@400001008000@100010P000`40000C0000H@001040004800<1:P030Bh00`430004
0@000BH00`4300040@000@8000@100015@0006800P4;00030@0002X000<10000;P000`4000020081
:@000`400002008110020AH0003o0240003o0240003o0240003o0240000X00@10`020Nl0000X0003
0@0000<000@10001103>0A`0000Y00030@00008000@1000110000`40000<00030@0000H000<10000
10000`40000200050@0000400P001040004200030@0100l000<100001P000`40000300030@000080
00D100000@0200040@000@8000<100403`000`40000600030@0000<000<100000P001@40000100<0
00H10001004200030@0000`000<100001P000`40000400030@00008000D100000@0200040@000@80
00<1004070000280104400050@0000400P000`40000200030@00030000<b0000<0000c80000a0003
<P00030000<100006P0002/000@100010P000`40000200030@00030000<b0000<0000c80000a0003
<P00030000<100006P0002P000@100010P001040004400030@00030000<b0000<0000c80000a0003
<P00030000<100006P0002T00P4400811@020C4000<b0000<0000c80000a0003<P0002l00P4L0000
=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000
<0000c80000a0003<P00030000<100006P0003H000<10000<0000c80000`0003<P00034000<b0000
<0000`40000J0000=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f0081
<@000c80000`0003<P00034000<b0000;`020A`0000f00030@00030000<b0000<0000c80000a0003
<P00030000<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000
<0000c80000a0003<P00030000<100006P0003H00P4a0003<P00030000<b0000<@000c80000_0081
700003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000=P000`40000`0003
<P00030000<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80000a0003
<P00030000<100006P0002P0104300811@000`40000`0003<P00030000<b00008@020@h000<b0000
<0000`40000J0000:P001@400001008000<100000P020HTb00<1<S80?S820A`0000Z00050@000040
0P000`40000200030@00030000<b0000<0000c80000a0003<P00030000<100006P00028010440003
0@0000800P4500030@00030000<b0000<0000c80000a0003<P00030000<100006P0002X000D10000
0@0200030@00008000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000:0030@<0
00@1000110000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000Z00030@000080
0P4500030@00030000<b0000<0000c80000a0003<P00030000<100006P0003H00P4a0003<P000300
00<b0000<@000c80000_0081700003H000<10000<0000c80000`0003<P00034000<b0000<0000`40
000J0000=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@000300
00<b0000<0000c80000a0003<P00030000<100006P0003H000<10000<0000c80000`0003<P000340
00<b0000<0000`40000J0000=P020C4000<b0000<0000c80000a0003<P0002l00P4L0000=P000`40
000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80
000a0003<P00030000<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40
000J0000=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f0081<@000c80
000`0003<P00034000<b0000;`020A`0000f00030@00030000<b0000<0000c80000a0003<P000300
00<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000:0040@<0
0P4500030@00030000<b0000<0000c80000a0003<P00030000<100006P0002X000D100000@020003
0@00008000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000:P001@4000010080
00<100000P020LXb0P4L00008P040@@000@100000`4500030@00030000<b0000<0000c80000a0003
<P00030000<100006P0002X000D100000@0700030@00030000<b0000<0000c80000a0003<P000300
00<100006P0002P00`4300030@0000D000<10000<0000c80000`0003<P00034000<b0000<0000`40
000J0000:P000`40000200<110000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0
000f0081<@000c80000`0003<P00034000<b0000;`020A`0000f00030@00030000<b0000<0000c80
000a0003<P00030000<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40
000J0000=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@000300
00<b0000<0000c80000a0003<P00030000<100006P0003H00P4a0003<P00030000<b0000<@000c80
000_0081700003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000=P000`40
000`0003<P00030000<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80
000a0003<P00030000<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40
000J0000=P020C4000<b0000<0000c80000a0003<P0002l00P4L0000=P000`40000`0003<P000300
00<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80000a0003<P000300
00<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000:0040@@0
00<100000`000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0000D008100<00@40
0`4>00030@0000<000<100000`020LXb0P4L00004`001040004200040@000@d000<100001@440003
0@00030000<b0000<0000c80000a0003<P00030000<100006P0001<000@100010P001040004500@1
10001040004200030@0000<000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
4`001040004200<13P001`40000100401@000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001X0000D00<10P001040004;00<110020@D000<10000<0000c80000`0003<P00034000<b0000
<0000`40000J00005P000`400@030@h000<100000`000`4000030081<@000c80000`0003<P000340
00<b0000;`020A`0000E00817`000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0
000f00030@00030000<b0000<0000c80000a0003<P00030000<100006P0003H000<10000<0000c80
000`0003<P00034000<b0000<0000`40000J0000=P000`40000`0003<P00030000<b0000<@000c80
000`00030@0001X0000f0081<@000c80000`0003<P0001P00P4G0003<P0002l00P4L0000=P000`40
000`0003<P00030000<b00006@000`40000E0003<P00030000<100006P0003H000<10000<0000c80
000`0003<P00034000<b0000<0000`40000J0000=P000`40000`0003<P00030000<b0000<@000c80
000`00030@0001X0000f00030@00030000<b0000<0000c80000a0003<P00030000<100006P0003H0
0P4a0003<P00030000<b0000<@000c80000_0081700003H000<10000<0000c80000`0003<P000340
00<b0000<0000`40000J0000=P000`40000`0003<P00030000<b0000<@000c80000`00030@0001X0
000f00030@00030000<b0000<0000c80000a0003<P00030000<100006P0002P0104200@110000`40
000`0003<P00030000<b0000<@000c80000`00030@0001X0000Z00050@0000401`020LXb0P4L0000
:P000`40000200030@0000@000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
8P040@@000<100000`000`40000300030@00030000<b0000<0000c80000a0003<P00030000<10000
6P0002X000<1000010000`40000200030@00030000<b0000<0000c80000a0003<P00030000<10000
6P0002P00`4300040@000@@000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
:P000`40000200811@020C4000<b00009`020@L000<b0000<@000c80000_0081700003H000<10000
<0000c80000R00811000104000020@8000<b0000<@000c80000`00030@0001X0000f00030@000300
00<b00008`000`40000700040@00<S<000<b0000<0000`40000J0000=P000`40000`0003<P000300
00<b0000<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80000a0003<P000300
00<100006P0003H00P4a0003<P0003000P4b0003<P0002l00P4L0000=P000`40000`0003<P000300
00<b0@00<@000c80000`00030@0001X0000f00030@00030000<b0000<0000c80000a0003<P000300
00<100006P0003H000<10000<0000c80000L00814P000c80000a0003<P00030000<100006P0003H0
00<10000<0000c80000M00030@00010000<b0000<@000c80000`00030@0001X0000f0081<@000c80
000`0003<P00034000<b0000;`020A`0000f00030@00030000<b0000<0000c80000a0003<P000300
00<100006P0003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000=P000`40
000`0003<P00030000@b00000P4^0003<P00030000<100006P0002P0104300811@000`40000`0003
<P00030000<b00000P000`40000/0003<P00030000<100006P0002X000D100000@0200030@000080
0P7:<P81700002X000D100000@0200030@00008000<10000<0000c80000`0003<P00034000<b0000
<0000`40000J00008P040@@000D100000@0200030@00008000<10000<0000c80000`0003<P000340
00<b0000<0000`40000J0000:P001@400001008000<100000P000`40000`0003<P00030000<b0000
4@020Ah000<b0000<0000`40000J0000:0030@<000@1000110000`40000`0003<P00030000<b0000
4P000`40000L0003<P00030000<100006P0002X000<100000P020@D00P4a0003<P00030000<b0000
<@000c80000_0081700003H000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
=P000`40000`0003<P0001D00P4I0003<P00034000<b0000<0000`40000J0000=P000`40000`0003
<P0001H000<100005`000c80000a0003<P00030000<100006P0003H000<10000<0000c80000`0003
<P00034000<b0000<0000`40000J0000=P020C4000<b0000<0000c8000060081:@000c80000_0081
700003H000<10000<0000c80000`0003<P0000L000<100009`000c80000`00030@0001X0000f0003
0@00030000<b0000<0000c80000a0003<P00030000<100006P0003H000<10000<0000c80000`0003
<P00034000<b0000<0000`40000J0000=P000`40000`0003<P00030000<b0000<@000c80000`0003
0@0001X0000f00030@00030000<b0000<0000c80000a0003<P00030000<100006P0003H00P4a0003
<P00030000<b00002`020B@000<b0000;`020A`0000f00030@00030000<b0000<0000c80000<0003
0@00028000<b0000<0000`40000J0000=P000`40000`0003<P0000`00P4R0003<P00034000<b0000
<0000`40000J0000;`020@D000<10000<0000c80000=00030@00020000<b0000<@000c80000`0003
0@0001X0000^00040@000@@000<10000<0000c80000`0003<P00034000<b0000<0000`40000J0000
;P00104000440081bS820A`0000X00@10`020@D000<10000;`020C8000<b0000<@000c80000`0003
0@0001X0000^00040@000@@000<10000<0000`40000`0003<P00034000<b0000<0000`40000J0000
;P001040004400030@00030000<b0000<0000c80000a0003<P00030000<100006P0002l00P450003
0@0000`000<100001P000`40000400030@00008000D100000@0200040@000@8000<100403`000`40
000600030@0000<000<100000P001@400001008000@100010P000`400@0?00030@0000H000<10000
0`000`40000200050@0000400`001P4000400@8000<1000030000`40000600030@0000@000<10000
0P001@400001008000@100010P000`400@0L0000=P3>0A`0003o0240003o0240003o0240003o0240
003o0240003o0240000/008110000`40000300810`040B400P4400030@000080104Y00@1;@040@<0
0P5<0000:`001040004800040@000@@000<100007`001040004:00030@0002X000<10000;P001@40
0001008000<10000B@0002/000@1000120001040004400030@0001l000@100012P000`40000Z0003
0@0002h000D100000@0200030@0004T0000[00040@000@P000@1000110000`40000O00040@000@X0
00<10000:P000`40000^00050@0000400P000`4000190000:`001040004800040@000@@000<10000
7`001040004:00030@0002X000<10000;P001@400001008000<10000B@0002/000@1000120001040
004200<18@001040004800<1:P030Bh00`4300040@000D/0000/00812P020@D000<1000080020@/0
00<10000:P000`40000^00030@0000800P5<0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{101, 388}, {693.812, 516.875}} -> {-5.02827, 27.9015, 
  0.0194952, 0.0989358}}]
}, Open  ]],

Cell["The default frequency band for the plot above is {0.01,100.}", "Text"],

Cell[CellGroupData[{

Cell["Options[BodeMagnitude]", "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
      \(GridPoints \[RuleDelayed] Grid[]\), ",", \(Var \[Rule] s\), ",", 
        RowBox[{"FrequencyBand", "\[Rule]", 
          RowBox[{"{", 
            RowBox[{
              StyleBox["0.01`",
                StyleBoxAutoDelete->True,
                PrintPrecision->1], ",", 
              StyleBox["100.`",
                StyleBoxAutoDelete->True,
                PrintPrecision->3]}], "}"}]}], ",", 
        \(PlotRange \[Rule] Automatic\)}], "}"}]], "Output"]
}, Open  ]],

Cell["\<\
The sampling frequency points belong to the band {0.09,10.1}.  Here \
is the plot of the phase of the list T:\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
BodePhase[T,FrequencyBand -> {0.1,10.},PlotRange -> \
{-200,10}];\
\>", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.5 0.588604 0.00294302 [
[.15051 -0.0125 -11 -15 ]
[.15051 -0.0125 11 0 ]
[.34949 -0.0125 -11 -15 ]
[.34949 -0.0125 11 0 ]
[.5 -0.0125 -5 -15 ]
[.5 -0.0125 5 0 ]
[.65051 -0.0125 -5 -15 ]
[.65051 -0.0125 5 0 ]
[.84949 -0.0125 -5 -15 ]
[.84949 -0.0125 5 0 ]
[1 -0.0125 -8 -15 ]
[1 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -29 -30 ]
[.5 -0.0125 29 -15 ]
[-0.0125 .14715 -24 -4.5 ]
[-0.0125 .14715 0 4.5 ]
[-0.0125 .2943 -24 -4.5 ]
[-0.0125 .2943 0 4.5 ]
[-0.0125 .44145 -18 -4.5 ]
[-0.0125 .44145 0 4.5 ]
[-0.0125 .5886 -6 -4.5 ]
[-0.0125 .5886 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -46 -7.5 ]
[-0.01875 .30902 -24 7.5 ]
[0 .63053 -11 0 ]
[0 .63053 11 15 ]
[.15051 .63053 -11 0 ]
[.15051 .63053 11 15 ]
[.34949 .63053 -11 0 ]
[.34949 .63053 11 15 ]
[.5 .63053 -5 0 ]
[.5 .63053 5 15 ]
[.65051 .63053 -5 0 ]
[.65051 .63053 5 15 ]
[.84949 .63053 -5 0 ]
[.84949 .63053 5 15 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 -0.0005 0 0 ]
[ 1 .61853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
0 0 m
0 .61803 L
s
.15051 0 m
.15051 .61803 L
s
.34949 0 m
.34949 .61803 L
s
.5 0 m
.5 .61803 L
s
.65051 0 m
.65051 .61803 L
s
.84949 0 m
.84949 .61803 L
s
1 0 m
1 .61803 L
s
0 .14715 m
1 .14715 L
s
0 .2943 m
1 .2943 L
s
0 .44145 m
1 .44145 L
s
0 g
.15051 0 m
.15051 .00625 L
s
gsave
.15051 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.2) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.34949 0 m
.34949 .00625 L
s
gsave
.34949 -0.0125 -72 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.65051 0 m
.65051 .00625 L
s
gsave
.65051 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.84949 0 m
.84949 .00625 L
s
gsave
.84949 -0.0125 -66 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(5) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -69 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(10) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.23856 0 m
.23856 .00375 L
s
.30103 0 m
.30103 .00375 L
s
.38908 0 m
.38908 .00375 L
s
.42255 0 m
.42255 .00375 L
s
.45154 0 m
.45154 .00375 L
s
.47712 0 m
.47712 .00375 L
s
.73856 0 m
.73856 .00375 L
s
.80103 0 m
.80103 .00375 L
s
.88908 0 m
.88908 .00375 L
s
.92255 0 m
.92255 .00375 L
s
.95154 0 m
.95154 .00375 L
s
.97712 0 m
.97712 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -90 -34 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(H) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(rad) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
99.000000 15.000000 moveto
(\\220) show
105.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(s) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
111.000000 15.000000 moveto
(L) show
117.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
0 .14715 m
.00625 .14715 L
s
[(-150)] -0.0125 .14715 1 0 Mshowa
0 .2943 m
.00625 .2943 L
s
[(-100)] -0.0125 .2943 1 0 Mshowa
0 .44145 m
.00625 .44145 L
s
[(-50)] -0.0125 .44145 1 0 Mshowa
0 .5886 m
.00625 .5886 L
s
[(0)] -0.0125 .5886 1 0 Mshowa
.125 Mabswid
0 .02943 m
.00375 .02943 L
s
0 .05886 m
.00375 .05886 L
s
0 .08829 m
.00375 .08829 L
s
0 .11772 m
.00375 .11772 L
s
0 .17658 m
.00375 .17658 L
s
0 .20601 m
.00375 .20601 L
s
0 .23544 m
.00375 .23544 L
s
0 .26487 m
.00375 .26487 L
s
0 .32373 m
.00375 .32373 L
s
0 .35316 m
.00375 .35316 L
s
0 .38259 m
.00375 .38259 L
s
0 .41202 m
.00375 .41202 L
s
0 .47088 m
.00375 .47088 L
s
0 .50031 m
.00375 .50031 L
s
0 .52974 m
.00375 .52974 L
s
0 .55917 m
.00375 .55917 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -107 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(deg) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 .61178 m
0 .61803 L
s
gsave
0 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.1) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.15051 .61178 m
.15051 .61803 L
s
gsave
.15051 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.2) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.34949 .61178 m
.34949 .61803 L
s
gsave
.34949 .63053 -72 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(0.5) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.5 .61178 m
.5 .61803 L
s
gsave
.5 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.65051 .61178 m
.65051 .61803 L
s
gsave
.65051 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.84949 .61178 m
.84949 .61803 L
s
gsave
.84949 .63053 -66 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(5) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.001 w
.23856 .61428 m
.23856 .61803 L
s
.30103 .61428 m
.30103 .61803 L
s
.38908 .61428 m
.38908 .61803 L
s
.42255 .61428 m
.42255 .61803 L
s
.45154 .61428 m
.45154 .61803 L
s
.47712 .61428 m
.47712 .61803 L
s
.73856 .61428 m
.73856 .61803 L
s
.80103 .61428 m
.80103 .61803 L
s
.88908 .61428 m
.88908 .61803 L
s
.92255 .61428 m
.92255 .61803 L
s
.95154 .61428 m
.95154 .61803 L
s
.97712 .61428 m
.97712 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .14715 m
1 .14715 L
s
.99375 .2943 m
1 .2943 L
s
.99375 .44145 m
1 .44145 L
s
.99375 .5886 m
1 .5886 L
s
.125 Mabswid
.99625 .02943 m
1 .02943 L
s
.99625 .05886 m
1 .05886 L
s
.99625 .08829 m
1 .08829 L
s
.99625 .11772 m
1 .11772 L
s
.99625 .17658 m
1 .17658 L
s
.99625 .20601 m
1 .20601 L
s
.99625 .23544 m
1 .23544 L
s
.99625 .26487 m
1 .26487 L
s
.99625 .32373 m
1 .32373 L
s
.99625 .35316 m
1 .35316 L
s
.99625 .38259 m
1 .38259 L
s
.99625 .41202 m
1 .41202 L
s
.99625 .47088 m
1 .47088 L
s
.99625 .50031 m
1 .50031 L
s
.99625 .52974 m
1 .52974 L
s
.99625 .55917 m
1 .55917 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.01 w
.14933 .53055 Mdot
.24097 .50516 Mdot
.30861 .49638 Mdot
.36398 .49943 Mdot
.41245 .5001 Mdot
.45709 .4955 Mdot
.5 .48672 Mdot
.54291 .47049 Mdot
.58755 .4318 Mdot
.63602 .35867 Mdot
.69139 .26325 Mdot
.75903 .16163 Mdot
.85067 .05414 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<ol08@000?l08@000?l08@0009T000<10000
7P000`40001S0000V0000`40000C00030@0000X000<10000HP0008X000<100402`00104000040@<0
0P40100100020@050040004010030@@000<10000HP0008X000<100402`000`40000200030@000080
00@100010P001040004300030@0000D000D100000@1T0000R@001@400@0100X000<100000P000`40
000300<10P0010400043008110020@@000<10000HP0008T000D100400@0:00030@00008000<10000
1@001040004200030@00008000D100000@0600030@000680002900040@0100812@001P40000100<1
0P020@@00`4400030@0000800`4300030@000680002I00030@0000l000<100000`000`4000060003
0@0006<0002Z0081M0000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l0
8@0005D00P4400030@000080104J008110000`40000200<15P040A/0104U00<16@040@<00P4A0000
E0001040004800030@0001X000@100012`000`40000E00030@0001X000<10000:@000`40000H0005
0@0000400P000`40000>0000E0001040004900030@0001T000@100012`000`40000E00030@0001/0
00<10000:0000`40000H00050@0000400P000`40000>0000E0001040004:00030@0001P000@10001
20030AP000<1000070000`40000T00<16`001@400001008000<100003P0005@000@100012`000`40
000G00040@000@P000<1000060000`40000M00030@0002<000<100006`001@400001008000<10000
3P0005@000@1000120001040004I00040@000@P000<100005P030A`000@100019@000`40000I00<1
0`001040004@0000E@020@X00P4K00812@040AL000<100006`020BH0104J00030@0000800P4A0000
o`0Q0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000?@3?0A@0000m00030@0001`000<100003`000`40
000:00030@0000L000<100001@000`40000400030@0000<000<100000P000`40000200030@0001`0
00<100003`000`40000:00030@0000L000<100001@000`40000400030@0000<000<100000P000`40
000200030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00817@000c80000V0003<P0001`000<b000070000c80000V0003<P0001/0
0P4D0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40
000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40
000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40
000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40
000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002D00P4N00030@000180
000m00817@000c80000V0003<P0001`000<b000070000c80000V00030@0001/00P4D0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@020Ad0
00<b00009P000c80000L0003<P0001`000<b00009P000c80000K0081500003d000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d00P4M0003<P0002H0
00<b000070000c80000L0003<P0002H000<b00006`020A@0000m00030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000Z00@10P030@@00P4400030@0001`0
00<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000/00030@0000@0
00@100010P00104000030LTb0`4D0000;0000`40000400040@000@8000D100000@0N0003<P0002H0
00<b000070000c80000L0003<P0002H000<b000070000`40000B000090040@@000@100000`430004
0@000@<000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<10000
4P0002`000D100000@0500040@000@<000<1000070000c80000V0003<P0001`000<b000070000c80
000C00814@000c80000L00030@000180000Z00<10`000`40000300040@000@<000<1000070000c80
000V0003<P0001`000<b000070000c80000D00030@0000l000<b000070000`40000B0000;0001040
00040@<00P4400030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L0003
0@000180000m00817@000c80000V0003<P0001`000<b000070000c80000V0003<P0001/00P4D0000
?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000
?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000
?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000
?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000
?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000
?@020Ad000<b00009P000c80000L0003<P0001`000<b00009P000c80000K0081500003d000<10000
70000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<10000
70000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<10000
70000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<10000
70000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<10000
70000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d00P4M0003
<P0002H000<b000070000c80000L0003<P0002H000<b00006`020A@0000m00030@0001`000<b0000
9P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b0000
9P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b0000
9P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b0000
9P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`000<b0000
9P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00817@000c80000V0003
<P0001`000<b000070000c80000500817`000c80000K0081500003d000<1000070000c80000V0003
<P0001`000<b000070000c80000600030@0001d000<b000070000`40000B0000?@000`40000L0003
<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40000L0003
<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40000L0003
<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B00006`030@`010430081
10020@@000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<10000
4P0001h000<100002`001@400001008000@100010P001@40000101h000<b00009P000c80000L0003
<P0001`000<b00009P000c80000L00030@000180000@008100@00@000`4300<13@001@4000010080
00@100010P00104000030LTb0`4D00003`001040004200030@0000<000@100011@040@@000D10000
0@0200040@000@8000D100000@0N0003<P0002H000<b000070000c80000L0003<P0002H000<b0000
70000`40000B00003`001040004200<10`001040004=00050@0000400P001040004200050@000040
7P000c80000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0000l000@10001
0P001040004200040@000@/00`4300040@000@8000@100010`000`40000L0003<P0002H000<b0000
70000c80000L0003<P0002H000<b000070000`40000B000040030@<00P44008100<00@002`000`40
0002008110020@@000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P00018000<10000:0000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B00004@020BX00P4M0003<P0002H000<b000070000c80000L0003<P0002H0
00<b00006`020A@0000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000m00817@000c80000V0003<P0001`000<b000060020@8000<b00009P000c80
000K0081500003d000<1000070000c80000V0003<P0001`000<b00006@001040038X0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d00P4M0003<P0002H000<b000070000c80000L0003<P0002H000<b00006`020A@0
000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180
000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180
000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180
000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180
000m00030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180
000m00817@000c80000V0003<P0001`000<b000070000c80000V0003<P0001/00P4D0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0002H000<b000070000c80000>008130000c80000V0003<P0001`000<100004P000300
0`44008110000`40000L0003<P0002H000<b000070000c80000?00030@0000X000<b00009P000c80
000L00030@000180000c00040@000@8000D100000@0N0003<P0002H000<b000070000c80000L0003
<P0002H000<b000070000`40000B0000<`001040004200040@0000<1bC830A@0000Z00@10P030@<0
00@100010`000`40000L0003<P0002H000<b000070000c80000L0003<P0002H000<b000070000`40
000B0000<0000`40000300040@000@<000<1000070000c80000V0003<P0001`000<b000070000c80
000V0003<P0001`000<100004P00030000<100000`001040004300030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000`00@10`020@@000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<b00001@020AD000<b00009P000c80000L00030@000180000m00817@000c80
000V0003<P0001`000<b00001P000`40000C0003<P0002H000<b00006`020A@0000m00030@0001`0
00<b00009P000c80000L0003<P0001`000<b00009P000c80000L00030@000180000m00030@0001`0
00<b00009P000c80000K00817P000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000V0003<P0001`000<1000070000c80000V0003<P0001`000<100004P0003d000<1000070000c80
000M00811`000c80000C00811`000c80000L0003<P0002H000<b000070000`40000B0000?@000`40
000L0003<P0001h000<100001@000c8000020@L00P4900030@0000D000<b000070000c80000V0003
<P0001`000<100004P0003d00P4M0003<P0000l00P4E0005<P00004020000`40000?0003<P0001`0
00<b00009P000c80000K0081500003d000<1000070000c80000@00030@0001<000<b000070000c80
000L0003<P0002H000<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80
000L0003<P0002H000<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80
000L0003<P0002H000<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80
000L0003<P0002H000<b000070000`40000B0000?@000`40000K0081:0000c80000L0003<P0001`0
00<b00009P000c80000L00030@000180000m00817@000`40000V0003<P0001`000<b000070000c80
000V0003<P0001/00P4D0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B0000?@000`40000L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b000070000`40000B0000?@020Ad000<b00009P000c80000L0003<P0001`000<b00009P000c80
000K0081500003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003d000<1000070000c80000V0003<P0001`000<b000070000c80000V0003<P0001`0
00<100004P0003L00P4400030@0001`000<b00009P000c80000L0003<P0001`000<b00009P000c80
000L00030@000180000f00040@000@<00`4L0003<P0002H000<b000070000c80000L0003<P0002H0
00<b00006P030A@0000f00040@000@<000<1000070000c80000V0003<P0001`000<b000070000c80
000V0003<P0001`000<100004P0003H000@100010`000`40000L0003<P0002H000<b000070000c80
000L0003<P0002H000<b000070000`40000B0000=P001040004300030@0001`000<b00009P000c80
000L0003<P0001`000<b00009P000c80000L00030@000180000f00040@000@<000<1000070000c80
000V0003<P0001`000<b000070000c80000V0003<P0001`000<100004P0003L00P4400030@0001`0
00<100003`000`40000:00030@0000L000<100001@000`40000400030@0000<000<100000P000`40
000200030@0001`000<100003`000`40000:00030@0000L000<100001@000`40000400030@0000<0
00<100000P000`40000200030@000180000m0<l150000?l08@000?l08@000?l08@000?l08@000?l0
8@000?l08@0003H00P4400030@000080104@008110000`40000200@16P020@@000<100000P030AH0
104K00@19@030C<0000e00040@000@X000<100003P001040004800030@0001X000@100012`000`40
000E00030@0001X000<10000:@000`40000`0000=@001040004:00030@0000h000@100012@000`40
000I00040@000@/000<100005@000`40000K00030@0002P000<10000<00003D000@100012P000`40
000>00040@000@X000<1000060001040004800<160000`40000L00030@0002@00`4c0000=@001040
004:00030@0000h000@100012`000`40000G00040@000@P000<1000060000`40000M00030@0002<0
00<10000<`0003D000@1000120030A0000@1000120001040004I00040@000@P000<100005P030A`0
00@100019@000`40000c0000=P020@/000<100003`020@X00P4K00812@040AL000<100006`020BH0
104b0000o`0Q0000o`0Q0000o`0Q0000o`0Q0000\
\>"],
  ImageRangeCache->{{{101, 388}, {366.938, 190}} -> {-2.58724, 60.72, 
  0.00974759, 1.65605}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["3D List plot of gridvalues of T", "Subsection"],

Cell["\<\
To plot in 3D discrete functions of frequence (i.e., lists of \
values such as T ), use the command\
\>", "Text"],

Cell[CellGroupData[{

Cell["RHPListPlot3D[T,PlotRange -> {{0,3},Automatic,Automatic}];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .57471 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics3D
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.030072 1.03007 -0.208785 1.03007 [
[.01823 .28309 -2.68831 -9 ]
[.01823 .28309 3.31169 0 ]
[.28372 .1918 -2.86282 -9 ]
[.28372 .1918 3.13718 0 ]
[.56985 .09344 -3.05918 -9 ]
[.56985 .09344 2.94082 0 ]
[.87912 -0.01282 -3.28178 -9 ]
[.87912 -0.01282 2.71822 0 ]
[.42496 .07918 -4.93004 -15 ]
[.42496 .07918 5.06996 0 ]
[.91811 .03414 0 -6.0157 ]
[.91811 .03414 24 2.9843 ]
[.9443 .07334 0 -5.9888 ]
[.9443 .07334 6 3.0112 ]
[.96975 .11141 0 -5.96283 ]
[.96975 .11141 18 3.03717 ]
[1.00542 .05311 0 -9.98133 ]
[1.00542 .05311 34 5.01867 ]
[.99918 .19857 0 -5.90901 ]
[.99918 .19857 6 3.09099 ]
[1.00526 .25186 0 -5.87832 ]
[1.00526 .25186 18 3.12168 ]
[1.01148 .30645 0 -5.84693 ]
[1.01148 .30645 18 3.15307 ]
[1.06412 .20967 0 -9.81942 ]
[1.06412 .20967 34 5.18058 ]
[ 0 0 0 0 ]
[ 1 .57471 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.0169 .2959 m
.88032 0 L
s
.0169 .2959 m
.01623 .3023 L
s
[(0)] .01823 .28309 -0.1039 1 Mshowa
.28313 .20466 m
.28284 .21109 L
s
[(1)] .28372 .1918 -0.04573 1 Mshowa
.5701 .10631 m
.57023 .11275 L
s
[(2)] .56985 .09344 .01973 1 Mshowa
.88032 0 m
.88092 .00641 L
s
[(3)] .87912 -0.01282 .09393 1 Mshowa
.125 Mabswid
.06859 .27818 m
.06824 .28203 L
s
.12105 .2602 m
.12074 .26405 L
s
.17428 .24196 m
.17401 .24582 L
s
.2283 .22345 m
.22808 .22731 L
s
.33879 .18558 m
.33867 .18944 L
s
.3953 .16622 m
.39522 .17008 L
s
.45268 .14656 m
.45265 .15042 L
s
.51094 .12659 m
.51096 .13045 L
s
.6302 .08572 m
.63033 .08958 L
s
.69125 .0648 m
.69143 .06866 L
s
.75327 .04354 m
.75351 .0474 L
s
.81628 .02195 m
.81658 .0258 L
s
gsave
.42496 .07918 -65.93 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.88032 0 m
.98098 .15049 L
s
.9059 .03825 m
.8998 .0403 L
s
[(-0.2)] .91811 .03414 -1 .33682 Mshowa
.93208 .07738 m
.92597 .0794 L
s
[(0)] .9443 .07334 -1 .33084 Mshowa
.9575 .11539 m
.95138 .11738 L
s
[(0.2)] .96975 .11141 -1 .32507 Mshowa
.125 Mabswid
.91252 .04814 m
.90886 .04937 L
s
.91909 .05796 m
.91542 .05918 L
s
.92561 .0677 m
.92194 .06892 L
s
.9385 .08698 m
.93483 .08819 L
s
.94488 .09652 m
.94121 .09772 L
s
.95121 .10599 m
.94754 .10719 L
s
.89924 .02828 m
.89558 .02952 L
s
.89253 .01825 m
.88887 .01949 L
s
.88576 .00813 m
.88211 .00938 L
s
.96374 .12472 m
.96007 .12591 L
s
.96994 .13399 m
.96627 .13517 L
s
.9761 .14319 m
.97242 .14437 L
s
gsave
1.00542 .05311 -61 -13.9813 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Im) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.98098 .15049 m
1 .31764 L
s
.98689 .20242 m
.98075 .20434 L
s
[(0)] .99918 .19857 -1 .31311 Mshowa
.99294 .25563 m
.98679 .25751 L
s
[(0.2)] 1.00526 .25186 -1 .30629 Mshowa
.99915 .31014 m
.99298 .31199 L
s
[(0.4)] 1.01148 .30645 -1 .29932 Mshowa
.125 Mabswid
.98244 .16333 m
.97876 .1645 L
s
.98392 .17628 m
.98023 .17745 L
s
.9854 .18931 m
.98172 .19047 L
s
.98839 .2156 m
.9847 .21675 L
s
.9899 .22886 m
.98621 .23001 L
s
.99142 .2422 m
.98773 .24334 L
s
.99448 .26913 m
.99079 .27026 L
s
.99603 .28272 m
.99233 .28384 L
s
.99758 .29639 m
.99388 .2975 L
s
gsave
1.06412 .20967 -61 -13.8194 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
.25 Mabswid
.0169 .2959 m
0 .45854 L
s
0 .45854 m
.1384 .57471 L
s
.1384 .57471 m
.14987 .41626 L
s
.14987 .41626 m
.0169 .2959 L
s
.88032 0 m
.98098 .15049 L
s
.98098 .15049 m
1 .31764 L
s
1 .31764 m
.8964 .17117 L
s
.8964 .17117 m
.88032 0 L
s
.0169 .2959 m
0 .45854 L
s
0 .45854 m
.8964 .17117 L
s
.8964 .17117 m
.88032 0 L
s
.88032 0 m
.0169 .2959 L
s
.14987 .41626 m
.98098 .15049 L
s
.98098 .15049 m
1 .31764 L
s
1 .31764 m
.1384 .57471 L
s
.1384 .57471 m
.14987 .41626 L
s
0 0 m
1 0 L
1 .57471 L
0 .57471 L
closepath
clip
newpath
.01 w
.07136 .51419 Mdot
.08549 .49379 Mdot
.10148 .4656 Mdot
.12504 .43489 Mdot
.15744 .41364 Mdot
.19296 .40127 Mdot
.22981 .38935 Mdot
.27019 .37459 Mdot
.31697 .35605 Mdot
.37217 .33085 Mdot
.43849 .28374 Mdot
.53552 .19577 Mdot
.70893 .10156 Mdot
.25 Mabswid
.88032 0 m
.98098 .15049 L
s
.98098 .15049 m
1 .31764 L
s
1 .31764 m
.8964 .17117 L
s
.8964 .17117 m
.88032 0 L
s
.0169 .2959 m
0 .45854 L
s
0 .45854 m
.8964 .17117 L
s
.8964 .17117 m
.88032 0 L
s
.88032 0 m
.0169 .2959 L
s
.0169 .2959 m
.88032 0 L
s
.0169 .2959 m
.01623 .3023 L
s
[(0)] .01823 .28309 -0.1039 1 Mshowa
.28313 .20466 m
.28284 .21109 L
s
[(1)] .28372 .1918 -0.04573 1 Mshowa
.5701 .10631 m
.57023 .11275 L
s
[(2)] .56985 .09344 .01973 1 Mshowa
.88032 0 m
.88092 .00641 L
s
[(3)] .87912 -0.01282 .09393 1 Mshowa
.125 Mabswid
.06859 .27818 m
.06824 .28203 L
s
.12105 .2602 m
.12074 .26405 L
s
.17428 .24196 m
.17401 .24582 L
s
.2283 .22345 m
.22808 .22731 L
s
.33879 .18558 m
.33867 .18944 L
s
.3953 .16622 m
.39522 .17008 L
s
.45268 .14656 m
.45265 .15042 L
s
.51094 .12659 m
.51096 .13045 L
s
.6302 .08572 m
.63033 .08958 L
s
.69125 .0648 m
.69143 .06866 L
s
.75327 .04354 m
.75351 .0474 L
s
.81628 .02195 m
.81658 .0258 L
s
gsave
.42496 .07918 -65.93 -19 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(w) show
69.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.25 Mabswid
.98098 .15049 m
1 .31764 L
s
.98689 .20242 m
.98075 .20434 L
s
[(0)] .99918 .19857 -1 .31311 Mshowa
.99294 .25563 m
.98679 .25751 L
s
[(0.2)] 1.00526 .25186 -1 .30629 Mshowa
.99915 .31014 m
.99298 .31199 L
s
[(0.4)] 1.01148 .30645 -1 .29932 Mshowa
.125 Mabswid
.98244 .16333 m
.97876 .1645 L
s
.98392 .17628 m
.98023 .17745 L
s
.9854 .18931 m
.98172 .19047 L
s
.98839 .2156 m
.9847 .21675 L
s
.9899 .22886 m
.98621 .23001 L
s
.99142 .2422 m
.98773 .24334 L
s
.99448 .26913 m
.99079 .27026 L
s
.99603 .28272 m
.99233 .28384 L
s
.99758 .29639 m
.99388 .2975 L
s
gsave
1.06412 .20967 -61 -13.8194 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Re) show
%%IncludeResource: font Math2Mono
%%IncludeFont: Math2Mono
/Math2Mono findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
75.000000 15.000000 moveto
(@) show
87.000000 15.000000 moveto
(D) show
93.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Math2Mono
%%+ font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Math2Mono
%%+ Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Math2Mono
%%+ font Courier
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 165.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000Y@P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<ol08@000?l08@000?l08@000?l08@000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000<l00P5?0000cP001040
005>0000d@000`40001<0000c`020Dl0003A00030@0004`0003>00040@000Dh0003?0081C`000?l0
8@000=000P5>0000c@030@03004104d0003:00<110020Dd0003700<11`000`400@1<0000a0030@X0
00@10001B`000<400`4=00040@010@d00P4400030@000080104_0000_P030@03004000h000@10001
2`001040004800030@0001000P48008150000;`00P4D00030@0100812P001040004900030@0000l0
00<1000020000`40000B0000^@030AH000D100000@0400@10P001040004:00030@0100@100D00@01
00020@<000<1000020000`40000B0000H`000`400@1@00<16@00104000020@T000@100012`001040
004300050@01004010000`40000800030@000180001S00030@0104d00`4L00040@000@8000<10000
1P001040004800040@000@8000T100000@01004010000`40000800030@000180001R00050@010040
B@040Ah000<100000`000`40000700812P020@<000T100000@01004010000`40000800030@000180
001R00050@010040AP030B8000<100000`020AT000@100000P400`01000400030@0000P000<10000
4P00068000@100400P5200<19@000`40000500030@0001@01@4:008120020A@0002W00<1:0000`40
00040081A@000:@00`4[00030@0000H000<1000010020C`0002Q00@1;P000`400005008110001040
004k0000Q@040AD00`4b00030@0000H000<100000P001040004k0000Q@000`40000C00<1=@000`40
000700050@0000400P000`40000i0000QP000`40000?00<1>0000`40000600030@01008000@10001
>`0008L000<100002`030C/000<1000020001040004200030@0003T0002800030@0000P00`4m0003
0@0000P00P420081?00008D000@100011`030D0000<100002P000`40000m0000QP020@D00`530003
0@0000T00P4o0000RP030AX00P4Z00030@0000/000<1000010020@@000<100000P040BT0002700<1
7P000`40000X00030@0000/00P4400040@000@P000<10000:P0008@0105<00030@0000/000<10000
0P001040004900030@0002T0002100<10`000`40001:00030@0000`000D100000@0200030@0000P0
00<10000:00007h00`5C00030@0000/000<100400P001040004;00030@0002L0001k00<1EP000`40
000=00040@000@8000<100001P001040004Y0000N0030ET000<100003@020@800P4:0081:P0007D0
0`400`01001I00030@0000l000<10000=`0007800`5O00030@0000h00P4i0000K`030F8000<10000
40000`40000f0000K0030FD000<1000040020CL0001Z0081J0000`40000A00030@0003D0001W00<1
00<00@00J0000`40000@0081=P0006@00`5^00030@0000d00`400`01000e0000H@030G4000<10000
2P030@@000<10000=00005d0105d00030@0000H010460081=P000440104F00<1M`000`40000300<1
2`000`40000d0000@`000`40000B00<1N@020@800`4?00030@0003<0001300030@0000l00`5i00<1
00@00@414@020CD0001300030@0000`00`5i00<100H00@41004C00030@0002400P4800811P0004<0
00<100002@030GT00`4010010@4500030@00010000<100008@000`40000800030@0000@0001100<1
20030@03004007D0104010010@4900030@0000h00P4400813P030@8000<100000`4400030@0000P0
00<10000100004<000<100000`030C800P5400<100<00@400P4=00030@0000h000D100000@020003
0@0000`000@100010P000`40000500030@0000P000<10000100004H00`4f00030@0003l00`401001
0@4A00030@0000h000D100000@0200030@0000`00`4300<11@000`40000800030@0000@0001300<1
N0030@0400410AD000<100003@001@400001008000<1000030001040004200040@000@@000<10000
20000`4000040000@@030GL00`4010010@4I00030@0000/00P4300040@000@h000@100010`020@D0
00<1000020000`4000040000?P030@8000<10000LP030@0400410A`000<1000030001@4000010080
00<100002`040@/00P4800811P0003/00`5g00<100@00@4180000`40000;00810`020Bl0000h00<1
MP040@0400410B@000<100002`000`40000b0000=@030GH00`400`010@020BL000<100002`000`40
000b0000<P030@03004007<00`4010010@4/00030@0000X000<10000<P0002l00`5f00<100@00@41
<0000`4000080081=00002`00`5f00<100@00@41=0000`40000800030@000380000Y00<1MP030@04
00410CL000<1000020000`400003008110000`40000200@1800002H0105e00<100@00@41>`000`40
0006008110001040004800030@000240000S00<1MP030@0400410Cl000<100001P000`4000020004
0@000@T000<10000800002000`5e00P1@P000`40000600030@00008000@100012P000`40000O0000
7@030GD00`4010010@5700030@0000@00`4300040@000@/000<100007P0000<00P4E00<1M@030@04
00410D/000<100001@001@400001008000<100001P001040004P00000P001040004B00810P000`40
001`00<100@00@41C`000`40000400030@0000800P4:00818@00008000@100013`030G@00`401001
0@5B00030@0000<00P4c00000P001040004<00<1M0030@0400410EH000<100000`000`40000a0000
0P001040004900<1M0030@0400410EX000<100000P000`40000a00000P001040004600<100<00@00
E@020AT0205M00040@000081<`0000<00P4400<1G0000`40000D00<100@00@41HP001@40000103<0
000600<1L`030@0400410FH000@10001<`0000@00P5c00<100@00@41J@000`4000020@@00P440003
0@0000@000<100007P0000@00P5`00<100@00@41K@0010400@4300040@000@X000<100007P0000<0
0P400`01001[00<100@00@41L@000`400@0300040@000@L01@4O000010001040005X00<100@00@41
M@020@<000@100011`001040004P000010001@40000106@01`5h00810`001040004800030@010200
000300030@0000<000<10000G@070G`00P4300040@000@T00P4P00000`000`40000400030@0005T0
0`4010010@5l00<11P020@/000<100007P0000<000<100001@020EH00`4010010@5k00@1=`0000<0
00<100001`000`40001700811`030@0400410G/00`4k00000`000`40000800030@0004L000<10000
0P030@0400410G/00`4n00000`000`40000900030@0004P00`4010010@5j00@1@@0000<000<10000
2P000`40001400L1NP030DD0000300030@0000/000<10000@0060G/00`5800000`000`40000<0003
0@0003/01`5j00@1B`0000<000<100003@000`40000d008100D00@4100030GX00`5?00000P000`40
000?00030@0003<00`4010010@5j00<1DP00008000<1000040000`40000_00<100@00@41N@040ED0
000200030@00014000<10000:`030@0400410GT00`5I00000P000`40000B00030@0002L01`5i00<1
G000008000<100004`000`40000S00H1NP030El0000200030@0001@000<100007P070GT0105R0000
0P000`40000E00030@0001X00`4010010@5i00<1IP00008000<100005P020AL00`4010010@5i00<1
J@00008000<1000060000`40000A00<100@00@41N0040F`0000200030@0001T000<100003@030@04
00410GP00`5`000000<00@0070000`40000900L1N0030G<000000`01000M00030@0000D01`5g00@1
MP000003004001h000@100001P5h00<1NP000003004001h01`5h00<1O@000003004001/00`4200<1
MP040H0000000`01000H00<11@000`40001c00<1Q0000003004001D00`4800030@0007000`670000
00<00@004P030@/000<10000K@030HX000000`01000?00<120020@@000<10000J@040Hd000000`01
000<00<130000`40000200030@0006H00`6A000000<1000020040A@000<10000H`030I@000000`40
000500<15`000`40001P00@1U`0000030@0000800`4J00030@0005d00`6K000000<100400P4M0003
0@0005X00`6N00000P4P00030@0005H0106Q000000<00@0050020@T000<10000D`030JD000020003
0@0001<000<100001`000`40001@00<1Z00000<00P4M00030@0004d00`6[00001@000`40000J0003
0@0004T0106^00001P000`40000I00030@0004H00`6b00001`000`40000H00030@0004<00`6e0000
20000`40000G00030@0003l0106h00002@000`40000700813@000`40000l00<1_00000X000<10000
1`000`40000;00030@0003T00`6o00002`020A@000<10000=P040L80000=00030@00014000<10000
<`030LH0000>00030@00010000<10000<0030LT0000?00<13`000`40000/00@1c00001000P4?0003
0@0002T00`7@00004@000`40000=00030@0002H00`7C00004P000`40000<00030@0002<00`7F0000
4`020@`000<100007`040MT0000E00030@0000T000<1000070030Md0000F00030@0000P000<10000
6@030N00000G00030@0000L000<100005@040N<0000H00030@0000H000<100004P030NL0000I0003
0@0000@000<1000040030NX0000J00030@0000<000<1000030040Nd0000K00810`000`40000900<1
l@0001d000@1000120030O@0000N00030@0100@0107g00007`020@0400410O/0000P0081oP000?l0
8@000?l08@000?l08@000?l08@000?l08@000?l08@000?l08@000001\
\>"],
  ImageRangeCache->{{{101, 388}, {627.438, 462.938}} -> {-0.383221, 2.01151, 
  0.00408327, 0.00408327}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Nyquist plot definition and examples", "Subsection"],

Cell[CellGroupData[{

Cell["Nyquist[L];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5836 0.55979 0.309017 0.722653 [
[.02381 .29652 -6 -9 ]
[.02381 .29652 6 0 ]
[.16376 .29652 -15 -9 ]
[.16376 .29652 15 0 ]
[.3037 .29652 -12 -9 ]
[.3037 .29652 12 0 ]
[.44365 .29652 -15 -9 ]
[.44365 .29652 15 0 ]
[.72355 .29652 -12 -9 ]
[.72355 .29652 12 0 ]
[.8635 .29652 -9 -9 ]
[.8635 .29652 9 0 ]
[.5711 .01996 -24 -4.5 ]
[.5711 .01996 0 4.5 ]
[.5711 .16449 -24 -4.5 ]
[.5711 .16449 0 4.5 ]
[.5711 .45355 -18 -4.5 ]
[.5711 .45355 0 4.5 ]
[.5711 .59808 -18 -4.5 ]
[.5711 .59808 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
[(-1)] .02381 .29652 0 1 Mshowa
.16376 .30902 m
.16376 .31527 L
s
[(-0.75)] .16376 .29652 0 1 Mshowa
.3037 .30902 m
.3037 .31527 L
s
[(-0.5)] .3037 .29652 0 1 Mshowa
.44365 .30902 m
.44365 .31527 L
s
[(-0.25)] .44365 .29652 0 1 Mshowa
.72355 .30902 m
.72355 .31527 L
s
[(0.25)] .72355 .29652 0 1 Mshowa
.8635 .30902 m
.8635 .31527 L
s
[(0.5)] .8635 .29652 0 1 Mshowa
.125 Mabswid
.0518 .30902 m
.0518 .31277 L
s
.07979 .30902 m
.07979 .31277 L
s
.10778 .30902 m
.10778 .31277 L
s
.13577 .30902 m
.13577 .31277 L
s
.19175 .30902 m
.19175 .31277 L
s
.21974 .30902 m
.21974 .31277 L
s
.24773 .30902 m
.24773 .31277 L
s
.27572 .30902 m
.27572 .31277 L
s
.33169 .30902 m
.33169 .31277 L
s
.35968 .30902 m
.35968 .31277 L
s
.38767 .30902 m
.38767 .31277 L
s
.41566 .30902 m
.41566 .31277 L
s
.47164 .30902 m
.47164 .31277 L
s
.49963 .30902 m
.49963 .31277 L
s
.52762 .30902 m
.52762 .31277 L
s
.55561 .30902 m
.55561 .31277 L
s
.61159 .30902 m
.61159 .31277 L
s
.63958 .30902 m
.63958 .31277 L
s
.66757 .30902 m
.66757 .31277 L
s
.69556 .30902 m
.69556 .31277 L
s
.75154 .30902 m
.75154 .31277 L
s
.77953 .30902 m
.77953 .31277 L
s
.80752 .30902 m
.80752 .31277 L
s
.83551 .30902 m
.83551 .31277 L
s
.89148 .30902 m
.89148 .31277 L
s
.91947 .30902 m
.91947 .31277 L
s
.94746 .30902 m
.94746 .31277 L
s
.97545 .30902 m
.97545 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5836 .01996 m
.58985 .01996 L
s
[(-0.4)] .5711 .01996 1 0 Mshowa
.5836 .16449 m
.58985 .16449 L
s
[(-0.2)] .5711 .16449 1 0 Mshowa
.5836 .45355 m
.58985 .45355 L
s
[(0.2)] .5711 .45355 1 0 Mshowa
.5836 .59808 m
.58985 .59808 L
s
[(0.4)] .5711 .59808 1 0 Mshowa
.125 Mabswid
.5836 .05609 m
.58735 .05609 L
s
.5836 .09222 m
.58735 .09222 L
s
.5836 .12835 m
.58735 .12835 L
s
.5836 .20062 m
.58735 .20062 L
s
.5836 .23675 m
.58735 .23675 L
s
.5836 .27288 m
.58735 .27288 L
s
.5836 .34515 m
.58735 .34515 L
s
.5836 .38128 m
.58735 .38128 L
s
.5836 .41741 m
.58735 .41741 L
s
.5836 .48968 m
.58735 .48968 L
s
.5836 .52581 m
.58735 .52581 L
s
.5836 .56195 m
.58735 .56195 L
s
.25 Mabswid
.5836 0 m
.5836 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.5836 .30902 Mdot
.56296 .33191 Mdot
.52146 .31101 Mdot
.49069 .24335 Mdot
.49055 .14327 Mdot
.54623 .03734 Mdot
.65237 .01472 Mdot
.71017 .07771 Mdot
.72222 .12118 Mdot
.72715 .14171 Mdot
.72979 .15126 Mdot
.73147 .1467 Mdot
.74111 .12039 Mdot
.78387 .08062 Mdot
.86999 .08337 Mdot
.95048 .17484 Mdot
.97619 .30902 Mdot
.95048 .4432 Mdot
.86999 .53466 Mdot
.78387 .53742 Mdot
.74111 .49764 Mdot
.73147 .47134 Mdot
.72979 .46677 Mdot
.72715 .47633 Mdot
.72222 .49686 Mdot
.71017 .54032 Mdot
.65237 .60332 Mdot
.54623 .58069 Mdot
.49055 .47476 Mdot
.49069 .37469 Mdot
.52146 .30702 Mdot
.56296 .28612 Mdot
.03 w
.02381 .30902 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{225, 139},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003Q0000R`P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<n40002200030@0005`0002200030@0005`0
002200030@0005`0001`008110000`40000400030@00008000<10000G00006l000@100012P000`40
000200<1G00006l000@100011`050@<000<10000G00006T0104200040@000@L000@1000110000`40
001L0000K`001040004800030@0100@000<10000G00006l000@100012@020@@000<10000G0000700
0P4;00030@00008000<10000G000088000<10000G000088000<10000G00008800P5M0000PP000`40
001L0000PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40
001L0000PP000`40001L0000PP020Ed0002200030@0005`0002200030@0005`0002200030@0005`0
002200030@0005`0002200030@0005`0002200030@0005`0002200030@0005`000220081G@000880
00<10000G000088000<10000G000088000<10000G000088000<10000G000088000<10000G0000880
00<10000G00007000P4400030@000080104300030@0005`0001_00040@000@P000<1000010030E`0
001_00040@000@T000<100000`000`40001L0000J@040@8000@100012P000`40000200030@0005`0
001_00040@000@/000D100000@1N0000K`001040004800040@000@<000<10000G00007000P4:0081
10000`40001L0000PP000`40001L0000PP000`40001L0000PP020Ed0002200030@0005`000220003
0@0005`0002200030@0005`0002200030@0005`0002200030@0005`0002200030@0005`000220003
0@0005`000220081G@00088000<10000G000088000<10000G000088000<10000G000088000<10000
G000088000<10000G000088000<10000G000088000<10000G00000L0104C008110000`4000030003
0@0000800`4>008110000`40000200<13P020@@000<100000P040@800`4B008150020@@000<10000
0P040@800`4>008110000`40000200<16@0000T000<100004@001040004900030@0000D000<10000
2P001040004;00030@0000X000@1000120000`40000600030@0000l000<100004P00104000480003
0@0000H000<100002P001040004;00030@0001H0000900030@00014000@100012P000`4000040003
0@0000X000@100012`000`40000:00040@000@T000<100001@000`40000?00030@00018000@10001
2@000`40000500030@0000X000@100012`000`40000F000000<00@400P4400030@0000/010420004
0@000@X000@100000`4700@10P001040004800<11`040@8000@100012P00104000030A8000<10000
4P001040004:00040@0000<13@001040004800<16@0000T000<100004@001040004;00040@000@l0
00@1000120000`40000=00040@000@/000@1000150000`40000B00040@000@/000@100013`001040
004800030@0001T0000700<14`001040004800040@000@8000<100003@001040004800030@0000d0
00@1000120001040004200030@00018000<100004P001040004800040@000@8000<100003@001040
004800030@0001T0000400H150020@T0104200@13@020@T0104=00812P020@<0104A00030@0001<0
0P4:00810`040@d00P4900@1600000@01@5i00030@0005`0003P0@40000400D10`000`4000030003
0@0000<000<1000010000`40000300030@0000<000<100000`000`40000400030@0000<000<10000
0`000`40000300030@0000@000<100000`000`40000300030@0000<000<1000010000`4000030003
0@0000<000<100000`000`40000300030@0000@000<100000`000`40000300030@0000<000<10000
10000`40000300030@0000<000<100000`000`40000400030@0000<000<100000`000`4000030003
0@0000@000<100000`000`400004000010050A`000<1000070000`40000L00030@0001`000<10000
7@000`40000L00030@0001d0000600030@0007T000<10000G000088000<10000G000088000<10000
G000088000<10000G000088000<10000G000088000<10000G00008800P5M0000PP000`40001L0000
PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40001L0000
PP000`40001L0000PP020Ed0002200030@0005`0002200030@0005`0002200030@0005`000220003
0@0005`0002200030@0005`0002200030@0005`0002200030@0005`000220081G@00088000<10000
G000088000<10000G000088000<10000G000088000<10000G000088000<10000G00007000P440003
0@000080104300030@0005`0001_00040@000@P000<1000010000`40001L0000K`00104000490003
0@0000<00`5L0000K`001040004:00030@00008000<10000G00006l000@100012`001@40000105h0
001_00040@000@P000@100010`000`40001L0000L0020@X00P4400030@0005`0002200030@0005`0
002200030@0005`0002200030@0005`000220081G@00088000<10000G000088000<10000G0000880
00<10000G000088000<10000G000088000<10000G000088000<10000G000088000<10000G0000880
0P5M0000PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40001L0000PP000`40
001L0000PP000`40001L0000PP000`40001L0000PP020Ed0002200030@0005`0002200030@0005`0
002200030@0005`0002200030@0005`0002200030@0005`0002200030@0005`0001`008110000`40
000400030@00008000<10000G00006l000@100012P000`40000200<1G00006l000@100011`050@<0
00<10000G00006l000@100011`001040004400030@0005`0001_00040@000@P000<1004010000`40
001L0000K`0010400049008110000`40001L0000\
\>"],
  ImageRangeCache->{{{101, 325}, {277, 139}} -> {-1.85734, 0.434115, 
  0.00800585, 0.00620159}}]
}, Open  ]],

Cell["The points can be joined by typing the command :", "Text"],

Cell[CellGroupData[{

Cell["Nyquist[L,PlotJoined -> True];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5836 0.55979 0.309017 0.722653 [
[.02381 .29652 -6 -9 ]
[.02381 .29652 6 0 ]
[.16376 .29652 -15 -9 ]
[.16376 .29652 15 0 ]
[.3037 .29652 -12 -9 ]
[.3037 .29652 12 0 ]
[.44365 .29652 -15 -9 ]
[.44365 .29652 15 0 ]
[.72355 .29652 -12 -9 ]
[.72355 .29652 12 0 ]
[.8635 .29652 -9 -9 ]
[.8635 .29652 9 0 ]
[.5711 .01996 -24 -4.5 ]
[.5711 .01996 0 4.5 ]
[.5711 .16449 -24 -4.5 ]
[.5711 .16449 0 4.5 ]
[.5711 .45355 -18 -4.5 ]
[.5711 .45355 0 4.5 ]
[.5711 .59808 -18 -4.5 ]
[.5711 .59808 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
[(-1)] .02381 .29652 0 1 Mshowa
.16376 .30902 m
.16376 .31527 L
s
[(-0.75)] .16376 .29652 0 1 Mshowa
.3037 .30902 m
.3037 .31527 L
s
[(-0.5)] .3037 .29652 0 1 Mshowa
.44365 .30902 m
.44365 .31527 L
s
[(-0.25)] .44365 .29652 0 1 Mshowa
.72355 .30902 m
.72355 .31527 L
s
[(0.25)] .72355 .29652 0 1 Mshowa
.8635 .30902 m
.8635 .31527 L
s
[(0.5)] .8635 .29652 0 1 Mshowa
.125 Mabswid
.0518 .30902 m
.0518 .31277 L
s
.07979 .30902 m
.07979 .31277 L
s
.10778 .30902 m
.10778 .31277 L
s
.13577 .30902 m
.13577 .31277 L
s
.19175 .30902 m
.19175 .31277 L
s
.21974 .30902 m
.21974 .31277 L
s
.24773 .30902 m
.24773 .31277 L
s
.27572 .30902 m
.27572 .31277 L
s
.33169 .30902 m
.33169 .31277 L
s
.35968 .30902 m
.35968 .31277 L
s
.38767 .30902 m
.38767 .31277 L
s
.41566 .30902 m
.41566 .31277 L
s
.47164 .30902 m
.47164 .31277 L
s
.49963 .30902 m
.49963 .31277 L
s
.52762 .30902 m
.52762 .31277 L
s
.55561 .30902 m
.55561 .31277 L
s
.61159 .30902 m
.61159 .31277 L
s
.63958 .30902 m
.63958 .31277 L
s
.66757 .30902 m
.66757 .31277 L
s
.69556 .30902 m
.69556 .31277 L
s
.75154 .30902 m
.75154 .31277 L
s
.77953 .30902 m
.77953 .31277 L
s
.80752 .30902 m
.80752 .31277 L
s
.83551 .30902 m
.83551 .31277 L
s
.89148 .30902 m
.89148 .31277 L
s
.91947 .30902 m
.91947 .31277 L
s
.94746 .30902 m
.94746 .31277 L
s
.97545 .30902 m
.97545 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5836 .01996 m
.58985 .01996 L
s
[(-0.4)] .5711 .01996 1 0 Mshowa
.5836 .16449 m
.58985 .16449 L
s
[(-0.2)] .5711 .16449 1 0 Mshowa
.5836 .45355 m
.58985 .45355 L
s
[(0.2)] .5711 .45355 1 0 Mshowa
.5836 .59808 m
.58985 .59808 L
s
[(0.4)] .5711 .59808 1 0 Mshowa
.125 Mabswid
.5836 .05609 m
.58735 .05609 L
s
.5836 .09222 m
.58735 .09222 L
s
.5836 .12835 m
.58735 .12835 L
s
.5836 .20062 m
.58735 .20062 L
s
.5836 .23675 m
.58735 .23675 L
s
.5836 .27288 m
.58735 .27288 L
s
.5836 .34515 m
.58735 .34515 L
s
.5836 .38128 m
.58735 .38128 L
s
.5836 .41741 m
.58735 .41741 L
s
.5836 .48968 m
.58735 .48968 L
s
.5836 .52581 m
.58735 .52581 L
s
.5836 .56195 m
.58735 .56195 L
s
.25 Mabswid
.5836 0 m
.5836 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.5836 .30902 m
.56296 .33191 L
.52146 .31101 L
.49069 .24335 L
.49055 .14327 L
.54623 .03734 L
.65237 .01472 L
.71017 .07771 L
.72222 .12118 L
.72715 .14171 L
.72979 .15126 L
.73147 .1467 L
.74111 .12039 L
.78387 .08062 L
.86999 .08337 L
.95048 .17484 L
.97619 .30902 L
.95048 .4432 L
.86999 .53466 L
.78387 .53742 L
.74111 .49764 L
.73147 .47134 L
.72979 .46677 L
.72715 .47633 L
.72222 .49686 L
.71017 .54032 L
.65237 .60332 L
.54623 .58069 L
.49055 .47476 L
.49069 .37469 L
.52146 .30702 L
.56296 .28612 L
s
.03 w
.02381 .30902 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{233, 143.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003Y0000S`P10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<hL000<10000G`0008L000<10000G`0008L0
00<10000G`0007@00P4400030@0000@000<100000`000`40000;00<1D@0007<000@100012P000`40
000300<11P050@<000<10000CP0007<000@100011`050@@000@100001@4900030@0004d0001]00@1
0P001040004700040@000@@01@4?00030@0004`0001c00040@000@P000<1004010400`01000A0003
0@0004`0001c00040@000@L0104500030@00014000<10000B`0007@00P4800040@000@D000<10000
4P000`40001:0000O@000`40000700030@0001<000<10000B@0007d000<100001`000`40000D0003
0@0004P0001l00030@0000P000<100005@000`4000170000O0000`40000800815`000`4000160000
N`000`40000900030@0001L000<10000A@0007/000<100002@000`40000G00030@0004D0001j0003
0@0000X000<1000060000`4000140000NP000`40000:00030@0001T000<10000@`0007T000<10000
2`000`40000J00030@0000h02`4Y0000N@000`40000;00030@0001X000<100003@000`40000900X1
7`0007P000<1000030000`40000K00030@0000/000<1000050000`40000L0000N0000`40000<0081
70000`40000:00030@0001H000<100006`0007L000<100003@000`40000K00030@0000T000<10000
60000`40000J0000M`000`40000=00030@0001`000<100001`000`40000J00030@0001T0001f0003
0@0000h000<1000070000`40000600030@0001`000<10000600007H000<100003P000`40000L0003
0@0000D000<100007@000`40000H0000M@000`40000?00030@0001`000<1000010000`40000O0003
0@0001L0001e00030@0000l000<100007@000`40000200030@00024000<100005P0007@000<10000
40000`40000M00030@00008000<100008P000`40000E0000M0000`40000@00817P001@40000102H0
00<10000500007<000<100004@000`40000N00040@000BL000<100004`0007<000<100004@000`40
000N00030@0102T000<100004P00078000<100004P000`40000N00030@0102X000<100004@000780
00<100004P000`40000N0081;0000`40000@0000LP000`40000B00030@0001l000<10000:`000`40
000?0000LP000`40000B00030@0001l000<10000:`000`40000?0000LP0010400@4400030@000080
104400030@0004h000<100003P0007800P4200030@0000H000<100001@000`40001?00030@0000d0
001b00810P000`40000700030@0000@00`5@00030@0000`0001]00@100<00@400P000`4000080003
0@0000<000<10000D@000`40000;0000LP020@8000<100002@000`40000200030@00058000<10000
2P0007800P4200030@0000H000@1000110000`40001B00030@0000X0001b00040@010@X00P450003
0@00058000<100002P00078000<100004P000`40001C00030@0000T0001b00030@00018000<10000
D`000`4000090000LP000`40000B00030@0005<000<100002@00078000<100004P020E@000<10000
2@00078000<100004P000`40001C00030@0000T0001b00030@00018000<10000E0000`4000080000
LP000`40000B00030@0005@000<100002000078000<100004P000`40001D00030@0000P0001b0003
0@00018000<10000E0000`4000080000LP000`40000B00030@0005@000<100002000078000<10000
4P000`40001E00030@0000L0001b00030@0001800P5F00030@0000L0001b00030@00018000<10000
E@000`4000070000LP000`40000B00030@0005D000<100001`00078000<100004P000`40001E0003
0@0000L0001c00030@00014000<10000E@000`4000070000L`000`40000A00030@0005H000<10000
1P0007@000<1000040000`40001F00030@0000H0001d00030@00010000<10000EP000`4000060000
M@000`40000?00030@0005H000<100001P0000L0104D008110000`40000300030@0000800`4?0081
10000`40000200<140020@@000<100000P040@800`400`01000@008150020@@000<100000P040@80
0`4@008110000`40000200<14@000`40000600002@000`40000B00040@000@T000<100001@000`40
000;00040@000@/000<1000030001040004800030@0000H000<1004040000`40000B00040@000@P0
00<100001P000`40000<00040@000@/000<100003`000`40000500002@000`40000B00040@000@X0
00<1000010000`40000;00040@000@/000<1000030001040004900030@0000D000<1004040000`40
000B00040@000@T000<100001@000`40000<00040@000@/000<100003`000`400005000000<00@40
0P4400030@0000`0104200040@000@X000@100000`4800@10P001040004800<12@040@8000@10001
2P00104000030@<000<1000020000`40000200030@00018000@100012P00104000030@l000@10001
20030A8000<100001@0000T000<100004P001040004;00040@000A0000@1000120000`40000?0004
0@000@/000@100011@000`40000600811@000`40000B00040@000@/000@100014@00104000480003
0@00018000<100001@0000L00`4D00040@000@P000@100010P000`40000>00040@000@P000<10000
3`001040004800040@000@8000<1000010000`40000300811`000`40000B00040@000@P000@10001
0P000`40000?00040@000@P000<100004P000`400005000010060AD00P4900@10P040@h00P4900@1
3`020@X00P4300@10`00104000020@T000<100004`020@X00P4300@13`020@T0104B00030@0000@0
000400D1L0030@/000<10000F0000`4000040000j041000010050@<000<1000010000`4000030003
0@0000@000<100000`000`40000300030@0000@000<100000`000`40000400030@0000<000<10000
10000`40000300030@0000@000<100000`000`40000400030@0000<000<1000010000`40000200<1
1@000`40000200811P000`40000300030@0000@000<100000`000`40000300030@0000@000<10000
0`000`40000400030@0000<000<1000010000`40000300030@0000@000<100000`000`4000040003
0@0000@0000400D1K`00104000020@L000<10040FP000`40000400001P000`40001_00030@0000<0
0P4400040@000ET000<100001@0007L000<100001P020@030040008000<10000E`000`4000050000
M`000`40000800030@00008000<10000E`000`4000050000MP000`40000>00030@0005L000<10000
1@0007H000<100003P000`40001G00030@0000D0001e00030@0000l00P5G00030@0000H0001e0003
0@0000l000<10000EP000`4000060000M0000`40000@00030@0005H000<100001P0007@000<10000
40000`40001F00030@0000H0001c00030@00014000<10000EP000`4000060000L`000`40000A0003
0@0005D000<100001`00078000<100004P000`40001E00030@0000L0001b00030@00018000<10000
E@000`4000070000LP000`40000B00030@0005D000<100001`00078000<100004P020EH000<10000
1`00078000<100004P000`40001E00030@0000L0001b00030@00018000<10000E0000`4000080000
LP000`40000B00030@0005@000<100002000078000<100004P000`40001D00030@0000P0001b0003
0@00018000<10000E0000`4000080000LP000`40000B00030@0005@000<100002000078000<10000
4P000`40001C00030@0000T0001b00030@0001800P5D00030@0000T0001b00030@00018000<10000
D`000`4000090000LP000`40000B00030@0005<000<100002@00078000<100004P000`40001C0003
0@0000T0001b00030@00018000<10000DP000`40000:0000LP000`40000B00030@00058000<10000
2P00078000<100004P000`40001B00030@0000X0001b00040@010@@000<100000P040@@000<10000
D@000`40000;0000LP020@8000<100001P000`40000500<1D0000`40000<0000LP020@8000<10000
1`000`40000400030@0004l000<100003@0007800P4200030@0000P000<100000`000`40001>0003
0@0000h0001b00810P000`40000900030@00008000<10000C@000`40000?0000LP020@8000<10000
1P001040004400030@0001l000<10000:`000`40000?0000LP0010400@4:00811@000`40000O0003
0@0002X000<100004000078000<100004P000`40000N00030@0102X000<100004@0007<000<10000
4@000`40000N00030@0102T000<100004P0007<000<100004@000`40000N00040@000BL000<10000
4`0007@000<1000040020Ah000D100000@0V00030@0001@0001d00030@00010000<100007@000`40
000200030@00028000<100005@0007D000<100003`000`40000M00030@00008000<100008@000`40
000F0000M@000`40000?00030@0001`000<1000010000`40000O00030@0001L0001f00030@0000h0
00<1000070000`40000500030@0001d000<10000600007H000<100003P000`40000L00030@0000H0
00<1000070000`40000H0000M`000`40000=00030@0001`000<100001`000`40000J00030@0001T0
001g00030@0000d000<100006`000`40000900030@0001P000<100006P0007P000<1000030020A`0
00<100002P000`40000F00030@0001/0001h00030@0000`000<100006`000`40000;00030@0001@0
00<10000700007T000<100002`000`40000K00030@0000`000<10000200;0Al0001i00030@0000/0
00<100006P000`40000>00X1:P0007X000<100002P000`40000J00030@000480001j00030@0000X0
00<100006@000`4000130000N`000`40000900030@0001P000<10000A00007/000<100002@000`40
000G00030@0004D0001l00030@0000P000<100005P000`4000160000O0000`40000800815P000`40
00170000O@000`40000700030@0001@000<10000B00007d000<100001`000`40000D00030@0004P0
001n00030@0000H000<100004`000`4000190000OP030@H000<100004P000`40001:0000P@050@03
0040018000<10000B`0008H01@4?00030@0004`0001d008110000`40000400030@0000<000@10000
1@4900030@0004d0001c00040@000@X000<100000`030@H01@4300030@0004h0001c00040@000@L0
1@4400030@0000/00`5A0000L`001040004700040@000@D000<10000G`0007<000@1000120000`40
0@0500030@0005l0001c00040@000@T00P4500030@0005l00000\
\>"],
  ImageRangeCache->{{{101, 333}, {474.938, 332}} -> {-1.82736, 1.55882, 
  0.00772401, 0.00598326}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Nichols plot of gridvalues", "Subsection"],

Cell[CellGroupData[{

Cell["Nichols[L,PlotJoined -> True];", "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.97619 0.00756266 0.833667 0.0747467 [
[.06867 -0.0125 -12 -9 ]
[.06867 -0.0125 12 0 ]
[.21992 -0.0125 -12 -9 ]
[.21992 -0.0125 12 0 ]
[.37118 -0.0125 -9 -9 ]
[.37118 -0.0125 9 0 ]
[.52243 -0.0125 -9 -9 ]
[.52243 -0.0125 9 0 ]
[.67368 -0.0125 -9 -9 ]
[.67368 -0.0125 9 0 ]
[.82494 -0.0125 -9 -9 ]
[.82494 -0.0125 9 0 ]
[.97619 -0.0125 -3 -9 ]
[.97619 -0.0125 3 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -11 -24 ]
[.5 -0.0125 11 -9 ]
[-0.0125 .0862 -18 -4.5 ]
[-0.0125 .0862 0 4.5 ]
[-0.0125 .23569 -12 -4.5 ]
[-0.0125 .23569 0 4.5 ]
[-0.0125 .38519 -12 -4.5 ]
[-0.0125 .38519 0 4.5 ]
[-0.0125 .53468 -12 -4.5 ]
[-0.0125 .53468 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34 -7.5 ]
[-0.01875 .30902 -18 7.5 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.06867 0 m
.06867 .61803 L
s
.21992 0 m
.21992 .61803 L
s
.37118 0 m
.37118 .61803 L
s
.52243 0 m
.52243 .61803 L
s
.67368 0 m
.67368 .61803 L
s
.82494 0 m
.82494 .61803 L
s
0 .0862 m
1 .0862 L
s
0 .23569 m
1 .23569 L
s
0 .38519 m
1 .38519 L
s
0 .53468 m
1 .53468 L
s
0 g
.06867 0 m
.06867 .00625 L
s
[(-120)] .06867 -0.0125 0 1 Mshowa
.21992 0 m
.21992 .00625 L
s
[(-100)] .21992 -0.0125 0 1 Mshowa
.37118 0 m
.37118 .00625 L
s
[(-80)] .37118 -0.0125 0 1 Mshowa
.52243 0 m
.52243 .00625 L
s
[(-60)] .52243 -0.0125 0 1 Mshowa
.67368 0 m
.67368 .00625 L
s
[(-40)] .67368 -0.0125 0 1 Mshowa
.82494 0 m
.82494 .00625 L
s
[(-20)] .82494 -0.0125 0 1 Mshowa
.97619 0 m
.97619 .00625 L
s
[(0)] .97619 -0.0125 0 1 Mshowa
.125 Mabswid
.10648 0 m
.10648 .00375 L
s
.1443 0 m
.1443 .00375 L
s
.18211 0 m
.18211 .00375 L
s
.25774 0 m
.25774 .00375 L
s
.29555 0 m
.29555 .00375 L
s
.33336 0 m
.33336 .00375 L
s
.40899 0 m
.40899 .00375 L
s
.4468 0 m
.4468 .00375 L
s
.48462 0 m
.48462 .00375 L
s
.56024 0 m
.56024 .00375 L
s
.59806 0 m
.59806 .00375 L
s
.63587 0 m
.63587 .00375 L
s
.7115 0 m
.7115 .00375 L
s
.74931 0 m
.74931 .00375 L
s
.78712 0 m
.78712 .00375 L
s
.86275 0 m
.86275 .00375 L
s
.90056 0 m
.90056 .00375 L
s
.93838 0 m
.93838 .00375 L
s
.03086 0 m
.03086 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -72 -28 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(deg) show
81.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 .0862 m
.00625 .0862 L
s
[(-10)] -0.0125 .0862 1 0 Mshowa
0 .23569 m
.00625 .23569 L
s
[(-8)] -0.0125 .23569 1 0 Mshowa
0 .38519 m
.00625 .38519 L
s
[(-6)] -0.0125 .38519 1 0 Mshowa
0 .53468 m
.00625 .53468 L
s
[(-4)] -0.0125 .53468 1 0 Mshowa
.125 Mabswid
0 .12357 m
.00375 .12357 L
s
0 .16095 m
.00375 .16095 L
s
0 .19832 m
.00375 .19832 L
s
0 .27307 m
.00375 .27307 L
s
0 .31044 m
.00375 .31044 L
s
0 .34781 m
.00375 .34781 L
s
0 .42256 m
.00375 .42256 L
s
0 .45993 m
.00375 .45993 L
s
0 .49731 m
.00375 .49731 L
s
0 .04883 m
.00375 .04883 L
s
0 .01145 m
.00375 .01145 L
s
0 .57205 m
.00375 .57205 L
s
0 .60943 m
.00375 .60943 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95 -11.5 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23 translate 1 -1 scale
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(dB) show
75.000000 15.000000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.06867 .61178 m
.06867 .61803 L
s
.21992 .61178 m
.21992 .61803 L
s
.37118 .61178 m
.37118 .61803 L
s
.52243 .61178 m
.52243 .61803 L
s
.67368 .61178 m
.67368 .61803 L
s
.82494 .61178 m
.82494 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.10648 .61428 m
.10648 .61803 L
s
.1443 .61428 m
.1443 .61803 L
s
.18211 .61428 m
.18211 .61803 L
s
.25774 .61428 m
.25774 .61803 L
s
.29555 .61428 m
.29555 .61803 L
s
.33336 .61428 m
.33336 .61803 L
s
.40899 .61428 m
.40899 .61803 L
s
.4468 .61428 m
.4468 .61803 L
s
.48462 .61428 m
.48462 .61803 L
s
.56024 .61428 m
.56024 .61803 L
s
.59806 .61428 m
.59806 .61803 L
s
.63587 .61428 m
.63587 .61803 L
s
.7115 .61428 m
.7115 .61803 L
s
.74931 .61428 m
.74931 .61803 L
s
.78712 .61428 m
.78712 .61803 L
s
.86275 .61428 m
.86275 .61803 L
s
.90056 .61428 m
.90056 .61803 L
s
.93838 .61428 m
.93838 .61803 L
s
.03086 .61428 m
.03086 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .0862 m
1 .0862 L
s
.99375 .23569 m
1 .23569 L
s
.99375 .38519 m
1 .38519 L
s
.99375 .53468 m
1 .53468 L
s
.125 Mabswid
.99625 .12357 m
1 .12357 L
s
.99625 .16095 m
1 .16095 L
s
.99625 .19832 m
1 .19832 L
s
.99625 .27307 m
1 .27307 L
s
.99625 .31044 m
1 .31044 L
s
.99625 .34781 m
1 .34781 L
s
.99625 .42256 m
1 .42256 L
s
.99625 .45993 m
1 .45993 L
s
.99625 .49731 m
1 .49731 L
s
.99625 .04883 m
1 .04883 L
s
.99625 .01145 m
1 .01145 L
s
.99625 .57205 m
1 .57205 L
s
.99625 .60943 m
1 .60943 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.97619 .60332 m
.85657 .58441 L
.73875 .50133 L
.66266 .35358 L
.65213 .21194 L
.67084 .146 L
.6745 .13397 L
.65798 .14365 L
.62537 .16857 L
.56204 .22545 L
.42249 .27871 L
.21941 .20858 L
.02381 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{231, 142.688},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40003W0000SPP10`4000010?ooo`0007=cLjjb`k9=
NYM1Imb]`ULU>oooojjb`i>GYMcNiEePJ@00076;YF1fS;o;ecQ5D_ooooogjMSAa/c5^YVDR`0009BD
U:f][B4Q8Kfm_MkNgV=SHol0003o0000oooo003oool0od92@^7QhLS8b:n_[iJFUWemOFATI4];Bc8b
<_ogjIfP[eMIHM7Ej@00?`00O`00_`0o000o?`0oO`0o_`0oo`1o001o?`1oO`1o_`1oo`2o002o?`2o
O`2o_`2oo`3o?`3oO`3o_cl003l0?cl0Ocl0_cl0oclo03lo?cloOclo_cloocmo03mo?cmoOcmo_cmo
ocno03no?cnoOcno_cnoocoo03oo?cooOcoo_cooogl007l0?gl0Ogl0_gl0oglo07lo?gloOglo_glo
ogmo07mo?gmoOgmo_gmoogno07no?gnoOgno_gnoogoo07oo?gooOgoo_goookl00;l0?kl0Okl0_kl0
oklo0;lo?kloOklo_klookmo0;mo?kmoOkmo_kmookno0;no?knoOkno_knookoo0;oo?kooOkoo_koo
ool0?ol0Ool0_olo0?lo?oloOolo_olooomo0?mo?omoOomo_omooono0?no?onoOono_onooooo?ooo
Oooo_mgNiUiPJIBGYZ0Pl=WIfNo_kgalO>KViYbLW9VIIYVIVIVIc9VIoiW<09W<<iW<IYW<VIW<c9W<
oiWo09Wo<iWoIYWoVIWoc9Wool`00<`0<l`0I/`0VL`0c<`0ol`c0<`c<l`cI/`cVL`cc<`colaV0<aV
<laVI/aVVLaVc<aVolbI0<bI<lbII/bIVLbIc<bIolc<0<c<<lc<I/c<VLc<c<c<olco0<co<lcoI/co
VLcoc<coool0<ol0I_l0VOl0c?lc0?lc<olcI_lcVOlcc?lcoomV0?mV<omVI_mVVOmVc?mVoonI0?nI
<onII_nIVOnIc?nIooo<0?o<<oo<I_o<VOo<c?o<oooo<nL0003W0000i`0008T00`5K0000S0000`40
001H0000OP020@04004000<10`030EX0001m00040@000@8000<100000`001040005J0000O@001040
004200<10`001040005J0000O@001040004200040@000@8000@10001FP0007h00`43008110020@03
004005P0002000030@0006@0001o0081IP000>L0003W0000i`000>L0003W0000i`000>L0000`00@1
0P040@<00P4=00@10`020@@00P4A008110020A@00P4400815P000`40000200814`040@<00P4D0081
2000038000D100000@0500040@000@h000D100000@0200040@000@8000<100003@00104000420004
0@000A8000@100010P001040004E00050@0000400P000`40000@00030@0000<000@100014P001040
00470000<P000`40000200030@00008000@100013P001@400001008000@100010P000`40000=0004
0@000@8000@100014P001040004200040@000A801@4200040@000A<000<100000P001040004B0004
0@000@L0000Z00@110000`40000300050@0000400P000`40000400@110001@400001008000@10001
0P000`40000700@10`020@<000@1000130040@800`4300040@000@d010401@01000100<000@10001
30040@@000D100000@0200030@00010000@100011`00038000<1000010001040004200030@0000`0
00D100000@0200040@000@8000<100003@001040004200040@000A8000<100000`001040004C0003
0@0100<000@100015@001040004200030@00010000@100011`0003000`4300040@000@8000@10001
30030@<000@100010P001040004?00040@000@8000@100014P000`40000300040@000A@00P430004
0@000A8000@100010P001040004B00040@000@L0000b00030@0000800P4400813`000`4000020081
10020A400P44008150030@<00P4F00030@0000800P4D008110020A@00P480000:02l0@<0000X0003
0@0000<000<1000010000`40000400030@0000@000<1000010000`40000400030@0000@000<10000
10000`40000400030@0000@000<1000010000`40000400030@0000@000<1000010000`4000040003
0@0000@000<100001@000`40000400030@0000@000<1000010000`40000400030@0000@000<10000
10000`40000400030@0000@000<1000010000`40000200030@000040000X00812`000c80000I0003
<P0001T000<b00006@000c80000J0003<P0001T000<b00007@020@<0000X00050@00004020000c80
000I0003<P0001T000<b00006@000c80000J0003<P0001T000<b00007P000`4000010000:0000`40
000200030@0000D000<b00006@000c80000I0003<P0001T000<b00006P000c80000I0003<P0001h0
00<100000@0002P000<100000`000`4000040003<P0001T000<b00006@000c80000I0003<P0001X0
00<b00006@000c80000N00030@000040000X00030@0000@000<100000`000c80000I0003<P0001T0
00<b00006@000c80000J0003<P0001T000<b00007P000`4000010000:0000`40000500030@000080
00<b00006@000c80000I0003<P0001T000<b00006P000c80000I0003<P0001h000<100000@0002P0
00<100001P001@40000b01/000<b00006@000c80000I0003<P0001X000<b00006@000c80000N0003
0@000040000X008120001040038K0003<P0001T000<b00006@000c80000J0003<P0001T000<b0000
7@020@<0000X00030@0000P000<103806`000c80000I0003<P0001T000<b00006P000c80000I0003
<P0001h000<100000@0002P000<100002@000`4b000J0003<P0001T000<b00006@000c80000J0003
<P0001T000<b00007P000`4000010000:0000`40000:00030@0001T000<b00006@000c80000I0003
<P0001X000<b00006@000c80000N00030@000040000X00030@0000X000<b0@006@000c80000I0003
<P0001T000<b00006P000c80000I0003<P0001h000<100000@0002P000<100002P000c800@0I0003
<P0001T000<b00006@000c80000J0003<P0001T000<b00007P000`40000100006`040@<00P440003
0@0000X000@b000160000c80000I0003<P0001T000<b00006P000c80000I0003<P0001h000<10000
0@0001d000D100000@0200040@0000813c800`4b<P2U<P<10`0001d000D100000@0200050@000040
30000c80000200030@0001@000<b00006@000c80000I0003<P0001X000<b00006@000c80000N0003
0@000040000E00@110001@400001008000D100000@0<0003<P0000<000<100004`000c80000I0003
<P0001T000<b00006P000c80000I0003<P0001h000<100000@0001d000D100000@0200050@000040
30000c80000400030@00018000<b00006@000c80000I0003<P0001X000<b00006@000c80000N0003
0@000040000K00<10`001040004300030@0000X000<b00001@000`40000A0003<P0001T000<b0000
6@000c80000J0003<P0001T000<b00007P000`40000100007@000`400002008110000`40000:0003
<P0000H00P4A0003<P0001T000<b00006@000c80000J0003<P0001T000<b00007P000`4000010000
:0000`40000:0003<P0000P000<100003P000c80000I0003<P0001T000<b00006P000c80000I0003
<P0001h000<100000@0002P00P4;0003<P0000T000<100003@000c80000I0003<P0001T000<b0000
6P000c80000I0003<P0001d00P430000:0000`40000:0003<P0000X000<1000030000c80000I0003
<P0001T000<b00006P000c80000I0003<P0001h000<100000@0002P000<100002P000c80000;0003
0@0000/000<b00006@000c80000I0003<P0001X000<100006@000c80000N00030@000040000X0003
0@0000X000<b000030000`40000:0003<P0001T000<b00006@000c80000H00<16`000c80000N0003
0@000040000X00030@0000X000<b00003@000`4000090003<P0001T000<b00006@000c80000G0004
0@01<Q/000<b00007P000`4000010000:0000`40000:0003<P0000h000<1000020000c80000I0003
<P0001T000<b00005P001@40004b01/000<b00007P000`4000010000:0000`40000:0003<P0000l0
00<100001`000c80000I0003<P0001T000<b000050020@<000<1<P006P000c80000N00030@000040
000X00812`000c80000@00030@0000H000<b00006@000c80000I0003<P0001<000<100000P000`40
<P0K0003<P0001d00P430000:0000`40000:0003<P00014000<100001@000c80000I0003<P0001T0
00<b00004P000`40000300030@0b01/000<b00007P000`4000010000:0000`40000:0003<P000180
00<1000010000c80000I0003<P0001T000<b00004@000`40000400030@0b01/000<b00007P000`40
00010000:0000`40000:0003<P0001<000<100000`000c80000I0003<P0001T000<b000040000`40
000500030@0b01/000<b00007P000`4000010000:0000`40000:0003<P0001@000<100000P000c80
000I0003<P0001T000<b00003P020@L000@1000b6`000c80000N00030@000040000X00030@0000X0
00<b00005@001@40000b01/000<b00006@000c80000=00030@0000L000@1000b6`000c80000N0003
0@000040000X00030@0000X000<b00005P001040038K0003<P0001T000<b000030000`4000080004
0@00<Q/000<b00007P000`4000010000:0020@/000<b00005`000`40<P0K0003<P0001T000<b0000
2`000`40000900040@00<Q/000<b00007@020@<0000X00030@0000X000<b000060000`4b000J0003
<P0001T000<b00002P000`40000900050@0003806`000c80000N00030@000040000X00030@0000X0
00<b00006@020AX000<b00006@000c800008008130001@40000b01/000<b00007P000`4000010000
:0000`40000:0003<P0001T000<b00400P4G0003<P0001T000<b00001`000`40000<00050@000380
6`000c80000N00030@000040000X00030@0000X000<b00006@000c80000200<150000c80000I0003
<P0000H000<100003@001@40000b01/000<b00007P000`4000010000:0000`40000:0003<P0001T0
00<b00001@030A4000<b00006@000c800004008140001@40000b01/000<b00007P000`4000010000
8P020@@000<100002P000c80000I0003<P0000P00`4>0003<P0001T000<b00000P020A8000D10000
<P0K0003<P0001h000<100000@00024000@100010`020CDb0`4Y<P<153800`4b<P0l<P<10`000240
00@100010`000`40000:0003<P0001T000<b00003P030@P000<b000060030AL000D10000<P0K0003
<P0001h000<100000@0001/01043008110000`40000:0003<P0001T000<b00004@020@H000<b0000
5P020@03038001L000D10000<P0K0003<P0001h000<100000@00024000@100010`000`40000:0003
<P0001T000<b00004`030@<000<b00004`030@<000<b00005`001040038K0003<P0001h000<10000
0@00024000@100010`000`40000:0003<P0001T000<b00005P030@03<P0001400P460003<P0001L0
00@1000b6`000c80000N00030@000040000R008110000`40000:0003<P0001T000<b00006@030@h0
0`480003<P0001L000@1000b6`000c80000N00030@000040000X00030@0000X000<b00006@000c80
000I0003<P0000<120030@/000<b00005`001040038K0003<P0001h000<100000@0002P00P4;0003
<P0001T000<b00006@000c80000300<10`020@h000<b00005`001040038K0003<P0001d00P430000
:0000`40000:0003<P0001T000<b00006@000c80000600<140000c80000G00040@00<Q/000<b0000
7P000`4000010000:0000`40000:0003<P0001T000<b00006@000c80000I0003<P0001L000@1000b
6`000c80000N00030@000040000X00030@0000X000<b00006@000c80000I0003<P0001T000<b0000
5`001040038K0003<P0001h000<100000@0000H00P400`010@030AX000<100002P000c80000I0003
<P0001T000<b00006@000c80000G00040@00<Q/000<b00007P000`40000100001@00104000420004
0@000AT000<100002P000c80000I0003<P0001T000<b00006@000c80000G00040@00<Q/000<b0000
7P000`40000100001@001040004200040@000AT000<100002P000c80000I0003<P0001T000<b0000
6@000c80000G00040@00<Q/000<b00007P000`40000100001@001040004200<16P020@/000<b0000
6@000c80000I0003<P0001T000<b00005`001040038K0003<P0001d00P4300001P030@8000@10001
6@000`40000:0003<P0001T000<b00006@000c80000I0003<P0001L000@1000b6`000c80000N0003
0@000040000800030@0100<16P000`40000:0003<P0001T000<b00006@000c80000I0003<P0001P0
00<103806`000c80000N00030@000040000700817`000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001P000<103806`000c80000N00030@000040000X00030@0000X000<b00006@000c80
000I0003<P0001T000<b000060000`40<P0K0003<P0001h000<100000@0002P000<100002P000c80
000I0003<P0001T000<b00006@000c80000H00030@0b01/000<b00007P000`4000010000:0000`40
000:0003<P0001T000<b00006@000c80000I0003<P0001P000<103806`000c80000N00030@000040
000X00812`000c80000I0003<P0001T000<b00006@000c80000H00030@0b01/000<b00007@020@<0
000X00030@0000X000<b00006@000c80000I0003<P0001T000<b000060000`40<P0K0003<P0001h0
00<100000@0002P000<100002P000c80000I0003<P0001T000<b00006@000c80000I00030C8001X0
00<b00007P000`4000010000:0000`40000:0003<P0001T000<b00006@000c80000I0003<P0001T0
00<1<P006P000c80000N00030@000040000X00030@0000X000<b00006@000c80000I0003<P0001T0
00<b00006P000`40000I0003<P0001h000<100000@0002P000<100002P000c80000I0003<P0001T0
00<b00006@000c80000J00030@0001T000<b00007P000`40000100008P020@@000<100002P000c80
000I0003<P0001T000<b00006@000c80000J0003<P4001T000<b00007P000`40000100008@001040
00430081OC800`4b<P0g<P<10`00024000@100010`000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001X000<b00406@000c80000N00030@000040000K00@10P030@@000<100002P000c80
000I0003<P0001T000<b00006@000c80000J0003<P0101T000<b00007P000`40000100008@000`40
000400030@0000X000<b00006@000c80000I0003<P0001T000<b00006P001380004H0003<P0001h0
00<100000@00024000<1000010000`40000:0003<P0001T000<b00006@000c80000I0003<P0001X0
00@b000160000c80000N00030@000040000R00<10`000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001X000Db00000@0G0003<P0001h000<100000@0002P000<100002P000c80000I0003
<P0001T000<b00006@000c80000J0005<P0000405`000c80000N00030@000040000X00812`000c80
000I0003<P0001T000<b00006@000c80000J0003<P00008000<1000050000c80000M00810`0002P0
00<100002P000c80000I0003<P0001T000<b00006@000c80000J0003<P00008000<1000050000c80
000N00030@000040000X00030@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80
000300030@0001<000<b00007P000`4000010000:0000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001X000<b00000`000`40000C0003<P0001h000<100000@0002P000<100002P000c80
000I0003<P0001T000<b00006@000c80000J0003<P0000@000<100004P000c80000N00030@000040
000X00030@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80000400030@000180
00<b00007P000`4000010000:0000`40000:0003<P0001T000<b00006@000c80000I0003<P0001X0
00<b00001@000`40000A0003<P0001h000<100000@0002P00P4;0003<P0001T000<b00006@000c80
000I0003<P0001X000<b00001@000`40000A0003<P0001d00P430000:0000`40000:0003<P0001T0
00<b00006@000c80000I0003<P0001X000<b00001P000`40000@0003<P0001h000<100000@0002P0
00<100002P000c80000I0003<P0001T000<b00006@000c80000J0003<P0000H000<1000040000c80
000N00030@000040000X00030@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80
000700030@0000l000<b00007P000`4000010000:0000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001X000<b00001`000`40000?0003<P0001h000<100000@0002P000<100002P000c80
000I0003<P0001T000<b00006@000c80000J0003<P0000P000<100003P000c80000N00030@000040
000X00030@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80000800030@0000h0
00<b00007P000`4000010000:0020@/000<b00006@000c80000I0003<P0001T000<b00006P000c80
000900030@0000d000<b00007@020@<0000X00030@0000X000<b00006@000c80000I0003<P0001T0
00<b00006P000c80000900030@0000d000<b00007P000`4000010000:0000`40000:0003<P0001T0
00<b00006@000c80000I0003<P0001X000<b00002P020@d000<b00007P000`4000010000:0000`40
000:0003<P0001T000<b00006@000c80000I0003<P0001X000<b000030000`40000:0003<P0001h0
00<100000@0002P000<100002P000c80000I0003<P0001T000<b00006@000c80000J0003<P0000d0
0P4:0003<P0001h000<100000@0002P000<100002P000c80000I0003<P0001T000<b00006@000c80
000J0003<P0000l000<100001`000c80000N00030@000040000S00030@00008000<100002P000c80
000I0003<P0001T000<b00006@000c80000J0003<P0001000P470003<P0001h000<100000@0002<0
00<100000P020I4b00<1<S808c830@<0000P00D10`000`40000:0003<P0001T000<b00006@000c80
000I0003<P0001X000<b00004`020@@000<b00007P000`40000100006`040@050040004010000`40
000:0003<P0001T000<b00006@000c80000I0003<P0001X000<b00005@001@40000b020000<10000
0@00024000<1004010000`40000:0003<P0001T000<b00006@000c80000I0003<P0001X000<b0000
5P001040038P00030@000040000R008110000`40000:0003<P0001T000<b00006@000c80000I0003
<P0001X000<b00005`020@03<P0001h000<100000@0002<000<100000P000`40000:0003<P0001T0
00<b00006@000c80000I0003<P0001X000<b00006@000`40000N00030@000040000X00030@0000X0
00<b00006@000c80000I0003<P0001T000<b00006P000c80000I0003<P4101h000<100000@0002P0
0P4;0003<P0001T000<b00006@000c80000I0003<P0001X000<b00006@001380004L00810`0002P0
00<100002P000c80000I0003<P0001T000<b00006@000c80000J0003<P0001T000@b00000P4K0003
0@000040000X00030@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80000I0003
<P0000<00`4H00030@000040000X00030@0000X000<b00006@000c80000I0003<P0001T000<b0000
6P000c80000I0003<P0000H01P4B00030@000040000X00030@0000X000<b00006@000c80000I0003
<P0001T000<b00006P000c80000I0003<P0000`01@4=00030@000040000X00030@0000X000<b0000
6@000c80000I0003<P0001T000<b00006P000c80000I0003<P0001401P4700030@000040000X0003
0@0000X000<b00006@000c80000I0003<P0001T000<b00006P000c80000I0003<P0001L00`440003
0@000040000X00812`000`40000I00030@0001T000<100006@000`40000J00030@0001T000<10000
6@00104000020@<0000X0;`10`000001\
\>"],
  ImageRangeCache->{{{101, 331}, {688.625, 546.938}} -> {-229.229, 26.1628, 
  0.708443, 0.0716783}}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox[
"Rational approximation and model reduction"]], "Section"],

Cell[CellGroupData[{

Cell[TextData[StyleBox[
"From data on the grid to rational functions"]], "Subsection"],

Cell["\<\
To find a stable rational functiona that corresponds to a closed \
loop function T generated by OPTDesign (and that satisfies the internal \
stability requirements,)  you may use\
\>", "Text"],

Cell[CellGroupData[{

Cell["Trat = RationalModel[s,DegreeOfDenominator -> 3]", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]0.115866709626662234`\),
      SequenceForm[ "Error = ", 0.11586670962666223],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(-5.`\) + s\)\/\((\(2.`\[InvisibleSpace]\) + s)\)\^2 + 
      \(\((\(-5.`\) + s)\)\ 
          \((\(1.50242647691855468`\[InvisibleSpace]\) + 
              1.21124609592095877`\ s)\)\)\/\(\((
            \(-1.21489196265210175`\) - 0.785108037347898246`\ s)\)\ 
          \((\(2.`\[InvisibleSpace]\) + s)\)\^2\)\)], "Output"]
}, Open  ]],

Cell["\<\
A general pupose Caratheodory-Fejer approximation by stable \
rationals is implemented as the function StableFit.
In addition to the function values, one must specify a  grid with a standard \
OPTDesign format as input (which you can generate outside of OPTDesign if you \
like)  and the degree of the denominator.  Here we use the list T1 that is a \
byproduct of an OPTDesign run:\
\>", "Text"],

Cell[CellGroupData[{

Cell["Dimensions[T1]", "Input"],

Cell[BoxData[
    \({32}\)], "Output"]
}, Open  ]],

Cell["wpts = Grid[];", "Input"],

Cell[CellGroupData[{

Cell["StableFit[T1,wpts,DegreeOfDenominator ->3]", "Input"],

Cell[BoxData[
    InterpretationBox[\("Error = "\[InvisibleSpace]0.00965834358065453812`\),
      SequenceForm[ "Error = ", 0.0096583435806545381],
      Editable->False]], "Print"],

Cell[BoxData[
    \(\(\(1.59462623924101665`\[InvisibleSpace]\) + 4.91306875041354196`\ s + 
        3.11098101269514781`\ s\^2 + 
        1.21405867639897646`\ s\^3\)\/\(\(-1.19052208185486652`\) - 
        3.9191370980044816`\ s - 2.06137360076802655`\ s\^2 - 
        0.828967219372625407`\ s\^3\)\)], "Output"]
}, Open  ]],

Cell["\<\
For more on rational approximation with a different algorithm see \
the NewtonFit notebook (Appendix G)\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["From a rational function to data on a grid", "Subsection"],

Cell["\<\
if you have a rational function Tr rather than a list of values, \
the following command  gives a list of values of Tr 
on the  OptDesign session grid.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Tdisc = Discretize[Trat]", "Input"],

Cell[BoxData[
    \({\(-5.35865436688346186`*^-18\), 
      \(-0.0456089453239055675`\) + 0.0339104632253719184`\ I, 
      \(-0.125632147963354201`\) - 0.0169765232314980085`\ I, 
      \(-0.14854326586993638`\) - 0.136930025664120869`\ I, 
      \(-0.0892608404879294248`\) - 0.252745765042546155`\ I, 
      \(0.0169444654017759921`\[InvisibleSpace]\) - 0.316023039052998111`\ I, 
      \(0.12637800168399349`\[InvisibleSpace]\) - 0.322324094055437493`\ I, 
      \(0.213755635380280484`\[InvisibleSpace]\) - 0.289864837630298133`\ I, 
      \(0.272190336468773885`\[InvisibleSpace]\) - 0.239305088140012056`\ I, 
      \(0.304968583276897797`\[InvisibleSpace]\) - 0.1854779493466493`\ I, 
      \(0.318878736037731513`\[InvisibleSpace]\) - 0.136634920933563552`\ I, 
      \(0.320733516145895602`\[InvisibleSpace]\) - 0.0962158386357098116`\ I, 
      \(0.316079112649184423`\[InvisibleSpace]\) - 0.0647562673012019907`\ I, 
      \(0.30897999469030446`\[InvisibleSpace]\) - 0.0412528275994110149`\ I, 
      \(0.302207248496346947`\[InvisibleSpace]\) - 0.0239817380709407856`\ I, 
      \(0.297510196711739958`\[InvisibleSpace]\) - 0.0109312695179981101`\ I, 
      \(0.295843707821112467`\[InvisibleSpace]\) + 0.`\ I, 
      \(0.297510196711739958`\[InvisibleSpace]\) + 0.0109312695179981101`\ I, 
      \(0.302207248496346947`\[InvisibleSpace]\) + 0.0239817380709407856`\ I, 
      \(0.30897999469030446`\[InvisibleSpace]\) + 0.0412528275994110149`\ I, 
      \(0.316079112649184423`\[InvisibleSpace]\) + 0.0647562673012019907`\ I, 
      \(0.320733516145895602`\[InvisibleSpace]\) + 0.0962158386357098116`\ I, 
      \(0.318878736037731513`\[InvisibleSpace]\) + 0.136634920933563552`\ I, 
      \(0.304968583276897797`\[InvisibleSpace]\) + 0.1854779493466493`\ I, 
      \(0.272190336468773885`\[InvisibleSpace]\) + 0.239305088140012056`\ I, 
      \(0.213755635380280484`\[InvisibleSpace]\) + 0.289864837630298133`\ I, 
      \(0.12637800168399349`\[InvisibleSpace]\) + 0.322324094055437493`\ I, 
      \(0.0169444654017759921`\[InvisibleSpace]\) + 0.316023039052998111`\ I, 
      \(-0.0892608404879294248`\) + 0.252745765042546155`\ I, 
      \(-0.14854326586993638`\) + 0.136930025664120869`\ I, 
      \(-0.125632147963354201`\) + 0.0169765232314980085`\ I, 
      \(-0.0456089453239055675`\) - 0.0339104632253719184`\ I}\)], "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1024}, {0, 768}},
WindowSize->{672, 654},
WindowMargins->{{Automatic, 155}, {5, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1},
Magnification->1
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 52, 0, 154, "Title"],
Cell[1786, 53, 74, 0, 27, "SmallText"],

Cell[CellGroupData[{
Cell[1885, 57, 242, 5, 96, "Subsubtitle"],

Cell[CellGroupData[{
Cell[2152, 66, 126, 2, 45, "Subsection"],
Cell[2281, 70, 164, 5, 32, "Text"],

Cell[CellGroupData[{
Cell[2470, 79, 559, 28, 417, "Input"],
Cell[3032, 109, 122, 2, 55, "Print"],
Cell[3157, 113, 115, 2, 39, "Print"],
Cell[3275, 117, 179, 3, 39, "Message"],
Cell[3457, 122, 185, 3, 39, "Message"],
Cell[3645, 127, 77, 1, 39, "Print"],
Cell[3725, 130, 79, 1, 23, "Print"],
Cell[3807, 133, 301, 9, 42, "Print"],
Cell[4111, 144, 301, 9, 42, "Print"],
Cell[4415, 155, 37, 1, 23, "Print"],
Cell[4455, 158, 69, 1, 23, "Print"],
Cell[4527, 161, 65, 1, 23, "Print"],
Cell[4595, 164, 66, 1, 23, "Print"],
Cell[4664, 167, 86, 1, 23, "Print"],
Cell[4753, 170, 36, 1, 23, "Print"],
Cell[4792, 173, 114, 2, 23, "Print"],
Cell[4909, 177, 109, 2, 23, "Print"],
Cell[5021, 181, 114, 2, 23, "Print"],
Cell[5138, 185, 112, 2, 23, "Print"],
Cell[5253, 189, 112, 2, 23, "Print"],
Cell[5368, 193, 36, 1, 23, "Print"],
Cell[5407, 196, 36, 1, 23, "Print"],
Cell[5446, 199, 42, 1, 23, "Print"],
Cell[5491, 202, 42, 1, 23, "Print"],
Cell[5536, 205, 68, 1, 23, "Print"],
Cell[5607, 208, 63, 1, 23, "Print"],
Cell[5673, 211, 63, 1, 23, "Print"],
Cell[5739, 214, 54, 1, 23, "Print"],
Cell[5796, 217, 67, 1, 23, "Print"],
Cell[5866, 220, 56, 1, 23, "Print"],
Cell[5925, 223, 41, 1, 23, "Print"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[6015, 230, 58, 0, 34, "Section"],

Cell[CellGroupData[{
Cell[6098, 234, 56, 0, 45, "Subsection"],
Cell[6157, 236, 489, 10, 86, "Text"],

Cell[CellGroupData[{
Cell[6671, 250, 19, 0, 27, "Input"],
Cell[6693, 252, 2309, 31, 171, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[9051, 289, 97, 3, 45, "Subsection"],
Cell[9151, 294, 146, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[9322, 301, 23, 0, 27, "Input"],
Cell[9348, 303, 1180, 19, 75, "Output"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[10577, 328, 77, 0, 45, "Subsection"],
Cell[10657, 330, 117, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[10799, 337, 44, 0, 27, "Input"],
Cell[10846, 339, 4220, 89, 267, "Output"]
}, Open  ]],
Cell[15081, 431, 56, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[15162, 435, 35, 0, 27, "Input"],
Cell[15200, 437, 41, 1, 27, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[15302, 445, 44, 0, 54, "Section"],

Cell[CellGroupData[{
Cell[15371, 449, 64, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[15460, 453, 90, 3, 27, "Input"],
Cell[15553, 458, 105809, 4399, 296, 74272, 4005, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[121377, 4860, 55, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[121457, 4864, 109, 3, 42, "Input"],
Cell[121569, 4869, 105909, 4440, 296, 74667, 4050, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[227527, 9315, 59, 0, 45, "Subsection"],
Cell[227589, 9317, 156, 3, 50, "Text"],

Cell[CellGroupData[{
Cell[227770, 9324, 34, 0, 27, "Input"],
Cell[227807, 9326, 28255, 1042, 186, 16350, 891, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[256077, 10371, 76, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[256178, 10375, 39, 0, 27, "Input"],
Cell[256220, 10377, 514, 13, 27, "Output"]
}, Open  ]],
Cell[256749, 10393, 133, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[256907, 10400, 90, 3, 27, "Input"],
Cell[257000, 10405, 32849, 1058, 186, 18553, 877, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[289898, 11469, 53, 0, 45, "Subsection"],
Cell[289954, 11471, 123, 3, 32, "Text"],

Cell[CellGroupData[{
Cell[290102, 11478, 75, 0, 27, "Input"],
Cell[290180, 11480, 19058, 708, 174, 11056, 605, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[309287, 12194, 58, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[309370, 12198, 28, 0, 27, "Input"],
Cell[309401, 12200, 9155, 333, 147, 3771, 262, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],
Cell[318571, 12536, 64, 0, 32, "Text"],

Cell[CellGroupData[{
Cell[318660, 12540, 47, 0, 27, "Input"],
Cell[318710, 12542, 12405, 375, 152, 3681, 263, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[331164, 12923, 48, 0, 45, "Subsection"],

Cell[CellGroupData[{
Cell[331237, 12927, 47, 0, 27, "Input"],
Cell[331287, 12929, 20902, 647, 151, 6770, 468, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[352250, 13583, 82, 1, 54, "Section"],

Cell[CellGroupData[{
Cell[352357, 13588, 86, 1, 45, "Subsection"],
Cell[352446, 13591, 202, 4, 50, "Text"],

Cell[CellGroupData[{
Cell[352673, 13599, 65, 0, 27, "Input"],
Cell[352741, 13601, 177, 3, 23, "Print"],
Cell[352921, 13606, 349, 6, 46, "Output"]
}, Open  ]],
Cell[353285, 13615, 406, 7, 86, "Text"],

Cell[CellGroupData[{
Cell[353716, 13626, 31, 0, 27, "Input"],
Cell[353750, 13628, 38, 1, 27, "Output"]
}, Open  ]],
Cell[353803, 13632, 31, 0, 27, "Input"],

Cell[CellGroupData[{
Cell[353859, 13636, 59, 0, 27, "Input"],
Cell[353921, 13638, 181, 3, 23, "Print"],
Cell[354105, 13643, 314, 5, 45, "Output"]
}, Open  ]],
Cell[354434, 13651, 127, 3, 32, "Text"]
}, Open  ]],

Cell[CellGroupData[{
Cell[354598, 13659, 64, 0, 45, "Subsection"],
Cell[354665, 13661, 175, 4, 50, "Text"],

Cell[CellGroupData[{
Cell[354865, 13669, 41, 0, 27, "Input"],
Cell[354909, 13671, 2348, 32, 189, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

